Package: fastml
Type: Package
Title: Guarded Resampling Workflows for Safe and Automated Machine
        Learning in R
Version: 0.7.9
Authors@R: c(person("Selcuk", "Korkmaz", email="selcukorkmaz@gmail.com", role=c("aut", "cre"),
                    comment = c(ORCID = "0000-0003-4632-6850")), 
             person("Dincer", "Goksuluk", email="dincer.goksuluk@gmail.com", role="aut",
                    comment = c(ORCID = "0000-0002-2752-7668")), 
             person("Eda", "Karaismailoglu", email="eda.karaismailoglu@sbu.edu.tr", role="aut",
                    comment = c(ORCID = "0000-0003-3085-7809")))
Description: Provides a guarded resampling workflow for training and evaluating machine-learning models. 
	When the guarded resampling path is used, preprocessing and model fitting are re-estimated within 
	each resampling split to reduce leakage risk. Supports multiple resampling schemes, integrates 
	with established engines in the 'tidymodels' ecosystem, and aims to improve evaluation reliability by 
	coordinating preprocessing, fitting, and evaluation within supported workflows. Offers a lightweight 
	AutoML-style workflow by automating model training, resampling, and tuning across multiple algorithms, 
	while keeping evaluation design explicit and user-controlled.
Encoding: UTF-8
License: MIT + file LICENSE
URL: https://selcukorkmaz.github.io/fastml-tutorial/,
        https://github.com/selcukorkmaz/fastml
BugReports: https://github.com/selcukorkmaz/fastml/issues
Depends: R (>= 4.1.0)
Imports: stats, recipes, dplyr, ggplot2, reshape2, rsample, parsnip,
        tune, workflows, yardstick, tibble, rlang, dials, RColorBrewer,
        doFuture, foreach, finetune, future, viridisLite, magrittr,
        pROC, janitor, stringr, broom, tidyr, purrr, survival, xgboost
Suggests: testthat (>= 3.0.0), withr, baguette, C50, DALEX, discrim,
        fairmodels, flexsurv, iBreakDown, iml, lime, modelStudio, pdp,
        plsmod, probably, ranger, rstpm2, survRM2, aorsf, censored,
        crayon, kernlab, klaR, kknn, keras, lightgbm, rstanarm,
        patchwork, GGally, glmnet, themis, DT, UpSetR, VIM, dbscan,
        gridExtra, htmlwidgets, kableExtra, moments, naniar, plotly,
        scales, skimr, sparsediscrim, knitr, rmarkdown, pec
RoxygenNote: 7.3.3
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-08-19 11:34:52 UTC; selcukkorkmaz
Author: Selcuk Korkmaz [aut, cre] (ORCID:
    <https://orcid.org/0000-0003-4632-6850>),
  Dincer Goksuluk [aut] (ORCID: <https://orcid.org/0000-0002-2752-7668>),
  Eda Karaismailoglu [aut] (ORCID:
    <https://orcid.org/0000-0003-3085-7809>)
Maintainer: Selcuk Korkmaz <selcukorkmaz@gmail.com>
Repository: CRAN
Date/Publication: 2026-08-19 12:50:02 UTC

# Generated by roxygen2: do not edit by hand

S3method(plot,fastml)
S3method(plot,fastml_stability)
S3method(predict,fastml)
S3method(predict_risk,default)
S3method(predict_risk,fastml_native_survival)
S3method(predict_risk,workflow)
S3method(predict_survival,default)
S3method(predict_survival,fastml_native_survival)
S3method(predict_survival,workflow)
S3method(print,fastml_stability)
S3method(summary,fastml)
export(availableMethods)
export(compare_defaults)
export(counterfactual_explain)
export(estimate_tuning_time)
export(explain_ale)
export(explain_dalex)
export(explain_lime)
export(explain_stability)
export(fastexplain)
export(fastexplore)
export(fastml)
export(fastml_compute_holdout_results)
export(flatten_and_rename_models)
export(get_best_model_idx)
export(get_best_model_names)
export(get_best_workflows)
export(get_default_differences)
export(get_default_engine)
export(get_default_params)
export(get_default_params_with_warnings)
export(get_default_tune_params)
export(get_engine_names)
export(get_expanded_tune_params)
export(get_model_engine_names)
export(get_parsnip_default_engine)
export(get_parsnip_default_params)
export(get_tuning_complexity)
export(get_tuning_params_for_complexity)
export(interaction_strength)
export(load_model)
export(plot_ice)
export(predict_model.model_fit)
export(predict_risk)
export(predict_survival)
export(print_default_differences)
export(print_tuning_presets)
export(process_model)
export(recommend_tuning_config)
export(reset_default_warnings)
export(sanitize)
export(save.fastml)
export(save_fastml)
export(surrogate_tree)
export(train_models)
export(validate_defaults_registry)
export(warn_default_override)
importFrom(RColorBrewer,brewer.pal)
importFrom(broom,glance)
importFrom(broom,tidy)
importFrom(dials,finalize)
importFrom(dials,grid_latin_hypercube)
importFrom(dials,grid_regular)
importFrom(dials,range_set)
importFrom(dials,value_set)
importFrom(dplyr,across)
importFrom(dplyr,all_of)
importFrom(dplyr,arrange)
importFrom(dplyr,bind_cols)
importFrom(dplyr,bind_rows)
importFrom(dplyr,distinct)
importFrom(dplyr,filter)
importFrom(dplyr,group_by)
importFrom(dplyr,if_else)
importFrom(dplyr,mutate)
importFrom(dplyr,n)
importFrom(dplyr,n_distinct)
importFrom(dplyr,pull)
importFrom(dplyr,rename)
importFrom(dplyr,rename_with)
importFrom(dplyr,row_number)
importFrom(dplyr,select)
importFrom(dplyr,starts_with)
importFrom(dplyr,summarise)
importFrom(dplyr,ungroup)
importFrom(dplyr,where)
importFrom(finetune,control_race)
importFrom(finetune,tune_race_anova)
importFrom(ggplot2,aes)
importFrom(ggplot2,autoplot)
importFrom(ggplot2,coord_equal)
importFrom(ggplot2,element_blank)
importFrom(ggplot2,element_text)
importFrom(ggplot2,facet_wrap)
importFrom(ggplot2,geom_abline)
importFrom(ggplot2,geom_bar)
importFrom(ggplot2,geom_col)
importFrom(ggplot2,geom_errorbarh)
importFrom(ggplot2,geom_histogram)
importFrom(ggplot2,geom_line)
importFrom(ggplot2,geom_path)
importFrom(ggplot2,geom_point)
importFrom(ggplot2,geom_rug)
importFrom(ggplot2,ggplot)
importFrom(ggplot2,labs)
importFrom(ggplot2,position_dodge)
importFrom(ggplot2,scale_color_manual)
importFrom(ggplot2,theme)
importFrom(ggplot2,theme_bw)
importFrom(ggplot2,theme_minimal)
importFrom(ggplot2,ylim)
importFrom(graphics,plot)
importFrom(janitor,make_clean_names)
importFrom(magrittr,"%>%")
importFrom(pROC,auc)
importFrom(pROC,multiclass.roc)
importFrom(pROC,roc)
importFrom(parsnip,bag_tree)
importFrom(parsnip,boost_tree)
importFrom(parsnip,decision_tree)
importFrom(parsnip,extract_parameter_set_dials)
importFrom(parsnip,fit)
importFrom(parsnip,linear_reg)
importFrom(parsnip,logistic_reg)
importFrom(parsnip,mlp)
importFrom(parsnip,multinom_reg)
importFrom(parsnip,naive_Bayes)
importFrom(parsnip,nearest_neighbor)
importFrom(parsnip,pls)
importFrom(parsnip,predict.model_fit)
importFrom(parsnip,proportional_hazards)
importFrom(parsnip,rand_forest)
importFrom(parsnip,set_engine)
importFrom(parsnip,set_mode)
importFrom(parsnip,survival_reg)
importFrom(parsnip,svm_linear)
importFrom(parsnip,svm_rbf)
importFrom(purrr,flatten)
importFrom(recipes,all_nominal_predictors)
importFrom(recipes,all_numeric_predictors)
importFrom(recipes,all_outcomes)
importFrom(recipes,all_predictors)
importFrom(recipes,bake)
importFrom(recipes,prep)
importFrom(recipes,recipe)
importFrom(recipes,step_center)
importFrom(recipes,step_dummy)
importFrom(recipes,step_impute_bag)
importFrom(recipes,step_impute_knn)
importFrom(recipes,step_impute_median)
importFrom(recipes,step_naomit)
importFrom(recipes,step_novel)
importFrom(recipes,step_rm)
importFrom(recipes,step_scale)
importFrom(recipes,step_unknown)
importFrom(recipes,step_zv)
importFrom(reshape2,dcast)
importFrom(reshape2,melt)
importFrom(rlang,.data)
importFrom(rlang,as_string)
importFrom(rlang,call2)
importFrom(rlang,call_modify)
importFrom(rlang,expr)
importFrom(rlang,get_env)
importFrom(rlang,get_expr)
importFrom(rlang,sym)
importFrom(rlang,syms)
importFrom(rsample,analysis)
importFrom(rsample,assessment)
importFrom(rsample,bootstraps)
importFrom(rsample,group_vfold_cv)
importFrom(rsample,initial_split)
importFrom(rsample,nested_cv)
importFrom(rsample,rolling_origin)
importFrom(rsample,testing)
importFrom(rsample,training)
importFrom(rsample,validation_split)
importFrom(rsample,vfold_cv)
importFrom(stats,AIC)
importFrom(stats,BIC)
importFrom(stats,aggregate)
importFrom(stats,approx)
importFrom(stats,approxfun)
importFrom(stats,as.formula)
importFrom(stats,ave)
importFrom(stats,coef)
importFrom(stats,complete.cases)
importFrom(stats,delete.response)
importFrom(stats,formula)
importFrom(stats,integrate)
importFrom(stats,logLik)
importFrom(stats,median)
importFrom(stats,model.frame)
importFrom(stats,model.matrix)
importFrom(stats,model.offset)
importFrom(stats,na.omit)
importFrom(stats,na.pass)
importFrom(stats,nobs)
importFrom(stats,pnorm)
importFrom(stats,predict)
importFrom(stats,qnorm)
importFrom(stats,quantile)
importFrom(stats,sd)
importFrom(stats,setNames)
importFrom(stringr,str_detect)
importFrom(survival,Surv)
importFrom(survival,basehaz)
importFrom(survival,coxph)
importFrom(survival,psurvreg)
importFrom(survival,strata)
importFrom(survival,survfit)
importFrom(survival,untangle.specials)
importFrom(tibble,as_tibble)
importFrom(tibble,is_tibble)
importFrom(tibble,tibble)
importFrom(tidyr,pivot_wider)
importFrom(tune,control_bayes)
importFrom(tune,control_grid)
importFrom(tune,extract_fit_parsnip)
importFrom(tune,finalize_model)
importFrom(tune,finalize_workflow)
importFrom(tune,select_best)
importFrom(tune,tune)
importFrom(tune,tune_bayes)
importFrom(tune,tune_grid)
importFrom(utils,capture.output)
importFrom(utils,head)
importFrom(utils,tail)
importFrom(viridisLite,viridis)
importFrom(workflows,add_model)
importFrom(workflows,add_recipe)
importFrom(workflows,extract_fit_parsnip)
importFrom(workflows,extract_recipe)
importFrom(workflows,pull_workflow_preprocessor)
importFrom(workflows,pull_workflow_spec)
importFrom(workflows,workflow)
importFrom(yardstick,accuracy)
importFrom(yardstick,conf_mat)
importFrom(yardstick,f_meas)
importFrom(yardstick,kap)
importFrom(yardstick,mae)
importFrom(yardstick,metric_set)
importFrom(yardstick,new_class_metric)
importFrom(yardstick,precision)
importFrom(yardstick,rmse)
importFrom(yardstick,roc_auc)
importFrom(yardstick,rsq)
importFrom(yardstick,sens)
importFrom(yardstick,spec)

                FastML - program for computing maximum likelihood
                         ancestral sequence reconstruction

The FastML program is a bioinformatics tool for the reconstruction of ancestral sequences based on the phylogenetic relations between homologous sequences. 
The program runs several algorithms that reconstruct the ancestral sequences with emphasis on an accurate reconstruction of both indels and characters. 

URL: http://fastml.tau.ac.il/

Authors: Haim Ashkenazy, Osnat Penn, Adi Doron-Faigenboim, Ofir Cohen, Gina Cannarozzi, Oren Zomer and Tal Pupko

When using the FastML algorithm please cite:
[1] Ashkenazy H, Penn O, Doron-Faigenboim A, Cohen O, Cannarozzi G, Zomer O, Pupko T. 2012
FastML: a web server for probabilistic reconstruction of ancestral sequences 
Nucleic Acids Res. 40(Web Server issue):W580-4.

[2] Pupko T, Pe'er I, Hasegawa M, Graur D, Friedman N. 2002
A branch-and-bound algorithm for the inference of ancestral amino-acid sequences when the replacement rate varies among sites: Application to the evolution of five gene families. 
Bioinformatics 18(8): 1116-1123. [pdf] [abs] 

[3] Pupko T, Pe'er I, Shamir R, Graur D. 2000.
A fast algorithm for joint reconstruction of ancestral amino-acid sequences.
Mol. Biol. Evol. 17(6): 890-896. [pdf] [abs] 

[4] Pupko, T. and Pe'er I. 2000.
Maximum likelihood reconstruction of ancestral amino-acid sequences.
Currents in Computational Molecular Biology. Ed. Miyano, S., Shamir, R, and Takagi, T. pp. 184-185. Universal Academy Press, Tokyo, Japan. [pdf]

Installation
============

  1. Unpack the archive by typing:
        % tar -xzf FastML.v3.1.tgz

  2. Compile the package by typing:
        % cd FastML.v3.1
        % make
     (Running `make' takes a while)

  3A. FastML uses Perl:
      Type "perl -v" and check that Perl is installed.
      If it's not installed, download and install it from: http://www.perl.org/

	3B. To reconstruct the ML tree during FastML run RAxML and BioPerl should be installed in your system.

	   RAxML: Type "which raxmlHPC" and check that the program is found
   	 If it's not installed, download and install RAxML from:  http://sco.h-its.org/exelixis/web/software/raxml/index.html
	   BioPerl: Type "perl -e 'use Bio::SeqIO'" to check that BioPerl is installed.
     If it's not installed, download and install it from: http://www.bioperl.org/
         

Usage
=====

Run the Perl script: FastML.v3.1/www/fastml/FastML_Wrapper.pl
(Note that you cannot move this script from of its directory, because it uses relative paths to other files in other directories. Sorry)
FastML uses flags in the command line arguments: (for help, type: "perl FastML_Wrapper.pl")

USAGE: perl FastML_Wrapper.pl --MSA_File MSA_File --seqType [AA|NUC|CODON] --outDir OUTDIR

Required parameters:
  --MSA_File     Input multiple sequence alignment in FASTA format
  --seqType      Sequence type may be either of: nuc (nucleotides), aa (amino acids),
                 or codon (nucleotides that will be treated as whole codons)
  --outDir       FULL PATH of the output directory where all output files will be created
                 (NOTE: EACH RUN must have its UNIQUE outDir.
                  In case the outDir does not exists it will be created automatically)
Optional parameters:
  --Tree <phylogenetic tree>
  --TreeAlg <NJ | RAxML> - How to reconstruct the tree when a tree is not provided by the user; default=NJ
  --SubMatrix <JTT | LG | mtREV | cpREV | WAG | DAYHOFF > amino acid options, the default is JTT.
              <JC_Nuc | T92 | HKY | GTR> nucleotide options, the default is JC_Nuc.
              <yang | empiriCodon> codon options, the default is yang.
  --OptimizeBL <yes | no> default: yes
  --UseGamma   <yes | no> default: yes
  --Alpha      <User provide alpha>   (relevant only when UseGamma==yes)
                                       user alpha parameter of the gamma distribution [if alpha is not given, alpha and branches will be evaluated from the data]
  --jointReconstruction <yes | no> default: yes
  --indelReconstruction <PARSIMONY|ML|BOTH> - which method is used for indel reconstruction
  --indelCutOff <Cutoff for indel vs Char> deafult = 0.5

EXAMPLE:
> perl FastML.v3.1/www/fastml/FastML_Wrapper.pl --MSA_File MSA.aln --outDir /home/MSA.FastML --seqType aa --Tree tree.newick    
Will reconstruct ancestral sequences (both "joint" and "marginal") based on the proteins MSA in "MSA.aln" and the tree in "tree,newick" and output all results to the diretory "MSA.FastML" at the home directory

Copyrights
==========
    * To modify the code, or use parts of it for other purposes, permission should be requested. Please contact Tal Pupko: talp@post.tau.ac.il
    * Please note that the use of the FastML program is for academic use only

<img src="man/figures/fastml_hex.png" align="right" width="85"/>

# fastml: Guarded Resampling Workflows for Safe and Automated Machine Learning in R

<!-- badges: start -->
[![CRAN status](https://www.r-pkg.org/badges/version/fastml)](https://CRAN.R-project.org/package=fastml)
[![CRAN downloads](https://cranlogs.r-pkg.org/badges/grand-total/fastml)](https://CRAN.R-project.org/package=fastml)
[![CRAN monthly](https://cranlogs.r-pkg.org/badges/fastml)](https://CRAN.R-project.org/package=fastml)
[![License: MIT](https://img.shields.io/badge/License-MIT-blue.svg)](https://opensource.org/licenses/MIT)
[![Lifecycle: stable](https://img.shields.io/badge/lifecycle-stable-brightgreen.svg)](https://lifecycle.r-lib.org/articles/stages.html#stable)
[![R](https://img.shields.io/badge/R%20%3E%3D-4.1.0-blue.svg)](https://cran.r-project.org/)
<!-- badges: end -->

**Guarded Resampling Workflows for Safe and Automated Machine Learning in R**

fastml is an R package for training, evaluating, and comparing machine learning models with a guarded resampling workflow. Rather than introducing new learning algorithms, fastml focuses on **reducing leakage risk** by keeping preprocessing, model fitting, and evaluation aligned within supported resampling paths.

In fastml, *fast* refers to the rapid construction of statistically valid workflows, not to computational shortcuts. By eliminating entire classes of user-induced errors — most notably preprocessing leakage — fastml allows practitioners to obtain reliable performance estimates with minimal configuration.

## Core Principles

- **Guarded resampling workflow** — When the guarded resampling path is used, preprocessing and model fitting are re-estimated independently within each resampling split. This reduces leakage risk, but does not prevent users from supplying preprocessed inputs.

- **Leakage risk reduction, not guarantees** — fastml can mitigate common leakage modes (e.g., global scaling or imputation before resampling) when workflows are fit within resamples. It does not universally prevent all leakage scenarios.

- **Single, unified interface** — Multiple models can be trained and benchmarked through a single call, while internally coordinating resampling, preprocessing, and evaluation in supported paths.

- **Compatibility with established engines** — fastml orchestrates existing modeling infrastructure (recipes, rsample, parsnip, yardstick) without modifying their statistical behavior.

## Installation

### From CRAN

```r
install.packages("fastml")
```

### From GitHub (development version)

```r
# install.packages("devtools")
devtools::install_github("selcukorkmaz/fastml")
```

### Optional Dependencies

fastml uses a lightweight core. Feature-specific packages are installed only when needed:

```r
# Explainability (SHAP, LIME, ALE, ICE, etc.)
install.packages(c("DALEX", "iml", "lime", "iBreakDown", "pdp"))

# Additional model engines
install.packages(c("ranger", "glmnet", "kernlab", "kknn", "lightgbm", "C50"))

# Discriminant analysis and bagged trees
install.packages(c("discrim", "baguette"))

# Survival analysis extensions
install.packages(c("censored", "flexsurv", "rstpm2", "survRM2", "aorsf"))

# Fairness and interactive dashboards
install.packages(c("fairmodels", "modelStudio"))
```

## Quick Start

```r
library(fastml)
library(dplyr)

data(iris)

iris_binary <- iris %>%
  filter(Species != "setosa") %>%
  mutate(Species = factor(Species))

# Train and evaluate models in one call
fit <- fastml(
  data = iris_binary,
  label = "Species",
  algorithms = c("rand_forest", "logistic_reg")
)

# Inspect results
summary(fit)

# Visualize
plot(fit, type = "bar")
plot(fit, type = "roc")
plot(fit, type = "calibration")

# Predict on new data
predict(fit, newdata = iris_binary[1:5, ])
```

## Supported Tasks

fastml auto-detects the task type based on the target variable:

| Task | Detection | Metrics |
|------|-----------|---------|
| **Classification** | Factor or character target (binary/multiclass) | Accuracy, ROC AUC, Sensitivity, Specificity, Precision, F1, Kappa, Logloss |
| **Regression** | Numeric target | RMSE, R-squared, MAE |
| **Survival** | Two-column label (time + status) | C-index, Integrated Brier Score, time-dependent Brier, RMST |

## Supported Algorithms

### Classification

| Algorithm | Default Engine | Alternative Engines |
|-----------|---------------|---------------------|
| `logistic_reg` | glm | glmnet, brulee, stan, keras, LiblineaR |
| `multinom_reg` | nnet | glmnet, brulee, keras |
| `rand_forest` | ranger | randomForest, h2o, partykit |
| `xgboost` | xgboost | — |
| `lightgbm` | lightgbm | — |
| `decision_tree` | rpart | C5.0, partykit |
| `bag_tree` | rpart | — |
| `svm_rbf` | kernlab | — |
| `svm_linear` | kernlab | LiblineaR |
| `nearest_neighbor` | kknn | — |
| `naive_Bayes` | klaR | naivebayes, h2o |
| `mlp` | nnet | brulee, keras, h2o |
| `discrim_linear` | MASS | — |
| `discrim_quad` | sparsediscrim | — |

### Regression

| Algorithm | Default Engine | Alternative Engines |
|-----------|---------------|---------------------|
| `linear_reg` | lm | — |
| `ridge_reg` | glmnet | — |
| `lasso_reg` | glmnet | — |
| `elastic_net` | glmnet | — |
| `rand_forest` | ranger | randomForest, h2o |
| `xgboost` | xgboost | — |
| `lightgbm` | lightgbm | — |
| `decision_tree` | rpart | — |
| `svm_rbf` | kernlab | — |
| `svm_linear` | kernlab | — |
| `nearest_neighbor` | kknn | — |
| `mlp` | nnet | brulee, keras, h2o |
| `pls` | mixOmics | — |
| `bayes_glm` | stan | — |

### Survival

| Algorithm | Default Engine |
|-----------|---------------|
| `cox_ph` | survival |
| `penalized_cox` | glmnet |
| `stratified_cox` | survival |
| `time_varying_cox` | survival |
| `survreg` | survival |
| `rand_forest` | aorsf / ranger |
| `xgboost` / `xgboost_aft` | xgboost |
| `parametric_surv` | flexsurv |
| `piecewise_exp` | flexsurv |
| `royston_parmar` | rstpm2 |

## Resampling Methods

```r
fit <- fastml(
  data = df,
  label = "target",
  resampling_method = "cv",  # default
  cv_folds = 10
)
```

| Method | Description |
|--------|-------------|
| `cv` | K-fold cross-validation (default) |
| `repeatedcv` | Repeated cross-validation |
| `boot` | Bootstrap resampling |
| `grouped_cv` | Grouped cross-validation (keeps groups intact) |
| `blocked_cv` | Blocked/time-series CV (respects temporal ordering) |
| `rolling_origin` | Rolling window resampling |
| `nested_cv` | Nested cross-validation (unbiased tuning) |
| `validation_split` | Simple train/validation split |
| `none` | No resampling (single holdout) |

## Hyperparameter Tuning

```r
fit <- fastml(
  data = iris_binary,
  label = "Species",
  algorithms = c("rand_forest", "xgboost"),
  tuning_strategy = "bayes",
  tuning_complexity = "balanced",
  tuning_iterations = 25
)
```

| Strategy | Description |
|----------|-------------|
| `grid` | Grid search (default) |
| `bayes` | Bayesian optimization |
| `none` | No tuning, use defaults |

**Tuning complexity presets** control search breadth:

| Preset | Grid Levels | Use Case |
|--------|------------|----------|
| `quick` | 2 | Prototyping, debugging |
| `balanced` | 3 | Most production use (default) |
| `thorough` | 5 | Final model selection, publications |
| `exhaustive` | 7 | Research, competitions |

## Preprocessing

fastml isolates preprocessing inside the resampling loop to prevent leakage:

```r
fit <- fastml(
  data = df,
  label = "target",
  impute_method = "knnImpute",   # medianImpute, bagImpute, remove, error
  scale = c("center", "scale"),
  balance = "upsample"           # downsample, or none (default)
)
```

Transformations applied per fold: imputation, dummy encoding, centering, scaling, zero-variance removal, novel/unknown level handling.

## Explainability

Model explainability is provided through `fastexplain()` with 10 methods:

```r
# DALEX-based: variable importance, SHAP values, partial dependence
fastexplain(fit, method = "dalex")

# LIME local explanations
fastexplain(fit, method = "lime", observation = df[1, ])

# Accumulated Local Effects
fastexplain(fit, method = "ale", features = "Sepal.Length")

# Individual Conditional Expectation curves
fastexplain(fit, method = "ice", features = "Sepal.Length")

# Surrogate decision tree
fastexplain(fit, method = "surrogate")

# Feature interaction strength
fastexplain(fit, method = "interaction")

# iBreakDown contributions
fastexplain(fit, method = "breakdown", observation = df[1, ])

# Counterfactual explanations
fastexplain(fit, method = "counterfactual", observation = df[1, ])

# Interactive modelStudio dashboard
fastexplain(fit, method = "studio")

# Fairness diagnostics
fastexplain(fit, method = "fairness", protected = df$gender)
```

### Feature Importance Stability

Analyze how feature importance varies across cross-validation folds:

```r
fit <- fastml(data = df, label = "target", store_fold_models = TRUE)
stability <- explain_stability(fit)
print(stability)
plot(stability)
```

## Exploratory Data Analysis

`fastexplore()` provides read-only diagnostics prior to model training:

```r
fastexplore(iris, label = "Species")
```

Generates: summary statistics, distribution plots (histograms, boxplots, bar charts), correlation heatmaps, Q-Q plots, missingness analysis, and more — without invoking any modeling.

## Visualization

```r
# Performance comparison
plot(fit, type = "bar")

# ROC curves (classification)
plot(fit, type = "roc")

# Calibration plot (classification)
plot(fit, type = "calibration")

# Residual diagnostics (regression)
plot(fit, type = "residual")

# All plots at once
plot(fit, type = "all")
```

## Parallel Processing

```r
fit <- fastml(
  data = df,
  label = "target",
  algorithms = c("rand_forest", "xgboost", "svm_rbf"),
  n_cores = 4
)
```

## Model Persistence

```r
# Save
save.fastml(fit, path = "my_model.rds")

# Load
fit <- load_model("my_model.rds")
```

## Survival Analysis

```r
library(survival)
data(lung)

fit_surv <- fastml(
  data = lung,
  label = c("time", "status"),
  algorithms = c("cox_ph", "rand_forest"),
  eval_times = c(180, 365, 730)
)

summary(fit_surv)

# Predict survival probabilities
predict_survival(fit_surv, newdata = lung[1:5, ], eval_times = c(365, 730))

# Predict risk scores
predict_risk(fit_surv, newdata = lung[1:5, ])
```

## Advanced Options

```r
fit <- fastml(
  data = df,
  label = "target",

  # Algorithms and engines
  algorithms = c("rand_forest", "xgboost"),
  algorithm_engines = list(rand_forest = "ranger", xgboost = "xgboost"),
  engine_params = list(xgboost = list(nthread = 4)),

  # Resampling
  resampling_method = "cv",
  cv_folds = 10,
  store_fold_models = TRUE,

  # Tuning
  tuning_strategy = "bayes",
  tuning_complexity = "thorough",
  tuning_iterations = 30,

  # Classification-specific
  class_threshold = "auto",          # auto-tune threshold
  multiclass_auc = "macro_weighted", # prevalence-weighted AUC

  # Bootstrap confidence intervals
  bootstrap_ci = TRUE,
  bootstrap_samples = 500,

  # Reproducibility
  seed = 42
)
```

## Scope

fastml is intended for users who require reliable performance estimation under cross-validation, particularly in:

- Multi-site or grouped data
- High-dimensional biomedical applications
- Workflows prone to preprocessing leakage

It prioritizes correctness-oriented defaults and workflow clarity over maximum flexibility.

## Links

- [Tutorial & Documentation](https://selcukorkmaz.github.io/fastml-tutorial/)
- [GitHub Repository](https://github.com/selcukorkmaz/fastml)
- [Bug Reports](https://github.com/selcukorkmaz/fastml/issues)
- [CRAN Page](https://CRAN.R-project.org/package=fastml)

## License

MIT License.  

#' Get Available Methods
#'
#' Returns a character vector of algorithm names available for classification,
#' regression or survival tasks.
#'
#' @param type A character string specifying the type of task. Must be one of
#'   \code{"classification"}, \code{"regression"}, or \code{"survival"}. Defaults to
#'   \code{c("classification", "regression", "survival")} and uses
#'   \code{\link[base]{match.arg}} to select one.
#' @param ... Additional arguments (currently not used).
#'
#' @return A character vector containing the names of the available algorithms for the specified task type.
#'
#' @details Depending on the specified \code{type}, the function returns a different set of algorithm names:
#' \itemize{
#'   \item For \code{"classification"}, it returns algorithms such as \code{"logistic_reg"}, \code{"multinom_reg"}, \code{"decision_tree"}, \code{"C5_rules"}, \code{"rand_forest"}, \code{"xgboost"}, \code{"lightgbm"}, \code{"svm_linear"}, \code{"svm_rbf"}, \code{"nearest_neighbor"}, \code{"naive_Bayes"}, \code{"mlp"}, \code{"discrim_linear"}, \code{"discrim_quad"}, and \code{"bag_tree"}.
#'   \item For \code{"regression"}, it returns algorithms such as \code{"linear_reg"}, \code{"ridge_reg"}, \code{"lasso_reg"}, \code{"elastic_net"}, \code{"decision_tree"}, \code{"rand_forest"}, \code{"xgboost"}, \code{"lightgbm"}, \code{"svm_linear"}, \code{"svm_rbf"}, \code{"nearest_neighbor"}, \code{"mlp"}, \code{"pls"}, and \code{"bayes_glm"}.
#'   \item For \code{"survival"}, it returns algorithms such as \code{"rand_forest"}, \code{"cox_ph"}, \code{"penalized_cox"}, \code{"stratified_cox"}, \code{"time_varying_cox"}, \code{"survreg"}, \code{"royston_parmar"}, \code{"parametric_surv"}, \code{"piecewise_exp"}, and \code{"xgboost"}.
#' }
#' 
#' @export
availableMethods <- function(type = c("classification", "regression", "survival"), ...){
    type <- match.arg(type)

    algorithms <- if (type == "classification"){
      c(
        "logistic_reg",
        "multinom_reg",
        "decision_tree",
        "C5_rules",
        "rand_forest",
        "xgboost",
        "lightgbm",
        "svm_linear",
        "svm_rbf",
        "nearest_neighbor",
        "naive_Bayes",
        "mlp",
        "discrim_linear",
        "discrim_quad",
        "bag_tree"
      )
    } else if (type == "regression") {
      c(
        "linear_reg",
        "ridge_reg",
        "lasso_reg",
        "elastic_net",
        "decision_tree",
        "rand_forest",
        "xgboost",
        "lightgbm",
        "svm_linear",
        "svm_rbf",
        "nearest_neighbor",
        "mlp",
        "pls",
        "bayes_glm"
      )
    } else {
      c(
        "rand_forest",
        "rand_forest_survival",
        "cox_ph",
        "penalized_cox",
        "stratified_cox",
        "time_varying_cox",
        "survreg",
        "royston_parmar",
        "parametric_surv",
        "piecewise_exp",
        "xgboost",
        "xgboost_aft"
      )
    }

    return(algorithms)
  }

#' Get Default Engine
#'
#' Returns the default engine corresponding to the specified algorithm.
#'
#' @param algo A character string specifying the name of the algorithm. The value should match one of the supported algorithm names.
#' @param task Optional task type (e.g., \code{"classification"}, \code{"regression"}, or \code{"survival"}). Used to determine defaults that depend on the task.
#'
#' @return A character string containing the default engine name associated with the provided algorithm.
#'
#' @details The function uses a \code{switch} statement to select the default engine based on the given algorithm. For survival random forests, the function defaults to \code{"aorsf"}. If the provided algorithm does not have a defined default engine, the function terminates with an error.
#'
#'
#' @export
get_default_engine <- function(algo, task = NULL) {
  if (algo == "rand_forest" && !is.null(task) && task == "survival") {
    return("aorsf")
  }
  if (algo == "rand_forest_survival" && !is.null(task) && task == "survival") {
    return("aorsf")
  }
  if (algo == "cox_ph" && !is.null(task) && task == "survival") {
    return("survival")
  }
  if (algo == "penalized_cox" && !is.null(task) && task == "survival") {
    return("glmnet")
  }
  if (algo == "stratified_cox" && !is.null(task) && task == "survival") {
    return("survival")
  }
  if (algo == "time_varying_cox" && !is.null(task) && task == "survival") {
    return("survival")
  }
  if (algo == "survreg" && !is.null(task) && task == "survival") {
    return("survival")
  }
  if (algo == "royston_parmar" && !is.null(task) && task == "survival") {
    return("rstpm2")
  }
  if (algo == "parametric_surv" && !is.null(task) && task == "survival") {
    return("flexsurvreg")
  }
  if (algo == "piecewise_exp" && !is.null(task) && task == "survival") {
    return("flexsurvreg")
  }
  if (algo == "aft" && !is.null(task) && task == "survival") {
    return("survival")
  }
  if (algo == "xgboost" && !is.null(task) && task == "survival") {
    return("aft")
  }
  if (algo == "xgboost_aft" && !is.null(task) && task == "survival") {
    return("aft")
  }

  switch(algo,
         "lightgbm" = "lightgbm",
         "xgboost" = "xgboost",
         "C5_rules" = "C5.0",
         "rand_forest" = "ranger",
         "ranger" = "ranger",
         "logistic_reg" = "glm",
         "multinom_reg" = "nnet",
         "decision_tree" = "rpart",
         "svm_linear" = "kernlab",
         "svm_rbf" = "kernlab",
         "nearest_neighbor" = "kknn",
         "naive_Bayes" = "klaR",
         "mlp" = "nnet",
         "discrim_linear" = "MASS",
         "discrim_quad" = "sparsediscrim",
         "bag_tree" = "rpart",
         "elastic_net" = "glmnet",
         "bayes_glm" = "stan",
         "pls" = "mixOmics",
         "linear_reg" = "lm",
         "ridge_reg" = "glmnet",
         "lasso_reg" = "glmnet",
         "stratified_cox" = "survival",
         "time_varying_cox" = "survival",
         "survreg" = "survival",
         "royston_parmar" = "rstpm2",
         "deep_learning" = "keras",
         "cox_ph" = "survival",
         stop("No default engine defined for algorithm: ", algo)
  )
}

resolve_engine_params <- function(engine_params, algo, engine) {
  if (is.null(engine_params)) {
    return(list())
  }
  if (!is.list(engine_params)) {
    stop("'engine_params' must be a list.")
  }

  algo_params <- engine_params[[algo]]
  if (is.null(algo_params)) {
    return(list())
  }
  if (!is.list(algo_params)) {
    stop(sprintf("Engine parameters for algorithm '%s' must be provided as a list.", algo))
  }

  engine_name <- if (length(engine) == 0 || is.na(engine)) "" else as.character(engine)

  candidate <- NULL
  if (!is.null(names(algo_params)) && engine_name %in% names(algo_params)) {
    candidate <- algo_params[[engine_name]]
  } else if (length(algo_params) > 0 && (is.null(names(algo_params)) || any(names(algo_params) == ""))) {
    candidate <- algo_params
  }

  if (is.null(candidate)) {
    return(list())
  }
  if (!is.list(candidate)) {
    detail <- if (nzchar(engine_name)) sprintf(" and engine '%s'", engine_name) else ""
    stop(sprintf("Engine parameters for algorithm '%s'%s must be provided as a list.", algo, detail))
  }
  if (length(candidate) > 0 && (is.null(names(candidate)) || any(names(candidate) == ""))) {
    detail <- if (nzchar(engine_name)) sprintf(" and engine '%s'", engine_name) else ""
    stop(sprintf("Engine parameters for algorithm '%s'%s must be a named list.", algo, detail))
  }

  candidate
}

merge_engine_args <- function(base = list(), overrides = list()) {
  if (is.null(base)) {
    base <- list()
  }
  if (is.null(overrides)) {
    overrides <- list()
  }
  if (!is.list(base)) {
    stop("'base' must be a list when merging engine arguments.")
  }
  if (!is.list(overrides)) {
    stop("'overrides' must be a list when merging engine arguments.")
  }
  if (length(overrides) == 0) {
    return(base)
  }
  if (is.null(names(overrides)) || any(names(overrides) == "")) {
    stop("Engine parameter overrides must be a named list.")
  }
  if (length(base) == 0) {
    return(overrides)
  }
  utils::modifyList(base, overrides, keep.null = TRUE)
}

call_with_engine_params <- function(fun, base_args, engine_args) {
  if (is.null(engine_args)) {
    engine_args <- list()
  }
  if (!is.list(base_args)) {
    stop("'base_args' must be provided as a list.")
  }
  if (!is.list(engine_args)) {
    stop("Engine parameters must be provided as a list.")
  }
  combined <- merge_engine_args(base_args, engine_args)
  do.call(fun, combined)
}

fastml_normalize_seed <- function(seed) {
  if (is.null(seed) || length(seed) == 0 || is.na(seed[[1]])) {
    return(NULL)
  }
  seed <- as.integer(seed[[1]])
  if (!is.finite(seed)) {
    return(NULL)
  }
  seed
}

#' Merge engine arguments onto a specification's existing ones
#'
#' `parsnip::set_engine()` replaces any engine arguments already attached to a
#' model specification rather than adding to them. When arguments are applied in
#' a second pass, the defaults a spec builder established would therefore be
#' lost. This helper evaluates the existing quosures and overlays `new_args` on
#' top, so spec defaults survive unless explicitly overridden.
#'
#' @param existing The `eng_args` element of a `model_spec` (a list of quosures),
#'   or `NULL`.
#' @param new_args Named list of engine arguments to apply.
#'
#' @return A named list suitable for `do.call(parsnip::set_engine, ...)`.
#' @keywords internal
#' @noRd
fastml_merge_engine_args <- function(existing, new_args) {
  if (is.null(new_args)) {
    new_args <- list()
  }
  if (is.null(existing) || length(existing) == 0) {
    return(new_args)
  }
  existing_vals <- list()
  for (nm in names(existing)) {
    if (!nzchar(nm)) next
    val <- tryCatch(rlang::eval_tidy(existing[[nm]]), error = function(e) NULL)
    if (!is.null(val)) {
      existing_vals[[nm]] <- val
    }
  }
  if (length(existing_vals) == 0) {
    return(new_args)
  }
  utils::modifyList(existing_vals, new_args)
}

fastml_normalize_threads <- function(n_cores) {
  if (is.null(n_cores) || length(n_cores) == 0 || is.na(n_cores[[1]])) {
    return(NULL)
  }
  n_cores <- as.integer(n_cores[[1]])
  if (!is.finite(n_cores) || n_cores < 1L) {
    return(NULL)
  }
  n_cores
}

fastml_engine_arg_value <- function(engine_args, name) {
  if (is.null(engine_args) || length(engine_args) == 0) {
    return(NULL)
  }
  if (!is.null(engine_args[[name]])) {
    return(engine_args[[name]])
  }
  params <- engine_args$params
  if (!is.null(params) && !is.null(params[[name]])) {
    return(params[[name]])
  }
  NULL
}

fastml_apply_engine_seed <- function(engine_args,
                                     algo,
                                     engine,
                                     seed,
                                     n_cores,
                                     task = NULL) {
  if (is.null(engine_args)) {
    engine_args <- list()
  }
  seed_val <- fastml_normalize_seed(seed)
  threads_val <- fastml_normalize_threads(n_cores)
  if (is.null(seed_val) && is.null(threads_val)) {
    return(engine_args)
  }

  engine_key <- tolower(as.character(engine))
  add_if_missing <- function(name, value) {
    if (is.null(engine_args[[name]])) {
      engine_args[[name]] <<- value
    }
  }

  if (engine_key %in% c("ranger")) {
    if (!is.null(seed_val)) {
      add_if_missing("seed", seed_val)
    }
    if (!is.null(threads_val)) {
      add_if_missing("num.threads", threads_val)
    }
  } else if (engine_key == "lightgbm") {
    if (!is.null(seed_val)) {
      add_if_missing("seed", seed_val)
    }
    add_if_missing("deterministic", TRUE)
    if (!is.null(threads_val)) {
      add_if_missing("num_threads", threads_val)
    }
  } else if (engine_key == "xgboost") {
    if (!is.null(threads_val)) {
      add_if_missing("nthread", threads_val)
    }
  } else if (engine_key == "h2o") {
    if (!is.null(seed_val)) {
      add_if_missing("seed", seed_val)
    }
  } else if (engine_key == "spark") {
    if (!is.null(seed_val)) {
      add_if_missing("seed", seed_val)
    }
  }

  engine_args
}

fastml_engine_determinism_warning <- function(algo,
                                              engine,
                                              task,
                                              n_cores,
                                              engine_args) {
  engine_key <- tolower(as.character(engine))
  if (engine_key %in% c("keras")) {
    return("Keras/TensorFlow uses nondeterministic ops by default; configure TensorFlow for determinism to reproduce results.")
  }
  if (engine_key %in% c("h2o")) {
    return("H2O engines can be nondeterministic across runs even with a seed due to distributed execution.")
  }
  if (engine_key %in% c("spark")) {
    return("Spark MLlib training can be nondeterministic across executors/partitions even with a seed.")
  }
  if (engine_key %in% c("aorsf")) {
    seed_val <- fastml_engine_arg_value(engine_args, "seed")
    if (is.null(seed_val)) {
      return("aorsf randomness is not controlled unless a fixed seed is supplied via engine_params.")
    }
  }

  threads_val <- fastml_engine_arg_value(engine_args, "num_threads")
  if (is.null(threads_val)) {
    threads_val <- fastml_engine_arg_value(engine_args, "nthread")
  }
  if (is.null(threads_val)) {
    threads_val <- fastml_normalize_threads(n_cores)
  }
  if (!is.null(threads_val)) {
    threads_val <- as.integer(threads_val[[1]])
    if (!is.finite(threads_val)) {
      threads_val <- NULL
    } else if (threads_val <= 0L) {
      threads_val <- 2L
    }
  }
  if (engine_key %in% c("lightgbm") && !is.null(threads_val) && threads_val > 1L) {
    return("LightGBM can be nondeterministic with multithreading; set num_threads = 1 for strict reproducibility.")
  }
  if (engine_key %in% c("xgboost") && !is.null(threads_val) && threads_val > 1L) {
    return("xgboost can be nondeterministic with multithreading; set nthread = 1 for strict reproducibility.")
  }

  NULL
}

fastml_emit_determinism_warnings <- function(entries) {
  if (length(entries) == 0) {
    return(invisible(NULL))
  }
  lines <- vapply(
    entries,
    function(entry) {
      sprintf("%s (%s): %s", entry$algo, entry$engine, entry$reason)
    },
    character(1)
  )
  warning(
    paste(
      "Determinism warnings for selected engines:",
      paste(lines, collapse = "\n"),
      sep = "\n"
    ),
    call. = FALSE
  )
  invisible(NULL)
}


#' Get Engine Names from Model Workflows
#'
#' Extracts and returns a list of unique engine names from a list of model workflows.
#'
#' @param models A list where each element is a list of model workflows. Each workflow is expected to contain a fitted model that can be processed with \code{tune::extract_fit_parsnip}.
#'
#' @return A list of character vectors. Each vector contains the unique engine names extracted from the corresponding element of \code{models}.
#'
#' @details The function applies \code{tune::extract_fit_parsnip} to each model workflow to extract the fitted model object. It then retrieves the engine name from the model specification (\code{spec$engine}). If the extraction fails, \code{NA_character_} is returned for that workflow. Finally, the function removes any duplicate engine names using \code{unique}.
#'
#' @importFrom tune extract_fit_parsnip
#'
#' @export
get_engine_names <- function(models) {
  extract_engine <- function(mod) {
    if (inherits(mod, "fastml_native_survival")) {
      eng <- mod$engine
      if (is.null(eng) || is.na(eng)) {
        return(NA_character_)
      }
      return(as.character(eng))
    }
    fit_obj <- tryCatch(
      extract_fit_parsnip(mod),
      error = function(e) NULL
    )
    if (!is.null(fit_obj) && !is.null(fit_obj$spec$engine)) {
      return(as.character(fit_obj$spec$engine))
    }
    NA_character_
  }

  lapply(models, function(model_entry) {
    if (inherits(model_entry, "fastml_native_survival")) {
      eng <- extract_engine(model_entry)
      if (is.na(eng)) NA_character_ else eng
    } else if (inherits(model_entry, "workflow") || inherits(model_entry, "model_fit")) {
      eng <- extract_engine(model_entry)
      if (is.na(eng)) NA_character_ else eng
    } else if (is.list(model_entry)) {
      if (length(model_entry) == 0) {
        return(NA_character_)
      }
      engines <- vapply(model_entry, extract_engine, character(1))
      engines <- unique(engines[!is.na(engines) & engines != ""])
      if (length(engines) == 0) {
        NA_character_
      } else {
        engines
      }
    } else {
      NA_character_
    }
  })
}

#' Get Model Engine Names
#'
#' Extracts and returns a named vector mapping algorithm names to engine names from a nested list of model workflows.
#'
#' @param models A nested list of model workflows. Each inner list should contain model objects from which a fitted model can be extracted using \code{tune::extract_fit_parsnip}.
#'
#' @return A named character vector where the names correspond to algorithm names (e.g., \code{"rand_forest"}, \code{"logistic_reg"}) and the values correspond to the associated engine names (e.g., \code{"ranger"}, \code{"glm"}).
#'
#' @details The function iterates over a nested list of model workflows and, for each workflow, attempts to extract the fitted model object using \code{tune::extract_fit_parsnip}. If successful, it retrieves the algorithm name from the first element of the class attribute of the model specification and the engine name from the specification. The results are combined into a named vector.
#'
#' @importFrom tune extract_fit_parsnip
#' @importFrom stats setNames
#'
#' @export
get_model_engine_names <- function(models) {
  result <- c()
  for(model_list in models) {
    for(mod in model_list) {
      fit_obj <- tryCatch(
        extract_fit_parsnip(mod),
        error = function(e) NULL
      )
      if (!is.null(fit_obj)) {
        algo   <- class(fit_obj$spec)[1]   # e.g., "rand_forest" or "logistic_reg"
        engine <- fit_obj$spec$engine       # e.g., "ranger", "randomForest", "glm"
        result <- c(result, setNames(engine, algo))
      }
    }
  }
  result
}

library(dplyr)
library(testthat)
library(survival)

data(iris)
iris <- iris[iris$Species != "setosa", ]  # Binary classification
iris$Species <- factor(iris$Species)


data(cancer)
test_that("fastml errors when data contains NAs and impute_method = 'error'", {
  expect_error(
    fastml(
      data = cancer,
      label = c("time", "status"),
      algorithms = "rand_forest",
      task = "survival",
      test_size = 0.3
    ),
    "Data contains NAs"
  )
})


test_that("rand_forest defaults to aorsf engine for survival", {
  expect_identical(get_default_engine("rand_forest", "survival"), "aorsf")
})

test_that("bootstrap CI columns are returned for classification and regression", {
  skip_on_cran()
  # Reduce separation risk in glm for stable bootstrap fits.
  iris_bootstrap <- iris[, c("Sepal.Length", "Sepal.Width", "Species")]

  fit_class <- fastml(
    data = iris_bootstrap,
    label = "Species",
    algorithms = "logistic_reg",
    task = "classification",
    resampling_method = "none",
    use_default_tuning = FALSE,
    test_size = 0.3,
    seed = 123,
    bootstrap_ci = TRUE,
    bootstrap_samples = 10,
    bootstrap_seed = 123
  )

  perf_class <- fit_class$performance[[1]]
  expect_true(all(c(".lower", ".upper", ".n_boot") %in% names(perf_class)))
  expect_true(all(perf_class$.n_boot == 10))
  expect_true(any(is.finite(perf_class$.lower)))
  expect_true(any(is.finite(perf_class$.upper)))

  fit_reg <- fastml(
    data = mtcars,
    label = "mpg",
    algorithms = "linear_reg",
    task = "regression",
    resampling_method = "none",
    use_default_tuning = FALSE,
    test_size = 0.3,
    seed = 123,
    bootstrap_ci = TRUE,
    bootstrap_samples = 10,
    bootstrap_seed = 123
  )

  perf_reg <- fit_reg$performance[[1]]
  expect_true(all(c(".lower", ".upper", ".n_boot") %in% names(perf_reg)))
  expect_true(all(perf_reg$.n_boot == 10))
  expect_true(any(is.finite(perf_reg$.lower)))
  expect_true(any(is.finite(perf_reg$.upper)))
})



test_that("advanced imputation options are rejected", {
  expect_error(
    fastml(
      data = iris,
      label = "Species",
      algorithms = "decision_tree",
      impute_method = "mice",
      resampling_method = "none",
      use_default_tuning = FALSE
    ),
    "no longer supported",
    ignore.case = TRUE
  )

  expect_error(
    fastml(
      data = iris,
      label = "Species",
      algorithms = "decision_tree",
      impute_method = "missForest",
      resampling_method = "none",
      use_default_tuning = FALSE
    ),
    "no longer supported",
    ignore.case = TRUE
  )

  expect_error(
    fastml(
      data = iris,
      label = "Species",
      algorithms = "decision_tree",
      impute_method = "custom",
      resampling_method = "none",
      use_default_tuning = FALSE
    ),
    "no longer supported",
    ignore.case = TRUE
  )
})

test_that("multinom_reg falls back to logistic_reg for binary outcomes", {
  skip_on_cran()
  binary_iris <- iris[, c("Sepal.Length", "Sepal.Width", "Species")]

  expect_warning(
    fastml(
      data = binary_iris,
      label = "Species",
      algorithms = "multinom_reg",
      resampling_method = "none",
      use_default_tuning = FALSE,
      seed = 123
    ) -> fit,
    "not applicable to two-class outcomes"
  )

  expect_s3_class(fit, "fastml")
  expect_true(any(startsWith(names(fit$models), "logistic_reg")))
  expect_false(any(startsWith(names(fit$models), "multinom_reg")))
})

test_that("grouped_cv requires grouping columns", {
  binary_iris <- iris[iris$Species != "virginica", ]
  binary_iris$Species <- droplevels(binary_iris$Species)

  expect_error(
    fastml(
      data = binary_iris,
      label = "Species",
      algorithms = c("logistic_reg"),
      resampling_method = "grouped_cv",
      folds = 3,
      use_default_tuning = FALSE
    ),
    "group_cols",
    ignore.case = TRUE
  )
})


test_that("blocked_cv enforces ordered data", {
  set.seed(99)
  blocked_df <- data.frame(
    time_id = sample(seq_len(60)),
    x1 = rnorm(60),
    response = factor(sample(c("yes", "no"), 60, replace = TRUE))
  )

  expect_error(
    fastml(
      data = blocked_df,
      label = "response",
      algorithms = c("logistic_reg"),
      resampling_method = "blocked_cv",
      block_col = "time_id",
      block_size = 5,
      folds = 4,
      use_default_tuning = FALSE
    ),
    "sorted",
    ignore.case = TRUE
  )
})


test_that("rolling_origin validates configuration", {
  rolling_df <- data.frame(
    time_id = seq_len(30),
    x = rnorm(30),
    y = rnorm(30)
  )

  expect_error(
    fastml(
      data = rolling_df,
      label = "y",
      algorithms = c("linear_reg"),
      resampling_method = "rolling_origin",
      block_col = "time_id",
      assess_window = 5,
      use_default_tuning = FALSE
    ),
    "initial_window",
    ignore.case = TRUE
  )
})



test_that("'label' is not available in the data", {
  expect_error({
    # Train models with Bayesian optimization
    model <- fastml(
      data = iris,
      label = "Unknown",
      algorithms = c("rand_forest")
    )
  })
})



test_that("model fails if reponse variable is not of supported type", {
  tmp <- iris %>%
    mutate(
      Date = as.Date("2024-12-19")
    )

  expect_error({
    # Train models with Bayesian optimization
    model <- fastml(
      data = iris,
      label = "Date",
      algorithms = c("rand_forest")
    )
  })
})


test_that("stop if requested metric is not allowed.", {
  expect_error({
    # Train models with Bayesian optimization
    model <- fastml(
      data = iris,
      label = "Species",
      algorithms = c("rand_forest"),
      metric = "unknown"
    )
  })

  expect_error({
    # Example 2: Using the mtcars dataset for regression
    # Train models
    model <- fastml(
      data = mtcars,
      label = "mpg",
      algorithms = c("rand_forest"),
      metric = "unknown"
    )
  })
})

test_that("check for supported algorithms", {
  expect_warning({
    # Train models with Bayesian optimization
    fastml(
      data = iris,
      label = "Species",
      algorithms = c("rand_forest", "unknown")
    )
  })

  expect_error({
    # Train models with Bayesian optimization
    fastml(
      data = iris,
      label = "Species",
      algorithms = c("unknown")
    )
  })
})

test_that("variables successfuly excluded", {
  expect_error({
    fastml(
      data = iris,
      label = "Species",
      algorithms = c("rand_forest"),
      exclude = "Species"
    )
  })

  expect_warning({
    fastml(
      data = iris,
      label = "Species",
      algorithms = c("rand_forest"),
      exclude = c("Sepal.Length", "unknown")
    )
  })
})

test_that("checks for impute_method", {
  skip_on_cran()
  expect_warning(
    fastml(
      data = iris,
      label = "Species",
      algorithms = c("rand_forest"),
      impute_method = "medianImpute"
    ),
    "Missing values in numeric predictors are being imputed using the median."
  )

  expect_warning(
    fastml(
      data = iris,
      label = "Species",
      algorithms = c("rand_forest"),
      impute_method = "knnImpute"
    ),
    "Missing values are being imputed using KNN"
  )

  expect_warning(
    fastml(
      data = iris,
      label = "Species",
      algorithms = c("rand_forest"),
      impute_method = "bagImpute"
    ),
    "Missing values are being imputed using bagging"
  )

  expect_warning(
    fastml(
      data = iris,
      label = "Species",
      algorithms = c("rand_forest"),
      impute_method = "remove"
    ),
    "Rows with missing values in predictors are being removed."
  )

  expect_error(
    fastml(
      data = iris,
      label = "Species",
      algorithms = c("rand_forest"),
      impute_method = "unknown"
    ),
    "Invalid impute_method specified."
  )
})

test_that("engine_params list does not break ranger training", {
  skip_on_cran()
  skip_if_not_installed("ranger")

  rf_recipe <- recipes::recipe(Species ~ ., data = iris)

  expect_error(
    train_models(
      train_data = iris,
      label = "Species",
      task = "classification",
      algorithms = "rand_forest",
      resampling_method = "none",
      folds = 3,
      repeats = NA,
      resamples = NULL,
      tune_params = NULL,
      engine_params = list(rand_forest = list(ranger = list(importance = "impurity"))),
      metric = "accuracy",
      summaryFunction = NULL,
      seed = 123,
      recipe = rf_recipe,
      use_default_tuning = FALSE,
      tuning_strategy = "none",
      tuning_iterations = 10,
      early_stopping = FALSE,
      adaptive = FALSE,
      algorithm_engines = list(rand_forest = "ranger"),
      event_class = "first"
    ),
    regexp = NA
  )
})


test_that("engine_params configure Cox model ties", {
  skip_on_cran()
  skip_if_not_installed("survival")

  data(cancer, package = "survival")
  cancer$status <- ifelse(cancer$status == 2, 1, 0)
  cancer <- na.omit(cancer)  # remove NAs to prevent imputation error

  cox_result <- suppressWarnings(
    fastml(
      data = cancer,
      label = c("time", "status"),
      algorithms = "cox_ph",
      task = "survival",
      test_size = 0.3,
      resampling_method = "none",
      tune_params = NULL,
      engine_params = list(
        cox_ph = list(survival = list(ties = "breslow"))
      ),
      use_default_tuning = FALSE,
      tuning_strategy = "none",
      bootstrap_ci = FALSE
    )
  )

  expect_s3_class(cox_result$models$cox_ph$fit, "coxph")
  expect_identical(cox_result$models$cox_ph$fit$method, "breslow")
})


test_that("stop if recipe is not correctly specified.", {
  expect_error({
    fastml(
      data = iris,
      label = "Species",
      algorithms = c("rand_forest"),
      recipe = "unknown"
    )
  })
})


test_that("process_model works without global variables", {
  rec <- recipes::recipe(Species ~ ., data = iris)
  spec <- parsnip::logistic_reg() %>% parsnip::set_engine("glm")
  wf <- workflows::workflow() %>%
    workflows::add_model(spec) %>%
    workflows::add_recipe(rec)
  fitted_wf <- parsnip::fit(wf, data = iris)

  res <- process_model(fitted_wf, model_id = "log_reg", task = "classification",
                       test_data = iris, label = "Species",
                       event_class = "second", engine = "glm",
                       train_data = iris, metric = "accuracy")
  expect_s3_class(res$performance, "tbl_df")
  expect_equal(nrow(res$predictions), nrow(iris))
})

test_that("regression model successful.", {
  skip_on_cran()
  res <- fastml(
    data = iris[, -5],
    label = "Sepal.Length",
    algorithms = c("linear_reg")
  )
  expect_s3_class(res, "fastml")
})



test_that("multicore tasks successful.", {
  # CRAN coverage of the parallel path is provided by test-parallel-setup.R,
  # whose structural tests catch the same defects in ~4s; this end-to-end run
  # costs ~20s, almost all of it multisession worker startup, so it is gated.
  skip_on_cran()
  skip_if_not_installed("future")
  skip_if_not_installed("doFuture")

  res <- fastml(
    data = iris[, -5],
    label = "Sepal.Length",
    algorithms = c("linear_reg"),
    n_cores = 2
  )
  expect_s3_class(res, "fastml")
})

test_that("stop if unsupported metric is selected.", {
  skip_on_cran()
  expect_error({
    fastml(
      data = iris[,-5],
      label = "Sepal.Length",
      algorithms = c("linear_reg"),
      n_cores = 2,
      metric = "unkown"
    )
  })

  expect_no_error({
    fastml(
      data = iris[,-5],
      label = "Sepal.Length",
      algorithms = c("linear_reg"),
      n_cores = 2,
      metric = "rmse"
    )
  })
})

test_that("invalid tuning_strategy triggers error", {
  expect_error(
    fastml(
      data = iris,
      label = "Species",
      algorithms = c("rand_forest"),
      tuning_strategy = "invalid"
    ),
    "should be one of"
  )
})

test_that("grid tuning executes successfully", {
  skip_on_cran()
  suppressWarnings({
    res <- fastml(
      data = iris,
      label = "Species",
      algorithms = c("logistic_reg"),
      use_default_tuning = TRUE,
      tuning_strategy = "grid",
      resampling_method = "cv",
      folds = 5
    )
  })
  expect_s3_class(res, "fastml")
  expect_true(length(res$models) > 0)
})

test_that("Bayesian tuning executes successfully", {
  skip_on_cran()
  suppressWarnings({
    res <- fastml(
      data = iris,
      label = "Species",
      algorithms = c("logistic_reg"),
      use_default_tuning = TRUE,
      tuning_strategy = "bayes",
      tuning_iterations = 2,
      resampling_method = "cv",
      folds = 5
    )
  })
  expect_s3_class(res, "fastml")
  expect_true(length(res$models) > 0)
})

test_that("adaptive tuning executes successfully", {
  skip_on_cran()
  skip_if_not_installed("lme4")

  suppressWarnings({
    res <- fastml(
      data = iris,
      label = "Species",
      algorithms = c("logistic_reg"),
      use_default_tuning = TRUE,
      tuning_strategy = "grid",
      adaptive = TRUE,
      resampling_method = "cv",
      folds = 5
    )
  })
  expect_s3_class(res, "fastml")
})

test_that("early_stopping does not warn with grid tuning", {
  skip_on_cran()
  suppressWarnings({
    res <- fastml(
      data = iris,
      label = "Species",
      algorithms = c("logistic_reg"),
      use_default_tuning = TRUE,
      tuning_strategy = "grid",
      early_stopping = TRUE,
      resampling_method = "cv",
      folds = 5
    )
  })
  expect_s3_class(res, "fastml")
})

test_that("early stopping with bayesian tuning works", {
  skip_on_cran()
  suppressWarnings({
    res <- fastml(
      data = iris,
      label = "Species",
      algorithms = c("logistic_reg"),
      use_default_tuning = TRUE,
      tuning_strategy = "bayes",
      tuning_iterations = 2,
      early_stopping = TRUE,
      resampling_method = "cv",
      folds = 5
    )
  })
  expect_s3_class(res, "fastml")
})

test_that("invalid tuning_iterations triggers error", {
  expect_error(
    fastml(
      data = iris,
      label = "Species",
      algorithms = c("rand_forest"),
      tuning_strategy = "bayes",
      tuning_iterations = 0,
      resampling_method = "cv",
      folds = 5
    ),
    "tuning_iterations"
  )
})

test_that("tuning_iterations ignored for non-bayesian strategies", {
  skip_on_cran()
  suppressWarnings({
    expect_error(
      fastml(
        data = iris,
        label = "Species",
        algorithms = "logistic_reg",
        use_default_tuning = TRUE,
        tuning_strategy = "grid",
        tuning_iterations = 0,
        resampling_method = "cv",
        folds = 5
      ),
      regexp = NA
    )

    expect_error(
      fastml(
        data = iris,
        label = "Species",
        algorithms = "logistic_reg",
        tuning_strategy = "none",
        tuning_iterations = -1,
        resampling_method = "none"
      ),
      regexp = NA
    )
  })
})


test_that("adaptive ignored with bayesian tuning", {
  skip_on_cran()
  # Capture all warnings and check that one matches "adaptive"
  warnings_caught <- character(0)
  withCallingHandlers(
    fastml(
      data = iris,
      label = "Species",
      algorithms = c("logistic_reg"),
      use_default_tuning = TRUE,
      tuning_strategy = "bayes",
      adaptive = TRUE,
      tuning_iterations = 1,
      resampling_method = "cv",
      folds = 5
    ),
    warning = function(w) {
      warnings_caught <<- c(warnings_caught, conditionMessage(w))
      invokeRestart("muffleWarning")
    }
  )
  expect_true(any(grepl("adaptive", warnings_caught, ignore.case = TRUE)))
})

test_that("fold-unsafe preprocessing is blocked during resampling", {
  skip_on_cran()
  skip_if_not_installed("rsample")

  set.seed(101)
  # Use fresh iris data (not the modified global one) for binary classification
  fresh_iris <- datasets::iris[datasets::iris$Species != "setosa", ]
  fresh_iris$Species <- factor(fresh_iris$Species)

  idx <- sample(seq_len(nrow(fresh_iris)), size = floor(0.7 * nrow(fresh_iris)))
  train_split <- fresh_iris[idx, , drop = FALSE]
  test_split <- fresh_iris[-idx, , drop = FALSE]

  guard_resamples <- rsample::apparent(train_split)

  # The guard triggers a training failure, which results in "No models were

  # successfully trained" error. We verify the guard message appears in warnings.
  expect_warning(
    expect_error(
      fastml(
        train_data = train_split,
        test_data = test_split,
        label = "Species",
        algorithms = "logistic_reg",
        resamples = guard_resamples,
        resampling_method = "cv",
        folds = 5,
        seed = 202
      ),
      "No models were successfully trained"
    ),
    "Detected preprocessing on the full training set"
  )
})

test_that("guarded folds avoid performance inflation", {
  skip_on_cran()
  set.seed(2024)
  noise_df <- data.frame(
    x1 = rnorm(80),
    x2 = rnorm(80),
    x3 = rnorm(80),
    y = factor(sample(c("A", "B"), 80, replace = TRUE))
  )

  res_noise <- fastml(
    data = noise_df,
    label = "y",
    algorithms = "logistic_reg",
    resampling_method = "cv",
    folds = 4,
    seed = 77
  )

  guard_metrics <- res_noise$resampling_results[["logistic_reg (glm)"]]$aggregated
  guard_folds <- res_noise$resampling_results[["logistic_reg (glm)"]]$folds
  noise_accuracy <- guard_metrics[guard_metrics$.metric == "accuracy", ".estimate", drop = TRUE]

  expect_true(is.numeric(noise_accuracy))
  expect_lt(noise_accuracy, 0.8)
  expect_false(any(c(".lower", ".upper", ".n_boot") %in% names(guard_folds)))
})

test_that("guarded resampling is reproducible with fixed seeds", {
  skip_on_cran()
  set.seed(404)
  synth_df <- data.frame(
    x1 = rnorm(60),
    x2 = rnorm(60),
    y = factor(sample(c("yes", "no"), 60, replace = TRUE))
  )

  fit_one <- fastml(
    data = synth_df,
    label = "y",
    algorithms = "logistic_reg",
    resampling_method = "cv",
    folds = 3,
    seed = 55
  )

  set.seed(404)
  fit_two <- fastml(
    data = synth_df,
    label = "y",
    algorithms = "logistic_reg",
    resampling_method = "cv",
    folds = 3,
    seed = 55
  )

  expect_equal(
    fit_one$resampling_results[["logistic_reg (glm)"]]$aggregated,
    fit_two$resampling_results[["logistic_reg (glm)"]]$aggregated
  )
})

test_that("warning when tune_params ignored with no tuning", {
  skip_on_cran()
  tune <- list(rand_forest = list(ranger = list(mtry = c(1, 2))))
  expect_warning(
    fastml(
      data = iris,
      label = "Species",
      algorithms = c("rand_forest"),
      tune_params = tune,
      tuning_strategy = "none",
      use_default_tuning = TRUE,
      resampling_method = "none"
    ),
    "tune_params"
  )
})

library(testthat)

# Regression tests for fastml_merge_engine_args().
# parsnip::set_engine() replaces engine arguments rather than adding to them, so
# applying pipeline arguments in a second pass used to discard the defaults a spec
# builder had set (LightGBM's counts/bagging_freq/verbose, xgboost's early-stopping
# configuration, sparsediscrim's regularization_method).

test_that("spec defaults survive a second application of engine args", {
  existing <- list(counts = rlang::quo(TRUE), verbose = rlang::quo(-1L))
  merged <- fastml:::fastml_merge_engine_args(
    existing, list(num_threads = 4L, seed = 2025L)
  )
  expect_setequal(names(merged), c("counts", "verbose", "num_threads", "seed"))
  expect_true(merged$counts)
  expect_identical(merged$verbose, -1L)
})

test_that("pipeline args override spec defaults on shared keys", {
  existing <- list(num_threads = rlang::quo(0L), seed = rlang::quo(123L))
  merged <- fastml:::fastml_merge_engine_args(
    existing, list(num_threads = 4L, seed = 2025L)
  )
  expect_identical(merged$num_threads, 4L)
  expect_identical(merged$seed, 2025L)
})

test_that("empty or absent inputs are handled", {
  expect_identical(
    fastml:::fastml_merge_engine_args(NULL, list(a = 1)), list(a = 1)
  )
  expect_identical(
    fastml:::fastml_merge_engine_args(list(), list(a = 1)), list(a = 1)
  )
  expect_length(fastml:::fastml_merge_engine_args(NULL, NULL), 0)
})

test_that("quosures that cannot be evaluated are dropped, not fatal", {
  existing <- list(good = rlang::quo(1L), bad = rlang::quo(stop("boom")))
  merged <- fastml:::fastml_merge_engine_args(existing, list(z = 2L))
  expect_identical(merged$good, 1L)
  expect_false("bad" %in% names(merged))
  expect_identical(merged$z, 2L)
})

library(testthat)
library(survival)

test_that("available survival methods include core algorithms", {
  methods <- availableMethods("survival")
  expect_true(all(c("rand_forest", "cox_ph", "penalized_cox", "xgboost") %in% methods))
})

test_that("get_default_engine works for survival algorithms", {
  expect_identical(get_default_engine("rand_forest", task = "survival"), "aorsf")
  expect_identical(get_default_engine("cox_ph", task = "survival"), "survival")
  expect_identical(get_default_engine("penalized_cox", task = "survival"), "glmnet")
  expect_identical(get_default_engine("xgboost", task = "survival"), "aft")
})

test_that("xgboost AFT survival model trains and evaluates", {
  skip_on_cran()
  skip_if_not_installed("xgboost")

  data(heart, package = "survival")
  set.seed(123)

  res <- suppressWarnings(
    fastml(
      data = heart,
      label = c("stop", "event"),
      algorithms = "xgboost",
      task = "survival",
      resampling_method = "none",
      test_size = 0.3,
      impute_method = "remove",
      algorithm_engines = list(xgboost = "aft")
    )
  )

  expect_s3_class(res, "fastml")
  expect_identical(res$engine_names$xgboost, "aft")

  perf <- res$performance[[1]]
  expect_true(all(c("c_index", "uno_c", "ibs", "rmst_diff") %in% perf$.metric))

  preds <- res$predictions[[1]]
  expect_true("surv_prob_curve" %in% names(preds))
  expect_true(is.numeric(attr(preds$surv_prob_curve, "eval_times")))
  expect_true("surv_time" %in% names(preds))
  expect_true(any(is.finite(preds$surv_time)))
})


test_that("cox_ph survival model trains and evaluates", {
  skip_on_cran()
  data(cancer, package = "survival")
  cancer$status <- cancer$status - 1  # Recode from 1/2 to 0/1

  res <- suppressWarnings(
    fastml(
      data = cancer,
      label = c("time", "status"),
      algorithms = "cox_ph",
      task = "survival",
      resampling_method = "none",
      impute_method = "remove",
      test_size = 0.3
    )
  )

  expect_s3_class(res, "fastml")

  # Check metrics
  perf <- res$performance[[1]]
  expect_true(all(c("c_index", "uno_c", "ibs", "rmst_diff") %in% perf$.metric))
  expect_true(all(c(".lower", ".upper", ".n_boot") %in% names(perf)))

  # Brier-related metrics
  brier_metrics <- perf$.metric[grepl("^brier_t", perf$.metric)]
  expect_true(length(brier_metrics) >= 1)
  brier_val <- perf$.estimate[perf$.metric %in% brier_metrics]
  expect_true(length(brier_val) > 0)
  expect_true(all(is.finite(brier_val)))
  expect_true(all(brier_val >= 0 & brier_val <= 1))

  ibs_val <- perf$.estimate[perf$.metric == "ibs"]
  expect_true(all(is.finite(ibs_val)))
  expect_true(all(ibs_val >= 0 & ibs_val <= 1))
  expect_true(length(res$survival_brier_times) >= 1)
  expect_true(all(names(res$survival_brier_times) %in% brier_metrics))
  expect_identical(res$engine_names$cox_ph, "survival")

  # Summary should run cleanly
  expect_no_error(capture.output(summary(res)))

  # Metrics summary without CI
  metrics_no_ci <- capture.output(summary(res, type = "metrics", show_ci = FALSE))
  expect_false(any(grepl("\\([0-9.]+, [0-9.]+\\)", metrics_no_ci)))

  # Metrics summary with CI (should print, even if CI differs)
  metrics_with_ci <- capture.output(summary(res, type = "metrics", show_ci = TRUE))
  expect_true(length(metrics_with_ci) > 0)

  # Brier time summaries — use existing grid and ensure call succeeds
  if (length(res$survival_brier_times) > 0) {
    first_time <- unname(res$survival_brier_times[1])
    expect_no_error(summary(res, type = "metrics", brier_times = first_time))
  }

  # Predictions
  preds <- res$predictions[[1]]
  expect_true("surv_prob_curve" %in% names(preds))
  eval_times_attr <- attr(preds$surv_prob_curve, "eval_times")
  expect_true(is.numeric(eval_times_attr))

  if (length(preds$surv_prob_curve) > 0) {
    first_curve <- preds$surv_prob_curve[[1]]
    if (!is.null(first_curve)) {
      expect_true(is.numeric(first_curve))
    }
  }

  # surv_time presence if censored is installed
  if (requireNamespace("censored", quietly = TRUE)) {
    expect_true("surv_time" %in% names(preds))
    expect_type(preds$surv_time, "double")
  }
})

test_that("survival resampling results omit .estimator", {
  skip_on_cran()
  data(cancer, package = "survival")
  cancer$status <- cancer$status - 1  # Recode from 1/2 to 0/1

  res <- suppressWarnings(
    fastml(
      data = cancer,
      label = c("time", "status"),
      algorithms = "cox_ph",
      task = "survival",
      resampling_method = "cv",
      folds = 3,
      impute_method = "remove",
      test_size = 0.3,
      seed = 123
    )
  )

  resampling <- res$resampling_results[["cox_ph (survival)"]]
  expect_true(is.list(resampling))
  expect_true("aggregated" %in% names(resampling))
  expect_true("folds" %in% names(resampling))
})

test_that("penalized Cox survival model trains and evaluates", {
  skip_on_cran()
  # Ensure required packages are installed
  skip_if_not_installed("censored")
  skip_if_not_installed("glmnet")

  data(cancer, package = "survival")
  cancer$status <- cancer$status - 1  # Recode from 1/2 to 0/1

  # Suppress expected warnings about status recoding and NA removal
  suppressWarnings({
    res <- fastml(
      data = cancer,
      label = c("time", "status"),
      algorithms = "penalized_cox",
      task = "survival",
      resampling_method = "none",
      impute_method = "remove",
      test_size = 0.3
    )
  })

  expect_s3_class(res, "fastml")
  perf <- res$performance[[1]]

  expect_true(all(c("c_index", "uno_c", "ibs", "rmst_diff") %in% perf$.metric))

  brier_metrics <- perf$.metric[grepl("^brier_t", perf$.metric)]
  expect_true(length(brier_metrics) >= 1)
  brier_val <- perf$.estimate[perf$.metric %in% brier_metrics]
  expect_true(length(brier_val) > 0)
  expect_true(all(is.finite(brier_val)))
  expect_true(all(brier_val >= 0 & brier_val <= 1))
  expect_identical(res$engine_names$penalized_cox, "glmnet")

  preds <- res$predictions[[1]]
  expect_true("surv_prob_curve" %in% names(preds))
  expect_true(is.numeric(attr(preds$surv_prob_curve, "eval_times")))

  if (requireNamespace("censored", quietly = TRUE)) {
    expect_true("surv_time" %in% names(preds))
    expect_type(preds$surv_time, "double")
  }
})

test_that("stratified Cox summary reports strata without coefficients", {
  skip_on_cran()
  data(cancer, package = "survival")
  cancer$status <- cancer$status - 1  # Recode from 1/2 to 0/1
  cancer$strata_inst <- factor(cancer$inst)
  set.seed(123)
  res <- suppressWarnings(
      fastml(
      data = cancer,
      label = c("time", "status"),
      algorithms = c("stratified_cox"),
      task = "survival",
      resampling_method = "none",
      test_size = 0.3,
      impute_method = "remove"
    )
  )
  expect_s3_class(res, "fastml")
  summary_lines <- capture.output(summary(res))
  expect_true(any(grepl("^  Stratified by:", summary_lines)))
  expect_false(any(grepl("^\\s*inst\\s+<NA>", summary_lines)))
  spec <- res$models[[1]]
  expect_true(inherits(spec, "fastml_native_survival"))
  strata_info <- spec$strata_info[["strata_inst"]]
  expect_true(is.list(strata_info))
  expected_levels <- strata_info$levels
  expected_levels <- expected_levels[!is.na(expected_levels) & expected_levels != ""]
  strat_line <- summary_lines[grepl("inst \\(", summary_lines)]
  expect_true(length(strat_line) >= 1)
  if (length(expected_levels) > 0) {
    printed_levels <- sub("^.*strata: ", "", strat_line[1])
    printed_levels <- sub("\\)$", "", printed_levels)
    printed_levels <- trimws(strsplit(printed_levels, ",")[[1]])
    expect_setequal(printed_levels, expected_levels)
  }
  perf_tbl <- res$performance[[1]]
  expect_true(is.data.frame(perf_tbl))
  c_index <- perf_tbl$.estimate[perf_tbl$.metric == "c_index"][1]
  conc_line <- summary_lines[grepl("^  Concordance \\(Harrell C-index\\):", summary_lines)]
  expect_true(length(conc_line) == 1)
  conc_value <- as.numeric(trimws(sub("^  Concordance \\(Harrell C-index\\):", "", conc_line[1])))
  expect_true(is.finite(conc_value))
  expect_equal(conc_value, signif(c_index, 4))
})

test_that("survreg survival model returns Brier scores", {
  skip_on_cran()
  data(cancer, package = "survival")
  cancer$status <- cancer$status - 1  # Recode from 1/2 to 0/1
  res <- suppressWarnings(
    fastml(
      data = cancer,
      label = c("time", "status"),
      algorithms = c("survreg"),
      task = "survival",
      resampling_method = "none",
      impute_method = "remove",
      test_size = 0.3
    )
  )
  expect_s3_class(res, "fastml")
  perf <- res$performance[[1]]
  expect_true(all(c("c_index", "uno_c", "ibs", "rmst_diff") %in% perf$.metric))
  brier_metrics <- perf$.metric[grepl("^brier_t", perf$.metric)]
  expect_true(length(brier_metrics) >= 1)
  brier_val <- perf$.estimate[perf$.metric %in% brier_metrics]
  expect_true(length(brier_val) > 0)
  expect_true(all(is.finite(brier_val)))
  expect_true(all(brier_val >= 0 & brier_val <= 1))
  preds <- res$predictions[[1]]
  expect_true("surv_prob_curve" %in% names(preds))
  expect_true(is.numeric(attr(preds$surv_prob_curve, "eval_times")))
})

test_that("parametric_surv flexsurv integration returns survival metrics", {
  skip_on_cran()
  skip_if_not_installed("flexsurv")

  suppressWarnings(data(lung, package = "survival"))
  lung_surv <- subset(lung, select = c(time, status, age, sex, ph.ecog))
  lung_surv <- stats::na.omit(lung_surv)
  lung_surv$status <- lung_surv$status - 1  # Recode from 1/2 to 0/1
  lung_surv$sex <- factor(lung_surv$sex, levels = 1:2, labels = c("male", "female"))

  set.seed(123)
  res <- suppressWarnings(
    fastml(
      data = lung_surv,
      label = c("time", "status"),
      task = "survival",
      algorithms = "parametric_surv",
      metric = "ibs",
      resampling_method = "none",
      test_size = 0.30,
      impute_method = "remove",
      eval_times = c(90, 180, 365),
      engine_params = list(
        parametric_surv = list(
          flexsurvreg = list(dist = "loglogistic")
        )
      )
    )
  )

  expect_s3_class(res, "fastml")
  perf <- res$performance[[1]]
  expect_true(any(perf$.metric == "ibs"))
  ibs_val <- perf$.estimate[perf$.metric == "ibs"][1]
  expect_true(is.finite(ibs_val))

  brier_metrics <- perf$.metric[grepl("^brier_t", perf$.metric)]
  expect_true(length(brier_metrics) >= 1)
  brier_vals <- perf$.estimate[perf$.metric %in% brier_metrics]
  expect_true(all(is.finite(brier_vals)))
  expect_true(all(brier_vals >= 0 & brier_vals <= 1))
})


test_that("piecewise_exp flexsurv generates default knots when none supplied", {
  skip_on_cran()
  skip_if_not_installed("flexsurv")
  skip_if_not_installed("survival")

  lung <- survival::lung
  lung_surv <- subset(lung, select = c(time, status, age, sex, ph.ecog))
  lung_surv <- stats::na.omit(lung_surv)
  lung_surv$status <- lung_surv$status - 1  # Recode from 1/2 to 0/1
  lung_surv$sex <- factor(lung_surv$sex, levels = 1:2, labels = c("male", "female"))

  set.seed(123)
  res <- suppressWarnings(
    fastml(
      data = lung_surv,
      label = c("time", "status"),
      task = "survival",
      algorithms = "piecewise_exp",
      metric = "ibs",
      resampling_method = "none",
      test_size = 0.30,
      impute_method = "remove",
      eval_times = c(90, 180, 365)
    )
  )

  expect_s3_class(res, "fastml")
  model_extras <- res$models[[1]]$extras
  expect_true(is.list(model_extras))
  expect_equal(model_extras$distribution_label, "piecewise exponential")
  expect_false(is.null(model_extras$breaks))
  expect_true(length(model_extras$breaks) >= 1)
  expect_true(all(is.finite(model_extras$breaks)))
})

test_that("survival random forest with aorsf engine trains when available", {
  skip_on_cran()
  skip_if_not_installed("aorsf")
  skip_if_not_installed("censored")

  data(cancer, package = "survival")
  cancer$status <- cancer$status - 1  # Recode from 1/2 to 0/1

  set.seed(123)
  res <- suppressWarnings(
    fastml(
      data = cancer,
      label = c("time", "status"),
      algorithms = c("rand_forest"),
      algorithm_engines = list(rand_forest = "aorsf"),
      task = "survival",
      resampling_method = "none",
      impute_method = "remove",
      test_size = 0.3
    )
  )

  expect_s3_class(res, "fastml")

  perf <- res$performance[[1]]
  expect_true(all(c("c_index", "uno_c", "ibs", "rmst_diff") %in% perf$.metric))
  brier_metrics <- perf$.metric[grepl("^brier_t", perf$.metric)]
  expect_true(length(brier_metrics) >= 1)
  brier_val <- perf$.estimate[perf$.metric %in% brier_metrics]
  expect_true(length(brier_val) > 0)
  expect_true(all(is.finite(brier_val)))
  expect_true(all(brier_val >= 0 & brier_val <= 1))
})

#' Guarded Resampling Utilities
#'
#' Internal helpers that enforce the Guarded Resampling Principle by
#' fitting preprocessing pipelines independently within each resampling
#' split. These functions are not exported.
#' @param indices Numeric vector of row indices for a resample split.
#' @param label Character string used to identify the index source in errors.
#'
#' @importFrom dplyr bind_rows group_by summarise mutate rename ungroup n
#' @importFrom tibble as_tibble
#' @importFrom rsample analysis assessment
fastml_guard_validate_indices <- function(indices, label) {
  if (!is.atomic(indices) || length(dim(indices)) > 0) {
    stop(sprintf("Guarded resampling requires '%s' to be a vector of indices.", label))
  }
  if (!is.numeric(indices)) {
    stop(sprintf("Guarded resampling requires '%s' to be numeric indices.", label))
  }
  if (length(indices) == 0) {
    stop(sprintf("Guarded resampling requires '%s' to contain at least one index.", label))
  }
  if (anyNA(indices)) {
    stop(sprintf("Guarded resampling requires '%s' to contain no missing values.", label))
  }
  if (any(indices < 1)) {
    stop(sprintf("Guarded resampling requires '%s' to contain positive indices.", label))
  }
  if (any(indices != floor(indices))) {
    stop(sprintf("Guarded resampling requires '%s' to contain integer indices.", label))
  }
  as.integer(indices)
}

fastml_guard_index_candidates <- function() {
  c("in_id", "analysis_id", "analysis_ids", "analysis_index", "analysis_indices", "analysis")
}

fastml_guard_split_index_candidates <- function(split) {
  candidates <- fastml_guard_index_candidates()
  valid <- character(0)
  for (candidate in candidates) {
    value <- split[[candidate]]
    if (is.null(value)) {
      next
    }
    if (!is.atomic(value) || length(dim(value)) > 0 || !is.numeric(value)) {
      next
    }
    ok <- tryCatch({
      fastml_guard_validate_indices(value, candidate)
      TRUE
    }, error = function(e) FALSE)
    if (ok) {
      valid <- c(valid, candidate)
    }
  }
  valid
}

fastml_guard_select_index_name <- function(splits) {
  candidates <- fastml_guard_index_candidates()
  valid_names <- lapply(splits, fastml_guard_split_index_candidates)
  has_valid <- vapply(valid_names, length, integer(1)) > 0
  if (!all(has_valid)) {
    stop(
      "Guarded resampling requires split index information (e.g., `in_id`) for every resample."
    )
  }
  common <- Reduce(intersect, valid_names)
  if (length(common) == 0) {
    stop(
      "Guarded resampling requires a consistent index field (e.g., `in_id`) across resamples."
    )
  }
  for (candidate in candidates) {
    if (candidate %in% common) {
      return(candidate)
    }
  }
  common[[1]]
}

fastml_guard_extract_in_id <- function(split, index_name = NULL) {
  if (!is.null(index_name)) {
    value <- split[[index_name]]
    if (is.null(value)) {
      stop(
        sprintf(
          "Guarded resampling requires '%s' index information for every resample.",
          index_name
        )
      )
    }
    return(fastml_guard_validate_indices(value, index_name))
  }
  candidates <- fastml_guard_index_candidates()
  for (candidate in candidates) {
    value <- split[[candidate]]
    if (is.null(value)) {
      next
    }
    if (!is.atomic(value) || length(dim(value)) > 0 || !is.numeric(value)) {
      next
    }
    return(fastml_guard_validate_indices(value, candidate))
  }
  NULL
}

fastml_guard_is_bootstrap_split <- function(split) {
  inherits(split, "boot_split")
}

fastml_guard_is_bootstrap_resamples <- function(resamples, plan = NULL) {
  if (!is.null(plan)) {
    method <- fastml_resample_method(plan)
    if (!is.null(method)) {
      method <- tolower(as.character(method))
      if (!is.na(method) && method %in% c("boot", "bootstrap", "bootstraps")) {
        return(TRUE)
      }
    }
  }
  if (!is.null(resamples) && inherits(resamples, "bootstraps")) {
    return(TRUE)
  }
  FALSE
}

fastml_guard_detect_full_analysis <- function(in_id, total_rows, is_bootstrap = FALSE) {
  if (is_bootstrap) {
    return(FALSE)
  }
  if (length(in_id) < total_rows) {
    return(FALSE)
  }
  sorted_unique <- sort(unique(in_id))
  if (length(sorted_unique) != total_rows) {
    return(FALSE)
  }
  all(sorted_unique == seq_len(total_rows))
}

fastml_guarded_resample_fit <- function(workflow_spec,
                                        resamples,
                                        original_train_rows,
                                        task,
                                        label,
                                        metric,
                                        event_class,
                                        class_threshold = 0.5,
                                        engine,
                                        start_col = NULL,
                                        time_col = NULL,
                                        status_col = NULL,
                                        eval_times = NULL,
                                        at_risk_threshold = 0.1,
                                        survival_metric_convention = "fastml",
                                        engine_args = list(),
                                        seed = NULL,
                                        summaryFunction = NULL,
                                        multiclass_auc = "macro",
                                        store_fold_models = FALSE) {
  survival_metric_convention <- fastml_normalize_survival_convention(survival_metric_convention)
  plan <- NULL
  if (fastml_is_resample_plan(resamples)) {
    plan <- fastml_resample_validate(resamples)
    resamples <- fastml_resample_splits(plan)
  }

  if (!inherits(resamples, "rset")) {
    stop("'resamples' must be an 'rset' object for guarded resampling.")
  }

  splits <- resamples$splits
  if (length(splits) == 0) {
    return(NULL)
  }

  index_name <- fastml_guard_select_index_name(splits)
  bootstrap_resamples <- fastml_guard_is_bootstrap_resamples(resamples, plan)

  fold_metrics <- vector("list", length(splits))
  fold_models <- if (store_fold_models) vector("list", length(splits)) else NULL
  fold_train_data <- if (store_fold_models) vector("list", length(splits)) else NULL

  for (i in seq_along(splits)) {
    split <- splits[[i]]

    in_id <- fastml_guard_extract_in_id(split, index_name = index_name)
    is_bootstrap_split <- bootstrap_resamples || fastml_guard_is_bootstrap_split(split)

    if (fastml_guard_detect_full_analysis(in_id, original_train_rows, is_bootstrap_split)) {
      stop(
        paste(
          "Detected preprocessing on the full training set during resampling.",
          "Each fold must train preprocessing exclusively on its analysis subset."
        )
      )
    }

    analysis_data <- rsample::analysis(split)
    assessment_data <- rsample::assessment(split)

    fit_args <- list(object = workflow_spec, data = analysis_data)
    fold_seed <- fastml_tuning_seed(seed, offset = i)
    fold_fit <- fastml_with_seed(fold_seed, function() {
      if (!is.null(engine_args) && length(engine_args) > 0) {
        call_with_engine_params(parsnip::fit, fit_args, engine_args)
      } else {
        parsnip::fit(workflow_spec, data = analysis_data)
      }
    })

    fold_result <- process_model(
      model_obj = fold_fit,
      model_id = paste0("fold_", i),
      task = task,
      test_data = assessment_data,
      label = label,
      event_class = event_class,
      class_threshold = class_threshold,
      start_col = start_col,
      time_col = time_col,
      status_col = status_col,
      engine = engine,
      train_data = analysis_data,
      metric = metric,
      eval_times_user = eval_times,
      bootstrap_ci = FALSE,
      bootstrap_samples = 0,
      bootstrap_seed = NULL,
      at_risk_threshold = at_risk_threshold,
      survival_metric_convention = survival_metric_convention,
      summaryFunction = summaryFunction,
      multiclass_auc = multiclass_auc
    )

    fold_metrics[[i]] <- fold_result$performance

    # Optionally store fold models for stability analysis
    if (store_fold_models) {
      fold_models[[i]] <- fold_fit
      fold_train_data[[i]] <- analysis_data
    } else {
      rm(fold_fit)
      rm(analysis_data)
    }
    rm(assessment_data)
    gc(verbose = FALSE)
  }

  fold_metrics_df <- dplyr::bind_rows(fold_metrics, .id = "fold")
  ci_cols <- c(".lower", ".upper", ".n_boot")
  keep_cols <- setdiff(names(fold_metrics_df), ci_cols)
  fold_metrics_df <- fold_metrics_df[, keep_cols, drop = FALSE]

  # Explicitly ungroup and convert to plain tibble to strip any lingering

  # grouping attributes that may have carried over from the fold processing
  fold_metrics_df <- dplyr::ungroup(fold_metrics_df)
  fold_metrics_df <- tibble::as_tibble(fold_metrics_df)

  if (identical(task, "survival")) {
    fold_metrics_df <- fold_metrics_df[, setdiff(names(fold_metrics_df), ".estimator"), drop = FALSE]
    aggregated <- fold_metrics_df %>%
      dplyr::group_by(.metric) %>%
      dplyr::summarise(
        mean_est = mean(.estimate, na.rm = TRUE),
        n_valid = sum(!is.na(.estimate)),
        sd_est = stats::sd(.estimate, na.rm = TRUE),
        .groups = "drop"
      ) %>%
      dplyr::mutate(
        std_err = ifelse(n_valid > 0, sd_est / sqrt(n_valid), NA_real_)
      ) %>%
      dplyr::rename(
        .estimate = mean_est,
        n = n_valid,
        std_dev = sd_est
      )
  } else {
    if (!".estimator" %in% names(fold_metrics_df)) {
      fold_metrics_df$.estimator <- NA_character_
    }
    aggregated <- fold_metrics_df %>%
      dplyr::group_by(.metric, .estimator) %>%
      dplyr::summarise(
        mean_est = mean(.estimate, na.rm = TRUE),
        n_valid = sum(!is.na(.estimate)),
        sd_est = stats::sd(.estimate, na.rm = TRUE),
        .groups = "drop"
      ) %>%
      dplyr::mutate(
        std_err = ifelse(n_valid > 0, sd_est / sqrt(n_valid), NA_real_)
      ) %>%
      dplyr::rename(
        .estimate = mean_est,
        n = n_valid,
        std_dev = sd_est
      )
  }

  result <- list(
    aggregated = aggregated,
    folds = fold_metrics_df
  )

  # Add fold models for stability analysis if requested
  if (store_fold_models && !is.null(fold_models)) {
    result$fold_models <- fold_models
    result$fold_train_data <- fold_train_data
  }

  if (!is.null(plan)) {
    result$metadata <- fastml_resample_metadata(plan)
  }

  result
}

library(testthat)

test_that("bootstrap coverage does not trigger the full-data guard", {
  in_id <- seq_len(10)
  expect_true(
    fastml:::fastml_guard_detect_full_analysis(in_id, 10, is_bootstrap = FALSE)
  )
  expect_false(
    fastml:::fastml_guard_detect_full_analysis(in_id, 10, is_bootstrap = TRUE)
  )
})

test_that("guarded resampling errors when split indices are missing", {
  skip_on_cran()
  skip_if_not_installed("rsample")

  set.seed(101)
  binary_iris <- iris[iris$Species != "virginica", ]
  binary_iris$Species <- factor(binary_iris$Species)

  idx <- sample(seq_len(nrow(binary_iris)), size = floor(0.7 * nrow(binary_iris)))
  train_split <- binary_iris[idx, , drop = FALSE]
  test_split <- binary_iris[-idx, , drop = FALSE]

  resamples <- rsample::vfold_cv(train_split, v = 3)
  resamples$splits[[1]]$in_id <- NULL

  # The error is now caught, reported as a warning, and tracked in failed_models
  # The final error is "No models were successfully trained"
  expect_error(
    expect_warning(
      fastml(
        train_data = train_split,
        test_data = test_split,
        label = "Species",
        algorithms = "logistic_reg",
        resamples = resamples,
        resampling_method = "cv",
        folds = 3,
        seed = 202
      ),
      "split index information"
    ),
    "No models were successfully trained"
  )
})

test_that("guarded resampling requires consistent index fields", {
  skip_on_cran()
  skip_if_not_installed("rsample")

  set.seed(202)
  binary_iris <- iris[iris$Species != "virginica", ]
  binary_iris$Species <- factor(binary_iris$Species)

  idx <- sample(seq_len(nrow(binary_iris)), size = floor(0.7 * nrow(binary_iris)))
  train_split <- binary_iris[idx, , drop = FALSE]
  test_split <- binary_iris[-idx, , drop = FALSE]

  resamples <- rsample::vfold_cv(train_split, v = 3)
  resamples$splits[[1]]$analysis_id <- resamples$splits[[1]]$in_id
  resamples$splits[[1]]$in_id <- NULL

  # The error is now caught, reported as a warning, and tracked in failed_models
  # The final error is "No models were successfully trained"
  expect_error(
    expect_warning(
      fastml(
        train_data = train_split,
        test_data = test_split,
        label = "Species",
        algorithms = "logistic_reg",
        resamples = resamples,
        resampling_method = "cv",
        folds = 3,
        seed = 303
      ),
      "consistent index field"
    ),
    "No models were successfully trained"
  )
})

#' Save Model Function
#'
#' Saves the trained model object to a file.
#'
#' @param model An object of class \code{fastml}.
#' @param filepath A string specifying the file path to save the model.
#' @return No return value, called for its side effect of saving the model object to a file.
#' @export
save_fastml <- function(model, filepath) {
  saveRDS(model, filepath)
}

#' Save Model Function (Deprecated)
#'
#' \code{save.fastml} is deprecated in favour of \code{\link{save_fastml}}.
#' The old name resembled an S3 method for \code{base::save()}, which is not
#' a generic, leading to dispatch confusion.
#'
#' @inheritParams save_fastml
#' @return No return value, called for its side effect of saving the model object to a file.
#' @export
save.fastml <- function(model, filepath) {
  .Deprecated("save_fastml")
  save_fastml(model, filepath)
}


#' Load Model Function
#'
#' Loads a trained model object from a file.
#'
#' @param filepath A string specifying the file path to load the model from.
#' @return An object of class \code{fastml}.
#' @export
load_model <- function(filepath) {
  model <- readRDS(filepath)
  return(model)
}

library(testthat)
library(dplyr)

# small dataset for tests
set.seed(123)
data(iris)
iris_binary <- iris[iris$Species != "setosa", ]
iris_binary$Species <- factor(iris_binary$Species)

test_that("save.fastml and load_model work correctly for classification", {
  skip_on_cran()

  model <- fastml(
    data = iris_binary,
    label = "Species",
    algorithms = "rand_forest",
    resampling_method = "cv",
    folds = 2
  )

  temp_file <- tempfile(fileext = ".rds")
  on.exit(unlink(temp_file), add = TRUE)

  # Test save
  expect_error(save.fastml(model, temp_file), regexp = NA)
  expect_true(file.exists(temp_file))

  # Test load
  loaded_model <- load_model(temp_file)
  expect_true(inherits(loaded_model, "fastml"))
  expect_equal(loaded_model$label, model$label)
  expect_equal(loaded_model$task, model$task)
})

test_that("save.fastml and load_model work correctly for regression", {
  skip_on_cran()

  set.seed(42)
  data(mtcars)

  model <- fastml(
    data = mtcars,
    label = "mpg",
    algorithms = "rand_forest",
    resampling_method = "cv",
    folds = 2
  )

  temp_file <- tempfile(fileext = ".rds")
  on.exit(unlink(temp_file), add = TRUE)

  save.fastml(model, temp_file)
  loaded_model <- load_model(temp_file)

  expect_true(inherits(loaded_model, "fastml"))
  expect_equal(loaded_model$label, "mpg")
  expect_equal(loaded_model$task, "regression")
})

test_that("loaded model can make predictions", {
  skip_on_cran()

  model <- fastml(
    data = iris_binary,
    label = "Species",
    algorithms = "rand_forest",
    resampling_method = "cv",
    folds = 2
  )

  temp_file <- tempfile(fileext = ".rds")
  on.exit(unlink(temp_file), add = TRUE)

  save.fastml(model, temp_file)
  loaded_model <- load_model(temp_file)

  # Test that predictions work on loaded model
  new_data <- iris_binary[1:5, ]
  new_data$Species <- NULL

  preds_original <- suppressWarnings(predict(model, new_data))
  preds_loaded <- suppressWarnings(predict(loaded_model, new_data))

  expect_equal(length(preds_original), 5)
  expect_equal(length(preds_loaded), 5)
})

test_that("save.fastml overwrites existing file", {
  skip_on_cran()

  model1 <- fastml(
    data = iris_binary,
    label = "Species",
    algorithms = "rand_forest",
    resampling_method = "cv",
    folds = 2
  )

  set.seed(42)
  data(mtcars)
  model2 <- fastml(
    data = mtcars,
    label = "mpg",
    algorithms = "rand_forest",
    resampling_method = "cv",
    folds = 2
  )

  temp_file <- tempfile(fileext = ".rds")
  on.exit(unlink(temp_file), add = TRUE)

  save.fastml(model1, temp_file)
  save.fastml(model2, temp_file)

  loaded <- load_model(temp_file)
  expect_equal(loaded$label, "mpg")
})

fastml_setup_parallel <- function(n_cores, seed) {
  if (!requireNamespace("future", quietly = TRUE)) {
    stop("The 'future' package is required but is not installed.")
  }

  old_plan <- future::plan()
  old_future_seed <- getOption("future.seed")

  n_cores_val <- fastml_normalize_threads(n_cores)
  if (is.null(n_cores_val)) {
    n_cores_val <- 1L
  }

  use_parallel <- n_cores_val > 1L
  old_do_par <- NULL

  if (use_parallel) {
    if (!requireNamespace("doFuture", quietly = TRUE)) {
      stop("The 'doFuture' package is required for parallel processing but is not installed.")
    }

    # CAPTURE: Register doFuture and save the PREVIOUS backend info.
    # registerDoFuture() takes no arguments and returns the old backend
    # (a list of fun/data/info) which we restore on cleanup.
    # See: https://dofuture.futureverse.org/reference/registerDoFuture.html
    old_do_par <- doFuture::registerDoFuture()

    seed_val <- fastml_normalize_seed(seed)
    options(future.seed = if (!is.null(seed_val)) seed_val else TRUE)
    future::plan(future::multisession, workers = n_cores_val)
  } else {
    # plan(sequential) is sufficient for sequential execution.
    # doFuture handles this automatically without needing registerDoSEQ().
    future::plan(future::sequential)
  }

  # RESTORE: Define the cleanup function
  restore <- function() {
    future::plan(old_plan)
    options(future.seed = old_future_seed)

    # If we changed the foreach backend, restore the previous one. When no
    # backend was registered before this call, registerDoFuture() reports the
    # previous data/info hooks as NULL; passing those straight back to
    # setDoPar() leaves foreach in a state where getDoParName() errors for the
    # rest of the session, so the sequential default is registered instead.
    if (!is.null(old_do_par)) {
      if (is.function(old_do_par$info)) {
        do.call(foreach::setDoPar, args = old_do_par)
      } else {
        foreach::registerDoSEQ()
      }
    }
  }

  list(restore = restore, use_parallel = use_parallel)
}

fastml_muffle_foreach_warning <- function(w) {
  msg <- conditionMessage(w)
  if (is.null(msg)) {
    return(invisible())
  }
  if (grepl("executing %dopar% sequentially: no parallel backend registered", msg, fixed = TRUE)) {
    invokeRestart("muffleWarning")
  }
  invisible()
}

library(testthat)

# Regression tests for fastml_setup_parallel(). These are intentionally not
# gated behind skip_on_cran(): they call the setup/restore helper directly
# rather than fitting models, so they are fast enough to run everywhere, and
# they cover two defects that shipped in 0.7.8 precisely because the only test
# touching the parallel path was CRAN-gated.
#
#   1. registerDoFuture() was called with a `flavor` argument it does not
#      accept, so every n_cores > 1 call failed with "unused argument".
#   2. When no foreach backend existed beforehand, the cleanup handler passed
#      NULL data/info hooks back to setDoPar(), after which getDoParName()
#      errored for the rest of the session.

test_that("sequential setup registers and restores without error", {
  skip_if_not_installed("future")

  handle <- fastml:::fastml_setup_parallel(n_cores = 1, seed = 42)
  expect_false(handle$use_parallel)
  expect_type(handle$restore, "closure")

  expect_silent(handle$restore())
})

test_that("parallel setup registers a doFuture backend", {
  skip_if_not_installed("future")
  skip_if_not_installed("doFuture")
  skip_if_not_installed("foreach")

  handle <- fastml:::fastml_setup_parallel(n_cores = 2, seed = 42)
  on.exit(handle$restore(), add = TRUE)

  expect_true(handle$use_parallel)
  expect_identical(foreach::getDoParName(), "doFuture")
})

test_that("restore leaves foreach in a usable state", {
  skip_if_not_installed("future")
  skip_if_not_installed("doFuture")
  skip_if_not_installed("foreach")

  handle <- fastml:::fastml_setup_parallel(n_cores = 2, seed = 42)
  handle$restore()

  # The failure mode this guards against is getDoParName() raising
  # "attempt to apply non-function" rather than returning a backend name.
  expect_error(foreach::getDoParName(), NA)
  expect_true(is.character(foreach::getDoParName()))
})

test_that("restore returns the future plan and future.seed option", {
  skip_if_not_installed("future")

  old_seed_option <- getOption("future.seed")
  handle <- fastml:::fastml_setup_parallel(n_cores = 1, seed = 7)
  handle$restore()

  expect_identical(getOption("future.seed"), old_seed_option)
  expect_true(inherits(future::plan(), "future"))
})

test_that("a foreach backend registered by the caller is restored", {
  skip_if_not_installed("future")
  skip_if_not_installed("doFuture")
  skip_if_not_installed("foreach")

  foreach::registerDoSEQ()
  before <- foreach::getDoParName()

  handle <- fastml:::fastml_setup_parallel(n_cores = 2, seed = 42)
  handle$restore()

  expect_identical(foreach::getDoParName(), before)
})

# n_cores sizes the worker pool; engine_threads sizes each engine's thread pool.
# Driving both from n_cores meant a request for k cores could demand k^2 threads.

test_that("engine_threads, not n_cores, sets the engine thread count", {
  for (et in c(1L, 2L, 4L)) {
    ea <- fastml:::fastml_apply_engine_seed(
      list(), "rand_forest", "ranger", seed = 2025, n_cores = et,
      task = "regression"
    )
    expect_identical(as.integer(ea$num.threads), et)
  }
})

test_that("both arguments are exposed with sequential defaults", {
  f <- formals(fastml::fastml)
  expect_true("engine_threads" %in% names(f))
  expect_identical(eval(f$n_cores), 1)
  expect_identical(eval(f$engine_threads), 1)
  expect_true("engine_threads" %in% names(formals(fastml::train_models)))
})

test_that("results are invariant to the worker/thread split", {
  # Gated: this fits three models to check a property, where the tests above
  # catch the defects. It costs ~20s against ~4s for the whole rest of the file.
  skip_on_cran()
  skip_if_not_installed("ranger")
  skip_if_not_installed("future")

  d <- iris[iris$Species != "virginica", ]
  d$Species <- factor(d$Species)
  est <- function(nc, et) {
    set.seed(7)
    m <- suppressWarnings(suppressMessages(fastml(
      data = d, label = "Species", algorithms = "rand_forest",
      n_cores = nc, engine_threads = et, seed = 42,
      verbose = FALSE, bootstrap_ci = FALSE
    )))
    p <- as.data.frame(m$performance[[1]])
    p$.estimate[p$.metric == "roc_auc"][1]
  }
  base <- est(1, 1)
  expect_equal(est(2, 1), base, tolerance = 0)
  expect_equal(est(1, 2), base, tolerance = 0)
})

#' Lightweight exploratory helper
#'
#' `fastexplore()` is an optional, lightweight exploratory data analysis (EDA)
#' helper. It returns summary tables and plot objects; it only writes to disk or
#' renders a report when you explicitly request it via `save_results` or
#' `render_report`.
#'
#' This helper is intentionally decoupled from the core modeling workflow. Most
#' of its heavy dependencies are treated as optional and loaded via
#' `requireNamespace()` when requested features are used.
#'
#' @param data A `data.frame` to explore.
#' @param label Optional column name of the target/label. If supplied and
#'   categorical, grouped plots and class balance summaries are produced.
#' @param visualize Character vector indicating which plot families to build.
#'   Defaults to `c("histogram", "boxplot", "barplot", "heatmap", "scatterplot")`.
#' @param save_results Logical; if `TRUE`, plots/results are saved under
#'   `output_dir` (defaults to the working directory). Default is `FALSE`.
#' @param render_report Logical; if `TRUE`, a short HTML report is rendered via
#'   `rmarkdown` (if available). Default is `FALSE`.
#' @param output_dir Directory to save results/report when `save_results` or
#'   `render_report` is `TRUE`.
#' @param sample_size Optional integer; if supplied, visualizations are produced
#'   on a random sample of this size.
#' @param interactive Logical; if `TRUE` and `plotly` is available, an
#'   interactive correlation heatmap is produced. Falls back to static ggplot
#'   output otherwise.
#' @param corr_threshold Absolute correlation threshold for flagging high
#'   correlations.
#' @param auto_convert_numeric Logical; convert factor/character columns that
#'   look numeric into numeric.
#' @param visualize_missing Logical; if `TRUE`, include simple missingness
#'   visualizations.
#' @param imputation_suggestions Logical; if `TRUE`, prints lightweight
#'   suggestions based on missingness patterns.
#' @param report_duplicate_details Logical; if `TRUE`, returns a small sample of
#'   duplicated rows when present.
#' @param detect_near_duplicates Placeholder for future fuzzy duplicate checks.
#' @param auto_convert_dates Logical; convert YYYY-MM-DD strings to `Date`.
#' @param feature_engineering Logical; if `TRUE`, derive day/month/year from date
#'   columns to aid inspection of temporal structure.
#' @param outlier_method One of `"iqr"`, `"zscore"`, `"dbscan"`, `"lof"`.
#' @param run_distribution_checks Logical; if `TRUE`, run normality tests on
#'   numeric columns.
#' @param normality_tests Character vector of normality tests to run; currently
#'   supports `"shapiro"` and `"ks"`.
#' @param pairwise_matrix Logical; if `TRUE` and `GGally` is available, returns a
#'   ggpairs scatterplot matrix for a subset of numeric columns.
#' @param max_scatter_cols Maximum number of numeric columns to include in the
#'   pairwise matrix.
#' @param grouped_plots Logical; if `TRUE` and `label` is a factor, group
#'   histograms/boxplots/density plots by label.
#' @param use_upset_missing Logical; retained for compatibility. When `TRUE` and
#'   `UpSetR` is installed, an UpSet plot of missingness is returned; otherwise a
#'   simpler missingness heatmap is used.
#'
#' @return A list of summaries (tables/tibbles) and plot objects (ggplot/plotly),
#'   plus any saved file paths when `save_results`/`render_report` are enabled.
#'
#' @importFrom stats na.omit
#' @export
fastexplore <- function(
  data,
  label = NULL,
  visualize = c("histogram", "boxplot", "barplot", "heatmap", "scatterplot"),
  save_results = FALSE,
  render_report = FALSE,
  output_dir = NULL,
  sample_size = NULL,
  interactive = FALSE,
  corr_threshold = 0.9,
  auto_convert_numeric = TRUE,
  visualize_missing = TRUE,
  imputation_suggestions = FALSE,
  report_duplicate_details = TRUE,
  detect_near_duplicates = FALSE,
  auto_convert_dates = FALSE,
  feature_engineering = FALSE,
  outlier_method = c("iqr", "zscore", "dbscan", "lof"),
  run_distribution_checks = TRUE,
  normality_tests = c("shapiro"),
  pairwise_matrix = TRUE,
  max_scatter_cols = 5,
  grouped_plots = TRUE,
  use_upset_missing = TRUE
) {
  outlier_method <- match.arg(outlier_method)
  visualize <- match.arg(visualize, several.ok = TRUE)
  normality_tests <- match.arg(normality_tests, several.ok = TRUE)

  if (!is.data.frame(data)) {
    stop("'data' must be a data.frame")
  }

  if (nrow(data) == 0) {
    stop("'data' has no rows to explore")
  }

  if (!is.null(label) && !label %in% colnames(data)) {
    stop(sprintf("Label column '%s' is not present in data", label))
  }

  data <- tibble::as_tibble(data)

  data <- dplyr::mutate(
    data,
    dplyr::across(where(is.integer), as.numeric),
    dplyr::across(where(is.character), as.factor)
  )

  if (auto_convert_numeric) {
    factor_like <- names(Filter(function(x) is.factor(x) || is.character(x), data))
    for (col in factor_like) {
      x <- data[[col]]
      if (all(grepl("^[0-9.-]+$", x[!is.na(x)])) && length(unique(x)) >= 6) {
        new_x <- suppressWarnings(as.numeric(as.character(x)))
        if (sum(is.na(new_x)) <= sum(is.na(x))) {
          data[[col]] <- new_x
        }
      }
    }
  }

  if (auto_convert_dates) {
    factor_like <- names(Filter(function(x) is.factor(x) || is.character(x), data))
    for (col in factor_like) {
      x <- data[[col]]
      if (all(grepl("^\\d{4}-\\d{2}-\\d{2}$", x[!is.na(x)]))) {
        converted_date <- as.Date(x, format = "%Y-%m-%d")
        if (!all(is.na(converted_date))) {
          data[[col]] <- converted_date
        }
      }
    }
  }

  if (!is.null(label)) {
    target_var <- data[[label]]
    if (is.numeric(target_var) && length(unique(target_var)) <= 5) {
      data[[label]] <- as.factor(target_var)
      warning(sprintf("Label '%s' converted to factor for grouped plots", label))
    }
  }

  if (feature_engineering) {
    date_cols <- names(Filter(function(x) inherits(x, c("Date", "POSIXct", "POSIXt")), data))
    for (col in date_cols) {
      data[[paste0(col, "_day")]] <- as.numeric(format(data[[col]], "%d"))
      data[[paste0(col, "_month")]] <- as.numeric(format(data[[col]], "%m"))
      data[[paste0(col, "_year")]] <- as.numeric(format(data[[col]], "%Y"))
    }
  }

  numeric_cols <- names(Filter(is.numeric, data))
  factor_cols <- names(Filter(function(x) is.factor(x) || is.character(x), data))

  if (is.null(sample_size) || sample_size >= nrow(data)) {
    data_vis <- data
  } else {
    set.seed(123)
    idx <- sample.int(nrow(data), sample_size)
    data_vis <- data[idx, , drop = FALSE]
  }

  data_dimensions <- tibble::tibble(Rows = nrow(data), Columns = ncol(data))
  data_types <- tibble::tibble(
    Column = names(data),
    Data_Type = vapply(data, function(x) paste(class(x), collapse = ","), character(1))
  )
  unique_values <- dplyr::summarise(data, dplyr::across(dplyr::everything(), dplyr::n_distinct)) |>
    tidyr::pivot_longer(dplyr::everything(), names_to = "Column", values_to = "Unique_Values")

  missing_data <- tibble::tibble(
    Column = names(data),
    Missing = vapply(data, function(x) sum(is.na(x)), integer(1)),
    Missing_Percent = round(Missing / nrow(data) * 100, 2)
  )

  data_summary <- NULL
  if (safe_require("skimr")) {
    data_summary <- skimr::skim(data)
  }

  data_overview <- list(
    dimensions = data_dimensions,
    types = data_types,
    unique_values = unique_values,
    summary = data_summary,
    head = utils::head(data, 6)
  )

  potential_id_cols <- Filter(function(col) dplyr::n_distinct(data[[col]]) == nrow(data), names(data))
  date_time_cols <- names(Filter(function(x) inherits(data[[x]], c("Date", "POSIXct", "POSIXt")), names(data)))

  summary_stats <- NULL
  if (length(numeric_cols) > 0) {
    base_stats <- tibble::tibble(
      Column = numeric_cols,
      Mean = vapply(data[numeric_cols], function(x) mean(x, na.rm = TRUE), numeric(1)),
      Median = vapply(data[numeric_cols], function(x) stats::median(x, na.rm = TRUE), numeric(1)),
      SD = vapply(data[numeric_cols], function(x) stats::sd(x, na.rm = TRUE), numeric(1)),
      Min = vapply(data[numeric_cols], function(x) min(x, na.rm = TRUE), numeric(1)),
      Max = vapply(data[numeric_cols], function(x) max(x, na.rm = TRUE), numeric(1)),
      NAs = vapply(data[numeric_cols], function(x) sum(is.na(x)), integer(1))
    )

    if (safe_require("moments")) {
      base_stats$Skewness <- vapply(data[numeric_cols], function(x) moments::skewness(x, na.rm = TRUE), numeric(1))
      base_stats$Kurtosis <- vapply(data[numeric_cols], function(x) moments::kurtosis(x, na.rm = TRUE), numeric(1))
    }

    summary_stats <- base_stats
  }

  zero_variance_cols <- character()
  if (length(numeric_cols) > 0) {
    variances <- vapply(data[numeric_cols], function(x) stats::var(x, na.rm = TRUE), numeric(1))
    zero_variance_cols <- names(variances[is.na(variances) | variances <= 1e-8])
  }

  freq_tables <- NULL
  if (length(factor_cols) > 0) {
    freq_tables <- lapply(factor_cols, function(col) {
      tab <- table(data[[col]], useNA = "ifany")
      tibble::tibble(value = names(tab), count = as.integer(tab))
    })
    names(freq_tables) <- factor_cols
  }

  duplicated_rows <- sum(duplicated(data))
  duplicated_examples <- NULL
  if (report_duplicate_details && duplicated_rows > 0) {
    duplicated_examples <- utils::head(unique(data[duplicated(data), , drop = FALSE]), 10)
  }

  class_imbalance <- NULL
  if (!is.null(label) && is.factor(data[[label]])) {
    tab <- table(data[[label]], useNA = "ifany")
    class_imbalance <- tibble::tibble(
      class = names(tab),
      count = as.integer(tab),
      percent = round(100 * as.integer(tab) / sum(tab), 2)
    )
  }

  correlation_matrix <- NULL
  high_corr_pairs <- NULL
  if (length(numeric_cols) > 1) {
    correlation_matrix <- stats::cor(data[numeric_cols], use = "pairwise.complete.obs")
    high_pairs_df <- reshape2::melt(correlation_matrix, varnames = c("Var1", "Var2"), value.name = "Correlation")
    high_pairs_df <- high_pairs_df[high_pairs_df$Var1 != high_pairs_df$Var2, ]
    high_corr_pairs <- subset(high_pairs_df, abs(Correlation) >= corr_threshold)
  }

  outlier_summary <- perform_outlier_detection(data, numeric_cols, outlier_method)

  normality_results <- NULL
  normality_plots <- NULL
  if (run_distribution_checks && length(numeric_cols) > 0) {
    normality_results <- perform_normality_tests(data, numeric_cols, normality_tests)
    normality_plots <- generate_normality_plots(data, numeric_cols)
  }

  missingness_plot <- NULL
  upset_plot <- NULL
  if (visualize_missing) {
    missingness_plot <- build_missingness_heatmap(data)
    has_missing <- any(is.na(data))
    if (use_upset_missing && has_missing && safe_require("UpSetR")) {
      missing_df <- data
      missing_df[] <- lapply(missing_df, function(x) as.integer(is.na(x)))
      upset_plot <- UpSetR::upset(as.data.frame(missing_df), nsets = min(ncol(missing_df), 6), keep.order = TRUE)
    }
  }

  heatmap_plot <- NULL
  if ("heatmap" %in% visualize && !is.null(correlation_matrix)) {
    heatmap_plot <- build_heatmap_plot(correlation_matrix, interactive)
  }

  histogram_plots <- NULL
  if ("histogram" %in% visualize && length(numeric_cols) > 0) {
    histogram_plots <- lapply(numeric_cols, function(col) {
      p <- ggplot2::ggplot(data_vis, ggplot2::aes(x = .data[[col]])) +
        ggplot2::geom_histogram(bins = 30, fill = "steelblue", color = "white") +
        ggplot2::labs(title = paste("Histogram of", col), x = col, y = "Count") +
        ggplot2::theme_minimal()
      if (!is.null(label) && label %in% factor_cols && grouped_plots) {
        p <- ggplot2::ggplot(data_vis, ggplot2::aes(x = .data[[col]], fill = .data[[label]])) +
          ggplot2::geom_histogram(position = "identity", alpha = 0.6, bins = 30, color = "white") +
          ggplot2::labs(title = paste("Histogram of", col, "by", label), x = col, y = "Count", fill = label) +
          ggplot2::theme_minimal()
      }
      p
    })
    names(histogram_plots) <- numeric_cols
  }

  boxplot_plots <- NULL
  if ("boxplot" %in% visualize && length(numeric_cols) > 0) {
    boxplot_plots <- lapply(numeric_cols, function(col) {
      p <- ggplot2::ggplot(data_vis, ggplot2::aes(y = .data[[col]])) +
        ggplot2::geom_boxplot(fill = "tomato") +
        ggplot2::labs(title = paste("Boxplot of", col), x = NULL, y = col) +
        ggplot2::theme_minimal()
      if (!is.null(label) && label %in% factor_cols && grouped_plots) {
        p <- ggplot2::ggplot(data_vis, ggplot2::aes(x = .data[[label]], y = .data[[col]], fill = .data[[label]])) +
          ggplot2::geom_boxplot() +
          ggplot2::labs(title = paste("Boxplot of", col, "by", label), x = label, y = col, fill = label) +
          ggplot2::theme_minimal()
      }
      p
    })
    names(boxplot_plots) <- numeric_cols
  }

  bar_plots <- NULL
  if ("barplot" %in% visualize && length(factor_cols) > 0) {
    bar_plots <- lapply(factor_cols, function(col) {
      ggplot2::ggplot(data_vis, ggplot2::aes(x = .data[[col]])) +
        ggplot2::geom_bar(fill = "forestgreen") +
        ggplot2::labs(title = paste("Bar plot of", col), x = col, y = "Count") +
        ggplot2::theme_minimal()
    })
    names(bar_plots) <- factor_cols
  }

  scatter_plots <- NULL
  if ("scatterplot" %in% visualize && length(numeric_cols) > 1) {
    pairs_df <- NULL
    if (!is.null(correlation_matrix)) {
      upper_tri <- correlation_matrix
      upper_tri[lower.tri(upper_tri, diag = TRUE)] <- NA
      pairs_df <- reshape2::melt(upper_tri, varnames = c("Var1", "Var2"), value.name = "Correlation")
      pairs_df <- pairs_df[!is.na(pairs_df$Correlation), ]
      pairs_df <- pairs_df[order(abs(pairs_df$Correlation), decreasing = TRUE), ]
    }

    if (!is.null(pairs_df) && nrow(pairs_df) > 0) {
      top_pairs <- utils::head(pairs_df, 3)
    } else {
      combos <- utils::combn(numeric_cols, 2)
      top_pairs <- tibble::tibble(Var1 = combos[1, ], Var2 = combos[2, ], Correlation = NA_real_)
      top_pairs <- utils::head(top_pairs, 3)
    }

    scatter_plots <- apply(top_pairs, 1, function(row) {
      varx <- row[["Var1"]]
      vary <- row[["Var2"]]
      p <- ggplot2::ggplot(data_vis, ggplot2::aes(x = .data[[varx]], y = .data[[vary]])) +
        ggplot2::geom_point(color = "darkblue", alpha = 0.7) +
        ggplot2::labs(title = paste("Scatterplot:", varx, "vs", vary),
                      subtitle = if (!is.na(row[["Correlation"]])) paste("Correlation =", round(as.numeric(row[["Correlation"]]), 3)) else NULL) +
        ggplot2::theme_minimal()
      if (!is.null(label) && label %in% factor_cols && grouped_plots) {
        p <- ggplot2::ggplot(data_vis, ggplot2::aes(x = .data[[varx]], y = .data[[vary]], color = .data[[label]])) +
          ggplot2::geom_point(alpha = 0.7) +
          ggplot2::labs(title = paste("Scatterplot:", varx, "vs", vary, "by", label), color = label,
                        subtitle = if (!is.na(row[["Correlation"]])) paste("Correlation =", round(as.numeric(row[["Correlation"]]), 3)) else NULL) +
          ggplot2::theme_minimal()
      }
      p
    })
    names(scatter_plots) <- paste0("scatter_", seq_along(scatter_plots))
  }

  pairwise_plot <- NULL
  if (pairwise_matrix && length(numeric_cols) > 1 && safe_require("GGally")) {
    numeric_subset <- numeric_cols[seq_len(min(length(numeric_cols), max_scatter_cols))]
    pairwise_plot <- GGally::ggpairs(data_vis[, numeric_subset, drop = FALSE])
  }

  results_list <- list(
    data_overview = data_overview,
    summary_stats = summary_stats,
    freq_tables = freq_tables,
    missing_data = missing_data,
    duplicated_rows = duplicated_rows,
    duplicated_examples = duplicated_examples,
    class_imbalance = class_imbalance,
    correlation_matrix = correlation_matrix,
    high_corr_pairs = high_corr_pairs,
    zero_variance_cols = zero_variance_cols,
    potential_id_cols = potential_id_cols,
    date_time_cols = date_time_cols,
    outlier_method = outlier_method,
    outlier_summary = outlier_summary,
    normality_tests = normality_results,
    plots = list(
      correlation_heatmap = heatmap_plot,
      histograms = histogram_plots,
      boxplots = boxplot_plots,
      barplots = bar_plots,
      scatterplots = scatter_plots,
      pairwise_matrix = pairwise_plot,
      missingness = missingness_plot,
      upset_missing = upset_plot,
      normality = normality_plots
    ),
    saved_paths = list()
  )

  if (imputation_suggestions && visualize_missing) {
    message("Imputation suggestion: consider median/mean for numeric columns with low missingness and mode/most frequent for factors.")
  }

  if (detect_near_duplicates) {
    message("Near-duplicate detection is not implemented yet.")
  }

  results_folder <- NULL
  if (save_results || render_report) {
    base_dir <- if (is.null(output_dir)) getwd() else output_dir
    results_folder <- file.path(base_dir, paste0("fastexplore_", format(Sys.time(), "%Y%m%d_%H%M%S")))
    dir.create(results_folder, showWarnings = FALSE, recursive = TRUE)
    results_list$saved_paths$results_folder <- results_folder
  }

  if (save_results && !is.null(results_folder)) {
    saveRDS(results_list, file.path(results_folder, "fastexplore_results.rds"))

    save_plot_collection(histogram_plots, file.path(results_folder, "histograms"))
    save_plot_collection(boxplot_plots, file.path(results_folder, "boxplots"))
    save_plot_collection(bar_plots, file.path(results_folder, "barplots"))
    save_plot_collection(scatter_plots, file.path(results_folder, "scatterplots"))

    if (!is.null(heatmap_plot)) {
      save_plot_collection(list(correlation_heatmap = heatmap_plot), file.path(results_folder, "heatmap"))
    }

    if (!is.null(pairwise_plot)) {
      save_plot_collection(list(pairwise_matrix = pairwise_plot), file.path(results_folder, "pairwise"))
    }

    if (!is.null(missingness_plot)) {
      save_plot_collection(list(missingness = missingness_plot), file.path(results_folder, "missingness"))
    }

    if (!is.null(normality_plots)) {
      save_plot_collection(normality_plots, file.path(results_folder, "normality"))
    }
  }

  if (render_report && !is.null(results_folder)) {
    if (!safe_require("rmarkdown")) {
      message("Package 'rmarkdown' is not installed; skipping report rendering.")
    } else {
      report_path <- file.path(results_folder, "fastexplore_report.Rmd")
      results_rds <- file.path(results_folder, "fastexplore_results.rds")
      if (!file.exists(results_rds)) {
        saveRDS(results_list, results_rds)
      }

      report_lines <- c(
        "---",
        "title: 'fastexplore EDA Report'",
        "output: html_document",
        "---",
        "",
        "```{r setup, include=FALSE}",
        "results <- readRDS('fastexplore_results.rds')",
        "knitr::opts_chunk$set(echo = FALSE, warning = FALSE, message = FALSE)",
        "```",
        "",
        "## Data Overview",
        "```{r}",
        "results$data_overview$dimensions",
        "results$data_overview$types",
        "head(results$data_overview$unique_values)",
        "```",
        "",
        "## Summary Stats",
        "```{r}",
        "if (!is.null(results$summary_stats)) knitr::kable(results$summary_stats)",
        "```",
        "",
        "## Missing Data",
        "```{r}",
        "knitr::kable(results$missing_data)",
        "```",
        "",
        "## Correlation",
        "```{r, fig.height=5, fig.width=6}",
        "if (!is.null(results$plots$correlation_heatmap)) results$plots$correlation_heatmap",
        "```"
      )

      writeLines(report_lines, con = report_path)
      rendered <- rmarkdown::render(report_path, output_dir = results_folder, quiet = TRUE)
      results_list$saved_paths$report <- rendered
    }
  }

  results_list
}

safe_require <- function(pkg) {
  requireNamespace(pkg, quietly = TRUE)
}

save_plot_collection <- function(plot_list, folder) {
  if (is.null(plot_list) || length(plot_list) == 0) {
    return(invisible(NULL))
  }
  dir.create(folder, showWarnings = FALSE, recursive = TRUE)
  for (nm in names(plot_list)) {
    plt <- plot_list[[nm]]
    file_path <- file.path(folder, paste0(nm, ".png"))
    if (inherits(plt, c("gg", "ggplot"))) {
      ggplot2::ggsave(filename = file_path, plot = plt, width = 7, height = 5)
    } else if (inherits(plt, "plotly") && safe_require("htmlwidgets")) {
      htmlwidgets::saveWidget(plt, file = sub("\\.png$", ".html", file_path))
    }
  }
  invisible(NULL)
}

perform_normality_tests <- function(data, numeric_cols, normality_tests) {
  if (length(numeric_cols) == 0) {
    return(NULL)
  }
  results_df <- tibble::tibble(
    Column = character(),
    Test = character(),
    P_Value = numeric(),
    Adjusted_P_Value = numeric(),
    Normal = character()
  )

  for (col in numeric_cols) {
    col_data <- na.omit(data[[col]])
    if (length(col_data) == 0) next

    if ("shapiro" %in% normality_tests && length(col_data) >= 3 && length(col_data) <= 5000) {
      sw <- stats::shapiro.test(col_data)
      results_df <- dplyr::bind_rows(results_df, tibble::tibble(
        Column = col,
        Test = "Shapiro-Wilk",
        P_Value = sw$p.value
      ))
    }

    if ("ks" %in% normality_tests && length(col_data) > 0) {
      ks <- suppressWarnings(stats::ks.test(x = scale(col_data), y = "pnorm"))
      results_df <- dplyr::bind_rows(results_df, tibble::tibble(
        Column = col,
        Test = "Kolmogorov-Smirnov",
        P_Value = ks$p.value
      ))
    }
  }

  if (nrow(results_df) == 0) {
    return(NULL)
  }

  results_df <- dplyr::group_by(results_df, .data$Test) |>
    dplyr::mutate(Adjusted_P_Value = stats::p.adjust(.data$P_Value, method = "BH")) |>
    dplyr::ungroup() |>
    dplyr::mutate(Normal = dplyr::if_else(.data$Adjusted_P_Value < 0.05, "No", "Yes"))

  results_df
}

generate_normality_plots <- function(data, numeric_cols) {
  if (length(numeric_cols) == 0) {
    return(NULL)
  }
  plots <- list()
  for (col in numeric_cols) {
    col_data <- data[[col]]
    if (all(is.na(col_data))) next
    hist_plot <- ggplot2::ggplot(data, ggplot2::aes(sample = .data[[col]])) +
      ggplot2::geom_histogram(ggplot2::aes(x = .data[[col]], y = ..density..), bins = 30, fill = "skyblue", color = "white") +
      ggplot2::theme_minimal() +
      ggplot2::labs(title = paste("Histogram of", col), x = col, y = "Density")

    qq_plot <- ggplot2::ggplot(data, ggplot2::aes(sample = .data[[col]])) +
      ggplot2::stat_qq() +
      ggplot2::stat_qq_line(color = "red") +
      ggplot2::theme_minimal() +
      ggplot2::labs(title = paste("Q-Q plot of", col), x = "Theoretical Quantiles", y = "Sample Quantiles")

    plots[[col]] <- list(histogram = hist_plot, qq = qq_plot)
  }
  plots
}

build_heatmap_plot <- function(correlation_matrix, interactive = FALSE) {
  if (is.null(correlation_matrix)) {
    return(NULL)
  }
  cor_df <- reshape2::melt(correlation_matrix, varnames = c("Var1", "Var2"), value.name = "Correlation")
  static_plot <- ggplot2::ggplot(cor_df, ggplot2::aes(x = Var1, y = Var2, fill = Correlation)) +
    ggplot2::geom_tile(color = "white") +
    ggplot2::scale_fill_gradient2(low = "#2563eb", high = "#dc2626", mid = "white", midpoint = 0, limits = c(-1, 1)) +
    ggplot2::coord_equal() +
    ggplot2::theme_minimal() +
    ggplot2::theme(axis.text.x = ggplot2::element_text(angle = 45, hjust = 1)) +
    ggplot2::labs(title = "Correlation heatmap", x = NULL, y = NULL)

  if (interactive && safe_require("plotly")) {
    return(plotly::ggplotly(static_plot))
  }
  static_plot
}

build_missingness_heatmap <- function(data) {
  missing_long <- dplyr::mutate(data, .row = dplyr::row_number()) |>
    tidyr::pivot_longer(
      cols = - .row,
      names_to = "variable",
      values_to = "value",
      values_transform = list(value = as.character)
    ) |>
    dplyr::mutate(missing = is.na(.data$value))

  ggplot2::ggplot(missing_long, ggplot2::aes(x = variable, y = .row, fill = missing)) +
    ggplot2::geom_tile(color = "white") +
    ggplot2::scale_fill_manual(values = c(`TRUE` = "#ef4444", `FALSE` = "#16a34a"), name = "Missing") +
    ggplot2::labs(title = "Missingness heatmap", x = NULL, y = "Row") +
    ggplot2::theme_minimal() +
    ggplot2::theme(axis.text.x = ggplot2::element_text(angle = 45, hjust = 1))
}

detect_outliers_iqr <- function(x) {
  q1 <- stats::quantile(x, 0.25, na.rm = TRUE)
  q3 <- stats::quantile(x, 0.75, na.rm = TRUE)
  iqr_val <- q3 - q1
  lower_bound <- q1 - 1.5 * iqr_val
  upper_bound <- q3 + 1.5 * iqr_val
  proportion <- mean(x < lower_bound | x > upper_bound, na.rm = TRUE)
  proportion
}

detect_outliers_zscore <- function(x, threshold = 3) {
  z <- scale(x)
  proportion <- mean(abs(z) > threshold, na.rm = TRUE)
  proportion
}

detect_outliers_dbscan <- function(data, eps = 0.5, minPts = 5) {
  db_res <- dbscan::dbscan(data, eps = eps, minPts = minPts)
  outlier_idx <- which(db_res$cluster == 0)
  length(outlier_idx) / nrow(data)
}

detect_outliers_lof <- function(data, minPts = 5, threshold = 1.5) {
  lof_vals <- dbscan::lof(data, minPts = minPts)
  outlier_idx <- which(lof_vals > threshold)
  length(outlier_idx) / nrow(data)
}

perform_outlier_detection <- function(data, numeric_cols, outlier_method) {
  if (length(numeric_cols) == 0) {
    return(NULL)
  }

  if (outlier_method == "iqr") {
    return(vapply(data[numeric_cols], detect_outliers_iqr, numeric(1)))
  }

  if (outlier_method == "zscore") {
    return(vapply(data[numeric_cols], detect_outliers_zscore, numeric(1)))
  }

  if (outlier_method %in% c("dbscan", "lof")) {
    if (!safe_require("dbscan")) {
      message(sprintf("Package 'dbscan' not installed; cannot run %s-based outlier detection.", outlier_method))
      return(NULL)
    }
    numeric_data <- data[, numeric_cols, drop = FALSE]
    numeric_data <- na.omit(numeric_data)
    if (nrow(numeric_data) == 0) {
      return(NULL)
    }
    if (outlier_method == "dbscan") {
      return(detect_outliers_dbscan(numeric_data))
    }
    if (outlier_method == "lof") {
      return(detect_outliers_lof(numeric_data))
    }
  }

  NULL
}

library(testthat)
library(dplyr)

# Prevent Rplots.pdf from being created by graphics calls
grDevices::pdf(file = NULL)
withr::defer(grDevices::dev.off())

# small dataset for tests
set.seed(123)
data(iris)

test_that("fastexplore errors with non-data.frame input", {
  expect_error(
    fastexplore("not a data frame"),
    "'data' must be a data.frame"
  )
})

test_that("fastexplore errors with empty data", {
  expect_error(
    fastexplore(iris[0, ]),
    "'data' has no rows to explore"
  )
})

test_that("fastexplore errors with missing label column", {
  expect_error(
    fastexplore(iris, label = "nonexistent"),
    "Label column 'nonexistent' is not present in data"
  )
})

test_that("fastexplore returns expected structure", {
  result <- fastexplore(iris, label = "Species", pairwise_matrix = FALSE)

  expect_true(is.list(result))
  expect_true("data_overview" %in% names(result))
  expect_true("summary_stats" %in% names(result))
  expect_true("missing_data" %in% names(result))
  expect_true("correlation_matrix" %in% names(result))
  expect_true("plots" %in% names(result))
})

test_that("fastexplore data_overview contains expected elements", {
  result <- fastexplore(iris, pairwise_matrix = FALSE)

  expect_true("dimensions" %in% names(result$data_overview))
  expect_true("types" %in% names(result$data_overview))
  expect_true("unique_values" %in% names(result$data_overview))
  expect_true("head" %in% names(result$data_overview))

  expect_equal(result$data_overview$dimensions$Rows, 150)
  expect_equal(result$data_overview$dimensions$Columns, 5)
})

test_that("fastexplore computes summary statistics", {
  result <- fastexplore(iris, pairwise_matrix = FALSE)

  expect_true(!is.null(result$summary_stats))
  expect_true("Column" %in% names(result$summary_stats))
  expect_true("Mean" %in% names(result$summary_stats))
  expect_true("Median" %in% names(result$summary_stats))
  expect_true("SD" %in% names(result$summary_stats))
})

test_that("fastexplore detects missing data", {
  iris_with_na <- iris
  iris_with_na$Sepal.Length[1:10] <- NA

  result <- fastexplore(iris_with_na, pairwise_matrix = FALSE, use_upset_missing = FALSE)

  missing_row <- result$missing_data[result$missing_data$Column == "Sepal.Length", ]
  expect_equal(missing_row$Missing[[1]], 10)
})

test_that("fastexplore computes correlation matrix", {
  result <- fastexplore(iris, pairwise_matrix = FALSE)

  expect_true(!is.null(result$correlation_matrix))
  expect_true(is.matrix(result$correlation_matrix))
  expect_equal(nrow(result$correlation_matrix), 4)  # 4 numeric columns
})

test_that("fastexplore detects high correlations", {
  result <- fastexplore(iris, corr_threshold = 0.8, pairwise_matrix = FALSE)

  expect_true(!is.null(result$high_corr_pairs))
  # Sepal.Length and Petal.Length should be highly correlated
  expect_true(nrow(result$high_corr_pairs) > 0)
})

test_that("fastexplore generates frequency tables for factors", {
  result <- fastexplore(iris, pairwise_matrix = FALSE)

  expect_true(!is.null(result$freq_tables))
  expect_true("Species" %in% names(result$freq_tables))
  expect_equal(nrow(result$freq_tables$Species), 3)  # 3 species
})

test_that("fastexplore detects class imbalance when label is provided", {
  result <- fastexplore(iris, label = "Species", pairwise_matrix = FALSE)

  expect_true(!is.null(result$class_imbalance))
  expect_equal(nrow(result$class_imbalance), 3)
  expect_equal(sum(result$class_imbalance$count), 150)
})

test_that("fastexplore detects duplicated rows", {
  # Create a dataset without pre-existing duplicates
  df_base <- data.frame(
    a = 1:100,
    b = rnorm(100),
    c = letters[rep(1:10, 10)]
  )
  df_with_dups <- rbind(df_base, df_base[1:5, ])

  result <- fastexplore(df_with_dups, pairwise_matrix = FALSE)

  expect_equal(result$duplicated_rows, 5)
  expect_true(!is.null(result$duplicated_examples))
})

test_that("fastexplore performs IQR outlier detection", {
  result <- fastexplore(iris, outlier_method = "iqr", pairwise_matrix = FALSE)

  expect_true(!is.null(result$outlier_summary))
  expect_true(is.numeric(result$outlier_summary))
})

test_that("fastexplore performs zscore outlier detection", {
  result <- fastexplore(iris, outlier_method = "zscore", pairwise_matrix = FALSE)

  expect_true(!is.null(result$outlier_summary))
  expect_true(is.numeric(result$outlier_summary))
})

test_that("fastexplore runs normality tests", {
  result <- fastexplore(iris, run_distribution_checks = TRUE, pairwise_matrix = FALSE)

  expect_true(!is.null(result$normality_tests))
  expect_true("Column" %in% names(result$normality_tests))
  expect_true("P_Value" %in% names(result$normality_tests))
})

test_that("fastexplore generates histogram plots", {
  result <- fastexplore(iris, visualize = "histogram", pairwise_matrix = FALSE)

  expect_true(!is.null(result$plots$histograms))
  expect_true(length(result$plots$histograms) == 4)  # 4 numeric columns
})

test_that("fastexplore generates boxplot plots", {
  result <- fastexplore(iris, visualize = "boxplot", pairwise_matrix = FALSE)

  expect_true(!is.null(result$plots$boxplots))
  expect_true(length(result$plots$boxplots) == 4)
})

test_that("fastexplore generates bar plots for factors", {
  result <- fastexplore(iris, visualize = "barplot", pairwise_matrix = FALSE)

  expect_true(!is.null(result$plots$barplots))
  expect_true("Species" %in% names(result$plots$barplots))
})

test_that("fastexplore generates correlation heatmap", {
  result <- fastexplore(iris, visualize = "heatmap", pairwise_matrix = FALSE)

  expect_true(!is.null(result$plots$correlation_heatmap))
  expect_true(inherits(result$plots$correlation_heatmap, c("gg", "ggplot")))
})

test_that("fastexplore generates scatterplots", {
  result <- fastexplore(iris, visualize = "scatterplot", pairwise_matrix = FALSE)

  expect_true(!is.null(result$plots$scatterplots))
})

test_that("fastexplore respects sample_size parameter", {
  result <- fastexplore(iris, sample_size = 50, pairwise_matrix = FALSE)

  # The result should still work

  expect_true(!is.null(result$data_overview))
})

test_that("fastexplore auto_convert_numeric converts numeric-like factors", {
  df <- data.frame(
    a = factor(c("1", "2", "3", "4", "5", "6", "7", "8")),
    b = c(1, 2, 3, 4, 5, 6, 7, 8)
  )

  result <- fastexplore(df, auto_convert_numeric = TRUE, pairwise_matrix = FALSE)

  # Column 'a' should be converted and appear in summary_stats
  expect_true("a" %in% result$summary_stats$Column)
})

test_that("fastexplore auto_convert_dates converts date-like strings", {
  df <- data.frame(
    date_col = c("2023-01-01", "2023-01-02", "2023-01-03"),
    value = c(1, 2, 3)
  )

  result <- fastexplore(df, auto_convert_dates = TRUE, pairwise_matrix = FALSE)

  # The function should run without error
  expect_true(!is.null(result))
})

test_that("fastexplore feature_engineering creates date features", {
  df <- data.frame(
    date_col = as.Date(c("2021-01-15", "2022-06-20", "2023-12-25")),
    value = c(1, 2, 3)
  )

  result <- fastexplore(df, feature_engineering = TRUE, pairwise_matrix = FALSE, run_distribution_checks = FALSE)

  # date features should be created
  expect_true("date_col_day" %in% result$data_overview$types$Column)
  expect_true("date_col_month" %in% result$data_overview$types$Column)
  expect_true("date_col_year" %in% result$data_overview$types$Column)
})

test_that("fastexplore detects zero variance columns", {
  df <- data.frame(
    constant = rep(5, 100),
    variable = rnorm(100)
  )

  # Suppress expected warning about zero standard deviation in correlation
  result <- suppressWarnings(fastexplore(df, pairwise_matrix = FALSE, run_distribution_checks = FALSE))

  expect_true("constant" %in% result$zero_variance_cols)
})

test_that("fastexplore detects potential ID columns", {
  df <- data.frame(
    id = 1:100,
    value = rnorm(100)
  )

  result <- fastexplore(df, pairwise_matrix = FALSE)

  expect_true("id" %in% result$potential_id_cols)
})

test_that("fastexplore generates grouped plots when label is factor", {
  result <- fastexplore(
    iris,
    label = "Species",
    visualize = c("histogram", "boxplot"),
    grouped_plots = TRUE,
    pairwise_matrix = FALSE
  )

  # Plots should exist
  expect_true(!is.null(result$plots$histograms))
  expect_true(!is.null(result$plots$boxplots))
})

#' Generate counterfactual explanations for a fastml model
#'
#' Uses DALEX ceteris-paribus profiles (`predict_profile`) to compute
#' counterfactual-style what-if explanations for a given observation.
#'
#' @param object A `fastml` object.
#' @param observation A single observation (data frame with one row) to compute
#'   counterfactuals for.
#' @param variables Optional character vector of candidate variables to vary.
#'   Only numeric variables are used for counterfactual profiling.
#' @param data Character string specifying which data to use for the explainer background:
#'   \code{"train"} (default) or \code{"test"}.
#' @param positive_class Optional string used to filter lines/points in the
#'   resulting profiles for classification tasks.
#' @param event_class Optional event class indicator propagated from
#'   `fastml_prepare_explainer_inputs()` (kept for compatibility).
#' @param label_levels Optional vector of label levels propagated from
#'   `fastml_prepare_explainer_inputs()` (kept for compatibility).
#' @param ... Additional arguments passed to `DALEX::predict_profile`.
#'
#' @return A list (returned invisibly) containing the DALEX profile, filtered
#'   lines/points when `positive_class` is supplied, and the plotted object if
#'   rendering succeeds.
#' @export
counterfactual_explain <- function(object,
                                   observation,
                                   variables = NULL,
                                   data = c("train", "test"),
                                   positive_class = NULL,
                                   event_class = NULL,
                                   label_levels = NULL,
                                   ...) {
  if (missing(observation)) {
    stop("'observation' must be provided for counterfactual explanations.")
  }
  if (nrow(observation) != 1) {
    stop("'observation' must contain exactly one row.")
  }
  if (!requireNamespace("DALEX", quietly = TRUE)) {
    stop("The 'DALEX' package is required for counterfactual explanations.")
  }

  data <- match.arg(data)
  obs_raw <- observation
  explainer <- NULL
  prep <- NULL

  # 1. Build/Extract Explainer
  if (inherits(object, "fastml")) {
    prep <- fastml_prepare_explainer_inputs(object, data = data)
    # align observation with explainer input schema (raw predictors, no label)
    if (!is.null(prep$label) && prep$label %in% names(obs_raw)) {
      obs_raw[[prep$label]] <- NULL
    }
    expected_cols <- colnames(prep$x_raw)
    missing_cols <- setdiff(expected_cols, names(obs_raw))
    if (length(missing_cols)) {
      stop(
        "The observation is missing required columns: ",
        paste(missing_cols, collapse = ", "),
        call. = FALSE
      )
    }
    # keep only expected columns and coerce types to match training schema
    obs_raw <- obs_raw[, expected_cols, drop = FALSE]
    obs_raw <- as.data.frame(obs_raw, stringsAsFactors = FALSE)
    for (nm in expected_cols) {
      ref_col <- prep$x_raw[[nm]]
      if (is.factor(ref_col)) {
        obs_raw[[nm]] <- factor(obs_raw[[nm]], levels = levels(ref_col))
      } else if (is.logical(ref_col)) {
        obs_raw[[nm]] <- as.logical(obs_raw[[nm]])
      } else if (inherits(ref_col, "Date")) {
        obs_raw[[nm]] <- as.Date(obs_raw[[nm]])
      } else if (inherits(ref_col, "POSIXt")) {
        obs_raw[[nm]] <- as.POSIXct(obs_raw[[nm]])
      } else if (is.integer(ref_col)) {
        obs_raw[[nm]] <- as.integer(obs_raw[[nm]])
      } else if (is.numeric(ref_col)) {
        obs_raw[[nm]] <- as.numeric(obs_raw[[nm]])
      } else if (is.character(ref_col)) {
        obs_raw[[nm]] <- as.character(obs_raw[[nm]])
      }
    }
    expl_list <- fastml_build_dalex_explainers(prep)$explainers
    explainer <- expl_list[[1]]
  } else if (inherits(object, "explainer")) {
    explainer <- object
  } else {
    stop("`object` must be either a fastml object or a DALEX explainer.")
  }

  # 2. Filter Variables: Keep ONLY Numerics
  all_numeric <- names(explainer$data)[sapply(explainer$data, is.numeric)]

  final_vars <- all_numeric
  if (!is.null(variables)) {
    # Intersect user request with allowed numeric variables
    final_vars <- intersect(variables, all_numeric)
    if (length(final_vars) == 0) {
      warning("No numeric variables found in the requested 'variables' list.")
    }
  }

  # Generate Profile
  profile <- DALEX::predict_profile(
    explainer = explainer,
    new_observation = obs_raw,
    variables = final_vars, # Use the filtered list
    ...
  )

  # If specific variables were requested, keep only those rows (defensive)
  if (length(final_vars) > 0 && "_vname_" %in% names(profile)) {
    profile <- profile[profile$`_vname_` %in% final_vars, , drop = FALSE]
  }

  # --- FIX: ROBUST FILTERING OF LINES AND DOTS ---
  if (!is.null(positive_class)) {
    # A. Capture attributes BEFORE subsetting the main object
    obs_attr <- attr(profile, "observation") # legacy name (unused by ingredients)
    obs_attr_plural <- attr(profile, "observations") # actual name used by ingredients

    # B. Define filters
    matches_lines <- grepl(positive_class, profile$`_label_`, fixed = TRUE)

    # C. Filter the Lines (Main Object)
    if (any(matches_lines)) {
      profile <- profile[matches_lines, ]

      # D. Filter the points attached via attributes
      if (!is.null(obs_attr_plural) && "_label_" %in% names(obs_attr_plural)) {
        matches_dots <- grepl(positive_class, obs_attr_plural$`_label_`, fixed = TRUE)
        obs_attr_plural <- obs_attr_plural[matches_dots, , drop = FALSE]
      }
      if (!is.null(obs_attr) && "_label_" %in% names(obs_attr)) {
        matches_dots_legacy <- grepl(positive_class, obs_attr$`_label_`, fixed = TRUE)
        obs_attr <- obs_attr[matches_dots_legacy, , drop = FALSE]
      }

      # E. Re-attach filtered attributes (ingredients uses "observations")
      if (!is.null(obs_attr_plural)) {
        attr(profile, "observations") <- obs_attr_plural
      }
      if (!is.null(obs_attr)) {
        attr(profile, "observation") <- obs_attr
      }
    }
  }

  # 4. Plot (UPDATED: Added specific Y-axis label)
  plot_obj <- tryCatch(
    suppressWarnings(
      plot(profile, variables = if (length(final_vars) > 0) final_vars else NULL) +
        ggplot2::labs(y = paste("Predicted Probability of", positive_class))
    ),
    error = function(e) NULL
  )

  if (!is.null(plot_obj)) print(plot_obj)

  invisible(list(counterfactuals = profile, profile = profile, plot = plot_obj))
}

#' Generate DALEX explanations for a fastml model
#'
#' Creates a DALEX explainer and computes permutation based variable importance,
#' partial dependence (model profiles) and Shapley values.
#'
#' @inheritParams fastexplain
#' @return Invisibly returns a list with variable importance, optional model profiles and SHAP values.
#' @export
explain_dalex <- function(object,
                          data = c("train", "test"),
                          features = NULL,
                          grid_size = 20,
                          shap_sample = 5,
                          vi_iterations = 10,
                          seed = 123,
                          loss_function = NULL) {
  data <- match.arg(data)
  prep <- fastml_prepare_explainer_inputs(object, data = data)
  explainers <- fastml_build_dalex_explainers(prep)$explainers
  explain_dalex_internal(
    explainers = explainers,
    prep = prep,
    features = features,
    grid_size = grid_size,
    shap_sample = shap_sample,
    vi_iterations = vi_iterations,
    seed = seed,
    loss_function = loss_function
  )
}

#' @keywords internal
fastml_build_dalex_explainers <- function(prep, classification_output = c("prob", "logit")) {
  if (!requireNamespace("DALEX", quietly = TRUE)) {
    stop("The 'DALEX' package is required.")
  }
  if (!requireNamespace("ggplot2", quietly = TRUE)) {
    stop("The 'ggplot2' package is required for plotting.")
  }

  classification_output <- match.arg(classification_output)
  model_info <- if (prep$task == "classification") list(type = "classification") else list(type = "regression")

  normalize_class_name <- function(x) {
    tolower(trimws(as.character(x)))
  }

  resolve_positive_candidates <- function() {
    candidates <- character()
    if (!is.null(prep$positive_class)) {
      candidates <- c(candidates, prep$positive_class)
    }
    if (!is.null(prep$label_levels) && length(prep$label_levels) >= 2) {
      if (!is.null(prep$event_class) && prep$event_class %in% c("first", "second")) {
        idx <- if (prep$event_class == "second") min(2L, length(prep$label_levels)) else 1L
        candidates <- c(candidates, prep$label_levels[idx])
      } else {
        candidates <- c(candidates, prep$label_levels[min(2L, length(prep$label_levels))])
      }
    }
    if (!is.null(prep$label_levels)) {
      candidates <- c(candidates, prep$label_levels)
    }
    unique(candidates[!is.na(candidates) & nzchar(candidates)])
  }

  pos_candidates <- resolve_positive_candidates()
  pos_class <- if (length(pos_candidates)) pos_candidates[[1]] else NULL
  warned_missing_prob <- FALSE

  # --- ROBUST BAKING HELPER ---
  # This function ensures data is properly preprocessed before prediction.
  # It handles three scenarios:
  # 1. No preprocessor exists -> return data as-is
  # 2. Baking succeeds -> return baked data
  # 3. Baking fails -> validate if data is already processed, error if not
  bake_newdata <- function(newdata) {
    if (is.null(prep$preprocessor)) return(newdata)

    processed_cols <- colnames(prep$x_processed)

    # Always attempt to bake when preprocessor exists - column names matching
    # does NOT mean data is already transformed (e.g., centering/scaling keeps
    # names but changes values)
    bake_error <- NULL
    baked <- tryCatch(
      recipes::bake(prep$preprocessor, new_data = newdata),
      error = function(e) {
        bake_error <<- e$message
        NULL
      }
    )

    # If baking failed, we need to validate whether input can be safely used
    if (is.null(baked)) {
      input_cols <- colnames(newdata)

      # Check if input has all expected processed columns
      if (is.null(processed_cols) || !all(processed_cols %in% input_cols)) {
        missing <- if (!is.null(processed_cols)) setdiff(processed_cols, input_cols) else character(0)
        stop(
          "Preprocessing failed: ", bake_error, "\n",
          "Input data is missing expected columns: ",
          paste(utils::head(missing, 5), collapse = ", "),
          if (length(missing) > 5) paste0(" ... and ", length(missing) - 5, " more") else "",
          "\nCannot safely generate explanations with incompatible data."
        )
      }

      # Validate column types match expected processed data types
      type_mismatches <- character(0)
      for (col in processed_cols) {
        if (col %in% colnames(prep$x_processed)) {
          expected_class <- class(prep$x_processed[[col]])[1]
          actual_class <- class(newdata[[col]])[1]
          # Allow numeric/integer/double interchangeability
          numeric_types <- c("numeric", "integer", "double")
          if (expected_class %in% numeric_types && actual_class %in% numeric_types) {
            next
          }
          if (expected_class != actual_class) {
            type_mismatches <- c(type_mismatches,
              sprintf("%s (expected %s, got %s)", col, expected_class, actual_class))
          }
        }
      }

      if (length(type_mismatches) > 0) {
        stop(
          "Preprocessing failed: ", bake_error, "\n",
          "Input data column types do not match expected processed data:\n",
          paste("  -", utils::head(type_mismatches, 5), collapse = "\n"),
          if (length(type_mismatches) > 5) paste0("\n  ... and ", length(type_mismatches) - 5, " more") else "",
          "\nCannot safely generate explanations with incompatible data."
        )
      }

      # Data appears to already be in processed format - use with warning
      warning(
        "Preprocessing failed (", bake_error, "). ",
        "Input data appears to already be processed (has expected columns with compatible types). ",
        "Using input data directly."
      )
      baked <- newdata
    }

    # Remove label column if present
    if (!is.null(prep$label) && prep$label %in% names(baked)) {
      baked[[prep$label]] <- NULL
    }

    # Ensure output has exactly the expected columns in the right order
    if (!is.null(processed_cols)) {
      missing_cols <- setdiff(processed_cols, colnames(baked))
      if (length(missing_cols) > 0) {
        for (col in missing_cols) baked[[col]] <- 0
      }
      baked <- baked[, processed_cols, drop = FALSE]
    }

    baked
  }

  # --- PREDICT FUNCTION ---
  logit_transform <- function(p) {
    eps <- 1e-7
    p <- pmin(pmax(p, eps), 1 - eps)
    log(p / (1 - p))
  }

  predict_function <- function(m, newdata) {
    newdata_processed <- bake_newdata(newdata)

    to_numeric_vec <- function(x) {
      if (is.null(x)) return(NULL)
      drop_if_empty <- function(v) if (length(v) == 0) NULL else as.numeric(v)

      if (is.numeric(x)) return(drop_if_empty(x))
      if (is.matrix(x) && is.numeric(x)) {
        return(drop_if_empty(x[, 1, drop = TRUE]))
      }
      if (is.data.frame(x)) {
        numeric_cols <- names(x)[sapply(x, function(col) is.numeric(col) && !is.list(col))]
        if (length(numeric_cols)) return(drop_if_empty(x[[numeric_cols[1]]]))
        list_cols <- names(x)[sapply(x, is.list)]
        if (length(list_cols)) {
          lst <- x[[list_cols[1]]]
          vec <- vapply(lst, function(el) if (is.null(el) || length(el) == 0) NA_real_ else as.numeric(el[[1]]), numeric(1))
          return(drop_if_empty(vec))
        }
      }
      if (is.list(x)) {
        if (length(x) == 0) return(NULL)
        if (all(vapply(x, function(el) is.numeric(el) && length(el) == 1, logical(1)))) {
          return(drop_if_empty(unlist(x, use.names = FALSE)))
        }
        if (all(vapply(x, function(el) is.numeric(el) && length(el) > 0, logical(1)))) {
          vec <- vapply(x, function(el) as.numeric(el[[1]]), numeric(1))
          return(drop_if_empty(vec))
        }
        flat <- unlist(x, recursive = TRUE, use.names = FALSE)
        if (is.numeric(flat)) return(drop_if_empty(flat))
      }
      NULL
    }

    if (prep$task == "classification") {
      p <- predict(m, new_data = newdata_processed, type = "prob")
      colnames(p) <- sub("^\\.pred_", "", colnames(p))
      if (ncol(p) > 2) {
        p_mat <- as.matrix(p)
        if (classification_output == "logit") {
          p_mat <- logit_transform(p_mat)
        }
        return(p_mat)
      }

      p <- as.data.frame(p)
      prob_cols <- colnames(p)
      match_idx <- NA_integer_
      chosen_class <- NULL

      if (length(pos_candidates)) {
        for (cand in pos_candidates) {
          idx <- match(normalize_class_name(cand), normalize_class_name(prob_cols))
          if (!is.na(idx)) {
            match_idx <- idx
            chosen_class <- prob_cols[idx]
            break
          }
        }
      }

      if (is.na(match_idx) && length(prob_cols) == 2) {
        fallback_idx <- if (!is.null(prep$event_class) && prep$event_class == "second") 2L else 1L
        match_idx <- min(fallback_idx, length(prob_cols))
        chosen_class <- prob_cols[match_idx]
      }

      if (is.na(match_idx)) {
        match_idx <- 1L
        chosen_class <- prob_cols[match_idx]
      }

      if (!warned_missing_prob &&
          length(pos_candidates) &&
          all(is.na(match(normalize_class_name(pos_candidates), normalize_class_name(prob_cols))))) {
        warning(
          "Positive class '", pos_candidates[[1]], "' not found in predicted probabilities; using '",
          chosen_class, "' instead."
        )
        warned_missing_prob <<- TRUE
      }

      p_vec <- p[[match_idx]]
      if (classification_output == "logit") {
        p_vec <- logit_transform(p_vec)
      }
      return(p_vec)

    } else if (prep$task == "survival") {
      try_type <- function(tp) {
        tryCatch(predict(m, new_data = newdata_processed, type = tp), error = function(e) NULL)
      }
      pred_vec <- NULL
      for (tp in c("time", "numeric", "risk", "linear_pred", "raw", "survival", "hazard")) {
        candidate <- to_numeric_vec(try_type(tp))
        if (!is.null(candidate)) {
          pred_vec <- candidate
          break
        }
      }
      if (is.null(pred_vec)) {
        pred_vec <- to_numeric_vec(tryCatch(predict(m, new_data = newdata_processed), error = function(e) NULL))
      }
      if (is.null(pred_vec)) stop("Unable to compute survival predictions.")
      return(pred_vec)
    } else {
      p <- predict(m, new_data = newdata_processed, type = "numeric")
      return(as.numeric(p$.pred))
    }
  }

  build_one <- function(model, name) {
    # --- FIX 2: Correctly Extract and Coerce Target for DALEX/modelStudio ---
    y_target <- prep$y_raw
    if (prep$task == "survival") {
      if (inherits(y_target, "Surv")) y_target <- y_target[, 1]
      y_target <- as.numeric(y_target)
    } else if (prep$task == "classification") {
      if (is.factor(y_target) || is.character(y_target) || is.logical(y_target)) {
        y_chr <- as.character(y_target)
        pos_for_y <- NULL
        if (!is.null(pos_class) && any(normalize_class_name(y_chr) == normalize_class_name(pos_class))) {
          pos_for_y <- pos_class
        } else if (!is.null(prep$label_levels) && length(prep$label_levels) == 2) {
          if (!is.null(prep$event_class) && prep$event_class == "first") {
            pos_for_y <- prep$label_levels[1]
          } else {
            pos_for_y <- prep$label_levels[2]
          }
        }

        if (!is.null(pos_for_y)) {
          y_target <- as.numeric(normalize_class_name(y_chr) == normalize_class_name(pos_for_y))
        } else {
          y_target <- as.numeric(as.factor(y_chr)) - 1
        }
      }
    } else {
      y_target <- as.numeric(y_target)
    }

    DALEX::explain(
      model = model,
      data = prep$x_raw,
      y = y_target,
      label = name,
      predict_function = predict_function,
      model_info = model_info,
      verbose = FALSE
    )
  }

  explainers <- if (length(prep$fits) == 1) {
    list(build_one(prep$fits[[1]], if (length(prep$model_names)) prep$model_names[[1]] else "model"))
  } else {
    mapply(
      build_one,
      model = prep$fits,
      name = if (length(prep$model_names)) prep$model_names else paste0("model_", seq_along(prep$fits)),
      SIMPLIFY = FALSE
    )
  }

  list(explainers = explainers)
}

#' @keywords internal
explain_dalex_internal <- function(explainers,
                                   prep,
                                   features = NULL,
                                   grid_size = 20,
                                   shap_sample = 5,
                                   vi_iterations = 10,
                                   seed = 123,
                                   loss_function = NULL) {

  vi <- NULL; mp <- NULL; shap <- NULL

  # Validate that user-provided features exist in the data

  # NOTE: Do NOT sanitize feature names here - the data (prep$x_raw) uses original
  # column names. Users should pass feature names exactly as they appear in their data.
  if (!is.null(features)) {
    available_features <- colnames(prep$x_raw)
    invalid_features <- setdiff(features, available_features)
    if (length(invalid_features) > 0) {
      stop(
        sprintf(
          "Feature name(s) not found in data: %s\nAvailable features: %s\nNote: feature names must match exactly (case-sensitive).",
          paste(invalid_features, collapse = ", "),
          paste(utils::head(available_features, 10), collapse = ", "),
          if (length(available_features) > 10) " ..." else ""
        ),
        call. = FALSE
      )
    }
  }

  model_names <- if (length(prep$model_names)) prep$model_names else paste0("model_", seq_along(explainers))

  # --- CRITICAL FIX 1: Safe Loss Function for Factors ---
  # DALEX::loss_cross_entropy requires 0/1 numerics.
  # If we feed it 1/2 (from R factors), the math is wrong.
  if (is.null(loss_function)) {
    if (prep$task == "classification") {
      loss_function <- function(observed, predicted) {
        eps <- 1e-15
        # Multiclass: take prob of the true class per row, then mean cross-entropy
        if (is.data.frame(predicted) || is.matrix(predicted)) {
          pred_mat <- as.matrix(predicted)
          obs_chr <- as.character(observed)
          # try column-name alignment first
          col_idx <- if (!is.null(colnames(pred_mat))) match(obs_chr, colnames(pred_mat)) else rep(NA_integer_, length(obs_chr))
          p_true <- numeric(length(obs_chr))
          for (i in seq_along(obs_chr)) {
            if (!is.na(col_idx[i])) {
              p_true[i] <- pred_mat[i, col_idx[i]]
            } else {
              # fallback: if observed is numeric-like and columns unnamed, treat as column index
              obs_num <- suppressWarnings(as.integer(obs_chr[i]))
              if (!is.na(obs_num) && obs_num >= 1 && obs_num <= ncol(pred_mat)) {
                p_true[i] <- pred_mat[i, obs_num]
              } else {
                p_true[i] <- NA_real_
              }
            }
          }
          p_true <- pmin(pmax(p_true, eps), 1 - eps)
          return(mean(-log(p_true), na.rm = TRUE))
        }
        # Binary: force 0/1 target, mean cross-entropy
        obs_num <- observed
        if (is.factor(obs_num)) obs_num <- as.integer(obs_num) - 1
        obs_num <- as.numeric(obs_num)

        # Get unique non-NA values for validation
        unique_vals <- unique(stats::na.omit(obs_num))
        n_unique <- length(unique_vals)

        # Handle edge cases
        if (n_unique == 0) {
          warning("No valid observed values for loss calculation")
          return(NA_real_)
        }

        if (n_unique > 2) {
          # More than 2 classes but predictions are scalar - shouldn't happen
          # but fall back to RMSE to avoid garbage results
          warning("More than 2 classes detected but predictions are scalar. ",
                  "Using RMSE as fallback loss.")
          return(sqrt(mean((obs_num - as.numeric(predicted))^2, na.rm = TRUE)))
        }

        # Normalize to {0, 1} - handle any 2-value or 1-value coding
        if (n_unique == 2) {
          min_val <- min(unique_vals)
          max_val <- max(unique_vals)
          if (min_val != 0 || max_val != 1) {
            obs_num <- ifelse(obs_num == min_val, 0, 1)
          }
        } else if (n_unique == 1) {
          # Single class present - map to 0 or 1 as appropriate
          if (!unique_vals[1] %in% c(0, 1)) {
            # Arbitrary single value - treat as class 0
            obs_num <- rep(0, length(obs_num))
            obs_num[is.na(observed)] <- NA  # Preserve original NAs
          }
        }

        pred <- pmin(pmax(as.numeric(predicted), eps), 1 - eps)

        # Compute binary cross-entropy with NA handling
        ce_vals <- obs_num * log(pred) + (1 - obs_num) * log(1 - pred)
        if (all(is.na(ce_vals))) {
          warning("All cross-entropy values are NA")
          return(NA_real_)
        }
        -mean(ce_vals, na.rm = TRUE)
      }
    } else {
      loss_function <- DALEX::loss_root_mean_square
    }
  }

  # --- CRITICAL FIX 2: Scientifically Valid SHAP Sampling ---
  # Head (1:5) is biased. We use stratified sampling for classification to ensure
  # minority classes are represented. For regression/survival, use random sampling.
  set.seed(seed)
  n_rows <- nrow(prep$x_raw)
  n_sample <- min(shap_sample, n_rows)

  sample_indices <- if (prep$task == "classification" && !is.null(prep$y_raw) && n_sample > 1) {
    # Stratified sampling: ensure each class is represented proportionally
    y_vec <- as.character(prep$y_raw)
    class_levels <- unique(y_vec)
    n_classes <- length(class_levels)

    if (n_classes > 1 && n_classes <= n_sample) {
      # Calculate proportional allocation per class (at least 1 per class if possible)
      class_counts <- table(y_vec)
      class_props <- as.numeric(class_counts) / sum(class_counts)

      # Allocate samples: at least 1 per class, rest proportionally
      base_allocation <- pmin(1, as.numeric(class_counts))  # 1 per class or class size if smaller
      remaining <- n_sample - sum(base_allocation)

      if (remaining > 0) {
        # Distribute remaining samples proportionally
        extra_allocation <- floor(class_props * remaining)
        # Handle rounding: distribute leftover to largest classes
        leftover <- remaining - sum(extra_allocation)
        if (leftover > 0) {
          order_idx <- order(class_props, decreasing = TRUE)
          for (i in seq_len(leftover)) {
            extra_allocation[order_idx[i]] <- extra_allocation[order_idx[i]] + 1
          }
        }
        final_allocation <- base_allocation + extra_allocation
      } else {
        final_allocation <- base_allocation
      }

      # Sample from each class
      sampled_indices <- integer(0)
      for (i in seq_along(class_levels)) {
        cls <- class_levels[i]
        cls_indices <- which(y_vec == cls)
        n_to_sample <- min(final_allocation[i], length(cls_indices))
        if (n_to_sample > 0) {
          sampled_indices <- c(sampled_indices, sample(cls_indices, n_to_sample))
        }
      }

      # If we somehow got fewer than requested (edge case), fill with random
      if (length(sampled_indices) < n_sample) {
        remaining_indices <- setdiff(seq_len(n_rows), sampled_indices)
        n_fill <- min(n_sample - length(sampled_indices), length(remaining_indices))
        if (n_fill > 0) {
          sampled_indices <- c(sampled_indices, sample(remaining_indices, n_fill))
        }
      }

      sampled_indices
    } else {
      # Single class or more classes than samples: fall back to random
      sample(seq_len(n_rows), n_sample)
    }
  } else {
    # Regression/survival or single sample: random sampling
    sample(seq_len(n_rows), n_sample)
  }

  shap_data <- prep$x_raw[sample_indices, , drop = FALSE]

  abort_shap <- function(message) {
    cond <- structure(
      list(message = message, call = sys.call(-1)),
      class = c("fastml_shap_value_error", "error", "condition")
    )
    stop(cond)
  }

  exp_try <- tryCatch({
    explainer <- if (length(explainers) == 1) explainers[[1]] else NULL
    explainer_list <- if (length(explainers) > 1) explainers else NULL

    cat("\n=== DALEX Variable Importance (with Boxplots) ===\n")
    set.seed(seed)

    # Use the custom safe loss_function defined above
    if (length(explainers) == 1) {
      vi <- DALEX::model_parts(explainer, B = vi_iterations, type = "raw", loss_function = loss_function)
    } else {
      vi <- lapply(explainer_list, function(expl) {
        DALEX::model_parts(expl, B = vi_iterations, type = "raw", loss_function = loss_function)
      })
      names(vi) <- model_names
    }

    # Plotting: Since we fixed the math, values are now Positive Cross Entropy.
    # We can label it clearly. Add explicit whiskers based on permutation draws
    # to ensure variability is visible even when boxplot whiskers collapse.

    # Helper to fix DALEX boxplot linewidth issue (ggplot2 3.4+ uses linewidth, not size)
    fix_boxplot_linewidth <- function(p, linewidth = 0.4, colour = "#371ea3") {
      for (i in seq_along(p$layers)) {
        if (inherits(p$layers[[i]]$geom, "GeomBoxplot")) {
          p$layers[[i]]$aes_params$linewidth <- linewidth
          p$layers[[i]]$aes_params$colour <- colour
        }
      }
      p
    }

    build_vi_plot <- function(vi_obj) {
      if (is.list(vi_obj) && !inherits(vi_obj, "data.frame")) {
        do.call(plot, c(vi_obj, list(show_boxplots = TRUE)))
      } else {
        plot(vi_obj, show_boxplots = TRUE)
      }
    }

    p_vi <- build_vi_plot(vi) + ggplot2::labs(y = "Mean Cross Entropy Loss")
    p_vi <- fix_boxplot_linewidth(p_vi)

    vi_df <- if (is.list(vi) && !inherits(vi, "data.frame")) {
      do.call(rbind, vi)
    } else {
      vi
    }

    if (inherits(vi_df, "data.frame") &&
        "permutation" %in% names(vi_df) &&
        "dropout_loss" %in% names(vi_df) &&
        any(vi_df$permutation > 0, na.rm = TRUE)) {
      vi_perm <- vi_df[vi_df$permutation > 0 & !grepl("^_", vi_df$variable), , drop = FALSE]
      if (nrow(vi_perm) > 0) {
        vi_stats <- stats::aggregate(
          dropout_loss ~ variable + label,
          data = vi_perm,
          FUN = function(x) c(min = min(x, na.rm = TRUE), max = max(x, na.rm = TRUE))
        )
        vi_stats <- do.call(data.frame, vi_stats)
        names(vi_stats) <- gsub("^dropout_loss\\.", "", names(vi_stats))

        if (!is.null(p_vi$data$variable) && is.factor(p_vi$data$variable)) {
          vi_stats$variable <- factor(vi_stats$variable, levels = levels(p_vi$data$variable))
        }

        p_vi <- p_vi + ggplot2::geom_linerange(
          data = vi_stats,
          ggplot2::aes(x = variable, ymin = min, ymax = max),
          inherit.aes = FALSE,
          linewidth = 0.4,
          color = "#371ea3"
        )
      }
    }

    suppressWarnings(print(p_vi))

    # ... (Rest of your code for model_profile and SHAP is fine) ...
    if (!is.null(features)) {
      cat("\n=== DALEX Model Profiles (Partial Dependence) ===\n")
      if (length(explainers) == 1) {
        mp <- DALEX::model_profile(explainer, variables = features, N = grid_size)
      } else {
        mp <- lapply(explainer_list, function(expl) {
          DALEX::model_profile(expl, variables = features, N = grid_size)
        })
        names(mp) <- model_names
      }
      suppressWarnings(print(plot(mp)))
    }

    cat("\n=== DALEX Shapley Values (SHAP) ===\n")
    if (length(explainers) == 1) {
      shap <- DALEX::predict_parts(explainer, new_observation = shap_data, type = "shap")
      p_shap <- plot(shap) + ggplot2::labs(title = "SHAP Values")
      p_shap <- fix_boxplot_linewidth(p_shap)
      suppressWarnings(print(p_shap))
    } else {
      shap <- lapply(explainer_list, function(expl) {
        DALEX::predict_parts(expl, new_observation = shap_data, type = "shap")
      })
      names(shap) <- model_names
      invisible(lapply(names(shap), function(nm) {
        p_shap <- plot(shap[[nm]]) + ggplot2::labs(title = paste("SHAP:", nm))
        p_shap <- fix_boxplot_linewidth(p_shap)
        suppressWarnings(print(p_shap))
      }))
    }

  }, error = function(e) {
    if (inherits(e, "fastml_shap_value_error")) {
      stop(e)
    }
    structure(e, class = c("try-error", class(e)))
  })

  if (inherits(exp_try, "try-error")) {
    cat("DALEX explanations not available for this model.\n")
    warning(exp_try)
  }

  invisible(list(variable_importance = vi, model_profiles = mp, shap_values = shap))
}

make_grouped_cv <- function(data, group_cols, v = 5, strata = NULL, repeats = 1) {
  if (is.null(group_cols) || length(group_cols) == 0) {
    stop("'group_cols' must be provided when using grouped resampling.")
  }
  missing_cols <- setdiff(group_cols, colnames(data))
  if (length(missing_cols) > 0) {
    stop(
      sprintf(
        "The following grouping columns were not found in the data: %s",
        paste(missing_cols, collapse = ", ")
      )
    )
  }
  if (!is.numeric(v) || length(v) != 1 || v < 2 || v != as.integer(v)) {
    stop("'folds' must be an integer greater than or equal to 2 for 'grouped_cv'.")
  }
  if (!is.numeric(repeats) || length(repeats) != 1 || repeats <= 0 ||
      repeats != as.integer(repeats)) {
    stop("'repeats' must be a positive integer when using 'grouped_cv'.")
  }

  prepared_data <- data
  if (length(group_cols) > 1) {
    prepared_data <- prepared_data %>%
      dplyr::mutate(
        .fastml_group = interaction(!!!rlang::syms(group_cols), drop = TRUE)
      )
    group_sym <- rlang::sym(".fastml_group")
  } else {
    group_sym <- rlang::sym(group_cols[[1]])
  }

  group_values <- prepared_data[[rlang::as_string(group_sym)]]
  if (any(is.na(group_values))) {
    stop("Grouping columns contain missing values, which are not supported for grouped resampling.")
  }

  n_groups <- dplyr::n_distinct(group_values)
  if (n_groups < v) {
    stop(
      sprintf(
        "'folds' cannot exceed the number of groups for 'grouped_cv'. Found %d unique groups for %d-fold CV.",
        n_groups, v
      )
    )
  }

  strata_sym <- NULL
  if (!is.null(strata)) {
    if (!strata %in% colnames(prepared_data)) {
      warning(
        sprintf(
          "Column '%s' requested for stratification was not found; proceeding without stratification.",
          strata
        )
      )
      strata <- NULL
    } else {
      strata_sym <- rlang::sym(strata)
      groups_per_stratum <- prepared_data %>%
        dplyr::distinct(!!group_sym, !!strata_sym) %>%
        dplyr::count(!!strata_sym, name = "n_groups")
      insufficient <- groups_per_stratum[groups_per_stratum$n_groups < v, , drop = FALSE]
      if (nrow(insufficient) > 0) {
        strata_values <- unique(insufficient[[rlang::as_string(strata_sym)]])
        warning(
          sprintf(
            paste0(
              "Unable to stratify grouped folds because the following strata have fewer than %d groups: %s.",
              " Proceeding without stratification."
            ),
            v,
            paste(strata_values, collapse = ", ")
          )
        )
        strata <- NULL
        strata_sym <- NULL
      }
    }
  } else {
    # Without strata, grouped CV may yield imbalanced class distributions across folds.
    # This is documented here because rsample currently lacks native grouped-stratified
    # splitting when no stratification column is supplied.
  }

  try_grouped_cv <- function(strata_column) {
    if (is.null(strata_column)) {
      rsample::group_vfold_cv(
        prepared_data,
        group = !!group_sym,
        v = v,
        repeats = repeats
      )
    } else {
      rsample::group_vfold_cv(
        prepared_data,
        group = !!group_sym,
        v = v,
        repeats = repeats,
        strata = !!strata_column
      )
    }
  }

  resamples <- tryCatch(
    try_grouped_cv(strata_sym),
    error = function(err) {
      if (!is.null(strata_sym)) {
        warning(
          sprintf(
            paste0(
              "Grouped stratification failed: %s. Proceeding without stratification."
            ),
            conditionMessage(err)
          )
        )
        return(try_grouped_cv(NULL))
      }
      stop(err)
    }
  )

  resamples
}

fastml_tuning_seed <- function(seed, offset = 0L) {
  if (is.null(seed) || length(seed) == 0 || is.na(seed[[1]])) {
    return(NULL)
  }
  base_seed <- as.integer(seed[[1]])
  if (!is.finite(base_seed)) {
    return(NULL)
  }
  offset <- as.integer(offset[[1]])
  if (!is.finite(offset)) {
    offset <- 0L
  }
  seed_out <- base_seed + offset
  seed_out <- abs(seed_out %% .Machine$integer.max)
  if (!is.finite(seed_out) || seed_out == 0L) {
    seed_out <- abs(base_seed)
  }
  seed_out
}

fastml_with_seed <- function(seed, fn) {
  if (is.null(seed) || length(seed) == 0 || is.na(seed[[1]])) {
    return(fn())
  }
  seed <- as.integer(seed[[1]])
  if (!is.finite(seed)) {
    return(fn())
  }
  has_seed <- exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE)
  old_seed <- if (has_seed) get(".Random.seed", envir = .GlobalEnv, inherits = FALSE) else NULL
  on.exit({
    if (has_seed) {
      assign(".Random.seed", old_seed, envir = .GlobalEnv)
    } else if (exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE)) {
      rm(".Random.seed", envir = .GlobalEnv)
    }
  }, add = TRUE)
  set.seed(seed)
  fn()
}

make_blocked_cv <- function(data, block_col, block_size) {
  if (is.null(block_col) || length(block_col) != 1) {
    stop("'block_col' must be provided when using blocked resampling.")
  }
  if (is.null(block_size)) {
    stop("'block_size' must be provided when using blocked resampling.")
  }
  if (!block_col %in% colnames(data)) {
    stop(sprintf("Column '%s' was not found in the data.", block_col))
  }
  if (!is.numeric(block_size) || length(block_size) != 1 || block_size <= 0 ||
      block_size != as.integer(block_size)) {
    stop("'block_size' must be a positive integer.")
  }
  if (any(is.na(data[[block_col]]))) {
    stop(sprintf("Column '%s' contains missing values which are not supported for blocked resampling.", block_col))
  }
  if (!identical(order(data[[block_col]]), seq_len(nrow(data)))) {
    stop(sprintf("Data must be sorted by '%s' in ascending order before applying blocked resampling.", block_col))
  }

  block_ids <- ceiling(seq_len(nrow(data)) / block_size)
  n_blocks <- length(unique(block_ids))
  if (n_blocks < 2) {
    stop("'block_size' results in fewer than two blocks, which is insufficient for blocked resampling.")
  }

  list(
    data = data,
    block_ids = block_ids,
    n_blocks = n_blocks
  )
}

make_rolling_origin_cv <- function(data,
                                   block_col,
                                   initial_window,
                                   assess_window,
                                   skip = 0) {
  if (is.null(block_col) || length(block_col) != 1) {
    stop("'block_col' must be provided when using rolling resampling.")
  }
  if (!block_col %in% colnames(data)) {
    stop(sprintf("Column '%s' was not found in the data.", block_col))
  }
  if (any(is.na(data[[block_col]]))) {
    stop(sprintf("Column '%s' contains missing values which are not supported for rolling resampling.", block_col))
  }
  if (!identical(order(data[[block_col]]), seq_len(nrow(data)))) {
    stop(sprintf("Data must be sorted by '%s' in ascending order before applying rolling resampling.", block_col))
  }

  window_args <- list(
    initial_window = initial_window,
    assess_window = assess_window,
    skip = skip
  )

  if (is.null(window_args$initial_window) || is.null(window_args$assess_window)) {
    stop("Both 'initial_window' and 'assess_window' must be provided for rolling resampling.")
  }

  for (nm in c("initial_window", "assess_window")) {
    value <- window_args[[nm]]
    if (!is.numeric(value) || length(value) != 1 || value <= 0 ||
        value != as.integer(value)) {
      stop(sprintf("'%s' must be a positive integer.", nm))
    }
  }

  if (!is.numeric(window_args$skip) || length(window_args$skip) != 1 ||
      window_args$skip < 0 || window_args$skip != as.integer(window_args$skip)) {
    stop("'skip' must be a non-negative integer.")
  }

  rsample::rolling_origin(
    data,
    initial = window_args$initial_window,
    assess = window_args$assess_window,
    skip = window_args$skip
  )
}

make_nested_cv <- function(data,
                            outer_folds = 5,
                            inner_folds = 5,
                            group_cols = NULL,
                            strata = NULL) {
  if (!is.numeric(outer_folds) || length(outer_folds) != 1 || outer_folds < 2 ||
      outer_folds != as.integer(outer_folds)) {
    stop("'outer_folds' must be an integer greater than or equal to 2.")
  }

  if (!is.numeric(inner_folds) || length(inner_folds) != 1 || inner_folds < 2 ||
      inner_folds != as.integer(inner_folds)) {
    stop("'inner_folds' must be an integer greater than or equal to 2.")
  }

  strata_sym <- if (!is.null(strata)) rlang::sym(strata) else NULL

  outer_resamples <- if (!is.null(group_cols) && length(group_cols) > 0) {
    make_grouped_cv(
      data = data,
      group_cols = group_cols,
      v = outer_folds,
      strata = strata,
      repeats = 1
    )
  } else {
    if (is.null(strata_sym)) {
      rsample::vfold_cv(data, v = outer_folds)
    } else {
      rsample::vfold_cv(data, v = outer_folds, strata = !!strata_sym)
    }
  }

  nested_tbl <- outer_resamples
  nested_tbl$inner_resamples <- vector("list", length(outer_resamples$splits))

  for (i in seq_along(outer_resamples$splits)) {
    outer_analysis <- rsample::analysis(outer_resamples$splits[[i]])
    if (!is.null(group_cols) && length(group_cols) > 0) {
      inner_resamples <- make_grouped_cv(
        data = outer_analysis,
        group_cols = group_cols,
        v = inner_folds,
        strata = strata,
        repeats = 1
      )
    } else {
      if (is.null(strata_sym)) {
        inner_resamples <- rsample::vfold_cv(outer_analysis, v = inner_folds)
      } else {
        inner_resamples <- rsample::vfold_cv(
          outer_analysis,
          v = inner_folds,
          strata = !!strata_sym
        )
      }
    }
    nested_tbl$inner_resamples[[i]] <- inner_resamples
  }

  class(nested_tbl) <- unique(c("nested_cv", class(outer_resamples)))
  nested_tbl
}

fastml_nested_strip_meta <- function(param_tbl) {
  meta_cols <- c(".metric", ".estimator", ".estimate", ".config", ".notes", ".iter", ".order",
                 ".eval_time", "mean", "n", "std_dev", "std_err", "id", "id2", "fold")
  keep <- setdiff(names(param_tbl), meta_cols)
  if (length(keep) == 0) {
    return(NULL)
  }
  tibble::as_tibble(param_tbl[, keep, drop = FALSE])
}

fastml_nested_signature <- function(param_tbl) {
  if (is.null(param_tbl) || nrow(param_tbl) == 0) {
    return(NA_character_)
  }
  vals <- vapply(
    param_tbl,
    function(col) {
      if (is.list(col)) {
        paste(vapply(col, function(x) paste(as.character(x), collapse = ","), character(1)), collapse = ";")
      } else {
        paste(as.character(col), collapse = ",")
      }
    },
    character(1)
  )
  paste(names(vals), vals, sep = "=", collapse = "||")
}

fastml_nested_select_params <- function(inner_results,
                                        metric,
                                        lower_is_better,
                                        summary_fn = mean) {
  if (length(inner_results) == 0) {
    return(NULL)
  }

  metrics_list <- lapply(inner_results, function(entry) {
    if (is.null(entry)) {
      return(NULL)
    }
    metrics_tbl <- NULL
    if (is.data.frame(entry) && all(c(".metric", ".estimate") %in% names(entry))) {
      metrics_tbl <- entry
    } else {
      metrics_tbl <- tryCatch(
        tune::collect_metrics(entry, summarize = FALSE),
        error = function(e) NULL
      )
    }
    if (is.null(metrics_tbl) || nrow(metrics_tbl) == 0) {
      return(NULL)
    }
    metrics_tbl
  })

  metrics_list <- metrics_list[!vapply(metrics_list, is.null, logical(1))]
  if (length(metrics_list) == 0) {
    return(NULL)
  }

  metrics_tbl <- dplyr::bind_rows(metrics_list)
  if (!".metric" %in% names(metrics_tbl) || !".estimate" %in% names(metrics_tbl)) {
    return(NULL)
  }

  metrics_tbl <- metrics_tbl[metrics_tbl$.metric == metric, , drop = FALSE]
  if (nrow(metrics_tbl) == 0) {
    return(NULL)
  }

  param_tbl <- fastml_nested_strip_meta(metrics_tbl)
  if (is.null(param_tbl) || nrow(param_tbl) == 0) {
    return(NULL)
  }

  param_cols <- sort(names(param_tbl))
  group_cols <- param_cols
  has_estimator <- ".estimator" %in% names(metrics_tbl)
  if (has_estimator) {
    group_cols <- c(group_cols, ".estimator")
  }

  aggregated <- metrics_tbl %>%
    dplyr::group_by(dplyr::across(dplyr::all_of(group_cols))) %>%
    dplyr::summarise(
      .estimate = summary_fn(.data$.estimate, na.rm = TRUE),
      n = sum(is.finite(.data$.estimate)),
      std_dev = stats::sd(.data$.estimate, na.rm = TRUE),
      std_err = ifelse(n > 0, std_dev / sqrt(n), NA_real_),
      .groups = "drop"
    )

  aggregated <- aggregated[is.finite(aggregated$.estimate), , drop = FALSE]
  if (nrow(aggregated) == 0) {
    return(NULL)
  }

  params_only <- fastml_nested_strip_meta(aggregated)
  if (is.null(params_only) || nrow(params_only) == 0) {
    return(NULL)
  }
  params_only <- params_only[, param_cols, drop = FALSE]

  signatures <- vapply(
    seq_len(nrow(params_only)),
    function(i) fastml_nested_signature(params_only[i, , drop = FALSE]),
    character(1)
  )
  aggregated$._signature <- signatures

  # Tie-breaking: best aggregate metric, then lower std_err, then estimator/signature ordering.
  if (has_estimator) {
    if (lower_is_better) {
      ordered <- dplyr::arrange(
        aggregated,
        .data$.estimate,
        .data$std_err,
        .data$.estimator,
        .data$._signature
      )
    } else {
      ordered <- dplyr::arrange(
        aggregated,
        dplyr::desc(.data$.estimate),
        .data$std_err,
        .data$.estimator,
        .data$._signature
      )
    }
  } else {
    if (lower_is_better) {
      ordered <- dplyr::arrange(
        aggregated,
        .data$.estimate,
        .data$std_err,
        .data$._signature
      )
    } else {
      ordered <- dplyr::arrange(
        aggregated,
        dplyr::desc(.data$.estimate),
        .data$std_err,
        .data$._signature
      )
    }
  }

  tibble::as_tibble(ordered[1, param_cols, drop = FALSE])
}

fastml_run_nested_cv <- function(workflow_spec,
                                  nested_resamples,
                                  train_data,
                                  label,
                                  task,
                                 metric,
                                 metrics,
                                 engine,
                                 engine_args,
                                 tune_params_template,
                                 engine_tune_params,
                                 tuning_strategy,
                                 tuning_iterations,
                                 adaptive,
                                 early_stopping,
                                 ctrl_grid,
                                 ctrl_bayes,
                                 ctrl_race,
                                 my_metrics,
                                 do_tuning,
                                 event_class,
                                 class_threshold,
                                 start_col,
                                 time_col,
                                 status_col,
                                 eval_times,
                                 at_risk_threshold,
                                 seed,
                                 update_params_fn,
                                 multiclass_auc = "macro",
                                 survival_metric_convention = "fastml",
                                 grid_levels = 3L) {
  if (!inherits(nested_resamples, "rset") ||
      is.null(nested_resamples$inner_resamples)) {
    stop("'nested_resamples' must be a nested resampling object with inner resamples.")
  }

  outer_ids <- nested_resamples$id
  inner_list <- nested_resamples$inner_resamples
  outer_splits <- nested_resamples$splits

  inner_results <- vector("list", length(outer_splits))
  outer_metrics <- vector("list", length(outer_splits))
  best_params_list <- vector("list", length(outer_splits))

  metric_direction <- tryCatch({
    if (requireNamespace("yardstick", quietly = TRUE) &&
        exists(metric, envir = asNamespace("yardstick"), inherits = FALSE)) {
      attr(get(metric, envir = asNamespace("yardstick")), "direction", exact = TRUE)
    } else {
      NA_character_
    }
  }, error = function(e) NA_character_)

  lower_is_better <- if (!is.na(metric_direction) && !is.null(metric_direction)) {
    identical(metric_direction, "minimize")
  } else {
    metric %in% c("rmse", "mae", "ibs", "logloss", "mse", "brier_score", "ece") ||
      grepl("brier", metric, fixed = TRUE)
  }

  for (i in seq_along(outer_splits)) {
    current_split <- outer_splits[[i]]
    if (is.null(current_split)) {
      next
    }

    tuning_seed <- fastml_tuning_seed(seed, offset = i)
    ctrl_bayes_current <- ctrl_bayes
    if (!is.null(tuning_seed) && inherits(ctrl_bayes_current, "control_bayes")) {
      ctrl_bayes_current$seed <- tuning_seed
    }

    inner_resamples <- inner_list[[i]]
    if (do_tuning && (is.null(inner_resamples) || !inherits(inner_resamples, "rset"))) {
      stop("Inner resamples must be an 'rset' object when tuning is enabled.")
    }

    outer_train <- rsample::analysis(current_split)
    outer_assess <- rsample::assessment(current_split)

    current_workflow <- workflow_spec
    best_params <- NULL

    if (do_tuning) {
      params_current <- tune_params_template
      if (!is.null(params_current)) {
        params_current <- finalize(
          params_current,
          x = outer_train %>% dplyr::select(-dplyr::all_of(label))
        )
      }

      if (!is.null(params_current) && !is.null(engine_tune_params)) {
        params_current <- update_params_fn(params_current, engine_tune_params)
      }

      grid_current <- NULL
      if (!is.null(params_current) && nrow(params_current) > 0) {
        if (tuning_strategy == "grid" && !adaptive) {
          grid_current <- grid_regular(params_current, levels = grid_levels)
        }
      }

      # Note: engine_args are already embedded in workflow_spec via set_engine(),
      # so we don't pass them here. Tuning functions don't accept extra arguments.
      tuning_args <- list(
        object = workflow_spec,
        resamples = inner_resamples,
        metrics = if (!is.null(my_metrics)) my_metrics else metrics
      )

      if (tuning_strategy == "bayes") {
        tuning_args$param_info <- params_current
        tuning_args$iter <- tuning_iterations
        tuning_args$control <- ctrl_bayes_current
        model_tuned <- fastml_with_seed(tuning_seed, function() {
          do.call(tune_bayes, tuning_args)
        })
      } else if (adaptive) {
        tuning_args$param_info <- params_current
        tuning_args$grid <- if (is.null(grid_current)) 20 else grid_current
        tuning_args$control <- ctrl_race
        model_tuned <- fastml_with_seed(tuning_seed, function() {
          do.call(tune_race_anova, tuning_args)
        })
      } else {
        grid_arg <- if (!is.null(grid_current)) {
          grid_current
        } else {
          5
        }
        tuning_args$grid <- grid_arg
        tuning_args$control <- ctrl_grid
        model_tuned <- fastml_with_seed(tuning_seed, function() {
          do.call(tune::tune_grid, tuning_args)
        })
      }

      inner_results[[i]] <- model_tuned
      best_params <- tryCatch({
        select_best(model_tuned, metric = metric)
      }, error = function(e) NULL)

      if (!is.null(best_params)) {
        best_params_list[[i]] <- best_params
        current_workflow <- finalize_workflow(workflow_spec, best_params)
      }
    }

    fitted_outer <- tryCatch({
      fastml_with_seed(tuning_seed, function() {
        parsnip::fit(current_workflow, data = outer_train)
      })
    }, error = function(e) {
      warning(sprintf("Outer fit failed for split '%s': %s", outer_ids[[i]], e$message))
      NULL
    })

    if (!is.null(fitted_outer)) {
        eval_result <- tryCatch({
          process_model(
            model_obj = fitted_outer,
            model_id = paste0("outer_", outer_ids[[i]]),
            task = task,
            test_data = outer_assess,
            label = label,
            event_class = event_class,
            class_threshold = class_threshold,
            start_col = start_col,
            time_col = time_col,
            status_col = status_col,
            engine = engine,
            train_data = outer_train,
            metric = metric,
            metrics = if (!is.null(my_metrics)) my_metrics else metrics,
            eval_times_user = eval_times,
            bootstrap_ci = FALSE,
            bootstrap_samples = 1,
            bootstrap_seed = seed,
            at_risk_threshold = at_risk_threshold,
            survival_metric_convention = survival_metric_convention,
            multiclass_auc = multiclass_auc
          )
        }, error = function(e) {
        warning(sprintf("Evaluation failed for outer split '%s': %s", outer_ids[[i]], e$message))
        NULL
      })

      if (!is.null(eval_result) && !is.null(eval_result$performance)) {
        outer_metrics[[i]] <- eval_result$performance %>%
          dplyr::mutate(
            outer_id = outer_ids[[i]],
            .n = nrow(outer_assess)
          )
      }
    }
  }

  outer_metrics <- outer_metrics[!vapply(outer_metrics, is.null, logical(1))]
  outer_performance <- if (length(outer_metrics) > 0) {
    dplyr::bind_rows(outer_metrics)
  } else {
    NULL
  }

  final_params <- NULL
  selected_outer_id <- NULL
  if (do_tuning) {
    final_params <- fastml_nested_select_params(
      inner_results = inner_results,
      metric = metric,
      lower_is_better = lower_is_better
    )
    if (!is.null(final_params) && is.data.frame(final_params) && nrow(final_params) > 0) {
      param_cols <- names(final_params)
      best_score <- if (lower_is_better) Inf else -Inf
      for (oi in seq_along(inner_results)) {
        entry <- inner_results[[oi]]
        if (is.null(entry)) next
        ir_metrics <- tryCatch(
          tune::collect_metrics(entry, summarize = TRUE),
          error = function(e) {
            if (is.data.frame(entry) && ".metric" %in% names(entry) &&
                any(c(".estimate", "mean") %in% names(entry))) {
              entry
            } else {
              NULL
            }
          }
        )
        if (is.null(ir_metrics) || nrow(ir_metrics) == 0) next
        ir_sub <- ir_metrics[ir_metrics$.metric == metric, , drop = FALSE]
        if (nrow(ir_sub) == 0) next
        shared <- intersect(param_cols, names(ir_sub))
        match_rows <- ir_sub
        for (pc in shared) {
          cmp <- match_rows[[pc]] == final_params[[pc]]
          cmp[is.na(cmp)] <- FALSE
          match_rows <- match_rows[cmp, , drop = FALSE]
        }
        if (nrow(match_rows) > 0) {
          # collect_metrics(summarize = TRUE) reports the score in `mean`;
          # `.estimate` appears only in unsummarized (per-resample) output. Read
          # whichever is present rather than assuming one, since indexing a tibble
          # for a missing column returns NULL and silently disables this branch.
          score_col <- intersect(c("mean", ".estimate"), names(match_rows))[1]
          fold_score <- if (!is.na(score_col)) match_rows[[score_col]][1] else NULL
          if (!is.null(fold_score) && length(fold_score) == 1 &&
              is.finite(fold_score) &&
              ((lower_is_better && fold_score < best_score) ||
               (!lower_is_better && fold_score > best_score))) {
            best_score <- fold_score
            selected_outer_id <- outer_ids[[oi]]
          }
        }
      }
    }
  }

  if (do_tuning && (is.null(final_params) || !is.data.frame(final_params) || nrow(final_params) == 0)) {
    params_final <- tune_params_template
    if (!is.null(params_final)) {
      params_final <- tryCatch(
        finalize(
          params_final,
          x = train_data %>% dplyr::select(-dplyr::all_of(label))
        ),
        error = function(e) NULL
      )
    }
    if (!is.null(params_final) && !is.null(engine_tune_params)) {
      params_final <- update_params_fn(params_final, engine_tune_params)
    }
    if (!is.null(params_final) && nrow(params_final) > 0) {
      fallback_grid <- tryCatch(
        grid_regular(params_final, levels = grid_levels),
        error = function(e) NULL
      )
      if (!is.null(fallback_grid) && nrow(fallback_grid) > 0) {
        final_params <- fallback_grid[1, , drop = FALSE]
        warning(
          "Nested tuning failed to select parameters; using the first available grid configuration.",
          call. = FALSE
        )
      }
    }
  }

  final_workflow <- if (!is.null(final_params) && is.data.frame(final_params) && nrow(final_params) > 0) {
    finalize_workflow(workflow_spec, final_params)
  } else {
    workflow_spec
  }

  final_model <- fastml_with_seed(seed, function() {
    parsnip::fit(final_workflow, data = train_data)
  })

  list(
    final_model = final_model,
    details = list(
      inner_results = inner_results,
      outer_performance = outer_performance,
      final_params = final_params,
      selected_outer_id = selected_outer_id
    )
  )
}

fastml_filter_tune_metrics <- function(metrics_tbl, best_params) {
  if (is.null(metrics_tbl) || !is.data.frame(metrics_tbl) || nrow(metrics_tbl) == 0) {
    return(metrics_tbl)
  }
  if (is.null(best_params) || !is.data.frame(best_params) || nrow(best_params) == 0) {
    return(metrics_tbl)
  }

  param_cols <- intersect(names(best_params), names(metrics_tbl))
  if (length(param_cols) == 0) {
    return(metrics_tbl)
  }

  non_param <- c(".metric", ".estimator", "mean", "n", "std_err", ".estimate",
                 "id", "fold", ".iter", ".order")
  param_cols <- setdiff(param_cols, non_param)
  if (length(param_cols) == 0 && ".config" %in% names(best_params) &&
      ".config" %in% names(metrics_tbl)) {
    param_cols <- ".config"
  }
  if (length(param_cols) == 0) {
    return(metrics_tbl)
  }

  dplyr::semi_join(metrics_tbl, best_params, by = param_cols)
}

fastml_attach_fold_sizes <- function(res_summary, resamples, task) {
  if (is.null(res_summary) || !is.list(res_summary)) {
    return(res_summary)
  }
  folds_tbl <- res_summary$folds
  if (is.null(folds_tbl) || !is.data.frame(folds_tbl) || nrow(folds_tbl) == 0) {
    return(res_summary)
  }
  if (".n" %in% names(folds_tbl)) {
    res_summary$aggregated <- fastml_aggregate_resample_metrics(folds_tbl, task)
    return(res_summary)
  }

  splits <- resamples
  if (fastml_is_resample_plan(splits)) {
    splits <- fastml_resample_splits(splits)
  }
  if (is.null(splits) || !inherits(splits, "rset")) {
    return(res_summary)
  }
  split_list <- splits$splits
  if (length(split_list) == 0) {
    return(res_summary)
  }

  fold_sizes <- vapply(split_list, function(split) {
    assess <- tryCatch(rsample::assessment(split), error = function(e) NULL)
    if (is.null(assess)) {
      return(NA_real_)
    }
    nrow(assess)
  }, numeric(1))
  size_tbl <- tibble::tibble(
    fold = as.character(seq_along(fold_sizes)),
    .n = as.numeric(fold_sizes)
  )

  if (!"fold" %in% names(folds_tbl)) {
    folds_tbl$fold <- as.character(seq_len(nrow(folds_tbl)))
  } else {
    folds_tbl$fold <- as.character(folds_tbl$fold)
  }
  folds_tbl <- dplyr::left_join(folds_tbl, size_tbl, by = "fold")
  res_summary$folds <- folds_tbl
  res_summary$aggregated <- fastml_aggregate_resample_metrics(folds_tbl, task)
  res_summary
}

fastml_collect_tune_resample_summary <- function(tune_results, best_params, task) {
  # First, collect per-fold metrics (unsummarized)
  folds_tbl <- tryCatch(
    tune::collect_metrics(tune_results, summarize = FALSE),
    error = function(e) NULL
  )

  if (is.null(folds_tbl) || nrow(folds_tbl) == 0) {
    return(NULL)
  }

  folds_tbl <- fastml_filter_tune_metrics(folds_tbl, best_params)
  if (is.null(folds_tbl) || nrow(folds_tbl) == 0) {
    return(NULL)
  }

  # Rename id to fold for consistency
 if ("id" %in% names(folds_tbl)) {
    folds_tbl <- dplyr::rename(folds_tbl, fold = .data$id)
  }

  # Drop unnecessary columns
  drop_cols <- intersect(names(folds_tbl), c(".config", ".iter", ".order"))
  if (length(drop_cols) > 0) {
    folds_tbl <- folds_tbl[, setdiff(names(folds_tbl), drop_cols), drop = FALSE]
  }

  if (identical(task, "survival") && ".estimator" %in% names(folds_tbl)) {
    folds_tbl <- folds_tbl[, setdiff(names(folds_tbl), ".estimator"), drop = FALSE]
  }

  # Compute aggregated metrics (mean, SD, SE) from fold-level data
  # This ensures proper CV statistics instead of using pooled metrics
  aggregated <- fastml_aggregate_resample_metrics(folds_tbl, task)

  list(aggregated = aggregated, folds = folds_tbl)
}

#' Train Specified Machine Learning Algorithms on the Training Data
#'
#' Trains specified machine learning algorithms on the preprocessed training data.
#'
#' @param train_data Preprocessed training data frame.
#' @param train_data Preprocessed training data frame.
#' @param label Name of the target variable.
#' @param task Type of task: "classification", "regression", or "survival".
#' @param algorithms Vector of algorithm names to train.
#' @param resampling_method Resampling method for cross-validation. Supported
#'   options include standard \code{"cv"}, \code{"repeatedcv"}, and
#'   \code{"boot"}, as well as grouped resampling via \code{"grouped_cv"},
#'   blocked/rolling schemes via \code{"blocked_cv"} or \code{"rolling_origin"},
#'   nested resampling via \code{"nested_cv"}, and the passthrough
#'   \code{"none"} option.
#' @param folds Number of folds for cross-validation.
#' @param repeats Number of times to repeat cross-validation (only applicable for methods like "repeatedcv").
#' @param resamples Optional rsample object. If provided, custom resampling splits
#'   will be used instead of those created internally.
#' @param group_cols Optional character vector of grouping columns used with
#'   `resampling_method = "grouped_cv"`. For classification problems the outcome
#'   column is used to request grouped stratification where supported; if class
#'   imbalance prevents stratification, grouped folds are still created and a
#'   warning is emitted to document the limitation.
#' @param block_col Optional name of the ordering column used with blocked or
#'   rolling resampling.
#' @param block_size Optional integer specifying the block size for
#'   `resampling_method = "blocked_cv"`.
#' @param initial_window Optional integer specifying the initial window size for
#'   rolling resampling.
#' @param assess_window Optional integer specifying the assessment window size
#'   for rolling resampling.
#' @param skip Optional integer number of resamples to skip between rolling
#'   resamples.
#' @param outer_folds Optional integer specifying the number of outer folds for
#'   `resampling_method = "nested_cv"`.
#' @param tune_params A named list of tuning ranges. For each algorithm, supply a
#'   list of engine-specific parameter values, e.g.
#'   \code{list(rand_forest = list(ranger = list(mtry = c(1, 3)))).}
#' @param engine_params A named list of fixed engine-level arguments passed
#'   directly to the model fitting call for each algorithm/engine combination.
#'   Use this to control options like \code{ties = "breslow"} for Cox models or
#'   \code{importance = "impurity"} for ranger. Unlike \code{tune_params}, these
#'   values are not tuned over a grid.
#' @param metric The performance metric to optimize. For classification, options
#'   include \code{"accuracy"}, \code{"roc_auc"}, \code{"logloss"},
#'   \code{"brier_score"}, and \code{"ece"} (plus other class metrics).
#' @param summaryFunction A custom summary function for model evaluation. Default is \code{NULL}.
#' @param seed An integer value specifying the random seed for reproducibility.
#' @param recipe A recipe object for preprocessing.
#' @param use_default_tuning Logical; if \code{TRUE} and \code{tune_params} is \code{NULL}, tuning is performed using default grids. Tuning also occurs when custom \code{tune_params} are supplied. When \code{FALSE} and no custom parameters are given, the model is fitted once with default settings.
#' @param tuning_strategy A string specifying the tuning strategy. Must be one of
#'   \code{"grid"}, \code{"bayes"}, or \code{"none"}. Adaptive methods may be
#'   used with \code{"grid"}. If \code{"none"} is selected, the workflow is fitted
#'   directly without tuning.
#'   If custom \code{tune_params} are supplied with \code{tuning_strategy = "none"},
#'   they will be ignored with a warning.
#' @param tuning_iterations Number of iterations for Bayesian tuning. Ignored
#'   when \code{tuning_strategy} is not \code{"bayes"}; validation occurs only
#'   for the Bayesian strategy.
#' @param tuning_complexity Character string specifying tuning complexity preset.
#'   One of "quick", "balanced", "thorough", or "exhaustive". Controls both
#'   grid density and parameter range width.
#' @param grid_levels Integer specifying number of levels per parameter for
#'   grid search. Higher values create denser grids but increase computation
#'   exponentially (grid size = levels^n_params).
#' @param early_stopping Logical for early stopping in Bayesian tuning.
#' @param adaptive Logical indicating whether to use adaptive/racing methods.
#' @param algorithm_engines A named list specifying the engine to use for each algorithm.
#' @param use_parsnip_defaults Logical. If \code{TRUE}, use parsnip's default engines
#'   instead of fastml's optimized defaults. Default is \code{FALSE}.
#' @param warn_engine_defaults Logical. If \code{TRUE} (default), warn when fastml's
#'   default engine differs from parsnip's default.
#' @param n_cores Integer number of parallel worker processes used to evaluate
#'   resamples and tuning candidates. Sizes the worker pool only, and decides
#'   whether tuning/resampling runs in parallel.
#' @param engine_threads Integer number of threads passed to engines that accept
#'   a thread count (\code{num.threads}, \code{num_threads}, \code{nthread}).
#'   Total CPU demand is approximately \code{n_cores * engine_threads}.
#' @param verbose Logical. If \code{TRUE}, print informational messages about
#'   engine selection and parameter overrides.
#' @param event_class Character string identifying the positive class when computing
#'   classification metrics ("first" or "second").
#' @param class_threshold For binary classification, controls how class probabilities
#'   are converted into hard class predictions during evaluation. The default is `0.5`
#'   (standard threshold). Numeric values in (0, 1) set a fixed threshold. Use `"auto"`
#'   to tune a threshold on training data to maximize F1; use `"model"` to keep the
#'   model's default predictions.
#' @param multiclass_auc For multiclass ROC AUC, the averaging method to use:
#'   `"macro"` (default, tidymodels) or `"macro_weighted"`. Macro weights each
#'   class equally, while macro_weighted weights by class prevalence and can
#'   change model rankings on imbalanced data.
#' @param start_col Optional name of the survival start time column passed through
#'   to downstream evaluation helpers.
#' @param time_col Optional name of the survival stop time column.
#' @param status_col Optional name of the survival status/event column.
#' @param eval_times Optional numeric vector of time horizons for survival metrics.
#' @param at_risk_threshold Numeric cutoff used to determine the evaluation window
#'   for survival metrics within guarded resampling.
#' @param survival_metric_convention Character string specifying which survival
#'   metric conventions to follow. `"fastml"` (default) uses fastml's internal
#'   defaults for evaluation horizons and t_max. `"tidymodels"` uses
#'   `eval_times` as the explicit evaluation grid and applies yardstick-style
#'   Brier/IBS normalization; when `eval_times` is `NULL`, time-dependent Brier
#'   metrics are omitted.
#' @param store_fold_models Logical. If \code{TRUE}, store the fitted fold models
#'   during resampling for later inspection or stability analysis.
#' @param audit_env Internal environment that tracks security audit findings when
#'   custom preprocessing hooks are executed. Typically supplied by
#'   \code{fastml()} and should be left as \code{NULL} when calling
#'   \code{train_models()} directly.
#' @importFrom magrittr %>%
#' @importFrom dplyr filter mutate select if_else starts_with arrange row_number
#'   n_distinct
#' @importFrom tibble tibble
#' @importFrom rlang sym syms call2 call_modify expr as_string
#' @importFrom dials range_set value_set grid_regular grid_latin_hypercube finalize
#' @importFrom parsnip fit extract_parameter_set_dials survival_reg
#' @importFrom workflows workflow add_model add_recipe
#' @importFrom tune tune_grid control_grid select_best finalize_workflow finalize_model tune_bayes control_grid control_bayes
#' @importFrom yardstick metric_set accuracy kap roc_auc sens spec precision f_meas rmse rsq mae new_class_metric
#' @importFrom rsample vfold_cv bootstraps validation_split group_vfold_cv
#'   rolling_origin nested_cv
#' @importFrom recipes all_nominal_predictors all_numeric_predictors all_outcomes all_predictors
#' @importFrom finetune control_race tune_race_anova
#' @return A list of trained model objects.
#' @export
train_models <- function(train_data,
                         label,
                         task,
                         algorithms,
                         resampling_method,
                         folds,
                         repeats,
                         group_cols = NULL,
                         block_col = NULL,
                         block_size = NULL,
                         initial_window = NULL,
                         assess_window = NULL,
                         skip = 0,
                         outer_folds = NULL,
                         resamples = NULL,
                         tune_params,
                         engine_params = list(),
                         metric,
                         summaryFunction = NULL,
                         seed = 123,
                         recipe,
                         use_default_tuning = FALSE,
                         tuning_strategy = "grid",
                         tuning_iterations = 10,
                         tuning_complexity = "balanced",
                         grid_levels = 3L,
                         early_stopping = FALSE,
                         adaptive = FALSE,
                         algorithm_engines = NULL,
                         use_parsnip_defaults = FALSE,
                         warn_engine_defaults = TRUE,
                         n_cores = 1,
                         engine_threads = 1,
                         verbose = FALSE,
                         event_class = "first",
                         class_threshold = 0.5,
                         start_col = NULL,
                         time_col = NULL,
                         status_col = NULL,
                         eval_times = NULL,
                         at_risk_threshold = 0.1,
                         survival_metric_convention = "fastml",
                         audit_env = NULL,
                         multiclass_auc = "macro",
                         store_fold_models = FALSE) {
  survival_metric_convention <- fastml_normalize_survival_convention(survival_metric_convention)

  .fastml_old_seed <- if (exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE))
    get(".Random.seed", envir = .GlobalEnv, inherits = FALSE) else NULL
  .fastml_had_seed <- !is.null(.fastml_old_seed)
  on.exit({
    if (.fastml_had_seed) {
      assign(".Random.seed", .fastml_old_seed, envir = .GlobalEnv)
    } else if (exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE)) {
      rm(".Random.seed", envir = .GlobalEnv)
    }
  }, add = TRUE)
  set.seed(seed)

  # `n_cores` sizes the worker pool; `engine_threads` sizes the thread pool each
  # engine is given. These are deliberately separate: driving both from one
  # argument meant a request for k cores could demand up to k^2 threads and
  # contend with itself on engines that thread aggressively.
  n_cores_val <- fastml_normalize_threads(n_cores)
  engine_threads_val <- fastml_normalize_threads(engine_threads)
  allow_par_base <- !is.null(n_cores_val) && n_cores_val > 1L
  determinism_warnings <- list()
  add_determinism_warning <- function(algo, engine, engine_args) {
    reason <- fastml_engine_determinism_warning(algo, engine, task, engine_threads_val, engine_args)
    if (!is.null(reason)) {
      determinism_warnings[[length(determinism_warnings) + 1]] <<- list(
        algo = algo,
        engine = engine,
        reason = reason
      )
    }
  }

  if (is.null(audit_env) || !is.environment(audit_env)) {
    audit_env <- fastml_init_audit_env(FALSE)
  }

  multiclass_auc <- fastml_normalize_multiclass_auc(multiclass_auc)
  roc_auc <- fastml_configured_roc_auc(multiclass_auc, event_class = event_class)
  accuracy_metric <- fastml_wrap_basic_metric(accuracy)
  kap_metric <- fastml_wrap_basic_metric(kap)
  sens_metric <- fastml_wrap_event_level_metric(sens, event_class)
  spec_metric <- fastml_wrap_event_level_metric(spec, event_class)
  precision_metric <- fastml_wrap_event_level_metric(precision, event_class)
  f_meas_metric <- fastml_wrap_event_level_metric(f_meas, event_class)
  logloss_metric <- fastml_configured_logloss(event_class = event_class)
  brier_metric <- fastml_configured_brier_score(event_class = event_class)
  ece_metric <- fastml_configured_ece(event_class = event_class)

  tuning_strategy <- match.arg(tuning_strategy, c("grid", "bayes", "none"))
  event_class <- match.arg(event_class, c("first", "second"))

  resample_data <- train_data

  supported_resampling <- c(
    "cv",
    "repeatedcv",
    "boot",
    "validation_split",
    "grouped_cv",
    "blocked_cv",
    "rolling_origin",
    "nested_cv",
    "none"
  )

  if (!resampling_method %in% supported_resampling) {
    stop("Unsupported resampling method.")
  }
  if (!is.null(repeats) && length(repeats) == 1 && is.na(repeats)) {
    repeats <- NULL
  }
  if (!is.null(repeats)) {
    if (!is.numeric(repeats) || length(repeats) != 1 ||
        !isTRUE(all.equal(repeats, round(repeats))) || repeats <= 0) {
      stop("'repeats' must be a positive integer when supplied.", call. = FALSE)
    }
    repeats <- as.integer(round(repeats))
  }

  if (tuning_strategy == "bayes" && adaptive) {
    warning("'adaptive' is not supported with Bayesian tuning. Setting adaptive = FALSE.")
    adaptive <- FALSE
  }

  if (tuning_strategy == "none") {
    has_tune_params <- !is.null(tune_params)
    has_default_tuning <- isTRUE(use_default_tuning)
    if (has_tune_params && has_default_tuning) {
      warning("'tune_params' and 'use_default_tuning = TRUE' are ignored because 'tuning_strategy' is 'none'")
    } else if (has_tune_params) {
      warning("'tune_params' are ignored when 'tuning_strategy' is 'none'")
    } else if (has_default_tuning) {
      warning("'use_default_tuning = TRUE' is ignored because 'tuning_strategy' is 'none'")
    }
  }

  if (tuning_strategy == "bayes") {
    if (!is.numeric(tuning_iterations) || length(tuning_iterations) != 1 ||
        tuning_iterations <= 0 || tuning_iterations != as.integer(tuning_iterations)) {
      stop("'tuning_iterations' must be a positive integer")
    }
  }

  if (early_stopping && tuning_strategy != "bayes") {
    message("Engine-level early stopping will be applied when supported")
  }

  if (task == "survival") {
    models <- list()
    failed_models <- list()

    update_params_surv <- function(params_model, new_params) {
      for (param_name in names(new_params)) {
        param_value <- new_params[[param_name]]
        param_row <- params_model %>% dplyr::filter(id == param_name)
        if (nrow(param_row) == 0) next

        param_obj <- param_row$object[[1]]

        try_update <- function(obj, value) {
          if (length(value) == 2) {
            if (inherits(obj, "integer_parameter")) {
              return(obj %>% dials::range_set(c(as.integer(value[1]), as.integer(value[2]))))
            } else {
              return(obj %>% dials::range_set(value))
            }
          } else {
            if (inherits(obj, "integer_parameter")) {
              return(obj %>% dials::value_set(as.integer(value)))
            } else {
              return(obj %>% dials::value_set(value))
            }
          }
        }

        updated_obj <- tryCatch({
          try_update(param_obj, param_value)
        }, error = function(e) {
          current_lb <- attr(param_obj, "range")$lower
          current_ub <- attr(param_obj, "range")$upper

          if (length(param_value) == 1) {
            new_val <- if (inherits(param_obj, "integer_parameter")) as.integer(param_value) else param_value
          } else {
            new_val <- c(min(param_value), max(param_value))
          }

          if (length(new_val) == 1) {
            new_lb <- min(current_lb, new_val)
            new_ub <- max(current_ub, new_val)
          } else {
            new_lb <- min(current_lb, new_val[1])
            new_ub <- max(current_ub, new_val[2])
          }

          param_obj %>% dials::range_set(c(new_lb, new_ub))
        })

        params_model$object[params_model$id == param_name] <- list(updated_obj)
      }
      params_model
    }

    status_strata <- if (!is.null(status_col) && status_col %in% names(resample_data)) {
      status_col
    } else {
      NULL
    }

    resample_plan <- NULL
    if (!is.null(resamples)) {
      if (fastml_is_resample_plan(resamples)) {
        resample_plan <- fastml_resample_validate(resamples)
      } else if (inherits(resamples, "rset")) {
        resample_plan <- fastml_new_resample_plan(
          splits = resamples,
          method = "custom",
          params = list(source = "user_provided")
        )
      } else {
        stop("'resamples' must be an 'rset' object or a fastml resampling plan for survival tasks.")
      }
    } else if (resampling_method == "cv") {
      if (nrow(resample_data) < folds) {
        stop(
          sprintf(
            "You requested %d-fold cross-validation, but your training set only has %d rows. \nThis prevents each fold from having at least one row. \nEither reduce 'folds', increase data, or use a different resampling method (e.g. 'boot').",
            folds,
            nrow(resample_data)
          )
        )
      }
      resamples_obj <- vfold_cv(
        resample_data,
        v = folds,
        repeats = 1,
        strata = if (!is.null(status_strata)) dplyr::all_of(status_strata) else NULL
      )
      resample_plan <- fastml_new_resample_plan(
        splits = resamples_obj,
        method = "cv",
        params = list(
          v = folds,
          repeats = 1,
          strata = status_strata
        )
      )
    } else if (resampling_method == "boot") {
      resamples_obj <- bootstraps(
        resample_data,
        times = folds,
        strata = if (!is.null(status_strata)) dplyr::all_of(status_strata) else NULL
      )
      resample_plan <- fastml_new_resample_plan(
        splits = resamples_obj,
        method = "boot",
        params = list(times = folds, strata = status_strata)
      )
    } else if (resampling_method == "validation_split") {
      strata_ok <- !is.null(status_strata)
      resamples_obj <- rsample::validation_split(
        resample_data,
        prop = 1 - 1 / folds,
        strata = if (strata_ok) dplyr::all_of(status_strata) else NULL
      )
      resample_plan <- fastml_new_resample_plan(
        splits = resamples_obj,
        method = "validation_split",
        params = list(prop = 1 - 1 / folds, strata = status_strata)
      )
    } else if (resampling_method == "repeatedcv") {
      if (nrow(resample_data) < folds) {
        stop(
          sprintf(
            "You requested %d-fold cross-validation, but your training set only has %d rows. \nThis prevents each fold from having at least one row. \nEither reduce 'folds', increase data, or use a different resampling method (e.g. 'boot').",
            folds,
            nrow(resample_data)
          )
        )
      }
      repeats_val <- if (is.null(repeats)) 1L else repeats
      resamples_obj <- vfold_cv(
        resample_data,
        v = folds,
        repeats = repeats_val,
        strata = if (!is.null(status_strata)) dplyr::all_of(status_strata) else NULL
      )
      resample_plan <- fastml_new_resample_plan(
        splits = resamples_obj,
        method = "repeatedcv",
        params = list(
          v = folds,
          repeats = repeats_val,
          strata = status_strata
        )
      )
    } else if (resampling_method == "grouped_cv") {
      repeats_arg <- if (is.null(repeats)) 1L else repeats
      resamples_obj <- make_grouped_cv(
        data = resample_data,
        group_cols = group_cols,
        v = folds,
        strata = status_strata,
        repeats = repeats_arg
      )
      resample_plan <- fastml_new_resample_plan(
        splits = resamples_obj,
        method = "grouped_cv",
        params = list(
          group_cols = group_cols,
          v = folds,
          repeats = repeats_arg,
          strata = status_strata
        )
      )
    } else if (resampling_method %in% c("blocked_cv", "rolling_origin")) {
      if (is.null(block_col) || length(block_col) != 1) {
        stop("'block_col' must be provided when using blocked or rolling resampling.")
      }

      if (resampling_method == "blocked_cv") {
        blocked_info <- make_blocked_cv(
          data = resample_data,
          block_col = block_col,
          block_size = block_size
        )

        if (!is.numeric(folds) || length(folds) != 1 || folds < 2 || folds != as.integer(folds)) {
          stop("'folds' must be an integer greater than or equal to 2 for 'blocked_cv'.")
        }

        if (folds > blocked_info$n_blocks) {
          stop("'folds' cannot exceed the number of derived blocks.")
        }

        block_group_col <- ".fastml_block"
        blocked_data <- blocked_info$data
        blocked_data[[block_group_col]] <- blocked_info$block_ids

        resamples_obj <- rsample::group_vfold_cv(
          blocked_data,
          group = !!rlang::sym(block_group_col),
          v = folds
        )

        resamples_obj$splits <- lapply(
          resamples_obj$splits,
          function(split) {
            split$data[[block_group_col]] <- NULL
            split
          }
        )

        resample_plan <- fastml_new_resample_plan(
          splits = resamples_obj,
          method = "blocked_cv",
          params = list(
            block_col = block_col,
            block_size = block_size,
            v = folds
          )
        )
      } else {
        resamples_obj <- make_rolling_origin_cv(
          data = resample_data,
          block_col = block_col,
          initial_window = initial_window,
          assess_window = assess_window,
          skip = skip
        )
        resample_plan <- fastml_new_resample_plan(
          splits = resamples_obj,
          method = "rolling_origin",
          params = list(
            block_col = block_col,
            initial_window = initial_window,
            assess_window = assess_window,
            skip = skip
          )
        )
      }
    } else if (resampling_method == "nested_cv") {
      stop("Nested resampling for survival tasks is not yet supported. Please use a different resampling_method or provide custom resamples.")
    } else if (resampling_method == "none") {
      resample_plan <- NULL
    }

    resamples <- if (!is.null(resample_plan)) fastml_resample_splits(resample_plan) else NULL

    if (is.null(resamples) && (use_default_tuning || !is.null(tune_params)) &&
        !identical(tuning_strategy, "none")) {
      warning("Tuning is skipped because no resamples were supplied (set resampling_method or provide resamples).")
    }

    resampling_summaries <- list()

    response_col <- label
    label_cols <- attr(train_data[[label]], "fastml_label_cols")
    if (is.null(label_cols)) {
      if (length(label) > 1) {
        label_cols <- label
      } else {
        stop("Survival training requires original time/status column names to be stored on the response.")
      }
    }
    start_col <- if (length(label_cols) == 3) label_cols[1] else NULL
    time_col <- if (length(label_cols) == 3) label_cols[2] else label_cols[1]
    status_col <- label_cols[length(label_cols)]

    get_engine <- function(algo, fastml_default_engine) {
      # If user explicitly specified engine, use it without warning
      if (!is.null(algorithm_engines) && !is.null(algorithm_engines[[algo]])) {
        if (verbose) {
          message(sprintf("[%s] Using user-specified engine: '%s'", algo, algorithm_engines[[algo]]))
        }
        return(algorithm_engines[[algo]])
      }

      # Determine which default to use
      if (use_parsnip_defaults) {
        parsnip_default <- get_parsnip_default_engine(algo, task)
        if (!is.null(parsnip_default)) {
          if (verbose) {
            message(sprintf("[%s] Using parsnip default engine: '%s'", algo, parsnip_default))
          }
          return(parsnip_default)
        }
        # Fall back to fastml default if no parsnip default exists
        if (verbose) {
          message(sprintf("[%s] No parsnip default found, using fastml default: '%s'", algo, fastml_default_engine))
        }
        return(fastml_default_engine)
      }

      # Using fastml default - warn if it differs from parsnip
      if (warn_engine_defaults) {
        warn_default_override(algo, task, fastml_default_engine, verbose = verbose)
      } else if (verbose) {
        message(sprintf("[%s] Using fastml default engine: '%s'", algo, fastml_default_engine))
      }
      return(fastml_default_engine)
    }

    rec_prep_cache <- NULL
    baked_train_cache <- NULL
    get_prepped_data <- function() {
      if (is.null(rec_prep_cache)) {
        rec_prep_cache <<- recipes::prep(recipe, training = train_data, retain = TRUE)
        baked_train_cache <<- recipes::bake(rec_prep_cache, new_data = NULL)
      }
      list(recipe = rec_prep_cache, data = baked_train_cache)
    }

    create_native_spec <- function(algo_name, engine_name, fit_obj, recipe_obj, extras = list()) {
      spec <- c(list(
        algo = algo_name,
        engine = if (!is.null(engine_name)) engine_name else NA_character_,
        fit = fit_obj,
        recipe = recipe_obj,
        response = response_col,
        label_cols = label_cols,
        time_col = time_col,
        status_col = status_col,
        start_col = start_col
      ), extras)
      class(spec) <- c("fastml_native_survival", "fastml_model")
      spec
    }

    resampling_summaries <- list()

    for (algo in algorithms) {
      engine <- get_engine(algo, get_default_engine(algo, task))
      engine_args <- resolve_engine_params(engine_params, algo, engine)
      fit_engine_args <- NULL
      prefer_parsnip <- !is.null(resamples) || !identical(resampling_method, "none") ||
        use_default_tuning || (!is.null(tune_params) && length(tune_params) > 0)
      spec_is_parsnip <- FALSE
      apply_seed_defaults <- function() {
        engine_args <<- fastml_apply_engine_seed(engine_args, algo, engine, seed, engine_threads_val, task)
        add_determinism_warning(algo, engine, engine_args)
      }

      if (algo %in% c("rand_forest", "rand_forest_survival")) {
        if (identical(engine, "aorsf") && !requireNamespace("aorsf", quietly = TRUE)) {
          if (requireNamespace("ranger", quietly = TRUE)) {
            warning("Engine 'aorsf' not installed. Falling back to 'ranger' for survival random forest.")
            engine <- "ranger"
          } else {
            warning("Neither 'aorsf' nor 'ranger' are available. Skipping survival random forest.")
            next
          }
        }
        if (!requireNamespace("censored", quietly = TRUE)) {
          warning("Package 'censored' not installed; skipping survival random forest parsnip model.")
          next
        }
        apply_seed_defaults()
        spec <- define_rand_forest_spec("survival", train_data, label,
                                       tuning = FALSE, engine = engine)$model_spec
        spec_is_parsnip <- TRUE
      } else if (algo == "elastic_net") {
        warning("Survival 'elastic_net' requires censored survival model specs not available in your setup. Skipping.")
        next
      } else if (algo == "cox_ph") {
        if (!requireNamespace("survival", quietly = TRUE)) {
          stop("The 'survival' package is required for Cox PH. Please install it.")
        }
        apply_seed_defaults()
        if (prefer_parsnip) {
          base_spec <- parsnip::proportional_hazards(
            penalty = NULL,
            mixture = NULL
          ) %>%
            parsnip::set_mode("censored regression")
          spec <- do.call(parsnip::set_engine, c(list(base_spec, engine = engine), engine_args))
          spec_is_parsnip <- TRUE
        } else {
          prep_dat <- get_prepped_data()
          baked_train <- prep_dat$data
          rec_prep <- prep_dat$recipe
          fit <- call_with_engine_params(
            survival::coxph,
            list(
              formula = as.formula(paste(response_col, "~ .")),
              data = baked_train,
              ties = "efron"
            ),
            engine_args
          )
          spec <- create_native_spec("cox_ph", engine, fit, rec_prep)
        }
      } else if (algo == "penalized_cox") {
        if (!requireNamespace("censored", quietly = TRUE)) {
          warning("Package 'censored' not installed; skipping penalized_cox.")
          next
        }
        if (!requireNamespace("glmnet", quietly = TRUE)) {
          stop("The 'glmnet' package is required for penalized_cox. Please install it.")
        }

        apply_seed_defaults()

        defaults <- get_default_params("penalized_cox", task, engine = engine)
        penalty_val <- defaults$penalty
        mixture_val <- defaults$mixture

        if (length(engine_args) > 0) {
          if ("penalty" %in% names(engine_args)) {
            penalty_val <- engine_args$penalty
            engine_args$penalty <- NULL
          }
          if ("mixture" %in% names(engine_args)) {
            mixture_val <- engine_args$mixture
            engine_args$mixture <- NULL
          }
        }

        spec_info <- define_penalized_cox_spec(
          task = task,
          penalty = penalty_val,
            mixture = mixture_val,
            engine = engine,
            engine_params = engine_args
          )
        spec <- spec_info$model_spec
        fit_engine_args <- spec_info$fit_args
        spec_is_parsnip <- TRUE
      } else if (algo %in% c("xgboost", "xgboost_aft")) {
        if (!requireNamespace("xgboost", quietly = TRUE)) {
          warning("Package 'xgboost' not installed; skipping xgboost survival model.")
          next
        }

        engine_name <- engine
        if (is.null(engine_name) || length(engine_name) == 0 || is.na(engine_name)) {
          engine_name <- "aft"
        }
        engine_name <- tolower(as.character(engine_name)[1])
        if (engine_name %in% c("xgboost", "default")) {
          engine_name <- "aft"
        }
        if (!identical(engine_name, "aft")) {
          warning(sprintf("XGBoost survival engine '%s' is not supported; skipping.", engine_name))
          next
        }
        engine <- engine_name
        apply_seed_defaults()

        if (!is.null(start_col)) {
          warning("XGBoost AFT does not support start-stop survival outcomes; skipping.")
          next
        }

        defaults <- get_default_params(
          "xgboost",
          task,
          num_predictors = ncol(train_data %>% dplyr::select(-dplyr::all_of(label))),
          engine = "xgboost"
        )

        if (prefer_parsnip) {
          warning(
            paste(
              "Parsnip does not currently expose a 'censored regression' mode for xgboost.",
              "Falling back to the native AFT implementation with resampling disabled."
            )
          )
          resamples <- NULL
          resample_plan <- NULL
          prefer_parsnip <- FALSE
        }

        if (prefer_parsnip) {
          tuning_flag <- (!is.null(resamples)) && (use_default_tuning || (!is.null(tune_params) &&
                                                   !is.null(tune_params[[algo]]) &&
                                                   !is.null(tune_params[[algo]][[engine]])))
          model_spec <- parsnip::boost_tree(
            trees = if (tuning_flag) tune() else defaults$trees,
            tree_depth = if (tuning_flag) tune() else defaults$tree_depth,
            learn_rate = if (tuning_flag) tune() else defaults$learn_rate,
            mtry = if (tuning_flag) tune() else defaults$mtry,
            min_n = if (tuning_flag) tune() else defaults$min_n,
            loss_reduction = if (tuning_flag) tune() else defaults$loss_reduction,
            sample_size = if (tuning_flag) tune() else defaults$sample_size
          ) %>%
            parsnip::set_mode("censored regression")

          engine_aug <- c(list(
            model_spec,
            engine = "xgboost",
            objective = "survival:aft",
            eval_metric = "aft-nloglik"
          ), engine_args)
          model_spec <- do.call(parsnip::set_engine, engine_aug)
          spec <- model_spec
          spec_is_parsnip <- TRUE
        } else {
          prep_dat <- get_prepped_data()
          baked_train <- prep_dat$data
          rec_prep <- prep_dat$recipe

          predictor_cols <- setdiff(names(baked_train), response_col)
          if (length(predictor_cols) == 0) {
            warning("No predictors available after preprocessing; skipping xgboost survival model.")
            next
          }
          predictors <- baked_train[, predictor_cols, drop = FALSE]
          feature_names <- predictor_cols
          design_mat <- fastml_prepare_xgb_matrix(predictors, feature_names)

          if (ncol(design_mat) == 0) {
            warning("Unable to construct a feature matrix for xgboost; skipping.")
            next
          }

          time_vec <- as.numeric(train_data[[time_col]])
          status_vec <- as.numeric(train_data[[status_col]])
          valid_idx <- is.finite(time_vec) & time_vec > 0 & is.finite(status_vec)
          if (!any(valid_idx)) {
            warning("No valid survival outcomes available for xgboost; skipping.")
            next
          }
          if (!all(valid_idx)) {
            design_mat <- design_mat[valid_idx, , drop = FALSE]
            time_vec <- time_vec[valid_idx]
            status_vec <- status_vec[valid_idx]
          }
          if (!any(status_vec > 0, na.rm = TRUE)) {
            warning("XGBoost AFT requires at least one observed event; skipping.")
            next
          }

          if (any(!is.finite(time_vec) | time_vec <= 0)) {
            stop(
              "XGBoost AFT requires strictly positive, finite survival times. Remove or correct invalid times before fitting.",
              call. = FALSE
            )
          }
          log_time <- log(time_vec)
          lower_bounds <- log_time
          upper_bounds <- log_time
          upper_bounds[status_vec <= 0 | !is.finite(status_vec)] <- Inf

          dtrain <- xgboost::xgb.DMatrix(data = design_mat, label = log_time)
          xgboost::setinfo(dtrain, "label_lower_bound", lower_bounds)
          xgboost::setinfo(dtrain, "label_upper_bound", upper_bounds)

          max_depth <- defaults$tree_depth
          if (is.null(max_depth) || !is.finite(max_depth)) max_depth <- 6
          max_depth <- as.integer(max(1, round(max_depth)))

          eta <- defaults$learn_rate
          if (is.null(eta) || !is.finite(eta)) eta <- 0.1

          gamma <- defaults$loss_reduction
          if (is.null(gamma) || !is.finite(gamma)) gamma <- 0

          min_child_weight <- defaults$min_n
          if (is.null(min_child_weight) || !is.finite(min_child_weight)) min_child_weight <- 2

          subsample <- defaults$sample_size
          if (is.null(subsample) || !is.finite(subsample)) subsample <- 0.5

          mtry <- defaults$mtry
          if (is.null(mtry) || !is.finite(mtry)) {
            colsample <- 1
          } else {
            colsample <- min(1, max(1, mtry) / max(1, ncol(design_mat)))
          }

          nrounds <- defaults$trees
          if (is.null(nrounds) || !is.finite(nrounds) || nrounds <= 0) nrounds <- 50
          nrounds <- as.integer(max(1, round(nrounds)))

          params <- list(
            objective = "survival:aft",
            eval_metric = "aft-nloglik",
            max_depth = max_depth,
            eta = eta,
            gamma = gamma,
            min_child_weight = min_child_weight,
            subsample = subsample,
            colsample_bytree = colsample,
            aft_loss_distribution = "logistic",
            aft_loss_distribution_scale = 1,
            verbosity = 0
          )

          seed_val <- fastml_normalize_seed(seed)
          if (!is.null(engine_args$seed)) {
            seed_val <- engine_args$seed[[1]]
            engine_args$seed <- NULL
          }
          if (!is.null(seed_val) && is.null(params$seed)) {
            params$seed <- seed_val
          }

          threads_val <- engine_threads_val
          if (!is.null(engine_args$nthread)) {
            threads_val <- as.integer(engine_args$nthread[[1]])
            engine_args$nthread <- NULL
          }
          if (!is.null(threads_val) && is.null(params$nthread)) {
            params$nthread <- threads_val
          }

          aft_quantiles <- c(0.25, 0.5, 0.75)
          early_stop <- defaults$stop_iter
          if (!is.null(early_stop) && is.finite(early_stop) && early_stop <= 0) {
            early_stop <- NULL
          }

          extra_args <- list()
          if (length(engine_args) > 0) {
            if (!is.null(engine_args$params)) {
              if (!is.list(engine_args$params)) {
                stop("Engine parameters for xgboost must provide 'params' as a list.")
              }
              params <- merge_engine_args(params, engine_args$params)
              engine_args$params <- NULL
            }
            if (!is.null(engine_args$nrounds)) {
              nrounds <- as.integer(engine_args$nrounds[1])
              engine_args$nrounds <- NULL
            }
            if (!is.null(engine_args$early_stopping_rounds)) {
              early_stop <- as.integer(engine_args$early_stopping_rounds[1])
              engine_args$early_stopping_rounds <- NULL
            }
            if (!is.null(engine_args$quantiles)) {
              aft_quantiles <- engine_args$quantiles
              engine_args$quantiles <- NULL
            }
            if (!is.null(engine_args$aft_quantiles)) {
              aft_quantiles <- engine_args$aft_quantiles
              engine_args$aft_quantiles <- NULL
            }
            extra_args <- merge_engine_args(extra_args, engine_args)
          }

          aft_quantiles <- as.numeric(aft_quantiles)
          aft_quantiles <- aft_quantiles[is.finite(aft_quantiles) & aft_quantiles > 0 & aft_quantiles < 1]
          aft_quantiles <- sort(unique(aft_quantiles))

          if (length(aft_quantiles) == 0) {
            aft_quantiles <- c(0.25, 0.5, 0.75)
          }

          if (!is.finite(nrounds) || nrounds <= 0) {
            warning("xgboost survival requires 'nrounds' > 0; skipping.")
            next
          }

          train_args <- c(
            list(
              params = params,
              data = dtrain,
              nrounds = nrounds
            ),
            extra_args
          )

          if (!is.null(early_stop) && is.finite(early_stop) && early_stop > 0) {
            train_args$early_stopping_rounds <- as.integer(early_stop)
          }

          booster <- fastml_with_seed(seed_val, function() {
            do.call(xgboost::xgb.train, train_args)
          })

          aft_distribution <- params$aft_loss_distribution
          if (is.null(aft_distribution) || !nzchar(as.character(aft_distribution)[1])) {
            aft_distribution <- "logistic"
          }
          aft_scale <- params$aft_loss_distribution_scale
          if (is.null(aft_scale) || !is.finite(aft_scale) || aft_scale <= 0) {
            aft_scale <- 1
          }

          fit <- list(
            booster = booster,
            objective = "survival:aft",
            feature_names = colnames(design_mat),
            aft_distribution = as.character(aft_distribution)[1],
            aft_scale = as.numeric(aft_scale)[1],
            aft_quantiles = aft_quantiles
          )
          class(fit) <- "fastml_xgb_survival"

          spec <- create_native_spec("xgboost", engine, fit, rec_prep)
        }
      } else if (algo == "stratified_cox") {
        if (!requireNamespace("survival", quietly = TRUE)) {
          stop("The 'survival' package is required for stratified Cox. Please install it.")
        }
        apply_seed_defaults()
        # Native-only model: warn if resampling was requested
        if (prefer_parsnip) {
          warning(
            "Stratified Cox uses native survival::coxph() which does not support ",
            "tidymodels resampling. Preprocessing will be applied to full training data. ",
            "Cross-validation metrics will not be available for this model.",
            call. = FALSE
          )
        }
        strata_cols <- names(train_data)[grepl("^strata", names(train_data))]
        if (length(strata_cols) == 0) {
          warning("No columns starting with 'strata' were found; skipping stratified Cox.")
          next
        }
        prep_dat <- get_prepped_data()
        baked_train <- prep_dat$data
        rec_prep <- prep_dat$recipe
        strata_cols_present <- character()
        strata_base_cols <- character()
        strata_info <- list()
        for (sc in strata_cols) {
          if (!(sc %in% names(baked_train)) && sc %in% names(train_data)) {
            baked_train[[sc]] <- train_data[[sc]]
          }
          if (!(sc %in% names(baked_train))) {
            next
          }
          source_vals <- NULL
          if (sc %in% names(train_data)) {
            source_vals <- train_data[[sc]]
          } else {
            source_vals <- baked_train[[sc]]
          }
          desired_levels <- NULL
          if (is.factor(source_vals)) {
            desired_levels <- levels(source_vals)
          } else if (!is.null(source_vals)) {
            desired_levels <- sort(unique(source_vals))
          }
          if (is.null(desired_levels) || length(desired_levels) == 0) {
            alt_vals <- baked_train[[sc]]
            if (is.factor(alt_vals)) {
              desired_levels <- levels(alt_vals)
            } else {
              desired_levels <- sort(unique(alt_vals))
            }
          }
          desired_levels <- desired_levels[!is.na(desired_levels)]
          if (length(desired_levels) == 0) {
            next
          }
          baked_train[[sc]] <- factor(baked_train[[sc]], levels = desired_levels)
          strata_cols_present <- c(strata_cols_present, sc)
          base_candidate <- gsub("^strata[._]*", "", sc)
          base_candidate <- gsub("^[._]+", "", base_candidate)
          display_name <- if (nzchar(base_candidate)) base_candidate else sc
          strata_info[[sc]] <- list(
            column = sc,
            display = display_name,
            levels = as.character(desired_levels)
          )
          if (nzchar(base_candidate) && base_candidate %in% names(baked_train)) {
            strata_base_cols <- c(strata_base_cols, base_candidate)
          }
        }
        strata_cols_present <- unique(strata_cols_present)
        strata_base_cols <- unique(strata_base_cols)
        if (length(strata_cols_present) == 0) {
          warning("Unable to identify usable strata columns after preprocessing; skipping stratified Cox.")
          next
        }
        strata_dummy_cols <- character(0)
        if (length(strata_cols_present) > 0) {
          for (sc in strata_cols_present) {
            prefix_underscore <- paste0(sc, "_")
            prefix_dot <- paste0(sc, ".")
            matches <- names(baked_train)[startsWith(names(baked_train), prefix_underscore) |
                                            startsWith(names(baked_train), prefix_dot)]
            if (length(matches) > 0) {
              strata_dummy_cols <- c(strata_dummy_cols, matches)
            }
          }
          strata_dummy_cols <- unique(strata_dummy_cols)
        }
        predictor_cols <- setdiff(names(baked_train),
                                  c(response_col, strata_cols_present, strata_dummy_cols,
                                    strata_base_cols))
        rhs_terms <- c(predictor_cols, paste0("strata(", strata_cols_present, ")"))
        formula_rhs <- if (length(rhs_terms) == 0) "1" else paste(rhs_terms, collapse = " + ")
        f <- as.formula(paste(response_col, "~", formula_rhs))
        fit <- call_with_engine_params(
          survival::coxph,
          list(
            formula = f,
            data = baked_train,
            ties = "efron"
          ),
          engine_args
        )
        spec <- create_native_spec("stratified_cox", engine, fit, rec_prep,
                                   extras = list(strata_cols = strata_cols_present,
                                                 strata_dummy_cols = strata_dummy_cols,
                                                 strata_info = strata_info,
                                                 strata_base_cols = strata_base_cols))
      } else if (algo == "time_varying_cox") {
        if (length(label_cols) != 3) {
          warning("time_varying_cox requires label = c(start, stop, status). Skipping.")
          next
        }
        if (!requireNamespace("survival", quietly = TRUE)) {
          stop("The 'survival' package is required for time-varying Cox. Please install it.")
        }
        apply_seed_defaults()
        # Native-only model: warn if resampling was requested
        if (prefer_parsnip) {
          warning(
            "Time-varying Cox uses native survival::coxph() which does not support ",
            "tidymodels resampling. Preprocessing will be applied to full training data. ",
            "Cross-validation metrics will not be available for this model.",
            call. = FALSE
          )
        }
        prep_dat <- get_prepped_data()
        baked_train <- prep_dat$data
        rec_prep <- prep_dat$recipe
        fit <- call_with_engine_params(
          survival::coxph,
          list(
            formula = as.formula(paste(response_col, "~ .")),
            data = baked_train,
            ties = "efron"
          ),
          engine_args
        )
        spec <- create_native_spec("time_varying_cox", engine, fit, rec_prep)
      } else if (algo == "survreg") {
        if (!requireNamespace("survival", quietly = TRUE)) {
          stop("The 'survival' package is required for survreg. Please install it.")
        }
        apply_seed_defaults()
        if (prefer_parsnip) {
          # Use parsnip workflow - preprocessing handled per fold during resampling
          dist_val <- "weibull"
          if (!is.null(engine_args$dist)) {
            dist_val <- engine_args$dist
            engine_args$dist <- NULL
          }
          survreg_spec <- parsnip::survival_reg(dist = dist_val) %>%
            parsnip::set_mode("censored regression")
          spec <- do.call(parsnip::set_engine, c(list(survreg_spec, engine = engine), engine_args))
          spec_is_parsnip <- TRUE
        } else {
          # Native fit - no resampling, so prepping on full data is valid
          prep_dat <- get_prepped_data()
          baked_train <- prep_dat$data
          rec_prep <- prep_dat$recipe
          fit <- call_with_engine_params(
            survival::survreg,
            list(
              formula = as.formula(paste(response_col, "~ .")),
              data = baked_train,
              dist = "weibull"
            ),
            engine_args
          )
          spec <- create_native_spec("survreg", engine, fit, rec_prep,
                                     extras = list(distribution = "weibull"))
        }
      } else if (algo == "parametric_surv") {
        if (!requireNamespace("flexsurv", quietly = TRUE)) {
          warning("Package 'flexsurv' not installed; skipping parametric_surv.")
          next
        }
        apply_seed_defaults()
        # Native-only model: warn if resampling was requested
        if (prefer_parsnip) {
          warning(
            "Parametric survival (flexsurv) does not have a tidymodels interface. ",
            "Preprocessing will be applied to full training data. ",
            "Cross-validation metrics will not be available for this model.",
            call. = FALSE
          )
        }
        prep_dat <- get_prepped_data()
        baked_train <- prep_dat$data
        rec_prep <- prep_dat$recipe
        dist_arg <- engine_args$dist
        if (!is.null(dist_arg)) {
          engine_args$dist <- NULL
        }
        if (is.null(dist_arg)) {
          dist_arg <- "weibull"
        }
        dist_arg <- as.character(dist_arg)[1]
        dist_arg <- tolower(dist_arg)
        dist_map <- c(
          weibull = "weibull",
          loglogistic = "llogis",
          llogis = "llogis",
          lognormal = "lognormal",
          exponential = "exp",
          exp = "exp"
        )
        label_map <- c(
          weibull = "weibull",
          loglogistic = "loglogistic",
          llogis = "loglogistic",
          lognormal = "lognormal",
          exponential = "exponential",
          exp = "exponential"
        )
        if (!(dist_arg %in% names(dist_map))) {
          stop("Unsupported distribution for parametric_surv. Choose one of 'weibull', 'loglogistic', 'lognormal', or 'exponential'.")
        }
        dist_flex <- dist_map[[dist_arg]]
        dist_label <- label_map[[dist_arg]]
        base_args <- list(
          formula = as.formula(paste(response_col, "~ .")),
          data = baked_train,
          dist = dist_flex
        )
        fit <- fastml_fit_flexsurvreg(base_args, engine_args)
        extras <- list(
          distribution = dist_flex,
          distribution_label = dist_label
        )
        if (!is.null(time_col) && time_col %in% names(train_data)) {
          extras$train_times <- train_data[[time_col]]
        }
        if (!is.null(status_col) && status_col %in% names(train_data)) {
          extras$train_status <- train_data[[status_col]]
        }
        extras$train_size <- nrow(baked_train)
        spec <- create_native_spec(
          "parametric_surv",
          engine,
          fit,
          rec_prep,
          extras = extras
        )
      } else if (algo == "piecewise_exp") {
        if (!requireNamespace("flexsurv", quietly = TRUE)) {
          warning("Package 'flexsurv' not installed; skipping piecewise_exp.")
          next
        }
        apply_seed_defaults()
        # Native-only model: warn if resampling was requested
        if (prefer_parsnip) {
          warning(
            "Piecewise exponential (flexsurv) does not have a tidymodels interface. ",
            "Preprocessing will be applied to full training data. ",
            "Cross-validation metrics will not be available for this model.",
            call. = FALSE
          )
        }

        prep_dat <- get_prepped_data()
        baked_train <- prep_dat$data
        rec_prep <- prep_dat$recipe

        breaks_val <- NULL
        if ("breaks" %in% names(engine_args)) {
          breaks_val <- engine_args$breaks
          engine_args$breaks <- NULL
        }
        if ("cuts" %in% names(engine_args) && is.null(breaks_val)) {
          breaks_val <- engine_args$cuts
          engine_args$cuts <- NULL
        }
        if ("knots" %in% names(engine_args) && is.null(breaks_val)) {
          breaks_val <- engine_args$knots
          engine_args$knots <- NULL
        }

        breaks_val <- fastml_piecewise_normalize_breaks(breaks_val)
        if (length(breaks_val) == 0) breaks_val <- NULL

        if (is.null(breaks_val)) {
          event_times <- NULL
          if (!is.null(status_col) && status_col %in% names(train_data)) {
            status_vals <- train_data[[status_col]]
            if (!is.null(time_col) && time_col %in% names(train_data)) {
              time_vals <- train_data[[time_col]]
            } else {
              time_vals <- NULL
            }
            if (!is.null(time_vals)) {
              event_idx <- which(is.finite(status_vals) & status_vals == 1)
              if (length(event_idx) > 0) {
                event_times <- as.numeric(time_vals[event_idx])
                event_times <- event_times[is.finite(event_times) & event_times > 0]
              }
            }
          }
          if (!is.null(event_times) && length(event_times) >= 2) {
            quantiles <- stats::quantile(
              event_times,
              probs = c(1 / 3, 2 / 3),
              names = FALSE,
              type = 7
            )
            breaks_val <- fastml_piecewise_normalize_breaks(quantiles)
          }
        }

        pw_spec <- fastml_piecewise_make_distribution(breaks_val)
        base_args <- list(
          formula = as.formula(paste(response_col, "~ .")),
          data = baked_train,
          dist = pw_spec$dlist,
          dfns = pw_spec$dfns
        )
        if (length(pw_spec$aux) > 0) {
          base_args$aux <- pw_spec$aux
        }

        fit <- fastml_fit_flexsurvreg(base_args, engine_args)

        extras <- list(
          distribution = "fastml_piecewise_exponential",
          distribution_label = "piecewise exponential",
          breaks = breaks_val
        )
        if (!is.null(time_col) && time_col %in% names(train_data)) {
          extras$train_times <- train_data[[time_col]]
        }
        if (!is.null(status_col) && status_col %in% names(train_data)) {
          extras$train_status <- train_data[[status_col]]
        }

        spec <- create_native_spec(
          "piecewise_exp",
          engine,
          fit,
          rec_prep,
          extras = extras
        )

        # ensure extras are attached to top-level spec
        if (inherits(spec, "fastml_native_survival")) {
          spec$extras <- extras
        }
      }
      else if (algo == "royston_parmar") {
        if (!requireNamespace("rstpm2", quietly = TRUE)) {
          warning("Package 'rstpm2' not installed; skipping royston_parmar.")
          next
        }
        apply_seed_defaults()
        # Native-only model: warn if resampling was requested
        if (prefer_parsnip) {
          warning(
            "Royston-Parmar (rstpm2) does not have a tidymodels interface. ",
            "Preprocessing will be applied to full training data. ",
            "Cross-validation metrics will not be available for this model.",
            call. = FALSE
          )
        }
        prep_dat <- get_prepped_data()
        baked_train <- as.data.frame(prep_dat$data)
        rec_prep <- prep_dat$recipe
        rp_train <- baked_train
        if (!(response_col %in% names(rp_train)) && response_col %in% names(train_data)) {
          rp_train[[response_col]] <- train_data[[response_col]]
        }
        if (length(label_cols) < 2) {
          warning("royston_parmar requires time and status columns; skipping.")
          next
        }
        for (lc in label_cols) {
          if (lc %in% names(train_data)) {
            rp_train[[lc]] <- train_data[[lc]]
          }
        }
        predictor_cols <- setdiff(names(rp_train), c(response_col, label_cols))
        predictor_terms <- if (length(predictor_cols) == 0) {
          "1"
        } else {
          paste(paste0("`", predictor_cols, "`"), collapse = " + ")
        }
        label_terms <- paste0("`", label_cols, "`")
        surv_lhs <- if (length(label_cols) == 3) {
          paste0("survival::Surv(", paste(label_terms, collapse = ", "), ")")
        } else {
          paste0("survival::Surv(", paste(label_terms[1:2], collapse = ", "), ")")
        }
        rp_formula <- as.formula(paste(surv_lhs, "~", predictor_terms))
        fit <- tryCatch({
          stpm2_fn <- get("stpm2", envir = asNamespace("rstpm2"))
          base_args <- list(
            formula = rp_formula,
            data = rp_train,
            df = 3,
            penalised = FALSE
          )
          fit_obj <- call_with_engine_params(
            stpm2_fn,
            base_args,
            engine_args
          )
          fit_obj
        }, error = function(e) e)
        if (inherits(fit, "error")) {
          failed_models[[length(failed_models) + 1]] <- list(
            algorithm = algo,
            engine = engine,
            reason = fit$message
          )
          warning(sprintf("royston_parmar training failed: %s", fit$message))
          next
        }
        if (!inherits(fit, c("stpm2", "pstpm2"))) {
          warning("rstpm2::stpm2 did not return an 'stpm2' compatible object; skipping royston_parmar.")
          next
        }
        loglik_val <- tryCatch({
          val <- stats::logLik(fit)
          if (length(val) > 0) as.numeric(val)[1] else NA_real_
        }, error = function(e) NA_real_)
        if (!is.finite(loglik_val)) {
          loglik_val <- tryCatch({
            val <- stats::logLik(fit)
            if (length(val) > 0) as.numeric(val)[1] else NA_real_
          }, error = function(e) NA_real_)
        }
        safe_stat_val <- function(expr) {
          val <- tryCatch(expr, error = function(e) NA_real_)
          if (length(val) == 0) {
            return(NA_real_)
          }
          as.numeric(val[1])
        }
        loglik_val <- safe_stat_val(stats::logLik(fit))
        aic_val <- safe_stat_val(stats::AIC(fit))
        bic_val <- safe_stat_val(stats::BIC(fit))
        spec <- create_native_spec("royston_parmar", engine, fit, rec_prep,
                                   extras = list(
                                     spline_df = 3,
                                     loglik = loglik_val,
                                     aic = aic_val,
                                     bic = bic_val
                                   ))
      } else if (algo == "aft") {
        warning("Survival 'aft' requires censored survival model specs not available in your setup. Skipping.")
        next
      } else {
        next
      }

      # Apply engine-level arguments to parsnip specs and avoid passing them via fit(...),
      # which would otherwise error in workflows.
      if (spec_is_parsnip && length(engine_args) > 0) {
        spec <- do.call(parsnip::set_engine, c(list(spec, engine = engine), engine_args))
      }

      if (is.null(fit_engine_args)) {
        fit_engine_args <- if (spec_is_parsnip) list() else engine_args
      }

      if (!is.null(resamples) && inherits(spec, "fastml_native_survival")) {
        stop(
          paste(
            "Guarded resampling for survival models is not yet implemented for native engines.",
            "Please supply parsnip-compatible specifications or disable resampling."
          )
        )
      }

      if (inherits(spec, "fastml_native_survival")) {
        # Announce the execution path: which of the two survival paths a method
        # takes is method- and configuration-dependent, so a user expecting
        # workflow-based fitting should be told when a native engine is used.
        if (verbose) {
          message(sprintf(
            paste0(
              "Algorithm '%s' is fitted through its native engine rather than a ",
              "parsnip workflow; preprocessing is applied once via the fitted ",
              "recipe rather than re-estimated per resample."
            ),
            algo
          ))
        }
        models[[algo]] <- spec
      } else {
        if (verbose) {
          message(sprintf(
            "Algorithm '%s' is fitted through a parsnip workflow (guarded resampling path).",
            algo
          ))
        }
        wf <- workflows::workflow() %>%
          workflows::add_recipe(recipe) %>%
          workflows::add_model(spec)

        user_params <- NULL
        if (!is.null(tune_params) &&
            !is.null(tune_params[[algo]]) &&
            !is.null(tune_params[[algo]][[engine]])) {
          user_params <- tune_params[[algo]][[engine]]
        }
        defaults <- NULL
        if (use_default_tuning) {
          defaults <- get_default_tune_params(
            algo,
            train_data,
            label,
            engine
          )
        }
        engine_tune_params <- if (is.null(defaults)) list() else defaults
        if (!is.null(user_params)) {
          for (nm in names(user_params)) {
            engine_tune_params[[nm]] <- user_params[[nm]]
          }
        }

        perform_tuning <- !is.null(resamples) &&
          (use_default_tuning || !is.null(user_params)) &&
          !identical(tuning_strategy, "none")

        tune_grid_values <- NULL
        if (perform_tuning) {
          param_set <- extract_parameter_set_dials(spec)
          param_set <- finalize(
            param_set,
            x = train_data %>% dplyr::select(-dplyr::all_of(label))
          )
          if (!is.null(engine_tune_params) && length(engine_tune_params) > 0) {
            param_set <- update_params_surv(param_set, engine_tune_params)
          }
          if (nrow(param_set) > 0) {
            if (tuning_strategy == "grid" && !adaptive) {
              tune_grid_values <- grid_regular(param_set, levels = grid_levels)
            } else {
              tune_grid_values <- grid_regular(param_set, levels = grid_levels)
            }
          } else {
            perform_tuning <- FALSE
          }
        }

        metric_direction <- tryCatch({
          if (requireNamespace("yardstick", quietly = TRUE) &&
              exists(metric, envir = asNamespace("yardstick"), inherits = FALSE)) {
            attr(get(metric, envir = asNamespace("yardstick")), "direction", exact = TRUE)
          } else {
            NA_character_
          }
        }, error = function(e) NA_character_)

        lower_is_better <- if (!is.na(metric_direction) && !is.null(metric_direction)) {
          identical(metric_direction, "minimize")
        } else {
          metric %in% c("rmse", "mae", "ibs", "logloss", "mse", "brier_score", "ece") ||
            grepl("brier", metric, fixed = TRUE)
        }

        best_params <- NULL
        res_summary_best <- NULL

        if (perform_tuning && !is.null(tune_grid_values)) {
          for (g in seq_len(nrow(tune_grid_values))) {
            wf_g <- finalize_workflow(wf, tune_grid_values[g, ])
            res_summary <- fastml_guarded_resample_fit(
              workflow_spec = wf_g,
              resamples = if (!is.null(resample_plan)) resample_plan else resamples,
              original_train_rows = nrow(train_data),
              task = task,
              label = label,
              metric = metric,
              event_class = event_class,
              class_threshold = class_threshold,
              engine = engine,
              start_col = start_col,
              time_col = time_col,
              status_col = status_col,
              eval_times = eval_times,
              at_risk_threshold = at_risk_threshold,
              survival_metric_convention = survival_metric_convention,
              engine_args = fit_engine_args,
              seed = seed,
              summaryFunction = summaryFunction,
              multiclass_auc = multiclass_auc
            )
            res_summary <- fastml_attach_fold_sizes(
              res_summary,
              if (!is.null(resample_plan)) resample_plan else resamples,
              task
            )
            if (is.null(res_summary) || is.null(res_summary$aggregated)) {
              next
            }
            score <- res_summary$aggregated %>%
              dplyr::filter(.data$.metric == metric) %>%
              dplyr::pull(.data$.estimate)
            if (length(score) > 0) score <- score[1] else score <- NA_real_
            if (is.null(res_summary_best) && !is.na(score)) {
              res_summary_best <- res_summary
              best_params <- tune_grid_values[g, , drop = FALSE]
            } else if (!is.na(score)) {
              best_score <- res_summary_best$aggregated %>%
                dplyr::filter(.data$.metric == metric) %>%
                dplyr::pull(.data$.estimate)
              if (length(best_score) == 0) best_score <- NA_real_
              choose_new <- if (is.na(best_score)) TRUE else {
                if (lower_is_better) score < best_score else score > best_score
              }
              if (choose_new) {
                res_summary_best <- res_summary
                best_params <- tune_grid_values[g, , drop = FALSE]
              }
            }
          }
          if (!is.null(best_params)) {
            wf <- finalize_workflow(wf, best_params)
          }
        } else if (!is.null(resamples)) {
          res_summary_best <- fastml_guarded_resample_fit(
            workflow_spec = wf,
            resamples = if (!is.null(resample_plan)) resample_plan else resamples,
            original_train_rows = nrow(train_data),
            task = task,
            label = label,
            metric = metric,
            event_class = event_class,
            class_threshold = class_threshold,
            engine = engine,
            start_col = start_col,
            time_col = time_col,
            status_col = status_col,
            eval_times = eval_times,
            at_risk_threshold = at_risk_threshold,
            survival_metric_convention = survival_metric_convention,
            engine_args = fit_engine_args,
            seed = seed,
            summaryFunction = summaryFunction,
            multiclass_auc = multiclass_auc,
            store_fold_models = store_fold_models
          )
          res_summary_best <- fastml_attach_fold_sizes(
            res_summary_best,
            if (!is.null(resample_plan)) resample_plan else resamples,
            task
          )
        }

        if (!is.null(res_summary_best)) {
          if (is.null(resampling_summaries[[algo]])) {
            resampling_summaries[[algo]] <- list()
          }
          resampling_summaries[[algo]][[engine]] <- res_summary_best
        } else if (perform_tuning && !is.null(tune_grid_values) && nrow(tune_grid_values) > 0) {
          # Ensure the workflow is finalized even if resampling failed to return summaries
          wf <- finalize_workflow(wf, tune_grid_values[1, , drop = FALSE])
        }

        models[[algo]] <- fastml_with_seed(seed, function() {
          do.call(
            parsnip::fit,
            c(list(object = wf, data = train_data), fit_engine_args)
          )
        })
      }
    }

    if (length(resampling_summaries) > 0) {
      attr(models, "guarded_resampling") <- resampling_summaries
    }

    if (!is.null(resample_plan)) {
      attr(models, "resampling_plan") <- resample_plan
    }

    # Report failed models prominently
    if (length(failed_models) > 0) {
      message("\n", paste(rep("=", 60), collapse = ""))
      message(sprintf("WARNING: %d model(s) failed to train:", length(failed_models)))
      message(paste(rep("-", 60), collapse = ""))
      for (fm in failed_models) {
        message(sprintf("  - %s (%s): %s", fm$algorithm, fm$engine, fm$reason))
      }
      message(paste(rep("=", 60), collapse = ""), "\n")
      attr(models, "failed_models") <- failed_models
    }

    fastml_emit_determinism_warnings(determinism_warnings)

    return(models)
  } else if (task == "classification") {
    if(is.null(summaryFunction)){
      metrics <- metric_set(
        accuracy_metric,
        kap_metric,
        sens_metric,
        spec_metric,
        precision_metric,
        f_meas_metric,
        roc_auc,
        logloss = logloss_metric,
        brier_score = brier_metric,
        ece = ece_metric
      )
    }else{

      newClassMetric <- new_class_metric(summaryFunction, "maximize")

      assign(metric, newClassMetric)

      metrics <- metric_set(
        accuracy_metric,
        kap_metric,
        sens_metric,
        spec_metric,
        precision_metric,
        f_meas_metric,
        roc_auc,
        logloss = logloss_metric,
        brier_score = brier_metric,
        ece = ece_metric,
        !!sym(metric)
      )


    }
  } else {
    metrics <- metric_set(rmse, rsq, mae)
  }

  # Ensure the requested metric is available in the current metric set
  metric_ids <- if (task == "classification") {
    base_ids <- c("accuracy", "kap", "sens", "spec", "precision", "f_meas", "roc_auc", "logloss", "brier_score", "ece")
    if (!is.null(summaryFunction)) {
      base_ids <- c(base_ids, metric)
    }
    base_ids
  } else if (task == "survival") {
    c("c_index", "uno_c", "ibs", "rmst_diff")
  } else {
    c("rmse", "rsq", "mae")
  }
  if (!(metric %in% metric_ids)) {
    warning(
      sprintf(
        "Metric '%s' not available for task '%s'; defaulting to '%s'.",
        metric, task, metric_ids[1]
      ),
      call. = FALSE
    )
    metric <- metric_ids[1]
  }

  resample_plan <- NULL

  if (!is.null(resamples)) {
    if (fastml_is_resample_plan(resamples)) {
      resample_plan <- fastml_resample_validate(resamples)
    } else if (inherits(resamples, "rset")) {
      resample_plan <- fastml_new_resample_plan(
        splits = resamples,
        method = "custom",
        params = list(source = "user_provided")
      )
    } else {
      stop("'resamples' must be an 'rset' object or a fastml resampling plan")
    }
  } else if (resampling_method == "cv") {
    n_data <- nrow(resample_data)
    if (n_data < 3) {
      stop(
        sprintf(
          "Cross-validation requires at least 3 rows, but your training set only has %d rows. \nUse a different resampling method (e.g. 'boot') or provide more data.",
          n_data
        )
      )
    }
    if (folds >= n_data) {
      old_folds <- folds
      folds <- max(2L, n_data - 1L)
      warning(
        sprintf(
          "Reducing folds from %d to %d because the training set has only %d rows (folds >= n would trigger leave-one-out CV, which is not supported by vfold_cv).",
          old_folds, folds, n_data
        ),
        call. = FALSE
      )
    }
    resamples_obj <- vfold_cv(
      resample_data,
      v = folds,
      repeats = 1,
      strata = if (task == "classification")
        all_of(label)
      else
        NULL
    )
    resample_plan <- fastml_new_resample_plan(
      splits = resamples_obj,
      method = "cv",
      params = list(v = folds, repeats = 1, strata = if (task == "classification") label else NULL)
    )
  } else if (resampling_method == "boot") {
    resamples_obj <- bootstraps(resample_data,
                                times = folds,
                                strata = if (task == "classification")
                                  all_of(label)
                                else
                                  NULL)
    resample_plan <- fastml_new_resample_plan(
      splits = resamples_obj,
      method = "boot",
      params = list(times = folds, strata = if (task == "classification") label else NULL)
    )
  } else if (resampling_method == "validation_split") {
    strata_col <- if (task == "classification") label else NULL
    resamples_obj <- rsample::validation_split(
      resample_data,
      prop = 1 - 1 / folds,
      strata = if (!is.null(strata_col)) all_of(strata_col) else NULL
    )
    resample_plan <- fastml_new_resample_plan(
      splits = resamples_obj,
      method = "validation_split",
      params = list(prop = 1 - 1 / folds, strata = strata_col)
    )
  } else if (resampling_method == "repeatedcv") {
    n_data <- nrow(resample_data)
    if (n_data < 3) {
      stop(
        sprintf(
          "Cross-validation requires at least 3 rows, but your training set only has %d rows. \nUse a different resampling method (e.g. 'boot') or provide more data.",
          n_data
        )
      )
    }
    if (folds >= n_data) {
      old_folds <- folds
      folds <- max(2L, n_data - 1L)
      warning(
        sprintf(
          "Reducing folds from %d to %d because the training set has only %d rows (folds >= n would trigger leave-one-out CV, which is not supported by vfold_cv).",
          old_folds, folds, n_data
        ),
        call. = FALSE
      )
    }
    repeats_val <- if (is.null(repeats)) 1L else repeats
    resamples_obj <- vfold_cv(
      resample_data,
      v = folds,
      repeats = repeats_val,
      strata = if (task == "classification")
        all_of(label)
      else
        NULL
    )
    resample_plan <- fastml_new_resample_plan(
      splits = resamples_obj,
      method = "repeatedcv",
      params = list(v = folds, repeats = repeats_val, strata = if (task == "classification") label else NULL)
    )
  } else if (resampling_method == "grouped_cv") {
    repeats_arg <- if (is.null(repeats)) 1L else repeats
    resamples_obj <- make_grouped_cv(
      data = resample_data,
      group_cols = group_cols,
      v = folds,
      strata = if (task == "classification") label else NULL,
      repeats = repeats_arg
    )
    resample_plan <- fastml_new_resample_plan(
      splits = resamples_obj,
      method = "grouped_cv",
      params = list(
        group_cols = group_cols,
        v = folds,
        repeats = repeats_arg,
        strata = if (task == "classification") label else NULL
      )
    )
  } else if (resampling_method %in% c("blocked_cv", "rolling_origin")) {
    if (is.null(block_col) || length(block_col) != 1) {
      stop("'block_col' must be provided when using blocked or rolling resampling.")
    }

    if (resampling_method == "blocked_cv") {
      blocked_info <- make_blocked_cv(
        data = resample_data,
        block_col = block_col,
        block_size = block_size
      )

      if (!is.numeric(folds) || length(folds) != 1 || folds < 2 || folds != as.integer(folds)) {
        stop("'folds' must be an integer greater than or equal to 2 for 'blocked_cv'.")
      }

      if (folds > blocked_info$n_blocks) {
        stop("'folds' cannot exceed the number of derived blocks.")
      }

      block_group_col <- ".fastml_block"
      blocked_data <- blocked_info$data
      blocked_data[[block_group_col]] <- blocked_info$block_ids

      resamples_obj <- rsample::group_vfold_cv(
        blocked_data,
        group = !!rlang::sym(block_group_col),
        v = folds
      )

      resamples_obj$splits <- lapply(
        resamples_obj$splits,
        function(split) {
          split$data[[block_group_col]] <- NULL
          split
        }
      )

      resample_plan <- fastml_new_resample_plan(
        splits = resamples_obj,
        method = "blocked_cv",
        params = list(
          block_col = block_col,
          block_size = block_size,
          v = folds
        )
      )
    } else {
      resamples_obj <- make_rolling_origin_cv(
        data = resample_data,
        block_col = block_col,
        initial_window = initial_window,
        assess_window = assess_window,
        skip = skip
      )
      resample_plan <- fastml_new_resample_plan(
        splits = resamples_obj,
        method = "rolling_origin",
        params = list(
          block_col = block_col,
          initial_window = initial_window,
          assess_window = assess_window,
          skip = skip
        )
      )
    }
  } else if (resampling_method == "nested_cv") {
    if (is.null(outer_folds)) {
      stop("'outer_folds' must be provided when using 'nested_cv'.")
    }
    if (!is.numeric(outer_folds) || length(outer_folds) != 1 || outer_folds < 2 ||
        outer_folds != as.integer(outer_folds)) {
      stop("'outer_folds' must be an integer greater than or equal to 2.")
    }
    inner_folds <- if (is.null(folds)) 10 else folds
    if (!is.numeric(inner_folds) || length(inner_folds) != 1 || inner_folds < 2 ||
        inner_folds != as.integer(inner_folds)) {
      stop("'folds' must be an integer greater than or equal to 2 for inner resampling.")
    }
    resamples_obj <- make_nested_cv(
      data = resample_data,
      outer_folds = outer_folds,
      inner_folds = inner_folds,
      group_cols = group_cols,
      strata = if (task == "classification") label else NULL
    )
    resample_plan <- fastml_new_resample_plan(
      splits = resamples_obj,
      method = "nested_cv",
      params = list(
        outer_folds = outer_folds,
        inner_folds = inner_folds,
        group_cols = group_cols,
        strata = if (task == "classification") label else NULL
      )
    )
  } else if (resampling_method == "none") {
    resample_plan <- NULL
  }

  resamples <- if (!is.null(resample_plan)) fastml_resample_splits(resample_plan) else NULL

  if (is.null(resamples) && (use_default_tuning || !is.null(tune_params)) &&
      !identical(tuning_strategy, "none")) {
    warning("Tuning is skipped because no resamples were supplied (set resampling_method or provide resamples).")
  }

  models <- list()
  failed_models <- list()
  resampling_summaries <- list()
  resample_method_meta <- if (!is.null(resample_plan)) {
    fastml_resample_method(resample_plan) %||% resampling_method
  } else {
    resampling_method
  }

  nested_mode <- identical(resample_method_meta, "nested_cv")
  nested_details <- list()

  # A helper function to choose the engine for an algorithm
  get_engine <- function(algo, fastml_default_engine) {
    # If user explicitly specified engine, use it without warning
    if (!is.null(algorithm_engines) && !is.null(algorithm_engines[[algo]])) {
      if (verbose) {
        message(sprintf("[%s] Using user-specified engine: '%s'", algo, algorithm_engines[[algo]]))
      }
      return(algorithm_engines[[algo]])
    }

    # Determine which default to use
    if (use_parsnip_defaults) {
      parsnip_default <- get_parsnip_default_engine(algo, task)
      if (!is.null(parsnip_default)) {
        if (verbose) {
          message(sprintf("[%s] Using parsnip default engine: '%s'", algo, parsnip_default))
        }
        return(parsnip_default)
      }
      # Fall back to fastml default if no parsnip default exists
      if (verbose) {
        message(sprintf("[%s] No parsnip default found, using fastml default: '%s'", algo, fastml_default_engine))
      }
      return(fastml_default_engine)
    }

    # Using fastml default - warn if it differs from parsnip
    if (warn_engine_defaults) {
      warn_default_override(algo, task, fastml_default_engine, verbose = verbose)
    } else if (verbose) {
      message(sprintf("[%s] Using fastml default engine: '%s'", algo, fastml_default_engine))
    }
    return(fastml_default_engine)
  }

  update_params <- function(params_model, new_params) {
    for (param_name in names(new_params)) {
      param_value <- new_params[[param_name]]
      param_row <- params_model %>% dplyr::filter(id == param_name)
      if (nrow(param_row) == 0) next

      param_obj <- param_row$object[[1]]

      # Helper function to update a parameter object
      try_update <- function(obj, value) {
        if (length(value) == 2) {
          if (inherits(obj, "integer_parameter")) {
            return(obj %>% dials::range_set(c(as.integer(value[1]), as.integer(value[2]))))
          } else {
            return(obj %>% dials::range_set(value))
          }
        } else {
          if (inherits(obj, "integer_parameter")) {
            return(obj %>% dials::value_set(as.integer(value)))
          } else {
            return(obj %>% dials::value_set(value))
          }
        }
      }

      updated_obj <- tryCatch({
        try_update(param_obj, param_value)
      }, error = function(e) {
        # If update fails, expand the allowed range to include the new value.
        current_lb <- attr(param_obj, "range")$lower
        current_ub <- attr(param_obj, "range")$upper

        # Ensure new value(s) are numeric
        if (length(param_value) == 1) {
          new_val <- if (inherits(param_obj, "integer_parameter")) as.integer(param_value) else param_value
        } else {
          new_val <- c(min(param_value), max(param_value))
        }

        # Compute new bounds that include the new value(s)
        if (length(new_val) == 1) {
          new_lb <- min(current_lb, new_val)
          new_ub <- max(current_ub, new_val)
        } else {
          new_lb <- min(current_lb, new_val[1])
          new_ub <- max(current_ub, new_val[2])
        }

        # Expand the range and update the parameter object
        param_obj %>% dials::range_set(c(new_lb, new_ub))
      })

      params_model$object[params_model$id == param_name] <- list(updated_obj)
    }
    return(params_model)
  }

  predictors_all_numeric <- function(df) {
    if (is.null(df) || ncol(df) == 0) {
      return(TRUE)
    }
    is_numeric <- function(col) is.numeric(col) || is.integer(col) || is.logical(col)
    all(vapply(df, is_numeric, logical(1)))
  }

  # NOTE: baked_predictors cache is used ONLY for tuning parameter finalization

  # (e.g., determining mtry range based on post-preprocessing column count).
  # This is a known limitation of dials::finalize() which needs the predictor
  # structure before model training. The actual model training uses workflows
  # which handle per-fold preprocessing correctly during resampling.
  baked_predictors <- NULL
  baked_predictors_ready <- FALSE

  get_baked_predictors <- function() {
    if (baked_predictors_ready) {
      return(baked_predictors)
    }
    baked_predictors_ready <<- TRUE
    if (is.null(recipe)) {
      baked_predictors <<- NULL
      return(NULL)
    }
    baked <- tryCatch(
      recipes::bake(recipe, new_data = train_data),
      error = function(e) NULL
    )
    if (is.null(baked)) {
      baked <- tryCatch({
        prepped <- recipes::prep(recipe, training = train_data, retain = TRUE)
        recipes::bake(prepped, new_data = NULL)
      }, error = function(e) NULL)
    }
    if (!is.null(baked) && !is.null(label) && label %in% names(baked)) {
      baked <- baked[, setdiff(names(baked), label), drop = FALSE]
    }
    baked_predictors <<- baked
    baked
  }

  finalize_predictors <- function(raw_predictors) {
    if (predictors_all_numeric(raw_predictors)) {
      return(raw_predictors)
    }
    baked <- get_baked_predictors()
    if (!is.null(baked) && predictors_all_numeric(baked)) {
      return(baked)
    }
    mm <- tryCatch(
      stats::model.matrix(~ . - 1, data = raw_predictors),
      error = function(e) NULL
    )
    if (!is.null(mm)) {
      return(as.data.frame(mm))
    }
    raw_predictors
  }

  n_class <- length(levels(train_data[[label]]))

  # Validate that there are at least 2 classes for classification
  if (n_class < 2) {
    # Check actual unique values in case factor levels are misleading
    actual_classes <- length(unique(na.omit(train_data[[label]])))
    if (actual_classes < 2) {
      stop(
        sprintf(
          "Classification requires at least 2 classes, but the outcome '%s' has only %d unique value(s). ",
          label, actual_classes
        ),
        "Check your data for class imbalance or filtering issues.",
        call. = FALSE
      )
    }
  }

  if (n_class == 2 && "multinom_reg" %in% algorithms) {
    warning(
      sprintf(
        "Multinomial regression ('multinom_reg') is not applicable to two-class outcomes. Detected a binary outcome for '%s'; using logistic regression ('logistic_reg') instead.",
        label
      ),
      call. = FALSE
    )
    algorithms[algorithms == "multinom_reg"] <- "logistic_reg"
    algorithms <- algorithms[!duplicated(algorithms)]
  } else if (n_class > 2 && "logistic_reg" %in% algorithms) {
    algorithms[algorithms == "logistic_reg"] <- "multinom_reg"
    algorithms <- algorithms[!duplicated(algorithms)]
    if (!is.null(engine_params[["logistic_reg"]]) &&
        is.null(engine_params[["multinom_reg"]])) {
      engine_params[["multinom_reg"]] <- engine_params[["logistic_reg"]]
    }
  }

  for (algo_index in seq_along(algorithms)) {
    algo <- algorithms[[algo_index]]
    set.seed(seed)

    engines <- get_engine(algo, get_default_engine(algo, task))


    # Create a nested list to store models by algorithm and engine.
    models[[algo]] <- list()

    # Loop over each engine provided
    for (engine_index in seq_along(engines)) {
      engine <- engines[[engine_index]]
      tuning_seed_base <- fastml_tuning_seed(seed, offset = algo_index * 100L + engine_index)
      engine_args <- resolve_engine_params(engine_params, algo, engine)
      engine_args <- fastml_apply_engine_seed(engine_args, algo, engine, seed, engine_threads_val, task)
      add_determinism_warning(algo, engine, engine_args)

      # Get default parameters for this engine
      if (use_default_tuning) {
        defaults <- get_default_tune_params(
          algo,
          train_data,
          label,
          engine
        )
      } else {
        defaults <- get_default_params(
          algo,
          task,
          num_predictors = ncol(train_data %>% dplyr::select(-!!sym(label))),
          engine = engine
        )
      }

      # User supplied tuning parameters for this algorithm/engine
      user_params <- NULL
      if (!is.null(tune_params) &&
          !is.null(tune_params[[algo]]) &&
          !is.null(tune_params[[algo]][[engine]])) {
        user_params <- tune_params[[algo]][[engine]]
      }

      # Merge defaults with user parameters
      engine_tune_params <- if (is.null(defaults)) list() else defaults
      if (!is.null(user_params)) {
        for (nm in names(user_params)) {
          engine_tune_params[[nm]] <- user_params[[nm]]
        }
      }

      if (algo == "logistic_reg" && engine %in% c("glm", "gee", "glmer", "stan", "stan_glmer")) {
        perform_tuning <- FALSE
      } else {
        has_custom <- !is.null(user_params)
        perform_tuning <- (use_default_tuning || has_custom) && !is.null(resamples)
        if (tuning_strategy == "none") {
          perform_tuning <- FALSE
        }
      }

        spec_context <- list(
          task = task,
          train_data = train_data,
          label = label,
          tuning = perform_tuning,
          engine = engine,
          early_stopping = early_stopping,
          n_class = n_class
        )
        model_info <- fastml_get_model_spec(algo, spec_context)
        if (is.null(model_info)) {
          next
        }

        # Assume the model specification is stored in model_info$model_spec
        model_spec <- model_info$model_spec

      if(!is.null(model_spec)){

      # Set up tuning parameters and grid (if needed)
      tune_params_model <- NULL
      tune_params_template <- NULL
      tune_grid_values <- NULL

      if (perform_tuning) {
        param_source <- if (inherits(model_spec, "model_spec")) {
          model_spec
        } else {
          model_spec[[1]]
        }

        param_set <- extract_parameter_set_dials(param_source)

        if (nested_mode) {
          tune_params_template <- param_set
        } else {
          raw_predictors <- train_data %>% dplyr::select(-dplyr::all_of(label))
          tune_params_model <- finalize(
            param_set,
            x = finalize_predictors(raw_predictors)
          )

          if (!is.null(engine_tune_params)) {
            tune_params_model <- update_params(tune_params_model, engine_tune_params)
          }

          if (nrow(tune_params_model) > 0 && tuning_strategy == "grid" && !adaptive) {
            tune_grid_values <- grid_regular(tune_params_model, levels = grid_levels)
          }
        }
      }

      do_tuning <- perform_tuning && !all(vapply(engine_tune_params, is.null, logical(1)))

      # Apply engine_args to model_spec via set_engine before creating workflow.
      # parsnip::set_engine() replaces the engine arguments already attached to a
      # specification rather than adding to them, so calling it again here would
      # silently discard the defaults the spec builder set -- LightGBM's
      # counts/bagging_freq/verbose, an xgboost early-stopping configuration, or
      # sparsediscrim's regularization_method. Merge instead, letting the values
      # assembled above win only on the keys they actually define.
      model_spec_base <- if (inherits(model_spec, "model_spec")) model_spec else model_spec[[1]]
      if (length(engine_args) > 0) {
        merged_args <- fastml_merge_engine_args(model_spec_base$eng_args, engine_args)
        model_spec_base <- do.call(
          parsnip::set_engine,
          c(list(model_spec_base, engine = engine), merged_args)
        )
      }

      # Create the workflow
      workflow_spec <- workflow() %>%
        add_model(model_spec_base) %>%
        add_recipe(recipe)

      # Fit the model (with tuning if requested)
      tryCatch({
        allow_par <- allow_par_base
        my_metrics <- NULL

        if (do_tuning) {
          if (algo == "rand_forest" && engine == "h2o") {

            roc_auc_h2o <- function(data, truth, ...) {
              # Rename probability columns from ".pred_p0"/".pred_p1" to ".pred_0"/".pred_1"
              data <- data %>%
                rename_with(~ sub("^\\.pred_p", ".pred_", .x), starts_with(".pred_p"))

              # Call the built-in roc_auc() with the renamed columns
                roc_auc(data, truth = {{truth}}, ...)
            }

            # Assign the same metadata as roc_auc() so metric ids stay consistent
            class(roc_auc_h2o) <- class(roc_auc)
            attr(roc_auc_h2o, "direction") <- attr(yardstick::roc_auc, "direction")
            attr(roc_auc_h2o, "metric_name") <- attr(yardstick::roc_auc, "metric_name")

            my_metrics <- metric_set(
              accuracy_metric,
              kap_metric,
              sens_metric,
              spec_metric,
              precision_metric,
              f_meas_metric,
              roc_auc = roc_auc_h2o,
              logloss = logloss_metric,
              brier_score = brier_metric,
              ece = ece_metric
            )

            allow_par <- FALSE
          }

          else if(engine == "LiblineaR"){

            my_metrics <- metric_set(accuracy_metric, kap_metric, sens_metric, spec_metric, precision_metric, f_meas_metric)
            allow_par <- allow_par_base

          }else{
            allow_par <- allow_par_base
            my_metrics <- NULL
          }
        }

        ctrl_grid <- control_grid(save_pred = TRUE, allow_par = allow_par)
        ctrl_bayes <- if (!is.null(tuning_seed_base)) {
          control_bayes(save_pred = TRUE, seed = tuning_seed_base)
        } else {
          control_bayes(save_pred = TRUE)
        }
        ctrl_race <- control_race(save_pred = TRUE)

        if (early_stopping && tuning_strategy == "bayes") {
          ctrl_bayes <- if (!is.null(tuning_seed_base)) {
            control_bayes(save_pred = TRUE, no_improve = 5, seed = tuning_seed_base)
          } else {
            control_bayes(save_pred = TRUE, no_improve = 5)
          }
        }

        if (nested_mode) {
          nested_fit <- fastml_run_nested_cv(
            workflow_spec = workflow_spec,
            nested_resamples = resamples,
            train_data = train_data,
            label = label,
            task = task,
            metric = metric,
            metrics = metrics,
            engine = engine,
            engine_args = engine_args,
            tune_params_template = if (do_tuning) tune_params_template else NULL,
            engine_tune_params = engine_tune_params,
            tuning_strategy = tuning_strategy,
            tuning_iterations = tuning_iterations,
            adaptive = adaptive,
            early_stopping = early_stopping,
            ctrl_grid = ctrl_grid,
            ctrl_bayes = ctrl_bayes,
            ctrl_race = ctrl_race,
            my_metrics = my_metrics,
            do_tuning = do_tuning,
            event_class = event_class,
            class_threshold = class_threshold,
            start_col = start_col,
            time_col = time_col,
            status_col = status_col,
            eval_times = eval_times,
            at_risk_threshold = at_risk_threshold,
            seed = tuning_seed_base,
            update_params_fn = update_params,
            multiclass_auc = multiclass_auc,
            survival_metric_convention = survival_metric_convention,
            grid_levels = grid_levels
          )

          model <- nested_fit$final_model

          if (!is.null(nested_fit$details)) {
            if (is.null(nested_details[[algo]])) {
              nested_details[[algo]] <- list()
            }
            nested_details[[algo]][[engine]] <- nested_fit$details
          }
        } else if (do_tuning) {
          # Set up control objects for tuning
          if (is.null(resamples)) {
            stop("Tuning cannot be performed without resamples.")
          }

          # Select tuning function based on strategy
          if (tuning_strategy == "bayes") {
            model_tuned <- fastml_with_seed(tuning_seed_base, function() {
              tune_bayes(
                object = workflow_spec,
                resamples = resamples,
                param_info = tune_params_model,
                iter = tuning_iterations,
                metrics = if(!is.null(my_metrics)) my_metrics else metrics,
                control = ctrl_bayes
              )
            })
          } else if (adaptive) {
            model_tuned <- fastml_with_seed(tuning_seed_base, function() {
              tune_race_anova(
                object = workflow_spec,
                resamples = resamples,
                param_info = tune_params_model,
                grid = if (is.null(tune_grid_values)) 20 else tune_grid_values,
                metrics = if(!is.null(my_metrics)) my_metrics else metrics,
                control = ctrl_race
              )
            })
          } else if (tuning_strategy == "grid") {
            if (is.null(tune_grid_values)) {
              tune_grid_values <- grid_regular(tune_params_model, levels = grid_levels)
            }
            model_tuned <- fastml_with_seed(tuning_seed_base, function() {
              tune::tune_grid(
                object = workflow_spec,
                resamples = resamples,
                grid = tune_grid_values,
                metrics = if(!is.null(my_metrics)) my_metrics else metrics,
                control = ctrl_grid
              )
            })
          } else {
            model_tuned <- fastml_with_seed(tuning_seed_base, function() {
              tune::tune_grid(
                object = workflow_spec,
                resamples = resamples,
                grid = if (is.null(tune_grid_values)) 5 else tune_grid_values,
                metrics = if(!is.null(my_metrics)) my_metrics else metrics,
                control = ctrl_grid
              )
            })
          }

          best_metric_used <- metric
          best_params <- tryCatch(
            select_best(model_tuned, metric = metric),
            error = function(e) NULL
          )
          if (is.null(best_params) || !is.data.frame(best_params) || nrow(best_params) == 0) {
            metrics_tbl <- tryCatch(
              tune::collect_metrics(model_tuned),
              error = function(e) NULL
            )
            if (!is.null(metrics_tbl) && ".metric" %in% names(metrics_tbl)) {
              candidate_metrics <- unique(metrics_tbl$.metric)
              candidate_metrics <- setdiff(candidate_metrics, metric)
              for (cand in candidate_metrics) {
                cand_best <- tryCatch(
                  select_best(model_tuned, metric = cand),
                  error = function(e) NULL
                )
                if (!is.null(cand_best) && nrow(cand_best) > 0) {
                  best_params <- cand_best
                  best_metric_used <- cand
                  break
                }
              }
            }
          }

          if (!identical(best_metric_used, metric)) {
            warning(
              sprintf(
                "Metric '%s' unavailable during tuning for %s (%s); selecting parameters using '%s'.",
                metric, algo, engine, best_metric_used
              ),
              call. = FALSE
            )
          }

          fallback_workflow <- NULL
          if (is.null(best_params) || !is.data.frame(best_params) || nrow(best_params) == 0) {
            fallback_grid <- tune_grid_values
            if (is.null(fallback_grid) || nrow(fallback_grid) == 0) {
              fallback_grid <- tryCatch(
                grid_regular(tune_params_model, levels = grid_levels),
                error = function(e) NULL
              )
            }
            if (!is.null(fallback_grid) && nrow(fallback_grid) > 0) {
              best_params <- fallback_grid[1, , drop = FALSE]
              warning(
                sprintf(
                  "Tuning failed to select parameters for %s (%s); using the first available grid configuration.",
                  algo, engine
                ),
                call. = FALSE
              )
            } else {
              warning(
                sprintf(
                  "Tuning failed to select parameters for %s (%s); refitting with default settings.",
                  algo, engine
                ),
                call. = FALSE
              )
              fallback_context <- spec_context
              fallback_context$tuning <- FALSE
              fallback_info <- fastml_get_model_spec(algo, fallback_context)
              fallback_spec <- if (!is.null(fallback_info)) fallback_info$model_spec else NULL
              if (!is.null(fallback_spec)) {
                fallback_spec_base <- if (inherits(fallback_spec, "model_spec")) {
                  fallback_spec
                } else {
                  fallback_spec[[1]]
                }
                if (length(engine_args) > 0) {
                  fallback_spec_base <- do.call(
                    parsnip::set_engine,
                    c(list(fallback_spec_base, engine = engine), engine_args)
                  )
                }
                fallback_workflow <- workflow() %>%
                  add_model(fallback_spec_base) %>%
                  add_recipe(recipe)
              }
            }
          }

          if (!is.null(best_params) && is.data.frame(best_params) && nrow(best_params) > 0) {
            final_workflow <- finalize_workflow(workflow_spec, best_params)
          } else if (!is.null(fallback_workflow)) {
            final_workflow <- fallback_workflow
          } else {
            stop(
              sprintf(
                "Tuning failed for %s (%s) and no fallback configuration is available.",
                algo, engine
              )
            )
          }
          if (!nested_mode && !is.null(resamples)) {
            res_summary <- fastml_guarded_resample_fit(
              workflow_spec = final_workflow,
              resamples = if (!is.null(resample_plan)) resample_plan else resamples,
              original_train_rows = nrow(train_data),
              task = task,
              label = label,
              metric = metric,
              event_class = event_class,
              class_threshold = class_threshold,
              engine = engine,
              start_col = start_col,
              time_col = time_col,
              status_col = status_col,
              eval_times = eval_times,
              at_risk_threshold = at_risk_threshold,
              survival_metric_convention = survival_metric_convention,
              engine_args = list(),  # engine_args already embedded in workflow via set_engine()
              seed = tuning_seed_base,
              summaryFunction = summaryFunction,
              multiclass_auc = multiclass_auc,
              store_fold_models = store_fold_models
            )
            res_summary <- fastml_attach_fold_sizes(
              res_summary,
              if (!is.null(resample_plan)) resample_plan else resamples,
              task
            )
            if (!is.null(res_summary)) {
              if (is.null(resampling_summaries[[algo]])) {
                resampling_summaries[[algo]] <- list()
              }
              resampling_summaries[[algo]][[engine]] <- res_summary
            }
          }
          model <- fastml_with_seed(tuning_seed_base, function() {
            parsnip::fit(final_workflow, data = train_data)
          })
        } else {
          # Use the workflow with engine parameters already applied
          final_workflow <- workflow_spec

          if (!nested_mode && !is.null(resamples)) {
            res_summary <- fastml_guarded_resample_fit(
              workflow_spec = final_workflow,
              resamples = if (!is.null(resample_plan)) resample_plan else resamples,
              original_train_rows = nrow(train_data),
            task = task,
            label = label,
            metric = metric,
            event_class = event_class,
            class_threshold = class_threshold,
            engine = engine,
              start_col = start_col,
              time_col = time_col,
              status_col = status_col,
              eval_times = eval_times,
              at_risk_threshold = at_risk_threshold,
              survival_metric_convention = survival_metric_convention,
              seed = tuning_seed_base,
              summaryFunction = summaryFunction,
              multiclass_auc = multiclass_auc,
              store_fold_models = store_fold_models
            )
            res_summary <- fastml_attach_fold_sizes(
              res_summary,
              if (!is.null(resample_plan)) resample_plan else resamples,
              task
            )
            if (!is.null(res_summary)) {
              if (is.null(resampling_summaries[[algo]])) {
                resampling_summaries[[algo]] <- list()
              }
              resampling_summaries[[algo]][[engine]] <- res_summary
            }
          }

          model <- fastml_with_seed(tuning_seed_base, function() {
            parsnip::fit(final_workflow, data = train_data)
          })
        }
        # Save the fitted model in the nested list under the current engine
        models[[algo]][[engine]] <- model
      }, error = function(e) {
        failed_models[[length(failed_models) + 1]] <<- list(
          algorithm = algo,
          engine = engine,
          reason = e$message
        )
        warning(paste("Training failed for algorithm:", algo, "with engine:", engine,
                      "\nError message:", e$message))
      })

      }else{

        models[[algo]] = NULL
      }

    }  # end of loop over engines

  }

  if (length(models) == 0) {
    stop("No models were successfully trained. Please check your data and parameters.")
  }

  if (length(resampling_summaries) > 0) {
    attr(models, "guarded_resampling") <- resampling_summaries
  }

  if (nested_mode && length(nested_details) > 0) {
    attr(models, "nested_cv_results") <- nested_details
  }

  if (!is.null(resample_plan)) {
    attr(models, "resampling_plan") <- resample_plan
  }

  # Report failed models prominently
  if (length(failed_models) > 0) {
    message("\n", paste(rep("=", 60), collapse = ""))
    message(sprintf("WARNING: %d model(s) failed to train:", length(failed_models)))
    message(paste(rep("-", 60), collapse = ""))
    for (fm in failed_models) {
      message(sprintf("  - %s (%s): %s", fm$algorithm, fm$engine, fm$reason))
    }
    message(paste(rep("=", 60), collapse = ""), "\n")
    attr(models, "failed_models") <- failed_models
  }

  fastml_emit_determinism_warnings(determinism_warnings)

  return(models)
}

library(dplyr)
library(testthat)
library(survival)

# Basic workflow smoke tests to cover core paths

test_that("fastml runs a minimal classification path", {
  skip_on_cran()
  required <- c("parsnip", "workflows", "recipes", "yardstick", "tune", "rsample", "dials")
  lapply(required, skip_if_not_installed)

  data <- iris[iris$Species != "setosa", , drop = FALSE]
  data$Species <- factor(data$Species)

  set.seed(123)
  fit <- fastml(
    data = data,
    label = "Species",
    algorithms = "logistic_reg",
    resampling_method = "none",
    stratify = FALSE,
    use_default_tuning = FALSE,
    tuning_strategy = "none",
    verbose = FALSE
  )

  expect_s3_class(fit, "fastml")
  expect_true("logistic_reg (glm)" %in% names(fit$models))

  grDevices::pdf(file = NULL)
  on.exit(grDevices::dev.off(), add = TRUE)
  suppressWarnings(expect_error(plot(fit, type = "bar"), NA))
})

test_that("train_models returns fitted models for classification", {
  skip_on_cran()
  required <- c("parsnip", "workflows", "recipes", "yardstick", "tune", "rsample", "dials")
  lapply(required, skip_if_not_installed)

  train_dat <- iris[iris$Species != "setosa", , drop = FALSE]
  train_dat$Species <- factor(train_dat$Species)
  train_dat <- train_dat[seq_len(40), ]

  recipe_obj <- recipes::recipe(Species ~ ., data = train_dat)

  models <- train_models(
    train_data = train_dat,
    label = "Species",
    task = "classification",
    algorithms = "logistic_reg",
    resampling_method = "none",
    folds = 3,
    repeats = NULL,
    group_cols = NULL,
    block_col = NULL,
    block_size = NULL,
    initial_window = NULL,
    assess_window = NULL,
    skip = 0,
    outer_folds = NULL,
    resamples = NULL,
    tune_params = NULL,
    engine_params = list(),
    metric = "accuracy",
    summaryFunction = NULL,
    seed = 123,
    recipe = recipe_obj,
    use_default_tuning = FALSE,
    tuning_strategy = "none",
    tuning_iterations = 10,
    early_stopping = FALSE,
    adaptive = FALSE,
    algorithm_engines = NULL,
    event_class = "first",
    start_col = NULL,
    time_col = NULL,
    status_col = NULL,
    eval_times = NULL,
    at_risk_threshold = 0.1,
    audit_env = NULL
  )

  expect_true(is.list(models))
  expect_true("logistic_reg" %in% names(models))
})

test_that("process_model evaluates fitted workflow", {
  skip_on_cran()
  required <- c("parsnip", "workflows", "recipes", "yardstick", "tune", "rsample", "dials")
  lapply(required, skip_if_not_installed)

  data <- iris[iris$Species != "setosa", , drop = FALSE]
  data$Species <- factor(data$Species)

  wf <- workflows::workflow() |>
    workflows::add_model(parsnip::logistic_reg() |> parsnip::set_engine("glm")) |>
    workflows::add_recipe(recipes::recipe(Species ~ ., data = data))

  fitted <- parsnip::fit(wf, data = data)

  res <- process_model(
    model_obj = fitted,
    model_id = "logit",
    task = "classification",
    test_data = data,
    label = "Species",
    event_class = "first",
    start_col = NULL,
    time_col = NULL,
    status_col = NULL,
    engine = "glm",
    train_data = data,
    metric = "accuracy",
    eval_times_user = NULL,
    bootstrap_ci = FALSE,
    bootstrap_samples = 10,
    bootstrap_seed = 123,
    at_risk_threshold = 0.1,
    precomputed_predictions = NULL
  )

  expect_true(is.list(res))
  expect_true(all(c("performance", "predictions") %in% names(res)))
  expect_s3_class(res$performance, "tbl_df")
})

test_that("fastml fits a simple survival model", {
  skip_on_cran()

  required <- c(
    "parsnip", "workflows", "recipes", "yardstick",
    "tune", "rsample", "dials", "survival", "censored"
  )
  lapply(required, skip_if_not_installed)

  lung <- survival::lung
  lung <- lung[complete.cases(lung),]

  # fastml expects: 0 = censored, 1 = event
  lung$status <- ifelse(lung$status == 2, 1, 0)

  fit <- fastml(
      data = lung,
      label = c("time", "status"),
      algorithms = "cox_ph",
      resampling_method = "none"
    )

  expect_s3_class(fit, "fastml")
  expect_true("cox_ph (survival)" %in% names(fit$models))
  expect_true(length(fit$performance) >= 1)
})

library(testthat)

test_that("grouped_cv rejects missing group values", {
  skip_if_not_installed("rsample")

  dat <- tibble::tibble(
    group = c("a", NA_character_, "b", "b"),
    value = 1:4
  )

  expect_error(
    fastml:::make_grouped_cv(
      dat,
      group_cols = "group",
      v = 2,
      strata = NULL,
      repeats = 1
    ),
    "missing values",
    ignore.case = TRUE
  )
})

test_that("grouped_cv supports multiple grouping columns", {
  skip_if_not_installed("rsample")

  set.seed(1)
  dat <- tibble::tibble(
    g1 = rep(c("A", "B", "C"), each = 4),
    g2 = rep(rep(1:2, each = 2), times = 3),
    value = rnorm(12)
  )

  resamples <- fastml:::make_grouped_cv(
    dat,
    group_cols = c("g1", "g2"),
    v = 3,
    strata = NULL,
    repeats = 1
  )

  purrr::walk(resamples$splits, function(split) {
    analysis_groups <- unique(rsample::analysis(split)$.fastml_group)
    assessment_groups <- unique(rsample::assessment(split)$.fastml_group)
    expect_length(intersect(analysis_groups, assessment_groups), 0)
  })
})

test_that("blocked_cv validates ordering and block size", {
  set.seed(2)
  dat <- tibble::tibble(
    time = seq_len(10),
    value = rnorm(10)
  )

  info <- fastml:::make_blocked_cv(
    dat,
    block_col = "time",
    block_size = 3
  )

  expect_equal(info$block_ids, c(1, 1, 1, 2, 2, 2, 3, 3, 3, 4))
  expect_equal(info$n_blocks, 4)

  unsorted <- dat[c(2, 1, 3:10), , drop = FALSE]
  expect_error(
    fastml:::make_blocked_cv(unsorted, block_col = "time", block_size = 3),
    "sorted",
    ignore.case = TRUE
  )

  expect_error(
    fastml:::make_blocked_cv(dat, block_col = "time", block_size = 10),
    "fewer than two blocks",
    ignore.case = TRUE
  )
})

test_that("rolling_origin creates leakage-safe expanding windows", {
  skip_if_not_installed("rsample")

  set.seed(3)
  dat <- tibble::tibble(
    time = seq_len(10),
    value = rnorm(10)
  )

  resamples <- fastml:::make_rolling_origin_cv(
    dat,
    block_col = "time",
    initial_window = 4,
    assess_window = 2,
    skip = 0
  )

  expect_equal(nrow(resamples), 5)

  purrr::walk(resamples$splits, function(split) {
    analysis_time <- rsample::analysis(split)$time
    assessment_time <- rsample::assessment(split)$time
    expect_true(max(analysis_time) < min(assessment_time))
    expect_true(all(diff(analysis_time) >= 0))
    expect_true(all(diff(assessment_time) >= 0))
  })

  unsorted <- dat[c(2, 1, 3:10), , drop = FALSE]
  expect_error(
    fastml:::make_rolling_origin_cv(
      unsorted,
      block_col = "time",
      initial_window = 4,
      assess_window = 2,
      skip = 0
    ),
    "sorted",
    ignore.case = TRUE
  )
})

test_that("nested_cv validates fold counts and supports grouped splits", {
  skip_if_not_installed("rsample")

  expect_error(
    fastml:::make_nested_cv(tibble::tibble(x = 1:5, y = 1:5), outer_folds = 1, inner_folds = 2),
    "outer_folds",
    ignore.case = TRUE
  )
  expect_error(
    fastml:::make_nested_cv(tibble::tibble(x = 1:5, y = 1:5), outer_folds = 2, inner_folds = 1),
    "inner_folds",
    ignore.case = TRUE
  )

  set.seed(4)
  dat <- tibble::tibble(
    id = seq_len(24),
    group = rep(paste0("g", 1:8), each = 3),
    x = rnorm(24),
    y = rnorm(24)
  )

  nested_resamples <- fastml:::make_nested_cv(
    dat,
    outer_folds = 3,
    inner_folds = 2,
    group_cols = "group",
    strata = NULL
  )

  expect_s3_class(nested_resamples, "nested_cv")
  expect_equal(nrow(nested_resamples), 3)
  expect_equal(length(nested_resamples$inner_resamples), 3)

  purrr::walk2(
    nested_resamples$splits,
    nested_resamples$inner_resamples,
    function(outer_split, inner_set) {
      outer_analysis <- rsample::analysis(outer_split)
      outer_assessment <- rsample::assessment(outer_split)

      expect_length(
        intersect(unique(outer_analysis$group), unique(outer_assessment$group)),
        0
      )

      purrr::walk(inner_set$splits, function(inner_split) {
        inner_analysis <- rsample::analysis(inner_split)
        inner_assessment <- rsample::assessment(inner_split)

        expect_true(all(inner_analysis$id %in% outer_analysis$id))
        expect_true(all(inner_assessment$id %in% outer_analysis$id))
        expect_length(
          intersect(unique(inner_analysis$group), unique(inner_assessment$group)),
          0
        )
        expect_length(
          intersect(unique(inner_assessment$group), unique(outer_assessment$group)),
          0
        )
      })
    }
  )
})

library(testthat)

# ============================================================================
# Unit tests for bugfix round 2
# ============================================================================

# ---------- Fix #2: event_class validation -----------------------------------

test_that("fastml rejects invalid event_class values", {
  expect_error(
    fastml(
      data = iris,
      label = "Species",
      algorithms = "logistic_reg",
      event_class = "fist"
    ),
    "'arg' should be one of"
  )
  expect_error(
    fastml(
      data = iris,
      label = "Species",
      algorithms = "logistic_reg",
      event_class = "third"
    ),
    "'arg' should be one of"
  )
})

test_that("fastml accepts valid event_class values without validation error", {
  # Valid event_class should pass validation. We trigger a DIFFERENT error
  # (missing label) to prove event_class itself was accepted.
  expect_error(
    fastml(
      data = iris,
      label = "nonexistent_col",
      algorithms = "logistic_reg",
      event_class = "first"
    ),
    "Label variable|not found|not present|does not exist"
  )
  expect_error(
    fastml(
      data = iris,
      label = "nonexistent_col",
      algorithms = "logistic_reg",
      event_class = "second"
    ),
    "Label variable|not found|not present|does not exist"
  )
})

test_that("train_models rejects invalid event_class values", {
  expect_error(
    train_models(
      train_data = iris,
      label = "Species",
      task = "classification",
      algorithms = "logistic_reg",
      event_class = "invalid"
    ),
    "'arg' should be one of"
  )
})

# ---------- Fix #6: dead perf_values statement removed -----------------------
# No behavioral test needed — this was dead code removal.
# Verified by the full test suite passing.

# ---------- Fix #7: unused true_labels removed -------------------------------
# No behavioral test needed — this was dead code removal.
# Verified by the full test suite passing.

# ---------- Fix #4: duplicate switch keys in engine_helpers ------------------

test_that("get_default_engine returns correct engines for previously-duplicated keys", {
  expect_equal(get_default_engine("survreg", "survival"), "survival")
  expect_equal(get_default_engine("royston_parmar", "survival"), "rstpm2")
  expect_equal(get_default_engine("cox_ph", "survival"), "survival")
})

test_that("get_default_engine errors for unknown algorithm", {
  expect_error(
    get_default_engine("nonexistent_algo", "classification"),
    "No default engine"
  )
})

# ---------- Fix #5: security guard avoids package-owned .GlobalEnv writes ----
# The sandbox intercepts direct rm/assign calls to .GlobalEnv. The regression
# guard here ensures the package itself does not write deleted objects back into
# the user's global environment while auditing hook side effects.

test_that("security guard does not restore deleted globals into .GlobalEnv", {
  src_path <- file.path(getwd(), "R", "security_guards.R")
  if (!file.exists(src_path)) {
    src_path <- system.file("R", "security_guards.R", package = "fastml")
  }
  if (!file.exists(src_path) || length(readLines(src_path)) == 0) {
    skip("Cannot locate security_guards.R source")
  }
  src <- readLines(src_path)

  assign_lines <- grep('assign.*pre_values.*\\.GlobalEnv', src)
  expect_length(assign_lines, 0)
})

test_that("added globals can still be cleaned up without restoring deletions", {
  env <- new.env(parent = emptyenv())
  assign("a", 1, envir = env)

  pre_names <- ls(env, all.names = TRUE)
  assign("extra", 99, envir = env)

  post_names <- ls(env, all.names = TRUE)
  added <- setdiff(post_names, pre_names)
  expect_equal(added, "extra")

  rm(list = added, envir = env)

  expect_false(exists("extra", envir = env))
  expect_true(exists("a", envir = env))
})

# ---------- Fix #3: logistic_reg/multinom_reg pre-loop swap ------------------

test_that("in-loop logistic_reg swap was removed from train_models", {
  src_path <- file.path(getwd(), "R", "train_models.R")
  if (!file.exists(src_path)) {
    src_path <- system.file("R", "train_models.R", package = "fastml")
  }
  if (!file.exists(src_path) || length(readLines(src_path)) == 0) {
    skip("Cannot locate train_models.R source")
  }
  src <- readLines(src_path)
  # The old in-loop swap line should no longer exist
  in_loop_swap <- grep(
    'n_class > 2 && algo == "logistic_reg".*algo.*=.*"multinom_reg"',
    src
  )
  expect_length(in_loop_swap, 0)
})

test_that("pre-loop multiclass swap exists in train_models", {
  src_path <- file.path(getwd(), "R", "train_models.R")
  if (!file.exists(src_path)) {
    src_path <- system.file("R", "train_models.R", package = "fastml")
  }
  if (!file.exists(src_path) || length(readLines(src_path)) == 0) {
    skip("Cannot locate train_models.R source")
  }
  src <- readLines(src_path)
  # Pre-loop swap should exist: algorithms[...logistic_reg...] <- "multinom_reg"
  pre_loop_swap <- grep(
    'algorithms\\[algorithms == "logistic_reg"\\] <- "multinom_reg"',
    src
  )
  expect_true(length(pre_loop_swap) > 0)
})

# ---------- Fix #1: selected_outer_id in nested CV --------------------------

test_that("fastml_nested_select_params returns correct best params", {
  # Create mock inner_results with two param configs across two folds
  mock_fold1 <- tibble::tibble(
    .metric = rep("rmse", 2),
    .estimator = rep("standard", 2),
    .estimate = c(1.5, 1.1),
    cost_complexity = c(0.01, 0.001)
  )
  mock_fold2 <- tibble::tibble(
    .metric = rep("rmse", 2),
    .estimator = rep("standard", 2),
    .estimate = c(1.3, 1.2),
    cost_complexity = c(0.01, 0.001)
  )
  inner_results <- list(mock_fold1, mock_fold2)

  result <- fastml_nested_select_params(
    inner_results = inner_results,
    metric = "rmse",
    lower_is_better = TRUE
  )

  expect_true(is.data.frame(result))
  expect_true("cost_complexity" %in% names(result))
  expect_equal(nrow(result), 1)
  # cost_complexity=0.001: avg rmse = (1.1+1.2)/2 = 1.15
  # cost_complexity=0.01:  avg rmse = (1.5+1.3)/2 = 1.40
  # Should pick 0.001
  expect_equal(result$cost_complexity[1], 0.001)
})

test_that("fastml_nested_select_params handles all-NULL inner results", {
  result <- fastml_nested_select_params(
    inner_results = list(NULL, NULL),
    metric = "rmse",
    lower_is_better = TRUE
  )
  expect_null(result)
})

library(testthat)

# Grouped CV tests

test_that("grouped CV keeps groups intact", {
  skip_if_not_installed("rsample")

  set.seed(123)
  grouped_data <- tibble::tibble(
    id = 1:12,
    group = rep(letters[1:4], each = 3),
    strata = rep(c("A", "B"), each = 6),
    value = rnorm(12)
  )

  resamples <- fastml:::make_grouped_cv(
    grouped_data,
    group_cols = "group",
    v = 4,
    strata = NULL,
    repeats = 1
  )

  purrr::walk(resamples$splits, function(split) {
    analysis_groups <- unique(rsample::analysis(split)$group)
    assessment_groups <- unique(rsample::assessment(split)$group)
    expect_length(intersect(analysis_groups, assessment_groups), 0)
  })
})

test_that("grouped CV errors when groups fewer than folds", {
  skip_if_not_installed("rsample")

  small_groups <- tibble::tibble(
    group = rep(letters[1:3], each = 2),
    value = runif(6)
  )

  expect_error(
    fastml:::make_grouped_cv(
      small_groups,
      group_cols = "group",
      v = 4,
      strata = NULL,
      repeats = 1
    ),
    "cannot exceed the number of groups"
  )
})

test_that("grouped CV stratification distributes strata across folds", {
  skip_if_not_installed("rsample")

  set.seed(456)
  strat_data <- tibble::tibble(
    id = 1:12,
    group = rep(paste0("g", 1:6), each = 2),
    strata = rep(c("A", "B"), each = 6),
    value = rnorm(12)
  )

  resamples <- fastml:::make_grouped_cv(
    strat_data,
    group_cols = "group",
    v = 3,
    strata = "strata",
    repeats = 1
  )

  purrr::walk(resamples$splits, function(split) {
    assessment_strata <- unique(rsample::assessment(split)$strata)
    expect_setequal(assessment_strata, unique(strat_data$strata))
  })
})

# Blocked CV tests

test_that("blocked CV preserves order and prevents leakage", {
  skip_if_not_installed("rsample")

  blocked_data <- tibble::tibble(
    time = 1:12,
    value = rnorm(12)
  )

  blocked_info <- fastml:::make_blocked_cv(
    blocked_data,
    block_col = "time",
    block_size = 3
  )

  expect_equal(blocked_info$data$time, blocked_data$time)
  expect_true(all(diff(blocked_info$block_ids) >= 0))

  blocked_plan <- blocked_info$data
  blocked_plan$.fastml_block <- blocked_info$block_ids

  resamples <- rsample::group_vfold_cv(
    blocked_plan,
    group = .fastml_block,
    v = blocked_info$n_blocks
  )

  purrr::walk(resamples$splits, function(split) {
    analysis <- rsample::analysis(split)
    assessment <- rsample::assessment(split)
    expect_true(all(diff(analysis$time) >= 0))
    expect_true(all(diff(assessment$time) >= 0))
    expect_true(all(!assessment$time %in% analysis$time))
  })
})

# Rolling origin tests

test_that("rolling origin uses expanding analysis and fixed assessment windows", {
  skip_if_not_installed("rsample")

  rolling_data <- tibble::tibble(
    time = 1:10,
    value = rnorm(10)
  )

  resamples <- fastml:::make_rolling_origin_cv(
    rolling_data,
    block_col = "time",
    initial_window = 4,
    assess_window = 2,
    skip = 0
  )

  analysis_sizes <- purrr::map_int(resamples$splits, ~ nrow(rsample::analysis(.x)))
  assessment_sizes <- purrr::map_int(resamples$splits, ~ nrow(rsample::assessment(.x)))

  expect_equal(analysis_sizes[1], 4)
  expect_true(all(diff(analysis_sizes) >= 0))
  expect_true(all(assessment_sizes == 2))
})

# Nested CV tests

test_that("nested CV inner folds remain within outer training data", {
  skip_if_not_installed("rsample")

  set.seed(789)
  nested_data <- tibble::tibble(
    id = 1:18,
    x = runif(18),
    y = rnorm(18)
  )

  nested_resamples <- fastml:::make_nested_cv(
    nested_data,
    outer_folds = 3,
    inner_folds = 2,
    group_cols = NULL,
    strata = NULL
  )

  purrr::walk2(
    nested_resamples$splits,
    nested_resamples$inner_resamples,
    function(outer_split, inner_set) {
      outer_analysis_ids <- rsample::analysis(outer_split)$id
      outer_assessment_ids <- rsample::assessment(outer_split)$id

      purrr::walk(inner_set$splits, function(inner_split) {
        inner_analysis_ids <- rsample::analysis(inner_split)$id
        inner_assessment_ids <- rsample::assessment(inner_split)$id

        expect_true(all(inner_analysis_ids %in% outer_analysis_ids))
        expect_true(all(inner_assessment_ids %in% outer_analysis_ids))
        expect_length(intersect(inner_assessment_ids, outer_assessment_ids), 0)
      })
    }
  )
})

test_that("nested CV tuning uses inner folds and outer metrics match manual evaluation", {
  skip_on_cran()
  skip_if_not_installed("rsample")
  skip_if_not_installed("workflows")
  skip_if_not_installed("recipes")
  skip_if_not_installed("parsnip")
  skip_if_not_installed("yardstick")
  skip_if_not_installed("tune")
  skip_if_not_installed("dials")
  skip_if_not_installed("rpart")
  if (requireNamespace("censored", quietly = TRUE) &&
      !requireNamespace("pec", quietly = TRUE)) {
    skip("Package 'pec' is required for decision_tree when censored is installed.")
  }

  set.seed(101)
  tuning_data <- tibble::tibble(
    id = 1:24,
    x = runif(24, -1, 1),
    y = 3 * x + rnorm(24, sd = 0.1)
  )

  nested_resamples <- fastml:::make_nested_cv(
    tuning_data,
    outer_folds = 3,
    inner_folds = 3,
    group_cols = NULL,
    strata = NULL
  )

  recipe_spec <- recipes::recipe(y ~ x, data = tuning_data)
  model_spec <- parsnip::decision_tree(
    cost_complexity = tune::tune(),
    tree_depth = tune::tune()
  ) %>%
    parsnip::set_mode("regression") %>%
    parsnip::set_engine("rpart")

  workflow_spec <- workflows::workflow() %>%
    workflows::add_recipe(recipe_spec) %>%
    workflows::add_model(model_spec)

  metrics <- yardstick::metric_set(yardstick::rmse)

  nested_fit <- fastml:::fastml_run_nested_cv(
    workflow_spec = workflow_spec,
    nested_resamples = nested_resamples,
    train_data = tuning_data,
    label = "y",
    task = "regression",
    metric = "rmse",
    metrics = metrics,
    engine = "rpart",
    engine_args = list(),
    tune_params_template = dials::parameters(
      dials::cost_complexity(),
      dials::tree_depth()
    ),
    engine_tune_params = list(),
    tuning_strategy = "grid",
    tuning_iterations = 5,
    adaptive = FALSE,
    early_stopping = FALSE,
    ctrl_grid = tune::control_grid(save_pred = TRUE),
    ctrl_bayes = tune::control_bayes(save_pred = TRUE),
    ctrl_race = NULL,
    my_metrics = NULL,
    do_tuning = TRUE,
    event_class = NULL,
    class_threshold = 0.5,
    start_col = NULL,
    time_col = NULL,
    status_col = NULL,
    eval_times = NULL,
    at_risk_threshold = 0.1,
    seed = 202,
    update_params_fn = function(params, new) params,
    multiclass_auc = "macro",
    grid_levels = 3L
  )

  inner_results <- nested_fit$details$inner_results

  purrr::walk2(
    nested_resamples$splits,
    seq_along(inner_results),
    function(outer_split, idx) {
      outer_analysis_ids <- rsample::analysis(outer_split)$id
      outer_assessment_ids <- rsample::assessment(outer_split)$id

      tune_splits <- inner_results[[idx]]$splits
      purrr::walk(tune_splits, function(inner_split) {
        inner_analysis_ids <- rsample::analysis(inner_split)$id
        inner_assessment_ids <- rsample::assessment(inner_split)$id

        expect_true(all(inner_analysis_ids %in% outer_analysis_ids))
        expect_true(all(inner_assessment_ids %in% outer_analysis_ids))
        expect_length(intersect(inner_assessment_ids, outer_assessment_ids), 0)
      })
    }
  )

  outer_performance <- nested_fit$details$outer_performance
  expect_equal(sort(unique(outer_performance$outer_id)), sort(nested_resamples$id))

  purrr::walk2(
    nested_resamples$splits,
    seq_along(inner_results),
    function(outer_split, idx) {
      best_params <- tune::select_best(inner_results[[idx]], metric = "rmse")
      finalized <- tune::finalize_workflow(workflow_spec, best_params)
      outer_train <- rsample::analysis(outer_split)
      outer_assess <- rsample::assessment(outer_split)
      fitted <- parsnip::fit(finalized, data = outer_train)
      preds <- predict(fitted, new_data = outer_assess)$.pred
      expected_rmse <- yardstick::rmse_vec(truth = outer_assess$y, estimate = preds)

      fold_id <- nested_resamples$id[idx]
      recorded <- outer_performance %>%
        dplyr::filter(outer_id == fold_id, .metric == "rmse")

      expect_equal(recorded$.estimate, expected_rmse)
    }
  )
})

# Stratification imbalance tests

test_that("stratified CV handles severe class imbalance (95/5)", {
  skip_if_not_installed("rsample")

  set.seed(123)
  # Create 95/5 imbalanced dataset with 100 observations
  n_majority <- 95
  n_minority <- 5
  imbalanced_data <- tibble::tibble(
    id = 1:(n_majority + n_minority),
    x = rnorm(n_majority + n_minority),
    outcome = factor(c(rep("majority", n_majority), rep("minority", n_minority)))
  )

  # Use stratified vfold_cv
  resamples <- rsample::vfold_cv(imbalanced_data, v = 5, strata = outcome)

  # Verify each fold has both classes represented
  purrr::walk(resamples$splits, function(split) {
    analysis_classes <- unique(rsample::analysis(split)$outcome)
    assessment_classes <- unique(rsample::assessment(split)$outcome)

    # Both classes should be present in analysis (training) set
    expect_length(analysis_classes, 2)
    # Assessment set should have at least minority class when possible
    # With 5 folds and 5 minority samples, each fold should have 1 minority
    expect_true("minority" %in% assessment_classes || length(assessment_classes) >= 1)
  })

  # Verify class proportions are approximately maintained (within reason)
  analysis_props <- purrr::map_dbl(resamples$splits, function(split) {
    analysis_data <- rsample::analysis(split)
    sum(analysis_data$outcome == "minority") / nrow(analysis_data)
  })

  # Proportions should be close to original 5%
  expect_true(all(analysis_props > 0.02 & analysis_props < 0.10))
})

test_that("stratified CV handles extreme imbalance (99/1)", {
  skip_if_not_installed("rsample")

  set.seed(456)
  # Create 99/1 imbalanced dataset
  n_majority <- 99
  n_minority <- 1
  extreme_data <- tibble::tibble(
    id = 1:(n_majority + n_minority),
    x = rnorm(n_majority + n_minority),
    outcome = factor(c(rep("majority", n_majority), rep("minority", n_minority)))
  )

  # Stratified CV with minority class < folds should either warn or handle gracefully
  # Different rsample versions may behave differently, so we test it doesn't error
  resamples <- tryCatch(
    rsample::vfold_cv(extreme_data, v = 5, strata = outcome),
    warning = function(w) {
      # If it warns, that's expected behavior
      suppressWarnings(rsample::vfold_cv(extreme_data, v = 5, strata = outcome))
    }
  )

  # The resamples should be created successfully

  expect_true(inherits(resamples, "rset"))
  expect_equal(nrow(resamples), 5)
})

test_that("grouped stratified CV handles imbalanced groups", {
  skip_if_not_installed("rsample")

  set.seed(789)
  # Create grouped data where some groups have only one class
  grouped_imbalanced <- tibble::tibble(
    id = 1:30,
    group = rep(paste0("g", 1:6), each = 5),
    strata = c(
      rep("A", 5),  # g1: all A
      rep("B", 5),  # g2: all B
      c("A", "A", "B", "B", "B"),  # g3: mixed
      c("A", "B", "A", "B", "A"),  # g4: mixed
      c("A", "A", "A", "B", "B"),  # g5: mixed
      rep("A", 5)   # g6: all A
    ),
    value = rnorm(30)
  )

  # When stratifying grouped CV, should handle uneven strata distribution

  # Expect a warning because strata is not constant within groups
  resamples <- expect_warning(
    fastml:::make_grouped_cv(
      grouped_imbalanced,
      group_cols = "group",
      v = 3,
      strata = "strata",
      repeats = 1
    ),
    "Proceeding without stratification"
  )

  # Verify groups remain intact
  purrr::walk(resamples$splits, function(split) {
    analysis_groups <- unique(rsample::analysis(split)$group)
    assessment_groups <- unique(rsample::assessment(split)$group)
    expect_length(intersect(analysis_groups, assessment_groups), 0)
  })
})

#' Internal helpers for xgboost-based survival models
#'
#' These utilities provide training-time summaries (e.g. baseline hazards)
#' and prediction helpers shared between the Cox and AFT objectives exposed
#' by the native xgboost survival integration.  They are not exported.
#'
#' @noRd
NULL

fastml_prepare_xgb_matrix <- function(predictors, feature_names) {
  if (is.null(predictors)) {
    predictors <- data.frame()
  }
  predictors <- as.data.frame(predictors)
  if (length(feature_names) == 0) {
    return(matrix(numeric(), nrow = nrow(predictors), ncol = 0))
  }
  if (ncol(predictors) == 0) {
    mat <- matrix(0, nrow = nrow(predictors), ncol = length(feature_names))
    colnames(mat) <- feature_names
    return(mat)
  }
  for (nm in names(predictors)) {
    col <- predictors[[nm]]
    if (is.factor(col)) {
      predictors[[nm]] <- as.numeric(col)
    } else if (is.logical(col)) {
      predictors[[nm]] <- as.numeric(col)
    }
  }
  mat <- as.matrix(predictors)
  if (is.null(colnames(mat))) {
    colnames(mat) <- paste0("V", seq_len(ncol(mat)))
  }
  present <- intersect(feature_names, colnames(mat))
  if (length(present) > 0) {
    mat <- mat[, present, drop = FALSE]
  } else {
    mat <- matrix(numeric(), nrow = nrow(mat), ncol = 0)
  }
  missing <- setdiff(feature_names, colnames(mat))
  if (length(missing) > 0) {
    if (nrow(mat) == 0) {
      add_cols <- matrix(numeric(), nrow = 0, ncol = length(missing))
    } else {
      add_cols <- matrix(0, nrow = nrow(mat), ncol = length(missing))
    }
    colnames(add_cols) <- missing
    if (ncol(mat) == 0) {
      mat <- add_cols
    } else {
      mat <- cbind(mat, add_cols)
    }
  }
  mat <- mat[, feature_names, drop = FALSE]
  storage.mode(mat) <- "numeric"
  mat
}

#' Compute Breslow Baseline Hazard and Survival
#'
#' Fits a Cox model with an offset linear predictor to estimate the Breslow
#' baseline cumulative hazard and corresponding survival function.
#'
#' @param time_vec Numeric vector of observed times.
#' @param status_vec Numeric event indicator (1 = event, 0 = censored).
#' @param lp_vec Numeric vector of linear predictors.
#'
#' @return A data frame containing \code{time}, \code{cumhaz}, and \code{surv}.
#' Returns \code{NULL} if fitting fails or no valid data remain.
#'
#' @importFrom survival coxph Surv basehaz
#' @keywords internal
#' @noRd
fastml_compute_breslow_baseline <- function(time_vec, status_vec, lp_vec) {
  df <- data.frame(
    time = as.numeric(time_vec),
    status = ifelse(is.na(status_vec), 0, ifelse(status_vec > 0, 1, 0)),
    lp = as.numeric(lp_vec)
  )
  df <- df[is.finite(df$time) & df$time >= 0, , drop = FALSE]
  if (nrow(df) == 0) {
    return(NULL)
  }
  suppressWarnings({
    fit <- survival::coxph(
      survival::Surv(time, status) ~ offset(lp),
      data = df,
      ties = "breslow"
    )
  })
  bh <- tryCatch(survival::basehaz(fit, centered = FALSE), error = function(e) NULL)
  if (is.null(bh) || nrow(bh) == 0) {
    return(NULL)
  }
  bh <- bh[order(bh$time), , drop = FALSE]
  bh$cumhaz <- as.numeric(bh$hazard)
  bh$surv <- exp(-bh$cumhaz)
  bh <- bh[, c("time", "cumhaz", "surv"), drop = FALSE]
  bh
}

#' Predict Linear Predictor from xgboost Booster
#'
#' Generates linear predictor (log-hazard or log-time) estimates from an
#' xgboost survival model, handling differences between Cox and AFT objectives
#' and ensuring shape consistency across xgboost versions.
#'
#' @param fit_obj A fitted xgboost survival model object containing a
#' \code{booster} and \code{feature_names}.
#' @param predictors A data frame or matrix of predictors for prediction.
#'
#' @return A numeric vector of predicted linear predictors (\code{lp}).
#' Returns \code{NA} if prediction fails or model is invalid.
#'
#' @keywords internal
#' @noRd
fastml_xgb_predict_lp <- function(fit_obj, predictors) {
  booster <- fit_obj$booster
  n_obs <- if (!is.null(predictors)) nrow(predictors) else 0L
  if (is.null(booster)) {
    return(rep(NA_real_, n_obs))
  }
  feature_names <- fit_obj$feature_names
  mat <- fastml_prepare_xgb_matrix(predictors, feature_names)
  # Ensure no non-finite values are passed to xgboost
  if (length(mat) > 0L) {
    mat[!is.finite(mat)] <- 0
  }
  if (nrow(mat) == 0 && length(feature_names) > 0) {
    return(rep(NA_real_, n_obs))
  }

  safe_margin_pred <- function(strict_shape = FALSE) {
    tryCatch(
      predict(
        booster,
        newdata = mat,
        outputmargin = TRUE,
        strict_shape = strict_shape
      ),
      error = function(e) NULL
    )
  }

  pred <- safe_margin_pred()
  if (is.null(pred) || length(pred) == 0 || all(!is.finite(pred))) {
    # Some versions of xgboost require strict_shape = TRUE to return a
    # vector when using the AFT objective. Retry with that flag before
    # falling back to non-margin predictions.
    pred <- safe_margin_pred(strict_shape = TRUE)
  }

  if (is.null(pred) || length(pred) == 0 || all(!is.finite(pred))) {
    # As a last resort, request the default prediction (expected survival
    # time) and convert it back to the location parameter on the log scale.
    pred_raw <- tryCatch(
      predict(booster, newdata = mat, strict_shape = TRUE),
      error = function(e) NULL
    )
    if (!is.null(pred_raw) && length(pred_raw) > 0) {
      pred_raw <- as.numeric(pred_raw)
      pred_raw[!is.finite(pred_raw) | pred_raw <= 0] <- NA_real_
      pred <- log(pred_raw)
    }
  }

  # If still missing, try an explicit DMatrix prediction before giving up.
  if (is.null(pred) || length(pred) == 0 || all(!is.finite(pred))) {
    dmat <- tryCatch(xgboost::xgb.DMatrix(data = mat), error = function(e) NULL)
    pred_raw <- if (!is.null(dmat)) {
      tryCatch(predict(booster, newdata = dmat, strict_shape = TRUE),
               error = function(e) NULL)
    } else NULL
    if (!is.null(pred_raw) && length(pred_raw) > 0) {
      pred_raw <- as.numeric(pred_raw)
      pred_raw[!is.finite(pred_raw) | pred_raw <= 0] <- NA_real_
      pred <- log(pred_raw)
    }
  }

  if (is.null(pred) || length(pred) == 0) {
    return(rep(NA_real_, nrow(mat)))
  }

  pred <- as.numeric(pred)
  if (length(pred) != nrow(mat)) {
    # The prediction output should align with the number of rows. If it
    # does not, coerce by recycling or truncating so downstream metrics are
    # well-defined rather than silently returning a mismatched vector.
    pred <- rep(pred, length.out = nrow(mat))
  }
  pred
}


  #' Compute Survival Matrix for Cox xgboost Model
  #'
  #' Calculates survival probabilities at specified times using the Cox model’s
  #' baseline cumulative hazard and predicted linear predictors from an
  #' xgboost-based survival fit.
  #'
  #' @param fit_obj A fitted xgboost Cox model containing a baseline hazard.
  #' @param lp_vec Numeric vector of linear predictors.
  #' @param times Numeric vector of evaluation times.
  #'
#' @return A numeric matrix of survival probabilities with rows corresponding
#' to observations and columns to times.
#'
#' @keywords internal
#' @noRd
fastml_xgb_survival_matrix_cox <- function(fit_obj, lp_vec, times) {
  baseline <- fit_obj$baseline
  if (is.null(baseline) || nrow(baseline) == 0 || length(times) == 0) {
    return(matrix(NA_real_, nrow = length(lp_vec), ncol = length(times)))
  }
  times <- as.numeric(times)
  times[!is.finite(times) | times < 0] <- NA_real_
  base_time <- baseline$time
  base_cumhaz <- baseline$cumhaz
  lookup <- function(t) {
    if (is.na(t)) {
      return(NA_real_)
    }
    if (t <= base_time[1]) {
      return(0)
    }
    idx <- findInterval(t, base_time)
    if (idx <= 0) {
      return(0)
    }
    if (idx > length(base_time)) {
      idx <- length(base_time)
    }
    base_cumhaz[idx]
  }
  cumhaz_vals <- vapply(times, lookup, numeric(1))
  cumhaz_vals[!is.finite(cumhaz_vals) | cumhaz_vals < 0] <- NA_real_
  if (length(lp_vec) == 0) {
    return(matrix(numeric(), nrow = 0, ncol = length(times)))
  }
  mat <- matrix(NA_real_, nrow = length(lp_vec), ncol = length(times))
  for (j in seq_along(times)) {
    ch <- cumhaz_vals[j]
    if (!is.finite(ch) || ch < 0) {
      next
    }
    mat[, j] <- exp(-ch * exp(lp_vec))
  }
  mat[mat < 0] <- 0
  mat[mat > 1] <- 1
  if (ncol(mat) > 1) {
    mat <- t(apply(mat, 1, function(row) {
      row <- cummin(pmin(pmax(row, 0), 1))
      row
    }))
  }
  mat
}

#' Compute Survival Probabilities for AFT Model
#'
#' Calculates survival probabilities for given times, location parameters,
#' and distribution family in accelerated failure time (AFT) models.
#'
#' @param times Numeric vector of evaluation times.
#' @param mu_vec Numeric vector of predicted location parameters.
#' @param dist Character string specifying the AFT distribution
#' (\code{"normal"}, \code{"logistic"}, or \code{"extreme"}).
#' @param scale Numeric scale parameter.
#'
#' @return A numeric matrix of survival probabilities with rows corresponding
#' to observations and columns to evaluation times.
#'
#' @importFrom stats pnorm
#' @keywords internal
#' @noRd
fastml_aft_survival_prob <- function(times, mu_vec, dist, scale) {
  if (length(times) == 0) {
    return(matrix(numeric(), nrow = length(mu_vec), ncol = 0))
  }
  scale <- ifelse(is.finite(scale) && scale > 0, scale, 1)
  times <- as.numeric(times)
  times[times <= 0 | !is.finite(times)] <- NA_real_
  if (length(mu_vec) == 0) {
    return(matrix(numeric(), nrow = 0, ncol = length(times)))
  }
  mat <- matrix(NA_real_, nrow = length(mu_vec), ncol = length(times))
  dist <- tolower(dist)
  for (j in seq_along(times)) {
    tval <- times[j]
    if (!is.finite(tval) || tval <= 0) {
      next
    }
    z <- (log(tval) - mu_vec) / scale
    surv <- switch(dist,
                   "logistic" = 1 / (1 + exp(z)),
                   "extreme" = 1 - exp(-exp(-z)),
                   "normal" = 1 - stats::pnorm(z),
                   1 - stats::pnorm(z))
    surv[!is.finite(surv)] <- NA_real_
    mat[, j] <- surv
  }
  mat[mat < 0] <- 0
  mat[mat > 1] <- 1
  if (ncol(mat) > 1) {
    mat <- t(apply(mat, 1, function(row) cummin(pmin(pmax(row, 0), 1))))
  }
  mat
}

#' Compute Quantile Times for AFT Model
#'
#' Calculates quantile survival times corresponding to specified probabilities
#' for an accelerated failure time (AFT) model with given distribution,
#' location, and scale parameters.
#'
#' @param probs Numeric vector of quantile probabilities (between 0 and 1).
#' @param mu_vec Numeric vector of predicted location parameters.
#' @param dist Character string specifying the AFT distribution
#' (\code{"normal"}, \code{"logistic"}, or \code{"extreme"}).
#' @param scale Numeric scale parameter.
#'
#' @return A numeric matrix of quantile survival times with rows corresponding
#' to observations and columns to probabilities.
#'
#' @importFrom stats qnorm
#' @keywords internal
#' @noRd
fastml_aft_quantile_times <- function(probs, mu_vec, dist, scale) {
  if (length(mu_vec) == 0 || length(probs) == 0) {
    return(matrix(numeric(), nrow = length(mu_vec), ncol = length(probs)))
  }
  scale <- ifelse(is.finite(scale) && scale > 0, scale, 1)
  probs <- as.numeric(probs)
  probs <- probs[is.finite(probs) & probs > 0 & probs < 1]
  if (length(probs) == 0) {
    return(matrix(numeric(), nrow = length(mu_vec), ncol = 0))
  }
  n <- length(mu_vec)
  k <- length(probs)
  z_vals <- switch(tolower(dist),
                   "logistic" = log(probs / (1 - probs)),
                   "extreme" = -log(-log(probs)),
                   "normal" = stats::qnorm(probs),
                   stats::qnorm(probs))
  z_vals[!is.finite(z_vals)] <- NA_real_
  log_time <- outer(mu_vec, rep(1, k)) + scale * matrix(z_vals, nrow = n, ncol = k, byrow = TRUE)
  times <- exp(log_time)
  times[!is.finite(times) | times <= 0] <- NA_real_
  colnames(times) <- format(probs, trim = TRUE, scientific = FALSE)
  times
}

#' Compute Survival from Quantile Times
#'
#' Converts quantile-based survival time predictions into survival probability
#' estimates for specified evaluation times.
#'
#' @param times Numeric vector of evaluation times.
#' @param quantile_mat Numeric matrix of quantile survival times (rows = subjects,
#' columns = quantiles).
#' @param probs Numeric vector of quantile probabilities corresponding to
#' columns of \code{quantile_mat}.
#'
#' @return A numeric matrix of survival probabilities with rows corresponding
#' to subjects and columns to evaluation times.
#'
#' @importFrom utils tail
#' @keywords internal
#' @noRd
fastml_aft_survival_from_quantiles <- function(times, quantile_mat, probs) {
  if (length(times) == 0) {
    return(matrix(numeric(), nrow = nrow(quantile_mat), ncol = 0))
  }
  n <- if (is.null(quantile_mat)) 0 else nrow(quantile_mat)
  if (n == 0) {
    return(matrix(numeric(), nrow = 0, ncol = length(times)))
  }
  if (is.null(probs) || length(probs) == 0) {
    return(matrix(NA_real_, nrow = n, ncol = length(times)))
  }
  probs <- as.numeric(probs)
  ord <- order(probs)
  probs_ord <- probs[ord]
  quant_ord <- quantile_mat[, ord, drop = FALSE]
  res <- matrix(NA_real_, nrow = n, ncol = length(times))
  for (i in seq_len(n)) {
    q_row <- as.numeric(quant_ord[i, ])
    valid <- is.finite(q_row)
    if (!any(valid)) {
      next
    }
    q_vals <- q_row[valid]
    p_vals <- probs_ord[valid]
    if (length(q_vals) < 1) {
      next
    }
    if (length(q_vals) > 1) {
      q_vals <- cummax(q_vals)
    }
    for (j in seq_along(times)) {
      tval <- times[j]
      if (!is.finite(tval) || tval <= 0) {
        next
      }
      idx <- findInterval(tval, q_vals)
      if (idx <= 0) {
        surv <- 1
      } else if (idx >= length(q_vals)) {
        surv <- 1 - tail(p_vals, 1)
      } else {
        q_low <- q_vals[idx]
        q_high <- q_vals[idx + 1]
        p_low <- p_vals[idx]
        p_high <- p_vals[idx + 1]
        if (!is.finite(q_low) || !is.finite(q_high) || q_high <= q_low) {
          surv <- 1 - p_high
        } else {
          weight <- (tval - q_low) / (q_high - q_low)
          weight <- min(max(weight, 0), 1)
          p_interp <- p_low + weight * (p_high - p_low)
          surv <- 1 - p_interp
        }
      }
      res[i, j] <- surv
    }
  }
  res[res < 0] <- 0
  res[res > 1] <- 1
  if (ncol(res) > 1) {
    res <- t(apply(res, 1, function(row) {
      row[!is.finite(row)] <- NA_real_
      cummin(pmin(pmax(row, 0), 1))
    }))
  }
  res
}

#' Predict AFT Survival Quantities from an XGBoost Model
#'
#' Computes location parameters, quantiles, and survival probabilities
#' for an accelerated failure time (AFT) model trained with XGBoost.
#'
#' @param fit_obj A fitted XGBoost AFT model object.
#' @param predictors A data frame or matrix of predictor variables.
#' @param eval_times Optional numeric vector of evaluation times for survival estimation.
#' @param quantile_probs Optional numeric vector of quantile probabilities
#' to compute quantile survival times.
#'
#' @return A list containing:
#' \item{mu}{Vector of location parameters.}
#' \item{quantiles}{Matrix of quantile survival times.}
#' \item{probs}{Vector of quantile probabilities.}
#' \item{surv}{Matrix of survival probabilities (if \code{eval_times} provided).}
#'
#' @keywords internal
#' @noRd
fastml_xgb_aft_predict <- function(fit_obj, predictors, eval_times = NULL, quantile_probs = NULL) {
  if (is.null(fit_obj) || is.null(fit_obj$booster)) {
    return(list(mu = numeric(0), quantiles = NULL, probs = numeric(0), surv = NULL))
  }
  mu_vec <- fastml_xgb_predict_lp(fit_obj, predictors = predictors)
  dist <- tryCatch(fit_obj$aft_distribution, error = function(e) "logistic")
  scale <- tryCatch(as.numeric(fit_obj$aft_scale), error = function(e) 1)
  if (!is.finite(scale) || scale <= 0) {
    scale <- 1
  }
  if (is.null(quantile_probs)) {
    quantile_probs <- tryCatch(fit_obj$aft_quantiles, error = function(e) NULL)
  }
  quantile_probs <- quantile_probs[is.finite(quantile_probs) & quantile_probs > 0 & quantile_probs < 1]
  if (length(quantile_probs) == 0) {
    quantile_probs <- c(0.25, 0.5, 0.75)
  }
  quantile_probs <- sort(unique(as.numeric(quantile_probs)))
  quantile_mat <- fastml_aft_quantile_times(quantile_probs, mu_vec, dist, scale)
  surv_mat <- NULL
  if (!is.null(eval_times)) {
    surv_mat <- fastml_aft_survival_from_quantiles(eval_times, quantile_mat, quantile_probs)
    invalid_surv <- is.null(surv_mat) || ncol(surv_mat) != length(eval_times) ||
      !any(is.finite(surv_mat))
    if (invalid_surv) {
      surv_mat <- fastml_aft_survival_prob(eval_times, mu_vec, dist, scale)
    }
  }
  list(mu = mu_vec, quantiles = quantile_mat, probs = quantile_probs, surv = surv_mat)
}

#' Predict Risk Scores from a Survival Model
#'
#' Provides a consistent interface for computing linear predictors (risk scores)
#' across various survival modeling engines, including native \pkg{fastml} models
#' (e.g., Cox proportional hazards, XGBoost Cox) and \pkg{parsnip}/workflow objects.
#'
#' @param fit A fitted survival model object.
#' @param newdata A data frame containing predictor variables for which to
#' compute risk scores.
#' @param ... Additional arguments passed to specific methods.
#'
#' @return A numeric vector of risk scores, where higher values indicate
#' greater predicted risk.
#'
#' @importFrom stats predict
#' @keywords internal
#' @export
predict_risk <- function(fit, newdata, ...) {
  UseMethod("predict_risk")
}

#' @rdname predict_risk
#' @method predict_risk fastml_native_survival
#' @export
#' @importFrom recipes bake
#' @importFrom stats predict model.matrix median
#' @keywords internal
predict_risk.fastml_native_survival <- function(fit, newdata, ...) {
  if (missing(newdata)) {
    stop("Please supply 'newdata' when calling predict_risk().")
  }
  baked <- fastml_bake_or_abort(fit$recipe, newdata, "predict_risk()")
  predictors <- fastml_prepare_native_survival_predictors(fit, baked, newdata)
  if (inherits(fit$fit, "coxph")) {
    res <- tryCatch(stats::predict(fit$fit, newdata = predictors, type = "lp"), error = function(e) NULL)
    return(as.numeric(res))
  }
  if (inherits(fit$fit, "survreg")) {
    res <- tryCatch(stats::predict(fit$fit, newdata = predictors, type = "lp"), error = function(e) NULL)
    return(as.numeric(res))
  }
  if (inherits(fit$fit, "flexsurvreg")) {
    default_time <- NA_real_
    if (!is.null(fit$train_times)) {
      default_time <- stats::median(as.numeric(fit$train_times), na.rm = TRUE)
    }
    if (!is.finite(default_time) || default_time <= 0) {
      default_time <- 1
    }
    pred_data <- predictors
    if ((is.null(pred_data) || nrow(pred_data) != nrow(newdata)) && nrow(newdata) > 0) {
      pred_data <- data.frame(matrix(nrow = nrow(newdata), ncol = 0))
    }
    surv_mat <- fastml_flexsurv_survival_matrix(fit$fit, pred_data, default_time)
    if (is.null(surv_mat) || nrow(surv_mat) != nrow(newdata) || ncol(surv_mat) == 0) {
      return(rep(NA_real_, nrow(newdata)))
    }
    surv_vals <- as.numeric(surv_mat[, ncol(surv_mat), drop = TRUE])
    return(1 - pmin(pmax(surv_vals, 0), 1))
  }
  if (inherits(fit$fit, "glmnet")) {
    if (!requireNamespace("glmnet", quietly = TRUE)) {
      return(rep(NA_real_, nrow(predictors)))
    }
    feature_names <- fit$feature_names
    x_terms <- fit$x_terms
    x_contrasts <- fit$x_contrasts
    prepare_glmnet_newx <- function(new_data) {
      if (is.null(new_data) || nrow(new_data) == 0) {
        return(matrix(0, nrow = ifelse(is.null(new_data), 0, nrow(new_data)), ncol = length(feature_names)))
      }
      mm <- tryCatch({
        if (!is.null(x_terms)) {
          stats::model.matrix(x_terms, new_data, contrasts.arg = x_contrasts)
        } else {
          stats::model.matrix(~ . - 1, data = new_data)
        }
      }, error = function(e) NULL)
      if (is.null(mm)) {
        mm <- tryCatch(stats::model.matrix(~ . - 1, data = new_data), error = function(e) NULL)
      }
      if (is.null(mm)) {
        mm <- matrix(0, nrow = nrow(new_data), ncol = 0)
      }
      mm <- as.matrix(mm)
      if (!is.null(colnames(mm)) && any(colnames(mm) == "(Intercept)")) {
        keep_cols <- colnames(mm) != "(Intercept)"
        mm <- mm[, keep_cols, drop = FALSE]
      }
      mm_full <- matrix(0, nrow = nrow(mm), ncol = length(feature_names))
      colnames(mm_full) <- feature_names
      overlap <- intersect(feature_names, colnames(mm))
      if (length(overlap) > 0) {
        mm_full[, overlap] <- mm[, overlap, drop = FALSE]
      }
      mm_full
    }
    mm <- prepare_glmnet_newx(predictors)
    penalty <- fit$penalty
    res <- tryCatch(predict(fit$fit, newx = mm, s = penalty, type = "link"), error = function(e) NULL)
    if (!is.null(res) && length(dim(res)) >= 2) {
      res <- as.matrix(res)[, 1, drop = TRUE]
    }
    if (!is.null(res)) {
      return(as.numeric(res))
    }
    coef_mat <- tryCatch(coef(fit$fit, s = penalty), error = function(e) NULL)
    if (is.null(coef_mat)) {
      return(rep(NA_real_, nrow(predictors)))
    }
    coef_dense <- as.matrix(coef_mat)
    intercept <- 0
    if ("(Intercept)" %in% rownames(coef_dense)) {
      intercept <- coef_dense["(Intercept)", 1]
      coef_dense <- coef_dense[setdiff(rownames(coef_dense), "(Intercept)"), , drop = FALSE]
    }
    coef_vec <- numeric(length(feature_names))
    names(coef_vec) <- feature_names
    overlap <- intersect(feature_names, rownames(coef_dense))
    if (length(overlap) > 0) {
      coef_vec[overlap] <- coef_dense[overlap, 1]
    }
    drop(as.numeric(mm %*% coef_vec + intercept))
  } else if (inherits(fit$fit, c("stpm2", "pstpm2"))) {
    if (!requireNamespace("rstpm2", quietly = TRUE)) {
      return(rep(NA_real_, nrow(predictors)))
    }
    base_newdata <- predictors
    time_var <- fit$time_col
    if (!is.null(time_var)) {
      if (!(time_var %in% names(base_newdata)) && time_var %in% names(newdata)) {
        base_newdata[[time_var]] <- newdata[[time_var]]
      }
      if (!(time_var %in% names(base_newdata))) {
        base_newdata[[time_var]] <- rep(1, nrow(base_newdata))
      }
    }
    return(as.numeric(rstpm2::predict(fit$fit, newdata = base_newdata, type = "lp")))
  } else if (inherits(fit$fit, "fastml_xgb_survival")) {
    lp <- fastml_xgb_predict_lp(fit$fit, predictors)
    if (identical(fit$fit$objective, "survival:aft")) {
      return(as.numeric(-lp))
    }
    return(as.numeric(lp))
  }
  rep(NA_real_, nrow(predictors))
}

#' @rdname predict_risk
#' @method predict_risk workflow
#' @export
predict_risk.workflow <- function(fit, newdata, ...) {
  if (missing(newdata)) {
    stop("Please supply 'newdata' when calling predict_risk().")
  }
  pred_lp <- tryCatch(predict(fit, new_data = newdata, type = "linear_pred"), error = function(e) NULL)
  if (is.null(pred_lp)) {
    pred_lp <- tryCatch(predict(fit, new_data = newdata, type = "lp"), error = function(e) NULL)
  }
  if (is.null(pred_lp)) {
    pred_lp <- tryCatch(predict(fit, new_data = newdata, type = "risk"), error = function(e) NULL)
  }
  if (is.null(pred_lp)) {
    stop("Unable to compute risk predictions for this workflow. Ensure the engine supports 'linear_pred' or 'lp'.")
  }
  if (is.data.frame(pred_lp)) {
    if (".pred" %in% names(pred_lp)) {
      return(as.numeric(pred_lp$.pred))
    }
    if (".pred_linear_pred" %in% names(pred_lp)) {
      return(as.numeric(pred_lp$.pred_linear_pred))
    }
  }
  as.numeric(pred_lp)
}

#' @rdname predict_risk
#' @method predict_risk default
#' @export
predict_risk.default <- function(fit, newdata, ...) {
  stop("predict_risk() is not implemented for objects of class ", paste(class(fit), collapse = ", "), ".")
}

#' Predict survival probabilities from a survival model
#'
#' @param fit A fitted survival model.
#' @param newdata A data frame of predictors for which to compute survival curves.
#' @param times Numeric vector of evaluation times.
#' @param ... Additional arguments passed to methods.
#'
#' @return A numeric matrix with one row per observation and one column per time.
#'
#' @export
predict_survival <- function(fit, newdata, times, ...) {
  UseMethod("predict_survival")
}

fastml_align_survival_output <- function(pred_obj, times, n) {
  if (is.null(pred_obj)) {
    return(matrix(NA_real_, nrow = n, ncol = length(times)))
  }
  if (is.matrix(pred_obj)) {
    mat <- pred_obj
    if (nrow(mat) != n) {
      stop("Returned matrix has an unexpected number of rows.")
    }
    if (ncol(mat) != length(times)) {
      if (length(times) == 1 && ncol(mat) == 1) {
        colnames(mat) <- NULL
      } else if (!is.null(colnames(mat))) {
        idx <- match(times, suppressWarnings(as.numeric(colnames(mat))))
        if (anyNA(idx)) {
          stop("Could not align survival predictions with requested times.")
        }
        mat <- mat[, idx, drop = FALSE]
      }
    }
    return(as.matrix(mat))
  }
  if (is.data.frame(pred_obj) && ".pred_survival" %in% names(pred_obj)) {
    lst <- pred_obj$.pred_survival
    res <- matrix(NA_real_, nrow = n, ncol = length(times))
    for (i in seq_along(lst)) {
      entry <- lst[[i]]
      if (is.null(entry)) next
      time_col <- intersect(c(".eval_time", "eval_time", "time"), names(entry))
      surv_col <- intersect(c(".pred", ".survival", "survival", "surv"), names(entry))
      if (length(time_col) == 0 || length(surv_col) == 0) next
      res[i, ] <- stats::approx(entry[[time_col[1]]], entry[[surv_col[1]]], xout = times, method = "constant", rule = 2, f = 0, ties = "ordered")$y
    }
    return(res)
  }
  stop("Unsupported survival prediction output structure.")
}

#' @rdname predict_survival
#' @method predict_survival fastml_native_survival
#' @export
predict_survival.fastml_native_survival <- function(fit, newdata, times, ...) {
  if (missing(newdata)) {
    stop("Please supply 'newdata' when calling predict_survival().")
  }
  if (missing(times)) {
    stop("Please supply 'times' when calling predict_survival().")
  }
  times <- as.numeric(times)
  times <- times[is.finite(times) & times >= 0]
  baked <- fastml_bake_or_abort(fit$recipe, newdata, "predict_survival()")
  predictors <- fastml_prepare_native_survival_predictors(fit, baked, newdata)
  n <- nrow(newdata)
  if (inherits(fit$fit, "coxph")) {
    base <- tryCatch(survival::basehaz(fit$fit, centered = FALSE), error = function(e) NULL)
    if (is.null(base) && !is.null(fit$baseline)) {
      base <- fit$baseline
      colnames(base) <- c("time", "hazard", "surv")
    }
    lp <- tryCatch(stats::predict(fit$fit, newdata = predictors, type = "lp"), error = function(e) rep(NA_real_, n))
    if (!is.null(base) && nrow(base) > 0) {
      bh <- data.frame(time = base$time, cumhaz = ifelse("hazard" %in% names(base), base$hazard, base$cumhaz))
      helper <- list(baseline = data.frame(time = bh$time, cumhaz = bh$cumhaz), feature_names = colnames(predictors))
      class(helper) <- "fastml_xgb_survival"
      helper$baseline$surv <- exp(-helper$baseline$cumhaz)
      helper$booster <- NULL
      return(fastml_xgb_survival_matrix_cox(helper, lp, times))
    }
    return(matrix(NA_real_, nrow = n, ncol = length(times)))
  }
  if (inherits(fit$fit, "survreg")) {
    lp <- tryCatch(stats::predict(fit$fit, newdata = predictors, type = "lp"), error = function(e) rep(NA_real_, n))
    scale <- tryCatch(fit$fit$scale, error = function(e) 1)
    dist <- tryCatch(fit$fit$dist, error = function(e) "weibull")
    if (identical(dist, "weibull")) {
      lambda <- exp(-lp / scale)
      k <- 1 / scale
      mat <- outer(lambda, times^k, FUN = function(l, t) exp(-l * t))
      mat[mat < 0] <- 0
      mat[mat > 1] <- 1
      if (ncol(mat) > 1) {
        mat <- t(apply(mat, 1, cummin))
      }
      return(mat)
    }
    return(matrix(NA_real_, nrow = n, ncol = length(times)))
  }
  if (inherits(fit$fit, "flexsurvreg")) {
    pred_data <- predictors
    if ((is.null(pred_data) || nrow(pred_data) != nrow(newdata)) && nrow(newdata) > 0) {
      pred_data <- data.frame(matrix(nrow = nrow(newdata), ncol = 0))
    }
    surv_mat <- fastml_flexsurv_survival_matrix(fit$fit, pred_data, times)
    if (is.null(surv_mat) || nrow(surv_mat) != nrow(newdata)) {
      return(matrix(NA_real_, nrow = nrow(newdata), ncol = length(times)))
    }
    return(surv_mat)
  }
  if (inherits(fit$fit, "fastml_xgb_survival")) {
    if (identical(fit$fit$objective, "survival:cox")) {
      stop("XGBoost survival:cox provides risk ranking only; survival curve predictions are unavailable.")
    }
    aft_pred <- fastml_xgb_aft_predict(fit$fit, predictors, eval_times = times)
    if (is.null(aft_pred$surv)) {
      n_obs <- if (is.null(predictors)) 0L else nrow(predictors)
      return(matrix(NA_real_, nrow = n_obs, ncol = length(times)))
    }
    return(aft_pred$surv)
  }
  stop("Survival prediction not implemented for this native engine.")
}

#' @rdname predict_survival
#' @method predict_survival workflow
#' @export
predict_survival.workflow <- function(fit, newdata, times, ...) {
  if (missing(newdata)) {
    stop("Please supply 'newdata' when calling predict_survival().")
  }
  if (missing(times)) {
    stop("Please supply 'times' when calling predict_survival().")
  }
  times <- as.numeric(times)
  times <- times[is.finite(times) & times >= 0]
  if (length(times) == 0) {
    stop("No valid evaluation times supplied.")
  }

  # Engines differ on the argument name for evaluation times (eval_time, time, times).
  # Try a small cascade of signatures, capturing the last error for debugging.
  extra_args <- list(...)
  last_err <- NULL
  attempt_predict <- function(arg_list) {
    do.call(predict, c(list(fit), arg_list, extra_args))
  }
  try_sig <- function(arg_list) {
    tryCatch(attempt_predict(arg_list), error = function(e) {
      last_err <<- e$message
      NULL
    })
  }

  pred <- NULL
  arg_variants <- list(
    list(new_data = newdata, type = "survival", eval_time = times),
    list(new_data = newdata, type = "survival", time = times),
    list(new_data = newdata, type = "survival", times = times),
    list(new_data = newdata, eval_time = times),
    list(new_data = newdata, time = times),
    list(new_data = newdata, times = times),
    list(newdata = newdata, type = "survival", eval_time = times), # some engines still use `newdata`
    list(newdata = newdata, type = "survival", time = times)
  )

  for (args in arg_variants) {
    pred <- try_sig(args)
    if (!is.null(pred)) break
  }

  # Single-time fallback for engines that only accept scalar time input
  if (is.null(pred) && length(times) == 1) {
    scalar_variants <- list(
      list(new_data = newdata, type = "survival", eval_time = times[1]),
      list(new_data = newdata, type = "survival", time = times[1]),
      list(new_data = newdata, eval_time = times[1]),
      list(newdata = newdata, type = "survival", eval_time = times[1])
    )
    for (args in scalar_variants) {
      pred <- try_sig(args)
      if (!is.null(pred)) break
    }
  }

  if (is.null(pred)) {
    msg <- "Underlying engine does not provide survival curve predictions."
    if (is.character(last_err) && nzchar(last_err)) {
      msg <- paste0(msg, " Last error: ", last_err)
    }
    stop(msg)
  }

  fastml_align_survival_output(pred, times, nrow(newdata))
}

#' @rdname predict_survival
#' @method predict_survival default
#' @export
predict_survival.default <- function(fit, newdata, times, ...) {
  stop("predict_survival() is not implemented for objects of class ", paste(class(fit), collapse = ", "), ".")
}

fastml_model_spec_registry <- function() {
  list(
    rand_forest = function(ctx) {
      define_rand_forest_spec(
        ctx$task,
        ctx$train_data,
        ctx$label,
        tuning = ctx$tuning,
        engine = ctx$engine
      )
    },
    logistic_reg = function(ctx) {
      define_logistic_reg_spec(
        ctx$task,
        tuning = ctx$tuning,
        engine = ctx$engine
      )
    },
    multinom_reg = function(ctx) {
      define_multinomial_reg_spec(
        ctx$task,
        tuning = ctx$tuning,
        engine = ctx$engine
      )
    },
    C5_rules = function(ctx) {
      define_C5_rules_spec(
        ctx$task,
        tuning = ctx$tuning,
        engine = ctx$engine
      )
    },
    xgboost = function(ctx) {
      define_xgboost_spec(
        ctx$task,
        ctx$train_data,
        ctx$label,
        tuning = ctx$tuning,
        engine = ctx$engine,
        early_stopping = ctx$early_stopping
      )
    },
    lightgbm = function(ctx) {
      define_lightgbm_spec(
        ctx$task,
        ctx$train_data,
        ctx$label,
        tuning = ctx$tuning,
        engine = ctx$engine,
        early_stopping = ctx$early_stopping
      )
    },
    decision_tree = function(ctx) {
      define_decision_tree_spec(
        ctx$task,
        tuning = ctx$tuning,
        engine = ctx$engine
      )
    },
    svm_linear = function(ctx) {
      define_svm_linear_spec(
        ctx$task,
        tuning = ctx$tuning,
        engine = ctx$engine
      )
    },
    svm_rbf = function(ctx) {
      define_svm_rbf_spec(
        ctx$task,
        tuning = ctx$tuning,
        engine = ctx$engine
      )
    },
    nearest_neighbor = function(ctx) {
      define_nearest_neighbor_spec(
        ctx$task,
        tuning = ctx$tuning,
        engine = ctx$engine
      )
    },
    naive_Bayes = function(ctx) {
      define_naive_Bayes_spec(
        ctx$task,
        tuning = ctx$tuning,
        engine = ctx$engine
      )
    },
    mlp = function(ctx) {
      define_mlp_spec(
        ctx$task,
        tuning = ctx$tuning,
        engine = ctx$engine
      )
    },
    discrim_linear = function(ctx) {
      define_discrim_linear_spec(
        ctx$task,
        engine = ctx$engine
      )
    },
    discrim_quad = function(ctx) {
      define_discrim_quad_spec(
        ctx$task,
        engine = ctx$engine
      )
    },
    bag_tree = function(ctx) {
      define_bag_tree_spec(
        ctx$task,
        tuning = ctx$tuning,
        engine = ctx$engine
      )
    },
    elastic_net = function(ctx) {
      define_elastic_net_spec(
        ctx$task,
        tuning = ctx$tuning,
        engine = ctx$engine
      )
    },
    bayes_glm = function(ctx) {
      define_bayes_glm_spec(
        ctx$task,
        engine = ctx$engine
      )
    },
    pls = function(ctx) {
      define_pls_spec(
        ctx$task,
        tuning = ctx$tuning,
        engine = ctx$engine
      )
    },
    linear_reg = function(ctx) {
      define_linear_reg_spec(
        ctx$task,
        engine = ctx$engine
      )
    },
    ridge_reg = function(ctx) {
      define_ridge_reg_spec(
        ctx$task,
        tuning = ctx$tuning,
        engine = ctx$engine
      )
    },
    lasso_reg = function(ctx) {
      define_lasso_reg_spec(
        ctx$task,
        tuning = ctx$tuning,
        engine = ctx$engine
      )
    }
  )
}

fastml_get_model_spec <- function(algo, context) {
  registry <- fastml_model_spec_registry()
  handler <- registry[[algo]]
  if (is.null(handler)) {
    warning(sprintf("Algorithm %s is not supported or failed to train.", algo))
    return(NULL)
  }
  handler(context)
}

#' Evaluate Models Function
#'
#' Evaluates the trained models on the test data and computes performance metrics.
#'
#' @param models A list of trained model objects.
#' @param train_data Preprocessed training data frame.
#' @param test_data Preprocessed test data frame.
#' @param label Name of the target variable. For survival analysis this should
#'   be a character vector of length two giving the names of the time and status
#'   columns.
#' @param start_col Optional string. The name of the column specifying the
#'   start time in counting process (e.g., `(start, stop, event)`) survival
#'   data. Only used when \code{task = "survival"}.
#' @param time_col String. The name of the column specifying the event or
#'   censoring time (the "stop" time in counting process data). Only used
#'   when \code{task = "survival"}.
#' @param status_col String. The name of the column specifying the event
#'   status (e.g., 0 for censored, 1 for event). Only used when
#'   \code{task = "survival"}.
#' @param task Type of task: "classification", "regression", or "survival".
#' @param metric The performance metric to optimize (e.g., "accuracy", "rmse").
#' @param event_class A single string. Either "first" or "second" to specify which level of truth to consider as the "event".
#' @param class_threshold For binary classification, controls how class probabilities
#'   are converted into hard class predictions. The default is `0.5` (standard threshold).
#'   Numeric values in (0, 1) set a fixed threshold. Use `"auto"` to tune a threshold
#'   on training data to maximize F1; use `"model"` to keep the model's default predictions.
#' @param multiclass_auc For multiclass ROC AUC, the averaging method to use:
#'   `"macro"` (default, tidymodels) or `"macro_weighted"`. Macro weights each
#'   class equally, while macro_weighted weights by class prevalence and can
#'   change model rankings on imbalanced data.
#' @importFrom dplyr filter bind_rows pull mutate select bind_cols
#' @importFrom yardstick metric_set accuracy kap roc_auc sens spec precision f_meas rmse rsq mae
#' @importFrom workflows pull_workflow_spec pull_workflow_preprocessor workflow add_model add_recipe
#' @importFrom parsnip fit predict.model_fit
#' @importFrom tune select_best finalize_model
#' @importFrom rlang sym syms
#' @importFrom tibble tibble
#' @param eval_times Optional numeric vector of evaluation horizons for survival
#'   metrics. Passed through to \code{process_model}.
#' @param bootstrap_ci Logical indicating whether bootstrap confidence intervals
#'   should be computed for the evaluation metrics.
#' @param bootstrap_samples Number of bootstrap resamples used when
#'   \code{bootstrap_ci = TRUE}.
#' @param bootstrap_seed Optional integer seed for the bootstrap procedure used
#'   in metric estimation.
#' @param at_risk_threshold Minimum proportion of subjects that must remain at
#'   risk to define \eqn{t_{max}} when computing survival metrics such as the
#'   integrated Brier score.
#' @param survival_metric_convention Character string specifying which survival
#'   metric conventions to follow. `"fastml"` (default) uses fastml's internal
#'   defaults for evaluation horizons and t_max. `"tidymodels"` uses
#'   `eval_times` as the explicit evaluation grid and applies yardstick-style
#'   Brier/IBS normalization; when `eval_times` is `NULL`, time-dependent Brier
#'   metrics are omitted.
#' @param precomputed_predictions Optional data frame or nested list of
#'   previously generated predictions (per algorithm/engine) to reuse instead
#'   of recomputing. This is mainly used when combining results across engines.
#' @param summaryFunction Optional custom classification metric function passed
#'   through to \code{process_model} for holdout evaluation.
#' @return A list with two elements:
#'   \describe{
#'     \item{performance}{A named list of performance metric tibbles for each model.}
#'     \item{predictions}{A named list of data frames with columns including truth, predictions, and probabilities per model.}
#'   }
#' @export
fastml_compute_holdout_results <- function(models,
                                           train_data,
                                           test_data,
                                           label,
                                           start_col = NULL,
                                           time_col = NULL,
                                           status_col = NULL,
                                           task,
                                           metric = NULL,
                                           event_class,
                                           class_threshold = 0.5,
                                           eval_times = NULL,
                                           bootstrap_ci = TRUE,
                                           bootstrap_samples = 500,
                                           bootstrap_seed = 1234,
                                           at_risk_threshold = 0.1,
                                           survival_metric_convention = "fastml",
                                           precomputed_predictions = NULL,
                                           summaryFunction = NULL,
                                           multiclass_auc = "macro") {
  survival_metric_convention <- fastml_normalize_survival_convention(survival_metric_convention)
  # Load required packages
  required_pkgs <- c("yardstick", "parsnip", "tune", "workflows",
                     "dplyr", "rlang", "tibble")
  if (task == "survival") {
    required_pkgs <- c(required_pkgs, "survival")
  }
  for(pkg in required_pkgs){
    if (!requireNamespace(pkg, quietly = TRUE)) {
      stop(sprintf("The '%s' package is required but not installed.", pkg))
    }
  }

  # Determine engine names for each algorithm
  engine_names <- get_engine_names(models)
  # Initialize performance and predictions lists
  performance <- list()
  predictions_list <- list()

  reuse_prediction <- function(algo_name, engine_name) {
    if (is.null(precomputed_predictions)) {
      return(NULL)
    }
    entry <- precomputed_predictions[[algo_name]]
    if (is.null(entry)) {
      return(NULL)
    }
    if (is.list(entry) && !is.data.frame(entry)) {
      return(entry[[engine_name]])
    }
    if (is.data.frame(entry) && (is.null(engine_name) || is.na(engine_name))) {
      return(entry)
    }
    NULL
  }

  # Iterate over the models object. Check if the model is nested (i.e. a list of engines)
  for (algo in names(models)) {
    performance[[algo]] <- list()
    predictions_list[[algo]] <- list()

    # If the model object has names, assume it is nested by engine.
    # But skip this for native survival models stored as lists.
    if (is.list(models[[algo]]) && !inherits(models[[algo]], "workflow") && !inherits(models[[algo]], "tune_results") && !inherits(models[[algo]], "fastml_native_survival")) {
      for (eng in names(models[[algo]])) {
        model_obj <- models[[algo]][[eng]]
        reused_preds <- reuse_prediction(algo, eng)
        result <- process_model(model_obj,
                                model_id = paste(algo, eng, sep = "_"),
                                task = task,
                                test_data = test_data,
                                label = label,
                                event_class = event_class,
                                class_threshold = class_threshold,
                                start_col = start_col,
                                time_col = time_col,
                                status_col = status_col,
                                engine = eng,
                                train_data = train_data,
                                metric = metric,
                                summaryFunction = summaryFunction,
                                eval_times_user = eval_times,
                                bootstrap_ci = bootstrap_ci,
                                bootstrap_samples = bootstrap_samples,
                                bootstrap_seed = bootstrap_seed,
                                at_risk_threshold = at_risk_threshold,
                                survival_metric_convention = survival_metric_convention,
                                precomputed_predictions = reused_preds,
                                multiclass_auc = multiclass_auc)
        if (!is.null(result)) {
          performance[[algo]][[eng]] <- result$performance
          predictions_list[[algo]][[eng]] <- result$predictions
        }
      }
    } else {
      # Otherwise, assume a single model (not nested)
      eng <- if (inherits(models[[algo]], "fastml_native_survival")) {
        models[[algo]]$engine
      } else if (!is.null(engine_names[[algo]])) {
        engine_names[[algo]][1]
      } else if (inherits(models[[algo]], "workflow")) {
        workflows::extract_fit_parsnip(models[[algo]])$spec$engine
      } else {
        NA
      }
      reused_preds <- reuse_prediction(algo, eng)
      result <- process_model(model_obj = models[[algo]],
                              model_id = algo,
                              task = task,
                              test_data = test_data,
                              label = label,
                              start_col = start_col,
                              time_col = time_col,
                              status_col = status_col,
                              event_class = event_class,
                              class_threshold = class_threshold,
                              engine = eng,
                              train_data = train_data,
                              metric = metric,
                              summaryFunction = summaryFunction,
                              eval_times_user = eval_times,
                              bootstrap_ci = bootstrap_ci,
                              bootstrap_samples = bootstrap_samples,
                              bootstrap_seed = bootstrap_seed,
                              at_risk_threshold = at_risk_threshold,
                              survival_metric_convention = survival_metric_convention,
                              precomputed_predictions = reused_preds,
                              multiclass_auc = multiclass_auc)



      if (!is.null(result)) {
        performance[[algo]] <- result$performance
        predictions_list[[algo]] <- result$predictions
      }
    }
  }

  unwrap_single_engine <- function(entry) {
    if (is.list(entry) && !inherits(entry, "data.frame") && length(entry) == 1) {
      single <- entry[[1]]
      if (inherits(single, "data.frame")) {
        return(single)
      }
    }
    entry
  }

  performance <- lapply(performance, unwrap_single_engine)
  predictions_list <- lapply(predictions_list, unwrap_single_engine)

  list(performance = performance, predictions = predictions_list)
}

evaluate_models <- function(models,
                            train_data,
                            test_data,
                            label,
                            start_col = NULL,
                            time_col = NULL,
                            status_col = NULL,
                            task,
                            metric = NULL,
                            event_class,
                            class_threshold = 0.5,
                            eval_times = NULL,
                            bootstrap_ci = TRUE,
                            bootstrap_samples = 500,
                            bootstrap_seed = 1234,
                            at_risk_threshold = 0.1,
                            survival_metric_convention = "fastml",
                            precomputed_predictions = NULL,
                            summaryFunction = NULL,
                            multiclass_auc = "macro") {
  fastml_compute_holdout_results(
    models = models,
    train_data = train_data,
    test_data = test_data,
    label = label,
    start_col = start_col,
    time_col = time_col,
    status_col = status_col,
    task = task,
    metric = metric,
    event_class = event_class,
    class_threshold = class_threshold,
    eval_times = eval_times,
    bootstrap_ci = bootstrap_ci,
    bootstrap_samples = bootstrap_samples,
    bootstrap_seed = bootstrap_seed,
    at_risk_threshold = at_risk_threshold,
    survival_metric_convention = survival_metric_convention,
    precomputed_predictions = precomputed_predictions,
    summaryFunction = summaryFunction,
    multiclass_auc = multiclass_auc
  )
}

#' Get Default Parameters for an Algorithm
#'
#' Returns a list of default tuning parameters for the specified algorithm based on the task type, number of predictors, and engine.
#'
#' @param algo A character string specifying the algorithm name. Supported values include:
#'   \code{"rand_forest"}, \code{"C5_rules"}, \code{"xgboost"}, \code{"lightgbm"},
#'   \code{"logistic_reg"}, \code{"multinom_reg"}, \code{"decision_tree"}, \code{"svm_linear"},
#'   \code{"svm_rbf"}, \code{"nearest_neighbor"}, \code{"naive_Bayes"}, \code{"mlp"},
#'   \code{"deep_learning"}, \code{"discrim_linear"}, \code{"discrim_quad"}, \code{"bag_tree"},
#'   \code{"elastic_net"}, \code{"bayes_glm"}, \code{"pls"}, \code{"linear_reg"},
#'   \code{"ridge_reg"}, \code{"lasso_reg"}, and \code{"penalized_cox"}.
#'
#' @param task A character string specifying the task type, typically \code{"classification"} or \code{"regression"}.
#'
#' @param num_predictors An optional numeric value indicating the number of predictors. This value is used to compute default values for parameters such as \code{mtry}. Defaults to \code{NULL}.
#'
#' @param engine An optional character string specifying the engine to use. If not provided, a default engine is chosen where applicable.
#'
#' @return A list of default parameter settings for the specified algorithm. If the algorithm is not recognized, the function returns \code{NULL}.
#'
#' @details The function employs a \code{switch} statement to select and return a list of default parameters tailored for the given algorithm, task, and engine. The defaults vary by algorithm and, in some cases, by engine. For example:
#'   \itemize{
#'     \item For \code{"rand_forest"}, if \code{engine} is not provided, it defaults to \code{"ranger"}. The parameters such as \code{mtry}, \code{trees}, and \code{min_n} are computed based on the task and the number of predictors.
#'     \item For \code{"C5_rules"}, the defaults include \code{trees}, \code{min_n}, and \code{sample_size}.
#'     \item For \code{"xgboost"} and \code{"lightgbm"}, default values are provided for parameters like tree depth, learning rate, and sample size.
#'     \item For \code{"logistic_reg"} and \code{"multinom_reg"}, the function returns defaults for regularization parameters (\code{penalty} and \code{mixture}) that vary with the specified engine.
#'     \item For \code{"decision_tree"}, the parameters (such as \code{tree_depth}, \code{min_n}, and \code{cost_complexity}) are set based on the engine (e.g., \code{"rpart"}, \code{"C5.0"}, \code{"partykit"}, \code{"spark"}).
#'     \item Other algorithms, including \code{"svm_linear"}, \code{"svm_rbf"}, \code{"nearest_neighbor"}, \code{"naive_Bayes"}, \code{"mlp"}, \code{"deep_learning"}, \code{"elastic_net"}, \code{"bayes_glm"}, \code{"pls"}, \code{"linear_reg"}, \code{"ridge_reg"}, and \code{"lasso_reg"}, have their respective default parameter lists.
#'   }
#'
#'@importFrom utils tail
#'
#' @export
get_default_params <- function(algo, task, num_predictors = NULL, engine = NULL) {
  switch(algo,
         # 1. Random Forest
         "rand_forest" = {
           # Set a default engine if not provided:
           if (is.null(engine)) engine <- "ranger"

           if (engine == "ranger") {
             list(
               mtry  = if (!is.null(num_predictors)) floor(sqrt(num_predictors)) else 2,
               trees = 500L,
               min_n = if (task == "regression") 5 else 10
             )
           } else if (engine == "aorsf") {
             list(
               mtry           = if (!is.null(num_predictors)) ceiling(sqrt(num_predictors)) else 2,
               trees          = 500L,
               min_n          = 5L,
               split_min_stat = 3.841459  # Engine-specific tuning parameter
             )
           } else if (engine == "h2o") {
             list(
               mtry  = if (!is.null(num_predictors)) {
                 if (task == "classification") floor(sqrt(num_predictors)) else floor(num_predictors / 3)
               } else 2,
               trees = 50L,
               min_n = 2
             )
           } else if (engine == "partykit") {
             list(
               mtry  = if (!is.null(num_predictors)) {
                 if (task == "classification") floor(sqrt(num_predictors)) else floor(num_predictors / 3)
               } else 2,
               trees = 500L,
               min_n = 20L
             )
           } else if (engine == "randomForest") {
             list(
               mtry  = if (!is.null(num_predictors)) {
                 if (task == "classification") floor(sqrt(num_predictors)) else floor(num_predictors / 3)
               } else 2,
               trees = 500L,
               min_n = if (task == "regression") 5 else 10
             )
           } else if (engine == "spark") {
             list(
               mtry  = if (!is.null(num_predictors)) {
                 if (task == "classification") floor(sqrt(num_predictors)) else floor(num_predictors / 3)
               } else 2,
               trees = 20L,
               min_n = 1L
             )
           } else {
             # Fallback defaults (similar to ranger)
             list(
               mtry  = if (!is.null(num_predictors)) floor(sqrt(num_predictors)) else 2,
               trees = 500L,
               min_n = if (task == "regression") 5 else 10
             )
           }
         },
         # 4. C5_rules
         "C5_rules" = list(
           trees = 50,
           min_n = 5,
           sample_size = 0.5
         ),
         # 5. XGBoost
        "xgboost" = list(
          tree_depth = 6L,
          trees = 15L,
          learn_rate = 0.1,
           mtry =  if (!is.null(num_predictors)) max(1, floor(sqrt(num_predictors))) else 2,
           min_n =  2,
           loss_reduction = 0.0,
           sample_size =  0.5,
           stop_iter =  Inf
         ),
         # 6. LightGBM
        "lightgbm" = list(
          trees = 100,
          tree_depth = 3,
          learn_rate = 0.1,
           loss_reduction = 0,
           min_n = 5,
           sample_size = 0.5,
           mtry = if (!is.null(num_predictors)) max(1, floor(sqrt(num_predictors))) else 2
         ),
         # 7. Logistic Regression default parameters

         "logistic_reg" = {
           if (engine %in% c("glm")) {
             list(
               penalty = NULL,
               mixture = NULL
             )
           } else if (engine %in% c("gee")) {
             list(
               penalty = NULL,
               mixture = NULL
             )
           } else if (engine %in% c("glmer")) {
             list(
               penalty = NULL,
               mixture = NULL
             )
           } else if (engine %in% c("stan", "stan_glmer")) {
             list(
               penalty = NULL,
               mixture = NULL
             )
           } else if (engine %in% c("brulee")) {
             list(
               penalty = -3,      # corresponds to a raw penalty of 0.001 (log10(0.001) = -3)
               mixture = 0.0
             )
           } else if (engine %in% c("glmnet")) {
             list(
               penalty = -2,      # corresponds to a raw penalty of 0.01 (log10(0.01) = -2)
               mixture = 1.0      # pure lasso
             )
           } else if (engine %in% c("h2o")) {
             list(
               penalty = NULL,
               mixture = NULL
             )
           } else if (engine %in% c("keras")) {
             list(
               penalty = 0.0,     # no regularization
               mixture = NULL
             )
           } else if (engine %in% c("LiblineaR")) {
             list(
               penalty = -2,      # corresponds to a raw penalty of 0.01 (log10(0.01) = -2)
               mixture = 0        # ridge regularization (mixture = 0)
             )
           } else if (engine %in% c("spark")) {
             list(
               penalty = 0.0,     # no regularization
               mixture = 0.0
             )
           } else {
             # Fallback: use engine defaults
             list(penalty = NULL, mixture = NULL)
           }
         },

         "multinom_reg" = {
           if (engine %in% c("nnet")) {
             # nnet::multinom() uses only a penalty (decay) parameter; its default is 0.0.
             list(penalty = 0.0, mixture = NULL)
           } else if (engine %in% c("brulee")) {
             # brulee defaults: penalty = 0.001 and mixture = 0.0.
             list(penalty = 0.0, mixture = 0.0)
           } else if (engine %in% c("glmnet")) {
             # glmnet: by convention, we use a penalty of 0.0 and a mixture default of 1.0 (pure lasso).
             list(penalty = 0.0, mixture = 1.0)
           } else if (engine %in% c("h2o")) {
             # h2o: if no fixed penalty is given, a heuristic is used; here we choose 0.0 with ridge (mixture = 0.0).
             # list(mixture = 0.0)
           } else if (engine %in% c("keras")) {
             # keras_mlp() only uses a penalty parameter (default 0.0); mixture is not applicable.
             list(penalty = 0.0, mixture = NULL)
           } else if (engine %in% c("spark")) {
             # sparklyr: defaults are typically 0.0 for both penalty and mixture.
             list(penalty = 0.0, mixture = 0.0)
           } else {
             # Fallback default if engine not recognized.
             list(penalty = 0.0, mixture = 0.0)
           }
         },

         # 9. Decision Tree
         "decision_tree" = {
           if (engine %in% c("rpart")) {
             # rpart uses three parameters:
             #   tree_depth: 10L, min_n: 5L, cost_complexity: -2 (on log10 scale, i.e., 10^(-2) = 0.01)
             list(tree_depth = 10L, min_n = 5L, cost_complexity = -2)
           } else if (engine %in% c("C5.0")) {
             # C5.0 uses only the minimal node size
             list(min_n = 5L)
           } else if (engine %in% c("partykit")) {
             # partykit uses two parameters:
             #   tree_depth: 10L, min_n: 20L
             list(tree_depth = 10L, min_n = 20L)
           } else if (engine %in% c("spark")) {
             # spark uses two parameters:
             #   tree_depth: 5L, min_n: 2L
             list(tree_depth = 5L, min_n = 2L)
           } else {
             # Fallback defaults (using rpart defaults)
             list(tree_depth = 10L, min_n = 5L, cost_complexity = -2)
           }
         },

         # 10. SVM Linear
         "svm_linear" = {
           if (is.null(engine)) engine <- "kernlab"  # set default engine if not provided

           if (engine %in% c("kernlab")) {
             list(cost = 1, margin = 0.1)
           } else if (engine %in% c("LiblineaR")) {
             list(cost = 1, margin = 0.1)
           }
         },


         # 11. SVM Radial
         "svm_rbf" = list(
           cost = 1,
           rbf_sigma = 0.1
         ),
         # 12. nearest_neighbor
         "nearest_neighbor" = list(
           neighbors = 5,
           weight_func = "rectangular",
           dist_power = 2
         ),

         # 13. Naive Bayes
         "naive_Bayes" = {
           if (is.null(engine)) engine <- "klaR"  # set default engine if not provided

           if (engine %in% c("klaR", "naivebayes")) {
             list(smoothness = 1.0, Laplace = 0.0)
           } else if (engine %in% c("h2o")) {
             list(Laplace = 0.0)
           } else {
             stop("Unsupported engine specified for naive_Bayes.")
           }
         },

         # 14. Neural Network (nnet)
         "mlp" = {
           if (is.null(engine)) engine <- "nnet"  # set default engine if not provided

           if (engine %in% c("nnet")) {
             list(
               hidden_units = 5L,
               penalty      = 0.1,
               epochs       = 100L
             )
           } else if (engine %in% c("brulee")) {
             list(
               hidden_units = 3L,
               penalty      = 0.1,
               epochs       = 100L,
               activation   = "relu",
               # mixture      = 0.0,
               dropout      = 0.0,
               learn_rate   = -1
             )
           } else if (engine %in% c("h2o")) {
             list(
               hidden_units = 200L,
               penalty      = 0.0,
               dropout      = 0.5,
               epochs       = 10L,
               activation   = "relu",
               learn_rate   = 0.005
             )
           } else if (engine %in% c("keras")) {
             list(
               hidden_units = 5L,
               penalty      = 0.1,
               dropout      = 0.0,
               epochs       = 20L,
               activation   = "softmax"
             )
           } else {
             stop("Unsupported engine specified for mlp.")
           }
         },

         # 15. Deep Learning (keras)
         "deep_learning" = list(
           hidden_units = 10,
           penalty = 0.1,
           epochs = 50
         ),
         # 16. discrim_linear
         "discrim_linear" = list(),
         # 17. discrim_quad
         "discrim_quad" = list(),
         # 18. bag_tree
         "bag_tree" = list(
           min_n = 5
         ),
         # 19. Elastic Net
         "elastic_net" = list(
           penalty = 0.01,
           mixture = 0.5
         ),
         # 20. Bayesian GLM
         "bayes_glm" = list(),
         # 21. PLS
         "pls" = list(
           num_comp = 2
         ),
         # 22. Linear Regression
         "linear_reg" = list(),
         # 23. Ridge Regression
         "ridge_reg" = list(
           penalty = 0.01,
           mixture = 0
         ),
         # 24. Lasso Regression
         "lasso_reg" = list(
           penalty = 0.01,
           mixture = 1
         ),
         "penalized_cox" = list(
           penalty = 0.01,
           mixture = 1
         ),
         NULL)
}

#' Get Default Tuning Parameters
#'
#' Returns a list of default tuning parameter ranges for a specified algorithm based on the provided training data, outcome label, and engine.
#'
#' @param algo A character string specifying the algorithm name. Supported values include: \code{"rand_forest"}, \code{"C5_rules"}, \code{"xgboost"}, \code{"lightgbm"}, \code{"logistic_reg"}, \code{"multinom_reg"}, \code{"decision_tree"}, \code{"svm_linear"}, \code{"svm_rbf"}, \code{"nearest_neighbor"}, \code{"naive_Bayes"}, \code{"mlp"}, \code{"deep_learning"}, \code{"discrim_linear"}, \code{"discrim_quad"}, \code{"bag_tree"}, \code{"elastic_net"}, \code{"bayes_glm"}, \code{"pls"}, \code{"linear_reg"}, \code{"ridge_reg"}, and \code{"lasso_reg"}.
#'
#' @param train_data A data frame containing the training data.
#'
#' @param label A character string specifying the name of the outcome variable in \code{train_data}. This column is excluded when calculating the number of predictors.
#'
#' @param engine A character string specifying the engine to be used for the algorithm. Different engines may have different tuning parameter ranges.
#'
#' @return A list of tuning parameter ranges for the specified algorithm. If no tuning parameters are defined for the given algorithm, the function returns \code{NULL}.
#'
#' @details The function first determines the number of predictors by removing the outcome variable (specified by \code{label}) from \code{train_data}. It then uses a \code{switch} statement to select a list of default tuning parameter ranges tailored for the specified algorithm and engine. The tuning ranges have been adjusted for efficiency and may include parameters such as \code{mtry}, \code{trees}, \code{min_n}, and others depending on the algorithm.
#'
#' @importFrom dplyr select
#' @importFrom rlang sym
#'
#' @export
get_default_tune_params <- function(algo, train_data, label, engine) {
  # Determine the number of predictors
  num_predictors <- ncol(train_data %>% select(-!!sym(label)))

  switch(algo,
         # 1. Random Forest
         "rand_forest" = list(
           mtry = c(1, max(1, floor(sqrt(num_predictors)))),
           trees = c(100, 200),  # Reduced upper limit for efficiency
           min_n = c(2, 5)
         ),

         # 4. C5_rules
         "C5_rules" = list(
           trees = c(1, 50),  # Reduced upper limit for efficiency
           min_n = c(2, 5)
         ),

         # 5. XGBoost
         "xgboost" = list(
           trees = c(50, 150),  # Reduced range for efficiency
           tree_depth = c(1, 5),  # Reduced maximum depth
           learn_rate = c(-2, -1),  # log scale
           loss_reduction = c(0, 5),  # Reduced upper limit
           min_n = c(2, 5),
           sample_size = c(0.5, 0.99),
           mtry = c(1, num_predictors)
         ),

         # 6. LightGBM
         "lightgbm" = list(
           trees = c(50, 150),  # Reduced range for efficiency
           tree_depth = c(1, 5),  # Reduced maximum depth
           learn_rate = c(-2, -1),  # log scale
           loss_reduction = c(0, 5),  # Reduced upper limit
           min_n = c(2, 5),
           sample_size = c(0.5, 0.99),
           mtry = c(1, num_predictors)
         ),

         # 7. Logistic Regression


         "logistic_reg" = {

           if (engine %in% c("glm", "gee", "glmer", "stan", "stan_glmer")) {
             list(penalty = NULL, mixture = NULL)
           } else if (engine %in% c("brulee", "glmnet", "h2o", "LiblineaR", "spark")) {
             list(penalty = c(-5, 0), mixture = c(0, 1))
           } else if (engine %in% c("keras")) {
             list(penalty = c(-5, 0), mixture = NULL)
           } else {
             # Default if engine not recognized
             list(penalty = c(-5, 0), mixture = c(0, 1))
           }

         },

         "multinom_reg" = {
           if (engine %in% c("nnet")) {
             # nnet::multinom() uses only a penalty (decay) parameter; its default is 0.0.
             list(penalty = 0.0, mixture = NULL)
           } else if (engine %in% c("brulee")) {
             # brulee defaults: penalty = 0.001 and mixture = 0.0.
             list(penalty = 0.001, mixture = 0.0)
           } else if (engine %in% c("glmnet")) {
             # glmnet: by convention, we use a penalty of 0.0 and a mixture default of 1.0 (pure lasso).
             list(penalty = 0.0, mixture = 1.0)
           } else if (engine %in% c("h2o")) {
             # h2o: if no fixed penalty is given, a heuristic is used; here we choose 0.0 with ridge (mixture = 0.0).
             list(penalty = 0.0, mixture = 0.0)
           } else if (engine %in% c("keras")) {
             # keras_mlp() only uses a penalty parameter (default 0.0); mixture is not applicable.
             list(penalty = 0.0, mixture = NULL)
           } else if (engine %in% c("spark")) {
             # sparklyr: defaults are typically 0.0 for both penalty and mixture.
             list(penalty = 0.0, mixture = 0.0)
           } else {
             # Fallback default if engine not recognized.
             list(penalty = 0.0, mixture = 0.0)
           }
         },


         # 9. Decision Tree
         "decision_tree" = list(
           cost_complexity = c(-5, 0),  # log scale
           tree_depth = c(1, 5),  # Reduced maximum depth
           min_n = c(2, 5)
         ),

         # 10. SVM Linear
         "svm_linear" = list(
           cost = c(-3, 3)  # log scale
         ),

         # 11. SVM Radial
         "svm_rbf" = list(
           cost = c(-3, 3),  # log scale
           rbf_sigma = c(-9, -1)  # log scale
         ),

         # 12. nearest_neighbor
         "nearest_neighbor" = list(
           neighbors = c(3, 7),  # Narrowed range for efficiency
           dist_power = c(1, 2)
         ),

         # 13. Naive Bayes
         "naive_Bayes" = list(
           smoothness = c(0, 1),
           Laplace = c(0, 1)
         ),

         # 14. Neural Network (nnet)
         "mlp" = list(
           hidden_units = c(1, 5),  # Reduced upper limit
           penalty = c(1e-05, 0.1),
           epochs = c(100, 150)  # Reduced upper limit
         ),

         # 15. Deep Learning (keras)
         "deep_learning" = list(
           hidden_units = c(10, 30),  # Reduced upper limit
           penalty = c(1e-05, 0.1),
           epochs = c(50, 100)  # Reduced upper limit
         ),

         # 16. discrim_linear
         "discrim_linear" = NULL,

         # 17. discrim_quad
         "discrim_quad" = NULL,

         # 18. bag_tree
         "bag_tree" = list(
           cost_complexity = c(-5, 0),  # log scale
           tree_depth = c(1, 5),  # Reduced maximum depth
           min_n = c(2, 5)
         ),

         # 19. Elastic Net
         "elastic_net" = list(
           penalty = c(-5, 0),  # log scale
           mixture = c(0, 1)
         ),

         # 20. Bayesian GLM
         "bayes_glm" = NULL,

         # 21. PLS
         "pls" = list(
           num_comp = c(1, min(5, num_predictors))  # Reduced upper limit
         ),

         # 22. Linear Regression
         "linear_reg" = NULL,

         # 23. Ridge Regression
         "ridge_reg" = list(
           penalty = c(-5, 0)  # log scale
         ),

         # 24. Lasso Regression
         "lasso_reg" = list(
           penalty = c(-5, 0)  # log scale
         ),

         # Default case
         NULL)
}

#' Get Default Parameters with Transparency Warnings
#'
#' Wrapper around \code{get_default_params} that optionally warns when fastml's
#' default parameters differ from parsnip/engine defaults.
#'
#' @param algo Algorithm name.
#' @param task Task type.
#' @param num_predictors Number of predictors (optional).
#' @param engine Engine name (optional).
#' @param warn_param_defaults Logical; if TRUE, warn about parameter differences.
#' @param verbose Logical; if TRUE, print parameter selections.
#'
#' @return A list of default parameters.
#'
#' @keywords internal
#' @export
get_default_params_with_warnings <- function(algo, task, num_predictors = NULL,
                                             engine = NULL, warn_param_defaults = TRUE,
                                             verbose = FALSE) {
  fastml_params <- get_default_params(algo, task, num_predictors, engine)

  if (is.null(fastml_params)) {
    return(fastml_params)
  }

  if (warn_param_defaults || verbose) {
    parsnip_params <- get_parsnip_default_params(algo, engine, task)

    if (!is.null(parsnip_params)) {
      differences <- list()
      common_params <- intersect(names(fastml_params), names(parsnip_params))

      for (param in common_params) {
        fastml_val <- fastml_params[[param]]
        parsnip_val <- parsnip_params[[param]]

        # Skip NULL comparisons
        if (is.null(fastml_val) && is.null(parsnip_val)) next

        # Check if different
        if (is.null(fastml_val) != is.null(parsnip_val) ||
            (!is.null(fastml_val) && !isTRUE(all.equal(fastml_val, parsnip_val)))) {
          differences[[param]] <- list(fastml = fastml_val, parsnip = parsnip_val)
        }
      }

      if (length(differences) > 0) {
        diff_msgs <- vapply(names(differences), function(p) {
          d <- differences[[p]]
          f_str <- if (is.null(d$fastml)) "NULL" else as.character(d$fastml)
          p_str <- if (is.null(d$parsnip)) "engine-default" else as.character(d$parsnip)
          sprintf("%s=%s (parsnip: %s)", p, f_str, p_str)
        }, character(1))

        msg <- sprintf("[%s/%s] fastml parameter defaults: %s",
                       algo, engine, paste(diff_msgs, collapse = ", "))

        if (verbose) {
          message(msg)
        } else if (warn_param_defaults) {
          # Only warn once per algo/engine combination
          warn_key <- paste0("params_", algo, "_", engine)
          if (!exists(warn_key, envir = .fastml_warned_defaults)) {
            warning(msg, "\nUse tune_params to override or engine_params for fixed values.", call. = FALSE)
            assign(warn_key, TRUE, envir = .fastml_warned_defaults)
          }
        }
      }
    }
  }

  fastml_params
}

#' Process Model and Compute Performance Metrics
#'
#' Finalizes a tuning result or utilizes an already fitted workflow to generate predictions on test data and compute performance metrics.
#'
#' @param model_obj A model object, which can be either a tuning result (an object inheriting from \code{"tune_results"}) or an already fitted workflow.
#' @param model_id A unique identifier for the model, used in warning messages if issues arise during processing.
#' @param task A character string indicating the type of task, either \code{"classification"} or \code{"regression"}.
#' @param test_data A data frame containing the test data on which predictions will be generated.
#' @param label A character string specifying the name of the outcome variable in \code{test_data}.
#' @param event_class For classification tasks, a character string specifying which event class to consider as positive (accepted values: \code{"first"} or \code{"second"}).
#' @param engine A character string specifying the modeling engine used. This parameter affects prediction types and metric computations.
#' @param train_data A data frame containing the training data used to fit tuned models.
#' @param metric A character string specifying the metric name used to select the best tuning parameters.
#'
#' @return A list with two components:
#'   \describe{
#'     \item{performance}{A data frame of performance metrics. For classification tasks, metrics include accuracy, kappa, sensitivity, specificity, precision, F-measure, and ROC AUC (when applicable). For regression tasks, metrics include RMSE, R-squared, and MAE.}
#'     \item{predictions}{A data frame containing the test data augmented with predicted classes and, when applicable, predicted probabilities.}
#'   }
#'
#' @details The function first checks if \code{model_obj} is a tuning result. If so, it attempts to:
#'   \itemize{
#'     \item Select the best tuning parameters using \code{tune::select_best} (note that the metric used for selection should be defined in the calling environment).
#'     \item Extract the model specification and preprocessor from \code{model_obj} using \code{workflows::pull_workflow_spec} and \code{workflows::pull_workflow_preprocessor}, respectively.
#'     \item Finalize the model specification with the selected parameters via \code{tune::finalize_model}.
#'     \item Rebuild the workflow using \code{workflows::workflow}, \code{workflows::add_recipe}, and \code{workflows::add_model}, and fit the finalized workflow with \code{parsnip::fit} on the supplied \code{train_data}.
#'   }
#'   If \code{model_obj} is already a fitted workflow, it is used directly.
#'
#'   For classification tasks, the function makes class predictions (and probability predictions if \code{engine} is not \code{"LiblineaR"}) and computes performance metrics using functions from the \code{yardstick} package. In binary classification, the positive class is determined based on the \code{event_class} argument and ROC AUC is computed accordingly. For multiclass classification, macro-averaged metrics are reported and ROC AUC uses the \code{multiclass_auc} estimator (default \code{"macro"}; \code{"macro_weighted"} weights by class prevalence).
#'
#'   For regression tasks, the function predicts outcomes and computes regression metrics (RMSE, R-squared, and MAE).
#'
#'   If the number of predictions does not match the number of rows in \code{test_data}, the function stops with an informative error message regarding missing values and imputation options.
#'
#' @importFrom tune select_best finalize_model
#' @importFrom workflows pull_workflow_spec pull_workflow_preprocessor workflow add_recipe add_model
#' @importFrom parsnip fit
#' @importFrom dplyr select mutate bind_cols bind_rows
#' @importFrom yardstick metric_set accuracy kap sens spec precision f_meas roc_auc rmse rsq mae
#' @importFrom tibble tibble
#' @importFrom rlang sym
#' @importFrom stats predict
#' @importFrom magrittr %>%
#'
#' @export

#' Define Penalized Cox Regression Model Specification
#'
#' Constructs a penalized Cox proportional hazards specification using
#' `parsnip::proportional_hazards()` with the glmnet engine.
#'
#' @param task Character string specifying the task type. Must be
#'   "survival".
#' @param penalty Numeric non-negative penalty value (lambda) passed to
#'   glmnet. Set to `NULL` to use the engine default.
#' @param mixture Numeric value in [0, 1] controlling the elastic-net
#'   mixing parameter (alpha). Set to `NULL` to use the engine default.
#' @param engine Character string identifying the model engine. Defaults to
#'   "glmnet".
#' @param engine_params Optional named list of additional arguments passed to
#'   `parsnip::set_engine()`. Entries named `"control"` or
#'   `"control_parsnip"` are forwarded to `parsnip::fit()` instead.
#'
#' @return A list containing the parsnip model specification (`model_spec`)
#'   and any arguments that should be forwarded to `parsnip::fit()`
#'   (`fit_args`).
#'
#' @importFrom parsnip proportional_hazards set_mode set_engine
#' @noRd
define_penalized_cox_spec <- function(task,
                                      penalty = 0.01,
                                      mixture = 1,
                                      engine = "glmnet",
                                      engine_params = list()) {
  if (!identical(task, "survival")) {
    stop("Penalized Cox regression is only applicable for survival tasks.")
  }

  if (!is.null(penalty)) {
    penalty <- as.numeric(penalty)[1]
    if (!is.finite(penalty) || penalty < 0) {
      stop("'penalty' must be a non-negative finite number or NULL.")
    }
  }

  if (!is.null(mixture)) {
    mixture <- as.numeric(mixture)[1]
    if (!is.finite(mixture) || mixture < 0 || mixture > 1) {
      stop("'mixture' must be between 0 and 1 or NULL.")
    }
  }

  if (is.null(engine_params)) {
    engine_params <- list()
  }
  if (!is.list(engine_params)) {
    stop("'engine_params' must be provided as a list when supplied.")
  }

  fit_args <- list()
  if (length(engine_params) > 0) {
    control_names <- intersect(names(engine_params), c("control", "control_parsnip"))
    if (length(control_names) > 0) {
      for (nm in control_names) {
        fit_args[[nm]] <- engine_params[[nm]]
        engine_params[[nm]] <- NULL
      }
    }
    empty_names <- names(engine_params) == ""
    if (length(empty_names) > 0 && any(empty_names, na.rm = TRUE)) {
      stop("All entries in 'engine_params' must be named.")
    }
    engine_params <- engine_params[!vapply(engine_params, is.null, logical(1))]
  }

  model_spec <- proportional_hazards(
    penalty = penalty,
    mixture = mixture
  ) %>%
    set_mode("censored regression")

  if (length(engine_params) > 0) {
    model_spec <- do.call(set_engine, c(list(model_spec, engine), engine_params))
  } else {
    model_spec <- set_engine(model_spec, engine)
  }

  list(model_spec = model_spec, fit_args = fit_args)
}

#' Explain a fastml model using various techniques
#'
#' Provides model explainability across several backends. With \code{method = "dalex"} it:
#' \itemize{
#'   \item Creates a DALEX explainer from the trained model.
#'   \item Computes permutation-based variable importance with \code{vi_iterations} permutations and displays the table and plot.
#'   \item Computes partial dependence-like model profiles when \code{features} are supplied.
#'   \item Computes Shapley values (SHAP) for \code{shap_sample} training rows, displays the SHAP table,
#'   and plots a canonical SHAP summary (beeswarm) plot colored by raw feature values and ordered by
#'   \eqn{\text{mean}(\vert \text{SHAP value} \vert)} per feature. For classification, separate panels per class are shown.
#' }
#'
#' @details
#'  \itemize{
#'    \item \bold{Data source selection:} By default, explanations are computed on training data
#'    (\code{data = "train"}), which reflects in-sample model behavior and may be influenced by

#'    overfitting. Set \code{data = "test"} to compute explanations on held-out test data for
#'    a more realistic assessment of how the model uses features on unseen data.
#'    \item \bold{Method dispatch:} \code{method} can route to LIME, ICE, ALE, surrogate tree, interaction strengths,
#'    DALEX/modelStudio dashboards, fairness diagnostics, iBreakDown contributions, or counterfactual search.
#'    \item \bold{Variable importance controls:} Use \code{vi_iterations} to tune permutation stability and \code{loss_function}
#'    to override the default DALEX loss (cross-entropy for classification, RMSE for regression).
#'    \item \bold{Fairness and breakdown support:} Provide \code{protected} for \code{method = "fairness"} and an \code{observation}
#'    for \code{method = "breakdown"} or \code{method = "counterfactual"}. Observations are aligned to the explainer data before scoring.
#'  }
#'
#' @note By default, explanations use training data. For unbiased feature importance estimates
#' that better reflect model generalization, use \code{data = "test"} to compute explanations
#' on held-out test data.
#'
#' @param object A \code{fastml} object.
#' @param data Character string specifying which data to use for explanations:
#'   \code{"train"} (default) uses training data, \code{"test"} uses held-out test data.
#'   Using test data provides explanations that better reflect model generalization,
#'   while training data explanations may be influenced by overfitting.
#' @param method Character string specifying the explanation method.
#'   Supported values are \code{"dalex"}, \code{"lime"}, \code{"ice"},
#'   \code{"ale"}, \code{"surrogate"}, \code{"interaction"}, \code{"studio"},
#'   \code{"fairness"}, \code{"breakdown"}, and \code{"counterfactual"}.
#'   Defaults to \code{"dalex"}.
#' @param features Character vector of feature names for partial dependence (model profiles). Default NULL.
#' @param n_features Number of features to show in the explanation (used for lime). Default 5.
#' @param variables Character vector. Variable names to compute explanations for (used for counterfactuals).
#' @param observation A single observation for methods that need a new data point
#'   (\code{method = "lime"}, \code{method = "counterfactual"}, or \code{method = "breakdown"}). Default NULL.
#' @param grid_size Number of grid points for partial dependence. Default 20.
#' @param shap_sample Integer number of observations from the selected data source to compute SHAP values for. Default 5.
#' @param vi_iterations Integer. Number of permutations for variable importance (B). Default 10.
#' @param seed Integer. A value specifying the random seed.
#' @param loss_function Function. The loss function for \code{model_parts}.
#'   \itemize{
#'     \item If \code{NULL} and task = 'classification', defaults to \code{DALEX::loss_cross_entropy}.
#'     \item If \code{NULL} and task = 'regression', defaults to \code{DALEX::loss_root_mean_square}.
#'   }
#' @param protected Character or factor vector of protected attribute(s) required for
#'   \code{method = "fairness"}. Default NULL.
#' @param ... Additional arguments passed to the underlying helper functions
#'   for the chosen \code{method}.
#'
#' @importFrom dplyr select
#' @importFrom tune extract_fit_parsnip
#' @importFrom ggplot2 labs
#' @importFrom stats predict
#'
#' @return For DALEX-based methods, prints variable importance, model profiles, and SHAP summaries.
#'   Other methods return their respective explainer objects (e.g., LIME explanations, ALE plot,
#'   surrogate tree, interaction strengths, modelStudio dashboard, fairmodels object, breakdown
#'   object, or counterfactual results), usually invisibly after plotting or printing.
#' @export
fastexplain <- function(object,
                          method = "dalex",
                          data = c("train", "test"),
                          features = NULL,
                          n_features = 5,
                          variables = NULL,
                          observation = NULL,
                          grid_size = 20,
                          shap_sample = 5,
                          vi_iterations = 10,
                          seed = 123,
                          loss_function = NULL,
                          protected = NULL,
                          ...) {

  if (!inherits(object, "fastml")) {
    stop("The input must be a 'fastml' object.")
  }

  data <- match.arg(data)

  if (data == "train") {
    message("Note: Explanations are computed on TRAINING data and reflect in-sample model behavior.")
    message("      Use data = 'test' for held-out explanations that better reflect generalization.")
  } else {
    message("Note: Explanations are computed on TEST (held-out) data.")
  }

  method <- tolower(method)

  # Helper to ensure new observations are preprocessed before scoring
  preprocess_observation <- function(obs) {
    if (is.null(obs)) return(NULL)
    if (!is.null(object$preprocessor)) {
      tryCatch(
        recipes::bake(object$preprocessor, new_data = obs),
        error = function(e) obs
      )
    } else {
      obs
    }
  }

  # Legacy/specific methods
  if (method == "lime") {
    if (is.null(observation)) {
      stop("'observation' must be provided for method = 'lime'.")
    }
    return(explain_lime(object, new_observation = observation, data = data, n_features = n_features, ...))
  }
  if (method == "ice") return(plot_ice(object, features = features, data = data, ...))
  if (method == "ale") {
    if (is.null(features) || length(features) == 0) {
      stop("'features' must contain a feature name for ALE explanations.")
    }
    return(explain_ale(object, feature = features[1], data = data, ...))
  }
  if (method == "surrogate") return(surrogate_tree(object, data = data, ...))
  if (method == "interaction") return(interaction_strength(object, data = data, ...))
  # DALEX ecosystem methods reuse centralized explainer builder
  prep <- fastml_prepare_explainer_inputs(object, data = data)
  dalex_res <- fastml_build_dalex_explainers(prep)
  explainer <- dalex_res$explainers[[1]]

  if (method == "dalex") {
    return(explain_dalex_internal(
      explainers = dalex_res$explainers,
      prep = prep,
      features = features,
      grid_size = grid_size,
      shap_sample = shap_sample,
      vi_iterations = vi_iterations,
      seed = seed,
      loss_function = loss_function
    ))
  } else if (method == "studio") {
    if (!requireNamespace("modelStudio", quietly = TRUE)) {
      stop("Package 'modelStudio' required for method = 'studio'.")
    }
    return(modelStudio::modelStudio(explainer, ...))
  } else if (method == "fairness") {
    if (is.null(protected)) {
      stop("Argument 'protected' is required for method = 'fairness'.")
    }
    if (!requireNamespace("fairmodels", quietly = TRUE)) {
      stop("Package 'fairmodels' required for method = 'fairness'.")
    }
    fairness_obj <- fairmodels::fairness_check(explainer, protected = protected, ...)
    print(plot(fairness_obj))
    return(invisible(fairness_obj))
  } else if (method == "breakdown") {
    if (is.null(observation)) {
      stop("'observation' must be provided for method = 'breakdown'.")
    }
    if (!requireNamespace("iBreakDown", quietly = TRUE)) {
      stop("Package 'iBreakDown' required for method = 'breakdown'.")
    }
    if (prep$task == "survival") {
      message("Breakdown plots are skipped for survival tasks because reducing a survival curve to a single number (e.g., median time) can be undefined or misleading. Variable importance and SHAP remain available.")
      return(invisible(NULL))
    }
    # Align observation to the explainer's expected columns (raw predictors).
    obs_aligned <- as.data.frame(observation)
    if (!is.null(explainer$data)) {
      expl_cols <- colnames(explainer$data)
      # Add any missing predictors so the recipe/predict_function can handle them.
      missing_cols <- setdiff(expl_cols, colnames(obs_aligned))
      if (length(missing_cols) > 0) {
        for (nm in missing_cols) {
          tmpl <- explainer$data[[nm]]
          obs_aligned[[nm]] <- if (is.factor(tmpl)) factor(NA, levels = levels(tmpl)) else NA
        }
      }
      obs_aligned <- obs_aligned[, expl_cols, drop = FALSE]
    }
    bd <- suppressWarnings(iBreakDown::break_down(explainer, new_observation = obs_aligned, ...))
    bd_plot <- suppressWarnings(tryCatch(
      plot(bd),
      error = function(e) {
        warning("Breakdown plot could not be rendered: ", e$message)
        NULL
      }
    ))
    if (!is.null(bd_plot)) {
      print(bd_plot)
    }
    return(invisible(bd))
  } else if (method == "counterfactual") {
    if (is.null(observation)) {
      stop("'observation' must be provided for counterfactual explanations.")
    }

    # --- FIX: Suppress Warnings for GLM Rank-Deficiency ---
    return(suppressWarnings(counterfactual_explain(
      object = explainer,
      observation = observation,
      variables = variables,
      positive_class = prep$positive_class,
      event_class = prep$event_class,
      label_levels = prep$label_levels,
      ...
    )))

  } else {
    stop("Unknown explanation method.")
  }
}

library(testthat)
library(dplyr)

skip_on_cran()
skip_if_not_installed("lime")
skip_if_not_installed("pdp")
skip_if_not_installed("iml")
skip_if_not_installed("ceterisParibus")

# Prevent Rplots.pdf from being created by graphics calls
pdf(NULL)
withr::defer(dev.off())

# small dataset for tests
set.seed(123)
data(iris)
iris <- iris[iris$Species != "setosa", ]
iris$Species <- factor(iris$Species)

model <- fastml(
  data = iris,
  label = "Species",
  algorithms = c("rand_forest"),
  resampling_method = "cv",
  folds = 2
)



test_that("ALE explanation runs", {
  expect_error(
    suppressMessages(suppressWarnings(fastexplain(model, method = "ale", features = "Sepal.Length"))),
    regexp = NA
  )
})

test_that("DALEX explanation runs cleanly", {
  expect_error(
    suppressMessages(suppressWarnings(fastexplain(model))),
    regexp = NA
  )
})

test_that("Counterfactual explanation computes counterfactuals", {
  cf <- suppressWarnings(fastexplain(model, method = "counterfactual", observation = iris[1, ]))
  expect_true(is.list(cf))
})

#' Predict method for fastml objects
#'
#' Generates predictions from a trained `fastml` object on new data.
#' Supports both single-model and multi-model workflows, and handles classification
#' and regression tasks with optional post-processing and verbosity.
#'
#' @param object A fitted `fastml` object created by the `fastml()` function.
#' @param newdata A data frame or tibble containing new predictor data for which to generate predictions.
#' @param type Type of prediction to return. One of `"auto"` (default), `"class"`, `"prob"`, `"numeric"`,
#'   `"survival"`, or `"risk"`.
#'   - `"auto"`: chooses `"class"` for classification, `"numeric"` for regression, and `"survival"` for survival.
#'   - `"prob"`: returns class probabilities (only for classification).
#'   - `"class"`: returns predicted class labels.
#'   - `"numeric"`: returns predicted numeric values (for regression).
#'   - `"survival"`: returns survival probabilities at the supplied `eval_time` horizons (for survival tasks).
#'   - `"risk"`: returns risk scores on the linear predictor scale (for survival tasks).
#' @param model_name (Optional) Name of a specific model to use when `object$best_model` contains multiple models.
#' @param verbose Logical; if `TRUE`, prints progress messages showing which models are used during prediction.
#' @param postprocess_fn (Optional) A function to apply to the final predictions (e.g., inverse transforms, thresholding).
#' @param eval_time Optional numeric vector of time points (on the original time
#'   scale) at which to return survival probabilities when `type = "survival"`.
#'   Required for survival tasks when requesting survival curves.
#' @param ... Additional arguments (currently unused).
#'
#' @return A vector of predictions, or a named list of predictions (if multiple models are used).
#'         If `postprocess_fn` is supplied, its output will be returned instead.
#'
#' @method predict fastml
#' @importFrom stats predict
#' @importFrom tibble is_tibble
#' @export
predict.fastml <- function(object, newdata,
                           type = "auto",
                           model_name = NULL,
                           verbose = FALSE,
                           postprocess_fn = NULL,
                           eval_time = NULL,
                           ...) {
  # 1. input checks -------------------------------------------------------
  if (missing(newdata)) {
    stop("Please provide newdata for prediction.")
  }

  # --- FIX 1: Cast integers to numeric to prevent recipe imputation errors ---
  # Recipes are strict about type. If training data was integer but imputation
  # adds decimals (median/mean), applying it to integer test data crashes.
  if (is.data.frame(newdata)) {
    newdata[] <- lapply(newdata, function(x) {
      if (is.integer(x)) as.numeric(x) else x
    })
  }

  # 2. preprocessing via recipe -------------------------------------------
  if (is.null(object$preprocessor)) {
    stop("No preprocessing recipe found in fastml object.")
  }

  # Pass full data to bake; the recipe handles label removal via step_rm
  new_proc <- recipes::bake(object$preprocessor, new_data = newdata)

  if (!is.null(object$feature_names)) {
    miss <- setdiff(object$feature_names, names(new_proc))
    if (length(miss)) {
      stop("Missing features after preprocessing: ", paste(miss, collapse = ", "))
    }
  }

  # 3. pick prediction type ------------------------------------------------
  predict_type <- switch(type,
                         auto = if (object$task == "classification") "class" else if (object$task == "survival") "survival" else "numeric",
                         prob = {
                           if (object$task != "classification") {
                             warning("Probabilities only for classification; using numeric.")
                             "numeric"
                           } else "prob"
                         },
                         survival = "survival",
                         risk = "risk",
                         type
  )

  # Helper: standardized risk/linear predictor extraction -----------------
  fastml_predict_risk <- function(model, new_data, ...) {
    if (inherits(model, "fastml_native_survival")) {
      native_lp <- tryCatch(
        predict_risk(model, newdata = new_data, ...),
        error = function(e) NULL
      )
      if (!is.null(native_lp)) {
        return(as.numeric(native_lp))
      }
      stop("Risk prediction for this native survival model is not available.", call. = FALSE)
    }

    lp <- tryCatch(
      predict(model, new_data = new_data, type = "linear_pred", ...),
      error = function(e) NULL
    )

    if (is.null(lp)) {
      stop("This model does not support risk/linear predictor scores.", call. = FALSE)
    }

    if (is.data.frame(lp) || tibble::is_tibble(lp)) {
      if (".pred" %in% names(lp)) return(as.numeric(lp$.pred))
      if (".pred_lp" %in% names(lp)) return(as.numeric(lp$.pred_lp))
      return(as.numeric(lp[[1]]))
    }

    as.numeric(lp)
  }

  # Helper: normalize nested model lists into a flat, named list ----------
  normalize_models <- function(mods, engine_names = NULL) {
    if (!is.list(mods)) return(list())
    if (all(vapply(mods, valid_model, logical(1)))) return(mods)

    out <- list()
    for (algo in names(mods)) {
      entry <- mods[[algo]]
      if (inherits(entry, "workflow") || inherits(entry, "fastml_native_survival") || inherits(entry, "fastml_royston")) {
        eng <- tryCatch(engine_names[[algo]], error = function(e) NULL)
        eng <- if (is.null(eng) || length(eng) == 0 || is.na(eng[1])) NULL else eng[1]
        nm  <- if (!is.null(eng)) paste0(algo, " (", eng, ")") else algo
        out[[nm]] <- entry
        next
      }
      if (is.list(entry)) {
        inner_names <- names(entry)
        for (j in seq_along(entry)) {
          wf <- entry[[j]]
          if (!(inherits(wf, "workflow") || inherits(wf, "fastml_native_survival") || inherits(wf, "fastml_royston"))) next
          eng <- if (!is.null(inner_names) && !is.na(inner_names[j]) && nzchar(inner_names[j])) inner_names[j] else NULL
          nm <- if (!is.null(eng)) paste0(algo, " (", eng, ")") else algo
          out[[nm]] <- wf
        }
      }
    }
    out
  }

  resolve_model_names <- function(requested, available) {
    if (length(requested) == 0) return(character())
    avail_set <- unique(available)
    base_map <- stats::setNames(avail_set, sub(" \\(.*\\)$", "", avail_set))
    vapply(requested, function(nm) {
      if (nm %in% avail_set) nm else if (nm %in% names(base_map)) base_map[[nm]] else NA_character_
    }, character(1))
  }

  # Helper: check if object is a valid model type for prediction

  valid_model <- function(x) {
    inherits(x, "workflow") ||
      inherits(x, "fastml_native_survival") ||
      inherits(x, "fastml_royston")
  }

  # 4. choose which workflows to use ---------------------------------------
  all_mods <- normalize_models(object$models, object$engine_names)
  if (!length(all_mods) || !all(vapply(all_mods, valid_model, logical(1)))) {
    stop("No valid `models` slot in fastml object.")
  }

  if (!is.null(model_name)) {
    resolved <- resolve_model_names(model_name, names(all_mods))
    bad <- setdiff(model_name, model_name[!is.na(resolved)])
    if (length(bad)) {
      stop("Requested model_name not found: ", paste(bad, collapse = ", "))
    }
    to_predict <- all_mods[na.omit(resolved)]
  } else {
    bm <- object$best_model
    if (inherits(bm, "workflow") || inherits(bm, "fastml_royston")) {
      to_predict <- list(bm)
      names(to_predict) <- if (!is.null(names(bm))) names(bm) else "best_model"
    } else if (is.list(bm) && all(sapply(bm, valid_model))) {
      to_predict <- bm
    } else {
      to_predict <- all_mods
    }
  }

  # 5. run predictions ------------------------------------------------------
  preds <- lapply(names(to_predict), function(nm) {
    wf <- to_predict[[nm]]
    if (verbose) message("Predicting with: ", nm)

    # Workflows and native survival models expect raw data; native handlers bake internally.
    new_data_for_predict <- if (inherits(wf, "fastml_native_survival") || inherits(wf, "workflow")) {
      newdata
    } else {
      new_proc
    }

    if (object$task == "survival") {
      if (identical(predict_type, "survival")) {
        times <- eval_time
        if (is.null(times)) {
          stop("Please supply 'eval_time' when requesting survival probabilities.")
        }
        times <- as.numeric(times)
        times <- sort(unique(times[is.finite(times) & times >= 0]))
        if (length(times) == 0) {
          stop("No valid evaluation times supplied in 'eval_time'.")
        }

        surv_pred <- tryCatch({
          # --- FIX 2: Handle Native vs Workflow Separately ---
          if (inherits(wf, "fastml_native_survival")) {
            # Native helper needs 'newdata'
            predict_survival(wf, newdata = new_data_for_predict, times = times, ...)
          } else {
            # Workflows use standard predict with 'new_data'
            # 1. Get nested tibble prediction
            raw_pred <- predict(wf, new_data = new_data_for_predict, type = "survival", eval_time = times)
            # 2. Convert nested list column to Matrix (Rows=Obs, Cols=Time)
            probs_list <- lapply(raw_pred$.pred, function(x) x$.pred_survival)
            do.call(rbind, probs_list)
          }
        }, error = function(e) {
          stop(sprintf("Failed to obtain survival predictions for model '%s': %s", nm, e$message), call. = FALSE)
        })

        if (is.matrix(surv_pred) && ncol(surv_pred) == length(times)) {
          colnames(surv_pred) <- format(times, trim = TRUE, scientific = FALSE)
        }
        if (!is.null(postprocess_fn)) {
          surv_pred <- postprocess_fn(surv_pred)
        }
        return(surv_pred)
      }

      if (identical(predict_type, "risk")) {
        risk_pred <- tryCatch({
          # --- FIX 3: Map user-facing newdata to internal new_data requirement ---
          fastml_predict_risk(wf, new_data = new_data_for_predict, ...)
        }, error = function(e) {
          stop(sprintf("Failed to obtain risk predictions for model '%s': %s", nm, e$message), call. = FALSE)
        })
        if (!is.null(postprocess_fn)) {
          risk_pred <- postprocess_fn(risk_pred)
        }
        return(as.numeric(risk_pred))
      }

      stop("Unsupported prediction type for survival task: ", predict_type)
    }

    # Standard predictions (Classification / Regression)
    p <- predict(wf, new_data = new_data_for_predict, type = predict_type, ...)
    if (is.data.frame(p) || tibble::is_tibble(p)) {
      p <- switch(predict_type,
                  class   = p$.pred_class,
                  numeric = p$.pred,
                  prob    = p
      )
    } else {
      stop("Unsupported model type for prediction.")
    }
    if (!is.null(postprocess_fn)) {
      p <- postprocess_fn(p)
    }
    p
  })
  names(preds) <- names(to_predict)

  # 6. return simple vs. list ---------------------------------------------
  if (length(preds) == 1) {
    return(preds[[1]])
  }
  class(preds) <- "fastml_prediction"
  preds
}

#' Define Random Forest Model Specification
#'
#' @param task Character string specifying the task type: "classification",
#'   "regression", or "survival".
#' @param train_data Data frame containing the training data.
#' @param label Character string specifying the name of the target variable.
#' @param tune Logical indicating whether to use tuning parameters.
#' @return List containing the model specification (`model_spec`).
#' @importFrom parsnip rand_forest set_mode set_engine
#' @importFrom dplyr select all_of
#' @importFrom tune tune
#' @noRd
define_rand_forest_spec <- function(task, train_data, label, tuning = FALSE, engine = "ranger") {
  # Remove the label column to get predictors and count them
  predictors <- train_data %>% dplyr::select(-dplyr::all_of(label))
  num_predictors <- ncol(predictors)

  # Retrieve default parameters for a random forest.
  # (Assumes get_default_params returns a list with mtry, trees, and min_n.)
  defaults <- get_default_params("rand_forest", task, num_predictors, engine)

  if (task == "survival") {

    # Ensure censored is available so it can register parsnip's survival mode/engines.
    if (!requireNamespace("censored", quietly = TRUE)) {
      stop("Package 'censored' is required for survival random forest.", call. = FALSE)
    }

    # Censored registers parsnip survival models on namespace load.

    # Build model spec
    if (tuning) {
      model_spec <- parsnip::rand_forest(
        mtry  = tune(),
        trees = tune(),
        min_n = tune()
      )
    } else {
      model_spec <- parsnip::rand_forest(
        mtry  = defaults$mtry,
        trees = defaults$trees,
        min_n = defaults$min_n
      )
    }

    model_spec <- model_spec %>%
      parsnip::set_mode("censored regression")

    if (engine == "ranger") {
      model_spec <- model_spec %>% parsnip::set_engine("ranger")
    } else if (engine == "aorsf") {
      if (!requireNamespace("aorsf", quietly = TRUE)) {
        stop("Package 'aorsf' is required for this engine.", call. = FALSE)
      }
      model_spec <- model_spec %>% parsnip::set_engine("aorsf")
    } else {
      stop("Unsupported engine for survival random forest.", call. = FALSE)
    }
  } else {

    # Build model specification
    if (tuning) {
      model_spec <- parsnip::rand_forest(
        mtry  = tune(),
        trees = tune(),
        min_n = tune()
      )
    } else {
      model_spec <- parsnip::rand_forest(
        mtry  = defaults$mtry,
        trees = defaults$trees,
        min_n = defaults$min_n
      )
    }

    # Set model mode
    model_spec <- model_spec %>%
      parsnip::set_mode(task)

    # Engine-specific parameter settings
    if (engine == "ranger") {
      if (task == "classification") {
        model_spec <- model_spec %>%
          parsnip::set_engine("ranger", probability = TRUE)
      } else {
        model_spec <- model_spec %>%
          parsnip::set_engine("ranger")
      }
    } else if (engine == "h2o") {
      model_spec <- model_spec %>%
        parsnip::set_engine("h2o")
    } else if (engine == "randomForest") {
      model_spec <- model_spec %>%
        parsnip::set_engine("randomForest")
    } else if (engine == "spark") {
      model_spec <- model_spec %>%
        parsnip::set_engine("spark", seed = 1234)
    } else if (engine == "partykit") {
      model_spec <- model_spec %>%
        parsnip::set_engine("partykit")
    } else {
      stop("Unsupported engine specified for rand_forest.", call. = FALSE)
    }
  }

  # Return a named list for consistency with fastml framework
  list(model_spec = model_spec)
}

#' Define bag_tree Model Specification
#'
#' @inheritParams define_decision_tree_spec
#' @return List containing the model specification (`model_spec`).
#' @importFrom parsnip bag_tree set_mode set_engine
#' @importFrom tune tune
#' @noRd
define_bag_tree_spec <- function(task, tuning = FALSE, engine = "rpart") {
  if (!requireNamespace("baguette", quietly = TRUE)) {
    stop("Package 'baguette' is required for bagged tree models. Install it with: install.packages('baguette')")
  }
  defaults <- get_default_params("bag_tree")
  if (tuning) {
    model_spec <- bag_tree(
      min_n = tune()
    ) %>%
      set_mode(task) %>%
      set_engine(engine, times = 25)
  } else {
    model_spec <- bag_tree(
      min_n = defaults$min_n
    ) %>%
      set_mode(task) %>%
      set_engine(engine, times = 25)
  }
  list(model_spec = model_spec)
}


#' Define Decision Tree Model Specification
#'
#' @inheritParams define_rand_forest_spec
#' @param task A character string indicating the mode (e.g., "classification" or "regression").
#' @param tuning Logical; if \code{TRUE}, the tuning parameters are set to \code{tune()}.
#' @param engine A character string specifying the computational engine. Supported engines include
#'   "rpart", "C5.0", and "partykit".
#'
#' @return List containing the model specification (`model_spec`).
#'
#' @details
#' The available parameters and their defaults vary by engine:
#'
#' - **rpart:**
#'   \itemize{
#'     \item \code{tree_depth}: Tree Depth (default: 30L)
#'     \item \code{min_n}: Minimal Node Size (default: 2L)
#'     \item \code{cost_complexity}: Cost-Complexity Parameter (default: 0.01)
#'   }
#'
#' - **C5.0:**
#'   \itemize{
#'     \item \code{min_n}: Minimal Node Size (default: 2L)
#'   }
#'
#' - **partykit:**
#'   \itemize{
#'     \item \code{tree_depth}: Tree Depth (default: 30L; adjust as needed)
#'     \item \code{min_n}: Minimal Node Size (default: 20L)
#'   }
#'
#' @importFrom parsnip decision_tree set_mode set_engine
#' @importFrom tune tune
#' @noRd
define_decision_tree_spec <- function(task, tuning = FALSE, engine = "rpart") {

  # Set engine-specific defaults
  defaults <- switch(engine,
                     "rpart" = list(tree_depth = 30L, min_n = 2L, cost_complexity = 0.01),
                     "C5.0" = list(min_n = 2L),
                     "partykit" = list(tree_depth = 30L, min_n = 20L),
                     # fallback: assume rpart defaults
                     list(tree_depth = 30L, min_n = 2L, cost_complexity = 0.01)
  )

  # Build the model specification based on the engine and tuning flag.
  if (engine == "rpart") {
    if (tuning) {
      model_spec <- decision_tree(
        tree_depth = tune(),
        min_n = tune(),
        cost_complexity = tune()
      ) %>%
        set_mode(task) %>%
        set_engine(engine)
    } else {
      model_spec <- decision_tree(
        tree_depth = defaults$tree_depth,
        min_n = defaults$min_n,
        cost_complexity = defaults$cost_complexity
      ) %>%
        set_mode(task) %>%
        set_engine(engine)
    }
  } else if (engine == "C5.0") {
    if (tuning) {
      model_spec <- decision_tree(
        min_n = tune()
      ) %>%
        set_mode(task) %>%
        set_engine(engine)
    } else {
      model_spec <- decision_tree(
        min_n = defaults$min_n
      ) %>%
        set_mode(task) %>%
        set_engine(engine)
    }
  } else if (engine == "partykit") {
    if (tuning) {
      model_spec <- decision_tree(
        tree_depth = tune(),
        min_n = tune()
      ) %>%
        set_mode(task) %>%
        set_engine(engine)
    } else {
      model_spec <- decision_tree(
        tree_depth = defaults$tree_depth,
        min_n = defaults$min_n
      ) %>%
        set_mode(task) %>%
        set_engine(engine)
    }
  } else {
    # Fallback for unknown engine: assume rpart defaults
    if (tuning) {
      model_spec <- decision_tree(
        tree_depth = tune(),
        min_n = tune(),
        cost_complexity = tune()
      ) %>%
        set_mode(task) %>%
        set_engine(engine)
    } else {
      model_spec <- decision_tree(
        tree_depth = defaults$tree_depth,
        min_n = defaults$min_n,
        cost_complexity = defaults$cost_complexity
      ) %>%
        set_mode(task) %>%
        set_engine(engine)
    }
  }

  list(model_spec = model_spec)
}


#' Define Logistic Regression Model Specification
#'
#' @param task Character string specifying the task type ("classification").
#' @param tune Logical indicating whether to use tuning parameters.
#' @return List containing the model specification (`model_spec`).
#' @importFrom parsnip logistic_reg set_mode set_engine
#' @importFrom tune tune
#' @importFrom tune tune
#' @noRd

#' Analyze Feature Importance Stability Across Cross-Validation Folds
#'
#' Computes feature importance for each fold model and aggregates results
#' to assess the stability of feature importance rankings across resamples.
#' This helps identify features that are consistently important vs those
#' whose importance varies across different data subsets.
#'
#' @param object A \code{fastml} object trained with \code{store_fold_models = TRUE}.
#' @param model_name Character string specifying which model to analyze. If NULL,
#'   uses the best model. Should match the format "algorithm (engine)", e.g.,
#'   "rand_forest (ranger)".
#' @param vi_iterations Integer. Number of permutations for variable importance
#'   per fold. Default is 10 for faster computation across many folds.
#' @param seed Integer. Random seed for reproducibility.
#' @param plot Logical. If TRUE, displays a stability plot showing
#'   mean importance with confidence intervals. Default is FALSE.
#' @param conf_level Numeric. Confidence level for intervals. Default is 0.95.
#'
#' @return A list with class \code{"fastml_stability"} containing:
#'   \describe{
#'     \item{importance_summary}{Data frame with aggregated feature importance
#'       (mean, sd, se, lower/upper CI) across folds.}
#'     \item{fold_importance}{List of per-fold variable importance results.}
#'     \item{rank_stability}{Data frame showing how feature ranks vary across folds.}
#'     \item{n_folds}{Number of folds analyzed.}
#'     \item{model_name}{Name of the model analyzed.}
#'   }
#'
#' @details
#' This function requires that the fastml model was trained with
#' \code{store_fold_models = TRUE}, which stores the models fitted on each
#' cross-validation fold. Without stored fold models, only the final best
#' model is available, and cross-fold stability analysis is not possible.
#'
#' The stability analysis computes permutation-based variable importance
#' for each fold's model using DALEX, then aggregates across folds to show:
#' \itemize{
#'   \item Mean importance and standard deviation
#'   \item Confidence intervals for importance
#'   \item Rank stability (how consistently features rank across folds)
#' }
#'
#' Features with high mean importance but also high variance may be
#' important for some data subsets but not others, suggesting potential
#' instability in the model's reliance on those features.
#'
#' @examples
#' \donttest{
#' # Train model with fold models stored
#' model <- fastml(
#'   data = iris,
#'   label = "Species",
#'   algorithms = "rand_forest",
#'   store_fold_models = TRUE
#' )
#'
#' # Analyze stability
#' stability <- explain_stability(model)
#' print(stability)
#' plot(stability)
#' }
#'
#' @importFrom rlang .data
#' @importFrom stats aggregate qnorm sd
#' @importFrom utils head
#' @importFrom tune extract_fit_parsnip
#' @importFrom workflows extract_recipe extract_fit_parsnip
#' @importFrom recipes bake
#' @importFrom ggplot2 ggplot aes geom_point geom_errorbarh labs theme_minimal theme element_text
#' @export
explain_stability <- function(object,
                              model_name = NULL,
                              vi_iterations = 10,
                              seed = 123,
                              plot = FALSE,
                              conf_level = 0.95) {
  if (!inherits(object, "fastml")) {
    stop("The input must be a 'fastml' object.")
  }

  if (!requireNamespace("DALEX", quietly = TRUE)) {
    stop("Package 'DALEX' is required for stability analysis.")
  }

  if (!requireNamespace("ggplot2", quietly = TRUE)) {
    stop("Package 'ggplot2' is required for plotting.")
  }

  if (!requireNamespace("workflows", quietly = TRUE)) {
    stop("Package 'workflows' is required for stability analysis.")
  }

  if (!requireNamespace("recipes", quietly = TRUE)) {
    stop("Package 'recipes' is required for stability analysis.")
  }

  # Check if fold models are available
  if (is.null(object$resampling_results)) {
    stop(
      "No resampling results found. ",
      "Ensure the model was trained with resampling (resampling_method != 'none')."
    )
  }

  # Determine which model to analyze
  if (is.null(model_name)) {
    # Use best model name
    if (!is.null(object$best_model_name)) {
      model_name <- paste0(
        names(object$best_model_name),
        " (",
        object$best_model_name,
        ")"
      )
    } else {
      model_name <- names(object$resampling_results)[1]
    }
  }

  if (!model_name %in% names(object$resampling_results)) {
    stop(
      sprintf(
        "Model '%s' not found in resampling results. Available models: %s",
        model_name,
        paste(names(object$resampling_results), collapse = ", ")
      )
    )
  }

  resample_entry <- object$resampling_results[[model_name]]

  # Check if fold models are stored
  if (is.null(resample_entry$fold_models)) {
    stop(
      "Fold models are not available for this model. ",
      "To enable stability analysis, retrain with store_fold_models = TRUE:\n",
      "  model <- fastml(..., store_fold_models = TRUE)"
    )
  }

  fold_models <- resample_entry$fold_models
  fold_train_data <- resample_entry$fold_train_data
  n_folds <- length(fold_models)

  if (n_folds < 2) {
    stop("At least 2 folds are required for stability analysis.")
  }

  message(sprintf("Computing variable importance for %d folds...", n_folds))

  # Prepare common inputs
  task <- object$task
  label <- object$label
  positive_class <- object$positive_class
  event_class <- object$event_class

  # Extract label levels from training data
  label_levels <- NULL
  if (!is.null(object$raw_train_data) && label %in% names(object$raw_train_data)) {
    y_train <- object$raw_train_data[[label]]
    if (is.factor(y_train)) {
      label_levels <- levels(y_train)
    }
  }

  # Build loss function (same as explain_dalex)
  loss_function <- if (task == "classification") {
    function(observed, predicted) {
      eps <- 1e-15
      if (is.data.frame(predicted) || is.matrix(predicted)) {
        pred_mat <- as.matrix(predicted)
        obs_chr <- as.character(observed)
        col_idx <- if (!is.null(colnames(pred_mat))) {
          match(obs_chr, colnames(pred_mat))
        } else {
          rep(NA_integer_, length(obs_chr))
        }
        p_true <- numeric(length(obs_chr))
        for (i in seq_along(obs_chr)) {
          if (!is.na(col_idx[i])) {
            p_true[i] <- pred_mat[i, col_idx[i]]
          } else {
            p_true[i] <- NA_real_
          }
        }
        p_true <- pmin(pmax(p_true, eps), 1 - eps)
        return(mean(-log(p_true), na.rm = TRUE))
      }
      obs_num <- if (is.factor(observed)) as.integer(observed) - 1 else as.numeric(observed)
      unique_vals <- unique(stats::na.omit(obs_num))
      if (length(unique_vals) == 2) {
        min_val <- min(unique_vals)
        max_val <- max(unique_vals)
        if (min_val != 0 || max_val != 1) {
          obs_num <- ifelse(obs_num == min_val, 0, 1)
        }
      }
      pred <- pmin(pmax(as.numeric(predicted), eps), 1 - eps)
      -mean(obs_num * log(pred) + (1 - obs_num) * log(1 - pred), na.rm = TRUE)
    }
  } else {
    DALEX::loss_root_mean_square
  }

  # Compute variable importance for each fold
  set.seed(seed)
  fold_vi_results <- vector("list", n_folds)

  for (i in seq_len(n_folds)) {
    fold_model <- fold_models[[i]]
    fold_data <- fold_train_data[[i]]

    if (is.null(fold_model) || is.null(fold_data)) {
      warning(sprintf("Fold %d has missing model or data, skipping.", i))
      next
    }

    # Use the workflow directly with raw data, similar to fastexplain()
    # The predict_function will handle preprocessing internally
    # This ensures DALEX permutes raw features (original feature names)
    model_fit <- fold_model

    # For non-workflow models, extract the parsnip fit
    if (!inherits(fold_model, "workflow")) {
      model_fit <- tryCatch(
        tune::extract_fit_parsnip(fold_model),
        error = function(e) {
          if (inherits(fold_model, "model_fit")) fold_model else NULL
        }
      )
    }

    if (is.null(model_fit)) {
      warning(sprintf("Could not extract model fit for fold %d, skipping.", i))
      next
    }

    # Use raw data (without label) for DALEX - it will permute original features
    x_data <- fold_data[, setdiff(names(fold_data), label), drop = FALSE]
    y_data <- fold_data[[label]]

    # Extract preprocessor for use in predict_function (workflows only)
    fold_preprocessor <- NULL
    if (inherits(fold_model, "workflow")) {
      fold_preprocessor <- tryCatch(
        workflows::extract_recipe(fold_model),
        error = function(e) NULL
      )
    }

    # Coerce target for DALEX
    y_target <- if (task == "classification") {
      if (is.factor(y_data) || is.character(y_data) || is.logical(y_data)) {
        y_chr <- as.character(y_data)
        class_levels <- if (!is.null(label_levels)) label_levels else unique(y_chr)
        n_classes <- length(unique(class_levels))

        if (n_classes > 2) {
          # Multiclass: keep labels to align with probability column names
          factor(y_chr, levels = class_levels)
        } else {
          # Binary: map to 0/1 using positive class when available
          pos_class <- NULL
          if (!is.null(positive_class)) {
            pos_class <- positive_class
          } else if (!is.null(label_levels) && length(label_levels) == 2) {
            pos_class <- if (!is.null(event_class) && event_class == "first") {
              label_levels[1]
            } else {
              label_levels[2]
            }
          }
          if (!is.null(pos_class)) {
            as.numeric(tolower(trimws(y_chr)) == tolower(trimws(pos_class)))
          } else {
            as.numeric(as.factor(y_chr)) - 1
          }
        }
      } else {
        y_num <- as.numeric(y_data)
        unique_vals <- unique(stats::na.omit(y_num))
        if (length(unique_vals) > 2) {
          as.character(y_num)
        } else {
          y_num
        }
      }
    } else {
      as.numeric(y_data)
    }

    # Build predict function that handles preprocessing internally
    # This allows DALEX to permute raw features while still getting correct predictions
    # Capture fold_preprocessor in closure for use during prediction
    local_preprocessor <- fold_preprocessor

    predict_function <- function(m, newdata) {
      # For workflows: predict handles baking internally, so pass data directly
      # For non-workflows with preprocessor: bake first, then predict
      processed_data <- newdata

      if (!inherits(m, "workflow") && !is.null(local_preprocessor)) {
        # Bake the permuted data before prediction
        processed_data <- tryCatch(
          recipes::bake(local_preprocessor, new_data = newdata),
          error = function(e) {
            warning(sprintf(
              paste0(
                "Could not apply the fitted recipe to the explanation data: %s\n",
                "Falling back to the unprocessed data; the explanation below ",
                "describes the model on raw features and may not be meaningful."
              ),
              conditionMessage(e)
            ), call. = FALSE)
            newdata
          }
        )
        # Remove label if present in baked data
        if (label %in% names(processed_data)) {
          processed_data[[label]] <- NULL
        }
      }

      if (task == "classification") {
        p <- predict(m, new_data = processed_data, type = "prob")
        colnames(p) <- sub("^\\.pred_", "", colnames(p))
        if (ncol(p) > 2) {
          return(as.matrix(p))
        }
        # Binary: return probability of positive class
        return(as.numeric(p[[ncol(p)]]))
      } else {
        p <- predict(m, new_data = processed_data, type = "numeric")
        return(as.numeric(p$.pred))
      }
    }

    # Create DALEX explainer
    model_info <- if (task == "classification") {
      list(type = "classification")
    } else {
      list(type = "regression")
    }

    explainer <- tryCatch(
      DALEX::explain(
        model = model_fit,
        data = x_data,
        y = y_target,
        label = paste0("fold_", i),
        predict_function = predict_function,
        model_info = model_info,
        verbose = FALSE
      ),
      error = function(e) {
        warning(sprintf("Could not create explainer for fold %d: %s", i, e$message))
        NULL
      }
    )

    if (is.null(explainer)) next

    # Compute variable importance
    fold_seed <- seed + i
    vi_result <- tryCatch({
      set.seed(fold_seed)
      DALEX::model_parts(
        explainer,
        B = vi_iterations,
        type = "raw",
        loss_function = loss_function
      )
    }, error = function(e) {
      warning(sprintf("Variable importance failed for fold %d: %s", i, e$message))
      NULL
    })

    if (!is.null(vi_result)) {
      fold_vi_results[[i]] <- vi_result
    }
  }

  # Filter out NULL results
  valid_folds <- which(!sapply(fold_vi_results, is.null))
  n_valid <- length(valid_folds)

  if (n_valid < 2) {
    stop("Insufficient valid fold results for stability analysis. Need at least 2 folds.")
  }

  message(sprintf("Successfully computed importance for %d of %d folds.", n_valid, n_folds))

  # Aggregate results across folds
  # Extract dropout_loss for each feature from each fold
  extract_vi_agg <- function(vi, fold_id) {
    if (!inherits(vi, "data.frame") ||
        !all(c("variable", "dropout_loss") %in% names(vi))) {
      return(NULL)
    }
    vi_sub <- vi[!vi$variable %in% c("_baseline_", "_full_model_"), , drop = FALSE]
    if (nrow(vi_sub) == 0) {
      return(NULL)
    }
    vi_sub <- vi_sub[
      !is.na(vi_sub$variable) &
        !is.na(vi_sub$dropout_loss) &
        is.finite(vi_sub$dropout_loss),
      ,
      drop = FALSE
    ]
    if (nrow(vi_sub) == 0) {
      return(NULL)
    }
    vi_agg <- stats::aggregate(
      dropout_loss ~ variable,
      data = vi_sub,
      FUN = mean
    )
    if (nrow(vi_agg) == 0) {
      return(NULL)
    }
    vi_agg$fold <- fold_id
    vi_agg
  }

  all_importance <- lapply(valid_folds, function(i) {
    extract_vi_agg(fold_vi_results[[i]], i)
  })
  all_importance <- Filter(Negate(is.null), all_importance)

  if (length(all_importance) < 2) {
    stop(
      "Insufficient variable-level importance results for stability analysis. ",
      "Ensure the model supports variable importance and DALEX returns per-feature values."
    )
  }

  importance_df <- do.call(rbind, all_importance)

  # Compute summary statistics per feature
  importance_summary <- stats::aggregate(
    dropout_loss ~ variable,
    data = importance_df,
    FUN = function(x) {
      c(
        mean = mean(x, na.rm = TRUE),
        sd = stats::sd(x, na.rm = TRUE),
        min = min(x, na.rm = TRUE),
        max = max(x, na.rm = TRUE)
      )
    }
  )

  # Flatten the matrix output from aggregate
  importance_summary <- data.frame(
    variable = importance_summary$variable,
    mean_importance = importance_summary$dropout_loss[, "mean"],
    sd_importance = importance_summary$dropout_loss[, "sd"],
    min_importance = importance_summary$dropout_loss[, "min"],
    max_importance = importance_summary$dropout_loss[, "max"]
  )

  # Add standard error and confidence intervals
  importance_summary$se_importance <- importance_summary$sd_importance / sqrt(n_valid)
  z_val <- stats::qnorm(1 - (1 - conf_level) / 2)
  importance_summary$lower_ci <- importance_summary$mean_importance - z_val * importance_summary$se_importance
  importance_summary$upper_ci <- importance_summary$mean_importance + z_val * importance_summary$se_importance

  # Sort by mean importance (descending)
  importance_summary <- importance_summary[order(-importance_summary$mean_importance), ]
  rownames(importance_summary) <- NULL

  # Compute rank stability
  # Get ranks for each fold
  rank_data <- lapply(valid_folds, function(i) {
    vi_agg <- extract_vi_agg(fold_vi_results[[i]], i)
    if (is.null(vi_agg)) {
      return(NULL)
    }
    vi_agg$rank <- rank(-vi_agg$dropout_loss, ties.method = "average")
    vi_agg[, c("variable", "rank", "fold")]
  })
  rank_data <- Filter(Negate(is.null), rank_data)

  if (length(rank_data) < 2) {
    stop(
      "Insufficient variable-level importance results for rank stability. ",
      "Ensure the model supports variable importance and DALEX returns per-feature values."
    )
  }

  rank_df <- do.call(rbind, rank_data)

  # Compute rank summary
  rank_summary <- stats::aggregate(
    rank ~ variable,
    data = rank_df,
    FUN = function(x) {
      c(
        mean_rank = mean(x, na.rm = TRUE),
        sd_rank = stats::sd(x, na.rm = TRUE),
        min_rank = min(x, na.rm = TRUE),
        max_rank = max(x, na.rm = TRUE)
      )
    }
  )

  rank_stability <- data.frame(
    variable = rank_summary$variable,
    mean_rank = rank_summary$rank[, "mean_rank"],
    sd_rank = rank_summary$rank[, "sd_rank"],
    min_rank = rank_summary$rank[, "min_rank"],
    max_rank = rank_summary$rank[, "max_rank"]
  )

  rank_stability <- rank_stability[order(rank_stability$mean_rank), ]
  rownames(rank_stability) <- NULL

  # Create result object
  result <- list(
    importance_summary = importance_summary,
    fold_importance = fold_vi_results[valid_folds],
    rank_stability = rank_stability,
    n_folds = n_valid,
    model_name = model_name,
    conf_level = conf_level
  )
  class(result) <- "fastml_stability"

  # Plot if requested
  if (plot) {
    p <- plot.fastml_stability(result)
    print(p)
  }

  invisible(result)
}

#' Plot method for fastml_stability objects
#'
#' @param x A fastml_stability object from explain_stability().
#' @param top_n Integer. Number of top features to display. Default is 15.
#' @param ... Additional arguments (ignored).
#'
#' @return A ggplot object.
#' @export
plot.fastml_stability <- function(x, top_n = 15, ...) {
  if (!inherits(x, "fastml_stability")) {
    stop("Input must be a 'fastml_stability' object.")
  }

  if (!requireNamespace("ggplot2", quietly = TRUE)) {
    stop("Package 'ggplot2' is required for plotting.")
  }

  # Select top features
  plot_data <- utils::head(x$importance_summary, top_n)

  # Reorder factor levels for plotting

  plot_data$variable <- factor(
    plot_data$variable,
    levels = rev(plot_data$variable)
  )

  # Create plot
  p <- ggplot2::ggplot(plot_data, ggplot2::aes(x = .data$mean_importance, y = .data$variable)) +
    ggplot2::geom_point(size = 3, color = "steelblue") +
    ggplot2::geom_errorbarh(
      ggplot2::aes(xmin = .data$lower_ci, xmax = .data$upper_ci),
      height = 0.3,
      color = "steelblue",
      alpha = 0.7
    ) +
    ggplot2::labs(
      title = sprintf("Feature Importance Stability (%d folds)", x$n_folds),
      subtitle = sprintf("Model: %s | %d%% CI shown", x$model_name, round(x$conf_level * 100)),
      x = "Mean Dropout Loss (higher = more important)",
      y = "Feature"
    ) +
    ggplot2::theme_minimal() +
    ggplot2::theme(
      plot.title = ggplot2::element_text(face = "bold"),
      axis.text.y = ggplot2::element_text(size = 9)
    )

  p
}

#' Print method for fastml_stability objects
#'
#' @param x A fastml_stability object from explain_stability().
#' @param top_n Integer. Number of top features to display. Default is 10.
#' @param ... Additional arguments (ignored).
#'
#' @return The input object invisibly.
#' @export
print.fastml_stability <- function(x, top_n = 10, ...) {
  if (!inherits(x, "fastml_stability")) {
    stop("Input must be a 'fastml_stability' object.")
  }

  cat("\n=== Feature Importance Stability Analysis ===\n")
  cat(sprintf("Model: %s\n", x$model_name))
  cat(sprintf("Number of folds: %d\n", x$n_folds))
  cat(sprintf("Confidence level: %d%%\n\n", round(x$conf_level * 100)))

  cat("Top features by mean importance:\n")
  top_features <- utils::head(x$importance_summary, top_n)
  top_features$mean_importance <- round(top_features$mean_importance, 4)
  top_features$sd_importance <- round(top_features$sd_importance, 4)
  top_features$lower_ci <- round(top_features$lower_ci, 4)
  top_features$upper_ci <- round(top_features$upper_ci, 4)
  print(top_features[, c("variable", "mean_importance", "sd_importance", "lower_ci", "upper_ci")])

  cat("\n\nRank stability (lower SD = more stable ranking):\n")
  top_ranks <- utils::head(x$rank_stability, top_n)
  top_ranks$mean_rank <- round(top_ranks$mean_rank, 2)
  top_ranks$sd_rank <- round(top_ranks$sd_rank, 2)
  print(top_ranks[, c("variable", "mean_rank", "sd_rank", "min_rank", "max_rank")])

  invisible(x)
}

fastml_init_audit_env <- function(enabled = FALSE) {
  env <- new.env(parent = emptyenv())
  env$enabled <- isTRUE(enabled)
  env$log <- list()
  env$unsafe <- FALSE
  env$warnings_emitted <- character()
  env
}

fastml_register_audit <- function(env, message, severity = c("info", "warning"), context = NULL) {
  if (is.null(env) || !is.environment(env)) {
    return(invisible(NULL))
  }
  severity <- match.arg(severity)
  entry <- list(message = message, severity = severity, context = context)
  env$log <- c(env$log, list(entry))
  if (identical(severity, "warning")) {
    env$unsafe <- TRUE
  }
  invisible(entry)
}

fastml_audit_emit_warning <- function(env, message) {
  if (is.null(env) || !is.environment(env) || !isTRUE(env$enabled)) {
    return(invisible(NULL))
  }
  if (message %in% env$warnings_emitted) {
    return(invisible(NULL))
  }
  env$warnings_emitted <- c(env$warnings_emitted, message)
  warning(message, call. = FALSE)
  invisible(NULL)
}

fastml_env_inherits_global <- function(env) {
  if (is.null(env) || !is.environment(env)) {
    return(FALSE)
  }
  current <- env
  while (!identical(current, emptyenv())) {
    if (identical(current, .GlobalEnv)) {
      return(TRUE)
    }
    parent <- tryCatch(parent.env(current), error = function(...) NULL)
    if (is.null(parent) || identical(parent, current)) {
      break
    }
    current <- parent
  }
  FALSE
}

fastml_global_env_helper_names <- function() {
  c("globalenv", "parent.frame", "sys.frame", "sys.frames", "parent.env")
}

fastml_language_invokes_global_env <- function(expr) {
  if (!is.language(expr)) {
    return(FALSE)
  }

  if (is.symbol(expr)) {
    sym <- as.character(expr)
    return(identical(sym, ".GlobalEnv") || sym %in% fastml_global_env_helper_names())
  }

  if (!is.call(expr)) {
    return(FALSE)
  }

  fn <- expr[[1]]
  helper_names <- fastml_global_env_helper_names()

  if (is.symbol(fn)) {
    fn_name <- as.character(fn)
    if (fn_name %in% helper_names) {
      return(TRUE)
    }
  } else if (is.call(fn)) {
    fn_head <- as.character(fn[[1]])
    if (fn_head %in% c("::", ":::")) {
      ns_symbol <- fn[[3]]
      if (is.symbol(ns_symbol) && as.character(ns_symbol) %in% helper_names) {
        return(TRUE)
      }
    }
  }

  args <- as.list(expr)[-1]
  for (arg in args) {
    if (fastml_language_invokes_global_env(arg)) {
      return(TRUE)
    }
  }

  FALSE
}

fastml_recipe_step_contains_external_reference <- function(step, depth = 0, max_depth = 6) {
  if (depth > max_depth) {
    return(FALSE)
  }

  if (is.environment(step)) {
    return(fastml_env_inherits_global(step))
  }

  if (is.function(step)) {
    if (fastml_env_inherits_global(environment(step))) {
      return(TRUE)
    }
    return(fastml_recipe_step_contains_external_reference(body(step), depth + 1, max_depth))
  }

  # Checked before the generic list branch below: data frames and tibbles are
  # themselves lists, so recursing into their columns would report FALSE and the
  # embedded-data pattern this guard targets would never be flagged. Only nested
  # components count as embedded data; the step object handed in at depth 0 is not
  # itself expected to be a data frame.
  if (depth > 0 && (is.data.frame(step) || inherits(step, "tbl_df"))) {
    return(TRUE)
  }

  if (is.list(step)) {
    for (element in step) {
      if (fastml_recipe_step_contains_external_reference(element, depth + 1, max_depth)) {
        return(TRUE)
      }
    }
    return(FALSE)
  }

  if (inherits(step, "quosure")) {
    expr <- tryCatch(rlang::quo_get_expr(step), error = function(...) NULL)
    if (!is.null(expr)) {
      return(fastml_recipe_step_contains_external_reference(expr, depth + 1, max_depth))
    }
    return(FALSE)
  }

  if (is.symbol(step)) {
    sym <- as.character(step)
    return(identical(sym, ".GlobalEnv") || sym %in% fastml_global_env_helper_names())
  }

  if (is.character(step)) {
    return(any(step %in% c(".GlobalEnv", fastml_global_env_helper_names())))
  }

  if (is.language(step)) {
    if (fastml_language_invokes_global_env(step)) {
      return(TRUE)
    }
    components <- as.list(step)
    for (component in components) {
      if (fastml_recipe_step_contains_external_reference(component, depth + 1, max_depth)) {
        return(TRUE)
      }
    }
    return(FALSE)
  }

  FALSE
}

fastml_detect_leaky_recipe_steps <- function(recipe) {
  flagged <- character()
  if (is.null(recipe$steps) || length(recipe$steps) == 0) {
    return(flagged)
  }
  safe_classes <- c(
    "step_dummy", "step_novel", "step_unknown", "step_zv",
    "step_normalize", "step_center", "step_scale",
    "step_impute_median", "step_impute_knn", "step_impute_bag", "step_naomit",
    "step_upsample", "step_downsample"
  )
  for (idx in seq_along(recipe$steps)) {
    step <- recipe$steps[[idx]]
    label <- if (!is.null(step$id)) step$id else paste0("step_", idx)
    if (inherits(step, safe_classes)) {
      next
    }
    if (fastml_recipe_step_contains_external_reference(step)) {
      flagged <- c(flagged, label)
    }
  }
  unique(flagged)
}

# A prepped recipe carries no `trained` flag of its own, so testing
# `recipe$trained` never fires. Trained state is read from the step flags via
# recipes::fully_trained(), with the training-set record `tr_info` as a fallback
# for a prepped recipe that has no steps.
fastml_recipe_is_trained <- function(recipe) {
  if (isTRUE(recipe$trained)) {
    return(TRUE)
  }
  if (!is.null(recipe$tr_info)) {
    return(TRUE)
  }
  isTRUE(tryCatch(recipes::fully_trained(recipe), error = function(...) FALSE))
}

fastml_validate_user_recipe <- function(recipe, audit_env = NULL) {
  if (!inherits(recipe, "recipe")) {
    stop("`recipe` must be a recipes::recipe object.")
  }
  if (fastml_recipe_is_trained(recipe)) {
    stop("Pretrained recipes are not allowed; provide an untrained recipe.")
  }
  leaky_steps <- fastml_detect_leaky_recipe_steps(recipe)
  if (length(leaky_steps) > 0) {
    msg <- paste0(
      "The supplied recipe contains steps that depend on external data: ",
      paste(leaky_steps, collapse = ", "),
      ". Remove or rewrite these steps to avoid leakage."
    )
    stop(msg)
  }
  fastml_register_audit(
    audit_env,
    "Recipe validation completed with no detected leakage references.",
    severity = "info",
    context = "recipe"
  )
  invisible(recipe)
}

fastml_audit_io_functions <- function() {
  c(
    "read.csv", "read.csv2", "read.table", "readRDS", "load", "scan",
    "source", "file", "gzfile", "bzfile", "xzfile", "unz", "url",
    "readLines", "write.csv", "write.table", "save", "saveRDS"
  )
}

fastml_sandbox_default_packages <- function() {
  c("stats", "utils", "graphics", "grDevices", "methods", "datasets")
}

fastml_populate_sandbox_defaults <- function(env) {
  for (pkg in fastml_sandbox_default_packages()) {
    if (!base::requireNamespace(pkg, quietly = TRUE)) {
      next
    }
    exports <- base::getNamespaceExports(pkg)
    for (nm in exports) {
      if (identical(nm, ".GlobalEnv")) {
        next
      }
      if (!exists(nm, envir = env, inherits = FALSE)) {
        env[[nm]] <- base::getExportedValue(pkg, nm)
      }
    }
  }
}

fastml_fetch_sandbox_function <- function(name) {
  if (exists(name, envir = baseenv(), inherits = FALSE)) {
    return(get(name, envir = baseenv(), inherits = FALSE))
  }
  for (pkg in fastml_sandbox_default_packages()) {
    if (!base::requireNamespace(pkg, quietly = TRUE)) {
      next
    }
    if (name %in% base::getNamespaceExports(pkg)) {
      return(base::getExportedValue(pkg, name))
    }
  }
  NULL
}

fastml_build_sandbox_env <- function(audit_env, context) {
  env <- new.env(parent = baseenv())

  makeActiveBinding(
    ".GlobalEnv",
    function(value) {
      stop(sprintf("Sandboxed %s cannot access .GlobalEnv.", context), call. = FALSE)
    },
    env
  )

  env$assign <- function(x, value, pos = -1, envir = as.environment(pos), inherits = FALSE, immediate = TRUE) {
    target <- if (!missing(envir)) envir else as.environment(pos)
    if (identical(target, .GlobalEnv)) {
      stop(sprintf("Sandboxed %s cannot assign into .GlobalEnv.", context), call. = FALSE)
    }
    base::assign(x, value, pos = pos, envir = envir, inherits = inherits, immediate = immediate)
  }
  environment(env$assign) <- baseenv()

  env$rm <- function(..., list = character(), pos = -1, envir = as.environment(pos), inherits = FALSE) {
    target <- if (!missing(envir)) envir else as.environment(pos)
    if (identical(target, .GlobalEnv)) {
      stop(sprintf("Sandboxed %s cannot remove objects from .GlobalEnv.", context), call. = FALSE)
    }
    base::rm(..., list = list, pos = pos, envir = envir, inherits = inherits)
  }
  environment(env$rm) <- baseenv()

  env$get <- function(x, pos = -1, envir = as.environment(pos), inherits = FALSE) {
    target <- if (!missing(envir)) envir else as.environment(pos)
    if (identical(target, .GlobalEnv)) {
      stop(sprintf("Sandboxed %s cannot read from .GlobalEnv.", context), call. = FALSE)
    }
    base::get(x, pos = pos, envir = envir, inherits = inherits)
  }
  environment(env$get) <- baseenv()

  fastml_populate_sandbox_defaults(env)

  if (!is.null(audit_env) && isTRUE(audit_env$enabled)) {
    for (fn_name in fastml_audit_io_functions()) {
      base_fun <- fastml_fetch_sandbox_function(fn_name)
      if (is.null(base_fun)) {
        next
      }
      env[[fn_name]] <- (function(fun, name) {
        function(...) {
          msg <- sprintf("fastml audit: %s attempted to call %s(), which may access external resources.", context, name)
          fastml_register_audit(audit_env, msg, severity = "warning", context = context)
          fastml_audit_emit_warning(audit_env, msg)
          fun(...)
        }
      })(base_fun, fn_name)
      environment(env[[fn_name]]) <- baseenv()
    }
  }

  env
}

fastml_prepare_sandboxed_function <- function(fn, audit_env, context) {
  if (!is.function(fn)) {
    stop(sprintf("%s must be a function.", context), call. = FALSE)
  }
  if (is.primitive(fn)) {
    stop(sprintf("%s must not be a primitive function.", context), call. = FALSE)
  }
  sandbox_env <- fastml_build_sandbox_env(audit_env, context)
  fn_copy <- fn
  environment(fn_copy) <- sandbox_env
  fn_copy
}

fastml_audit_function_signature <- function(fn, context, audit_env) {
  if (is.null(audit_env) || !isTRUE(audit_env$enabled)) {
    return(invisible(NULL))
  }
  fn_env <- tryCatch(environment(fn), error = function(...) NULL)
  if (fastml_env_inherits_global(fn_env)) {
    msg <- sprintf("fastml audit: %s retains a reference to .GlobalEnv.", context)
    fastml_register_audit(audit_env, msg, severity = "warning", context = context)
    fastml_audit_emit_warning(audit_env, msg)
  }
  if (!is.primitive(fn)) {
    body_symbols <- tryCatch(all.names(body(fn), functions = TRUE), error = function(...) character())
    if (".GlobalEnv" %in% body_symbols) {
      msg <- sprintf("fastml audit: %s references .GlobalEnv, which is disallowed.", context)
      fastml_register_audit(audit_env, msg, severity = "warning", context = context)
      fastml_audit_emit_warning(audit_env, msg)
    }
    io_hits <- intersect(body_symbols, fastml_audit_io_functions())
    if (length(io_hits) > 0) {
      msg <- sprintf(
        "fastml audit: %s references I/O helper(s): %s. External access voids leakage guarantees.",
        context,
        paste(unique(io_hits), collapse = ", ")
      )
      fastml_register_audit(audit_env, msg, severity = "warning", context = context)
      fastml_audit_emit_warning(audit_env, msg)
    }
  }
  invisible(NULL)
}

fastml_run_user_hook <- function(fn, data, context, audit_env, extra_args = list()) {
  fn_prepared <- fastml_prepare_sandboxed_function(fn, audit_env, context)
  fastml_audit_function_signature(fn, context, audit_env)

  data_copy <- if (is.data.frame(data)) {
    data[ , , drop = FALSE]
  } else {
    data
  }

  args <- c(list(data_copy), extra_args)

  pre_globals <- ls(.GlobalEnv, all.names = TRUE)
  result <- tryCatch({
    do.call(fn_prepared, args)
  }, error = function(e) {
    stop(sprintf("Custom preprocessing step failed while executing %s: %s", context, e$message), call. = FALSE)
  })
  post_globals <- ls(.GlobalEnv, all.names = TRUE)
  added <- setdiff(post_globals, pre_globals)
  removed <- setdiff(pre_globals, post_globals)
  if (length(added) > 0) {
    rm(list = added, envir = .GlobalEnv)
  }
  if (length(added) > 0 || length(removed) > 0) {
    msg <- sprintf("Sandboxed %s attempted to modify the global environment.", context)
    fastml_register_audit(audit_env, msg, severity = "warning", context = context)
    stop(msg, call. = FALSE)
  }
  result
}

fastml_validate_custom_hook <- function(hook, context) {
  if (is.list(hook) || is.environment(hook)) {
    fit <- hook$fit
    transform <- hook$transform
  } else {
    fit <- NULL
    transform <- NULL
  }
  if (!is.function(fit) || !is.function(transform)) {
    stop(sprintf("%s must supply both 'fit' and 'transform' functions.", context), call. = FALSE)
  }
  list(fit = fit, transform = transform)
}

fastml_validate_transformed_data <- function(data, context) {
  if (!is.data.frame(data)) {
    stop(sprintf("%s must return a data.frame.", context), call. = FALSE)
  }
  data
}

fastml_process_custom_fit_result <- function(result, context) {
  if (!is.list(result) || is.null(result$state)) {
    stop(sprintf("%s fit must return a list with a 'state' element.", context), call. = FALSE)
  }
  list(
    state = result$state,
    transformed = result$transformed
  )
}

library(testthat)

# Behavioural tests for the recipe-scanning guard in R/security_guards.R.
# The guard skips a safe-list of standard steps and, for any remaining step,
# flags two patterns: a reference to the global environment / a parent frame,
# and an embedded external data object.

make_df <- function(n = 50) {
  data.frame(
    x = seq_len(n) / n,
    y = rep(c(0, 1), length.out = n),
    g = factor(rep(letters[1:3], length.out = n))
  )
}

# Attaches a synthetic non-standard step carrying `payload` to `rec`, so the
# scanner sees a step outside its safe-list.
add_custom_step <- function(rec, payload, id) {
  step <- rec$steps[[1]]
  class(step) <- c(paste0("step_", id), "step")
  step$payload <- payload
  step$id <- id
  rec$steps <- c(rec$steps, list(step))
  rec
}

safe_recipe <- function(df = make_df()) {
  recipes::recipe(y ~ ., data = df) |>
    recipes::step_impute_median(recipes::all_numeric_predictors()) |>
    recipes::step_normalize(recipes::all_numeric_predictors()) |>
    recipes::step_dummy(recipes::all_nominal_predictors()) |>
    recipes::step_novel(recipes::all_nominal_predictors()) |>
    recipes::step_zv(recipes::all_predictors())
}

test_that("safe-listed standard steps are not flagged", {
  skip_if_not_installed("recipes")

  expect_length(fastml:::fastml_detect_leaky_recipe_steps(safe_recipe()), 0)
  expect_silent(fastml:::fastml_validate_user_recipe(safe_recipe()))
})

test_that("standard steps outside the safe-list are not false positives", {
  skip_if_not_installed("recipes")

  # These steps are not on the safe-list, so they are scanned in full; none of
  # them embeds external data or reaches into the global environment, and the
  # scanner must not reject them.
  rec <- recipes::recipe(y ~ ., data = make_df()) |>
    recipes::step_pca(recipes::all_numeric_predictors(), num_comp = 2) |>
    recipes::step_log(x, offset = 1) |>
    recipes::step_YeoJohnson(recipes::all_numeric_predictors()) |>
    recipes::step_corr(recipes::all_numeric_predictors())

  expect_length(fastml:::fastml_detect_leaky_recipe_steps(rec), 0)
})

test_that("a step embedding an external data.frame is flagged", {
  skip_if_not_installed("recipes")

  # Regression test: data.frames are themselves lists, so a scanner that tests
  # is.list() before the data.frame check recurses into the columns and reports
  # no finding. This pattern must be flagged, not silently accepted.
  lookup <- data.frame(key = 1:3, value = c(0.1, 0.2, 0.3))
  rec <- add_custom_step(safe_recipe(), lookup, "custom_join")

  expect_identical(
    fastml:::fastml_detect_leaky_recipe_steps(rec),
    "custom_join"
  )
})

test_that("a step embedding an external tibble is flagged", {
  skip_if_not_installed("recipes")
  skip_if_not_installed("tibble")

  rec <- add_custom_step(safe_recipe(), tibble::tibble(key = 1:3), "custom_tbl")

  expect_identical(
    fastml:::fastml_detect_leaky_recipe_steps(rec),
    "custom_tbl"
  )
})

test_that("steps reaching into the global environment are flagged", {
  skip_if_not_installed("recipes")

  global_env <- add_custom_step(
    safe_recipe(), function(z) get("external_object", envir = globalenv()),
    "custom_globalenv"
  )
  parent_frame <- add_custom_step(
    safe_recipe(), quote(parent.frame()), "custom_parent_frame"
  )
  dot_global <- add_custom_step(safe_recipe(), quote(.GlobalEnv), "custom_dot_global")

  expect_identical(
    fastml:::fastml_detect_leaky_recipe_steps(global_env), "custom_globalenv"
  )
  expect_identical(
    fastml:::fastml_detect_leaky_recipe_steps(parent_frame), "custom_parent_frame"
  )
  expect_identical(
    fastml:::fastml_detect_leaky_recipe_steps(dot_global), "custom_dot_global"
  )
})

test_that("validation aborts on a flagged recipe rather than dropping the step", {
  skip_if_not_installed("recipes")

  rec <- add_custom_step(safe_recipe(), data.frame(key = 1:3), "custom_join")

  expect_error(
    fastml:::fastml_validate_user_recipe(rec),
    "depend on external data"
  )
  # The offending step is named so the user can act on the message.
  expect_error(fastml:::fastml_validate_user_recipe(rec), "custom_join")
})

test_that("pretrained recipes are rejected", {
  skip_if_not_installed("recipes")

  df <- make_df()
  trained <- recipes::prep(safe_recipe(df), training = df)

  expect_error(
    fastml:::fastml_validate_user_recipe(trained),
    "Pretrained recipes are not allowed"
  )
})

test_that("non-recipe input is rejected", {
  expect_error(
    fastml:::fastml_validate_user_recipe(make_df()),
    "must be a recipes::recipe object",
    fixed = TRUE
  )
})

#' Define Quadratic Discriminant Analysis Model Specification
#'
#' @return List containing the model specification (`model_spec`).
#' @importFrom parsnip set_mode set_engine
#' @noRd
define_discrim_quad_spec <- function(task, engine = "sparsediscrim") {
  if (task != "classification") {
    stop("discrim_quad is only applicable for classification tasks.")
  }
  if (!requireNamespace("discrim", quietly = TRUE)) {
    stop("Package 'discrim' is required for discriminant analysis models. Install it with: install.packages('discrim')")
  }
  if (identical(engine, "sparsediscrim") &&
      !requireNamespace("sparsediscrim", quietly = TRUE)) {
    stop(
      "Engine 'sparsediscrim' for discrim_quad requires the 'sparsediscrim' package. Please install it or choose a different engine."
    )
  }

  model_spec <- parsnip::discrim_quad() %>%
    set_mode("classification")

  if (identical(engine, "sparsediscrim")) {
    model_spec <- model_spec %>%
      set_engine(engine, regularization_method = "diagonal")
  } else {
    model_spec <- model_spec %>%
      set_engine(engine)
  }
  list(model_spec = model_spec)
}

#' Define Linear Discriminant Analysis Model Specification
#'
#' @inheritParams define_logistic_reg_spec
#' @return List containing the model specification (`model_spec`).
#' @importFrom parsnip set_mode set_engine
#' @noRd
define_discrim_linear_spec <- function(task, engine = "MASS") {
  if (task != "classification") {
    stop("discrim_linear is only applicable for classification tasks.")
  }
  if (!requireNamespace("discrim", quietly = TRUE)) {
    stop("Package 'discrim' is required for discriminant analysis models. Install it with: install.packages('discrim')")
  }
  model_spec <- parsnip::discrim_linear() %>%
    set_mode("classification") %>%
    set_engine(engine)
  list(model_spec = model_spec)
}

#' Define Neural Network Model Specification (nnet)
#'
#' @inheritParams define_rand_forest_spec
#' @return List containing the model specification (`model_spec`).
#' @importFrom parsnip mlp set_mode set_engine
#' @importFrom tune tune
#' @noRd
define_mlp_spec <- function(task, tuning = FALSE, engine = "nnet") {

  defaults <- get_default_params("mlp", task, num_predictors = NULL, engine)

  if (tuning) {
    if (engine == "nnet") {
      model_spec <- mlp(
        hidden_units = tune(),
        penalty = tune(),
        epochs = tune()
      ) %>%
        set_mode(task) %>%
        set_engine(engine)
    } else if (engine == "brulee") {
      model_spec <- mlp(
        hidden_units = tune(),
        penalty = tune(),
        epochs = tune(),
        activation = tune(),
        dropout = tune(),
        learn_rate = tune()
        # mixture = tune()
      ) %>%
        set_mode(task) %>%
        set_engine(engine)
    } else if (engine == "h2o") {
      model_spec <- mlp(
        hidden_units = tune(),
        penalty = tune(),
        dropout = tune(),
        epochs = tune(),
        activation = tune(),
        learn_rate = tune()
      ) %>%
        set_mode(task) %>%
        set_engine(engine)
    } else if (engine == "keras") {
      model_spec <- mlp(
        hidden_units = tune(),
        penalty = tune(),
        dropout = tune(),
        epochs = tune(),
        activation = tune()
      ) %>%
        set_mode(task) %>%
        set_engine(engine)
    }
  } else {
    if (engine == "nnet") {
      model_spec <- mlp(
        hidden_units = defaults$hidden_units,
        penalty = defaults$penalty,
        epochs = defaults$epochs
      ) %>%
        set_mode(task) %>%
        set_engine(engine)
    } else if (engine == "brulee") {
      model_spec <- mlp(
        hidden_units = defaults$hidden_units,
        penalty = defaults$penalty,
        epochs = defaults$epochs,
        activation = defaults$activation,
        dropout = defaults$dropout,
        learn_rate = defaults$learn_rate
        # mixture = defaults$mixture
      ) %>%
        set_mode(task) %>%
        set_engine(engine)
    } else if (engine == "h2o") {
      model_spec <- mlp(
        hidden_units = defaults$hidden_units,
        penalty = defaults$penalty,
        dropout = defaults$dropout,
        epochs = defaults$epochs,
        activation = defaults$activation,
        learn_rate = defaults$learn_rate
      ) %>%
        set_mode(task) %>%
        set_engine(engine)
    } else if (engine == "keras") {
      model_spec <- mlp(
        hidden_units = defaults$hidden_units,
        penalty = defaults$penalty,
        dropout = defaults$dropout,
        epochs = defaults$epochs,
        activation = defaults$activation
      ) %>%
        set_mode(task) %>%
        set_engine(engine)
    }
  }

  list(model_spec = model_spec)
}

#' Define Naive Bayes Model Specification
#'
#' @inheritParams define_logistic_reg_spec
#' @return List containing the model specification (`model_spec`).
#' @importFrom parsnip naive_Bayes set_mode set_engine
#' @importFrom tune tune
#' @noRd
define_naive_Bayes_spec <- function(task, tuning = FALSE, engine = "klaR") {
  if (task != "classification") {
    stop("Naive Bayes is only applicable for classification tasks.")
  }
  if (!requireNamespace("discrim", quietly = TRUE)) {
    stop("Package 'discrim' is required for Naive Bayes models. Install it with: install.packages('discrim')")
  }

  # Retrieve default parameters for naive_Bayes based on engine.
  defaults <- get_default_params("naive_Bayes", task, engine)

  # Build the base model spec with tuning or default values.
  if (tuning) {
    if (engine %in% c("klaR", "naivebayes")) {
      model_spec <- naive_Bayes(
        smoothness = tune(),
        Laplace = tune()
      )
    }else if (engine == "h2o") {

      model_spec <- naive_Bayes(
        Laplace = tune()
      )
    }
  } else {
    if (engine %in% c("klaR", "naivebayes")) {
      model_spec <- naive_Bayes(
        smoothness = defaults$smoothness,
        Laplace = defaults$Laplace
      )
    }else if (engine == "h2o") {
      model_spec <- naive_Bayes(
        Laplace = defaults$Laplace
      )
    }
  }

  # Set the model mode.
  model_spec <- model_spec %>% set_mode("classification")

  # Engine-specific parameter settings.
  # For the klaR and naivebayes engines, usekernel is set to TRUE by default.
  if (engine %in% c("klaR", "naivebayes")) {
    model_spec <- model_spec %>% set_engine(engine, usekernel = TRUE)
  } else if (engine == "h2o") {
    model_spec <- model_spec %>% set_engine(engine)
  } else {
    stop("Unsupported engine specified for naive_Bayes.")
  }

  list(model_spec = model_spec)
}

#' Define K-Nearest Neighbors Model Specification
#'
#' @inheritParams define_rand_forest_spec
#' @return List containing the model specification (`model_spec`).
#' @importFrom parsnip nearest_neighbor set_mode set_engine
#' @importFrom tune tune
#' @noRd
define_nearest_neighbor_spec <- function(task, tuning = FALSE, engine = "kknn") {
  defaults <- get_default_params("nearest_neighbor")

  if (tuning) {
    model_spec <- nearest_neighbor(
      neighbors = tune(),
      weight_func = tune(),
      dist_power = tune()
    ) %>%
      set_mode(task) %>%
      set_engine(engine)
  } else {
    model_spec <- nearest_neighbor(
      neighbors = defaults$neighbors,
      weight_func = defaults$weight_func,
      dist_power = defaults$dist_power
    ) %>%
      set_mode(task) %>%
      set_engine(engine)
  }
  list(model_spec = model_spec)
}

#' Define SVM Radial Model Specification
#'
#' @inheritParams define_svm_linear_spec
#' @return List containing the model specification (`model_spec`).
#' @importFrom parsnip svm_rbf set_mode set_engine
#' @importFrom tune tune
#' @noRd
define_svm_rbf_spec <- function(task, tuning = FALSE, engine = "kernlab") {
  defaults <- get_default_params("svm_rbf")

  if (tuning) {
    model_spec <- svm_rbf(
      cost = tune(),
      rbf_sigma = tune()
    ) %>%
      set_mode(task) %>%
      set_engine(engine)
  } else {
    model_spec <- svm_rbf(
      cost = defaults$cost,
      rbf_sigma = defaults$rbf_sigma
    ) %>%
      set_mode(task) %>%
      set_engine(engine)
  }
  list(model_spec = model_spec)
}

#' Define SVM Linear Model Specification
#'
#' @inheritParams define_rand_forest_spec
#' @return List containing the model specification (`model_spec`).
#' @importFrom parsnip svm_linear set_mode set_engine
#' @importFrom tune tune
#' @noRd
define_svm_linear_spec <- function(task, tuning = FALSE, engine = "kernlab") {
  defaults <- get_default_params("svm_linear")

  if (tuning) {
    if (task == "regression") {
      # For regression, both cost and margin are tunable
      model_spec <- svm_linear(
        cost   = tune(),
        margin = tune()
      ) %>%
        set_mode(task) %>%
        set_engine(engine)
    } else {  # classification
      # For classification, only cost is tuned; margin is not applicable
      model_spec <- svm_linear(
        cost = tune()
      ) %>%
        set_mode(task) %>%
        set_engine(engine)
    }
  } else {
    if (task == "regression") {
      model_spec <- svm_linear(
        cost   = defaults$cost,
        margin = defaults$margin  # defaults$margin will be 0.1 for kernlab or missing for LiblineaR
      ) %>%
        set_mode(task) %>%
        set_engine(engine)
    } else {  # classification
      model_spec <- svm_linear(
        cost = defaults$cost
      ) %>%
        set_mode(task) %>%
        set_engine(engine)
    }
  }
  list(model_spec = model_spec)
}


#' Define Decision Tree Model Specification
#'
#' @inheritParams define_rand_forest_spec
#' @param task A character string indicating the mode (e.g., "classification" or "regression").
#' @param tuning Logical; if \code{TRUE}, the tuning parameters are set to \code{tune()}.
#' @param engine A character string specifying the computational engine. Supported engines include
#'   "rpart", "C5.0", and "partykit".
#'
#' @return List containing the model specification (`model_spec`).
#'
#' @details
#' The available parameters and their defaults vary by engine:
#'
#' - **rpart:**
#'   \itemize{
#'     \item \code{tree_depth}: Tree Depth (default: 30L)
#'     \item \code{min_n}: Minimal Node Size (default: 2L)
#'     \item \code{cost_complexity}: Cost-Complexity Parameter (default: 0.01)
#'   }
#'
#' - **C5.0:**
#'   \itemize{
#'     \item \code{min_n}: Minimal Node Size (default: 2L)
#'   }
#'
#' - **partykit:**
#'   \itemize{
#'     \item \code{tree_depth}: Tree Depth (default: 30L; adjust as needed)
#'     \item \code{min_n}: Minimal Node Size (default: 20L)
#'   }
#'
#' @importFrom parsnip decision_tree set_mode set_engine
#' @importFrom tune tune
#' @noRd

# Internal helpers for handling resampling plans and metadata

fastml_is_resample_plan <- function(x) {
  inherits(x, "fastml_resample_plan")
}

fastml_new_resample_plan <- function(splits, method, params = list(), extras = list()) {
  if (is.null(splits)) {
    stop("'splits' must be provided when creating a resampling plan.")
  }
  if (!inherits(splits, "rset")) {
    stop("'splits' must be an 'rset' object when creating a resampling plan.")
  }
  metadata <- list(
    method = method,
    params = params
  )
  plan <- c(list(splits = splits, metadata = metadata), extras)
  class(plan) <- c("fastml_resample_plan", class(plan))
  plan
}

fastml_resample_splits <- function(plan) {
  if (fastml_is_resample_plan(plan)) {
    return(plan$splits)
  }
  plan
}

fastml_resample_metadata <- function(plan) {
  if (fastml_is_resample_plan(plan)) {
    return(plan$metadata)
  }
  NULL
}

fastml_resample_method <- function(plan) {
  meta <- fastml_resample_metadata(plan)
  if (is.null(meta)) {
    return(NULL)
  }
  meta$method
}

fastml_resample_params <- function(plan) {
  meta <- fastml_resample_metadata(plan)
  if (is.null(meta)) {
    return(list())
  }
  meta$params %||% list()
}

fastml_resample_update_splits <- function(plan, splits) {
  if (!fastml_is_resample_plan(plan)) {
    stop("Cannot update splits on a non-resampling plan object.")
  }
  if (!inherits(splits, "rset")) {
    stop("'splits' must be an 'rset' object when updating a resampling plan.")
  }
  plan$splits <- splits
  plan
}

fastml_resample_validate <- function(plan, allow_null = FALSE) {
  if (is.null(plan)) {
    if (allow_null) {
      return(NULL)
    }
    stop("Resampling plan cannot be NULL in this context.")
  }
  splits <- fastml_resample_splits(plan)
  if (is.null(splits) || !inherits(splits, "rset")) {
    stop("Resampling plan must contain an 'rset' object under the 'splits' element.")
  }
  plan
}

fastml_describe_resampling <- function(plan) {
  if (is.null(plan)) {
    return("none")
  }
  method <- fastml_resample_method(plan)
  params <- fastml_resample_params(plan)
  if (is.null(method)) {
    method <- "custom"
  }
  if (length(params) == 0) {
    return(method)
  }
  if (is.null(names(params))) {
    names(params) <- paste0("param", seq_along(params))
  }
  formatted <- vapply(
    names(params),
    function(nm) {
      value <- params[[nm]]
      if (is.null(value)) {
        value <- "NULL"
      } else if (length(value) > 1) {
        value <- paste(value, collapse = ", ")
      }
      paste0(nm, "=", value)
    },
    character(1)
  )
  paste0(method, " [", paste(formatted, collapse = ", "), "]")
}

`%||%` <- function(x, y) {
  if (is.null(x)) y else x
}

#' Plot ICE curves for a fastml model
#'
#' Generates Individual Conditional Expectation (ICE) plots for selected features
#' using the `pdp` package (ggplot2 engine), and returns both the underlying data
#' and the plot object.
#'
#' @param object A `fastml` object.
#' @param features Character vector of feature names to plot.
#' @param data Character string specifying which data to use: \code{"train"} (default) or \code{"test"}.
#' @param target_class For classification, which class probability to plot. If NULL (default),
#'   uses the positive class determined by the model settings. For multiclass problems,
#'   this shows the probability of the specified class vs all others.
#' @param ... Additional arguments passed to `pdp::partial`.
#'
#' @return A list with two elements: `data` (the ICE data frame) and `plot` (the ggplot object).
#' @importFrom ggplot2 autoplot
#' @importFrom recipes bake
#' @export
#' @examples
#' \dontrun{
#' data(iris)
#' iris <- iris[iris$Species != "setosa", ]
#' iris$Species <- factor(iris$Species)
#' model <- fastml(data = iris, label = "Species")
#' plot_ice(model, features = "Sepal.Length")
#' }
plot_ice <- function(object, features, data = c("train", "test"), target_class = NULL, ...) {
  if (!inherits(object, "fastml")) {
    stop("The input must be a 'fastml' object.")
  }
  if (missing(features)) {
    stop("'features' must be specified for ICE plots.")
  }
  if (!requireNamespace("pdp", quietly = TRUE)) {
    stop("The 'pdp' package is required for ICE plots.")
  }

  data <- match.arg(data)

  # Use the same helper as other explainer methods for consistency

  prep <- fastml_prepare_explainer_inputs(object, data = data)
  if (is.null(prep$train_data) || !(prep$label %in% names(prep$train_data))) {
    stop(sprintf("%s data not available for ICE plots.", tools::toTitleCase(data)))
  }

  # Use raw data (consistent with ALE, interaction_strength, surrogate_tree)
  x <- prep$x_raw
  y <- prep$y_raw

  # Validate that user-provided features exist in the raw data
  available_features <- colnames(x)
  invalid_features <- setdiff(features, available_features)
  if (length(invalid_features) > 0) {
    stop(
      sprintf(
        "Feature name(s) not found in data: %s\nAvailable features: %s%s\nNote: feature names must match exactly (case-sensitive).",
        paste(invalid_features, collapse = ", "),
        paste(utils::head(available_features, 10), collapse = ", "),
        if (length(available_features) > 10) " ..." else ""
      ),
      call. = FALSE
    )
  }

  parsnip_fit <- prep$fits[[1]]
  if (is.null(parsnip_fit)) {
    stop("Unable to extract parsnip model for ICE plots.")
  }

  # Determine target class for classification tasks
  # For multiclass, warn user that ICE shows probability for one class vs all others
  positive_class <- NULL
  is_multiclass <- FALSE
  if (prep$task == "classification" && (is.factor(y) || is.character(y))) {
    y_factor <- if (is.factor(y)) y else factor(y)
    y_levels <- levels(y_factor)
    is_multiclass <- length(y_levels) > 2

    if (!is.null(target_class)) {
      # User specified a target class
      if (!(target_class %in% y_levels)) {
        stop(
          sprintf(
            "target_class '%s' not found in class levels.\nAvailable classes: %s",
            target_class,
            paste(y_levels, collapse = ", ")
          ),
          call. = FALSE
        )
      }
      positive_class <- target_class
    } else {
      # Use default positive class
      positive_class <- resolve_positive_class(prep, y_levels)
    }

    if (is_multiclass) {
      message(
        sprintf(
          "Note: Multiclass ICE plot showing probability for class '%s' vs all others.\n",
          positive_class
        ),
        "Use target_class parameter to specify a different class."
      )
    }
  }

  # Custom predict function that preprocesses raw data before prediction
  # This ensures consistency with other explainer methods (ALE, interaction, surrogate)
  # Note: pdp::partial may pass either a parsnip model_fit or the raw underlying model,

  # so we need to handle both cases
  pred_fun <- function(object, newdata) {
    # Preprocess raw data using the stored recipe
    newdata_processed <- tryCatch(
      {
        baked <- recipes::bake(prep$preprocessor, new_data = newdata)
        if (!is.null(prep$label) && prep$label %in% names(baked)) {
          baked[[prep$label]] <- NULL
        }
        baked
      },
      error = function(e) {
        warning(sprintf(
          paste0(
            "Could not apply the fitted recipe to the explanation data: %s\n",
            "Falling back to the unprocessed data; the explanation below ",
            "describes the model on raw features and may not be meaningful."
          ),
          conditionMessage(e)
        ), call. = FALSE)
        newdata
      }
    )

    # Check if object is a parsnip model_fit or raw model
    is_parsnip <- inherits(object, "model_fit")

    if (is_parsnip) {
      # Try probabilities first (classification) using parsnip's predict
      prob <- tryCatch(
        predict(object, new_data = newdata_processed, type = "prob"),
        error = function(e) NULL
      )
      if (!is.null(prob)) {
        # Prefer column that matches the positive class; otherwise fall back to first numeric column
        prob_col <- paste0(".pred_", positive_class)
        if (!is.null(positive_class) && prob_col %in% names(prob)) {
          return(as.numeric(prob[[prob_col]]))
        }
        num_cols <- names(prob)[vapply(prob, is.numeric, logical(1))]
        if (length(num_cols) > 0) {
          return(as.numeric(prob[[num_cols[1]]]))
        }
      }

      # Fallback to standard predictions (regression)
      preds <- predict(object, new_data = newdata_processed)
      return(as.numeric(preds[[1]]))
    }

    # Handle raw model objects (e.g., rpart, glm, lm, ranger)
    # Try classification (probabilities) first
    prob <- tryCatch({
      p <- predict(object, newdata = newdata_processed, type = "prob")
      if (is.matrix(p) || is.data.frame(p)) {
        # Return probability for positive class or second column
        if (!is.null(positive_class) && positive_class %in% colnames(p)) {
          as.numeric(p[, positive_class])
        } else if (ncol(p) >= 2) {
          as.numeric(p[, 2])
        } else {
          as.numeric(p[, 1])
        }
      } else {
        as.numeric(p)
      }
    }, error = function(e) NULL)

    if (!is.null(prob)) {
      return(prob)
    }

    # Try response predictions
    preds <- tryCatch({
      p <- predict(object, newdata = newdata_processed)
      if (is.matrix(p) || is.data.frame(p)) {
        as.numeric(p[, 1])
      } else {
        as.numeric(p)
      }
    }, error = function(e) {
      # Last resort: try without type argument
      p <- predict(object, newdata_processed)
      as.numeric(if (is.matrix(p) || is.data.frame(p)) p[, 1] else p)
    })

    preds
  }

  # Compute ICE data using raw data and custom predict function
  # This makes ICE consistent with ALE, interaction_strength, and surrogate_tree
  pd <- pdp::partial(
    parsnip_fit,
    pred.var = features,
    ice = TRUE,
    train = x,
    pred.fun = pred_fun,
    plot = FALSE,
    ...
  )
  p <- ggplot2::autoplot(pd)

  # Convert deprecated `size` aesthetics on line geoms to `linewidth` to avoid ggplot2 warnings
  p$layers <- lapply(p$layers, function(layer) {
    if (inherits(layer$geom, c("GeomLine", "GeomPath", "GeomSegment", "GeomStep"))) {
      mapping <- layer$mapping
      if (!is.null(mapping$size)) {
        mapping$linewidth <- mapping$size
        mapping$size <- NULL
        layer$mapping <- mapping
      }
      if (!is.null(layer$aes_params$size)) {
        layer$aes_params$linewidth <- layer$aes_params$size
        layer$aes_params$size <- NULL
      }
    }
    layer
  })
  if (!is.null(p$labels$size)) {
    p$labels$linewidth <- p$labels$size
    p$labels$size <- NULL
  }
  print(p)
  invisible(list(data = pd, plot = p))
}

library(testthat)
library(dplyr)

skip_if_not_installed("pdp")
skip_if_not_installed("rpart")

# Prevent Rplots.pdf from being created by graphics calls
grDevices::pdf(file = NULL)
withr::defer(grDevices::dev.off())

# small dataset for tests
set.seed(123)
data(iris)
iris_binary <- iris[iris$Species != "setosa", ]
iris_binary$Species <- factor(iris_binary$Species)

# Helper to train model and skip if training fails
train_model_or_skip <- function(data, label, algorithms = "decision_tree") {
  model <- tryCatch(
    suppressWarnings(fastml(
      data = data,
      label = label,
      algorithms = algorithms,
      resampling_method = "none"
    )),
    error = function(e) NULL
  )
  if (is.null(model) || is.null(model$models) || length(model$models) == 0) {
    skip("Model training failed")
  }
  model
}

test_that("plot_ice errors with non-fastml object", {
  expect_error(
    plot_ice(list(a = 1), features = "Sepal.Length"),
    "must be a 'fastml' object"
  )
})

test_that("plot_ice errors when features is missing", {
  skip_on_cran()

  model <- train_model_or_skip(
    data = iris_binary,
    label = "Species",
    algorithms = "decision_tree"
  )

  expect_error(
    plot_ice(model),
    "'features' must be specified"
  )
})

test_that("plot_ice runs on classification model", {
  skip_on_cran()

  # Use decision_tree with rpart engine which has better pdp compatibility
  model <- train_model_or_skip(
    data = iris_binary,
    label = "Species",
    algorithms = "decision_tree"
  )

  result <- suppressWarnings(suppressMessages(plot_ice(model, features = "Sepal.Length")))
  expect_true(is.list(result))
  expect_true("data" %in% names(result))
  expect_true("plot" %in% names(result))
  expect_true(inherits(result$plot, c("gg", "ggplot")))
})

test_that("plot_ice runs on regression model", {
  skip_on_cran()

  set.seed(42)
  data(mtcars)

  # Use decision_tree with rpart engine which has better pdp compatibility
  model <- train_model_or_skip(
    data = mtcars,
    label = "mpg",
    algorithms = "decision_tree"
  )

  result <- suppressWarnings(suppressMessages(plot_ice(model, features = "hp")))
  expect_true(is.list(result))
  expect_true("data" %in% names(result))
  expect_true("plot" %in% names(result))
})

test_that("plot_ice works with different features", {
  skip_on_cran()

  # Use decision_tree with rpart engine which has better pdp compatibility
  model <- train_model_or_skip(
    data = iris_binary,
    label = "Species",
    algorithms = "decision_tree"
  )

  # Note: pdp::partial with ice=TRUE only supports single features,
  # so we test each feature individually
  result1 <- suppressWarnings(suppressMessages(
    plot_ice(model, features = "Sepal.Length")
  ))
  expect_true(is.list(result1))

  result2 <- suppressWarnings(suppressMessages(
    plot_ice(model, features = "Sepal.Width")
  ))
  expect_true(is.list(result2))
})

#' Get Best Model Names
#'
#' Extracts and returns the best engine names from a named list of model workflows.
#'
#' @param models A named list where each element corresponds to an algorithm and contains a list of model workflows.
#'   Each workflow should be compatible with \code{tune::extract_fit_parsnip}.
#'
#' @return A named character vector. The names of the vector correspond to the algorithm names, and the values represent the chosen best engine name for that algorithm.
#'
#' @details For each algorithm, the function extracts the engine names from the model workflows using \code{tune::extract_fit_parsnip}.
#'   It then chooses \code{"randomForest"} if it is available; otherwise, it selects the first non-\code{NA} engine.
#'   If no engine names can be extracted for an algorithm, \code{NA_character_} is returned.
#'
#' @importFrom tune extract_fit_parsnip
#'
#' @export
get_best_model_names <- function(models) {
  bests <- sapply(names(models), function(algo) {
    model_list <- models[[algo]]
    engines <- sapply(model_list, function(mod) {
      fit_obj <- tryCatch(
        extract_fit_parsnip(mod),
        error = function(e) NULL
      )
      if (!is.null(fit_obj)) {
        fit_obj$spec$engine
      } else {
        NA_character_
      }
    })
    if ("randomForest" %in% engines) {
      "randomForest"
    } else {
      non_na_engines <- engines[!is.na(engines)]
      if (length(non_na_engines) > 0) non_na_engines[1] else NA_character_
    }
  })
  bests
}

#' Get Best Workflows
#'
#' Extracts the best workflows from a nested list of model workflows based on the provided best model names.
#'
#' @param models A nested list of model workflows.
#' @param best_model_name A named character vector of chosen best engines.
#'
#' @export
get_best_workflows <- function(models, best_model_name) {
  best_list <- lapply(seq_along(best_model_name), function(i) {
    algo   <- names(best_model_name)[i]
    engine <- best_model_name[i]
    if (!is.null(models[[algo]]) && !is.null(models[[algo]][[engine]])) {
      models[[algo]][[engine]]
    } else {
      warning(paste("No workflow found for", algo, "with engine", engine))
      NULL
    }
  })
  names(best_list) <- paste0(names(best_model_name), " (", best_model_name, ")")
  best_list
}

#' Flatten and Rename Models
#'
#' @param models A named list of model objects or nested named lists of model
#'   objects. Nested entries are flattened and renamed using the pattern
#'   `algorithm (engine)`.
#'
#' @return A flat named list of model objects.
#'
#' @importFrom stats setNames
#' @export
flatten_and_rename_models <- function(models) {
  flattened <- list()
  for (outer_name in names(models)) {
    inner <- models[[outer_name]]
    if (is.list(inner) && !is.null(names(inner))) {
      new_names <- paste0(outer_name, " (", names(inner), ")")
      flattened <- c(flattened, setNames(inner, new_names))
    } else {
      flattened[[outer_name]] <- inner
    }
  }
  flattened
}

#' Get Best Model Indices by Metric and Group
#'
#' @param df A data frame containing model performance summaries.
#' @param metric A single character string naming the column in `df` used to
#'   identify the best model.
#' @param group_cols Character vector of column names used to define model
#'   groups before selecting the best metric value. Defaults to
#'   `c("Model", "Engine")`.
#'
#' @return Integer vector of row indices in `df` corresponding to the best
#'   model group or groups.
#'
#' @importFrom stats ave
#' @export
get_best_model_idx <- function(df, metric, group_cols = c("Model", "Engine")) {
  metric_values <- as.numeric(as.character(df[[metric]]))
  group_values <- interaction(df[, group_cols], drop = TRUE)

  lower_is_better <- metric %in% c("rmse", "mae", "ibs", "logloss", "mse",
                                   "brier_score", "ece") ||
    grepl("^brier_t", metric) ||
    grepl("loss", metric)

  if (lower_is_better) {
    group_val <- ave(metric_values, group_values, FUN = min)
    overall_val <- min(metric_values)
  } else {
    group_val <- ave(metric_values, group_values, FUN = max)
    overall_val <- max(metric_values)
  }

  best_groups <- unique(group_values[group_val == overall_val])
  which(group_values %in% best_groups)
}

fastml_aggregate_resample_metrics <- function(perf, task) {
  if (is.null(perf) || !is.data.frame(perf) || nrow(perf) == 0) {
    return(NULL)
  }

  perf <- tibble::as_tibble(as.data.frame(perf))

  if (!".estimate" %in% names(perf)) {
    return(NULL)
  }

  perf$.estimate <- suppressWarnings(as.numeric(perf$.estimate))
  if (!".estimator" %in% names(perf)) {
    perf$.estimator <- NA_character_
  }

  group_cols <- ".metric"
  if (!identical(task, "survival") && any(!is.na(perf$.estimator))) {
    group_cols <- c(group_cols, ".estimator")
  }

  aggregated <- perf %>%
    dplyr::group_by(dplyr::across(dplyr::all_of(group_cols))) %>%
    dplyr::summarise(
      mean_est = mean(.estimate, na.rm = TRUE),
      n_valid  = sum(is.finite(.estimate)),
      sd_est   = stats::sd(.estimate, na.rm = TRUE),
      .groups  = "drop"
    ) %>%
    dplyr::mutate(
      std_err = ifelse(n_valid > 0, sd_est / sqrt(n_valid), NA_real_)
    ) %>%
    dplyr::rename(
      .estimate = mean_est,
      n = n_valid,
      std_dev = sd_est
    )

  if (!".estimator" %in% names(aggregated)) {
    aggregated$.estimator <- NA_character_
  }

  aggregated
}

fastml_extract_selection_performance <- function(resampling_results,
                                                 nested_results,
                                                 task) {
  selection_perf <- list()
  selection_source <- NULL

  if (!is.null(nested_results) && length(nested_results) > 0) {
    selection_perf <- lapply(nested_results, function(entry) {
      outer_perf <- if (is.list(entry) && !is.null(entry$outer_performance)) {
        entry$outer_performance
      } else if (is.data.frame(entry)) {
        entry
      } else {
        NULL
      }
      fastml_aggregate_resample_metrics(outer_perf, task)
    })
    selection_perf <- selection_perf[!vapply(selection_perf, is.null, logical(1))]
    if (length(selection_perf) > 0) {
      selection_source <- "nested_cv"
    }
  }

  if (is.null(selection_source) &&
      !is.null(resampling_results) &&
      length(resampling_results) > 0) {
    selection_perf <- lapply(resampling_results, function(entry) {
      if (is.list(entry) && !inherits(entry, "data.frame")) {
        folds_tbl <- entry$folds
        if (!is.null(folds_tbl) && is.data.frame(folds_tbl) && nrow(folds_tbl) > 0) {
          return(fastml_aggregate_resample_metrics(folds_tbl, task))
        }
        NULL
      } else if (is.data.frame(entry)) {
        fastml_aggregate_resample_metrics(entry, task)
      } else {
        NULL
      }
    })
    selection_perf <- selection_perf[!vapply(selection_perf, is.null, logical(1))]
    if (length(selection_perf) > 0) {
      selection_source <- "resampling"
    }
  }

  list(performance = selection_perf, source = selection_source)
}

library(testthat)
library(dplyr)

# ========================
# Tests for availableMethods
# ========================

test_that("availableMethods returns valid methods for classification", {
  methods <- availableMethods(type = "classification")

  expect_true(is.character(methods))
  expect_true(length(methods) > 0)
  expect_true("logistic_reg" %in% methods)
  expect_true("rand_forest" %in% methods)
})

test_that("availableMethods returns valid methods for regression", {
  methods <- availableMethods(type = "regression")

  expect_true(is.character(methods))
  expect_true(length(methods) > 0)
  expect_true("linear_reg" %in% methods)
  expect_true("rand_forest" %in% methods)
})

test_that("availableMethods returns valid methods for survival", {
  methods <- availableMethods(type = "survival")

  expect_true(is.character(methods))
  expect_true(length(methods) > 0)
})

# ========================
# Tests for get_default_engine
# ========================

test_that("get_default_engine returns correct engine for rand_forest", {
  engine <- get_default_engine("rand_forest")
  expect_equal(engine, "ranger")
})

test_that("get_default_engine returns correct engine for logistic_reg", {
  engine <- get_default_engine("logistic_reg")
  expect_equal(engine, "glm")
})

test_that("get_default_engine returns correct engine for linear_reg", {
  engine <- get_default_engine("linear_reg")
  expect_equal(engine, "lm")
})

test_that("get_default_engine returns correct engine for xgboost", {
  engine <- get_default_engine("xgboost")
  expect_equal(engine, "xgboost")
})

# ========================
# Tests for get_default_params
# ========================

test_that("get_default_params returns list for rand_forest", {
  params <- get_default_params("rand_forest", "classification")

  expect_true(is.list(params))
})

test_that("get_default_params returns list for linear_reg", {
  params <- get_default_params("linear_reg", "regression")

  expect_true(is.list(params))
})

# ========================
# Tests for get_best_model_idx
# ========================

test_that("get_best_model_idx returns correct index for higher-better metric", {
  df <- data.frame(
    Model = c("A", "B", "C"),
    Engine = c("e1", "e1", "e1"),
    accuracy = c(0.8, 0.9, 0.7)
  )

  idx <- get_best_model_idx(df, "accuracy")
  expect_equal(idx, 2)  # B has highest accuracy
})

test_that("get_best_model_idx returns correct index for lower-better metric", {
  df <- data.frame(
    Model = c("A", "B", "C"),
    Engine = c("e1", "e1", "e1"),
    rmse = c(0.5, 0.3, 0.8)
  )

  idx <- get_best_model_idx(df, "rmse")
  expect_equal(idx, 2)  # B has lowest rmse
})

test_that("get_best_model_idx handles data without NA values", {
  # Note: get_best_model_idx uses max/min without na.rm, so all values must be non-NA
  df <- data.frame(
    Model = c("A", "B", "C"),
    Engine = c("e1", "e1", "e1"),
    accuracy = c(0.8, 0.85, 0.7)
  )

  idx <- get_best_model_idx(df, "accuracy")
  expect_equal(idx, 2)  # B has highest accuracy
})

# ========================
# Tests for get_best_model_names
# ========================

test_that("get_best_model_names returns named character vector", {
  skip_on_cran()

  set.seed(123)
  data(iris)
  iris_binary <- iris[iris$Species != "setosa", ]
  iris_binary$Species <- factor(iris_binary$Species)

  model <- suppressWarnings(fastml(
    data = iris_binary,
    label = "Species",
    algorithms = c("logistic_reg", "rand_forest"),
    resampling_method = "none",
    use_default_tuning = FALSE
  ))

  names <- get_best_model_names(model$models)

  expect_true(is.character(names))
  expect_true(length(names) > 0)
})

# ========================
# Tests for get_best_workflows
# ========================

test_that("get_best_workflows returns list of workflows", {
  skip_on_cran()

  set.seed(123)
  data(iris)
  iris_binary <- iris[iris$Species != "setosa", ]
  iris_binary$Species <- factor(iris_binary$Species)

  model <- suppressWarnings(fastml(
    data = iris_binary,
    label = "Species",
    algorithms = c("logistic_reg", "rand_forest"),
    resampling_method = "none",
    use_default_tuning = FALSE
  ))

  workflows <- suppressWarnings(get_best_workflows(model$models, model$best_model_name))

  expect_true(is.list(workflows))
  expect_true(length(workflows) > 0)
})

# ========================
# Tests for flatten_and_rename_models
# ========================

test_that("flatten_and_rename_models works on fastml models", {
  skip_on_cran()

  set.seed(123)
  data(iris)
  iris_binary <- iris[iris$Species != "setosa", ]
  iris_binary$Species <- factor(iris_binary$Species)

  model <- suppressWarnings(fastml(
    data = iris_binary,
    label = "Species",
    algorithms = c("logistic_reg", "rand_forest"),
    resampling_method = "none",
    use_default_tuning = FALSE
  ))

  flattened <- flatten_and_rename_models(model$models)

  expect_true(is.list(flattened))
  expect_true(length(flattened) > 0)
})

# ========================
# Tests for get_engine_names
# ========================

test_that("get_engine_names extracts engine names from models", {
  skip_on_cran()

  set.seed(123)
  data(iris)
  iris_binary <- iris[iris$Species != "setosa", ]
  iris_binary$Species <- factor(iris_binary$Species)

  model <- suppressWarnings(fastml(
    data = iris_binary,
    label = "Species",
    algorithms = c("logistic_reg", "rand_forest"),
    resampling_method = "none",
    use_default_tuning = FALSE
  ))

  engines <- get_engine_names(model$models)

  expect_true(is.list(engines) || is.character(engines))
  expect_true(length(engines) > 0)
})

# ========================
# Tests for get_tuning_complexity
# ========================

test_that("get_tuning_complexity returns valid config for quick", {
  config <- get_tuning_complexity("quick")

  expect_true(is.list(config))
  expect_true("grid_levels" %in% names(config))
  expect_true("recommended_folds" %in% names(config))
})

test_that("get_tuning_complexity returns valid config for balanced", {
  config <- get_tuning_complexity("balanced")

  expect_true(is.list(config))
  expect_true(config$grid_levels > get_tuning_complexity("quick")$grid_levels)
})

test_that("get_tuning_complexity returns valid config for thorough", {
  config <- get_tuning_complexity("thorough")

  expect_true(is.list(config))
  expect_true(config$grid_levels > get_tuning_complexity("balanced")$grid_levels)
})

test_that("get_tuning_complexity returns valid config for exhaustive", {
  config <- get_tuning_complexity("exhaustive")

  expect_true(is.list(config))
  expect_true(config$grid_levels > get_tuning_complexity("thorough")$grid_levels)
})

# ========================
# Tests for estimate_tuning_time
# ========================

test_that("estimate_tuning_time returns list with time estimates", {
  estimate <- estimate_tuning_time(n_params = 3, n_folds = 5, n_rows = 1000)

  expect_true(is.list(estimate))
  expect_true("estimated_seconds" %in% names(estimate))
  expect_true("estimated_minutes" %in% names(estimate))
  expect_true(is.numeric(estimate$estimated_seconds))
  expect_true(estimate$estimated_seconds > 0)
})

test_that("estimate_tuning_time increases with more params", {
  est1 <- estimate_tuning_time(n_params = 2)
  est2 <- estimate_tuning_time(n_params = 3)

  expect_true(est2$estimated_seconds > est1$estimated_seconds)
})

# ========================
# Tests for print_tuning_presets
# ========================

test_that("print_tuning_presets runs without error", {
  output <- capture.output(print_tuning_presets())

  expect_true(length(output) > 0)
  expect_true(any(grepl("quick|balanced|thorough|exhaustive", output, ignore.case = TRUE)))
})

# ========================
# Tests for validate_defaults_registry
# ========================

test_that("validate_defaults_registry returns list", {
  skip_if_not_installed("parsnip")

  result <- validate_defaults_registry()

  expect_true(is.list(result))
})

# ========================
# Tests for get_default_differences
# ========================

test_that("get_default_differences returns list", {
  result <- get_default_differences(c("rand_forest", "logistic_reg"), task = "classification")

  expect_true(is.list(result))
})

# ========================
# Tests for compare_defaults
# ========================

test_that("compare_defaults identifies matching engines", {
  comparison <- compare_defaults(
    algo = "rand_forest",
    task = "classification",
    fastml_engine = "ranger"
  )

  expect_false(comparison$engine_differs)
})

test_that("compare_defaults identifies differing engines", {
  comparison <- compare_defaults(
    algo = "boost_tree",
    task = "classification",
    fastml_engine = "lightgbm"
  )

  expect_true(comparison$engine_differs)
})

#' @title Defaults Registry for Engine and Parameter Transparency
#' @description Functions to track, compare, and warn about differences between
#'   fastml defaults and parsnip defaults, providing users with full transparency
#'   and control over model configuration.
#' @name defaults_registry
NULL

#' Get Parsnip Default Engine for an Algorithm
#'
#' Returns the default engine that parsnip would use for a given algorithm,
#' allowing comparison with fastml's optimized defaults.
#'
#' @param algo Character string specifying the algorithm name.
#' @param task Character string specifying the task type ("classification",
#'   "regression", or "survival").
#'
#' @return Character string of the parsnip default engine, or NULL if unknown.
#'
#' @details This function documents parsnip's default engine choices as of
#'   tidymodels 1.x. These may change with future parsnip versions.
#'
#' @keywords internal
#' @export
get_parsnip_default_engine <- function(algo, task = NULL) {
  # Parsnip default engines as of tidymodels 1.x

  # Source: parsnip package documentation and show_engines() output
  parsnip_defaults <- list(
    # Classification/Regression models
    logistic_reg = "glm",
    multinom_reg = "glmnet",
    linear_reg = "lm",
    decision_tree = "rpart",
    rand_forest = "ranger",
    boost_tree = "xgboost",
    xgboost = "xgboost",
    lightgbm = "lightgbm",
    svm_linear = "LiblineaR",
    svm_rbf = "kernlab",
    svm_poly = "kernlab",
    nearest_neighbor = "kknn",
    naive_Bayes = "klaR",
    mlp = "keras",
    bag_tree = "rpart",
    C5_rules = "C5.0",
    discrim_linear = "MASS",
    discrim_quad = "MASS",
    pls = "mixOmics",

    # Survival models (via censored package)
    proportional_hazards = "survival",
    survival_reg = "survival",
    cox_ph = "survival",
    bag_mars = "earth"
  )

  # Task-specific overrides for survival
  if (!is.null(task) && task == "survival") {
    survival_overrides <- list(
      rand_forest = "partykit",  # censored default for survival RF
      rand_forest_survival = "partykit"
    )
    if (algo %in% names(survival_overrides)) {
      return(survival_overrides[[algo]])
    }
  }

  if (algo %in% names(parsnip_defaults)) {
    return(parsnip_defaults[[algo]])
  }

  # Return NULL for algorithms not in parsnip
  NULL
}

#' Get Parsnip Default Parameters for an Algorithm
#'
#' Returns the default parameter values that parsnip would use for a given
#' algorithm and engine combination.
#'
#' @param algo Character string specifying the algorithm name.
#' @param engine Character string specifying the engine.
#' @param task Character string specifying the task type.
#'
#' @return A named list of default parameter values, or NULL if unknown.
#'
#' @details These defaults are based on parsnip's internal defaults and the
#'   underlying engine defaults. They may differ from fastml's optimized
#'   defaults which are tuned for better out-of-box performance.
#'
#' @keywords internal
#' @export
get_parsnip_default_params <- function(algo, engine = NULL, task = NULL) {
  # Parsnip typically uses NULL (engine defaults) for most parameters

  # These are documented defaults from parsnip model specifications
  parsnip_defaults <- list(
    rand_forest = list(
      ranger = list(mtry = NULL, trees = 500L, min_n = NULL),
      randomForest = list(mtry = NULL, trees = 500L, min_n = NULL),
      partykit = list(mtry = NULL, trees = 500L, min_n = 20L),
      aorsf = list(mtry = NULL, trees = 500L, min_n = 5L)
    ),
    xgboost = list(
      xgboost = list(
        tree_depth = 6L,
        trees = 15L,
        learn_rate = 0.3,
        mtry = NULL,
        min_n = 1L,
        loss_reduction = 0,
        sample_size = 1,
        stop_iter = NULL
      )
    ),
    lightgbm = list(
      lightgbm = list(
        trees = 100L,
        tree_depth = -1L,  # unlimited
        learn_rate = 0.1,
        min_n = 20L,
        loss_reduction = 0,
        sample_size = 1,
        mtry = NULL
      )
    ),
    logistic_reg = list(
      glm = list(penalty = NULL, mixture = NULL),
      glmnet = list(penalty = 0, mixture = 1)  # pure lasso by default
    ),
    decision_tree = list(
      rpart = list(tree_depth = 30L, min_n = 2L, cost_complexity = 0.01)
    ),
    svm_linear = list(
      LiblineaR = list(cost = 1, margin = NULL),
      kernlab = list(cost = 1, margin = NULL)
    ),
    nearest_neighbor = list(
      kknn = list(neighbors = 5L, weight_func = "optimal", dist_power = 2)
    ),
    mlp = list(
      nnet = list(hidden_units = 5L, penalty = 0, epochs = 100L)
    )
  )

  if (algo %in% names(parsnip_defaults)) {
    algo_defaults <- parsnip_defaults[[algo]]
    if (!is.null(engine) && engine %in% names(algo_defaults)) {
      return(algo_defaults[[engine]])
    }
    # Return first engine's defaults if engine not specified
    if (length(algo_defaults) > 0) {
      return(algo_defaults[[1]])
    }
  }

  NULL
}

#' Compare fastml and parsnip defaults
#'
#' Compares the default engine and parameter choices between fastml and parsnip
#' for a given algorithm.
#'
#' @param algo Character string specifying the algorithm name.
#' @param task Character string specifying the task type.
#' @param fastml_engine Character string of the fastml default engine.
#' @param fastml_params List of fastml default parameters.
#'
#' @return A list with components:
#'   \describe{
#'     \item{engine_differs}{Logical indicating if engines differ.}
#'     \item{fastml_engine}{The fastml default engine.}
#'     \item{parsnip_engine}{The parsnip default engine.}
#'     \item{param_differences}{Named list of parameters that differ.}
#'   }
#'
#' @keywords internal
#' @export
compare_defaults <- function(algo, task, fastml_engine, fastml_params = NULL) {
  parsnip_engine <- get_parsnip_default_engine(algo, task)
  parsnip_params <- get_parsnip_default_params(algo, fastml_engine, task)

  result <- list(
    engine_differs = FALSE,
    fastml_engine = fastml_engine,
    parsnip_engine = parsnip_engine,
    param_differences = list()
  )

  # Check if engines differ

if (!is.null(parsnip_engine) && !identical(fastml_engine, parsnip_engine)) {
    result$engine_differs <- TRUE
  }

  # Check parameter differences
  if (!is.null(fastml_params) && !is.null(parsnip_params)) {
    common_params <- intersect(names(fastml_params), names(parsnip_params))
    for (param in common_params) {
      fastml_val <- fastml_params[[param]]
      parsnip_val <- parsnip_params[[param]]

      # Skip if both are NULL
      if (is.null(fastml_val) && is.null(parsnip_val)) next

      # Check if values differ (handle NULL cases)
      if (is.null(fastml_val) != is.null(parsnip_val) ||
          (!is.null(fastml_val) && !isTRUE(all.equal(fastml_val, parsnip_val)))) {
        result$param_differences[[param]] <- list(
          fastml = fastml_val,
          parsnip = parsnip_val
        )
      }
    }
  }

  result
}

#' Format Default Override Warning Message
#'
#' Creates a human-readable warning message about default overrides.
#'
#' @param algo Algorithm name.
#' @param comparison Result from compare_defaults().
#' @param show_params Logical; whether to include parameter differences.
#'
#' @return Character string with the warning message.
#'
#' @keywords internal
format_default_override_warning <- function(algo, comparison, show_params = TRUE) {
  messages <- character(0)

  if (comparison$engine_differs) {
    messages <- c(messages, sprintf(
      "fastml uses '%s' engine (parsnip default: '%s')",
      comparison$fastml_engine,
      comparison$parsnip_engine
    ))
  }

  if (show_params && length(comparison$param_differences) > 0) {
    param_msgs <- vapply(names(comparison$param_differences), function(p) {
      diff <- comparison$param_differences[[p]]
      fastml_str <- if (is.null(diff$fastml)) "NULL" else as.character(diff$fastml)
      parsnip_str <- if (is.null(diff$parsnip)) "NULL/engine-default" else as.character(diff$parsnip)
      sprintf("  %s: %s (parsnip: %s)", p, fastml_str, parsnip_str)
    }, character(1))
    messages <- c(messages, param_msgs)
  }

  if (length(messages) == 0) {
    return(NULL)
  }

  paste0("[", algo, "] ", paste(messages, collapse = "; "))
}

#' Environment for Tracking Warned Defaults
#'
#' Internal environment to track which default warnings have been shown
#' in the current session to avoid duplicate warnings.
#'
#' @keywords internal
.fastml_warned_defaults <- new.env(parent = emptyenv())

#' Reset Default Override Warnings
#'
#' Resets the tracking of which default override warnings have been shown.
#' Useful for testing or when starting a new analysis session.
#'
#' @export
reset_default_warnings <- function() {
  rm(list = ls(.fastml_warned_defaults), envir = .fastml_warned_defaults)
  invisible(NULL)
}

#' Warn About Default Overrides
#'
#' Issues a warning if fastml defaults differ from parsnip defaults and
#' the warning hasn't been shown yet in this session.
#'
#' @param algo Algorithm name.
#' @param task Task type.
#' @param fastml_engine fastml's default engine for this algorithm.
#' @param fastml_params fastml's default parameters (optional).
#' @param verbose If TRUE, always show the message (as a message, not warning).
#' @param warn_once If TRUE (default), only warn once per algorithm per session.
#'
#' @return Invisibly returns the comparison result.
#'
#' @export
warn_default_override <- function(algo, task, fastml_engine, fastml_params = NULL,
                                  verbose = FALSE, warn_once = TRUE) {
  comparison <- compare_defaults(algo, task, fastml_engine, fastml_params)

  # Check if there's anything to warn about
  has_differences <- comparison$engine_differs || length(comparison$param_differences) > 0
  if (!has_differences) {
    return(invisible(comparison))
  }

  # Check if we've already warned for this algorithm
  warn_key <- paste0(algo, "_", task)
  if (warn_once && exists(warn_key, envir = .fastml_warned_defaults)) {
    return(invisible(comparison))
  }

  # Format the warning message
  msg <- format_default_override_warning(algo, comparison, show_params = TRUE)
  if (is.null(msg)) {
    return(invisible(comparison))
  }

  full_msg <- paste0(
    msg, "\n",
    "Set algorithm_engines or use_parsnip_defaults=TRUE for explicit control."
  )

  if (verbose) {
    message(full_msg)
  } else {
    warning(full_msg, call. = FALSE)
  }

  # Mark as warned
  if (warn_once) {
    assign(warn_key, TRUE, envir = .fastml_warned_defaults)
  }

  invisible(comparison)
}

#' Get All Default Differences Summary
#'
#' Returns a summary of all differences between fastml and parsnip defaults
#' for the specified algorithms.
#'
#' @param algorithms Character vector of algorithm names.
#' @param task Task type ("classification", "regression", or "survival").
#'
#' @return A data frame summarizing the differences.
#'
#' @export
get_default_differences <- function(algorithms, task = "classification") {
  results <- lapply(algorithms, function(algo) {
    fastml_engine <- tryCatch(
      get_default_engine(algo, task),
      error = function(e) NA_character_
    )
    if (is.na(fastml_engine)) {
      return(data.frame(
        algorithm = algo,
        fastml_engine = NA_character_,
        parsnip_engine = NA_character_,
        engine_differs = NA,
        param_differences = NA_character_,
        stringsAsFactors = FALSE
      ))
    }

    parsnip_engine <- get_parsnip_default_engine(algo, task)
    comparison <- compare_defaults(algo, task, fastml_engine, NULL)

    data.frame(
      algorithm = algo,
      fastml_engine = fastml_engine,
      parsnip_engine = if (is.null(parsnip_engine)) NA_character_ else parsnip_engine,
      engine_differs = comparison$engine_differs,
      param_differences = if (length(comparison$param_differences) == 0) {
        NA_character_
      } else {
        paste(names(comparison$param_differences), collapse = ", ")
      },
      stringsAsFactors = FALSE
    )
  })

  do.call(rbind, results)
}

#' Print Default Differences Table
#'
#' Prints a formatted table showing differences between fastml and parsnip
#' defaults for the specified task type.
#'
#' @param task Task type ("classification", "regression", or "survival").
#' @param algorithms Optional character vector of algorithms to check.
#'   If NULL, checks all available algorithms for the task.
#'
#' @return Invisibly returns the differences data frame.
#'
#' @export
print_default_differences <- function(task = "classification", algorithms = NULL) {
  if (is.null(algorithms)) {
    algorithms <- availableMethods(type = task)
  }

  diffs <- get_default_differences(algorithms, task)

  # Only show rows where there are differences
  diffs_only <- diffs[diffs$engine_differs == TRUE | !is.na(diffs$param_differences), ]

  if (nrow(diffs_only) == 0) {
    message("No differences between fastml and parsnip defaults for ", task, " task.")
    return(invisible(diffs))
  }

  message("\nfastml vs parsnip default differences (", task, "):\n")
  message(sprintf("%-20s %-15s %-15s", "Algorithm", "fastml", "parsnip"))
  message(paste(rep("-", 50), collapse = ""))

  for (i in seq_len(nrow(diffs_only))) {
    row <- diffs_only[i, ]
    parsnip_eng <- if (is.na(row$parsnip_engine)) "(not in parsnip)" else row$parsnip_engine
    marker <- if (row$engine_differs) " *" else ""
    message(sprintf("%-20s %-15s %-15s%s",
                    row$algorithm, row$fastml_engine, parsnip_eng, marker))
  }

  message("\n* = engine differs from parsnip default")
  message("\nTip: Use algorithm_engines parameter to explicitly set engines,")
  message("or use_parsnip_defaults=TRUE to use parsnip's default engines.\n")

  invisible(diffs)
}

#' Validate Defaults Registry Against Parsnip
#'
#' Compares the hardcoded parsnip default engines in fastml's registry against
#' the actual defaults reported by \code{parsnip::show_engines()}. Returns a
#' list of any mismatches found, which may indicate that parsnip has updated
#' its defaults since fastml's registry was last updated.
#'
#' @return A list of mismatches. Each element is a list with components:
#'   \describe{
#'     \item{algorithm}{The algorithm name.}
#'     \item{fastml_default}{The default engine recorded in fastml's registry.}
#'     \item{parsnip_default}{The actual default engine from parsnip.}
#'   }
#'   Returns an empty list if no mismatches are found.
#'
#' @details This function queries parsnip for model specifications and compares
#'   against the hardcoded \code{parsnip_defaults} list in
#'   \code{get_parsnip_default_engine()}. Mismatches may occur when:
#'   \itemize{
#'     \item Parsnip updates its default engine for a model type

#'     \item New engines are added to parsnip that become the new default
#'     \item fastml's registry has not been updated after a parsnip release
#'   }
#'
#'   This validation is intended for package maintenance and testing purposes.
#'
#' @export
#' @examples
#' \dontrun{
#' mismatches <- validate_defaults_registry()
#' if (length(mismatches) > 0) {
#'   message("Found ", length(mismatches), " mismatch(es) with parsnip defaults")
#' }
#' }
validate_defaults_registry <- function() {
  mismatches <- list()

  # Algorithms to check - those that have parsnip model functions
  algorithms_to_check <- c(
    "logistic_reg", "multinom_reg", "linear_reg", "decision_tree",
    "rand_forest", "boost_tree", "svm_linear", "svm_rbf", "svm_poly",
    "nearest_neighbor", "naive_Bayes", "mlp", "bag_tree",
    "discrim_linear", "discrim_quad"
  )

  for (algo in algorithms_to_check) {
    # Get fastml's recorded parsnip default
    fastml_default <- get_parsnip_default_engine(algo, task = NULL)

    if (is.null(fastml_default)) {
      next
    }

    # Try to get parsnip's actual default by querying show_engines
    parsnip_default <- tryCatch({
      # Check if the model function exists in parsnip
      if (!exists(algo, envir = asNamespace("parsnip"), inherits = FALSE)) {
        NULL
      } else {
        # Get available engines
        engines_df <- tryCatch(
          parsnip::show_engines(algo),
          error = function(e) NULL
        )

        if (is.null(engines_df) || nrow(engines_df) == 0) {
          NULL
        } else {
          # The first engine listed is typically the default
          # parsnip doesn't explicitly mark defaults, but the first registered
          # engine is conventionally the default
          engines_df$engine[1]
        }
      }
    }, error = function(e) {
      NULL
    })

    # Compare if we got a result from parsnip
    if (!is.null(parsnip_default) && !identical(fastml_default, parsnip_default)) {
      mismatches[[length(mismatches) + 1]] <- list(
        algorithm = algo,
        fastml_default = fastml_default,
        parsnip_default = parsnip_default
      )
    }
  }

  mismatches
}

library(testthat)

test_that("validate_defaults_registry returns a list", {
  skip_if_not_installed("parsnip")

  result <- validate_defaults_registry()

  expect_type(result, "list")
})

test_that("validate_defaults_registry mismatches have correct structure", {

  result <- validate_defaults_registry()

  # Result should always be a list (even if empty)
  expect_type(result, "list")

  # If there are mismatches, verify their structure
  if (length(result) > 0) {
    for (mismatch in result) {
      expect_true("algorithm" %in% names(mismatch))
      expect_true("fastml_default" %in% names(mismatch))
      expect_true("parsnip_default" %in% names(mismatch))
      expect_type(mismatch$algorithm, "character")
      expect_type(mismatch$fastml_default, "character")
      expect_type(mismatch$parsnip_default, "character")
    }
  }
})

test_that("get_parsnip_default_engine returns expected values for known algorithms", {
  # Test a few known defaults that are unlikely to change
  expect_equal(get_parsnip_default_engine("logistic_reg"), "glm")
  expect_equal(get_parsnip_default_engine("linear_reg"), "lm")
  expect_equal(get_parsnip_default_engine("decision_tree"), "rpart")
  expect_equal(get_parsnip_default_engine("rand_forest"), "ranger")

  # Unknown algorithms should return NULL
  expect_null(get_parsnip_default_engine("nonexistent_algorithm"))
})

test_that("compare_defaults detects engine differences", {
  # Test with a known difference (fastml uses xgboost for boost_tree,
  # and parsnip also uses xgboost)
  comparison <- compare_defaults(
    algo = "boost_tree",
    task = "classification",
    fastml_engine = "lightgbm"  # hypothetical different engine
  )

  expect_true(comparison$engine_differs)
  expect_equal(comparison$fastml_engine, "lightgbm")
  expect_equal(comparison$parsnip_engine, "xgboost")
})

test_that("compare_defaults returns no difference when engines match", {
  comparison <- compare_defaults(
    algo = "rand_forest",
    task = "classification",
    fastml_engine = "ranger"
  )

  expect_false(comparison$engine_differs)
  expect_equal(comparison$fastml_engine, "ranger")
  expect_equal(comparison$parsnip_engine, "ranger")
})

utils::globalVariables(c(
  ".", "..density..",
  "Var1", "Var2", "Correlation",
  "variable", "Missing",
  "Model", "Engine",
  "truth", "estimate",
  "id",  "valid_model",
  ".estimate", ".metric", ".estimator",
  "mean_est", "n_valid", "sd_est", "std_dev",
  "survival_metric_convention"
))

.onAttach <- function(libname, pkgname) {
  # Validate parsnip defaults registry on package attach (interactive only)
  if (interactive()) {
    mismatches <- tryCatch(
      validate_defaults_registry(),
      error = function(e) list()
    )

    if (length(mismatches) > 0) {
      packageStartupMessage(
        sprintf(
          "Note: fastml's parsnip defaults registry may be out of date. Found %d mismatch(es).\n",
          length(mismatches)
        ),
        "Run validate_defaults_registry() for details."
      )
    }
  }
}

#' Native survival prediction helpers
#'
#' The functions in this file provide shared preprocessing logic used across
#' the package whenever predictions are generated from models that operate on
#' fastml's "native" survival objects.  They are intentionally kept internal
#' (not exported) so they can be re-used without polluting the public API.
#'
#' @noRd
NULL

#' Apply a stored recipe, or abort
#'
#' Applies a fitted recipe to new data on a prediction path. Failure is fatal
#' rather than silently ignored: falling back to unpreprocessed data would make
#' the model predict on features it was not trained on, producing
#' plausible-looking but meaningless output. A `NULL` recipe means there is
#' nothing to apply, and `new_data` is returned unchanged.
#'
#' @param recipe A prepped recipe, or `NULL`.
#' @param new_data Data to bake.
#' @param context Short description of the calling path, used in the message.
#'
#' @return The baked data, or `new_data` when `recipe` is `NULL`.
#' @keywords internal
#' @noRd
fastml_bake_or_abort <- function(recipe, new_data, context) {
  if (is.null(recipe)) {
    return(new_data)
  }
  tryCatch(
    recipes::bake(recipe, new_data = new_data),
    error = function(e) {
      stop(sprintf(
        paste0(
          "Could not apply the stored preprocessing recipe in %s: %s\n",
          "The new data must contain the same columns, types, and factor levels ",
          "as the training data. fastml does not fall back to unpreprocessed ",
          "data, because doing so would return predictions from a model applied ",
          "to features it was not trained on."
        ),
        context,
        conditionMessage(e)
      ), call. = FALSE)
    }
  )
}

fastml_prepare_native_survival_predictors <- function(model, baked_newdata, original_newdata) {
  if (is.null(baked_newdata)) {
    baked_newdata <- data.frame()
  }

  pred_predictors <- as.data.frame(baked_newdata)
  keep_cols <- names(pred_predictors)

  drop_cols <- c(model$response, model$time_col, model$status_col, model$start_col)
  drop_cols <- unique(drop_cols[!is.na(drop_cols)])
  drop_cols <- intersect(drop_cols, keep_cols)
  if (length(drop_cols) > 0) {
    keep_cols <- setdiff(keep_cols, drop_cols)
  }

  if (!is.null(model$strata_dummy_cols) && length(model$strata_dummy_cols) > 0) {
    keep_cols <- setdiff(keep_cols, model$strata_dummy_cols)
  }
  if (!is.null(model$strata_base_cols) && length(model$strata_base_cols) > 0) {
    keep_cols <- setdiff(keep_cols, model$strata_base_cols)
  }

  if (length(keep_cols) == 0) {
    pred_predictors <- pred_predictors[, 0, drop = FALSE]
  } else {
    pred_predictors <- pred_predictors[, keep_cols, drop = FALSE]
  }

  extra_cols <- unique(c(model$start_col, model$time_col, model$status_col))
  extra_cols <- extra_cols[!is.na(extra_cols)]
  if (length(extra_cols) > 0) {
    for (ec in extra_cols) {
      if (!(ec %in% names(pred_predictors)) && ec %in% names(original_newdata)) {
        pred_predictors[[ec]] <- original_newdata[[ec]]
      }
    }
  }

  if (!is.null(model$strata_cols) && length(model$strata_cols) > 0) {
    strata_levels <- NULL
    fit_obj <- tryCatch(model$fit, error = function(e) NULL)
    if (!is.null(fit_obj)) {
      strata_levels <- tryCatch(fit_obj$xlevels, error = function(e) NULL)
    }
    for (sc in model$strata_cols) {
      if (!(sc %in% names(pred_predictors)) && sc %in% names(original_newdata)) {
        pred_predictors[[sc]] <- original_newdata[[sc]]
      }
      if (sc %in% names(pred_predictors)) {
        pred_predictors[[sc]] <- as.factor(pred_predictors[[sc]])
        if (!is.null(strata_levels) && sc %in% names(strata_levels)) {
          pred_predictors[[sc]] <- factor(pred_predictors[[sc]], levels = strata_levels[[sc]])
        }
      }
    }
  }

  pred_predictors
}

#' Internal helpers for survival-specific preprocessing
#'
#' These utilities standardize survival status indicators so that downstream
#' metrics always receive the conventional coding (0 = censored, 1 = event).
#' The functions are intentionally unexported and are used across multiple
#' internal modules.

#' Normalize survival status coding to 0/1 representation
#'
#' This helper attempts to coerce a status vector into a numeric format where
#' 0 represents censoring and 1 represents the event indicator. It accepts a
#' variety of common encodings such as 1/2, logical values, factors, or
#' character labels. When the supplied values deviate from the canonical
#' coding, the function records that a recode was performed so callers can
#' communicate this to the user (once).
#'
#' @param status_vec A vector containing survival status information. May be
#'   numeric, logical, factor, or character.
#' @param reference_length Optional integer specifying the desired length of the
#'   returned vector. When `status_vec` is `NULL`, this value controls the
#'   length of the output (defaulting to 0 when not supplied).
#'
#' @return A list with two elements: `status`, the recoded numeric vector, and
#'   `recoded`, a logical flag indicating whether a non-standard encoding was
#'   detected.
fastml_normalize_survival_status <- function(status_vec, reference_length = NULL) {
  if (is.null(reference_length)) {
    reference_length <- length(status_vec)
  }

  if (is.null(status_vec)) {
    stop(
      "Survival status column is missing. Provide a 0/1 event indicator (0=censored, 1=event).",
      call. = FALSE
    )
  }

  status_raw <- status_vec

  if (is.data.frame(status_raw)) {
    status_raw <- status_raw[[1]]
  }

  if (is.factor(status_raw) || is.character(status_raw)) {
    stop(
      paste(
        "Survival status must be coded as 0/1 (0=censored, 1=event) or logical.",
        "Please recode the status column explicitly before modeling."
      ),
      call. = FALSE
    )
  }

  if (is.logical(status_raw)) {
    status_num <- as.integer(status_raw)
  } else {
    status_num <- as.numeric(status_raw)
  }

  if (length(status_num) != reference_length) {
    status_num <- rep_len(status_num, reference_length)
  }

  if (anyNA(status_num)) {
    stop(
      "Survival status contains missing values. Remove rows with missing status before modeling.",
      call. = FALSE
    )
  }

  finite_vals <- sort(unique(status_num[is.finite(status_num)]))
  if (length(finite_vals) == 0) {
    stop("Survival status is empty after cleaning. Provide a 0/1 event indicator.", call. = FALSE)
  }

  if (!all(finite_vals %in% c(0, 1))) {
    stop(
      paste(
        "Survival status must be coded as 0/1 (0=censored, 1=event) or logical.",
        "Please recode the status column explicitly before modeling."
      ),
      call. = FALSE
    )
  }

  list(status = as.integer(status_num), recoded = FALSE)
}

fastml_normalize_survival_convention <- function(convention) {
  if (is.null(convention) || length(convention) == 0 || is.na(convention[[1]])) {
    return("fastml")
  }

  convention <- tolower(as.character(convention[[1]]))
  allowed <- c("fastml", "tidymodels")
  if (!convention %in% allowed) {
    stop(
      sprintf(
        "Invalid survival_metric_convention '%s'. Choose one of: %s.",
        convention,
        paste(allowed, collapse = ", ")
      ),
      call. = FALSE
    )
  }
  convention
}

#' Fast Machine Learning Function
#'
#' Trains and evaluates multiple classification or regression models automatically detecting the task based on the target variable type.
#'
#' Fast Machine Learning Function
#'
#' Trains and evaluates multiple classification or regression models. The function automatically
#' detects the task based on the target variable type and can perform advanced hyperparameter tuning
#' using various tuning strategies.
#'
#' @param data A data frame containing the complete dataset. If both `train_data` and `test_data` are `NULL`, `fastml()` will split this into training and testing sets according to `test_size` and `stratify`. When `group_cols` is supplied, the holdout keeps groups intact; when `block_col` is supplied, the holdout uses the last rows in time order. Defaults to `NULL`.
#' @param train_data A data frame pre-split for model training. If provided, `test_data` must also be supplied, and no internal splitting will occur. Defaults to `NULL`.
#' @param test_data A data frame pre-split for model evaluation. If provided, `train_data` must also be supplied, and no internal splitting will occur. Defaults to `NULL`.
#' @param label A string specifying the name of the target variable. For
#'   survival analysis, supply a character vector with the names of the time and
#'   status columns.
#' @param algorithms A vector of algorithm names to use. Default is \code{"all"} to run all supported algorithms.
#' @param task Character string specifying model type selection. Use "auto" to
#'   let the function detect whether the target is for classification, regression,
#'   or survival based on the data. Survival is detected when `label` is a
#'   character vector of length 2 that matches time and status columns in the data.
#'   You may also explicitly set to "classification", "regression", or "survival".
#' @param test_size A numeric value between 0 and 1 indicating the proportion of the data to use for testing. For grouped holdout, this is applied to groups; for time-ordered holdout, it selects the final proportion of rows. Default is \code{0.2}.
#' @param resampling_method A string specifying the resampling method for model evaluation. Default is \code{"cv"}
#'   (cross-validation) for classification/regression. Other options include \code{"none"}, \code{"boot"},
#'   \code{"repeatedcv"}, \code{"grouped_cv"}, \code{"blocked_cv"}, \code{"rolling_origin"}, and \code{"nested_cv"}.
#'   For survival tasks, resampling is supported for parsnip-compatible engines (e.g., censored/ranger, glmnet).
#'   Native survival engines (flexsurv/rstpm2/custom xgboost) ignore resampling and will error if custom resamples
#'   are supplied. When the task auto-detects survival and \code{resampling_method} is omitted, it defaults to
#'   \code{"none"} so native engines continue to run; set it explicitly to enable resampling for parsnip survival fits.
#' @param folds An integer specifying the number of folds for cross-validation
#'   (default \code{10} for methods containing "cv", \code{25} otherwise).
#'   When \code{resampling_method = "boot"}, this controls the number of
#'   bootstrap resamples.  When \code{resampling_method = "validation_split"},
#'   the proportion held out for validation is derived as \code{1 - 1/folds}.
#' @param repeats Number of times to repeat cross-validation (only applicable for methods like "repeatedcv").
#' @param group_cols Character vector naming one or more grouping columns used when
#'   \code{resampling_method = "grouped_cv"} or when grouped nested cross-validation is desired.
#'   All rows that share the same combination of values remain together in every fold. Columns must exist
#'   in the training data and cannot contain missing values.
#' @param block_col Single column name that defines the ordering variable for
#'   \code{resampling_method = "blocked_cv"} or \code{"rolling_origin"}. Data must already be sorted in
#'   ascending order by this column to avoid leakage from future observations.
#' @param block_size Positive integer specifying the block size for \code{"blocked_cv"}.
#' @param initial_window Positive integer giving the number of observations in the initial training
#'   window for \code{"rolling_origin"} resampling.
#' @param assess_window Positive integer giving the number of observations in each assessment window for
#'   \code{"rolling_origin"} resampling.
#' @param skip Non-negative integer specifying how many potential rolling windows to skip between
#'   successive resamples when \code{resampling_method = "rolling_origin"}.
#' @param outer_folds Positive integer giving the number of outer folds to use when
#'   \code{resampling_method = "nested_cv"} and no custom \code{resamples} object is supplied.
#' @param event_class A single string. Either "first" or "second" to specify which
#'   level of the binary outcome factor to treat as the positive class (the "event").
#'   For binary classification, "first" treats the first factor level as the positive
#'   class, "second" treats the second level as positive. Use
#'   \code{levels(your_data$outcome)} to check level order before training. Default is "first".
#' @param multiclass_auc For multiclass ROC AUC, the averaging method to use:
#'   `"macro"` (default, tidymodels) or `"macro_weighted"`. Macro weights each
#'   class equally, while macro_weighted weights by class prevalence and can
#'   change model rankings on imbalanced data.
#' @param exclude A character vector specifying the names of the columns to be excluded from the training process.
#' @param recipe A user-defined \code{recipe} object for custom preprocessing. If provided, internal recipe steps (imputation, encoding, scaling) are skipped.
#' @param tune_params A named list of tuning ranges for each algorithm and engine
#'   pair. Example: \code{list(rand_forest = list(ranger = list(mtry = c(1, 3))))}
#'   will override the defaults for the ranger engine. Default is \code{NULL}.
#' @param engine_params A named list of engine-level arguments to pass directly
#'   to the underlying model fitting functions. Use this for fixed settings that
#'   should apply whenever an engine is fitted (for example,
#'   \code{list(royston_parmar = list(rstpm2 = list(link = "PO")))},
#'   \code{list(cox_ph = list(survival = list(ties = "breslow")))}, or
#'   \code{list(rand_forest = list(ranger = list(importance = "impurity")))}).
#'   These arguments are distinct from \code{tune_params}, which define ranges of
#'   hyperparameters to explore during tuning. Default is an empty list.
#' @param metric The performance metric to optimize during training. For
#'   classification, options include \code{"accuracy"}, \code{"roc_auc"},
#'   \code{"logloss"}, \code{"brier_score"}, and \code{"ece"} (plus other class metrics).
#' @param class_threshold For binary classification, controls how class probabilities
#'   are converted into hard class predictions during holdout evaluation. The default
#'   is \code{0.5} (standard threshold). Numeric values in (0, 1) set a fixed threshold.
#'   Use \code{"auto"} to tune a threshold on training data to maximize F1, or specify
#'   a metric name (e.g., \code{"sens"}, \code{"spec"}, \code{"youden"}) to optimize
#'   for that metric. Use \code{"model"} to keep the model's default predictions.
#' @param algorithm_engines A named list specifying the engine to use for each algorithm.
#' @param use_parsnip_defaults Logical. If \code{TRUE}, fastml uses parsnip's default
#'   engines instead of fastml's optimized defaults. This provides compatibility with
#'   standard tidymodels behavior. If \code{FALSE} (default), fastml uses its own
#'   curated engine defaults which may differ from parsnip. When engines differ,
#'   a warning is issued unless \code{algorithm_engines} explicitly specifies the engine.
#'   Use \code{print_default_differences()} to see all differences.
#' @param warn_engine_defaults Logical. If \code{TRUE} (default), warns when fastml's
#'   default engine differs from parsnip's default. Set to \code{FALSE} to suppress
#'   these warnings. Warnings are only shown once per algorithm per session.
#' @param n_cores An integer specifying the number of parallel worker processes used to
#'   evaluate resamples and tuning candidates. Default is \code{1} (sequential). This sizes
#'   the worker pool only; it does not change how many threads an individual engine uses,
#'   which is controlled by \code{engine_threads}.
#' @param engine_threads An integer specifying the number of threads passed to engines that
#'   accept a thread count (for example \code{num.threads} for \pkg{ranger},
#'   \code{num_threads} for \pkg{lightgbm}, \code{nthread} for \pkg{xgboost}). Default is
#'   \code{1}. Total CPU demand is approximately \code{n_cores * engine_threads}, so raising
#'   both above the number of available cores will oversubscribe the machine and can be
#'   slower than either alone. Values greater than \code{1} may also make some engines
#'   nondeterministic; a warning is issued where that applies.
#' @param stratify Logical indicating whether to use stratified sampling when splitting the data. Only applied to random holdout splitting. Default is \code{TRUE} for classification and \code{FALSE} for regression.
#' @param impute_method Method for handling missing values. Options include:
#'   \describe{
#'     \item{\code{"medianImpute"}}{Impute missing values using median imputation (recipe-based).}
#'     \item{\code{"knnImpute"}}{Impute missing values using k-nearest neighbors (recipe-based).}
#'     \item{\code{"bagImpute"}}{Impute missing values using bagging (recipe-based).}
#'     \item{\code{"remove"}}{Remove rows with missing values from the data (recipe-based).}
#'     \item{\code{"error"}}{Do not perform imputation; if missing values are detected, stop execution with an error.}
#'     \item{\code{NULL}}{Equivalent to \code{"error"}. No imputation is performed, and the function will stop if missing values are present.}
#'   }
#'   All imputation occurs inside the recipe so the same trained preprocessing
#'   can be applied at prediction time. Default is \code{"error"}.
#' @param encode_categoricals Logical indicating whether to encode categorical variables. Default is \code{TRUE}.
#' @param scaling_methods Vector of scaling methods to apply. Default is \code{c("center", "scale")}.
#' @param balance_method Method to handle class imbalance. One of \code{"none"},
#'   \code{"upsample"}, or \code{"downsample"}. Applied inside the preprocessing
#'   recipe so each resampling split is balanced independently (requires the
#'   \code{themis} package when enabled). Default is \code{"none"}.
#' @param resamples Optional rsample object providing custom resampling splits.
#'   If supplied, \code{resampling_method}, \code{folds}, and \code{repeats} are
#'   ignored.
#' @param summaryFunction A custom summary function for model evaluation. Default is \code{NULL}.
#' @param use_default_tuning Logical. Tuning only runs when resamples are supplied and
#'   \code{tuning_strategy} is not \code{"none"}. If \code{TRUE} and
#'   \code{tune_params} is \code{NULL}, default grids are used; if
#'   \code{tune_params} is provided, those values override/extend defaults. When
#'   \code{FALSE} and no custom parameters are given, models are fitted once with
#'   default settings. If no resamples are available or \code{tuning_strategy =
#'   "none"}, tuning requests are ignored with a warning. Default is \code{FALSE}.
#' @param tuning_strategy A string specifying the tuning strategy. Must be one of
#'   \code{"grid"}, \code{"bayes"}, or \code{"none"}. Default is \code{"grid"}.
#'   If custom \code{tune_params} are provided while \code{tuning_strategy = "none"},
#'   they will be ignored with a warning.
#' @param tuning_iterations Number of iterations for Bayesian tuning. Ignored when
#'   \code{tuning_strategy} is not \code{"bayes"}. Validation of this argument only
#'   occurs for the Bayesian strategy. Default is \code{10}.
#' @param tuning_complexity Character string specifying a tuning complexity preset
#'   that controls grid density and parameter range width. One of:
#'   \describe{
#'     \item{\code{"quick"}}{Minimal tuning (2 levels/param, ~32 combinations for 5 params).
#'       Best for: prototyping, debugging, time-constrained scenarios.}
#'     \item{\code{"balanced"}}{Standard tuning (3 levels/param, ~243 combinations).
#'       Best for: most production use cases. This is the default.}
#'     \item{\code{"thorough"}}{Comprehensive tuning (5 levels/param, ~3,125 combinations).
#'       Best for: final model selection, publications.}
#'     \item{\code{"exhaustive"}}{Maximum coverage (7 levels/param, ~16,807 combinations).
#'       Best for: research, competitions. Consider Bayesian tuning instead.}
#'   }
#'   See \code{\link{print_tuning_presets}} for detailed comparison.
#'   Ignored if \code{grid_levels} is explicitly set.
#' @param grid_levels Integer specifying the number of levels per parameter for
#'   grid search. Higher values create denser grids but increase computation time
#'   exponentially (grid size = levels^n_params). Typical values:
#'   \itemize{
#'     \item 2: Very fast, minimal coverage
#'     \item 3: Balanced (default via \code{tuning_complexity = "balanced"})
#'     \item 5: Thorough coverage
#'     \item 7+: Exhaustive (consider Bayesian tuning instead)
#'   }
#'   If \code{NULL} (default), determined by \code{tuning_complexity}.
#' @param early_stopping Logical indicating whether to use early stopping in Bayesian tuning methods (if supported). Default is \code{FALSE}.
#' @param adaptive Logical indicating whether to use adaptive/racing methods for tuning. Default is \code{FALSE}.
#' @param learning_curve Logical. If TRUE, generate learning curves (performance vs. training size).
#' @param seed An integer value specifying the random seed for reproducibility.
#'   fastml also configures parallel backends for deterministic RNG streams when
#'   possible; some external engines (e.g., h2o, spark, keras) may still be
#'   nondeterministic and will emit a warning.
#' @param verbose Logical; if TRUE, prints progress messages during the training
#'   and evaluation process.
#' @param eval_times Optional numeric vector of evaluation horizons for survival
#'   models. When \code{NULL}, defaults to the median and 75th percentile of the
#'   observed follow-up times (rounded to the dataset's time unit).
#' @param survival_metric_convention Character string specifying which survival
#'   metric conventions to follow. `"fastml"` (default) uses fastml's internal
#'   defaults for evaluation horizons and t_max. `"tidymodels"` uses
#'   `eval_times` as the explicit evaluation grid and applies yardstick-style
#'   Brier/IBS normalization; when `eval_times` is `NULL`, time-dependent Brier
#'   metrics are omitted.
#' @param bootstrap_ci Logical indicating whether bootstrap confidence intervals
#'   should be computed for performance metrics. Applies to all task types.
#' @param bootstrap_samples Integer giving the number of bootstrap resamples to
#'   use when \code{bootstrap_ci = TRUE}. Defaults to 500.
#' @param bootstrap_seed Optional seed passed to the bootstrap procedure used to
#'   estimate confidence intervals. When omitted, defaults to `seed` for
#'   reproducible intervals; set to `NULL` to allow random bootstrap draws.
#' @param at_risk_threshold Numeric value between 0 and 1 used for survival
#'   metrics to determine the last follow-up time (\eqn{t_{max}}). The maximum
#'   time is set to the largest observed time where at least this proportion of
#'   subjects remain at risk.
#' @param audit_mode Logical; if \code{TRUE}, enables runtime auditing of custom
#'   preprocessing hooks and records potentially unsafe behaviour (such as global
#'   environment access or file I/O) while flagging the run as potentially
#'   unsafe.
#' @param store_fold_models Logical. If \code{TRUE}, stores the models trained
#'   on each cross-validation fold (memory intensive). This enables
#'   \code{\link{explain_stability}} to compute feature importance across folds
#'   and assess explanation stability. Default is \code{FALSE}.
#' @importFrom magrittr %>%
#' @importFrom rsample initial_split training testing
#' @importFrom recipes recipe step_impute_median step_impute_knn step_impute_bag step_naomit step_dummy step_center step_scale prep bake all_numeric_predictors all_predictors all_nominal_predictors all_outcomes step_zv step_rm step_novel step_unknown
#' @importFrom dplyr filter pull rename_with mutate across where select all_of
#' @importFrom rlang sym .data
#' @importFrom stats as.formula complete.cases
#' @importFrom janitor make_clean_names
#' @importFrom stringr str_detect
#' @importFrom purrr flatten
#' @details
#' Model selection is based exclusively on resampling metrics (cross-validation
#' or nested CV). The holdout split is reserved for final performance
#' estimation and is never used to choose the best model, mirroring
#' \code{tidymodels::last_fit()} semantics.
#'
#' For multiclass ROC AUC, fastml defaults to macro averaging (tidymodels).
#' Macro treats each class equally, while macro_weighted weights by class
#' prevalence and can change model rankings on imbalanced data. Keep the same
#' setting when comparing runs.
#'
#' ## Tuning: Speed vs Robustness Trade-offs
#'
#' Hyperparameter tuning involves a fundamental trade-off between computational
#' cost and the likelihood of finding optimal hyperparameters. fastml provides
#' presets via \code{tuning_complexity} to make this trade-off explicit:
#'
#' \tabular{lllll}{
#'   \strong{Level} \tab \strong{Grid Size*} \tab \strong{Time} \tab \strong{Quality} \tab \strong{Use Case} \cr
#'   quick \tab ~32 \tab ~1x \tab Low \tab Prototyping, debugging \cr
#'   balanced \tab ~243 \tab ~10x \tab Medium \tab Most production use \cr
#'   thorough \tab ~3,125 \tab ~100x \tab High \tab Final models, papers \cr
#'   exhaustive \tab ~16,807 \tab ~1000x \tab Very High \tab Research, competitions \cr
#' }
#' *Grid size shown for 5 tunable parameters (levels^5)
#'
#' **Recommendations:**
#' \itemize{
#'   \item Start with \code{tuning_complexity = "quick"} during development
#'   \item Use \code{"balanced"} (default) for most production pipelines
#'   \item Switch to \code{"thorough"} for final model selection
#'   \item Consider \code{tuning_strategy = "bayes"} instead of exhaustive grid search
#'   \item Enable \code{adaptive = TRUE} for early stopping of poor configurations
#' }
#'
#' Use \code{\link{print_tuning_presets}} to see all presets and
#' \code{\link{estimate_tuning_time}} to estimate runtime before starting.
#'
#' @section Factor Level Warning:
#' For binary classification, the interpretation of metrics like sensitivity, specificity,
#' and ROC AUC depends on which factor level is treated as the "positive" class (the event
#' of interest). The \code{event_class} parameter controls this:
#' \itemize{
#'   \item \code{"first"} (default): The first factor level is treated as positive
#'   \item \code{"second"}: The second factor level is treated as positive
#' }
#'
#' \strong{Important:} Recipe preprocessing steps like \code{step_other()} or
#' \code{step_unknown()} can modify factor levels, potentially changing which level
#' is "first" or "second". Always verify factor levels after preprocessing.
#'
#' To ensure consistent behavior, explicitly set factor levels before calling fastml:
#' \preformatted{
#' # Ensure "positive" is the second level (event_class = "second")
#' data$outcome <- factor(data$outcome, levels = c("negative", "positive"))
#'
#' # Or ensure "positive" is the first level (event_class = "first")
#' data$outcome <- factor(data$outcome, levels = c("positive", "negative"))
#' }
#'
#' @return An object of class \code{fastml} containing the best model, performance metrics, and other information.
#' @examples
#' \donttest{
#' # Example 1: Using the iris dataset for binary classification (excluding 'setosa')
#' data(iris)
#' iris <- iris[iris$Species != "setosa", ]  # Binary classification
#' iris$Species <- factor(iris$Species)
#'
#' # Define a custom tuning grid for the ranger engine
#' tune <- list(
#'   rand_forest = list(
#'     ranger = list(mtry = c(1, 3))
#'   )
#' )
#'
#' # Train models with custom tuning
#' model <- fastml(
#'   data = iris,
#'   label = "Species",
#'   algorithms = "rand_forest",
#'   tune_params = tune,
#'   use_default_tuning = TRUE
#' )
#'
#' # View model summary
#' summary(model)
#'
#'
#'   }
#'
#' @export
fastml <- function(data = NULL,
                   train_data = NULL,
                   test_data = NULL,
                   label,
                   algorithms = "all",
                   task = "auto",
                   test_size = 0.2,
                   resampling_method = if (identical(task, "survival")) "none" else "cv",
                   folds = ifelse(grepl("cv", resampling_method), 10, 25),
                   repeats = NULL,
                   group_cols = NULL,
                   block_col = NULL,
                   block_size = NULL,
                   initial_window = NULL,
                   assess_window = NULL,
                   skip = 0,
                   outer_folds = NULL,
                   event_class = "first",
                   exclude = NULL,
                   recipe = NULL,
                   tune_params = NULL,
                   engine_params = list(),
                   metric = NULL,
                   class_threshold = 0.5,
                   algorithm_engines = NULL,
                   use_parsnip_defaults = FALSE,
                   warn_engine_defaults = TRUE,
                   n_cores = 1,
                   engine_threads = 1,
                   stratify = TRUE,
                   impute_method = "error",
                   encode_categoricals = TRUE,
                   scaling_methods = c("center", "scale"),
                   balance_method = "none",
                   resamples = NULL,
                   summaryFunction = NULL,
                   use_default_tuning = FALSE,
                   tuning_strategy = "grid",
                   tuning_iterations = 10,
                   tuning_complexity = "balanced",
                   grid_levels = NULL,
                   early_stopping = FALSE,
                   adaptive = FALSE,
                   learning_curve = FALSE,
                   seed = 123,
                   verbose = FALSE,
                   eval_times = NULL,
                   survival_metric_convention = "fastml",
                   bootstrap_ci = TRUE,
                   bootstrap_samples = 500,
                   bootstrap_seed = NULL,
                   at_risk_threshold = 0.1,
                   audit_mode = FALSE,
                   multiclass_auc = "macro",
                   store_fold_models = FALSE) {

  resampling_method_missing <- missing(resampling_method)
  bootstrap_seed_missing <- missing(bootstrap_seed)
  audit_env <- fastml_init_audit_env(audit_mode)

  .fastml_old_seed <- if (exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE))
    get(".Random.seed", envir = .GlobalEnv, inherits = FALSE) else NULL
  .fastml_had_seed <- !is.null(.fastml_old_seed)
  on.exit({
    if (.fastml_had_seed) {
      assign(".Random.seed", .fastml_old_seed, envir = .GlobalEnv)
    } else if (exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE)) {
      rm(".Random.seed", envir = .GlobalEnv)
    }
  }, add = TRUE)
  set.seed(seed)
  if (bootstrap_seed_missing) {
    bootstrap_seed <- seed
  }

  task <- match.arg(task, c("auto", "classification", "regression", "survival"))
  event_class <- match.arg(event_class, c("first", "second"))
  tuning_strategy <- match.arg(tuning_strategy, c("grid", "bayes", "none"))
  tuning_complexity <- match.arg(tuning_complexity, c("balanced", "quick", "thorough", "exhaustive"))

  # Resolve grid_levels from tuning_complexity if not explicitly set
  if (is.null(grid_levels)) {
    tuning_config <- get_tuning_complexity(tuning_complexity)
    grid_levels <- tuning_config$grid_levels
    # Also update tuning_iterations for Bayesian if using complexity preset
    if (tuning_strategy == "bayes" && missing(tuning_iterations)) {
      tuning_iterations <- tuning_config$bayes_iterations
    }
    if (verbose) {
      message(sprintf("Using '%s' tuning complexity: %d grid levels, %d Bayesian iterations",
                      tuning_complexity, grid_levels, tuning_config$bayes_iterations))
    }
  } else {
    # Validate user-provided grid_levels
    if (!is.numeric(grid_levels) || length(grid_levels) != 1 ||
        grid_levels < 2 || grid_levels != as.integer(grid_levels)) {
      stop("'grid_levels' must be an integer >= 2.", call. = FALSE)
    }
    grid_levels <- as.integer(grid_levels)
    if (verbose) {
      message(sprintf("Using custom grid_levels: %d", grid_levels))
    }
  }

  survival_metric_convention <- fastml_normalize_survival_convention(survival_metric_convention)
  if (is.null(resampling_method)) {
    resampling_method <- "none"
  } else {
    resampling_method <- tolower(resampling_method)
  }
  if (task == "survival" && resampling_method_missing && !identical(resampling_method, "none")) {
    resampling_method <- "none"
  }
  custom_resamples <- !is.null(resamples)

  unsupported_imputers <- c("mice", "missforest", "custom")
  if (!is.null(impute_method) &&
      is.character(impute_method) &&
      tolower(impute_method) %in% unsupported_imputers) {
    stop(
      paste0(
        "Advanced imputation methods ('mice', 'missForest', 'custom') are no longer supported. ",
        "For leak-free predictions, specify a recipe-based imputation step such as ",
        "'medianImpute', 'knnImpute', 'bagImpute', or use `impute_method = \"remove\"`."
      ),
      call. = FALSE
    )
  }

  coerce_count <- function(value, name, allow_zero = FALSE) {
    if (is.null(value)) {
      return(NULL)
    }
    if (!is.numeric(value) || length(value) != 1 || is.na(value)) {
      stop(
        sprintf(
          "'%s' must be a %s integer.",
          name,
          if (allow_zero) "non-negative" else "positive"
        ),
        call. = FALSE
      )
    }
    if (!isTRUE(all.equal(value, round(value)))) {
      stop(
        sprintf(
          "'%s' must be supplied as a whole number.",
          name
        ),
        call. = FALSE
      )
    }
    if (!allow_zero && value <= 0) {
      stop(sprintf("'%s' must be a positive integer.", name), call. = FALSE)
    }
    if (allow_zero && value < 0) {
      stop(sprintf("'%s' must be a non-negative integer.", name), call. = FALSE)
    }
    as.integer(round(value))
  }

  ensure_columns_present <- function(df, cols, arg_label) {
    if (is.null(cols) || length(cols) == 0) {
      return(invisible(NULL))
    }
    missing_cols <- setdiff(cols, colnames(df))
    if (length(missing_cols) > 0) {
      stop(
        sprintf(
          "Column(s) %s required by %s were not found in the training data.",
          paste(missing_cols, collapse = ", "),
          arg_label
        ),
        call. = FALSE
      )
    }
    invisible(NULL)
  }

  if (!is.null(group_cols)) {
    group_cols <- unique(as.character(group_cols))
    group_cols <- group_cols[nzchar(group_cols)]
    if (length(group_cols) == 0) {
      group_cols <- NULL
    }
  }

  if (!is.null(block_col)) {
    block_col <- as.character(block_col)
    block_col <- block_col[nzchar(block_col)]
    if (length(block_col) == 0) {
      block_col <- NULL
    } else if (length(block_col) > 1) {
      stop("'block_col' must be a single column name.", call. = FALSE)
    } else {
      block_col <- block_col[1]
    }
  }

  if (!is.null(repeats) && length(repeats) == 1 && is.na(repeats)) {
    repeats <- NULL
  }
  repeats <- coerce_count(repeats, "repeats")
  block_size <- coerce_count(block_size, "block_size")
  initial_window <- coerce_count(initial_window, "initial_window")
  assess_window <- coerce_count(assess_window, "assess_window")
  outer_folds <- coerce_count(outer_folds, "outer_folds")
  skip <- coerce_count(skip, "skip", allow_zero = TRUE)
  if (is.null(skip)) {
    skip <- 0L
  }

  if (!custom_resamples && identical(resampling_method, "grouped_cv") && is.null(group_cols)) {
    stop("`group_cols` must be provided when `resampling_method = \"grouped_cv\"`.", call. = FALSE)
  }
  if (!custom_resamples && resampling_method %in% c("blocked_cv", "rolling_origin") && is.null(block_col)) {
    stop("`block_col` must be provided for blocked or rolling resampling.", call. = FALSE)
  }
  if (!custom_resamples && identical(resampling_method, "blocked_cv") && is.null(block_size)) {
    stop("`block_size` must be provided when `resampling_method = \"blocked_cv\"`.", call. = FALSE)
  }
  if (!custom_resamples && identical(resampling_method, "rolling_origin") &&
      (is.null(initial_window) || is.null(assess_window))) {
    stop("`initial_window` and `assess_window` must be provided when `resampling_method = \"rolling_origin\"`.", call. = FALSE)
  }
  if (!custom_resamples && identical(resampling_method, "nested_cv") && is.null(outer_folds)) {
    stop("`outer_folds` must be provided when `resampling_method = \"nested_cv\"` unless custom resamples are supplied.", call. = FALSE)
  }
  if (is.null(repeats) && identical(resampling_method, "repeatedcv") && !custom_resamples) {
    repeats <- 1L
  }

  # If explicit train/test provided, ensure both are given
  if (!is.null(train_data) || !is.null(test_data)) {
    if (is.null(train_data) || is.null(test_data)) {
      stop("Both 'train_data' and 'test_data' must be provided together.")
    }
    # use provided splits; require label in both
    data <- NULL  # skip full-data splitting
  } else {
    # require data for splitting
    if (is.null(data)) stop("Either 'data' or both 'train_data' and 'test_data' must be provided.")
  }


  # Determine source for target variable and track whether task detection is pending
  source_data <- if (!is.null(data)) data else train_data
  target_var <- NULL
  pending_auto_detection <- FALSE

  # Auto-detect task if requested, including survival when label has two columns
  if (task == "auto") {
    # survival detection: label is c(time_col, status_col)
    if (is.character(label) && length(label) %in% c(2, 3)) {
      if (!all(label %in% names(source_data))) {
        missing_vars <- setdiff(label, names(source_data))
        stop(paste0(
          "When task='auto' and 'label' has length ",
          length(label),
          ", the specified columns must exist in the data for survival detection. Missing: ",
          paste(missing_vars, collapse = ", ")
        ))
      }
      time_cols <- if (length(label) == 3) label[1:2] else label[1]
      status_col <- label[length(label)]
      time_ok <- all(vapply(time_cols, function(col) is.numeric(source_data[[col]]), logical(1)))
      status_vec <- source_data[[status_col]]
      uniq_status <- unique(status_vec[!is.na(status_vec)])
      status_ok <- is.logical(status_vec) || length(uniq_status) == 2
      if (time_ok && status_ok) {
        task <- "survival"
        target_var <- NULL
      } else {
        stop("Unable to detect survival task automatically: ensure time/start/stop are numeric and status has two unique values.")
      }
    } else {
      if (!(label %in% names(source_data))) {
        stop("Label variable must exist in the data source.")
      }
      pending_auto_detection <- TRUE
    }
  } else {
    # Non-auto: validate label(s) exist and set target_var when applicable
    if (task == "survival") {
      if (!(length(label) %in% c(2, 3)) || !all(label %in% names(source_data))) {
        stop("For survival tasks, 'label' must contain the time/status columns present in the data (length 2 or 3).")
      }
      target_var <- NULL
    } else {
      if (!(label %in% names(source_data))) {
        stop("Label variable must exist in the data source.")
      }
    }
  }

  # Early check: label column must not contain missing values for non-survival tasks
  if (!is.null(label) &&
      length(label) == 1 &&
      task != "survival" &&
      label %in% names(source_data)) {
    missing_targets <- sum(is.na(source_data[[label]]))
    if (missing_targets > 0) {
      stop(
        sprintf(
          "Error: The label variable '%s' contains %d missing values (NA).
          Rows with missing targets cannot be used for training.
          Please filter these rows out before running fastml.",
          label,
          missing_targets
        )
      )
    }
  }

  # determine positive_class after data split when factor levels are available
  positive_class <- NULL
  # ---------------- END TASK DETECTION ----------------


  holdout_mode <- "random"
  if (!is.null(block_col)) {
    if (!is.null(group_cols)) {
      warning(
        "Both 'block_col' and 'group_cols' were provided; holdout splitting will use 'block_col' ordering and ignore grouping.",
        call. = FALSE
      )
    }
    holdout_mode <- "time"
  } else if (!is.null(group_cols)) {
    holdout_mode <- "group"
  }

  split_holdout <- function(df, use_strata = FALSE) {
    seed_val <- fastml_normalize_seed(seed)
    if (!is.null(seed_val)) {
      set.seed(seed_val)
    }
    if (use_strata && !identical(holdout_mode, "random")) {
      warning(
        "Stratified holdout is only supported for random splitting; proceeding without stratification.",
        call. = FALSE
      )
      use_strata <- FALSE
    }

    if (identical(holdout_mode, "time")) {
      if (is.null(block_col) || !block_col %in% names(df)) {
        stop("`block_col` must be present in the data for time-ordered holdout splitting.", call. = FALSE)
      }
      block_vals <- df[[block_col]]
      if (any(is.na(block_vals))) {
        stop(sprintf("Column '%s' contains missing values which are not supported for time-ordered holdout splitting.", block_col), call. = FALSE)
      }
      if (!identical(order(block_vals), seq_len(nrow(df)))) {
        stop(sprintf("Data must be sorted by '%s' in ascending order before time-ordered holdout splitting.", block_col), call. = FALSE)
      }

      n_rows <- nrow(df)
      n_test <- max(1L, floor(n_rows * test_size))
      if (n_rows - n_test < 1L) {
        stop("`test_size` leaves no data for training after time-ordered splitting; reduce `test_size`.", call. = FALSE)
      }
      test_idx <- seq.int(n_rows - n_test + 1L, n_rows)
      train_idx <- seq_len(n_rows - n_test)
      return(list(
        train_data = df[train_idx, , drop = FALSE],
        test_data = df[test_idx, , drop = FALSE]
      ))
    }

    if (identical(holdout_mode, "group")) {
      if (is.null(group_cols) || !all(group_cols %in% names(df))) {
        stop("`group_cols` must be present in the data for grouped holdout splitting.", call. = FALSE)
      }
      group_df <- df[, group_cols, drop = FALSE]
      group_id <- if (length(group_cols) == 1) {
        group_df[[1]]
      } else {
        interaction(group_df, drop = TRUE)
      }
      if (any(is.na(group_id))) {
        stop("Grouping columns contain missing values, which are not supported for grouped holdout splitting.", call. = FALSE)
      }
      unique_groups <- unique(group_id)
      n_groups <- length(unique_groups)
      n_test_groups <- max(1L, floor(n_groups * test_size))
      if (n_groups - n_test_groups < 1L) {
        stop("`test_size` leaves no groups for training after grouped splitting; reduce `test_size`.", call. = FALSE)
      }
      test_groups <- sample(unique_groups, size = n_test_groups, replace = FALSE)
      in_test <- group_id %in% test_groups
      return(list(
        train_data = df[!in_test, , drop = FALSE],
        test_data = df[in_test, , drop = FALSE]
      ))
    }

    split <- if (use_strata) {
      rsample::initial_split(
        df,
        prop = 1 - test_size,
        strata = dplyr::all_of(label)
      )
    } else {
      rsample::initial_split(df, prop = 1 - test_size)
    }
    list(
      train_data = rsample::training(split),
      test_data = rsample::testing(split)
    )
  }

  provisional_split_used <- FALSE
  # If initial data provided, perform exclusion and checks, then split
  if (!is.null(data)) {
    if (task == "survival") {
      if (!all(label %in% names(data))) {
        stop("The specified label(s) do not exist in the data.")
      }
    } else {
      if (!(label %in% names(data))) {
        stop("The specified label does not exist in the data.")
      }
    }
    if (!is.null(exclude)) {
      if (task == "survival") {
        if (any(label %in% exclude)) {
          stop("Label variable(s) cannot be excluded: ", paste(label[label %in% exclude], collapse = ", "))
        }
      } else {
        if (label %in% exclude) stop("Label variable cannot be excluded: ", label)
      }
      missing_vars <- setdiff(exclude, colnames(data))
      if (length(missing_vars) > 0) {
        warning("Variables not in data: ", paste(missing_vars, collapse = ", "))
        exclude <- setdiff(exclude, missing_vars)
      }
      data <- dplyr::select(data, -dplyr::all_of(exclude))
    }
    if (!is.null(impute_method) && impute_method == "error" && anyNA(data)) {
      stop("Data contains NAs and 'impute_method = \"error\"'. Handle missing values first.")
    }

    integer_cols <- names(data)[vapply(data, is.integer, logical(1))]
    label_cols <- if (task == "survival") label else label
    integer_predictors <- setdiff(integer_cols, label_cols)
    if (length(integer_predictors) > 0) {
      low_cardinality <- integer_predictors[vapply(
        data[integer_predictors],
        function(col) length(unique(col[!is.na(col)])) <= 10,
        logical(1)
      )]
      if (length(low_cardinality) > 0) {
        warning(
          paste(
            "Integer predictors are treated as numeric.",
            "If any of these are categorical codes, convert them to factor before calling fastml:",
            paste(low_cardinality, collapse = ", ")
          ),
          call. = FALSE
        )
      }
    }

    data <- data %>%
      dplyr::mutate(
        dplyr::across(where(is.character), as.factor),
        dplyr::across(where(is.integer), as.numeric)
      )

    if (verbose) message("Splitting data into training and test sets...")
    # Split into train/test
    if (pending_auto_detection && task == "auto") {
      split_result <- split_holdout(data, use_strata = FALSE)
      provisional_split_used <- TRUE
    } else if (stratify && task == "classification") {
      split_result <- split_holdout(data, use_strata = TRUE)
    } else {
      split_result <- split_holdout(data, use_strata = FALSE)
    }

    train_data <- split_result$train_data
    test_data  <- split_result$test_data
  }

  if (task == "survival") {
    if (!(length(label) %in% c(2, 3))) {
      stop("For survival tasks, 'label' must contain the time/status columns present in the data (length 2 or 3).")
    }
    if (length(label) == 3) {
      stop(
        paste(
          "Start/stop (left-truncated) survival data are not supported for evaluation in fastml.",
          "Provide right-censored inputs as c(time, status) or evaluate externally."
        ),
        call. = FALSE
      )
    }
    outcome_cols_check <- if (length(label) == 2) label else label[1:3]
    drop_missing_outcomes <- function(df, which_set) {
      missing_idx <- !stats::complete.cases(df[, outcome_cols_check, drop = FALSE])
      if (any(missing_idx)) {
        warning(
          sprintf(
            "Removed %d row(s) with missing survival outcomes from the %s set.",
            sum(missing_idx),
            which_set
          ),
          call. = FALSE
        )
        df <- df[!missing_idx, , drop = FALSE]
      }
      df
    }
    train_data <- drop_missing_outcomes(train_data, "training")
    test_data  <- drop_missing_outcomes(test_data, "test")
    if (nrow(train_data) == 0 || nrow(test_data) == 0) {
      stop("No data remain after removing rows with missing survival outcomes.")
    }
    label_surv <- label
    status_warning_emitted <- FALSE
    normalize_status <- function(vec, reference_length) {
      res <- fastml_normalize_survival_status(vec, reference_length)
      if (res$recoded && !status_warning_emitted) {
        warning("Detected non-standard survival status coding; recoding to 0 = censored and 1 = event.", call. = FALSE)
        status_warning_emitted <<- TRUE
      }
      res$status
    }
    if (length(label) == 2) {
      start_col = NULL
      time_col <- label[1]
      status_col <- label[2]
      train_status <- normalize_status(train_data[[status_col]], nrow(train_data))
      test_status  <- normalize_status(test_data[[status_col]], nrow(test_data))
      train_data[[status_col]] <- train_status
      test_data[[status_col]]  <- test_status
      surv_train <- survival::Surv(train_data[[time_col]], train_status)
      surv_test  <- survival::Surv(test_data[[time_col]], test_status)
    } else {
      start_col <- label[1]
      stop_col <- label[2]
      status_col <- label[3]
      train_status <- normalize_status(train_data[[status_col]], nrow(train_data))
      test_status  <- normalize_status(test_data[[status_col]], nrow(test_data))
      train_data[[status_col]] <- train_status
      test_data[[status_col]]  <- test_status
      surv_train <- survival::Surv(train_data[[start_col]], train_data[[stop_col]], train_status)
      surv_test  <- survival::Surv(test_data[[start_col]], test_data[[stop_col]], test_status)
    }
    attr(surv_train, "fastml_label_cols") <- label_surv
    attr(surv_test, "fastml_label_cols") <- label_surv
    train_data$surv_obj <- surv_train
    test_data$surv_obj <- surv_test
    label <- "surv_obj"
  } else {
    label_surv <- label
  }


  if (task != "survival") {
    target_var <- train_data[[label]]
  }

  if (pending_auto_detection && task != "survival") {
    if (is.numeric(target_var) && length(unique(target_var)) <= 2) {
      train_data[[label]] <- factor(train_data[[label]])
      test_data[[label]]  <- factor(test_data[[label]], levels = levels(train_data[[label]]))
      task <- "classification"
      warning(sprintf(
        "The target variable '%s' is numeric with %d unique values. Converted to factor; task set to 'classification'.",
        label, length(unique(target_var))
      ))
    } else if (is.factor(target_var) || is.character(target_var) || is.logical(target_var)) {
      task <- "classification"
    } else if (is.numeric(target_var)) {
      task <- "regression"
    } else {
      stop("Unable to detect task type. The target variable must be numeric, factor, character, or logical.")
    }
    pending_auto_detection <- FALSE
  }

  if (provisional_split_used && stratify && task == "classification") {
    if (identical(holdout_mode, "random")) {
      set.seed(seed)
      split_result <- split_holdout(data, use_strata = TRUE)
      train_data <- split_result$train_data
      test_data  <- split_result$test_data
      target_var <- train_data[[label]]
    } else {
      warning(
        "Stratified holdout is not supported for grouped/time-based splitting; keeping the existing holdout split.",
        call. = FALSE
      )
    }
  }

  multiclass_auc <- fastml_normalize_multiclass_auc(multiclass_auc)

  # Set default metric now that task has been resolved and validate it
  if (is.null(metric)) {
    metric <- if (task == "classification") {
      "roc_auc"
    } else if (task == "regression") {
      "rmse"
    } else {
      "ibs"
    }
  }

  allowed_metrics_classification <- c("accuracy", "kap", "sens", "spec", "precision", "f_meas", "roc_auc", "logloss", "brier_score", "ece")
  allowed_metrics_regression <- c("rmse", "rsq", "mae")
  allowed_metrics_survival <- c("c_index", "uno_c", "ibs", "rmst_diff")

  if (task == "classification") {
    if (!(metric %in% allowed_metrics_classification) && is.null(summaryFunction)) {
      stop(paste0("Invalid metric for classification task. Choose one of: ",
                  paste(allowed_metrics_classification, collapse = ", "), "."))
    }
  } else if (task == "regression") {
    if (!(metric %in% allowed_metrics_regression) && is.null(summaryFunction)) {
      stop(paste0("Invalid metric for regression task. Choose one of: ",
                  paste(allowed_metrics_regression, collapse = ", "), "."))
    }
  } else {
    if (!(metric %in% allowed_metrics_survival || grepl("^brier_t", metric))) {
      stop(paste0("Invalid metric for survival task. Choose one of: ",
                  paste(c(allowed_metrics_survival, "brier_t*"), collapse = ", "), "."))
    }
  }

  supported_algorithms <- availableMethods(type = task)

  if ("all" %in% algorithms) {
    algorithms <- supported_algorithms
  } else {
    if (length(intersect(algorithms, supported_algorithms)) == 0) {
      stop("No valid algorithms specified.")
    }
    invalid_algos <- setdiff(algorithms, supported_algorithms)
    if (length(invalid_algos) > 0) {
      warning("Invalid algorithm(s) specified: ", paste(invalid_algos, collapse = ", "))
      algorithms <- intersect(algorithms, supported_algorithms)
    }
  }

  if (task == "survival") {
    warned_xgb <- FALSE
    warned_rf <- FALSE
    for (i in seq_along(algorithms)) {
      alg <- algorithms[[i]]
      if (identical(alg, "xgboost")) {
        if (!warned_xgb) {
          warning("Survival 'xgboost' is AFT-only; prefer 'xgboost_aft' (alias added).")
          warned_xgb <- TRUE
        }
        algorithms[[i]] <- "xgboost_aft"
      } else if (identical(alg, "rand_forest")) {
        if (!warned_rf) {
          warning("Survival 'rand_forest' uses censored RF engines; prefer 'rand_forest_survival' (alias added).")
          warned_rf <- TRUE
        }
        algorithms[[i]] <- "rand_forest_survival"
      }
    }
  }


  if(!is.null(impute_method) && impute_method == "remove") {
    if (anyNA(train_data)) {
      train_data <- train_data %>%
        filter(complete.cases(.))

      warning("Rows with missing values have been removed from the training set.")

    }

    if (anyNA(test_data)) {
      test_data <- test_data %>%
        filter(complete.cases(.))

      warning("Rows with missing values have been removed from the test set.")

    }

  }



  if (task == "classification") {
    train_data[[label]] <- as.factor(train_data[[label]])
    test_data[[label]] <- factor(test_data[[label]], levels = levels(train_data[[label]]))

    n_classes <- length(levels(train_data[[label]]))
    if (n_classes == 2 && "multinom_reg" %in% algorithms) {
      warning(
        sprintf(
          "Multinomial regression ('multinom_reg') is not applicable to two-class outcomes. Detected a binary outcome for '%s'; using logistic regression ('logistic_reg') instead.",
          label
        ),
        call. = FALSE
      )

      algorithms[algorithms == "multinom_reg"] <- "logistic_reg"
      algorithms <- algorithms[!duplicated(algorithms)]

      if (!is.null(algorithm_engines) &&
          !is.null(algorithm_engines[["multinom_reg"]]) &&
          is.null(algorithm_engines[["logistic_reg"]])) {
        allowed_logistic_engines <- c(
          "glm",
          "gee",
          "glmer",
          "stan",
          "stan_glmer",
          "brulee",
          "glmnet",
          "h2o",
          "LiblineaR",
          "spark",
          "keras"
        )
        candidate_engines <- intersect(
          as.character(algorithm_engines[["multinom_reg"]]),
          allowed_logistic_engines
        )
        if (length(candidate_engines) > 0) {
          algorithm_engines[["logistic_reg"]] <- candidate_engines
        }
      }

      if (!is.null(tune_params) &&
          !is.null(tune_params[["multinom_reg"]]) &&
          is.null(tune_params[["logistic_reg"]])) {
        tune_params[["logistic_reg"]] <- tune_params[["multinom_reg"]]
      }

      if (!is.null(engine_params) &&
          !is.null(engine_params[["multinom_reg"]]) &&
          is.null(engine_params[["logistic_reg"]])) {
        engine_params[["logistic_reg"]] <- engine_params[["multinom_reg"]]
      }
    }

    if (n_classes >= 2) {
      positive_class <- ifelse(event_class == "first",
                               levels(train_data[[label]])[1],
                               levels(train_data[[label]])[2])
    }
  } else if (task == "regression") {
    train_data[[label]] <- as.numeric(train_data[[label]])
    test_data[[label]] <- as.numeric(test_data[[label]])
    positive_class <- NULL
  } else {
    positive_class <- NULL
  }

  skip_recipe_imputation <- FALSE
  # If user has provided a custom recipe, skip internal imputation logic
  if (!is.null(recipe)) {
    fastml_validate_user_recipe(recipe, audit_env)
    skip_recipe_imputation <- TRUE
  } else {
    skip_recipe_imputation <- FALSE
  }

  if (is.null(recipe) && task == "classification" && "discrim_quad" %in% algorithms) {
    predictor_cols <- setdiff(names(train_data), label)
    numeric_predictors <- predictor_cols[vapply(train_data[predictor_cols], is.numeric, logical(1))]

    if (length(numeric_predictors) > 0) {
      zero_var_by_class <- vapply(numeric_predictors, function(col) {
        values <- train_data[[col]]
        split_vals <- split(values, train_data[[label]])
        any(vapply(split_vals, function(class_values) {
          class_values <- class_values[!is.na(class_values)]
          length(class_values) <= 1 || isTRUE(all.equal(stats::var(class_values), 0))
        }, logical(1)))
      }, logical(1))

      cols_to_drop <- names(zero_var_by_class)[zero_var_by_class]

      if (length(cols_to_drop) > 0) {
        warning(
          sprintf(
            "Removed %d predictor%s with zero variance within at least one class to stabilize Quadratic Discriminant Analysis: %s.",
            length(cols_to_drop),
            ifelse(length(cols_to_drop) == 1, "", "s"),
            paste(cols_to_drop, collapse = ", ")
          )
        )

        train_data <- dplyr::select(train_data, -dplyr::all_of(cols_to_drop))
        test_data <- dplyr::select(test_data, -dplyr::all_of(cols_to_drop))
      }
    }
  }

  reference_resample_data <- train_data
  ensure_columns_present(reference_resample_data, group_cols, "`group_cols`")
  ensure_columns_present(reference_resample_data, block_col, "`block_col`")
  ###############################################################################

  # Set up parallel processing using future and restore it on exit.
  parallel_ctx <- fastml_setup_parallel(n_cores = n_cores, seed = seed)
  on.exit(parallel_ctx$restore(), add = TRUE)

  ########################################################################
  # Build or use recipe if user hasn't provided it
  ########################################################################
  if (is.null(recipe)) {
    if (verbose) message("Creating preprocessing recipe...")
    # Build a default recipe, possibly skipping recipe-based imputation if advanced used
    recipe <- recipe(as.formula(paste(label, "~ .")), data = train_data)
    if (task == "survival") {
      recipe <- recipe %>% step_rm(all_of(label_surv))
    }

    # Remove zero-variance predictors before any additional transformations
    recipe <- recipe %>%
      step_zv(all_predictors())

    # Apply recipe-based imputation steps only if skip_recipe_imputation = FALSE
    if (!skip_recipe_imputation) {
      if (!is.null(impute_method) && impute_method == "medianImpute") {
        recipe <- recipe %>% step_impute_median(all_numeric_predictors())

        warning("Missing values in numeric predictors are being imputed using the median.")

      } else if (!is.null(impute_method) && impute_method == "knnImpute") {
        recipe <- recipe %>% step_impute_knn(all_predictors())

        warning("Missing values are being imputed using KNN (k-Nearest Neighbors).")

      } else if (!is.null(impute_method) && impute_method == "bagImpute") {
        recipe <- recipe %>% step_impute_bag(all_predictors())

        warning("Missing values are being imputed using bagging (bootstrap aggregation)")


      } else if (!is.null(impute_method) && impute_method == "remove") {
        recipe <- recipe %>% step_naomit(all_predictors(), skip = TRUE)

        warning("Rows with missing values in predictors are being removed.")

      } else if (impute_method == "error" || is.null(impute_method)) {
        # We'll detect if there's still NA after prep/bake, then stop
        # so do nothing special here
      } else {
        # If it's a leftover method not recognized
        stop("Invalid impute_method specified.")
      }
    } else {
      # skip_recipe_imputation = TRUE means we are using advanced or custom
      # so if "error" was set but we do have NAs after that, it won't be handled here
      # do nothing in recipe for imputation
    }

    # If encoding needed
    if (encode_categoricals) {
      recipe <- recipe %>%
        step_novel(all_nominal_predictors(), -all_outcomes()) %>%
        step_unknown(all_nominal_predictors(), -all_outcomes()) %>%
        step_dummy(all_nominal_predictors(), -all_outcomes())

      # Encoding can introduce zero-variance predictors (e.g., unused levels),
      # which would otherwise trigger warnings during subsequent scaling steps.
      recipe <- recipe %>% step_zv(all_predictors())
    }

    # scaling
    if (!is.null(scaling_methods)) {
      if ("center" %in% scaling_methods) {
        recipe <- recipe %>% step_center(all_numeric_predictors())
      }
      if ("scale" %in% scaling_methods) {
        recipe <- recipe %>% step_scale(all_numeric_predictors())
      }
    }

    # do not prep yet
  }

  if (task == "classification" && balance_method != "none") {
    if (!requireNamespace("themis", quietly = TRUE)) {
      stop("The 'themis' package is required for balance_method = \"upsample\" or \"downsample\".")
    }

    has_balance_step <- FALSE
    if (!is.null(recipe$steps) && length(recipe$steps) > 0) {
      has_balance_step <- any(vapply(recipe$steps, inherits, logical(1), "step_upsample")) ||
        any(vapply(recipe$steps, inherits, logical(1), "step_downsample"))
    }

    if (has_balance_step) {
      warning(
        "balance_method was ignored because the recipe already includes a sampling step.",
        call. = FALSE
      )
    } else if (balance_method == "upsample") {
      recipe <- recipe %>%
        themis::step_upsample(all_outcomes(), seed = seed, skip = TRUE)
    } else if (balance_method == "downsample") {
      recipe <- recipe %>%
        themis::step_downsample(all_outcomes(), seed = seed, skip = TRUE)
    } else {
      stop("Invalid balance_method. Use \"none\", \"upsample\", or \"downsample\".")
    }
  }

  if (verbose) message("Training models: ", paste(algorithms, collapse = ", "))

  models <- withCallingHandlers(
    train_models(
      train_data = train_data,
      label = label,
      task = task,
      algorithms = algorithms,
      resampling_method = resampling_method,
      folds = folds,
      repeats = repeats,
      group_cols = group_cols,
      block_col = block_col,
      block_size = block_size,
      initial_window = initial_window,
      assess_window = assess_window,
      skip = skip,
      outer_folds = outer_folds,
      resamples = resamples,
      tune_params = tune_params,
      engine_params = engine_params,
      metric = metric,
      summaryFunction = summaryFunction,
      seed = seed,
      recipe = recipe,
      use_default_tuning = use_default_tuning,
      tuning_strategy = tuning_strategy,
      tuning_iterations = tuning_iterations,
      tuning_complexity = tuning_complexity,
      grid_levels = grid_levels,
      early_stopping = early_stopping,
      adaptive = adaptive,
      algorithm_engines = algorithm_engines,
      use_parsnip_defaults = use_parsnip_defaults,
      warn_engine_defaults = warn_engine_defaults,
      n_cores = n_cores,
      engine_threads = engine_threads,
      verbose = verbose,
      event_class = event_class,
      class_threshold = class_threshold,
      start_col = start_col,
      time_col = time_col,
      status_col = status_col,
      eval_times = eval_times,
      at_risk_threshold = at_risk_threshold,
      survival_metric_convention = survival_metric_convention,
      audit_env = audit_env,
      multiclass_auc = multiclass_auc,
      store_fold_models = store_fold_models
    ),
    warning = fastml_muffle_foreach_warning
  )

  resampling_results <- attr(models, "guarded_resampling")
  attr(models, "guarded_resampling") <- NULL

  nested_results <- attr(models, "nested_cv_results")
  attr(models, "nested_cv_results") <- NULL

  resampling_plan <- attr(models, "resampling_plan")
  attr(models, "resampling_plan") <- NULL

  models <- models[sapply(models, function(x) length(x) > 0)]

  engine_names <- get_engine_names(models)

  if (length(models) == 0) {
    stop("No models were successfully trained.")
  }

  if (verbose) message("Evaluating models...")
  eval_output <- fastml_compute_holdout_results(models,
                                               train_data,
                                               test_data,
                                               label_surv,
                                               start_col,
                                               time_col,
                                               status_col,
                                               task,
                                               metric,
                                               event_class,
                                               class_threshold = class_threshold,
                                               eval_times = eval_times,
                                               bootstrap_ci = bootstrap_ci,
                                               bootstrap_samples = bootstrap_samples,
                                               bootstrap_seed = bootstrap_seed,
                                               at_risk_threshold = at_risk_threshold,
                                               survival_metric_convention = survival_metric_convention,
                                               summaryFunction = summaryFunction,
                                               multiclass_auc = multiclass_auc)
  performance <- eval_output$performance
  predictions <- eval_output$predictions

  # Unwrap single-engine performance entries for easier access.
  performance <- lapply(performance, function(entry) {
    if (is.list(entry) && !inherits(entry, "data.frame") && length(entry) == 1) {
      single <- entry[[1]]
      if (inherits(single, "data.frame")) {
        return(single)
      }
    }
    entry
  })

  # metric_values <- sapply(performance, function(x) x %>% filter(.metric == metric) %>% pull(.estimate))

  # Build a normalized performance list and align model names as "algorithm (engine)"
  combined_performance  <- list()
  model_map <- list()
  display_algo <- function(algo, task) {
    if (task == "survival") {
      if (algo %in% c("xgboost", "xgboost_aft")) return("xgboost (AFT)")
      if (algo %in% c("rand_forest", "rand_forest_survival")) return("rand_forest (censored)")
    }
    algo
  }
  for (alg in names(performance)) {
    perf_alg <- performance[[alg]]
    # Case 1: nested by engine (list of tibbles)
    if (is.list(perf_alg) && !inherits(perf_alg, "data.frame")) {
      for (eng in names(perf_alg)) {
        combined_name <- paste0(display_algo(alg, task), " (", eng, ")")
        combined_performance[[combined_name]] <- perf_alg[[eng]]
        # Map corresponding model
        if (is.list(models[[alg]]) && !inherits(models[[alg]], "workflow") && !inherits(models[[alg]], "tune_results")) {
          model_map[[combined_name]] <- models[[alg]][[eng]]
        } else {
          model_map[[combined_name]] <- models[[alg]]
        }
      }
    } else if (inherits(perf_alg, "data.frame")) {
      # Case 2: single workflow (tibble). Infer engine name when possible
      eng_candidates <- tryCatch(engine_names[[alg]], error = function(e) NULL)
      eng <- if (!is.null(eng_candidates) && length(eng_candidates) >= 1 && !is.na(eng_candidates[1])) {
        eng_candidates[1]
      } else {
        tryCatch(get_default_engine(alg, task), error = function(e) "unknown")
      }
      combined_name <- paste0(display_algo(alg, task), " (", eng, ")")
      combined_performance[[combined_name]] <- perf_alg
      model_entry <- models[[alg]]
      if (is.list(model_entry) &&
          !inherits(model_entry, c("workflow", "tune_results", "fastml_native_survival", "model_fit"))) {
        if (!is.null(eng) && !is.na(eng) && !is.null(model_entry[[eng]])) {
          model_map[[combined_name]] <- model_entry[[eng]]
        } else if (length(model_entry) == 1) {
          model_map[[combined_name]] <- model_entry[[1]]
        } else {
          model_map[[combined_name]] <- model_entry
        }
      } else {
        model_map[[combined_name]] <- model_entry
      }
    }
  }

  survival_brier_times <- NULL
  survival_t_max <- NULL
  if (task == "survival") {
    for (entry in combined_performance) {
      if (is.data.frame(entry)) {
        times_attr <- attr(entry, "brier_times")
        tmax_attr <- attr(entry, "t_max")
        if (!is.null(times_attr) || !is.null(tmax_attr)) {
          survival_brier_times <- times_attr
          survival_t_max <- tmax_attr
          break
        }
      }
    }
  }

  # Replace models with normalized, consistently named map
  models <- model_map

  if (!is.null(resampling_results)) {
    resampling_named <- list()
    for (alg in names(resampling_results)) {
      alg_entry <- resampling_results[[alg]]
      if (is.list(alg_entry) && !inherits(alg_entry, "data.frame")) {
        for (eng in names(alg_entry)) {
          combined_name <- paste0(display_algo(alg, task), " (", eng, ")")
          resampling_named[[combined_name]] <- alg_entry[[eng]]
        }
      } else if (!is.null(alg_entry)) {
        eng_candidates <- tryCatch(engine_names[[alg]], error = function(e) NULL)
        eng <- if (!is.null(eng_candidates) && length(eng_candidates) >= 1 && !is.na(eng_candidates[1])) {
          eng_candidates[1]
        } else {
          tryCatch(get_default_engine(alg, task), error = function(e) "unknown")
        }
        combined_name <- paste0(display_algo(alg, task), " (", eng, ")")
        resampling_named[[combined_name]] <- alg_entry
      }
    }
    resampling_results <- resampling_named
  }

  if (!is.null(resampling_results)) {
    resampling_results <- lapply(resampling_results, function(entry) {
      if (!is.list(entry) || inherits(entry, "data.frame")) {
        return(entry)
      }
      folds_tbl <- entry$folds
      if (!is.null(folds_tbl) && is.data.frame(folds_tbl) && nrow(folds_tbl) > 0) {
        # Always recompute aggregated from folds to ensure proper CV statistics
        # (mean and SD across folds) instead of pooled metrics
        agg <- fastml_aggregate_resample_metrics(folds_tbl, task)
        if (!is.null(agg) && is.data.frame(agg) && nrow(agg) > 0) {
          entry$aggregated <- agg
        }
      }
      entry
    })
  }

  if (!is.null(nested_results)) {
    nested_named <- list()
    for (alg in names(nested_results)) {
      alg_entry <- nested_results[[alg]]
      if (is.list(alg_entry) && !inherits(alg_entry, "data.frame")) {
        for (eng in names(alg_entry)) {
          combined_name <- paste0(alg, " (", eng, ")")
          nested_named[[combined_name]] <- alg_entry[[eng]]
        }
      } else if (!is.null(alg_entry)) {
        eng_candidates <- tryCatch(engine_names[[alg]], error = function(e) NULL)
        eng <- if (!is.null(eng_candidates) && length(eng_candidates) >= 1 && !is.na(eng_candidates[1])) {
          eng_candidates[1]
        } else {
          tryCatch(get_default_engine(alg, task), error = function(e) "unknown")
        }
        combined_name <- paste0(alg, " (", eng, ")")
        nested_named[[combined_name]] <- alg_entry
      }
    }
    nested_results <- nested_named
  }


  selection_bundle <- fastml_extract_selection_performance(
    resampling_results = resampling_results,
    nested_results = nested_results,
    task = task
  )
  selection_performance <- selection_bundle$performance
  selection_source <- if (is.null(selection_bundle$source)) "none" else selection_bundle$source

  # Select best model strictly from resampling metrics (CV or nested CV).
  lower_is_better_metrics <- c("rmse", "mae", "ibs", "logloss", "mse", "brier_score", "ece")
  extract_metric_values <- function(metric_name) {
    sapply(selection_performance, function(x) {
      if (is.data.frame(x)) {
        vals <- x[x$.metric == metric_name, ".estimate", drop = TRUE]
        if (length(vals) == 0) {
          NA_real_
        } else {
          as.numeric(vals[1])
        }
      } else {
        NA_real_
      }
    })
  }

  compute_lower_is_better <- function(metric_name) {
    if (is.null(metric_name) || length(metric_name) == 0 || is.na(metric_name)) {
      return(FALSE)
    }
    direction <- tryCatch({
      if (requireNamespace("yardstick", quietly = TRUE) &&
          exists(metric_name, envir = asNamespace("yardstick"), inherits = FALSE)) {
        attr(get(metric_name, envir = asNamespace("yardstick")), "direction", exact = TRUE)
      } else {
        NA_character_
      }
    }, error = function(e) NA_character_)

    if (!is.na(direction) && !is.null(direction)) {
      identical(direction, "minimize")
    } else {
      metric_name %in% lower_is_better_metrics ||
        grepl("^brier_t", metric_name) ||
        grepl("brier", metric_name, fixed = TRUE)
    }
  }

  if (length(selection_performance) == 0) {
    if (length(models) > 1) {
      warning(
        paste(
          "No resampling metrics are available for model selection.",
          "Holdout evaluation is reserved for final performance estimation.",
          "Defaulting to the first trained model."
        ),
        call. = FALSE
      )
    }
    best_model_idx <- names(models)[1]
  } else {
    metric_values <- extract_metric_values(metric)
    lower_is_better <- compute_lower_is_better(metric)

    if (!any(is.finite(metric_values))) {
      available_metrics <- unique(unlist(lapply(selection_performance, function(x) {
        if (is.data.frame(x)) {
          as.character(x$.metric)
        } else {
          character(0)
        }
      })))

      fallback_priority <- if (task == "classification") {
        c("logloss", "brier_score", "ece", "roc_auc", "accuracy", "kap", "sens", "spec", "precision", "f_meas")
      } else if (task == "regression") {
        c("rmse", "mae", "rsq")
      } else {
        c("c_index", "uno_c", "ibs", "rmst_diff")
      }

      candidate_metrics <- setdiff(fallback_priority, metric)
      candidate_metrics <- candidate_metrics[candidate_metrics %in% available_metrics]
      remaining_metrics <- setdiff(available_metrics, c(metric, candidate_metrics))
      candidate_metrics <- c(candidate_metrics, remaining_metrics)

      fallback_metric <- NULL
      for (cand in candidate_metrics) {
        cand_values <- extract_metric_values(cand)
        if (any(is.finite(cand_values))) {
          fallback_metric <- cand
          metric_values <- cand_values
          lower_is_better <- compute_lower_is_better(cand)
          break
        }
      }

      if (!is.null(fallback_metric)) {
        warning(sprintf("Metric '%s' unavailable across models; falling back to '%s'.", metric, fallback_metric), call. = FALSE)
        metric <- fallback_metric
      }
    }

    if (!any(is.finite(metric_values))) {
      stop("None of the models returned the specified metric from resampling.")
    }

    if (any(is.na(metric_values))) {
      warning("Some models did not return the specified metric from resampling.")
      metric_values[is.na(metric_values)] <- if (lower_is_better) Inf else -Inf
    }

    if (all(metric_values == if (lower_is_better) Inf else -Inf)) {
      stop("None of the models returned the specified metric from resampling.")
    }

    best_model_idx <- if (lower_is_better) {
      names(metric_values[metric_values == min(metric_values)])
    } else {
      names(metric_values[metric_values == max(metric_values)])
    }
  }

  # model_names <- get_model_engine_names(models)
  # best_model_name <- model_names[model_names %in% best_model_idx]

  # Split each compound name into algorithm and engine parts
  best_model_components <- lapply(best_model_idx, function(x) {
    parts <- strsplit(x, " \\(")[[1]]
    algo <- parts[1]
    engine <- gsub("\\)", "", parts[2])
    list(algo = algo, engine = engine)
  })

  # Build a named vector: names are the algorithm and values are the engine
  best_model_name <- sapply(best_model_components, function(comp) comp$engine)
  names(best_model_name) <- sapply(best_model_components, function(comp) comp$algo)

  if (verbose) {
    msg <- paste0(names(best_model_name), " (", best_model_name, ")", collapse = ", ")
    if (selection_source == "none") {
      message("Best model defaulted to first trained model (no resampling metrics available).")
    } else {
      message("Best model selected using ", selection_source, " metrics: ", msg)
    }
  }


  if (verbose) message("Preparing preprocessing recipe for downstream use...")
  # Now store processed training data for explainability:
  trained_recipe <- prep(recipe, training = train_data, retain = TRUE)
  processed_train_data <- bake(trained_recipe, new_data = NULL)
  raw_train_data <- train_data

  if (learning_curve) {
    # Define the fractions to test
    fractions <- c(0.1, 0.2, 0.4, 0.6, 0.8, 1.0)

    # Helper function to run the learning curve step for a given fraction
    run_curve <- function(fraction) {
      if (verbose) message(sprintf("Learning curve: using %.0f%% of training data", fraction * 100))
      set.seed(seed)

      # Select training subset based on the fraction value
      sub_train <- if (fraction >= 1.0) {
        train_data
      } else {
        sub_split <- rsample::initial_split(
          train_data,
          prop = fraction,
          strata = if (task == "classification") label else NULL
        )
        rsample::training(sub_split)
      }

      n_sub <- nrow(sub_train)
      if (is.null(n_sub) || n_sub < 2) {
        return(data.frame(Fraction = fraction, Performance = NA_real_))
      }

      adjust_folds <- function(n_rows, folds_val) {
        if (is.null(folds_val)) {
          return(NULL)
        }
        folds_int <- suppressWarnings(as.integer(folds_val))
        if (!is.finite(folds_int)) {
          return(folds_val)
        }
        n_rows <- suppressWarnings(as.integer(n_rows))
        if (!is.finite(n_rows) || n_rows <= 2L) {
          return(NA_integer_)
        }
        max_allowed <- max(2L, n_rows - 1L)
        max(2L, min(folds_int, max_allowed))
      }

      sub_resampling_method <- resampling_method
      sub_folds <- folds
      sub_outer_folds <- outer_folds
      folds_adjusted <- FALSE
      outer_adjusted <- FALSE
      resampling_disabled <- FALSE
      if (!is.null(resampling_method)) {
        if (resampling_method %in% c("cv", "repeatedcv", "grouped_cv", "nested_cv")) {
          sub_folds <- adjust_folds(n_sub, folds)
          folds_adjusted <- !is.null(folds) && !is.na(sub_folds) &&
            !identical(as.integer(folds), as.integer(sub_folds))
        }
        if (resampling_method == "nested_cv") {
          sub_outer_folds <- adjust_folds(n_sub, outer_folds)
          outer_adjusted <- !is.null(outer_folds) && !is.na(sub_outer_folds) &&
            !identical(as.integer(outer_folds), as.integer(sub_outer_folds))
        }
      }
      if (resampling_method %in% c("cv", "repeatedcv", "grouped_cv") &&
          is.na(sub_folds)) {
        sub_resampling_method <- "none"
        sub_folds <- folds
        resampling_disabled <- TRUE
      }
      if (resampling_method == "nested_cv" &&
          (is.na(sub_folds) || is.na(sub_outer_folds))) {
        sub_resampling_method <- "none"
        sub_folds <- folds
        sub_outer_folds <- outer_folds
        resampling_disabled <- TRUE
      }
      if (resampling_disabled) {
        warning(
          sprintf(
            "Learning curve: disabling resampling for %.0f%% subset (n=%d) to avoid invalid CV settings.",
            fraction * 100, n_sub
          ),
          call. = FALSE
        )
      } else {
        if (folds_adjusted) {
          warning(
            sprintf(
              "Learning curve: reducing folds from %d to %d for %.0f%% subset (n=%d).",
              as.integer(folds), as.integer(sub_folds), fraction * 100, n_sub
            ),
            call. = FALSE
          )
        }
        if (outer_adjusted) {
          warning(
            sprintf(
              "Learning curve: reducing outer_folds from %d to %d for %.0f%% subset (n=%d).",
              as.integer(outer_folds), as.integer(sub_outer_folds), fraction * 100, n_sub
            ),
            call. = FALSE
          )
        }
      }

      # Train models on the subset
      sub_models <- withCallingHandlers(
        train_models(
          train_data = sub_train,
          label = label,
          task = task,
          algorithms = algorithms,
          resampling_method = sub_resampling_method,
          folds = sub_folds,
          repeats = repeats,
          group_cols = group_cols,
          block_col = block_col,
          block_size = block_size,
          initial_window = initial_window,
          assess_window = assess_window,
          skip = skip,
          outer_folds = sub_outer_folds,
          resamples = resamples,
          tune_params = tune_params,
          engine_params = engine_params,
          metric = metric,
          summaryFunction = summaryFunction,
          seed = seed,
          recipe = recipe,
          use_default_tuning = use_default_tuning,
          tuning_strategy = tuning_strategy,
          tuning_iterations = tuning_iterations,
          tuning_complexity = tuning_complexity,
          grid_levels = grid_levels,
          early_stopping = early_stopping,
          adaptive = adaptive,
          algorithm_engines = algorithm_engines,
          use_parsnip_defaults = use_parsnip_defaults,
          warn_engine_defaults = warn_engine_defaults,
          n_cores = n_cores,
          engine_threads = engine_threads,
          verbose = verbose,
          event_class = event_class,
          start_col = start_col,
          time_col = time_col,
          status_col = status_col,
          eval_times = eval_times,
          at_risk_threshold = at_risk_threshold,
          survival_metric_convention = survival_metric_convention,
          audit_env = audit_env,
          multiclass_auc = multiclass_auc
        ),
        warning = fastml_muffle_foreach_warning
      )



      # Evaluate models on the subset
      sub_eval <- evaluate_models(
        models = sub_models,
        train_data = sub_train,
        test_data = test_data,
        label = label_surv,
        start_col = start_col,
        time_col = time_col,
        status_col = status_col,
        task = task,
        metric = metric,
        event_class = event_class,
        class_threshold = class_threshold,
        eval_times = eval_times,
        bootstrap_ci = bootstrap_ci,
        bootstrap_samples = bootstrap_samples,
        bootstrap_seed = bootstrap_seed,
        at_risk_threshold = at_risk_threshold,
        survival_metric_convention = survival_metric_convention,
        multiclass_auc = multiclass_auc
      )


      # Extract the performance metric from each model evaluation
      perf_values <- sapply(sub_eval$performance, function(entry) {
        # entry may be a data frame (unwrapped single-engine) or a list of data frames
        perf_df <- if (inherits(entry, "data.frame")) entry else entry[[1]]
        val <- perf_df[perf_df$.metric == metric, ".estimate", drop = TRUE]
        if (length(val) == 0) NA_real_ else as.numeric(val[[1]])
      })

      # Compute the average performance (ignoring any missing values)
      avg_perf <- mean(perf_values, na.rm = TRUE)

      # Return a data frame for this fraction
      data.frame(Fraction = fraction, Performance = avg_perf)
    }

    # Apply the helper function to each fraction and combine the results
    df_lc <- do.call(rbind, lapply(fractions, run_curve))

    # Prepare the learning-curve plot without printing (avoid side effects)
    lc_plot <- ggplot(df_lc, aes(x = .data$Fraction, y = .data$Performance)) +
      geom_line(color = "blue") +
      geom_point(color = "blue") +
      labs(
        title = "Learning Curve",
        x = "Training Set Size (fraction)",
        y = paste("Mean", metric, "across models")
      ) +
      theme_minimal()
  } else {
    df_lc <- NULL
    lc_plot <- NULL
  }


  processed_test_data <- test_data
  if (!is.null(trained_recipe)) {
    processed_test_data <- tryCatch(
      {
        baked <- recipes::bake(trained_recipe, new_data = test_data)
        if (!is.null(label) && label %in% names(baked)) {
          baked <- baked[, setdiff(names(baked), label), drop = FALSE]
        }
        baked
      },
      # Training has already succeeded here, so this is reported rather than
      # fatal; but the stored copy is left unprocessed, and downstream use of
      # processed_test_data would otherwise silently mix the two representations.
      error = function(e) {
        warning(sprintf(
          paste0(
            "Could not apply the fitted recipe to the test data when assembling ",
            "the result object: %s\n",
            "`processed_test_data` therefore holds the unprocessed test data. ",
            "Reported model performance is unaffected, but treat that component ",
            "as raw rather than preprocessed."
          ),
          conditionMessage(e)
        ), call. = FALSE)
        test_data
      }
    )
  }

  result <- list(
    best_model = models[best_model_idx],
    best_model_name = best_model_name,
    performance = performance,
    predictions = predictions,
    preprocessor = trained_recipe,
    raw_train_data = raw_train_data,
    processed_train_data = processed_train_data,
    raw_test_data = test_data,
    processed_test_data = processed_test_data,
    label = label,
    task = task,
    models = models,
    metric = metric,
    class_threshold = class_threshold,
    multiclass_auc = multiclass_auc,
    selection = list(source = selection_source, metric = metric),
    positive_class = positive_class,
    event_class = event_class,
    engine_names = engine_names,
    learning_curve = list(
      data = df_lc,
      plot = lc_plot
    ),
    survival_brier_times = survival_brier_times,
    survival_t_max = survival_t_max,
    survival_metric_convention = survival_metric_convention,
    metric_bootstrap = list(enabled = bootstrap_ci, samples = bootstrap_samples, seed = bootstrap_seed),
    resampling_results = resampling_results,
    resampling_plan = resampling_plan,
    nested_cv = nested_results,
    audit = if (audit_env$enabled) list(log = audit_env$log, flagged = audit_env$unsafe) else NULL
  )
  class(result) <- "fastml"
  if (verbose) message("Training complete.")
  return(result)
}

library(testthat)

test_that("grouped holdout keeps groups intact", {
  skip_on_cran()
  required <- c("parsnip", "workflows", "recipes", "yardstick", "tune", "rsample", "dials")
  lapply(required, skip_if_not_installed)

  set.seed(123)
  grouped_data <- tibble::tibble(
    group = rep(letters[1:5], each = 4),
    x = rnorm(20),
    y = rnorm(20)
  )

  fit <- fastml(
    data = grouped_data,
    label = "y",
    algorithms = "linear_reg",
    task = "regression",
    resampling_method = "none",
    group_cols = "group",
    test_size = 0.2,
    use_default_tuning = FALSE,
    tuning_strategy = "none",
    seed = 123
  )

  train_groups <- unique(as.character(fit$raw_train_data$group))
  test_groups <- unique(as.character(fit$raw_test_data$group))
  expect_length(intersect(train_groups, test_groups), 0)
})

test_that("time-ordered holdout uses tail rows", {
  skip_on_cran()
  required <- c("parsnip", "workflows", "recipes", "yardstick", "tune", "rsample", "dials")
  lapply(required, skip_if_not_installed)

  time_data <- tibble::tibble(
    time_id = 1:20,
    x = rnorm(20),
    y = rnorm(20)
  )

  fit <- fastml(
    data = time_data,
    label = "y",
    algorithms = "linear_reg",
    task = "regression",
    resampling_method = "none",
    block_col = "time_id",
    test_size = 0.2,
    use_default_tuning = FALSE,
    tuning_strategy = "none",
    seed = 123
  )

  expect_equal(max(fit$raw_train_data$time_id), 16)
  expect_equal(min(fit$raw_test_data$time_id), 17)
})

library(testthat)
library(dplyr)

# small dataset for tests
set.seed(123)
data(iris)
iris_binary <- iris[iris$Species != "setosa", ]
iris_binary$Species <- factor(iris_binary$Species)

# ========================
# Input Validation Tests
# ========================

test_that("fastml errors when label column doesn't exist", {
  expect_error(
    fastml(
      data = iris_binary,
      label = "nonexistent_column",
      algorithms = "logistic_reg"
    ),
    "Label variable must exist|not found|not present|does not exist"
  )
})

test_that("fastml errors with empty data", {
  expect_error(
    fastml(
      data = iris_binary[0, ],
      label = "Species",
      algorithms = "logistic_reg"
    )
  )
})

test_that("fastml errors with invalid algorithm name", {
  expect_error(
    fastml(
      data = iris_binary,
      label = "Species",
      algorithms = "nonexistent_algorithm"
    )
  )
})

test_that("fastml errors with invalid resampling method", {
  expect_error(
    fastml(
      data = iris_binary,
      label = "Species",
      algorithms = "logistic_reg",
      resampling_method = "invalid_method"
    )
  )
})

# ========================
# Task Detection Tests
# ========================

test_that("fastml detects classification task from factor label", {
  skip_on_cran()

  model <- suppressWarnings(fastml(
    data = iris_binary,
    label = "Species",
    algorithms = "logistic_reg",
    resampling_method = "none",
    use_default_tuning = FALSE
  ))

  expect_equal(model$task, "classification")
})

test_that("fastml detects regression task from numeric label", {
  skip_on_cran()

  data(mtcars)
  model <- fastml(
    data = mtcars,
    label = "mpg",
    algorithms = "linear_reg",
    resampling_method = "none",
    use_default_tuning = FALSE
  )

  expect_equal(model$task, "regression")
})

test_that("fastml uses explicitly specified task", {
  skip_on_cran()

  model <- suppressWarnings(fastml(
    data = iris_binary,
    label = "Species",
    algorithms = "logistic_reg",
    task = "classification",
    resampling_method = "none",
    use_default_tuning = FALSE
  ))

  expect_equal(model$task, "classification")
})

# ========================
# Data Splitting Tests
# ========================

test_that("fastml respects test_size parameter", {
  skip_on_cran()

  model <- suppressWarnings(fastml(
    data = iris_binary,
    label = "Species",
    algorithms = "logistic_reg",
    resampling_method = "none",
    use_default_tuning = FALSE,
    test_size = 0.3
  ))

  # Check that raw_train_data and raw_test_data exist
  expect_true(!is.null(model$raw_train_data))
  expect_true(!is.null(model$raw_test_data))
})

test_that("fastml uses provided train_data and test_data", {
  skip_on_cran()

  set.seed(123)
  idx <- sample(nrow(iris_binary), floor(0.7 * nrow(iris_binary)))
  train_data <- iris_binary[idx, ]
  test_data <- iris_binary[-idx, ]

  model <- suppressWarnings(fastml(
    train_data = train_data,
    test_data = test_data,
    label = "Species",
    algorithms = "logistic_reg",
    resampling_method = "none",
    use_default_tuning = FALSE
  ))

  expect_equal(nrow(model$raw_train_data), nrow(train_data))
  expect_equal(nrow(model$raw_test_data), nrow(test_data))
})

# ========================
# Resampling Tests
# ========================

test_that("fastml works with cv resampling", {
  skip_on_cran()

  model <- suppressWarnings(fastml(
    data = iris_binary,
    label = "Species",
    algorithms = "logistic_reg",
    resampling_method = "cv",
    folds = 3,
    use_default_tuning = FALSE
  ))

  expect_true(inherits(model, "fastml"))
})

test_that("fastml works with none resampling", {
  skip_on_cran()

  model <- suppressWarnings(fastml(
    data = iris_binary,
    label = "Species",
    algorithms = "logistic_reg",
    resampling_method = "none",
    use_default_tuning = FALSE
  ))

  expect_true(inherits(model, "fastml"))
})

# ========================
# Multiple Algorithms Tests
# ========================

test_that("fastml trains multiple algorithms", {
  skip_on_cran()

  model <- suppressWarnings(fastml(
    data = iris_binary,
    label = "Species",
    algorithms = c("logistic_reg", "rand_forest"),
    resampling_method = "none",
    use_default_tuning = FALSE
  ))

  expect_true(length(model$models) >= 2)
})

test_that("fastml returns best model among multiple", {
  skip_on_cran()

  model <- suppressWarnings(fastml(
    data = iris_binary,
    label = "Species",
    algorithms = c("logistic_reg", "rand_forest"),
    resampling_method = "none",
    use_default_tuning = FALSE
  ))

  expect_true(!is.null(model$best_model))
  expect_true(!is.null(model$best_model_name))
})

# ========================
# Metric Tests
# ========================

test_that("fastml uses specified metric for classification", {
  skip_on_cran()

  model <- suppressWarnings(fastml(
    data = iris_binary,
    label = "Species",
    algorithms = "logistic_reg",
    metric = "accuracy",
    resampling_method = "none",
    use_default_tuning = FALSE
  ))

  expect_equal(model$metric, "accuracy")
})

test_that("fastml uses specified metric for regression", {
  skip_on_cran()

  data(mtcars)
  model <- fastml(
    data = mtcars,
    label = "mpg",
    algorithms = "linear_reg",
    metric = "rmse",
    resampling_method = "none",
    use_default_tuning = FALSE
  )

  expect_equal(model$metric, "rmse")
})

# ========================
# Output Structure Tests
# ========================

test_that("fastml returns complete object structure", {
  skip_on_cran()

  model <- suppressWarnings(fastml(
    data = iris_binary,
    label = "Species",
    algorithms = "logistic_reg",
    resampling_method = "none",
    use_default_tuning = FALSE
  ))

  expect_true("models" %in% names(model))
  expect_true("performance" %in% names(model))
  expect_true("predictions" %in% names(model))
  expect_true("best_model" %in% names(model))
  expect_true("label" %in% names(model))
  expect_true("task" %in% names(model))
  expect_true("preprocessor" %in% names(model))
})

test_that("fastml performance contains expected metrics for classification", {
  skip_on_cran()

  model <- suppressWarnings(fastml(
    data = iris_binary,
    label = "Species",
    algorithms = "logistic_reg",
    resampling_method = "none",
    use_default_tuning = FALSE
  ))

  perf <- model$performance[[1]]
  if (is.list(perf) && !is.data.frame(perf)) {
    perf <- perf[[1]]
  }

  expect_true("accuracy" %in% perf$.metric || "roc_auc" %in% perf$.metric)
})

test_that("fastml performance contains expected metrics for regression", {
  skip_on_cran()

  data(mtcars)
  model <- fastml(
    data = mtcars,
    label = "mpg",
    algorithms = "linear_reg",
    resampling_method = "none",
    use_default_tuning = FALSE
  )

  perf <- model$performance[[1]]
  if (is.list(perf) && !is.data.frame(perf)) {
    perf <- perf[[1]]
  }

  expect_true("rmse" %in% perf$.metric || "rsq" %in% perf$.metric)
})

# ========================
# Seed Reproducibility Tests
# ========================

test_that("fastml results are reproducible with same seed", {
  skip_on_cran()

  model1 <- suppressWarnings(fastml(
    data = iris_binary,
    label = "Species",
    algorithms = "logistic_reg",
    resampling_method = "none",
    use_default_tuning = FALSE,
    seed = 42
  ))

  model2 <- suppressWarnings(fastml(
    data = iris_binary,
    label = "Species",
    algorithms = "logistic_reg",
    resampling_method = "none",
    use_default_tuning = FALSE,
    seed = 42
  ))

  # Performance should be identical with same seed
  perf1 <- model1$performance[[1]]
  perf2 <- model2$performance[[1]]

  if (is.list(perf1) && !is.data.frame(perf1)) {
    perf1 <- perf1[[1]]
  }
  if (is.list(perf2) && !is.data.frame(perf2)) {
    perf2 <- perf2[[1]]
  }

  expect_equal(perf1$.estimate, perf2$.estimate)
})

# ========================
# Special Data Types Tests
# ========================

test_that("fastml handles character columns by converting to factor", {
  skip_on_cran()

  df <- iris_binary
  df$char_col <- sample(c("A", "B"), nrow(df), replace = TRUE)

  model <- suppressWarnings(fastml(
    data = df,
    label = "Species",
    algorithms = "logistic_reg",
    resampling_method = "none",
    use_default_tuning = FALSE
  ))

  expect_true(inherits(model, "fastml"))
})

test_that("fastml handles data with NAs via imputation", {
  skip_on_cran()

  df <- iris_binary
  df$Sepal.Length[1:5] <- NA

  model <- suppressWarnings(fastml(
    data = df,
    label = "Species",
    algorithms = "logistic_reg",
    resampling_method = "none",
    use_default_tuning = FALSE,
    impute_method = "medianImpute"
  ))

  expect_true(inherits(model, "fastml"))
})

# ========================
# Event Class Tests
# ========================

test_that("fastml uses specified event_class", {
  skip_on_cran()

  model <- suppressWarnings(fastml(
    data = iris_binary,
    label = "Species",
    algorithms = "logistic_reg",
    event_class = "first",
    resampling_method = "none",
    use_default_tuning = FALSE
  ))

  expect_equal(model$event_class, "first")
})

# ========================
# Verbose Tests
# ========================

test_that("fastml verbose=TRUE produces output", {
  skip_on_cran()

  output <- capture.output(type = "message", {
    model <- suppressWarnings(fastml(
      data = iris_binary,
      label = "Species",
      algorithms = "logistic_reg",
      resampling_method = "none",
      use_default_tuning = FALSE,
      verbose = TRUE
    ))
  })

  expect_true(length(output) > 0)
})

test_that("fastml verbose=FALSE suppresses output", {
  skip_on_cran()

  output <- capture.output({
    model <- suppressMessages(suppressWarnings(fastml(
      data = iris_binary,
      label = "Species",
      algorithms = "logistic_reg",
      resampling_method = "none",
      use_default_tuning = FALSE,
      verbose = FALSE
    )))
  })

  # Verbose=FALSE should produce minimal/no output
  expect_true(length(output) == 0 || all(output == ""))
})

library(testthat)
library(dplyr)

skip_if_not_installed("iml")

# small dataset for tests
set.seed(123)
data(iris)
iris_binary <- iris[iris$Species != "setosa", ]
iris_binary$Species <- factor(iris_binary$Species)

test_that("surrogate_tree errors with non-fastml object", {
  expect_error(
    surrogate_tree(list(a = 1)),
    "must be a 'fastml' object"
  )
})

test_that("surrogate_tree runs on classification model", {
  skip_on_cran()

  model <- fastml(
    data = iris_binary,
    label = "Species",
    algorithms = "rand_forest",
    resampling_method = "cv",
    folds = 2
  )

  result <- suppressWarnings(suppressMessages(surrogate_tree(model)))
  expect_true(inherits(result, "TreeSurrogate"))
})

test_that("surrogate_tree respects maxdepth parameter", {
  skip_on_cran()

  model <- fastml(
    data = iris_binary,
    label = "Species",
    algorithms = "rand_forest",
    resampling_method = "cv",
    folds = 2
  )

  result <- suppressWarnings(suppressMessages(surrogate_tree(model, maxdepth = 2)))
  expect_true(inherits(result, "TreeSurrogate"))
})

test_that("surrogate_tree runs on regression model", {
  skip_on_cran()

  set.seed(42)
  data(mtcars)

  model <- fastml(
    data = mtcars,
    label = "mpg",
    algorithms = "rand_forest",
    resampling_method = "cv",
    folds = 2
  )

  result <- suppressWarnings(suppressMessages(surrogate_tree(model)))
  expect_true(inherits(result, "TreeSurrogate"))
})

library(testthat)
library(dplyr)

skip_if_not_installed("iml")

# small dataset for tests
set.seed(123)
data(iris)
iris_binary <- iris[iris$Species != "setosa", ]
iris_binary$Species <- factor(iris_binary$Species)

test_that("interaction_strength errors with non-fastml object", {
  expect_error(
    interaction_strength(list(a = 1)),
    "must be a 'fastml' object"
  )
})

test_that("interaction_strength runs on classification model", {
  skip_on_cran()

  model <- fastml(
    data = iris_binary,
    label = "Species",
    algorithms = "rand_forest",
    resampling_method = "cv",
    folds = 2
  )

  expect_error(
    suppressWarnings(suppressMessages(interaction_strength(model))),
    regexp = NA
  )
})

test_that("interaction_strength runs on regression model", {
  skip_on_cran()

  set.seed(42)
  data(mtcars)

  model <- fastml(
    data = mtcars,
    label = "mpg",
    algorithms = "rand_forest",
    resampling_method = "cv",
    folds = 2
  )

  result <- suppressWarnings(suppressMessages(interaction_strength(model)))
  expect_true(inherits(result, "Interaction"))
})

library(testthat)

test_that("custom summaryFunction metric is included in holdout performance", {
  skip_on_cran()
  required <- c("parsnip", "workflows", "recipes", "yardstick", "tune", "rsample", "dials")
  lapply(required, skip_if_not_installed)

  data <- iris[iris$Species != "setosa", , drop = FALSE]
  data$Species <- factor(data$Species)
  # Reduce separation risk in glm for stable fits.
  data <- data[, c("Sepal.Length", "Sepal.Width", "Species")]

  custom_metric <- function(data, truth, estimate, ...) {
    truth <- rlang::enquo(truth)
    estimate <- rlang::enquo(estimate)
    truth_vec <- dplyr::pull(data, !!truth)
    estimate_vec <- dplyr::pull(data, !!estimate)
    tibble::tibble(
      .metric = "custom_metric",
      .estimator = "binary",
      .estimate = mean(truth_vec == estimate_vec, na.rm = TRUE)
    )
  }

  fit <- fastml(
    data = data,
    label = "Species",
    algorithms = "logistic_reg",
    resampling_method = "none",
    use_default_tuning = FALSE,
    tuning_strategy = "none",
    summaryFunction = custom_metric,
    metric = "custom_metric",
    seed = 123
  )

  perf <- fit$performance[[1]]
  if (is.list(perf) && !is.data.frame(perf)) {
    perf <- perf[[1]]
  }
  expect_true("custom_metric" %in% perf$.metric)
})

library(testthat)
library(dplyr)

# small dataset for tests
set.seed(123)
data(iris)
iris_binary <- iris[iris$Species != "setosa", ]
iris_binary$Species <- factor(iris_binary$Species)

test_that("predict.fastml errors when newdata is missing", {
  skip_on_cran()

  model <- suppressWarnings(fastml(
    data = iris_binary,
    label = "Species",
    algorithms = "logistic_reg",
    resampling_method = "none",
    use_default_tuning = FALSE
  ))

  expect_error(
    predict(model),
    "Please provide newdata for prediction"
  )
})

test_that("predict.fastml returns class predictions for classification", {
  skip_on_cran()

  model <- suppressWarnings(fastml(
    data = iris_binary,
    label = "Species",
    algorithms = "logistic_reg",
    resampling_method = "none",
    use_default_tuning = FALSE
  ))

  new_data <- iris_binary[1:10, ]
  new_data$Species <- NULL

  preds <- predict(model, newdata = new_data, type = "class")

  expect_true(is.factor(preds) || is.character(preds))
  expect_equal(length(preds), 10)
})

test_that("predict.fastml returns probabilities for classification", {
  skip_on_cran()

  model <- suppressWarnings(fastml(
    data = iris_binary,
    label = "Species",
    algorithms = "logistic_reg",
    resampling_method = "none",
    use_default_tuning = FALSE
  ))

  new_data <- iris_binary[1:10, ]
  new_data$Species <- NULL

  preds <- predict(model, newdata = new_data, type = "prob")

  expect_true(is.data.frame(preds) || tibble::is_tibble(preds))
  expect_equal(nrow(preds), 10)
  expect_true(any(grepl("^\\.pred_", names(preds))))
})

test_that("predict.fastml returns numeric predictions for regression", {
  skip_on_cran()

  data(mtcars)
  model <- fastml(
    data = mtcars,
    label = "mpg",
    algorithms = "linear_reg",
    resampling_method = "none",
    use_default_tuning = FALSE
  )

  new_data <- mtcars[1:5, ]
  new_data$mpg <- NULL

  preds <- predict(model, newdata = new_data, type = "numeric")

  expect_true(is.numeric(preds))
  expect_equal(length(preds), 5)
})

test_that("predict.fastml auto type selects appropriately for classification", {
  skip_on_cran()

  model <- suppressWarnings(fastml(
    data = iris_binary,
    label = "Species",
    algorithms = "logistic_reg",
    resampling_method = "none",
    use_default_tuning = FALSE
  ))

  new_data <- iris_binary[1:5, ]
  new_data$Species <- NULL

  preds <- predict(model, newdata = new_data, type = "auto")

  expect_true(is.factor(preds) || is.character(preds))
})

test_that("predict.fastml auto type selects appropriately for regression", {
  skip_on_cran()

  data(mtcars)
  model <- fastml(
    data = mtcars,
    label = "mpg",
    algorithms = "linear_reg",
    resampling_method = "none",
    use_default_tuning = FALSE
  )

  new_data <- mtcars[1:5, ]
  new_data$mpg <- NULL

  preds <- predict(model, newdata = new_data, type = "auto")

  expect_true(is.numeric(preds))
})

test_that("predict.fastml applies postprocess_fn", {
  skip_on_cran()

  data(mtcars)
  model <- fastml(
    data = mtcars,
    label = "mpg",
    algorithms = "linear_reg",
    resampling_method = "none",
    use_default_tuning = FALSE
  )

  new_data <- mtcars[1:5, ]
  new_data$mpg <- NULL

  # Postprocess function that rounds predictions
  postprocess <- function(x) round(x)

  preds <- predict(model, newdata = new_data, postprocess_fn = postprocess)

  expect_true(all(preds == round(preds)))
})

test_that("predict.fastml handles model_name parameter", {
  skip_on_cran()

  model <- suppressWarnings(fastml(
    data = iris_binary,
    label = "Species",
    algorithms = c("logistic_reg", "rand_forest"),
    resampling_method = "none",
    use_default_tuning = FALSE
  ))

  new_data <- iris_binary[1:5, ]
  new_data$Species <- NULL

  # Predict using specific model
  preds <- predict(model, newdata = new_data, model_name = "logistic_reg")

  expect_true(is.factor(preds) || is.character(preds))
  expect_equal(length(preds), 5)
})

test_that("predict.fastml errors on invalid model_name", {
  skip_on_cran()

  model <- suppressWarnings(fastml(
    data = iris_binary,
    label = "Species",
    algorithms = "logistic_reg",
    resampling_method = "none",
    use_default_tuning = FALSE
  ))

  new_data <- iris_binary[1:5, ]
  new_data$Species <- NULL

  expect_error(
    predict(model, newdata = new_data, model_name = "nonexistent_model"),
    "not found"
  )
})

test_that("predict.fastml verbose prints messages", {
  skip_on_cran()

  model <- suppressWarnings(fastml(
    data = iris_binary,
    label = "Species",
    algorithms = "logistic_reg",
    resampling_method = "none",
    use_default_tuning = FALSE
  ))

  new_data <- iris_binary[1:5, ]
  new_data$Species <- NULL

  expect_message(
    predict(model, newdata = new_data, verbose = TRUE),
    "Predicting with"
  )
})

test_that("predict.fastml warns when requesting prob for regression", {
  skip_on_cran()

  data(mtcars)
  model <- fastml(
    data = mtcars,
    label = "mpg",
    algorithms = "linear_reg",
    resampling_method = "none",
    use_default_tuning = FALSE
  )

  new_data <- mtcars[1:5, ]
  new_data$mpg <- NULL

  expect_warning(
    predict(model, newdata = new_data, type = "prob"),
    "Probabilities only for classification"
  )
})

test_that("predict.fastml handles integer columns in newdata", {
  skip_on_cran()

  # Create dataset with integer columns
  df <- data.frame(
    x1 = as.integer(1:50),
    x2 = as.integer(50:1),
    y = factor(rep(c("A", "B"), 25))
  )

  model <- suppressWarnings(fastml(
    data = df,
    label = "y",
    algorithms = "logistic_reg",
    resampling_method = "none",
    use_default_tuning = FALSE
  ))

  new_data <- data.frame(
    x1 = as.integer(1:5),
    x2 = as.integer(5:1)
  )

  # Suppress rank-deficient prediction warnings
  preds <- suppressWarnings(predict(model, newdata = new_data))

  expect_equal(length(preds), 5)
})

library(testthat)
library(dplyr)

# small dataset for tests
set.seed(123)
data(iris)
iris_binary <- iris[iris$Species != "setosa", ]
iris_binary$Species <- factor(iris_binary$Species)

test_that("plot.fastml errors with non-fastml object", {
  expect_error(
    plot.fastml(list(a = 1)),
    "must be a 'fastml' object"
  )
})

test_that("plot.fastml type='bar' works for classification", {
  skip_on_cran()

  model <- suppressWarnings(fastml(
    data = iris_binary,
    label = "Species",
    algorithms = "logistic_reg",
    resampling_method = "none",
    use_default_tuning = FALSE
  ))

  # Use pdf device to capture plot
  grDevices::pdf(file = NULL)
  on.exit(grDevices::dev.off(), add = TRUE)

  expect_error(suppressWarnings(plot(model, type = "bar")), NA)
})

test_that("plot.fastml type='bar' works for regression", {
  skip_on_cran()

  data(mtcars)
  model <- suppressWarnings(fastml(
    data = mtcars,
    label = "mpg",
    algorithms = "linear_reg",
    resampling_method = "none",
    use_default_tuning = FALSE
  ))

  grDevices::pdf(file = NULL)
  on.exit(grDevices::dev.off(), add = TRUE)

  expect_error(suppressWarnings(plot(model, type = "bar")), NA)
})

test_that("plot.fastml type='roc' works for binary classification", {
  skip_on_cran()

  model <- suppressWarnings(fastml(
    data = iris_binary,
    label = "Species",
    algorithms = "logistic_reg",
    resampling_method = "none",
    use_default_tuning = FALSE
  ))

  grDevices::pdf(file = NULL)
  on.exit(grDevices::dev.off(), add = TRUE)

  expect_error(suppressWarnings(plot(model, type = "roc")), NA)
})

test_that("plot.fastml type='calibration' works for classification", {
  skip_on_cran()
  skip_if_not_installed("probably")

  model <- suppressWarnings(fastml(
    data = iris_binary,
    label = "Species",
    algorithms = "logistic_reg",
    resampling_method = "none",
    use_default_tuning = FALSE
  ))

  grDevices::pdf(file = NULL)
  on.exit(grDevices::dev.off(), add = TRUE)

  expect_error(suppressWarnings(plot(model, type = "calibration")), NA)
})

test_that("plot.fastml type='residual' works for regression", {
  skip_on_cran()

  data(mtcars)
  model <- suppressWarnings(fastml(
    data = mtcars,
    label = "mpg",
    algorithms = "linear_reg",
    resampling_method = "none",
    use_default_tuning = FALSE
  ))

  grDevices::pdf(file = NULL)
  on.exit(grDevices::dev.off(), add = TRUE)

  output <- capture.output(suppressWarnings(plot(model, type = "residual")))
  expect_true(TRUE)  # If we get here without error, test passes
})

test_that("plot.fastml type='all' generates multiple plots", {
  skip_on_cran()

  model <- suppressWarnings(fastml(
    data = iris_binary,
    label = "Species",
    algorithms = "logistic_reg",
    resampling_method = "none",
    use_default_tuning = FALSE
  ))

  grDevices::pdf(file = NULL)
  on.exit(grDevices::dev.off(), add = TRUE)

  expect_error(suppressWarnings(plot(model, type = "all")), NA)
})

test_that("plot.fastml returns invisible object", {
  skip_on_cran()

  model <- suppressWarnings(fastml(
    data = iris_binary,
    label = "Species",
    algorithms = "logistic_reg",
    resampling_method = "none",
    use_default_tuning = FALSE
  ))

  grDevices::pdf(file = NULL)
  on.exit(grDevices::dev.off(), add = TRUE)

  result <- suppressWarnings(plot(model, type = "bar"))
  expect_s3_class(result, "fastml")
})

test_that("plot.fastml handles multiple algorithms", {
  skip_on_cran()

  model <- suppressWarnings(fastml(
    data = iris_binary,
    label = "Species",
    algorithms = c("logistic_reg", "rand_forest"),
    resampling_method = "none",
    use_default_tuning = FALSE
  ))

  grDevices::pdf(file = NULL)
  on.exit(grDevices::dev.off(), add = TRUE)

  expect_error(suppressWarnings(plot(model, type = "bar")), NA)
})

test_that("plot.fastml algorithm parameter filters ROC curves", {
  skip_on_cran()

  model <- suppressWarnings(fastml(
    data = iris_binary,
    label = "Species",
    algorithms = c("logistic_reg", "rand_forest"),
    resampling_method = "none",
    use_default_tuning = FALSE
  ))

  grDevices::pdf(file = NULL)
  on.exit(grDevices::dev.off(), add = TRUE)

  expect_error(suppressWarnings(plot(model, type = "roc", algorithm = "logistic_reg")), NA)
})

test_that("plot.fastml prints message for residual on classification", {
  skip_on_cran()

  model <- suppressWarnings(fastml(
    data = iris_binary,
    label = "Species",
    algorithms = "logistic_reg",
    resampling_method = "none",
    use_default_tuning = FALSE
  ))

  grDevices::pdf(file = NULL)
  on.exit(grDevices::dev.off(), add = TRUE)

  output <- capture.output(suppressWarnings(plot(model, type = "residual")))
  expect_true(any(grepl("regression", output, ignore.case = TRUE)))
})

test_that("plot.fastml prints message for calibration on regression", {
  skip_on_cran()

  data(mtcars)
  model <- suppressWarnings(fastml(
    data = mtcars,
    label = "mpg",
    algorithms = "linear_reg",
    resampling_method = "none",
    use_default_tuning = FALSE
  ))

  grDevices::pdf(file = NULL)
  on.exit(grDevices::dev.off(), add = TRUE)

  output <- capture.output(suppressWarnings(plot(model, type = "calibration")))
  expect_true(any(grepl("classification", output, ignore.case = TRUE)))
})

test_that("plot.fastml handles learning_curve type gracefully", {
  skip_on_cran()

  model <- suppressWarnings(fastml(
    data = iris_binary,
    label = "Species",
    algorithms = "logistic_reg",
    resampling_method = "none",
    use_default_tuning = FALSE
  ))

  grDevices::pdf(file = NULL)
  on.exit(grDevices::dev.off(), add = TRUE)

  output <- capture.output(suppressWarnings(plot(model, type = "learning_curve")))
  expect_true(any(grepl("learning curve", output, ignore.case = TRUE)))
})

library(testthat)
library(dplyr)

# small dataset for tests
set.seed(123)
data(iris)
iris_binary <- iris[iris$Species != "setosa", ]
iris_binary$Species <- factor(iris_binary$Species)

test_that("summary.fastml errors with non-fastml object", {
  expect_error(
    summary.fastml(list(a = 1)),
    "must be a 'fastml' object"
  )
})

test_that("summary.fastml runs for classification without error", {
  skip_on_cran()

  model <- suppressWarnings(fastml(
    data = iris_binary,
    label = "Species",
    algorithms = "logistic_reg",
    resampling_method = "none",
    use_default_tuning = FALSE
  ))

  output <- capture.output(summary(model))
  expect_true(length(output) > 0)
  expect_true(any(grepl("fastml Model Summary", output)))
  expect_true(any(grepl("classification", output, ignore.case = TRUE)))
})

test_that("summary.fastml runs for regression without error", {
  skip_on_cran()

  data(mtcars)
  model <- suppressWarnings(fastml(
    data = mtcars,
    label = "mpg",
    algorithms = "linear_reg",
    resampling_method = "none",
    use_default_tuning = FALSE
  ))

  output <- capture.output(summary(model))
  expect_true(length(output) > 0)
  expect_true(any(grepl("fastml Model Summary", output)))
  expect_true(any(grepl("regression", output, ignore.case = TRUE)))
})

test_that("summary.fastml type='metrics' shows only metrics", {
  skip_on_cran()

  model <- suppressWarnings(fastml(
    data = iris_binary,
    label = "Species",
    algorithms = "logistic_reg",
    resampling_method = "none",
    use_default_tuning = FALSE
  ))

  output <- capture.output(summary(model, type = "metrics"))
  expect_true(length(output) > 0)
  expect_true(any(grepl("Metric|Performance", output, ignore.case = TRUE)))
})

test_that("summary.fastml type='params' shows hyperparameters", {
  skip_on_cran()

  model <- suppressWarnings(fastml(
    data = iris_binary,
    label = "Species",
    algorithms = "logistic_reg",
    resampling_method = "none",
    use_default_tuning = FALSE
  ))

  output <- capture.output(summary(model, type = "params"))
  expect_true(length(output) > 0)
  expect_true(any(grepl("hyperparameter|param", output, ignore.case = TRUE)))
})

test_that("summary.fastml type='conf_mat' shows confusion matrix", {
  skip_on_cran()

  model <- suppressWarnings(fastml(
    data = iris_binary,
    label = "Species",
    algorithms = "logistic_reg",
    resampling_method = "none",
    use_default_tuning = FALSE
  ))

  output <- capture.output(summary(model, type = "conf_mat"))
  expect_true(length(output) > 0)
  expect_true(any(grepl("Confusion|Matrix", output, ignore.case = TRUE)))
})

test_that("summary.fastml type='all' shows all sections", {
  skip_on_cran()

  model <- suppressWarnings(fastml(
    data = iris_binary,
    label = "Species",
    algorithms = "logistic_reg",
    resampling_method = "none",
    use_default_tuning = FALSE
  ))

  output <- capture.output(summary(model, type = "all"))
  expect_true(length(output) > 0)
  expect_true(any(grepl("Metric|Performance", output, ignore.case = TRUE)))
  expect_true(any(grepl("hyperparameter|param", output, ignore.case = TRUE)))
})

test_that("summary.fastml algorithm parameter filters models", {
  skip_on_cran()

  model <- suppressWarnings(fastml(
    data = iris_binary,
    label = "Species",
    algorithms = c("logistic_reg", "rand_forest"),
    resampling_method = "none",
    use_default_tuning = FALSE
  ))

  output <- capture.output(summary(model, algorithm = "logistic_reg", type = "metrics"))
  expect_true(length(output) > 0)
  # Should mention logistic_reg in output
  expect_true(any(grepl("logistic", output, ignore.case = TRUE)))
})

test_that("summary.fastml errors on invalid algorithm", {
  skip_on_cran()

  model <- suppressWarnings(fastml(
    data = iris_binary,
    label = "Species",
    algorithms = "logistic_reg",
    resampling_method = "none",
    use_default_tuning = FALSE
  ))

  expect_error(
    summary(model, algorithm = "nonexistent"),
    "algorithms entered correctly"
  )
})

test_that("summary.fastml returns invisible object", {
  skip_on_cran()

  model <- suppressWarnings(fastml(
    data = iris_binary,
    label = "Species",
    algorithms = "logistic_reg",
    resampling_method = "none",
    use_default_tuning = FALSE
  ))

  result <- capture.output(res <- summary(model))
  expect_s3_class(res, "fastml")
})

test_that("summary.fastml handles multiple algorithms", {
  skip_on_cran()

  model <- suppressWarnings(fastml(
    data = iris_binary,
    label = "Species",
    algorithms = c("logistic_reg", "rand_forest"),
    resampling_method = "none",
    use_default_tuning = FALSE
  ))

  output <- capture.output(summary(model))
  expect_true(any(grepl("logistic", output, ignore.case = TRUE)))
  expect_true(any(grepl("rand_forest|forest", output, ignore.case = TRUE)))
})

test_that("summary.fastml shows correct number of models", {
  skip_on_cran()

  model <- suppressWarnings(fastml(
    data = iris_binary,
    label = "Species",
    algorithms = c("logistic_reg", "rand_forest"),
    resampling_method = "none",
    use_default_tuning = FALSE
  ))

  output <- capture.output(summary(model))
  expect_true(any(grepl("2", output)))  # 2 models
})

test_that("summary.fastml works with sort_metric parameter", {
  skip_on_cran()

  model <- suppressWarnings(fastml(
    data = iris_binary,
    label = "Species",
    algorithms = c("logistic_reg", "rand_forest"),
    resampling_method = "none",
    use_default_tuning = FALSE
  ))

  output <- capture.output(summary(model, sort_metric = "accuracy"))
  expect_true(length(output) > 0)
})

test_that("summary.fastml errors on invalid sort_metric", {
  skip_on_cran()

  model <- suppressWarnings(fastml(
    data = iris_binary,
    label = "Species",
    algorithms = "logistic_reg",
    resampling_method = "none",
    use_default_tuning = FALSE
  ))

  expect_error(
    summary(model, sort_metric = "nonexistent_metric"),
    "Invalid sort_metric"
  )
})

#' Internal predict_model method for parsnip fits
#'
#' Shim for parsnip model objects so that lime's predict_model generic ignores
#' unused arguments passed via `...`.
#' @keywords internal
#' @export
predict_model.model_fit <- function(x, newdata, type, ...) {
  # LIME passes extra args that parsnip doesn't like; ignore them
  predict(x, new_data = newdata, type = type)
}

#' Generate LIME explanations for a fastml model
#'
#' Creates a `lime` explainer using processed (encoded, scaled) data and returns
#' feature explanations for new observations. The new observation is automatically
#' preprocessed using the same recipe to ensure alignment with the explainer background.
#'
#' @param object A `fastml` object.
#' @param new_observation A data frame containing the new observation(s) to explain.
#'   Must contain the same columns as the original training data (before preprocessing).
#'   The function will apply the stored preprocessor to transform it.
#' @param n_features Number of features to show in the explanation. Default 5.
#' @param n_labels Number of labels to explain (classification only). Default 1.
#' @param ... Additional arguments passed to `lime::explain`.
#'
#' @param data Character string specifying which data to use for the LIME explainer background:
#'   \code{"train"} (default) or \code{"test"}.
#' @return An object produced by `lime::explain`.
#' @importFrom recipes bake
#' @export
#' @examples
#' \dontrun{
#' data(iris)
#' iris <- iris[iris$Species != "setosa", ]
#' iris$Species <- factor(iris$Species)
#' model <- fastml(data = iris, label = "Species")
#' explain_lime(model, new_observation = iris[1, ])
#' }
explain_lime <- function(object, new_observation, data = c("train", "test"), n_features = 5, n_labels = 1, ...) {
  if (!inherits(object, "fastml")) {
    stop("The input must be a 'fastml' object.")
  }
  if (missing(new_observation)) {
    stop("'new_observation' must be provided for LIME.")
  }
  if (!requireNamespace("lime", quietly = TRUE)) {
    stop("The 'lime' package is required for LIME explanations.")
  }

  # 1. Get data to build the explainer (train or test based on parameter)
  data <- match.arg(data)
  prep <- fastml_prepare_explainer_inputs(object, data = data)
  if (length(prep$fits) == 0) {
    stop("Unable to extract parsnip model for LIME.")
  }

  # 2. Use PROCESSED data for explainer background (this is what the model sees)
 background_x <- prep$x_processed
  expected_cols <- colnames(background_x)

  # 3. Preprocess the new observation to match the processed feature space
  preprocess_failed <- FALSE
  if (!is.null(object$preprocessor)) {
    to_explain <- tryCatch(
      {
        baked <- recipes::bake(object$preprocessor, new_data = new_observation)
        # Remove label if present
        if (!is.null(object$label) && object$label %in% names(baked)) {
          baked[[object$label]] <- NULL
        }
        baked
      },
      error = function(e) {
        preprocess_failed <<- TRUE
        NULL
      }
    )
  } else {
    # No preprocessor - use raw observation but remove label
    to_explain <- new_observation
    if (!is.null(object$label) && object$label %in% names(to_explain)) {
      to_explain[[object$label]] <- NULL
    }
  }

  # 4. Validate alignment between background and observation
 if (preprocess_failed || is.null(to_explain)) {
    stop(
      "LIME preprocessing failed for the new observation. ",
      "Ensure the observation has the same columns as the training data.",
      call. = FALSE
    )
  }

  # Check for column mismatch
  obs_cols <- colnames(to_explain)
  missing_in_obs <- setdiff(expected_cols, obs_cols)
  extra_in_obs <- setdiff(obs_cols, expected_cols)

  if (length(missing_in_obs) > 0 || length(extra_in_obs) > 0) {
    msg_parts <- character()
    if (length(missing_in_obs) > 0) {
      msg_parts <- c(msg_parts, paste0(
        "Missing columns in observation: ",
        paste(head(missing_in_obs, 5), collapse = ", "),
        if (length(missing_in_obs) > 5) paste0(" (and ", length(missing_in_obs) - 5, " more)") else ""
      ))
    }
    if (length(extra_in_obs) > 0) {
      msg_parts <- c(msg_parts, paste0(
        "Extra columns in observation: ",
        paste(head(extra_in_obs, 5), collapse = ", "),
        if (length(extra_in_obs) > 5) paste0(" (and ", length(extra_in_obs) - 5, " more)") else ""
      ))
    }
    stop(
      "LIME column mismatch between explainer background and observation. ",
      paste(msg_parts, collapse = ". "),
      call. = FALSE
    )
  }

  # Ensure column order matches
  to_explain <- to_explain[, expected_cols, drop = FALSE]

  # 5. Build explainer and generate explanations
  explainer <- lime::lime(background_x, prep$fits[[1]])
  explanation <- lime::explain(
    x = to_explain,
    explainer = explainer,
    n_features = n_features,
    n_labels = n_labels,
    ...
  )
  print(lime::plot_features(explanation))
  invisible(explanation)
}

#' Process and Evaluate a Model Workflow
#'
#' This function processes a fitted model or a tuning result, finalizes the model if tuning was used,
#' makes predictions on the test set, and computes performance metrics depending on the task type
#' (classification or regression). It supports binary and multiclass classification, and handles
#' probabilistic outputs when supported by the modeling engine.
#'
#' @param model_obj A fitted model or a tuning result (`tune_results` object).
#' @param model_id A character identifier for the model (used in warnings).
#' @param task Type of task, either `"classification"`, `"regression"`, or
#'   `"survival"`.
#' @param test_data A data frame containing the test data.
#' @param label The name of the outcome variable (as a character string).
#' @param event_class For binary classification, specifies which class is considered the positive class:
#'   `"first"` or `"second"`.
#' @param class_threshold For binary classification, controls how class probabilities
#'   are converted into hard class predictions. The default is `0.5` (standard threshold).
#'   Numeric values in (0, 1) set a fixed threshold. Use `"auto"` to tune a threshold
#'   on training data to maximize F1; use `"model"` to keep the model's default predictions.
#' @param multiclass_auc For multiclass ROC AUC, the averaging method to use:
#'   `"macro"` (default, tidymodels) or `"macro_weighted"`. Macro weights each
#'   class equally, while macro_weighted weights by class prevalence and can
#'   change model rankings on imbalanced data.
#' @param start_col Optional string. The name of the column specifying the
#'   start time in counting process (e.g., `(start, stop, event)`) survival
#'   data. Only used when \code{task = "survival"}.
#' @param time_col String. The name of the column specifying the event or
#'   censoring time (the "stop" time in counting process data). Only used
#'   when \code{task = "survival"}.
#' @param status_col String. The name of the column specifying the event
#'   status (e.g., 0 for censored, 1 for event). Only used when
#'   \code{task = "survival"}.
#' @param engine A character string indicating the model engine (e.g., `"xgboost"`, `"randomForest"`). Used
#'   to determine if class probabilities are supported. If `NULL`, probabilities are skipped.
#' @param train_data A data frame containing the training data, required to refit finalized workflows.
#' @param metric The name of the metric (e.g., `"roc_auc"`, `"accuracy"`, `"rmse"`) used for selecting the best tuning result.
#' @param metrics Optional yardstick metric set (e.g., `yardstick::metric_set(yardstick::rmse)`) used for computing regression performance.
#' @param summaryFunction Optional custom classification metric function passed to
#'   `yardstick::new_class_metric()` and included in holdout evaluation.
#' @param eval_times_user Optional numeric vector of time horizons at which to
#'   evaluate survival Brier scores. When `NULL`, sensible defaults based on the
#'   observed follow-up distribution are used.
#' @param bootstrap_ci Logical; if `TRUE`, bootstrap confidence intervals are
#'   estimated for performance metrics.
#' @param bootstrap_samples Integer giving the number of bootstrap resamples
#'   used when computing confidence intervals.
#' @param bootstrap_seed Optional integer seed applied before bootstrap
#'   resampling to make interval estimates reproducible.
#' @param at_risk_threshold Numeric value between 0 and 1 defining the minimum
#'   proportion of subjects required to remain at risk when determining the
#'   maximum follow-up time used in survival metrics.
#' @param survival_metric_convention Character string specifying which survival
#'   metric conventions to follow. `"fastml"` (default) uses fastml's internal
#'   defaults for evaluation horizons and t_max. `"tidymodels"` uses
#'   `eval_times_user` as the explicit evaluation grid and applies
#'   yardstick-style Brier/IBS normalization; when `eval_times_user` is `NULL`,
#'   time-dependent Brier metrics are omitted.
#' @param precomputed_predictions Optional data frame or nested list of
#'   previously generated predictions (per algorithm/engine) to reuse instead
#'   of re-predicting; primarily used when combining results across engines.
#'
#' @return A list with two elements:
#' \describe{
#'   \item{performance}{A tibble with computed performance metrics.}
#'   \item{predictions}{A tibble with predicted values and corresponding truth values, and probabilities (if applicable).}
#' }
#'
#' @details
#' - If the input `model_obj` is a `tune_results` object, the function finalizes the model using the
#'   best hyperparameters according to the specified `metric`, and refits the model on the full training data.
#'
#' - For classification tasks, performance metrics include accuracy, kappa, sensitivity, specificity, precision,
#'   F1-score, and ROC AUC (if probabilities are available).
#'
#' - For multiclass ROC AUC, the estimator is controlled by `multiclass_auc`.
#'
#' - For regression tasks, RMSE, R-squared, and MAE are returned.
#' 
#' - For models with missing prediction lengths, a helpful imputation error is thrown to guide data preprocessing.
#' 
#' @importFrom survival psurvreg survfit
#' @export
#' @importFrom stats model.matrix formula median quantile setNames
#' @importFrom utils head tail
process_model <- function(model_obj,
                          model_id,
                          task,
                          test_data,
                          label,
                          event_class,
                          class_threshold = 0.5,
                          start_col = NULL,
                          time_col = NULL,
                          status_col = NULL,
                          engine,
                          train_data,
                          metric,
                          eval_times_user = NULL,
                          bootstrap_ci = TRUE,
                          bootstrap_samples = 500,
                          bootstrap_seed = 1234,
                          at_risk_threshold = 0.1,
                          survival_metric_convention = "fastml",
                          metrics = NULL,
                          summaryFunction = NULL,
                          precomputed_predictions = NULL,
                          multiclass_auc = "macro") {
  # If the model object is a tuning result, finalize the workflow
  if (inherits(model_obj, "tune_results")) {
    final_model <- finalize_tuned_model(
      model_obj = model_obj,
      model_id = model_id,
      task = task,
      metric = metric,
      train_data = train_data
    )
    if (is.null(final_model)) {
      return(NULL)
    }
  } else if (inherits(model_obj, "fastml_native_survival")) {
    # Native survival model fitted outside parsnip/workflows
    final_model <- model_obj
  } else {
    # Otherwise, assume the model is already a fitted workflow
    final_model <- model_obj
  }

  normalize_estimator <- function(df) {
    if (!".estimator" %in% names(df)) {
      df$.estimator <- NA_character_
    }
    df
  }

  metric_key <- function(df) {
    df <- normalize_estimator(df)
    estimator <- df$.estimator
    estimator[is.na(estimator)] <- ""
    paste(df$.metric, estimator, sep = "::")
  }

  add_bootstrap_ci <- function(perf_df, data_df, boot_fun) {
    n_obs <- nrow(data_df)
    ci_lower <- rep(NA_real_, nrow(perf_df))
    ci_upper <- rep(NA_real_, nrow(perf_df))
    n_boot_used <- 0L
    if (isTRUE(bootstrap_ci) &&
        bootstrap_samples > 1 && n_obs > 1) {
      .boot_old_seed <- if (exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE))
        get(".Random.seed", envir = .GlobalEnv, inherits = FALSE) else NULL
      .boot_had_seed <- !is.null(.boot_old_seed)
      if (!is.null(bootstrap_seed)) {
        set.seed(bootstrap_seed)
      }
      perf_key <- metric_key(perf_df)
      boot_matrix <- matrix(NA_real_,
                            nrow = bootstrap_samples,
                            ncol = nrow(perf_df))
      colnames(boot_matrix) <- perf_key
      for (b in seq_len(bootstrap_samples)) {
        idx <- sample.int(n_obs, size = n_obs, replace = TRUE)
        boot_perf <- suppressWarnings(boot_fun(data_df[idx, , drop = FALSE]))
        if (!is.null(boot_perf) && nrow(boot_perf) > 0) {
          boot_key <- metric_key(boot_perf)
          match_idx <- match(boot_key, perf_key)
          valid <- !is.na(match_idx)
          if (any(valid)) {
            boot_matrix[b, match_idx[valid]] <- as.numeric(boot_perf$.estimate[valid])
          }
        }
      }
      for (j in seq_len(nrow(perf_df))) {
        vals <- boot_matrix[, j]
        vals <- vals[is.finite(vals)]
        if (length(vals) > 0) {
          ci_lower[j] <- stats::quantile(vals,
                                         probs = 0.025,
                                         names = FALSE,
                                         na.rm = TRUE)
          ci_upper[j] <- stats::quantile(vals,
                                         probs = 0.975,
                                         names = FALSE,
                                         na.rm = TRUE)
        }
      }
      n_boot_used <- bootstrap_samples
      if (.boot_had_seed) {
        assign(".Random.seed", .boot_old_seed, envir = .GlobalEnv)
      } else if (exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE)) {
        rm(".Random.seed", envir = .GlobalEnv)
      }
    }
    perf_df$.lower <- as.numeric(ci_lower)
    perf_df$.upper <- as.numeric(ci_upper)
    perf_df$.n_boot <- n_boot_used
    perf_df
  }

  # Make predictions and compute performance metrics
  if (task == "classification") {
    pred_class <- predict(final_model, new_data = test_data, type = "class")$.pred_class

    pred_prob <- NULL
    pred_prob_error <- NULL
    if (!is.null(engine) &&
        !is.na(engine) && engine != "LiblineaR") {
      pred_prob <- tryCatch(
        predict(final_model, new_data = test_data, type = "prob"),
        error = function(e) {
          pred_prob_error <<- e
          NULL
        }
      )
      if (!is.null(pred_prob) && !is.data.frame(pred_prob)) {
        pred_prob <- NULL
      }
      if (is.null(pred_prob) && !is.null(pred_prob_error)) {
        warning(
          sprintf(
            "Probability predictions failed for model '%s' (engine: %s); ROC AUC will be skipped. %s",
            model_id,
            engine,
            conditionMessage(pred_prob_error)
          ),
          call. = FALSE
        )
      }
    }

    if (nrow(test_data) != length(pred_class)) {
      stop(
        'The dataset has missing values. To handle this, set impute_method = "remove" to delete rows with missing values,
             or use recipe-based imputation such as "medianImpute", "knnImpute", or "bagImpute" so the preprocessing can be applied during predict().'
      )
    }

    data_metrics <- test_data %>%
      dplyr::select(truth = !!rlang::sym(label)) %>%
      dplyr::mutate(estimate = pred_class)

    prob_available <- !is.null(pred_prob) && is.data.frame(pred_prob) && ncol(pred_prob) > 0
    if (prob_available) {
      data_metrics <- dplyr::bind_cols(data_metrics, pred_prob)
    }

    prob_cols_all <- setdiff(names(data_metrics), c("truth", "estimate"))
    prob_cols <- prob_cols_all
    if (length(prob_cols_all) > 0) {
      prob_cols <- prob_cols_all[
        vapply(data_metrics[prob_cols_all], is.numeric, logical(1))
      ]
    }
    prob_available <- prob_available && length(prob_cols) > 0
    has_probabilities <- !is.null(engine) &&
      !is.na(engine) && engine != "LiblineaR" &&
      length(prob_cols) > 0

    compute_calibration_rows <- function(df) {
      if (!prob_available || length(prob_cols) == 0) {
        return(NULL)
      }
      prob_info <- fastml_prepare_prob_matrix(df, prob_cols)
      if (is.null(prob_info)) {
        return(NULL)
      }
      calib <- fastml_class_calibration_metrics(
        truth_vec = df$truth,
        prob_mat = prob_info$mat,
        prob_cols = prob_info$cols,
        event_class = event_class
      )
      if (!is.finite(calib$logloss) &&
          !is.finite(calib$brier_score) &&
          !is.finite(calib$ece)) {
        return(NULL)
      }
      tibble::tibble(
        .metric = c("logloss", "brier_score", "ece"),
        .estimator = rep(calib$estimator, 3),
        .estimate = as.numeric(c(calib$logloss, calib$brier_score, calib$ece))
      )
    }

    ensure_calibration_metrics <- function(perf_df, estimator_value) {
      metric_names <- c("logloss", "brier_score", "ece")
      missing <- setdiff(metric_names, perf_df$.metric)
      if (length(missing) == 0) {
        return(perf_df)
      }
      add_rows <- tibble::tibble(
        .metric = missing,
        .estimator = rep(estimator_value, length(missing)),
        .estimate = NA_real_
      )
      dplyr::bind_rows(perf_df, add_rows)
    }

    resolve_class_threshold <- function(class_threshold, metric_name) {
      if (is.null(class_threshold)) {
        return(list(mode = "model", metric = NULL, value = NA_real_))
      }
      if (is.numeric(class_threshold)) {
        if (length(class_threshold) != 1 || !is.finite(class_threshold)) {
          return(list(mode = "model", metric = NULL, value = NA_real_))
        }
        return(list(
          mode = "fixed",
          metric = "fixed",
          value = min(max(class_threshold, 0), 1)
        ))
      }
      if (!is.character(class_threshold) || length(class_threshold) != 1 || !nzchar(class_threshold)) {
        return(list(mode = "model", metric = NULL, value = NA_real_))
      }
      method <- tolower(class_threshold)
      if (method %in% c("f1", "f1_score", "f1-score")) {
        method <- "f_meas"
      }
      if (method %in% c("model", "none", "default")) {
        return(list(mode = "model", metric = NULL, value = NA_real_))
      }
      if (method == "auto") {
        threshold_metrics <- c("accuracy", "sens", "spec", "precision", "f_meas")
        if (!is.null(metric_name) && metric_name %in% threshold_metrics) {
          method <- metric_name
        } else {
          method <- "f_meas"
        }
      }
      if (method %in% c("accuracy", "sens", "spec", "precision", "f_meas", "youden")) {
        return(list(mode = "tuned", metric = method, value = NA_real_))
      }
      warning(
        sprintf("Unsupported class_threshold '%s'; using model default predictions.", class_threshold),
        call. = FALSE
      )
      list(mode = "model", metric = NULL, value = NA_real_)
    }

    compute_optimal_threshold <- function(truth_vec, prob_vec, positive_class, method) {
      if (is.null(positive_class)) {
        return(NA_real_)
      }
      valid <- is.finite(prob_vec) & !is.na(truth_vec)
      if (!any(valid)) {
        return(NA_real_)
      }
      truth_vec <- as.character(truth_vec[valid])
      prob_vec <- prob_vec[valid]
      if (length(unique(truth_vec)) < 2) {
        return(NA_real_)
      }
      pos <- truth_vec == positive_class
      n_pos <- sum(pos)
      n_neg <- length(pos) - n_pos
      if (n_pos == 0 || n_neg == 0) {
        return(NA_real_)
      }
      ord <- order(prob_vec, decreasing = TRUE)
      prob_sorted <- prob_vec[ord]
      pos_sorted <- pos[ord]
      rle_probs <- rle(prob_sorted)
      idx_end <- cumsum(rle_probs$lengths)
      tp <- cumsum(pos_sorted)[idx_end]
      fp <- cumsum(!pos_sorted)[idx_end]
      fn <- n_pos - tp
      tn <- n_neg - fp

      accuracy <- (tp + tn) / (n_pos + n_neg)
      sens <- ifelse((tp + fn) > 0, tp / (tp + fn), NA_real_)
      spec <- ifelse((tn + fp) > 0, tn / (tn + fp), NA_real_)
      precision <- ifelse((tp + fp) > 0, tp / (tp + fp), NA_real_)
      f_meas <- ifelse((precision + sens) > 0, 2 * precision * sens / (precision + sens), NA_real_)
      youden <- sens + spec - 1

      metric_vals <- switch(
        method,
        accuracy = accuracy,
        sens = sens,
        spec = spec,
        precision = precision,
        f_meas = f_meas,
        youden = youden,
        accuracy
      )
      if (!any(is.finite(metric_vals))) {
        return(NA_real_)
      }
      best_val <- max(metric_vals, na.rm = TRUE)
      best_idx <- which(metric_vals == best_val)
      if (length(best_idx) > 1) {
        best_idx <- best_idx[which.min(abs(rle_probs$values[best_idx] - 0.5))]
      }
      rle_probs$values[best_idx]
    }

    tune_class_threshold <- function(positive_class, method) {
      if (is.null(train_data) || !is.data.frame(train_data)) {
        return(NA_real_)
      }
      train_truth <- train_data[[label]]
      if (is.null(train_truth)) {
        return(NA_real_)
      }
      train_prob <- tryCatch(
        predict(final_model, new_data = train_data, type = "prob"),
        error = function(e) NULL
      )
      if (is.null(train_prob) || !is.data.frame(train_prob) || nrow(train_prob) == 0) {
        return(NA_real_)
      }
      prob_cols_train <- names(train_prob)
      prob_cols_train <- prob_cols_train[
        vapply(train_prob[prob_cols_train], is.numeric, logical(1))
      ]
      if (length(prob_cols_train) == 0) {
        return(NA_real_)
      }
      prob_meta_train <- fastml_resolve_binary_prob_column(
        prob_cols = prob_cols_train,
        truth_levels = truth_levels,
        event_class = event_class
      )
      prob_col_train <- prob_meta_train$prob_col
      if (is.null(prob_col_train) || !is.numeric(train_prob[[prob_col_train]])) {
        return(NA_real_)
      }
      if (!any(is.finite(train_prob[[prob_col_train]]))) {
        return(NA_real_)
      }
      train_truth <- factor(train_truth, levels = truth_levels)
      compute_optimal_threshold(
        truth_vec = train_truth,
        prob_vec = train_prob[[prob_col_train]],
        positive_class = positive_class,
        method = method
      )
    }

    truth_levels <- levels(data_metrics$truth)
    if (is.null(truth_levels) || length(truth_levels) == 0) {
      truth_levels <- unique(as.character(data_metrics$truth))
    }
    num_classes <- length(truth_levels)
    observed_classes <- length(unique(data_metrics$truth))

    metric_name <- metric
    multiclass_auc <- fastml_normalize_multiclass_auc(multiclass_auc)
    build_class_metrics <- function() {
      if (is.null(summaryFunction)) {
        return(yardstick::metric_set(
          yardstick::accuracy,
          yardstick::kap,
          yardstick::sens,
          yardstick::spec,
          yardstick::precision,
          yardstick::f_meas
        ))
      }
      if (is.null(metric_name) || !nzchar(metric_name)) {
        metric_name <- "custom_metric"
      }
      new_class_metric <- yardstick::new_class_metric(summaryFunction, "maximize")
      assign(metric_name, new_class_metric, envir = environment())
      yardstick::metric_set(
        yardstick::accuracy,
        yardstick::kap,
        yardstick::sens,
        yardstick::spec,
        yardstick::precision,
        yardstick::f_meas,
        !!rlang::sym(metric_name)
      )
    }

    safe_metrics_class <- function(df, ...) {
      withCallingHandlers(
        metrics_class(df, ...),
        warning = function(w) {
          if (inherits(w, "yardstick_warning_precision_undefined")) {
            invokeRestart("muffleWarning")
          }
        }
      )
    }

    threshold_used <- NA_real_
    threshold_metric <- NA_character_

    if (num_classes == 2) {
      if (!event_class %in% c("first", "second")) {
        stop("Invalid event_class argument. It should be either 'first' or 'second'.")
      }

      prob_meta <- fastml_resolve_binary_prob_column(
        prob_cols = prob_cols,
        truth_levels = truth_levels,
        event_class = event_class
      )
      prob_col <- prob_meta$prob_col
      positive_class <- prob_meta$positive_class

      if (!is.null(prob_col) && !is.numeric(data_metrics[[prob_col]])) {
        prob_col <- NULL
      }
      if (!is.null(prob_col) && !any(is.finite(data_metrics[[prob_col]]))) {
        prob_col <- NULL
      }
      if (prob_available && !is.null(prob_col) && isTRUE(prob_meta$used_fallback)) {
        warning(
          sprintf(
            "Probability column for positive class '%s' not found; using '%s' based on column order.",
            positive_class,
            prob_col
          ),
          call. = FALSE
        )
      }

      if (prob_available && !is.null(prob_col)) {
        threshold_spec <- resolve_class_threshold(class_threshold, metric_name)
        if (identical(threshold_spec$mode, "fixed")) {
          threshold_used <- threshold_spec$value
          threshold_metric <- threshold_spec$metric
        } else if (identical(threshold_spec$mode, "tuned")) {
          threshold_metric <- threshold_spec$metric
          threshold_used <- tune_class_threshold(positive_class, threshold_spec$metric)
        }
        if (is.finite(threshold_used) && !is.null(positive_class)) {
          negative_class <- setdiff(truth_levels, positive_class)[1]
          if (!is.null(negative_class) && length(negative_class) == 1 && !is.na(negative_class)) {
            data_metrics$estimate <- ifelse(
              data_metrics[[prob_col]] >= threshold_used,
              positive_class,
              negative_class
            )
            data_metrics$estimate <- factor(data_metrics$estimate, levels = truth_levels)
          }
        }
      }

      metrics_class <- build_class_metrics()
      perf_class <- safe_metrics_class(
        data_metrics,
        truth = truth,
        estimate = estimate,
        event_level = event_class
      )

      if (prob_available) {
        if (observed_classes < 2) {
          warning(
            sprintf(
              "Only one class present in truth for model '%s'; ROC AUC will be skipped.",
              model_id
            ),
            call. = FALSE
          )
          perf <- perf_class
        } else if (!is.null(prob_col)) {
          # Compute ROC AUC using the probability column for the positive class
          roc_auc_value <- yardstick::roc_auc(
            data_metrics,
            truth = truth,!!rlang::sym(prob_col),
            event_level = event_class
          )
          perf <- dplyr::bind_rows(perf_class, roc_auc_value)
        } else {
          warning(
            sprintf(
              "Probability column for positive class '%s' not found for model '%s'; ROC AUC will be skipped. Available columns: %s",
              ifelse(is.null(positive_class), "", positive_class),
              model_id,
              paste(prob_cols, collapse = ", ")
            ),
            call. = FALSE
          )
          perf <- perf_class
        }
      } else {
        perf <- perf_class
      }

      if (!any(perf$.metric == "roc_auc")) {
        perf <- dplyr::bind_rows(
          perf,
          tibble::tibble(
            .metric = "roc_auc",
            .estimator = "binary",
            .estimate = NA_real_
          )
        )
      }

      calib_rows <- compute_calibration_rows(data_metrics)
      if (!is.null(calib_rows)) {
        perf <- dplyr::bind_rows(perf, calib_rows)
      }
      perf <- ensure_calibration_metrics(perf, NA_character_)

      compute_boot_perf <- function(df) {
        perf_boot <- tryCatch(
          suppressWarnings(safe_metrics_class(
            df,
            truth = truth,
            estimate = estimate,
            event_level = event_class
          )),
          error = function(e) NULL
        )
        if (is.null(perf_boot)) {
          perf_boot <- perf[0, , drop = FALSE]
        }
        if (prob_available && !is.null(prob_col) && prob_col %in% names(df)) {
          roc_boot <- tryCatch(
            suppressWarnings(yardstick::roc_auc(
              df,
              truth = truth,!!rlang::sym(prob_col),
              event_level = event_class
            )),
            error = function(e) NULL
          )
          if (!is.null(roc_boot)) {
            perf_boot <- dplyr::bind_rows(perf_boot, roc_boot)
          }
        }
        calib_boot <- compute_calibration_rows(df)
        if (!is.null(calib_boot)) {
          perf_boot <- dplyr::bind_rows(perf_boot, calib_boot)
        }
        perf_boot
      }
    } else {
      # Multiclass classification (macro metrics; ROC AUC uses configured estimator)
      metrics_class <- build_class_metrics()
      perf_class <- safe_metrics_class(
        data_metrics,
        truth = truth,
        estimate = estimate,
        estimator = multiclass_auc
      )

      if (has_probabilities) {
        perf_roc_auc <- yardstick::roc_auc(
          data_metrics,
          truth = truth,!!!rlang::syms(prob_cols),
          estimator = multiclass_auc
        )
        perf <- dplyr::bind_rows(perf_class, perf_roc_auc)
      } else {
        perf <- perf_class
      }

      if (!any(perf$.metric == "roc_auc")) {
        perf <- dplyr::bind_rows(
          perf,
          tibble::tibble(
            .metric = "roc_auc",
            .estimator = multiclass_auc,
            .estimate = NA_real_
          )
        )
      }

      calib_rows <- compute_calibration_rows(data_metrics)
      if (!is.null(calib_rows)) {
        perf <- dplyr::bind_rows(perf, calib_rows)
      }
      calib_estimator <- if (!is.null(calib_rows) &&
                             ".estimator" %in% names(calib_rows)) {
        calib_rows$.estimator[1]
      } else {
        "macro"
      }
      perf <- ensure_calibration_metrics(perf, calib_estimator)

      compute_boot_perf <- function(df) {
        perf_boot <- tryCatch(
          suppressWarnings(safe_metrics_class(
            df,
            truth = truth,
            estimate = estimate,
            estimator = multiclass_auc
          )),
          error = function(e) NULL
        )
        if (is.null(perf_boot)) {
          perf_boot <- perf[0, , drop = FALSE]
        }
        if (has_probabilities) {
          roc_boot <- tryCatch(
            suppressWarnings(yardstick::roc_auc(
              df,
              truth = truth,!!!rlang::syms(prob_cols),
              estimator = multiclass_auc
            )),
            error = function(e) NULL
          )
          if (!is.null(roc_boot)) {
            perf_boot <- dplyr::bind_rows(perf_boot, roc_boot)
          }
        }
        calib_boot <- compute_calibration_rows(df)
        if (!is.null(calib_boot)) {
          perf_boot <- dplyr::bind_rows(perf_boot, calib_boot)
        }
        perf_boot
      }
    }

    perf <- add_bootstrap_ci(perf, data_metrics, compute_boot_perf)
    if (is.finite(threshold_used)) {
      attr(perf, "class_threshold") <- threshold_used
      attr(perf, "class_threshold_metric") <- threshold_metric
      attr(data_metrics, "class_threshold") <- threshold_used
      attr(data_metrics, "class_threshold_metric") <- threshold_metric
    }
  } else if (task == "survival") {
    survival_metric_convention <- fastml_normalize_survival_convention(survival_metric_convention)
    status_warning_emitted <- FALSE
    normalize_status <- function(status_vec, reference_length) {
      res <- fastml_normalize_survival_status(status_vec, reference_length)
      if (res$recoded && !status_warning_emitted) {
        warning(
          "Detected non-standard survival status coding; recoding to 0 = censored and 1 = event.",
          call. = FALSE
        )
        status_warning_emitted <<- TRUE
      }
      res$status
    }
    # if (length(label) == 3) {
    #   start_col <- label[1]
    #   time_col <- label[2]
    #   status_col <- label[3]
    # } else {
    #   start_col <- NULL
    #   time_col <- label[1]
    #   status_col <- label[2]
    # }
    test_status_clean <- normalize_status(test_data[[status_col]], nrow(test_data))
    train_status_clean <- normalize_status(train_data[[status_col]], nrow(train_data))
    test_data[[status_col]] <- test_status_clean
    train_data[[status_col]] <- train_status_clean
    if (!is.null(start_col)) {
      surv_obj <- survival::Surv(test_data[[start_col]], test_data[[time_col]], test_status_clean)
    } else {
      surv_obj <- survival::Surv(test_data[[time_col]], test_status_clean)
    }

    # Prepare data for prediction depending on model type.
    # Failure to apply the stored recipe is fatal rather than silently ignored:
    # predicting on unpreprocessed features would yield plausible-looking but
    # meaningless metrics, which is worse than no result at all.
    pred_new_data <- test_data
    if (inherits(final_model, "fastml_native_survival") &&
        !is.null(final_model$recipe)) {
      pred_new_data <- tryCatch(
        recipes::bake(final_model$recipe, new_data = test_data),
        error = function(e) {
          stop(sprintf(
            paste0(
              "Could not apply the stored preprocessing recipe for model '%s' to ",
              "the evaluation data: %s\n",
              "The evaluation data must contain the same columns, types, and ",
              "factor levels as the training data. fastml does not fall back to ",
              "unpreprocessed data, because doing so would report metrics for a ",
              "model evaluated on features it was not trained on."
            ),
            model_id,
            conditionMessage(e)
          ), call. = FALSE)
        }
      )
    }

    pred_predictors <- NULL
    if (inherits(final_model, "fastml_native_survival")) {
      pred_predictors <- fastml_prepare_native_survival_predictors(final_model, pred_new_data, test_data)
    }

    identify_survival_model <- function(obj) {
      fit_obj <- NULL
      if (inherits(obj, "fastml_native_survival")) {
        fit_obj <- tryCatch(
          obj$fit,
          error = function(e)
            NULL
        )
      }
      if (is.null(fit_obj)) {
        fit_obj <- tryCatch({
          workflows::extract_fit_engine(obj)
        }, error = function(e)
          NULL)
      }
      if (is.null(fit_obj)) {
        fit_obj <- tryCatch({
          workflows::extract_fit_parsnip(obj)$fit
        }, error = function(e)
          NULL)
      }
      if (is.null(fit_obj)) {
        fit_obj <- tryCatch({
          parsnip::extract_fit_engine(obj)
        }, error = function(e)
          NULL)
      }
      if (is.null(fit_obj)) {
        fit_obj <- tryCatch(
          obj$fit,
          error = function(e)
            NULL
        )
      }

      model_type <- "other"
      if (inherits(fit_obj, "coxph")) {
        model_type <- "coxph"
      } else if (inherits(fit_obj, c("stpm2", "pstpm2"))) {
        model_type <- "rstpm2"
      } else if (inherits(fit_obj, "survreg")) {
        model_type <- "survreg"
      } else if (inherits(fit_obj, "flexsurvreg")) {
        model_type <- "flexsurv"
      } else if (inherits(fit_obj, "glmnet")) {
        model_type <- "glmnet"
      } else if (inherits(fit_obj, "fastml_xgb_survival")) {
        objective <- tryCatch(
          fit_obj$objective,
          error = function(e)
            NULL
        )
        if (identical(objective, "survival:aft")) {
          model_type <- "xgboost_aft"
        } else if (identical(objective, "survival:cox")) {
          model_type <- "xgboost_cox"
        } else {
          model_type <- "xgboost_survival"
        }
      }

      list(type = model_type, fit = fit_obj)
    }

    survival_model_info <- identify_survival_model(final_model)
    survival_model_type <- survival_model_info$type

    default_time_val <- NA_real_
    time_candidates <- list()
    if (!is.null(final_model$time_col) &&
        final_model$time_col %in% names(train_data)) {
      time_candidates[[length(time_candidates) + 1]] <- train_data[[final_model$time_col]]
    }
    if (time_col %in% names(train_data)) {
      time_candidates[[length(time_candidates) + 1]] <- train_data[[time_col]]
    }
    if (time_col %in% names(test_data)) {
      time_candidates[[length(time_candidates) + 1]] <- test_data[[time_col]]
    }
    if (length(time_candidates) > 0) {
      med_vals <- vapply(time_candidates, function(x) {
        stats::median(as.numeric(x), na.rm = TRUE)
      }, numeric(1))
      default_time_val <- med_vals[is.finite(med_vals) &
                                     med_vals > 0]
      if (length(default_time_val) > 0) {
        default_time_val <- default_time_val[1]
      } else {
        default_time_val <- NA_real_
      }
    }
    if (!is.finite(default_time_val) || default_time_val <= 0) {
      default_time_val <- 1
    }

    extract_pred <- function(pred) {
      # Return early if NULL
      if (is.null(pred)) {
        return(pred)
      }

      # If a tibble/data.frame, pick the appropriate column first
      if (is.data.frame(pred)) {
        if (".pred" %in% names(pred)) {
          pred <- pred[[".pred"]]
        } else if (".pred_survival" %in% names(pred)) {
          pred <- pred[[".pred_survival"]]
        } else {
          pred <- pred[[1]]
        }
      }

      # If a list (including list-column extracted from tibble), unlist robustly
      # Engines for survival often return a list-column where each element is a
      # numeric of length 1 when a single eval_time is provided.
      if (is.list(pred)) {
        pred <- vapply(pred, function(x) {
          if (is.null(x))
            return(NA_real_)
          # If it's a scalar numeric, take it; otherwise try first element
          if (is.numeric(x) &&
              length(x) == 1L)
            return(as.numeric(x))
          if (is.atomic(x) &&
              length(x) >= 1L)
            return(as.numeric(x[[1L]]))
          # If it's a list wrapping a scalar, try to dig one level
          if (is.list(x) && length(x) >= 1L) {
            x1 <- x[[1L]]
            if (is.numeric(x1) &&
                length(x1) >= 1L)
              return(as.numeric(x1[[1L]]))
          }
          NA_real_
        }, numeric(1))
      }

      # Coerce remaining vector-like objects to numeric
      as.numeric(pred)
    }

    risk <- tryCatch({
      if (inherits(final_model, "fastml_native_survival")) {
        # Prepare predictor-only data for native survival fits
        if (is.null(pred_predictors)) {
          pred_predictors <- data.frame()
        }
        if (inherits(final_model$fit, "coxph")) {
          as.numeric(stats::predict(
            final_model$fit,
            newdata = pred_predictors,
            type = "lp"
          ))
        } else if (inherits(final_model$fit, "survreg")) {
          as.numeric(stats::predict(
            final_model$fit,
            newdata = pred_predictors,
            type = "lp"
          ))
        } else if (inherits(final_model$fit, "flexsurvreg")) {
          tryCatch({
            # Use the original (unprocessed) test data for flexsurvreg predictions
            mm <- model.matrix(formula(final_model$fit), data = test_data)
            coefs <- final_model$fit$res.t[, "est"]
            common <- intersect(names(coefs), colnames(mm))
            lp <- as.numeric(mm[, common, drop = FALSE] %*% coefs[common])
            if ("(Intercept)" %in% names(coefs)) {
              lp <- lp + coefs["(Intercept)"]
            }
            lp
          }, error = function(e) {
            warning("Failed to compute linear predictors for flexsurvreg: ", e$message)
            rep(NA_real_, nrow(test_data))
          })
        } else if (inherits(final_model$fit, "glmnet")) {
          if (!requireNamespace("glmnet", quietly = TRUE)) {
            rep(NA_real_, nrow(test_data))
          } else {
            feature_names <- final_model$feature_names
            x_terms <- final_model$x_terms
            x_contrasts <- final_model$x_contrasts
            n_obs <- if (is.null(pred_predictors))
              0L
            else
              nrow(pred_predictors)

            if (is.null(feature_names) ||
                length(feature_names) == 0) {
              rep(NA_real_, nrow(test_data))
            } else if (n_obs == 0) {
              numeric(0)
            } else {
              prepare_glmnet_newx <- function(new_data) {
                if (is.null(new_data)) {
                  new_data <- data.frame()
                }
                n_local <- nrow(new_data)
                if (n_local == 0) {
                  out <- matrix(0,
                                nrow = 0,
                                ncol = length(feature_names))
                  colnames(out) <- feature_names
                  return(out)
                }
                if (ncol(new_data) == 0) {
                  mm <- matrix(0, nrow = n_local, ncol = 0)
                } else {
                  mm <- tryCatch({
                    if (!is.null(x_terms)) {
                      stats::model.matrix(x_terms, new_data, contrasts.arg = x_contrasts)
                    } else {
                      stats::model.matrix( ~ . - 1, data = new_data)
                    }
                  }, error = function(e)
                    NULL)
                  if (is.null(mm)) {
                    mm <- tryCatch(
                      stats::model.matrix( ~ . - 1, data = new_data),
                      error = function(e)
                        NULL
                    )
                  }
                  if (is.null(mm)) {
                    mm <- matrix(0, nrow = n_local, ncol = 0)
                  }
                }
                mm <- as.matrix(mm)
                mm_colnames <- colnames(mm)
                if (!is.null(mm_colnames) &&
                    any(mm_colnames == "(Intercept)")) {
                  keep_cols <- mm_colnames != "(Intercept)"
                  mm <- mm[, keep_cols, drop = FALSE]
                  mm_colnames <- colnames(mm)
                }
                mm_full <- matrix(0,
                                  nrow = n_local,
                                  ncol = length(feature_names))
                colnames(mm_full) <- feature_names
                if (ncol(mm) > 0) {
                  overlap <- intersect(feature_names, mm_colnames)
                  if (length(overlap) > 0) {
                    mm_full[, overlap] <- mm[, overlap, drop = FALSE]
                  }
                }
                mm_full[!is.finite(mm_full)] <- 0
                mm_full
              }

              mm_full <- prepare_glmnet_newx(pred_predictors)
              n_obs <- nrow(mm_full)
              penalty <- final_model$penalty
              pred_lp <- tryCatch({
                predict(
                  final_model$fit,
                  newx = mm_full,
                  s = penalty,
                  type = "link"
                )
              }, error = function(e)
                NULL)
              if (!is.null(pred_lp) && length(dim(pred_lp)) >= 2) {
                pred_lp <- as.matrix(pred_lp)[, 1, drop = TRUE]
              }
              if (!is.null(pred_lp)) {
                pred_lp <- as.numeric(pred_lp)
              }

              needs_manual_lp <- is.null(pred_lp) ||
                length(pred_lp) == 0 || any(!is.finite(pred_lp))

              if (needs_manual_lp) {
                coef_mat <- tryCatch({
                  coef(final_model$fit, s = penalty)
                }, error = function(e)
                  NULL)

                if (!is.null(coef_mat)) {
                  coef_dense <- as.matrix(coef_mat)
                  intercept <- 0
                  if ("(Intercept)" %in% rownames(coef_dense)) {
                    intercept <- coef_dense["(Intercept)", 1]
                    coef_dense <- coef_dense[setdiff(rownames(coef_dense), "(Intercept)"), , drop = FALSE]
                  }

                  coef_vec <- numeric(length(feature_names))
                  names(coef_vec) <- feature_names
                  coef_overlap <- intersect(feature_names, rownames(coef_dense))
                  if (length(coef_overlap) > 0) {
                    coef_vals <- coef_dense[coef_overlap, 1]
                    coef_vals[!is.finite(coef_vals)] <- 0
                    coef_vec[coef_overlap] <- coef_vals
                  }
                  if (!is.finite(intercept)) {
                    intercept <- 0
                  }
                  pred_lp <- as.numeric(mm_full %*% coef_vec + intercept)
                } else {
                  pred_lp <- rep(NA_real_, n_obs)
                }
              }

              pred_lp
            }
          }
        } else if (inherits(final_model$fit, "fastml_xgb_survival")) {

          # --- FIX: Robust Matrix Preparation and Error Reporting ---
          expected_features <- final_model$fit$feature_names
          current_predictors <- pred_predictors
          if(is.null(current_predictors)) current_predictors <- data.frame()

          # Ensure strictly numeric matrix with correct columns
          # 1. Initialize matrix with NAs/0
          n_rows_pred <- nrow(current_predictors)
          if(n_rows_pred == 0 && nrow(test_data) > 0) {
            # If predictors are empty but test_data exists (e.g. only outcome cols were dropped?)
            # This assumes predictors should match test rows.
            n_rows_pred <- nrow(test_data)
          }

          mat_data <- matrix(0, nrow = n_rows_pred, ncol = length(expected_features))
          colnames(mat_data) <- expected_features

          # 2. Fill available columns
          common_cols <- intersect(names(current_predictors), expected_features)
          if(length(common_cols) > 0) {
            # Convert valid columns to numeric safely
            for(col in common_cols) {
              val <- current_predictors[[col]]
              # Factor handling: if bake() didn't handle it, ensure numeric
              if(is.factor(val) || is.character(val)) {
                val <- as.numeric(as.factor(val))
              }
              mat_data[, col] <- as.numeric(val)
            }
          }

          # 3. Create DMatrix (add missing=NA to be safe, though 0 is filled above)
          dtest <- tryCatch(
            xgboost::xgb.DMatrix(data = mat_data, missing = NA),
            error = function(e) {
              warning("xgb.DMatrix creation failed: ", e$message)
              NULL
            }
          )

          lp_vals <- NULL
          if (!is.null(dtest)) {
            lp_vals <- tryCatch(
              predict(final_model$fit$booster, dtest),
              error = function(e) {
                warning("XGBoost predict failed: ", e$message)
                NULL
              }
            )
          }

          if (is.null(lp_vals)) {
            lp_vals <- rep(NA_real_, nrow(test_data))
          }

          # Return risk (negative LP for AFT because higher time = lower risk)
          if (identical(final_model$fit$objective, "survival:aft")) {
            as.numeric(-lp_vals)
          } else {
            as.numeric(lp_vals)
          }
          # --- END FIX ---
        } else if (inherits(final_model$fit, c("stpm2", "pstpm2"))) {
          if (!requireNamespace("rstpm2", quietly = TRUE)) {
            rep(NA_real_, nrow(test_data))
          } else {
            rp_newdata <- pred_predictors
            if (is.null(rp_newdata)) {
              rp_newdata <- data.frame()
            }
            time_var <- final_model$time_col
            if (!is.null(time_var)) {
              if (!(time_var %in% names(rp_newdata)) &&
                  time_var %in% names(test_data)) {
                rp_newdata[[time_var]] <- test_data[[time_var]]
              }
              if (!(time_var %in% names(rp_newdata))) {
                rp_newdata[[time_var]] <- rep(default_time_val, nrow(test_data))
              } else {
                rp_newdata[[time_var]][!is.finite(rp_newdata[[time_var]])] <- default_time_val
              }
            }
            if (!is.null(final_model$start_col) &&
                !(final_model$start_col %in% names(rp_newdata)) &&
                final_model$start_col %in% names(test_data)) {
              rp_newdata[[final_model$start_col]] <- test_data[[final_model$start_col]]
            }
            as.numeric(predict(
              final_model$fit,
              newdata = rp_newdata,
              type = "lp"
            ))
          }
        } else {
          rep(NA_real_, nrow(test_data))
        }
      } else {
        pred_vals <- NULL
        if (identical(survival_model_type, "rstpm2")) {
          pred_vals <- tryCatch(
            predict(final_model, new_data = test_data, type = "lp"),
            error = function(e)
              NULL
          )
        }
        if (is.null(pred_vals)) {
          pred_vals <- tryCatch(
            predict(final_model, new_data = test_data, type = "linear_pred"),
            error = function(e)
              NULL
          )
        }
        extract_pred(pred_vals)
      }
    }, error = function(e) {
      # The original condition is reported rather than discarded: without it the
      # user sees only NA risk scores (or a second, unrelated failure) and has no
      # way to tell which engine call went wrong.
      primary_msg <- conditionMessage(e)
      if (inherits(final_model, "fastml_native_survival")) {
        warning(sprintf(
          paste0(
            "Risk prediction failed for model '%s'; risk scores are NA and ",
            "survival metrics for this model will be unavailable. Original error: %s"
          ),
          model_id,
          primary_msg
        ), call. = FALSE)
        rep(NA_real_, nrow(test_data))
      } else {
        warning(sprintf(
          paste0(
            "Risk prediction failed for model '%s'; retrying with a default ",
            "prediction call. Original error: %s"
          ),
          model_id,
          primary_msg
        ), call. = FALSE)
        tryCatch(
          extract_pred(predict(final_model, new_data = test_data)),
          error = function(e2) {
            stop(sprintf(
              paste0(
                "Risk prediction failed for model '%s'. Original error: %s\n",
                "The fallback prediction call also failed: %s"
              ),
              model_id,
              primary_msg,
              conditionMessage(e2)
            ), call. = FALSE)
          }
        )
      }
    })

    if (length(risk) == 0) {
      risk <- rep(NA_real_, nrow(test_data))
    } else if (length(risk) == 1 && nrow(test_data) > 1) {
      risk <- rep(risk, nrow(test_data))
    }
    risk <- as.numeric(risk)

    train_surv_matrix <- tryCatch(
      as.matrix(train_data$surv_obj),
      error = function(e)
        NULL
    )
    train_times_raw <- as.numeric(train_data[[time_col]])
    train_status_default <- normalize_status(train_data[[status_col]], length(train_times_raw))
    train_info <- get_surv_info(train_surv_matrix, train_times_raw, train_status_default)
    train_times <- train_info$time
    train_status_event <- ifelse(is.na(train_info$status),
                                 0,
                                 ifelse(train_info$status > 0, 1, 0))

    test_surv_matrix <- tryCatch(
      as.matrix(surv_obj),
      error = function(e)
        NULL
    )
    test_times_raw <- as.numeric(test_data[[time_col]])
    test_status_default <- normalize_status(test_data[[status_col]], length(test_times_raw))
    test_info <- get_surv_info(test_surv_matrix, test_times_raw, test_status_default)
    obs_time <- test_info$time
    status_event <- ifelse(is.na(test_info$status),
                           0,
                           ifelse(test_info$status > 0, 1, 0))

    t0 <- stats::median(train_times, na.rm = TRUE)
    threshold <- min(max(at_risk_threshold, 0.01), 0.5)

    if (identical(survival_metric_convention, "tidymodels")) {
      eval_horizons <- if (is.null(eval_times_user)) {
        numeric(0)
      } else {
        sort(unique(as.numeric(eval_times_user)))
      }
      eval_horizons <- eval_horizons[is.finite(eval_horizons) &
                                       eval_horizons >= 0]
      if (length(eval_horizons) == 0) {
        warning(
          "No valid eval_times supplied; time-dependent Brier metrics will be omitted.",
          call. = FALSE
        )
      }
      brier_times <- eval_horizons
      brier_metric_names <- if (length(brier_times) > 0) {
        paste0("brier_t", seq_along(brier_times))
      } else {
        character(0)
      }
      brier_time_map <- if (length(brier_times) > 0) {
        stats::setNames(brier_times, brier_metric_names)
      } else {
        numeric(0)
      }
      eval_times <- brier_times
      tau_max <- if (length(eval_times) > 0) {
        max(eval_times)
      } else {
        suppressWarnings(max(obs_time[is.finite(obs_time)], na.rm = TRUE))
      }
      if (!is.finite(tau_max) || tau_max <= 0) {
        tau_max <- NA_real_
      }
    } else {
      tau_max <- compute_tau_limit(obs_time, threshold)
      if (!is.finite(tau_max) || tau_max <= 0) {
        tau_max <- suppressWarnings(max(obs_time[is.finite(obs_time)], na.rm = TRUE))
      }
      if (!is.finite(tau_max) || tau_max <= 0) {
        tau_max <- NA_real_
      }

      digits_round <- determine_round_digits(obs_time)
      if (is.null(eval_times_user)) {
        eval_horizons <- unique(round(c(
          stats::median(obs_time, na.rm = TRUE),
          as.numeric(
            stats::quantile(obs_time, 0.75, na.rm = TRUE, names = FALSE)
          )
        ), digits_round))
        eval_horizons <- eval_horizons[is.finite(eval_horizons) &
                                         eval_horizons > 0]
      } else {
        eval_horizons <- sort(unique(as.numeric(eval_times_user)))
        eval_horizons <- eval_horizons[is.finite(eval_horizons) &
                                         eval_horizons > 0]
      }
      if (length(eval_horizons) > 0 && is.finite(tau_max)) {
        too_late <- eval_horizons > tau_max
        if (any(too_late)) {
          eval_horizons <- eval_horizons[!too_late]
          if (length(eval_horizons) == 0) {
            warning(
              "All requested eval_times exceed t_max; horizon-specific Brier scores will be omitted."
            )
          } else {
            warning(
              "Some requested eval_times exceed t_max and were removed for Brier score computation."
            )
          }
        }
      }
      brier_times <- eval_horizons
      brier_metric_names <- if (length(brier_times) > 0) {
        paste0("brier_t", seq_along(brier_times))
      } else {
        character(0)
      }
      brier_time_map <- if (length(brier_times) > 0) {
        stats::setNames(brier_times, brier_metric_names)
      } else {
        numeric(0)
      }

      combined_times <- c(train_times, obs_time)
      combined_times <- combined_times[is.finite(combined_times) &
                                         combined_times > 0]
      if (length(combined_times) > 0) {
        eval_times <- sort(unique(combined_times))
        if (length(eval_times) > 200) {
          probs <- seq(0, 1, length.out = 200)
          eval_times <- sort(unique(as.numeric(
            stats::quantile(
              eval_times,
              probs = probs,
              na.rm = TRUE,
              type = 1
            )
          )))
        }
      } else {
        eval_times <- numeric(0)
      }
      special_points <- c(brier_times, tau_max, t0)
      special_points <- special_points[is.finite(special_points) &
                                         special_points > 0]
      eval_times <- sort(unique(c(eval_times, special_points)))
      if (is.finite(tau_max)) {
        eval_times <- eval_times[eval_times <= tau_max + 1e-08]
      }
      if (length(eval_times) == 0 &&
          is.finite(tau_max) && tau_max > 0) {
        eval_times <- tau_max
      }
    }

    censor_eval_test <- create_censor_eval(obs_time, status_event)
    censor_eval_train <- create_censor_eval(train_times, train_status_event)

    n_obs <- nrow(test_data)
    surv_prob_mat <- NULL
    aft_quantile_mat <- NULL
    aft_probs <- numeric(0)
    aft_mu <- NULL
    if (length(eval_times) > 0 && n_obs > 0) {
      if (inherits(final_model, "fastml_native_survival")) {
        newdata_survfit <- pred_predictors
        if (is.null(newdata_survfit) ||
            nrow(newdata_survfit) != n_obs) {
          newdata_survfit <- tryCatch({
            if (is.null(pred_new_data)) {
              NULL
            } else {
              drop_cols <- c(
                final_model$response,
                final_model$time_col,
                final_model$status_col,
                final_model$start_col
              )
              drop_cols <- drop_cols[!is.na(drop_cols)]
              keep_cols <- setdiff(names(pred_new_data), drop_cols)
              pred_new_data[, keep_cols, drop = FALSE]
            }
          }, error = function(e)
            NULL)
        }
        if (is.null(newdata_survfit) && n_obs > 0) {
          newdata_survfit <- data.frame(matrix(nrow = n_obs, ncol = 0))
        }
        if (inherits(final_model$fit, "survreg")) {
          surv_prob_mat <- compute_survreg_matrix(final_model$fit, newdata_survfit, eval_times)
          if (is.null(surv_prob_mat)) {
            surv_fit <- tryCatch(
              survival::survfit(final_model$fit, newdata = newdata_survfit),
              error = function(e)
                NULL
            )
            surv_prob_mat <- build_survfit_matrix(surv_fit, eval_times, n_obs)
          }
        } else if (inherits(final_model$fit, "flexsurvreg")) {
          # flexsurvreg models expect the covariates used during fitting.
          # Prefer the preprocessed predictor set when available, but fall back
          # to the raw test data if preprocessing failed or yielded mismatched
          # rows (e.g. due to recipes that only keep the outcome columns).
          flexsurv_newdata <- newdata_survfit
          if (is.null(flexsurv_newdata) || nrow(flexsurv_newdata) != n_obs) {
            flexsurv_newdata <- test_data
          }
          surv_prob_mat <- fastml_flexsurv_survival_matrix(
            final_model$fit,
            flexsurv_newdata,
            eval_times
          )

          # If risk has not been computed yet, derive it from the survival probabilities
          if (!any(is.finite(risk)) &&
              !is.null(surv_prob_mat) &&
              ncol(surv_prob_mat) > 0) {
            mid_idx <- ceiling(ncol(surv_prob_mat) / 2)
            if (mid_idx > 0 && mid_idx <= ncol(surv_prob_mat)) {
              surv_mid <- surv_prob_mat[, mid_idx]
              risk <- -log(pmax(surv_mid, .Machine$double.eps))
            }
          }
        } else if (inherits(final_model$fit, "stpm2")) {
          if (requireNamespace("rstpm2", quietly = TRUE) &&
              length(eval_times) > 0) {
            base_newdata <- newdata_survfit
            if (is.null(base_newdata)) {
              base_newdata <- data.frame()
            }
            time_var <- final_model$time_col
            if (!is.null(time_var)) {
              if (!(time_var %in% names(base_newdata)) &&
                  time_var %in% names(test_data)) {
                base_newdata[[time_var]] <- test_data[[time_var]]
              }
              if (!(time_var %in% names(base_newdata))) {
                base_newdata[[time_var]] <- rep(default_time_val, n_obs)
              } else {
                base_newdata[[time_var]][!is.finite(base_newdata[[time_var]])] <- default_time_val
              }
            }
            if (!is.null(final_model$start_col) &&
                !(final_model$start_col %in% names(base_newdata)) &&
                final_model$start_col %in% names(test_data)) {
              base_newdata[[final_model$start_col]] <- test_data[[final_model$start_col]]
            }
            surv_prob_mat <- matrix(NA_real_,
                                    nrow = n_obs,
                                    ncol = length(eval_times))
            for (j in seq_along(eval_times)) {
              nd <- base_newdata
              if (!is.null(time_var)) {
                nd[[time_var]] <- eval_times[j]
              }
              preds <- tryCatch({
                as.numeric(predict(
                  final_model$fit,
                  newdata = nd,
                  type = "surv"
                ))
              }, error = function(e)
                rep(NA_real_, n_obs))
              if (length(preds) == n_obs) {
                surv_prob_mat[, j] <- preds
              }
            }
          }
        } else if (inherits(final_model$fit, "fastml_xgb_survival")) {
          # --- START: Corrected XGBoost AFT Block ---
          if (identical(final_model$fit$objective, "survival:cox")) {
            surv_prob_mat <- NULL

          } else if (identical(final_model$fit$objective, "survival:aft")) {

            # --- FIX: Robust Matrix Preparation (Same as risk calculation) ---
            expected_features <- final_model$fit$feature_names
            current_predictors <- pred_predictors
            if(is.null(current_predictors)) current_predictors <- data.frame()

            n_rows_pred <- nrow(current_predictors)
            if(n_rows_pred == 0 && nrow(test_data) > 0) n_rows_pred <- nrow(test_data)

            mat_data <- matrix(0, nrow = n_rows_pred, ncol = length(expected_features))
            colnames(mat_data) <- expected_features

            common_cols <- intersect(names(current_predictors), expected_features)
            if(length(common_cols) > 0) {
              for(col in common_cols) {
                val <- current_predictors[[col]]
                if(is.factor(val) || is.character(val)) val <- as.numeric(as.factor(val))
                mat_data[, col] <- as.numeric(val)
              }
            }

            predictor_matrix <- mat_data # Use the safe matrix

            # Create DMatrix
            dtest <- tryCatch(
              xgboost::xgb.DMatrix(data = predictor_matrix, missing = NA),
              error = function(e) NULL
            )

            lp <- NULL
            if(!is.null(dtest)) {
              lp <- tryCatch(predict(final_model$fit$booster, dtest), error = function(e) NULL)
            }
            # --- END OF FIX ---

            if(!is.null(lp)) {
              # Get model parameters stored by fastml.
              dist <- final_model$fit$aft_distribution
              scale <- final_model$fit$aft_scale

              # Calculate the survival probability matrix.
              if (dist == "logistic") {
                surv_prob_mat <- matrix(NA_real_, nrow = n_obs, ncol = length(eval_times))
                for (i in seq_along(eval_times)) {
                  t <- eval_times[i]
                  if (t > 0) {
                    z <- (log(t) - lp) / scale
                    surv_prob_mat[, i] <- 1 / (1 + exp(z))
                  } else {
                    surv_prob_mat[, i] <- 1
                  }
                }
              } else {
                warning(paste("AFT distribution", dist, "is not supported. Skipping survival matrix."))
                surv_prob_mat <- NULL
              }
              aft_mu <- lp
            } else {
              surv_prob_mat <- NULL
            }

          } else {
            surv_prob_mat <- NULL
          }
          # --- END: Corrected XGBoost AFT Block ---
        } else {
          surv_fit <- tryCatch(
            survival::survfit(final_model$fit, newdata = newdata_survfit),
            error = function(e)
              NULL
          )
          surv_prob_mat <- build_survfit_matrix(surv_fit, eval_times, n_obs)
        }
      } else {
        surv_pred <- tryCatch(
          predict(
            final_model,
            new_data = test_data,
            type = "survival",
            eval_time = eval_times
          ),
          error = function(e)
            NULL
        )
        if (is.null(surv_pred) && length(eval_times) == 1) {
          surv_pred <- tryCatch(
            predict(
              final_model,
              new_data = test_data,
              type = "survival",
              eval_time = eval_times[1]
            ),
            error = function(e)
              NULL
          )
        }
        surv_prob_mat <- convert_survival_predictions(surv_pred, eval_times, n_obs)
      }
    }

    surv_prob <- rep(NA_real_, n_obs)
    if (!is.null(surv_prob_mat) &&
        nrow(surv_prob_mat) == n_obs &&
        ncol(surv_prob_mat) == length(eval_times) &&
        length(eval_times) > 0) {
      idx_t0 <- 1L
      if (is.finite(t0) && t0 > 0) {
        idx_t0 <- which.min(abs(eval_times - t0))
      }
      idx_t0 <- max(1L, min(idx_t0, ncol(surv_prob_mat)))
      surv_prob <- as.numeric(surv_prob_mat[, idx_t0])
    }

    if (identical(survival_model_type, "rstpm2")) {
      if (length(surv_prob) == n_obs && any(is.finite(surv_prob))) {
        risk <- -log(pmax(surv_prob, .Machine$double.eps))
      }
    } else if ((!any(is.finite(risk)) ||
                all(is.na(risk))) &&
               length(surv_prob) == n_obs && any(is.finite(surv_prob))) {
      risk <- -log(pmax(surv_prob, .Machine$double.eps))
    }

    if (identical(survival_model_type, "xgboost_aft") &&
        (!any(is.finite(risk)) || all(is.na(risk)))) {
      if (!is.null(aft_mu) && length(aft_mu) == n_obs) {
        aft_mu_vec <- as.numeric(aft_mu)
        if (any(is.finite(aft_mu_vec))) {
          risk <- -aft_mu_vec
        }
      }
    }
    if (length(risk) != n_obs) {
      risk <- rep(risk, length.out = n_obs)
    }
    risk[!is.finite(risk)] <- NA_real_

    # Retry computing risk for xgboost AFT when everything is missing.
    if (!any(is.finite(risk)) &&
        inherits(final_model, "fastml_native_survival") &&
        inherits(final_model$fit, "fastml_xgb_survival") &&
        identical(final_model$fit$objective, "survival:aft")) {

      # --- FIX: Robust Matrix Preparation in Retry Block ---
      expected_features <- final_model$fit$feature_names
      retry_predictors <- pred_predictors
      if (is.null(retry_predictors)) retry_predictors <- data.frame()

      n_rows_pred <- nrow(retry_predictors)
      if(n_rows_pred == 0 && nrow(test_data) > 0) n_rows_pred <- nrow(test_data)

      mat_data <- matrix(0, nrow = n_rows_pred, ncol = length(expected_features))
      colnames(mat_data) <- expected_features

      common_cols <- intersect(names(retry_predictors), expected_features)
      if(length(common_cols) > 0) {
        for(col in common_cols) {
          val <- retry_predictors[[col]]
          if(is.factor(val) || is.character(val)) val <- as.numeric(as.factor(val))
          mat_data[, col] <- as.numeric(val)
        }
      }

      dretry <- tryCatch(
        xgboost::xgb.DMatrix(data = mat_data, missing = NA),
        error = function(e) NULL
      )

      lp_retry <- if(!is.null(dretry)) {
        tryCatch(
          predict(final_model$fit$booster, dretry),
          error = function(e) NULL
        )
      } else {
        NULL
      }
      # --- END FIX ---
      if (!is.null(lp_retry)) {
        lp_retry <- as.numeric(lp_retry)
        if (length(lp_retry) != n_obs && length(lp_retry) > 0) {
          lp_retry <- rep(lp_retry, length.out = n_obs)
        }
        if (length(lp_retry) == n_obs && any(is.finite(lp_retry))) {
          risk <- -lp_retry
        }
      }
    }

    # Absolute fallback: if risk is entirely missing, use a flat baseline so metrics stay defined.
    if (!any(is.finite(risk)) && n_obs > 0) {
      warning(
        "Predicted risk scores are all missing; using a flat baseline for metrics. Check preprocessing and feature alignment.",
        call. = FALSE
      )
      risk <- rep(0, n_obs)
    }

    # Align risk direction: higher values should imply worse survival.
    # If predicted risk is positively associated with survival probability,
    # flip its sign to keep concordance metrics consistent.
    if (length(risk) == n_obs &&
        length(surv_prob) == n_obs &&
        any(is.finite(risk)) &&
        any(is.finite(surv_prob))) {
      cor_dir <- suppressWarnings(stats::cor(risk, surv_prob,
                                             use = "pairwise.complete.obs"))
      if (is.finite(cor_dir) && cor_dir > 0) {
        risk <- -risk
      }
    }

    surv_curve_list <- vector("list", n_obs)
    if (!is.null(surv_prob_mat) &&
        nrow(surv_prob_mat) == n_obs &&
        ncol(surv_prob_mat) == length(eval_times) &&
        length(eval_times) > 0) {
      formatted_times <- format(eval_times, trim = TRUE, scientific = FALSE)
      for (i in seq_len(n_obs)) {
        row_vals <- as.numeric(surv_prob_mat[i, ])
        names(row_vals) <- formatted_times
        surv_curve_list[[i]] <- row_vals
      }
    }
    attr(surv_curve_list, "eval_times") <- eval_times

    limited_metrics <- identical(survival_model_type, "xgboost_cox")
    if (limited_metrics) {
      attr(surv_curve_list, "eval_times") <- numeric(0)
      brier_times <- numeric(0)
      brier_metric_names <- character(0)
      brier_time_map <- numeric(0)
    }

    surv_time <- rep(NA_real_, n_obs)
    if (!is.null(surv_prob_mat) &&
        nrow(surv_prob_mat) == n_obs &&
        ncol(surv_prob_mat) > 0) {

      median_indices <- apply(surv_prob_mat, 1, function(surv_row) {
        idx <- which(surv_row <= 0.5)
        if (length(idx) > 0) {
          min(idx)
        } else {
          NA_integer_
        }
      })

      valid_indices <- !is.na(median_indices)
      if (any(valid_indices)) {
        surv_time[valid_indices] <- eval_times[median_indices[valid_indices]]
      }
    }

    if (identical(survival_model_type, "xgboost_aft")) {
      if (!is.null(aft_quantile_mat) &&
          nrow(aft_quantile_mat) == n_obs && length(aft_probs) > 0) {
        median_idx <- which.min(abs(aft_probs - 0.5))
        if (length(median_idx) == 0) {
          median_idx <- ceiling(length(aft_probs) / 2)
        } else {
          median_idx <- median_idx[1]
        }
        median_vals <- as.numeric(aft_quantile_mat[, median_idx, drop = TRUE])
        surv_time <- median_vals
        surv_time[!is.finite(surv_time)] <- NA_real_
      } else if (!is.null(aft_mu) && length(aft_mu) == n_obs) {
        surv_time <- exp(as.numeric(aft_mu))
        surv_time[!is.finite(surv_time)] <- NA_real_
      }
    } else if (identical(survival_model_type, "xgboost_cox")) {
      surv_time <- rep(NA_real_, n_obs)
    }

    risk_group <- assign_risk_group(risk)
    ibrier_normalize <- if (identical(survival_metric_convention, "tidymodels")) {
      "n"
    } else {
      "non_missing"
    }
    ibrier_include_zero <- !identical(survival_metric_convention, "tidymodels")

    ibs_curve_full <- rep(NA_real_, length(eval_times))
    ibs_point <- NA_real_
    brier_time_values <- if (length(brier_metric_names) > 0)
      rep(NA_real_, length(brier_metric_names))
    else
      numeric(0)
    rmst_diff <- NA_real_
    if (!limited_metrics &&
        !is.null(surv_prob_mat) && length(eval_times) > 0) {
      ibs_res_full <- compute_ibrier(eval_times,
                                     surv_prob_mat,
                                     obs_time,
                                     status_event,
                                     tau_max,
                                     censor_eval_test,
                                     normalize_by = ibrier_normalize,
                                     include_zero = ibrier_include_zero)
      ibs_point <- ibs_res_full$ibs
      ibs_curve_full <- ibs_res_full$curve
      if (identical(survival_metric_convention, "tidymodels") &&
          length(eval_times) < 2) {
        ibs_point <- NA_real_
      }
      brier_time_values <- map_brier_values(ibs_curve_full, eval_times, brier_times)
      rmst_diff <- compute_rmst_difference(
        obs_time,
        status_event,
        risk,
        tau_max,
        surv_mat = surv_prob_mat,
        eval_times_full = eval_times,
        model_type = survival_model_type
      )
    }

    harrell_c <- NA_real_
    risk_valid <- is.finite(risk) & is.finite(obs_time) & is.finite(status_event)
    if (any(risk_valid)) {
      risk_vals <- risk[risk_valid]
      if (length(unique(risk_vals)) > 1) {
        surv_valid <- surv_obj[risk_valid]
        harrell_c <- tryCatch(
          survival::concordance(surv_valid ~ risk_vals)$concordance,
          error = function(e)
            NA_real_
        )
      } else {
        harrell_c <- 0.5
      }
    }
    harrell_c <- clamp01(harrell_c)

    uno_c <- compute_uno_c_index(
      train_times,
      train_status_event,
      obs_time,
      status_event,
      risk,
      tau_max,
      censor_eval_train
    )
    uno_c <- clamp01(uno_c)

    # If Harrell and Uno disagree on direction (e.g., Harrell < 0.5 but Uno > 0.5),
    # mirror Harrell so that higher still means worse survival.
    if (is.finite(harrell_c) && is.finite(uno_c)) {
      if (harrell_c < 0.5 && uno_c > 0.5) {
        harrell_c <- 1 - harrell_c
      } else if (harrell_c > 0.5 && uno_c < 0.5) {
        harrell_c <- 1 - harrell_c
      }
    }
    if (!limited_metrics) {
      ibs_point <- clamp01(ibs_point)
      if (length(brier_time_values) > 0) {
        brier_time_values <- clamp01(brier_time_values)
      }
    }

    if (limited_metrics) {
      metric_names <- c("c_index", "uno_c")
      metrics_point <- c(harrell_c, uno_c)
    } else {
      metric_names <- c("c_index",
                        "uno_c",
                        "ibs",
                        "rmst_diff",
                        brier_metric_names)
      metrics_point <- c(harrell_c,
                         uno_c,
                         ibs_point,
                         rmst_diff,
                         brier_time_values)
    }
    names(metrics_point) <- metric_names
    metrics_point[is.nan(metrics_point)] <- NA_real_

    compute_metrics_boot <- function(idx) {
      idx <- as.integer(idx)
      obs_sub <- obs_time[idx]
      status_sub <- status_event[idx]
      risk_sub <- risk[idx]
      surv_sub <- surv_obj[idx]
      surv_mat_sub <- if (!is.null(surv_prob_mat))
        surv_prob_mat[idx, , drop = FALSE]
      else
        NULL
      harrell_sub <- NA_real_
      risk_valid_sub <- is.finite(risk_sub) &
        is.finite(obs_sub) &
        is.finite(status_sub)
      if (any(risk_valid_sub)) {
        risk_vals_sub <- risk_sub[risk_valid_sub]
        if (length(unique(risk_vals_sub)) > 1) {
          surv_valid_sub <- surv_sub[risk_valid_sub]
          harrell_sub <- tryCatch(
            survival::concordance(surv_valid_sub ~ risk_vals_sub)$concordance,
            error = function(e)
              NA_real_
          )
        } else {
          harrell_sub <- 0.5
        }
      }
      harrell_sub <- clamp01(harrell_sub)
      uno_sub <- compute_uno_c_index(
        train_times,
        train_status_event,
        obs_sub,
        status_sub,
        risk_sub,
        tau_max,
        censor_eval_train
      )
      uno_sub <- clamp01(uno_sub)

      if (is.finite(harrell_sub) && is.finite(uno_sub)) {
        if (harrell_sub < 0.5 && uno_sub > 0.5) {
          harrell_sub <- 1 - harrell_sub
        } else if (harrell_sub > 0.5 && uno_sub < 0.5) {
          harrell_sub <- 1 - harrell_sub
        }
      }
      if (limited_metrics) {
        vals <- c(harrell_sub, uno_sub)
        names(vals) <- metric_names
        vals[is.nan(vals)] <- NA_real_
        return(vals)
      }
      ibs_sub <- NA_real_
      curve_sub <- rep(NA_real_, length(eval_times))
      if (!is.null(surv_mat_sub) &&
          nrow(surv_mat_sub) == length(idx)) {
        ibs_sub_res <- compute_ibrier(eval_times,
                                      surv_mat_sub,
                                      obs_sub,
                                      status_sub,
                                      tau_max,
                                      censor_eval_test,
                                      normalize_by = ibrier_normalize,
                                      include_zero = ibrier_include_zero)
        ibs_sub <- clamp01(ibs_sub_res$ibs)
        curve_sub <- ibs_sub_res$curve
        if (identical(survival_metric_convention, "tidymodels") &&
            length(eval_times) < 2) {
          ibs_sub <- NA_real_
        }
      }
      brier_sub <- map_brier_values(curve_sub, eval_times, brier_times)
      if (length(brier_sub) > 0) {
        brier_sub <- clamp01(brier_sub)
      }
      rmst_sub <- compute_rmst_difference(
        obs_sub,
        status_sub,
        risk_sub,
        tau_max,
        surv_mat = surv_mat_sub,
        eval_times_full = eval_times,
        model_type = survival_model_type
      )
      vals <- c(harrell_sub, uno_sub, ibs_sub, rmst_sub, brier_sub)
      names(vals) <- metric_names
      vals[is.nan(vals)] <- NA_real_
      vals
    }

    ci_lower <- rep(NA_real_, length(metric_names))
    ci_upper <- rep(NA_real_, length(metric_names))
    n_boot_used <- 0L
    if (isTRUE(bootstrap_ci) &&
        bootstrap_samples > 1 && n_obs > 1) {
      .boot_old_seed2 <- if (exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE))
        get(".Random.seed", envir = .GlobalEnv, inherits = FALSE) else NULL
      .boot_had_seed2 <- !is.null(.boot_old_seed2)
      if (!is.null(bootstrap_seed)) {
        set.seed(bootstrap_seed)
      }
      boot_matrix <- matrix(NA_real_,
                            nrow = bootstrap_samples,
                            ncol = length(metric_names))
      colnames(boot_matrix) <- metric_names
      for (b in seq_len(bootstrap_samples)) {
        idx <- sample.int(n_obs, size = n_obs, replace = TRUE)
        boot_matrix[b, ] <- compute_metrics_boot(idx)
      }
      for (j in seq_along(metric_names)) {
        vals <- boot_matrix[, j]
        vals <- vals[is.finite(vals)]
        if (length(vals) > 0) {
          ci_lower[j] <- stats::quantile(vals,
                                         probs = 0.025,
                                         names = FALSE,
                                         na.rm = TRUE)
          ci_upper[j] <- stats::quantile(vals,
                                         probs = 0.975,
                                         names = FALSE,
                                         na.rm = TRUE)
        }
      }
      clamp_idx <- metric_names %in% c("c_index", "uno_c", "ibs") |
        grepl("^brier_t", metric_names)
      if (any(clamp_idx)) {
        ci_lower[clamp_idx] <- clamp01(ci_lower[clamp_idx])
        ci_upper[clamp_idx] <- clamp01(ci_upper[clamp_idx])
      }
      n_boot_used <- bootstrap_samples
      if (.boot_had_seed2) {
        assign(".Random.seed", .boot_old_seed2, envir = .GlobalEnv)
      } else if (exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE)) {
        rm(".Random.seed", envir = .GlobalEnv)
      }
    }

    perf <- tibble::tibble(
      .metric = metric_names,
      .estimate = as.numeric(metrics_point),
      .lower = as.numeric(ci_lower),
      .upper = as.numeric(ci_upper),
      .n_boot = n_boot_used
    )

    perf$.estimate <- as.numeric(perf$.estimate)

    if (!is.null(start_col) && start_col %in% names(test_data)) {
      data_metrics <- tibble::tibble(
        start = test_data[[start_col]],
        time = test_data[[time_col]],
        status = test_data[[status_col]],
        risk = risk,
        risk_group = risk_group,
        surv_prob = surv_prob,
        surv_time = surv_time,
        surv_prob_curve = surv_curve_list
      )
    } else {
      data_metrics <- tibble::tibble(
        time = test_data[[time_col]],
        status = test_data[[status_col]],
        risk = risk,
        risk_group = risk_group,
        surv_prob = surv_prob,
        surv_time = surv_time,
        surv_prob_curve = surv_curve_list
      )
    }

    attr(perf, "brier_times") <- brier_time_map
    attr(perf, "t_max") <- tau_max
    attr(perf, "at_risk_threshold") <- threshold
    attr(perf, "survival_metric_convention") <- survival_metric_convention
    attr(data_metrics, "eval_times") <- if (limited_metrics)
      numeric(0)
    else
      eval_times
    if (limited_metrics) {
      attr(data_metrics, "brier_curve") <- NULL
      attr(data_metrics, "brier_times") <- numeric(0)
    } else {
      attr(data_metrics, "brier_curve") <- tibble::tibble(eval_time = eval_times, brier = ibs_curve_full)
      attr(data_metrics, "brier_times") <- brier_time_map
    }
    attr(data_metrics, "t_max") <- tau_max
    attr(data_metrics, "survival_metric_convention") <- survival_metric_convention

  } else {
    # Regression task
    predictions <- predict(final_model, new_data = test_data)
    pred <- predictions$.pred
    data_metrics <- tibble::tibble(truth = test_data[[label]], estimate = pred)
    metrics_set <- if (is.null(metrics)) {
      yardstick::metric_set(yardstick::rmse, yardstick::rsq, yardstick::mae)
    } else {
      if (!is.function(metrics)) {
        stop("'metrics' must be a yardstick::metric_set() function or NULL.")
      }
      metrics
    }
    safe_metrics_regression <- function(df, ...) {
      withCallingHandlers(
        metrics_set(df, ...),
        warning = function(w) {
          if (inherits(w, "yardstick_warning_correlation_undefined")) {
            invokeRestart("muffleWarning")
          }
        }
      )
    }
    perf <- safe_metrics_regression(data_metrics, truth = truth, estimate = estimate)
    compute_boot_perf <- function(df) {
      tryCatch(
        safe_metrics_regression(df, truth = truth, estimate = estimate),
        error = function(e) perf[0, , drop = FALSE]
      )
    }
    perf <- add_bootstrap_ci(perf, data_metrics, compute_boot_perf)
  }

  return(list(performance = perf, predictions = data_metrics))
}

#' Finalize a tuned workflow
#'
#' Selects the best hyperparameter configuration from a `tune_results` object and
#' refits the corresponding workflow on the full training data. If the requested
#' metric is unavailable, other metrics are tried in a task-specific priority
#' order and the substitution is reported with a warning.
#'
#' @param model_obj A `tune_results` object.
#' @param model_id Identifier used in diagnostic messages.
#' @param task One of `"classification"`, `"regression"`, or `"survival"`.
#' @param metric Name of the metric used to select the best configuration.
#' @param train_data Training data used to refit the finalized workflow.
#'
#' @return A fitted workflow, or `NULL` (with a warning) when no configuration
#'   could be selected.
#'
#' @keywords internal
#' @noRd
finalize_tuned_model <- function(model_obj, model_id, task, metric, train_data) {
  select_metric <- metric
  select_best_safe <- function(metric_name) {
    tryCatch(tune::select_best(model_obj, metric = metric_name),
             error = function(e) e)
  }
  best_params <- select_best_safe(select_metric)
  select_error <- NULL
  if (inherits(best_params, "error")) {
    select_error <- best_params
    available_metrics <- tryCatch(
      unique(tune::collect_metrics(model_obj)$.metric),
      error = function(e) NULL
    )
    if (!is.null(available_metrics) && length(available_metrics) > 0) {
      fallback_priority <- switch(task,
        classification = c("logloss", "brier_score", "roc_auc", "accuracy", "kap",
                           "sens", "spec", "precision", "f_meas"),
        regression = c("rmse", "rsq", "mae"),
        survival = c("c_index", "uno_c", "ibs", "rmst_diff"),
        c()
      )
      candidate_metrics <- c(
        fallback_priority[fallback_priority %in% available_metrics],
        setdiff(available_metrics, fallback_priority)
      )
      candidate_metrics <- unique(setdiff(candidate_metrics, select_metric))
      for (cand in candidate_metrics) {
        res <- select_best_safe(cand)
        if (!inherits(res, "error")) {
          best_params <- res
          warning(sprintf(
            "Metric '%s' unavailable for model '%s'; selecting best by '%s'.",
            select_metric,
            model_id,
            cand
          ), call. = FALSE)
          break
        }
      }
    }
  }
  if (inherits(best_params, "error") || is.null(best_params)) {
    warning(paste(
      "Could not select best parameters for model",
      model_id,
      ":",
      conditionMessage(select_error)
    ))
    return(NULL)
  }

  model_spec <- workflows::pull_workflow_spec(model_obj)
  model_recipe <- workflows::pull_workflow_preprocessor(model_obj)

  final_model_spec <- tune::finalize_model(model_spec, best_params)
  final_workflow <- workflows::workflow() %>%
    workflows::add_recipe(model_recipe) %>%
    workflows::add_model(final_model_spec)

  parsnip::fit(final_workflow, data = train_data)
}

library(testthat)

# Regression tests for the three failure-handling contracts in process_model():
#   1. a stored recipe that cannot be applied aborts rather than silently
#      falling back to unpreprocessed data;
#   2. risk-prediction failures surface the original error message;
#   3. tuning finalization is factored into finalize_tuned_model(), which
#      returns NULL (with a warning) when no configuration can be selected.

# A tune_results object whose metrics cannot be collected: selection fails for
# the requested metric and for every fallback candidate. Feeding tune a malformed
# object also emits incidental warnings of its own, so warnings are collected and
# inspected rather than matched one at a time.
broken_tune_results <- function() structure(list(a = 1), class = "tune_results")

collect_warnings <- function(expr) {
  warns <- character()
  value <- withCallingHandlers(
    expr,
    warning = function(w) {
      warns <<- c(warns, conditionMessage(w))
      invokeRestart("muffleWarning")
    }
  )
  list(value = value, warnings = warns)
}

test_that("finalize_tuned_model returns NULL and warns when selection fails", {
  res <- collect_warnings(
    fastml:::finalize_tuned_model(
      model_obj = broken_tune_results(),
      model_id = "broken_model",
      task = "classification",
      metric = "roc_auc",
      train_data = data.frame(y = factor(c("a", "b")), x = c(1, 2))
    )
  )

  expect_null(res$value)
  expect_true(any(grepl("Could not select best parameters", res$warnings, fixed = TRUE)))
})

test_that("finalize_tuned_model names the offending model in its warning", {
  res <- collect_warnings(
    fastml:::finalize_tuned_model(
      broken_tune_results(), "rand_forest_ranger", "regression", "rmse",
      data.frame(y = c(1, 2), x = c(1, 2))
    )
  )

  expect_true(any(grepl("rand_forest_ranger", res$warnings, fixed = TRUE)))
})

test_that("process_model aborts when the stored recipe cannot be applied", {
  skip_if_not_installed("recipes")
  skip_if_not_installed("survival")

  set.seed(42)
  n <- 60
  train_data <- data.frame(
    time = rexp(n, 0.1),
    status = rbinom(n, 1, 0.7),
    x = rnorm(n)
  )

  rec <- recipes::recipe(~ x, data = train_data)
  rec_prep <- recipes::prep(rec, training = train_data, retain = TRUE)

  fit <- survival::coxph(
    survival::Surv(time, status) ~ x,
    data = train_data
  )

  native <- structure(
    list(
      algo = "cox_ph", engine = "survival", fit = fit, recipe = rec_prep,
      response = "surv", label_cols = NULL,
      time_col = "time", status_col = "status", start_col = NULL
    ),
    class = c("fastml_native_survival", "fastml_model")
  )

  # Evaluation data missing the predictor the recipe requires.
  bad_test <- data.frame(
    time = rexp(20, 0.1),
    status = rbinom(20, 1, 0.7),
    wrong_name = rnorm(20)
  )

  expect_error(
    fastml:::process_model(
      model_obj = native, model_id = "cox_ph_survival", task = "survival",
      test_data = bad_test, label = "surv", event_class = "first",
      time_col = "time", status_col = "status",
      engine = "survival", train_data = train_data, metric = "c_index",
      bootstrap_ci = FALSE
    ),
    "Could not apply the stored preprocessing recipe"
  )
})

test_that("the recipe-failure message explains why there is no fallback", {
  skip_if_not_installed("recipes")
  skip_if_not_installed("survival")

  set.seed(43)
  n <- 60
  train_data <- data.frame(
    time = rexp(n, 0.1),
    status = rbinom(n, 1, 0.7),
    x = rnorm(n)
  )
  rec_prep <- recipes::prep(
    recipes::recipe(~ x, data = train_data),
    training = train_data, retain = TRUE
  )
  fit <- survival::coxph(survival::Surv(time, status) ~ x, data = train_data)

  native <- structure(
    list(
      algo = "cox_ph", engine = "survival", fit = fit, recipe = rec_prep,
      response = "surv", label_cols = NULL,
      time_col = "time", status_col = "status", start_col = NULL
    ),
    class = c("fastml_native_survival", "fastml_model")
  )

  bad_test <- data.frame(
    time = rexp(20, 0.1), status = rbinom(20, 1, 0.7), wrong_name = rnorm(20)
  )

  err <- tryCatch(
    fastml:::process_model(
      model_obj = native, model_id = "cox_ph_survival", task = "survival",
      test_data = bad_test, label = "surv", event_class = "first",
      time_col = "time", status_col = "status",
      engine = "survival", train_data = train_data, metric = "c_index",
      bootstrap_ci = FALSE
    ),
    error = function(e) conditionMessage(e)
  )

  # Names the model, and states that no unpreprocessed fallback is attempted.
  expect_match(err, "cox_ph_survival")
  expect_match(err, "does not fall back to")
})

test_that("a well-formed native survival model still evaluates cleanly", {
  skip_if_not_installed("recipes")
  skip_if_not_installed("survival")

  set.seed(44)
  n <- 120
  make_data <- function(m) {
    data.frame(time = rexp(m, 0.1), status = rbinom(m, 1, 0.7), x = rnorm(m))
  }
  train_data <- make_data(n)
  test_data <- make_data(40)

  rec_prep <- recipes::prep(
    recipes::recipe(~ x, data = train_data),
    training = train_data, retain = TRUE
  )
  fit <- survival::coxph(survival::Surv(time, status) ~ x, data = train_data)

  native <- structure(
    list(
      algo = "cox_ph", engine = "survival", fit = fit, recipe = rec_prep,
      response = "surv", label_cols = NULL,
      time_col = "time", status_col = "status", start_col = NULL
    ),
    class = c("fastml_native_survival", "fastml_model")
  )

  res <- fastml:::process_model(
    model_obj = native, model_id = "cox_ph_survival", task = "survival",
    test_data = test_data, label = "surv", event_class = "first",
    time_col = "time", status_col = "status",
    engine = "survival", train_data = train_data, metric = "c_index",
    bootstrap_ci = FALSE
  )

  expect_type(res, "list")
  expect_true("performance" %in% names(res))
  expect_true(nrow(res$performance) > 0)
})

#' Internal helpers for piecewise exponential survival models
#'
#' These utilities construct a custom piecewise exponential distribution for
#' flexsurv::flexsurvreg() along with supporting density, distribution,
#' quantile, hazard, and simulation routines. They are internal to fastml.
#'
#' @noRd

fastml_piecewise_make_distribution <- function(breaks = NULL) {
  cuts <- fastml_piecewise_normalize_breaks(breaks)
  n_intervals <- length(cuts) + 1L
  param_names <- c("log_rate", if (n_intervals > 1L) {
    paste0("log_ratio_", seq_len(n_intervals - 1L))
  })
  transforms <- replicate(length(param_names), identity, simplify = FALSE)
  inv_transforms <- transforms

  inits_fun <- function(t) {
    tt <- as.numeric(t)
    tt <- tt[is.finite(tt) & tt > 0]
    base_log_rate <- if (length(tt) == 0) 0 else log(1 / mean(tt))
    c(base_log_rate, rep(0, length(param_names) - 1L))
  }

  dlist <- list(
    name = "fastml_piecewise_exponential",
    pars = param_names,
    location = "log_rate",
    transforms = transforms,
    inv.transforms = inv_transforms,
    inits = inits_fun
  )

  dfns <- fastml_piecewise_dfns_factory(cuts, param_names)

  aux <- if (length(cuts) > 0) list(cuts = cuts) else list()

  list(dlist = dlist, dfns = dfns, aux = aux)
}

fastml_piecewise_normalize_breaks <- function(breaks) {
  if (is.null(breaks) || length(breaks) == 0) {
    return(numeric(0))
  }
  vals <- as.numeric(breaks)
  vals <- vals[is.finite(vals) & vals > 0]
  vals <- sort(unique(vals))
  vals
}

fastml_piecewise_dfns_factory <- function(cuts, param_names) {
  delta_names <- if (length(param_names) > 1L) param_names[-1L] else character(0)

  list(
    d = fastml_piecewise_density_fn(cuts, delta_names),
    p = fastml_piecewise_cdf_fn(cuts, delta_names),
    q = fastml_piecewise_quantile_fn(cuts, delta_names),
    r = fastml_piecewise_random_fn(cuts, delta_names),
    h = fastml_piecewise_hazard_fn(cuts, delta_names),
    H = fastml_piecewise_cumhaz_fn(cuts, delta_names)
  )
}

fastml_piecewise_extract_deltas <- function(extra, delta_names) {
  if (length(delta_names) == 0L) {
    return(numeric(0))
  }
  vapply(delta_names, function(nm) {
    val <- extra[[nm]]
    if (is.null(val)) {
      0
    } else {
      as.numeric(val)[1]
    }
  }, numeric(1))
}

fastml_piecewise_rates <- function(log_rate, log_ratios) {
  log_rate <- as.numeric(log_rate)
  log_ratios <- as.numeric(log_ratios)

  if (length(log_rate) > 1L) {
    log_ratios <- c(log_rate[-1L], log_ratios)
    log_rate <- log_rate[1L]
  }

  pars <- c(log_rate, log_ratios)
  if (any(!is.finite(pars))) {
    return(rep(NA_real_, length(log_ratios) + 1L))
  }

  base_rate <- exp(log_rate)
  c(base_rate, base_rate * exp(log_ratios))
}

fastml_piecewise_interval_starts <- function(cuts) {
  if (length(cuts) == 0L) {
    return(0)
  }
  c(0, cuts)
}

fastml_piecewise_cumhaz_starts <- function(rates, cuts) {
  if (length(rates) <= 1L || length(cuts) == 0L) {
    return(rep(0, length(rates)))
  }
  seg_lengths <- diff(c(0, cuts))
  if (length(seg_lengths) >= length(rates)) {
    seg_lengths <- seg_lengths[seq_len(length(rates) - 1L)]
  }
  accum <- cumsum(rates[seq_len(length(seg_lengths))] * seg_lengths)
  c(0, accum)
}

fastml_piecewise_cumhaz_core <- function(times, rates, cuts) {
  times <- as.numeric(times)
  res <- rep(NA_real_, length(times))
  if (length(times) == 0L) {
    return(res)
  }
  if (any(is.na(rates))) {
    return(res)
  }
  idx_valid <- which(is.finite(times) & times >= 0)
  if (length(idx_valid) > 0L) {
    tvals <- times[idx_valid]
    interval_idx <- findInterval(tvals, cuts) + 1L
    starts <- fastml_piecewise_interval_starts(cuts)
    cumhaz_starts <- fastml_piecewise_cumhaz_starts(rates, cuts)
    contrib <- rates[interval_idx] * (tvals - starts[interval_idx])
    res[idx_valid] <- cumhaz_starts[interval_idx] + contrib
  }
  res[is.finite(times) & times < 0] <- 0
  res
}

fastml_piecewise_survival_core <- function(times, rates, cuts) {
  H <- fastml_piecewise_cumhaz_core(times, rates, cuts)
  ifelse(is.na(H), NA_real_, exp(-pmax(H, 0)))
}

fastml_piecewise_density_core <- function(times, rates, cuts) {
  surv <- fastml_piecewise_survival_core(times, rates, cuts)
  times <- as.numeric(times)
  dens <- rep(NA_real_, length(times))
  idx_valid <- which(is.finite(times) & times >= 0)
  if (length(idx_valid) > 0L && !any(is.na(rates))) {
    tvals <- times[idx_valid]
    interval_idx <- findInterval(tvals, cuts) + 1L
    dens[idx_valid] <- rates[interval_idx] * surv[idx_valid]
  }
  dens[is.finite(times) & times < 0] <- 0
  dens
}

fastml_piecewise_cdf_core <- function(times, rates, cuts) {
  surv <- fastml_piecewise_survival_core(times, rates, cuts)
  ifelse(is.na(surv), NA_real_, pmin(pmax(1 - surv, 0), 1))
}

fastml_piecewise_hazard_core <- function(times, rates, cuts) {
  times <- as.numeric(times)
  haz <- rep(NA_real_, length(times))
  idx_valid <- which(is.finite(times) & times >= 0)
  if (length(idx_valid) > 0L && !any(is.na(rates))) {
    tvals <- times[idx_valid]
    interval_idx <- findInterval(tvals, cuts) + 1L
    haz[idx_valid] <- rates[interval_idx]
  }
  haz[is.finite(times) & times < 0] <- 0
  haz
}

fastml_piecewise_quantile_core <- function(probs, rates, cuts) {
  probs <- as.numeric(probs)
  res <- rep(NA_real_, length(probs))
  if (length(probs) == 0L || any(is.na(rates))) {
    return(res)
  }
  invalid <- probs < 0 | probs > 1
  res[invalid] <- NaN
  zero_idx <- which(!invalid & probs <= 0)
  if (length(zero_idx) > 0L) {
    res[zero_idx] <- 0
  }
  one_idx <- which(!invalid & probs >= 1)
  if (length(one_idx) > 0L) {
    res[one_idx] <- Inf
  }
  idx <- which(!invalid & probs > 0 & probs < 1)
  if (length(idx) > 0L) {
    target <- -log1p(-probs[idx])
    cumhaz_starts <- fastml_piecewise_cumhaz_starts(rates, cuts)
    cumhaz_ends <- if (length(rates) > 1L && length(cuts) > 0L) {
      cumhaz_starts[-1L]
    } else {
      numeric(0)
    }
    cumhaz_bounds <- c(cumhaz_ends, Inf)
    interval_idx <- findInterval(target, cumhaz_bounds) + 1L
    starts <- fastml_piecewise_interval_starts(cuts)
    res[idx] <- starts[interval_idx] +
      (target - cumhaz_starts[interval_idx]) / rates[interval_idx]
  }
  res
}

fastml_piecewise_random_core <- function(n, rates, cuts) {
  if (n <= 0 || any(is.na(rates))) {
    return(numeric(0))
  }
  u <- stats::runif(n)
  fastml_piecewise_quantile_core(u, rates, cuts)
}

fastml_piecewise_density_fn <- function(cuts, delta_names) {
  function(x, log_rate, ..., log = FALSE) {
    extra <- list(...)
    deltas <- fastml_piecewise_extract_deltas(extra, delta_names)
    rates <- fastml_piecewise_rates(log_rate, deltas)
    dens <- fastml_piecewise_density_core(x, rates, cuts)
    if (log) {
      log(dens)
    } else {
      dens
    }
  }
}

fastml_piecewise_cdf_fn <- function(cuts, delta_names) {
  function(q, log_rate, ..., lower.tail = TRUE, log.p = FALSE) {
    extra <- list(...)
    deltas <- fastml_piecewise_extract_deltas(extra, delta_names)
    rates <- fastml_piecewise_rates(log_rate, deltas)
    cdf <- fastml_piecewise_cdf_core(q, rates, cuts)
    if (!lower.tail) {
      cdf <- 1 - cdf
    }
    cdf <- pmin(pmax(cdf, 0), 1)
    if (log.p) {
      log(cdf)
    } else {
      cdf
    }
  }
}

fastml_piecewise_quantile_fn <- function(cuts, delta_names) {
  function(p, log_rate, ..., lower.tail = TRUE, log.p = FALSE) {
    extra <- list(...)
    deltas <- fastml_piecewise_extract_deltas(extra, delta_names)
    rates <- fastml_piecewise_rates(log_rate, deltas)
    probs <- as.numeric(p)
    if (log.p) {
      probs <- exp(probs)
    }
    if (!lower.tail) {
      probs <- 1 - probs
    }
    fastml_piecewise_quantile_core(probs, rates, cuts)
  }
}

fastml_piecewise_random_fn <- function(cuts, delta_names) {
  function(n, log_rate, ...) {
    extra <- list(...)
    deltas <- fastml_piecewise_extract_deltas(extra, delta_names)
    rates <- fastml_piecewise_rates(log_rate, deltas)
    n_val <- as.integer(n[1])
    if (is.na(n_val) || n_val <= 0) {
      return(numeric(0))
    }
    fastml_piecewise_random_core(n_val, rates, cuts)
  }
}

fastml_piecewise_hazard_fn <- function(cuts, delta_names) {
  function(x, log_rate, ...) {
    extra <- list(...)
    deltas <- fastml_piecewise_extract_deltas(extra, delta_names)
    rates <- fastml_piecewise_rates(log_rate, deltas)
    fastml_piecewise_hazard_core(x, rates, cuts)
  }
}

fastml_piecewise_cumhaz_fn <- function(cuts, delta_names) {
  function(x, log_rate, ...) {
    extra <- list(...)
    deltas <- fastml_piecewise_extract_deltas(extra, delta_names)
    rates <- fastml_piecewise_rates(log_rate, deltas)
    fastml_piecewise_cumhaz_core(x, rates, cuts)
  }
}

library(testthat)

test_that("fastml_piecewise_rates handles scalar inputs", {
  skip_if_not_installed("flexsurv")
  base_log_rate <- log(0.1)
  log_ratios <- log(c(2, 3))
  res <- fastml:::fastml_piecewise_rates(base_log_rate, log_ratios)
  expect_length(res, length(log_ratios) + 1L)
  expect_equal(res, exp(base_log_rate) * c(1, exp(log_ratios)))
})

test_that("fastml_piecewise_rates accepts vectorised log_rate from flexsurv",
          {
            skip_if_not_installed("flexsurv")
            base_log_rate <- log(0.2)
            extra_params <- log(c(0.3, 0.4))
            res <-
              fastml:::fastml_piecewise_rates(c(base_log_rate, extra_params), numeric(0))
            expect_length(res, length(extra_params) + 1L)
            expect_equal(res[1], exp(base_log_rate))
            expect_equal(res[-1], exp(base_log_rate) * exp(extra_params))
          })

#' Fit a surrogate decision tree for a fastml model
#'
#' Builds an interpretable tree approximating the behaviour of the underlying
#' model using the `iml` package.
#'
#' @param object A `fastml` object.
#' @param maxdepth Maximum depth of the surrogate tree. Default 3.
#' @param ... Additional arguments passed to `iml::TreeSurrogate`.
#'
#' @param data Character string specifying which data to use: \code{"train"} (default) or \code{"test"}.
#' @return An `iml::TreeSurrogate` object.
#' @importFrom recipes bake
#' @export
#' @examples
#' \dontrun{
#' data(iris)
#' iris <- iris[iris$Species != "setosa", ]
#' iris$Species <- factor(iris$Species)
#' model <- fastml(data = iris, label = "Species")
#' surrogate_tree(model)
#' }
surrogate_tree <- function(object, maxdepth = 3, data = c("train", "test"), ...) {
  if (!inherits(object, "fastml")) {
    stop("The input must be a 'fastml' object.")
  }
  if (!requireNamespace("iml", quietly = TRUE)) {
    stop("The 'iml' package is required for surrogate models.")
  }

  data <- match.arg(data)
  prep <- fastml_prepare_explainer_inputs(object, data = data)
  raw_data <- prep$train_data
  if (is.null(raw_data) || !(prep$label %in% names(raw_data))) {
    stop("Training data not available for surrogate models.")
  }
  x <- prep$x_raw
  y <- prep$y_raw

  # Handle classification targets for iml
  # Determine if this is multiclass (>2 classes)
  is_multiclass <- FALSE
  positive_class <- NULL
  if (is.factor(y) || is.character(y)) {
    y_factor <- if (is.factor(y)) y else factor(y)
    n_classes <- length(levels(y_factor))
    is_multiclass <- n_classes > 2

    if (is_multiclass) {
      # Multiclass: keep y as factor, iml will handle it properly
      # when predict function returns all class probabilities
      y <- y_factor
    } else {
      # Binary classification: convert to numeric (0/1) for iml
      # Use resolve_positive_class to respect event_class settings from fastml()
      positive_class <- resolve_positive_class(prep, levels(y_factor))
      y <- as.numeric(y_factor == positive_class)
    }
  } else {
    positive_class <- prep$positive_class
  }

  parsnip_fit <- prep$fits[[1]]

  # Custom predict function
  # For multiclass: returns data.frame with all class probabilities
  # For binary: returns numeric probability vector for the positive class
  predict_fun <- function(model, newdata) {
    newdata_processed <- tryCatch(
      {
        baked <- recipes::bake(prep$preprocessor, new_data = newdata)
        if (!is.null(prep$label) && prep$label %in% names(baked)) {
          baked[[prep$label]] <- NULL
        }
        baked
      },
      error = function(e) {
        warning(sprintf(
          paste0(
            "Could not apply the fitted recipe to the explanation data: %s\n",
            "Falling back to the unprocessed data; the explanation below ",
            "describes the model on raw features and may not be meaningful."
          ),
          conditionMessage(e)
        ), call. = FALSE)
        newdata
      }
    )

    prob <- tryCatch(
      predict(model, new_data = newdata_processed, type = "prob"),
      error = function(e) NULL
    )
    if (!is.null(prob)) {
      # Multiclass: return full probability data.frame (iml handles this)
      if (is_multiclass) {
        # Clean column names: remove .pred_ prefix
        colnames(prob) <- sub("^\\.pred_", "", colnames(prob))
        return(as.data.frame(prob))
      }
      # Binary: return probability for positive class
      prob_col <- paste0(".pred_", positive_class)
      if (!is.null(positive_class) && prob_col %in% names(prob)) {
        return(as.numeric(prob[[prob_col]]))
      }
      num_cols <- names(prob)[vapply(prob, is.numeric, logical(1))]
      if (length(num_cols) > 0) {
        return(as.numeric(prob[[num_cols[1]]]))
      }
    }

    preds <- predict(model, new_data = newdata_processed)
    as.numeric(preds[[1]])
  }

  predictor <- iml::Predictor$new(parsnip_fit, data = x, y = y, predict.fun = predict_fun)
  tree <- iml::TreeSurrogate$new(predictor, maxdepth = maxdepth, ...)
  p <- plot(tree)
  print(p)
  invisible(tree)
}

#' Helper functions for flexsurv-based survival models
#'
#' These utilities are internal to fastml and support training-time summaries
#' and prediction helpers for models fitted with flexsurv::flexsurvreg().
#'
#' @noRd
NULL

#' Fit flexsurvreg with error handling
#'
#' A wrapper around \code{flexsurv::flexsurvreg} that merges base and engine
#' arguments and includes a fallback mechanism for singular hessian errors
#' by retrying with \code{hessian = FALSE}.
#'
#' @param base_args A list of base arguments for \code{flexsurvreg}.
#' @param engine_args A list of engine-specific arguments that can override
#'   \code{base_args}.
#'
#' @return A fitted \code{flexsurvreg} object.
#' @keywords internal
#' @noRd

fastml_fit_flexsurvreg <- function(base_args, engine_args) {
  attempt <- tryCatch(
    call_with_engine_params(
      flexsurv::flexsurvreg,
      base_args,
      engine_args
    ),
    error = function(e) e
  )

  if (inherits(attempt, "error")) {
    msg <- conditionMessage(attempt)
    singular <- grepl("system is exactly singular", msg, fixed = TRUE)
    user_overrode_hessian <- !is.null(engine_args) && "hessian" %in% names(engine_args)
    base_has_hessian <- !is.null(base_args$hessian)

    if (singular && !user_overrode_hessian && !base_has_hessian) {
      base_args_fallback <- base_args
      base_args_fallback$hessian <- FALSE

      attempt <- tryCatch(
        call_with_engine_params(
          flexsurv::flexsurvreg,
          base_args_fallback,
          engine_args
        ),
        error = function(e) e
      )
    }
  }

  if (inherits(attempt, "error")) {
    stop(attempt)
  }

  attempt
}

#' Generate Survival Matrix from a flexsurvreg Fit
#'
#' Robustly extracts survival probabilities from a \code{flexsurvreg} object
#' for new data at specified times. It attempts multiple prediction methods
#' (e.g., \code{predict}, \code{summary}) to create a standardized
#' [n_obs, n_times] matrix.
#'
#' @param fit A fitted \code{flexsurvreg} object.
#' @param newdata A data frame of new observations for prediction.
#' @param times A numeric vector of evaluation times.
#'
#' @return A matrix of survival probabilities with rows corresponding to
#'   \code{newdata} and columns to \code{times}. Returns \code{NULL} or a
#'   matrix of \code{NA} on failure.
#' @importFrom stats setNames
#' @importFrom utils head tail
#' @keywords internal
#' @noRd
fastml_flexsurv_survival_matrix <- function(fit, newdata, times) {
  if (!inherits(fit, "flexsurvreg")) {
    return(NULL)
  }
  if (!requireNamespace("flexsurv", quietly = TRUE)) {
    return(NULL)
  }

  times <- as.numeric(times)
  times <- times[is.finite(times) & times >= 0]
  if (length(times) == 0) {
    n_rows <- if (!is.null(newdata)) nrow(newdata) else 0L
    return(matrix(numeric(0), nrow = n_rows, ncol = 0))
  }

  if (is.null(newdata)) {
    return(matrix(numeric(), nrow = 0, ncol = length(times)))
  }

  newdata <- as.data.frame(newdata)
  n_obs <- nrow(newdata)
  if (n_obs == 0) {
    return(matrix(numeric(), nrow = 0, ncol = length(times)))
  }

  align_curve <- function(curve_times, curve_surv, eval_times) {
    if (length(eval_times) == 0) {
      return(numeric(0))
    }
    if (length(curve_times) == 0 || length(curve_surv) == 0) {
      return(rep(NA_real_, length(eval_times)))
    }
    curve_times <- as.numeric(curve_times)
    curve_surv <- as.numeric(curve_surv)
    ord <- order(curve_times)
    curve_times <- curve_times[ord]
    curve_surv <- curve_surv[ord]
    idx <- findInterval(eval_times, curve_times)
    res <- rep(NA_real_, length(eval_times))
    if (any(idx == 0)) {
      res[idx == 0] <- 1
    }
    pos_idx <- which(idx > 0)
    if (length(pos_idx) > 0) {
      mapped <- pmin(idx[pos_idx], length(curve_surv))
      res[pos_idx] <- curve_surv[mapped]
    }
    if (any(idx > length(curve_surv))) {
      last_val <- curve_surv[length(curve_surv)]
      res[idx > length(curve_surv)] <- last_val
    }
    res <- pmin(pmax(res, 0), 1)
    res
  }

  convert_predict_survival <- function(pred_tbl, times, n_obs) {
    if (is.null(pred_tbl)) {
      return(NULL)
    }

    if (inherits(pred_tbl, "tbl_df")) {
      pred_tbl <- as.data.frame(pred_tbl)
    }

    extract_frames <- function(tbl) {
      if (is.null(tbl)) {
        return(list())
      }
      if (".pred" %in% names(tbl)) {
        res <- tbl[[".pred"]]
        if (!is.list(res)) {
          res <- list(res)
        }
        return(res)
      }
      split(tbl, seq_len(nrow(tbl)))
    }

    frames <- extract_frames(pred_tbl)
    if (length(frames) == 0) {
      return(NULL)
    }

    res <- matrix(NA_real_, nrow = length(frames), ncol = length(times))
    for (i in seq_along(frames)) {
      df <- frames[[i]]
      if (is.null(df)) {
        next
      }
      df <- coerce_to_dataframe(df)
      if (is.null(df)) {
        next
      }
      time_col <- intersect(c(".time", "time", "t", ".eval_time"), names(df))
      surv_col <- intersect(c(".pred_survival", ".pred", "survival", "est", ".pred_surv"), names(df))
      if (length(time_col) == 0 || length(surv_col) == 0) {
        next
      }
      res[i, ] <- align_curve(df[[time_col[1]]], df[[surv_col[1]]], times)
    }

    if (nrow(res) == 1 && n_obs > 1) {
      res <- res[rep(1, n_obs), , drop = FALSE]
    }

    if (nrow(res) != n_obs) {
      if (length(frames) >= n_obs) {
        res <- res[seq_len(n_obs), , drop = FALSE]
      } else {
        res <- rbind(res, matrix(NA_real_, nrow = max(0, n_obs - nrow(res)), ncol = length(times)))
      }
    }

    if (nrow(res) != n_obs) {
      res <- res[seq_len(min(nrow(res), n_obs)), , drop = FALSE]
      if (nrow(res) < n_obs) {
        res <- rbind(res, matrix(NA_real_, nrow = n_obs - nrow(res), ncol = length(times)))
      }
    }

    if (nrow(res) == n_obs) {
      res
    } else {
      NULL
    }
  }

  coerce_to_dataframe <- function(x) {
    if (is.null(x)) {
      return(NULL)
    }
    if (is.data.frame(x)) {
      return(x)
    }
    if (inherits(x, "tbl_df")) {
      return(as.data.frame(x))
    }
    if (is.matrix(x)) {
      return(as.data.frame(x))
    }
    if (is.list(x) && !is.null(x$time) && !is.null(x$survival)) {
      df <- data.frame(time = x$time, survival = x$survival)
      other_cols <- setdiff(names(x), c("time", "survival"))
      for (nm in other_cols) {
        if (is.atomic(x[[nm]]) && length(x[[nm]]) == nrow(df)) {
          df[[nm]] <- x[[nm]]
        }
      }
      return(df)
    }
    NULL
  }

  extract_curves <- function(summary_list, times, n_obs) {
    if (inherits(summary_list, "data.frame")) {
      row_cols <- intersect(c(".row", "row", "obs", "case", "id", "ID", "index"),
                            names(summary_list))
      if (length(row_cols) > 0) {
        split_col <- summary_list[[row_cols[1]]]
        if (!is.null(split_col)) {
          split_keys <- unique(split_col)
          split_map <- split(summary_list, split_col)
          summary_list <- lapply(seq_along(split_keys), function(idx) {
            key_chr <- as.character(split_keys[idx])
            if (!is.null(names(split_map)) && key_chr %in% names(split_map)) {
              df <- split_map[[key_chr]]
            } else {
              df <- split_map[[1]]
            }
            drop_cols <- intersect(row_cols, names(df))
            if (length(drop_cols) > 0) {
              df <- df[, setdiff(names(df), drop_cols), drop = FALSE]
            }
            df
          })
          summary_list <- unname(summary_list)
        }
      }
    }

    if (!is.list(summary_list)) {
      summary_list <- list(summary_list)
    }

    summary_list <- lapply(summary_list, function(df) {
      df <- coerce_to_dataframe(df)
      if (!is.null(df) && nrow(df) > 0) {
        drop_cols <- intersect(c(".row", "row", "obs", "case", "id", "ID", "index"), names(df))
        if (length(drop_cols) > 0) {
          df <- df[, setdiff(names(df), drop_cols), drop = FALSE]
        }
      }
      df
    })

    if (length(summary_list) == 1L && n_obs > 1L) {
      summary_list <- rep(summary_list, length.out = n_obs)
    }

    res <- matrix(NA_real_, nrow = n_obs, ncol = length(times))

    max_iter <- min(length(summary_list), n_obs)
    for (i in seq_len(max_iter)) {
      df <- summary_list[[i]]
      if (!is.null(df) && !is.data.frame(df)) {
        df <- coerce_to_dataframe(df)
      }
      if (is.null(df)) {
        next
      }
      time_col <- intersect(c("time", "t", ".eval_time"), names(df))
      surv_col <- intersect(c("est", "survival", "S", ".pred_survival", ".pred"), names(df))
      if (length(time_col) == 0 || length(surv_col) == 0) {
        next
      }
      curve_times <- df[[time_col[1]]]
      curve_surv <- df[[surv_col[1]]]
      res[i, ] <- align_curve(curve_times, curve_surv, times)
    }

    if (n_obs > length(summary_list) && length(summary_list) >= 1) {
      filled <- res[seq_len(max_iter), , drop = FALSE]
      valid_row <- which(rowSums(!is.na(filled)) > 0)
      if (length(valid_row) >= 1) {
        template <- filled[valid_row[1], , drop = TRUE]
        for (i in seq((max_iter + 1), n_obs)) {
          res[i, ] <- template
        }
      }
    }

    res <- pmin(pmax(res, 0), 1)
    res
  }

  compute_rowwise <- function(newdata, times) {
    n_obs <- nrow(newdata)
    n_times <- length(times)
    res <- matrix(NA_real_, nrow = n_obs, ncol = n_times)

    if (n_obs == 0 || n_times == 0) {
      return(res)
    }

    res_list <- tryCatch({
      summary(
        fit,
        type = "survival",
        t = times,
        newdata = newdata,
        ci = FALSE
      )
    }, error = function(e) {
      warning("summary failed during prediction: ", e$message)
      NULL
    })

    if (is.null(res_list)) {
      return(res)
    }

    if (is.data.frame(res_list)) {
      res_list <- list(res_list)
    }

    if (!is.list(res_list)) {
      return(res)
    }

    row_values <- lapply(res_list, function(df) {
      df <- coerce_to_dataframe(df)
      if (is.null(df)) {
        return(rep(NA_real_, n_times))
      }
      time_col <- intersect(c("time", "t", ".eval_time"), names(df))
      surv_col <- intersect(c("est", "survival", "S", ".pred_survival", ".pred"), names(df))
      if (length(time_col) == 0 || length(surv_col) == 0) {
        return(rep(NA_real_, n_times))
      }
      align_curve(df[[time_col[1]]], df[[surv_col[1]]], times)
    })

    if (length(row_values) == 0) {
      return(res)
    }

    row_mat <- do.call(rbind, row_values)
    if (!is.matrix(row_mat)) {
      row_mat <- matrix(as.numeric(row_mat), ncol = n_times, byrow = TRUE)
    }

    if (nrow(row_mat) < n_obs) {
      pad <- matrix(NA_real_, nrow = n_obs - nrow(row_mat), ncol = n_times)
      row_mat <- rbind(row_mat, pad)
    }
    if (nrow(row_mat) > n_obs) {
      row_mat <- row_mat[seq_len(n_obs), , drop = FALSE]
    }

    if (ncol(row_mat) != n_times) {
      row_mat <- matrix(NA_real_, nrow = n_obs, ncol = n_times)
    }

    row_mat <- pmin(pmax(row_mat, 0), 1)
    row_mat
  }

  summary_tidy <- tryCatch(
    summary(
      fit,
      type = "survival",
      t = times,
      newdata = newdata,
      ci = FALSE,
      tidy = TRUE
    ),
    error = function(e) NULL
  )

  convert_tidy_summary <- function(df, times, n_obs) {
    if (!is.data.frame(df) || !(".row" %in% names(df))) {
      return(NULL)
    }
    time_col <- intersect(c("time", "t", ".eval_time"), names(df))
    surv_col <- intersect(c("est", "survival", "S", ".pred_survival", ".pred"), names(df))
    if (length(time_col) == 0 || length(surv_col) == 0) {
      return(NULL)
    }
    df$.row <- as.integer(df$.row)
    df <- df[is.finite(df$.row) & df$.row >= 1, , drop = FALSE]
    if (nrow(df) == 0) {
      return(NULL)
    }
    res <- matrix(NA_real_, nrow = n_obs, ncol = length(times))
    split_map <- split(df, df$.row)
    max_iter <- min(length(split_map), n_obs)
    for (i in seq_len(max_iter)) {
      entry <- split_map[[i]]
      res[i, ] <- align_curve(entry[[time_col[1]]], entry[[surv_col[1]]], times)
    }
    res
  }

  predict_tbl <- NULL
  if (length(times) > 0 && n_obs > 0) {
    predict_tbl <- tryCatch(
      predict(
        fit,
        newdata = newdata,
        type = "survival",
        times = times,
        conf.int = FALSE,
        se.fit = FALSE
      ),
      error = function(e) NULL
    )
  }

  res <- convert_predict_survival(predict_tbl, times, n_obs)

  if (is.null(res)) {
    res <- convert_tidy_summary(summary_tidy, times, n_obs)
  }

  if (is.null(res) || !is.matrix(res)) {
    summary_list <- tryCatch(
      summary(
        fit,
        type = "survival",
        t = times,
        newdata = newdata,
        ci = FALSE
      ),
      error = function(e) NULL
    )

    if (!is.null(summary_list)) {
      res <- extract_curves(summary_list, times, n_obs)
      has_finite <- is.matrix(res) && any(rowSums(is.finite(res)) > 0)
      if (!has_finite) {
        res <- compute_rowwise(newdata, times)
      }
    } else {
      res <- compute_rowwise(newdata, times)
    }
  }

  if (length(times) > 0 && is.matrix(res) &&
      nrow(res) == n_obs && ncol(res) == length(times)) {
    na_rows <- which(rowSums(is.finite(res)) == 0)
    if (length(na_rows) > 0) {
      fallback <- compute_rowwise(newdata, times)
      if (is.matrix(fallback) && nrow(fallback) == n_obs &&
          ncol(fallback) == length(times)) {
        res[na_rows, ] <- fallback[na_rows, , drop = FALSE]
      }
    }
  }

  expected_dim <- c(n_obs, length(times))
  if (!is.matrix(res)) {
    res <- matrix(as.numeric(res), nrow = expected_dim[1], ncol = expected_dim[2])
  }

  if (!all(dim(res) == expected_dim)) {
    res <- compute_rowwise(newdata, times)
    if (!is.matrix(res)) {
      res <- matrix(as.numeric(res), nrow = expected_dim[1], ncol = expected_dim[2])
    }
  }

  if (!is.matrix(res) || !all(dim(res) == expected_dim)) {
    res <- matrix(NA_real_, nrow = expected_dim[1], ncol = expected_dim[2])
  }

  if (length(times) > 0) {
    res[is.nan(res)] <- NA_real_
    res <- pmin(pmax(res, 0), 1)
  }
  res
}

#' Standardized Prediction Function for Parametric Survival Models
#'
#' A wrapper for \code{flexsurvreg} predictions that conforms to the
#' \code{fastml} internal prediction API. It returns both a survival
#' probability matrix and a single risk score vector.
#'
#' @param fit A fitted \code{flexsurvreg} object.
#' @param newdata A data frame of new observations.
#' @param eval_times A numeric vector of times to evaluate survival.
#' @param risk_time An optional specific time point used to calculate the
#'   risk score. If \code{NULL}, defaults to the median or max of
#'   \code{eval_times}.
#'
#' @return A list with two elements:
#'   \describe{
#'     \item{surv}{A matrix of survival probabilities (rows=subjects,
#'       cols=eval_times) with an \code{eval_times} attribute.}
#'     \item{risk}{A numeric vector of risk scores, typically \eqn{-log(S(t))}
#'       at the \code{risk_time}.}
#'   }
#'
#' @importFrom stats median
#' @keywords internal
#' @noRd
fastml_parametric_surv_predict <- function(fit,
                                           newdata,
                                           eval_times,
                                           risk_time = NULL) {
  n_obs <- if (!is.null(newdata)) nrow(newdata) else 0L

  if (!inherits(fit, "flexsurvreg") ||
      !requireNamespace("flexsurv", quietly = TRUE)) {
    empty_surv <- if (length(eval_times) > 0) {
      matrix(NA_real_, nrow = n_obs, ncol = length(eval_times))
    } else {
      matrix(numeric(), nrow = n_obs, ncol = 0)
    }
    return(list(surv = empty_surv, risk = rep(NA_real_, n_obs)))
  }

  eval_times <- as.numeric(eval_times)
  eval_times <- sort(unique(eval_times[is.finite(eval_times) & eval_times >= 0]))

  if (is.null(newdata)) {
    newdata <- data.frame(matrix(nrow = 0, ncol = 0))
  }

  newdata <- as.data.frame(newdata)
  n_obs <- nrow(newdata)

  if (n_obs == 0) {
    surv_mat <- if (length(eval_times) > 0) {
      matrix(numeric(), nrow = 0, ncol = length(eval_times))
    } else {
      matrix(numeric(), nrow = 0, ncol = 0)
    }
    attr(surv_mat, "eval_times") <- eval_times
    return(list(surv = surv_mat, risk = numeric(0)))
  }

  if (length(eval_times) == 0) {
    surv_mat <- matrix(NA_real_, nrow = n_obs, ncol = 0)
    attr(surv_mat, "eval_times") <- eval_times
    return(list(surv = surv_mat, risk = rep(NA_real_, n_obs)))
  }

  surv_mat <- fastml_flexsurv_survival_matrix(fit, newdata, times = eval_times)
  if (is.null(surv_mat) || !is.matrix(surv_mat) ||
      nrow(surv_mat) != n_obs || ncol(surv_mat) != length(eval_times)) {
    surv_mat <- matrix(NA_real_, nrow = n_obs, ncol = length(eval_times))
  }

  if (ncol(surv_mat) == length(eval_times)) {
    colnames(surv_mat) <- format(eval_times, trim = TRUE, scientific = FALSE)
  }
  attr(surv_mat, "eval_times") <- eval_times

  risk_vec <- rep(NA_real_, n_obs)
  valid_cols <- which(colSums(is.finite(surv_mat)) > 0)
  if (length(valid_cols) > 0) {
    if (is.null(risk_time) || !is.finite(risk_time) || risk_time <= 0) {
      med_time <- stats::median(eval_times)
      risk_time <- if (is.finite(med_time) && med_time > 0) med_time else max(eval_times)
    }
    if (!is.finite(risk_time) || risk_time <= 0) {
      risk_time <- eval_times[valid_cols[length(valid_cols)]]
    }
    idx <- which.min(abs(eval_times - risk_time))
    idx <- max(1L, min(idx, ncol(surv_mat)))
    surv_vals <- as.numeric(surv_mat[, idx])
    surv_vals[!is.finite(surv_vals)] <- NA_real_
    risk_vec <- -log(pmax(surv_vals, .Machine$double.eps))
    risk_vec[!is.finite(risk_vec)] <- NA_real_
  }

  list(surv = surv_mat, risk = risk_vec)
}

library(testthat)


test_that("fastml_flexsurv_survival_matrix preserves factor structure", {
  skip_if_not_installed("flexsurv")
  skip_if_not_installed("survival")

  lung <- survival::lung
  lung <- stats::na.omit(lung[, c("time", "status", "age", "sex", "ph.ecog")])
  lung$sex <- factor(lung$sex, levels = c(1, 2), labels = c("male", "female"))
  lung$ph.ecog <- factor(lung$ph.ecog)
  set.seed(123)
  lung$group <- factor(sample(c("A", "B", "C"), nrow(lung), replace = TRUE),
                       levels = c("A", "B", "C", "D"))

  suppressWarnings({
    fit <- flexsurv::flexsurvreg(
      survival::Surv(time, status) ~ age + sex + ph.ecog + group,
      data = lung,
      dist = "weibull"
    )
  })

  newdata <- lung[seq_len(min(6L, nrow(lung))), c("age", "sex", "ph.ecog", "group")]
  levels(newdata$group) <- levels(lung$group)
  times <- c(30, 90, 180, 365)

  surv_mat <- fastml:::fastml_flexsurv_survival_matrix(fit, newdata, times)

  expect_true(is.matrix(surv_mat))
  expect_identical(dim(surv_mat), c(nrow(newdata), length(times)))
  expect_true(all(rowSums(is.finite(surv_mat)) > 0))
  expect_true(all(surv_mat[is.finite(surv_mat)] >= 0))
  expect_true(all(surv_mat[is.finite(surv_mat)] <= 1))
})



test_that("fastml_flexsurv_survival_matrix handles empty newdata", {
  skip_if_not_installed("flexsurv")
  skip_if_not_installed("survival")

  lung <- survival::lung
  lung <- stats::na.omit(lung[, c("time", "status", "age", "sex")])
  lung$sex <- factor(lung$sex, levels = c(1, 2), labels = c("male", "female"))

  fit <- flexsurv::flexsurvreg(
    survival::Surv(time, status) ~ age + sex,
    data = lung,
    dist = "weibull"
  )

  times <- c(30, 90, 180)
  empty_newdata <- lung[0, c("age", "sex")]

  surv_mat <- fastml:::fastml_flexsurv_survival_matrix(fit, empty_newdata, times)

  expect_true(is.matrix(surv_mat))
  expect_identical(dim(surv_mat), c(0L, length(times)))
})

#' Determine rounding digits for time horizons
#'
#' Computes a sensible number of decimal digits to round time horizons based on the minimal
#' positive separation between unique finite times.
#'
#' @param times Numeric vector of times.
#'
#' @return Integer number of digits between 0 and 6.
#'
#' @details Uses the smallest strictly positive difference among sorted unique finite times,
#' then returns \code{ceiling(-log10(min_diff))} truncated to \eqn{[0, 6]}.
#'
#' @keywords internal
#' @examples
#' # Not run: determine_round_digits(c(0.1, 0.12, 0.125))
#' NULL
#'
#' @name determine_round_digits
#' @rdname determine_round_digits
#' @noMd
#'
determine_round_digits <- function(times) {
  times <- times[is.finite(times) & times > 0]
  if (length(times) <= 1) {
    return(0)
  }
  diffs <- diff(sort(unique(times)))
  diffs <- diffs[diffs > 0]
  if (length(diffs) == 0) {
    return(0)
  }
  min_diff <- min(diffs)
  if (!is.finite(min_diff) || min_diff <= 0) {
    return(0)
  }
  digits <- ceiling(-log10(min_diff))
  digits <- max(0, digits)
  digits <- min(digits, 6)
  digits
}

#' Compute Tau Limit (t_max)
#'
#' Finds the latest time point \eqn{t_{max}} such that at least a certain proportion
#' of subjects remain at risk.
#'
#' @param times Numeric vector of survival times.
#' @param threshold Minimum proportion of subjects that must remain at risk.
#'
#' @return The computed \eqn{t_{max}} value, or \code{NA_real_} if no valid
#'   times are provided.
#' @keywords internal
#' @noMd
compute_tau_limit <- function(times, threshold) {
  times_valid <- sort(unique(times[is.finite(times) & times > 0]))
  if (length(times_valid) == 0) {
    return(NA_real_)
  }
  n <- length(times)
  props <- vapply(times_valid, function(t) {
    mean(times >= t, na.rm = TRUE)
  }, numeric(1))
  idx <- which(props >= threshold)
  if (length(idx) == 0) {
    max(times_valid)
  } else {
    max(times_valid[idx])
  }
}

#' Create Censoring Distribution Evaluator
#'
#' Creates a function to evaluate the survival function of the censoring
#' distribution, \eqn{G(t) = P(C > t)}, using a Kaplan-Meier estimator.
#'
#' @param time_vec Numeric vector of survival/censoring times.
#' @param status_vec Numeric vector of event statuses (1=event, 0=censored).
#'
#' @return A function that takes a numeric vector of times and returns the
#'   estimated censoring survival probabilities \eqn{G(t)} at those times.
#'
#' @importFrom survival survfit Surv
#' @importFrom utils head
#' @keywords internal
#' @noMd
create_censor_eval <- function(time_vec, status_vec) {
  time_vec <- as.numeric(time_vec)
  status_vec <- ifelse(is.na(status_vec), 0, ifelse(status_vec > 0, 1, 0))
  valid_idx <- which(is.finite(time_vec) & time_vec >= 0)
  if (length(valid_idx) == 0) {
    return(function(times)
      rep(NA_real_, length(times)))
  }
  censor_indicator <- 1 - status_vec[valid_idx]
  fit <- tryCatch({
    survival::survfit(survival::Surv(time_vec[valid_idx], censor_indicator) ~ 1)
  }, error = function(e)
    NULL)
  if (is.null(fit)) {
    return(function(times)
      rep(NA_real_, length(times)))
  }
  function(times) {
    if (length(times) == 0)
      return(numeric(0))
    if (length(fit$time) == 0 || length(fit$surv) == 0) {
      return(rep(1, length(times)))
    }
    align_survival_curve(fit$time, fit$surv, times)
  }
}

#' Assign Risk Groups
#'
#' Dichotomizes a continuous risk vector into "low" and "high" risk groups
#' based on the median.
#'
#' @param risk_vec Numeric vector of predicted risk scores.
#'
#' @return A character vector of "low", "high", or \code{NA}.
#'
#' @importFrom stats median
#' @keywords internal
assign_risk_group <- function(risk_vec) {
  group <- rep(NA_character_, length(risk_vec))
  valid <- is.finite(risk_vec)
  if (sum(valid) == 0) {
    return(group)
  }
  risk_vals <- risk_vec[valid]
  if (length(unique(risk_vals)) > 1) {
    threshold <- stats::median(risk_vals)
    group[valid] <- ifelse(risk_vals > threshold, "high", "low")
  } else {
    group[valid] <- "low"
  }
  group
}

#' Compute Uno's C-index (Time-Dependent AUC)
#'
#' Calculates Uno's C-index (a time-dependent AUC measure) for survival data,
#' weighted by the inverse probability of censoring (IPCW).
#'
#' @param train_time Numeric vector of training times (used for censor model).
#' @param train_status Numeric vector of training statuses (used for censor model).
#' @param test_time Numeric vector of test times.
#' @param test_status Numeric vector of test statuses.
#' @param risk_vec Numeric vector of predicted risk scores for test data.
#' @param tau The time horizon \eqn{\tau} for evaluation. If \code{NA} or
#'   \code{<= 0}, the maximum finite test time is used.
#' @param censor_eval_fn A function (from \code{create_censor_eval}) that
#'   evaluates the censoring survival function \eqn{G(t)}.
#'
#' @return The computed Uno's C-index, or \code{NA_real_} on failure.
#' @keywords internal
#' @noMd
compute_uno_c_index <- function(train_time,
                                train_status,
                                test_time,
                                test_status,
                                risk_vec,
                                tau,
                                censor_eval_fn) {
  if (!is.function(censor_eval_fn)) {
    return(NA_real_)
  }
  if (!is.finite(tau) || tau <= 0) {
    tau <- suppressWarnings(max(test_time[is.finite(test_time)], na.rm = TRUE))
  }
  valid <- which(is.finite(test_time) & is.finite(risk_vec))
  if (length(valid) <= 1) {
    return(NA_real_)
  }
  time_val <- test_time[valid]
  status_val <- ifelse(is.na(test_status[valid]), 0, ifelse(test_status[valid] > 0, 1, 0))
  risk_val <- risk_vec[valid]
  G_vec <- censor_eval_fn(time_val)
  if (length(G_vec) == 0) {
    return(NA_real_)
  }
  G_vec[!is.finite(G_vec) | G_vec <= 0] <- NA_real_
  event_idx <- which(status_val == 1 &
                       is.finite(G_vec) & time_val <= tau)
  if (length(event_idx) == 0) {
    return(NA_real_)
  }
  numerator <- 0
  denominator <- 0
  for (ii in event_idx) {
    w_i <- 1 / (G_vec[ii]^2)
    if (!is.finite(w_i) || w_i <= 0)
      next
    later <- which(time_val > time_val[ii] &
                     time_val <= tau & is.finite(G_vec))
    later <- later[G_vec[later] > 0]
    if (length(later) == 0)
      next
    denominator <- denominator + w_i * length(later)
    diff_scores <- risk_val[ii] - risk_val[later]
    concordant <- sum(diff_scores > 0) + 0.5 * sum(diff_scores == 0)
    numerator <- numerator + w_i * concordant
  }
  if (denominator <= 0) {
    return(NA_real_)
  }
  val <- numerator / denominator
  val <- min(max(val, 0), 1)
  val
}

#' Compute Difference in Restricted Mean Survival Time (RMST)
#'
#' Calculates the difference in RMST between "low" and "high" risk groups up to
#' a time horizon \eqn{\tau}. Groups are defined by median-splitting the
#' \code{risk_vec}.
#'
#' @param time_vec Numeric vector of test times.
#' @param status_vec Numeric vector of test statuses.
#' @param risk_vec Numeric vector of predicted risk scores for test data.
#' @param tau The time horizon \eqn{\tau} for integration.
#' @param surv_mat Optional. A matrix of individual survival predictions
#'   (rows=subjects, cols=times) used for model-based RMST calculation.
#' @param eval_times_full Optional. A numeric vector of time points
#'   corresponding to the columns of \code{surv_mat}.
#' @param model_type Optional string (e.log., "rstpm2", "flexsurv") indicating
#'   if a model-based RMST calculation should be attempted.
#'
#' @return The RMST difference (RMST_low - RMST_high), or \code{NA_real_}.
#'
#' @importFrom survival survfit Surv
#' @importFrom stats approxfun integrate median
#' @keywords internal
#' @noMd
#' @importFrom utils head
#' @importFrom utils tail
compute_rmst_difference <- function(time_vec,
                                    status_vec,
                                    risk_vec,
                                    tau,
                                    surv_mat = NULL,
                                    eval_times_full = NULL,
                                    model_type = "other") {
  if (!is.finite(tau) || tau <= 0) {
    return(NA_real_)
  }

  risk_group <- assign_risk_group(risk_vec)
  valid <- which(is.finite(time_vec) & !is.na(risk_group))
  if (length(valid) < 2) {
    return(NA_real_)
  }

  time_val <- time_vec[valid]
  status_val <- ifelse(is.na(status_vec[valid]), 0, ifelse(status_vec[valid] > 0, 1, 0))
  group_val <- risk_group[valid]
  if (length(unique(group_val)) < 2) {
    return(NA_real_)
  }

  compute_rmst_from_curve <- function(curve_times, curve_surv) {
    if (length(curve_times) == 0 || length(curve_surv) == 0) {
      return(NA_real_)
    }
    ord <- order(curve_times)
    curve_times <- curve_times[ord]
    curve_surv <- curve_surv[ord]
    keep <- is.finite(curve_times) &
      curve_times >= 0 & is.finite(curve_surv)
    curve_times <- curve_times[keep]
    curve_surv <- curve_surv[keep]
    if (length(curve_times) == 0) {
      return(NA_real_)
    }
    if (curve_times[1] > 0) {
      curve_times <- c(0, curve_times)
      curve_surv <- c(1, curve_surv)
    } else {
      curve_surv[1] <- 1
    }
    if (curve_times[length(curve_times)] < tau) {
      tail_val <- tail(curve_surv, 1)
      curve_times <- c(curve_times, tau)
      curve_surv <- c(curve_surv, tail_val)
    }
    curve_surv <- pmin(pmax(curve_surv, 0), 1)
    curve_surv <- cummin(curve_surv)
    fn <- stats::approxfun(
      curve_times,
      curve_surv,
      method = "linear",
      yleft = 1,
      yright = tail(curve_surv, 1)
    )
    rmst_val <- tryCatch(
      stats::integrate(fn, lower = 0, upper = tau)$value,
      error = function(e)
        NA_real_
    )
    rmst_val
  }

  if ((identical(model_type, "rstpm2") ||
       identical(model_type, "flexsurv")) && !is.null(surv_mat) &&
      !is.null(eval_times_full) &&
      length(eval_times_full) > 0 &&
      is.matrix(surv_mat)) {
    surv_mat_use <- surv_mat
    if (nrow(surv_mat_use) == length(time_vec)) {
      surv_mat_use <- surv_mat_use[valid, , drop = FALSE]
    } else if (nrow(surv_mat_use) != length(valid)) {
      surv_mat_use <- tryCatch(
        surv_mat_use[valid, , drop = FALSE],
        error = function(e)
          NULL
      )
    }

    if (!is.null(surv_mat_use) &&
        nrow(surv_mat_use) == length(valid)) {
      eval_times_full <- as.numeric(eval_times_full)
      prepare_group_curve <- function(group_label) {
        idx <- which(group_val == group_label)
        if (length(idx) == 0) {
          return(NULL)
        }
        curves <- surv_mat_use[idx, , drop = FALSE]
        curve_mean <- if (nrow(curves) == 1) {
          as.numeric(curves[1, ])
        } else {
          colMeans(curves, na.rm = TRUE)
        }
        if (all(is.na(curve_mean))) {
          return(NULL)
        }
        times_full <- eval_times_full
        keep_full <- is.finite(times_full)
        times_full <- times_full[keep_full]
        curve_full <- curve_mean[keep_full]
        keep_vals <- is.finite(curve_full)
        times_full <- times_full[keep_vals]
        curve_full <- curve_full[keep_vals]
        if (length(times_full) == 0) {
          return(NULL)
        }
        sel <- times_full <= tau + 1e-08
        if (!any(sel)) {
          sel <- seq_along(times_full)
        }
        times_sel <- times_full[sel]
        surv_sel <- curve_full[sel]
        if (length(times_sel) == 0) {
          return(NULL)
        }
        if (max(times_sel) < tau) {
          surv_tau <- align_survival_curve(times_full, curve_full, tau)
          if (is.finite(surv_tau)) {
            times_sel <- c(times_sel, tau)
            surv_sel <- c(surv_sel, surv_tau)
          }
        }
        if (length(times_sel) == 0) {
          return(NULL)
        }
        unique_times <- sort(unique(times_sel))
        surv_unique <- vapply(unique_times, function(tt) {
          vals <- surv_sel[abs(times_sel - tt) < 1e-12]
          vals[length(vals)]
        }, numeric(1))
        surv_unique <- pmin(pmax(surv_unique, 0), 1)
        surv_unique <- cummin(surv_unique)
        if (unique_times[1] > 0) {
          unique_times <- c(0, unique_times)
          surv_unique <- c(1, surv_unique)
        } else {
          surv_unique[1] <- 1
        }
        list(times = unique_times,
             surv = surv_unique,
             size = length(idx))
      }

      build_pseudo_data <- function(curve_info, arm_label) {
        if (is.null(curve_info) || length(curve_info$times) < 1) {
          return(NULL)
        }
        times_vec <- curve_info$times
        surv_vec <- curve_info$surv
        if (length(times_vec) < 2) {
          total_n <- max(curve_info$size, 10L)
          return(data.frame(
            time = rep(tau, total_n),
            status = rep(0L, total_n),
            arm = rep.int(as.integer(arm_label), total_n)
          ))
        }
        total_n <- max(curve_info$size, 10L)
        surv_vec <- pmin(pmax(surv_vec, 0), 1)
        surv_vec <- cummin(surv_vec)
        interval_surv_prev <- head(surv_vec, -1)
        interval_surv_next <- tail(surv_vec, -1)
        interval_times <- tail(times_vec, -1)
        expected_counts <- pmax(0,
                                (interval_surv_prev - interval_surv_next) * total_n)
        base_counts <- floor(expected_counts + 1e-08)
        base_counts <- pmin(base_counts, total_n)
        remainder <- expected_counts - base_counts
        leftover <- total_n - sum(base_counts)
        if (leftover > 0 && any(remainder > 0)) {
          order_idx <- order(remainder, decreasing = TRUE)
          add_idx <- order_idx[seq_len(min(leftover, length(order_idx)))]
          base_counts[add_idx] <- base_counts[add_idx] + 1L
          leftover <- total_n - sum(base_counts)
        }
        event_times <- rep(interval_times, base_counts)
        status_vec <- rep(1L, length(event_times))
        if (leftover > 0) {
          event_times <- c(event_times, rep(tau, leftover))
          status_vec <- c(status_vec, rep(0L, leftover))
        }
        data.frame(
          time = as.numeric(event_times),
          status = as.integer(status_vec),
          arm = rep.int(as.integer(arm_label), length(event_times))
        )
      }

      low_curve <- prepare_group_curve("low")
      high_curve <- prepare_group_curve("high")

      if (!is.null(low_curve) && !is.null(high_curve)) {
        if (requireNamespace("survRM2", quietly = TRUE)) {
          pseudo_low <- build_pseudo_data(low_curve, arm_label = 1L)
          pseudo_high <- build_pseudo_data(high_curve, arm_label = 0L)
          if (!is.null(pseudo_low) && !is.null(pseudo_high) &&
              nrow(pseudo_low) > 0 && nrow(pseudo_high) > 0) {
            pseudo_df <- rbind(pseudo_low, pseudo_high)
            rmst_obj <- tryCatch(
              survRM2::rmst2(
                time = pseudo_df$time,
                status = pseudo_df$status,
                arm = pseudo_df$arm,
                tau = tau
              ),
              error = function(e)
                NULL
            )
            if (!is.null(rmst_obj) &&
                !is.null(rmst_obj$unadjusted.result)) {
              diff_row <- rmst_obj$unadjusted.result
              if (is.matrix(diff_row)) {
                rmst_row <- diff_row[grepl("RMST (arm=1)-(arm=0)",
                                           rownames(diff_row),
                                           fixed = TRUE), , drop = FALSE]
                if (nrow(rmst_row) >= 1 &&
                    "Est." %in% colnames(rmst_row)) {
                  diff_val <- rmst_row[1, "Est."]
                  if (is.finite(diff_val)) {
                    return(as.numeric(diff_val))
                  }
                }
              }
            }
          }
        }

        rmst_low <- compute_rmst_from_curve(low_curve$times, low_curve$surv)
        rmst_high <- compute_rmst_from_curve(high_curve$times, high_curve$surv)
        if (is.finite(rmst_low) && is.finite(rmst_high)) {
          return(rmst_low - rmst_high)
        }
      }
    }
  }

  rmst_group <- function(times, status) {
    fit <- tryCatch({
      survival::survfit(survival::Surv(times, status) ~ 1)
    }, error = function(e)
      NULL)
    if (is.null(fit)) {
      return(NA_real_)
    }
    if (length(fit$time) == 0 || length(fit$surv) == 0) {
      return(tau)
    }
    idx <- fit$time < tau
    surv_vals <- fit$surv[idx]
    time_vals <- fit$time[idx]
    surv_at_tau <- align_survival_curve(fit$time, fit$surv, tau)
    times_step <- c(0, time_vals, tau)
    surv_step <- c(1, surv_vals, surv_at_tau)
    sum(surv_step[-length(surv_step)] * diff(times_step))
  }

  low_idx <- group_val == "low"
  high_idx <- group_val == "high"
  if (!any(low_idx) || !any(high_idx)) {
    return(NA_real_)
  }

  rmst_low <- rmst_group(time_val[low_idx], status_val[low_idx])
  rmst_high <- rmst_group(time_val[high_idx], status_val[high_idx])
  if (!is.finite(rmst_low) || !is.finite(rmst_high)) {
    return(NA_real_)
  }
  rmst_low - rmst_high
}

#' Extract Time and Status from Survival Matrix
#'
#' Helper function to extract "time" and "status" columns from a matrix
#' (like one returned by \code{survival::Surv()}), falling back to defaults.
#'
#' @param surv_matrix_vals A matrix, typically from \code{Surv(time, status)}.
#' @param default_time Default time vector if not found.
#' @param default_status Default status vector if not found.
#'
#' @return A list with elements \code{time} and \code{status}.
#' @keywords internal
get_surv_info <- function(surv_matrix_vals,
                          default_time,
                          default_status) {
  time_out <- default_time
  status_out <- default_status
  if (!is.null(surv_matrix_vals)) {
    if ("time" %in% colnames(surv_matrix_vals)) {
      time_out <- as.numeric(surv_matrix_vals[, "time"])
    } else if ("time2" %in% colnames(surv_matrix_vals)) {
      time_out <- as.numeric(surv_matrix_vals[, "time2"])
    }
    if ("status" %in% colnames(surv_matrix_vals)) {
      status_out <- as.numeric(surv_matrix_vals[, "status"])
    }
  }
  list(time = time_out, status = status_out)
}

#' Map Brier Curve Values to Specific Horizons
#'
#' Extracts Brier score values from a pre-computed curve at specific
#' time horizons by finding the closest matching evaluation time.
#'
#' @param curve Numeric vector of Brier scores from \code{compute_ibrier}.
#' @param eval_times Numeric vector of times corresponding to \code{curve}.
#' @param horizons Numeric vector of target time horizons to extract.
#'
#' @return A numeric vector of Brier scores corresponding to \code{horizons}.
#' @keywords internal
map_brier_values <- function(curve, eval_times, horizons) {
  if (length(horizons) == 0) {
    return(numeric(0))
  }
  if (length(curve) != length(eval_times)) {
    return(rep(NA_real_, length(horizons)))
  }
  sapply(horizons, function(h) {
    if (!is.finite(h))
      return(NA_real_)
    idx <- which.min(abs(eval_times - h))
    if (length(idx) == 0)
      return(NA_real_)
    if (abs(eval_times[idx] - h) > max(1e-08, 1e-06 * max(1, abs(h)))) {
      return(NA_real_)
    }
    curve[idx]
  })
}

#' Clamp Values to [0, 1]
#'
#' Truncates a numeric vector so all values lie within the [0, 1] interval.
#'
#' @param x A numeric vector.
#'
#' @return The clamped numeric vector.
#' @keywords internal
clamp01 <- function(x) {
  x <- pmax(pmin(x, 1), 0)
  x
}

#' Align Survival Curve to Evaluation Times
#'
#' Aligns a survival curve (defined by time points and survival probabilities)
#' to a new set of evaluation times using constant interpolation (last value
#' carried forward). Ensures \eqn{S(0) = 1} and monotonicity.
#'
#' @param curve_times Numeric vector of time points from the survival curve.
#' @param curve_surv Numeric vector of survival probabilities corresponding to
#'   \code{curve_times}.
#' @param eval_times Numeric vector of new time points to evaluate at.
#'
#' @return A numeric vector of survival probabilities at \code{eval_times}.
#'
#' @importFrom stats approx
#' @keywords internal
#' @noMd
align_survival_curve <- function(curve_times, curve_surv, eval_times) {
  if (length(eval_times) == 0) {
    return(numeric(0))
  }
  if (length(curve_times) == 0 || length(curve_surv) == 0) {
    return(rep(NA_real_, length(eval_times)))
  }
  curve_times <- as.numeric(curve_times)
  curve_surv <- as.numeric(curve_surv)
  ord <- order(curve_times)
  curve_times <- curve_times[ord]
  curve_surv <- curve_surv[ord]
  keep <- is.finite(curve_times) & is.finite(curve_surv)
  curve_times <- curve_times[keep]
  curve_surv <- curve_surv[keep]
  if (length(curve_times) == 0) {
    return(rep(NA_real_, length(eval_times)))
  }
  if (curve_times[1] > 0) {
    curve_times <- c(0, curve_times)
    curve_surv <- c(1, curve_surv)
  } else {
    curve_times[1] <- 0
    curve_surv[1] <- 1
  }
  if (length(curve_times) > 1) {
    curve_surv <- cummin(pmin(pmax(curve_surv, 0), 1))
  } else {
    curve_surv <- pmin(pmax(curve_surv, 0), 1)
  }
  approx_res <- stats::approx(
    x = curve_times,
    y = curve_surv,
    xout = eval_times,
    method = "constant",
    f = 0,
    rule = 2,
    ties = "ordered"
  )
  res <- approx_res$y
  res <- pmin(pmax(res, 0), 1)
  res
}

#' Build Survival Matrix from survfit Object
#'
#' Extracts survival probabilities from a \code{survfit} object and aligns
#' them to a common set of evaluation times, creating a matrix.
#'
#' @param fit_obj A \code{survfit} object.
#' @param eval_times Numeric vector of evaluation times.
#' @param n_obs Expected number of observations (rows).
#'
#' @return A matrix (rows=subjects, cols=eval_times) of survival
#'   probabilities, or \code{NULL} on failure.
#' @keywords internal
build_survfit_matrix <- function(fit_obj, eval_times, n_obs) {
  if (is.null(fit_obj) || length(eval_times) == 0 || n_obs == 0) {
    return(NULL)
  }
  surv_times <- fit_obj$time
  surv_vals <- fit_obj$surv
  if (is.null(surv_times) || is.null(surv_vals)) {
    return(NULL)
  }
  if (is.matrix(surv_vals)) {
    n_curves <- ncol(surv_vals)
    res <- matrix(NA_real_,
                  nrow = n_curves,
                  ncol = length(eval_times))
    for (j in seq_len(n_curves)) {
      res[j, ] <- align_survival_curve(surv_times, surv_vals[, j], eval_times)
    }
    if (n_curves != n_obs) {
      if (n_curves == 1 && n_obs > 1) {
        res <- matrix(
          res[1, ],
          nrow = n_obs,
          ncol = length(eval_times),
          byrow = TRUE
        )
      } else {
        res <- res[seq_len(min(n_curves, n_obs)), , drop = FALSE]
        if (n_curves < n_obs) {
          res <- rbind(res,
                       matrix(
                         NA_real_,
                         nrow = n_obs - n_curves,
                         ncol = length(eval_times)
                       ))
        }
      }
    }
    return(res)
  }
  if (is.numeric(surv_vals)) {
    curve <- align_survival_curve(surv_times, surv_vals, eval_times)
    return(matrix(
      rep(curve, each = n_obs),
      nrow = n_obs,
      ncol = length(eval_times)
    ))
  }
  NULL
}

#' Extract survreg Linear Predictor and Scale
#'
#' Computes the linear predictor (lp) and scale parameter(s) for new data
#' from a fitted \code{survreg} model.
#'
#' @param fit_obj A fitted \code{survreg} object.
#' @param new_data A data frame with predictor variables.
#'
#' @return A list with elements \code{lp} (numeric vector) and \code{scale}
#'   (numeric vector), or \code{NULL} on failure.
#'
#' @importFrom stats delete.response model.frame model.offset model.matrix na.pass
#' @importFrom survival untangle.specials strata
#' @keywords internal
extract_survreg_components <- function(fit_obj, new_data) {
  if (is.null(fit_obj) || is.null(new_data)) {
    return(NULL)
  }

  Terms <- fit_obj$terms
  if (is.null(Terms) || !inherits(Terms, "terms")) {
    return(NULL)
  }

  Terms_noy <- stats::delete.response(Terms)

  model_frame <- tryCatch({
    stats::model.frame(
      Terms_noy,
      data = new_data,
      na.action = stats::na.pass,
      xlev = fit_obj$xlevels
    )
  }, error = function(e)
    NULL)

  if (is.null(model_frame) || nrow(model_frame) == 0) {
    return(NULL)
  }

  offset_vals <- tryCatch(
    stats::model.offset(model_frame),
    error = function(e)
      NULL
  )
  if (length(offset_vals) == 0 || all(is.na(offset_vals))) {
    offset_vals <- rep(0, nrow(model_frame))
  } else {
    offset_vals <- as.numeric(offset_vals)
    offset_vals[is.na(offset_vals)] <- 0
  }

  model_matrix <- tryCatch({
    stats::model.matrix(fit_obj, model_frame)
  }, error = function(e)
    NULL)

  if (is.null(model_matrix)) {
    return(NULL)
  }

  coefs <- fit_obj$coefficients
  mm_cols <- colnames(model_matrix)
  if (!is.null(mm_cols) && !is.null(names(coefs))) {
    missing_cols <- setdiff(mm_cols, names(coefs))
    if (length(missing_cols) > 0) {
      complete_coefs <- numeric(length(mm_cols))
      names(complete_coefs) <- mm_cols
      overlap <- intersect(mm_cols, names(coefs))
      if (length(overlap) > 0) {
        complete_coefs[overlap] <- coefs[overlap]
      }
      coefs <- complete_coefs
    } else {
      coefs <- coefs[mm_cols]
    }
  } else if (length(coefs) != ncol(model_matrix)) {
    return(NULL)
  }

  coefs[!is.finite(coefs)] <- 0
  lp <- as.numeric(model_matrix %*% coefs) + offset_vals

  strata_special <- attr(Terms, "specials")$strata
  if (!is.null(strata_special) && length(strata_special) > 0) {
    temp <- survival::untangle.specials(Terms, "strata", 1)
    if (length(temp$vars) == 1) {
      strata_vals <- model_frame[[temp$vars]]
    } else {
      strata_vals <- survival::strata(model_frame[, temp$vars], shortlabel = TRUE)
    }
    scale_lookup <- fit_obj$scale
    if (is.null(scale_lookup)) {
      scale_vec <- rep(1, length(lp))
    } else {
      scale_names <- names(scale_lookup)
      if (!is.null(scale_names)) {
        strata_index <- match(as.character(strata_vals), scale_names)
        strata_index[is.na(strata_index)] <- 1L
      } else {
        strata_index <- as.integer(factor(strata_vals))
        strata_index[!is.finite(strata_index)] <- 1L
        if (length(scale_lookup) < max(strata_index, na.rm = TRUE)) {
          scale_lookup <- rep(scale_lookup,
                              length.out = max(strata_index, na.rm = TRUE))
        }
      }
      scale_vec <- scale_lookup[strata_index]
    }
  } else {
    scale_lookup <- fit_obj$scale
    if (length(scale_lookup) <= 1) {
      scale_vec <- rep(scale_lookup, length(lp))
    } else {
      scale_vec <- scale_lookup[rep(1L, length(lp))]
    }
  }

  scale_vec <- as.numeric(scale_vec)
  scale_vec[!is.finite(scale_vec) | scale_vec <= 0] <- NA_real_

  list(lp = lp, scale = scale_vec)
}

#' Compute Survival Matrix from survreg Model
#'
#' Generates a matrix of survival probabilities (rows=subjects, cols=times)
#' from a fitted \code{survreg} model for new data.
#'
#' @param fit_obj A fitted \code{survreg} object.
#' @param new_data A data frame with predictor variables.
#' @param eval_times Numeric vector of evaluation times.
#'
#' @return A matrix of survival probabilities, or \code{NULL} on failure.
#'
#' @importFrom survival psurvreg
#' @importFrom stats median
#' @keywords internal
compute_survreg_matrix <- function(fit_obj, new_data, eval_times) {
  if (!inherits(fit_obj, "survreg") || length(eval_times) == 0) {
    return(NULL)
  }

  components <- extract_survreg_components(fit_obj, new_data)
  if (is.null(components)) {
    return(NULL)
  }

  lp <- components$lp
  scale_vec <- components$scale
  n_obs <- length(lp)

  if (n_obs == 0) {
    return(matrix(
      numeric(0),
      nrow = 0,
      ncol = length(eval_times)
    ))
  }

  finite_scale <- scale_vec[is.finite(scale_vec) &
                              scale_vec > 0]
  fallback_scale <- if (length(finite_scale) > 0)
    stats::median(finite_scale)
  else
    1
  if (length(scale_vec) != n_obs ||
      any(!is.finite(scale_vec) | scale_vec <= 0)) {
    scale_vec <- rep(fallback_scale, n_obs)
  }

  dist_name <- fit_obj$dist
  if (is.list(dist_name) && !is.null(dist_name$dist)) {
    dist_name <- dist_name$dist
  }
  if (is.null(dist_name)) {
    dist_name <- "weibull"
  }
  dist_name <- as.character(dist_name)
  parms <- fit_obj$parms

  res <- matrix(NA_real_, nrow = n_obs, ncol = length(eval_times))
  for (j in seq_along(eval_times)) {
    t_val <- eval_times[j]
    if (!is.finite(t_val)) {
      next
    }
    if (t_val <= 0) {
      res[, j] <- 1
      next
    }
    q_vec <- rep(t_val, n_obs)
    surv_vals <- tryCatch({
      if (is.null(parms)) {
        survival::psurvreg(
          q_vec,
          mean = lp,
          scale = scale_vec,
          distribution = dist_name
        )
      } else {
        survival::psurvreg(
          q_vec,
          mean = lp,
          scale = scale_vec,
          distribution = dist_name,
          parms = parms
        )
      }
    }, error = function(e)
      rep(NA_real_, n_obs))

    surv_vals <- 1 - surv_vals
    surv_vals[!is.finite(surv_vals)] <- NA_real_
    res[, j] <- pmin(pmax(surv_vals, 0), 1)
  }

  res
}

#' Convert Various Prediction Formats to Survival Matrix
#'
#' Attempts to convert various survival prediction formats (e.g., list of
#' data frames from \code{predict.model_fit} with type "survival", matrices)
#' into a standardized \code{[n_obs, n_eval_times]} matrix.
#'
#' @param pred_obj The prediction object.
#' @param eval_times Numeric vector of evaluation times.
#' @param n_obs Expected number of observations (rows).
#'
#' @return A standardized matrix of survival probabilities, or \code{NULL}
#'   on failure.
#' @keywords internal
convert_survival_predictions <- function(pred_obj, eval_times, n_obs) {
  if (is.null(pred_obj) || length(eval_times) == 0 || n_obs == 0) {
    return(NULL)
  }

  if (is.matrix(pred_obj)) {
    if (nrow(pred_obj) == n_obs &&
        ncol(pred_obj) == length(eval_times)) {
      return(as.matrix(pred_obj))
    }
    if (ncol(pred_obj) == n_obs &&
        nrow(pred_obj) == length(eval_times)) {
      return(t(pred_obj))
    }
  }

  if (is.numeric(pred_obj) &&
      length(pred_obj) == n_obs * length(eval_times)) {
    return(matrix(
      pred_obj,
      nrow = n_obs,
      ncol = length(eval_times)
    ))
  }

  extract_list <- NULL
  if (is.data.frame(pred_obj)) {
    if (".pred_survival" %in% names(pred_obj)) {
      extract_list <- pred_obj$.pred_survival
    } else if (".pred" %in% names(pred_obj)) {
      extract_list <- pred_obj$.pred
    } else {
      extract_list <- pred_obj[[1]]
    }
  } else if (is.list(pred_obj)) {
    extract_list <- pred_obj
  }

  if (is.null(extract_list)) {
    return(NULL)
  }

  if (!is.list(extract_list)) {
    if (is.numeric(extract_list) &&
        length(extract_list) == n_obs * length(eval_times)) {
      return(matrix(
        extract_list,
        nrow = n_obs,
        ncol = length(eval_times)
      ))
    }
    if (is.numeric(extract_list) &&
        length(extract_list) == n_obs) {
      return(matrix(extract_list, nrow = n_obs, ncol = 1))
    }
    return(NULL)
  }

  res <- matrix(NA_real_, nrow = n_obs, ncol = length(eval_times))
  max_iter <- min(length(extract_list), n_obs)
  for (i in seq_len(max_iter)) {
    entry <- extract_list[[i]]
    if (is.null(entry))
      next
    if (is.data.frame(entry)) {
      time_col <- intersect(c(".eval_time", ".time", "time"), names(entry))
      surv_col <- intersect(c(".survival", "survival", ".pred_survival"),
                            names(entry))
      if (length(time_col) > 0 && length(surv_col) > 0) {
        res[i, ] <- align_survival_curve(entry[[time_col[1]]], entry[[surv_col[1]]], eval_times)
        next
      }
      if (ncol(entry) == length(eval_times)) {
        vals <- as.numeric(entry[1, , drop = TRUE])
        if (length(vals) == length(eval_times)) {
          res[i, ] <- vals
          next
        }
      }
    }
    if (is.numeric(entry)) {
      vals <- as.numeric(entry)
      if (length(vals) == length(eval_times)) {
        res[i, ] <- vals
        next
      }
      if (length(vals) == 1 && length(eval_times) == 1) {
        res[i, ] <- rep(vals, length(eval_times))
        next
      }
      if (length(vals) > 1) {
        take <- min(length(vals), length(eval_times))
        res[i, seq_len(take)] <- vals[seq_len(take)]
        next
      }
    }
    if (is.list(entry) && length(entry) > 0) {
      inner <- entry[[1]]
      if (is.data.frame(inner)) {
        time_col <- intersect(c(".eval_time", ".time", "time"), names(inner))
        surv_col <- intersect(c(".survival", "survival", ".pred_survival"),
                              names(inner))
        if (length(time_col) > 0 && length(surv_col) > 0) {
          res[i, ] <- align_survival_curve(inner[[time_col[1]]], inner[[surv_col[1]]], eval_times)
          next
        }
      }
      if (is.numeric(inner)) {
        vals <- as.numeric(inner)
        if (length(vals) == length(eval_times)) {
          res[i, ] <- vals
        }
      }
    }
  }
  res
}

#' Compute Integrated Brier Score and Curve
#'
#' Calculates the Brier score at specified evaluation times and the
#' Integrated Brier Score (IBS) up to \eqn{\tau}, using IPCW to handle
#' censoring.
#'
#' @param eval_times Numeric vector of evaluation time points.
#' @param surv_mat Matrix of predicted survival probabilities
#'   (rows=subjects, cols=eval_times).
#' @param time_vec Numeric vector of test times.
#' @param status_vec Numeric vector of test statuses.
#' @param tau The time horizon \eqn{\tau} for integration.
#' @param censor_eval_fn A function (from \code{create_censor_eval}) that
#'   evaluates the censoring survival function \eqn{G(t)}.
#' @param normalize_by Character string specifying how to normalize Brier
#'   scores. Use \code{"non_missing"} (default) to divide by the number of
#'   non-missing contributions or \code{"n"} to divide by the total sample size.
#' @param include_zero Logical; if \code{TRUE}, integrates from time 0 with a
#'   Brier score of 0 when the evaluation grid starts after 0.
#'
#' @return A list with \code{ibs} (the scalar IBS value) and \code{curve} (a
#'   numeric vector of Brier scores at \code{eval_times}).
#' @keywords internal
#' @noMd
#' @importFrom utils head
#' @importFrom utils tail
compute_ibrier <- function(eval_times,
                           surv_mat,
                           time_vec,
                           status_vec,
                           tau,
                           censor_eval_fn,
                           normalize_by = c("non_missing", "n"),
                           include_zero = TRUE) {
  normalize_by <- match.arg(normalize_by)
  include_zero <- isTRUE(include_zero)
  n <- length(time_vec)
  m <- length(eval_times)
  if (n == 0 ||
      m == 0 ||
      is.null(surv_mat) || nrow(surv_mat) != n || ncol(surv_mat) != m) {
    return(list(ibs = NA_real_, curve = rep(NA_real_, m)))
  }

  status_vec <- ifelse(is.na(status_vec), 0, status_vec)
  status_vec <- ifelse(status_vec > 0, 1, 0)
  time_vec <- as.numeric(time_vec)

  valid_idx <- which(!is.na(time_vec) & !is.na(status_vec))
  if (length(valid_idx) == 0) {
    return(list(ibs = NA_real_, curve = rep(NA_real_, m)))
  }

  if (!is.function(censor_eval_fn)) {
    return(list(ibs = NA_real_, curve = rep(NA_real_, m)))
  }

  G_t <- censor_eval_fn(eval_times)
  G_t[!is.finite(G_t) | G_t <= 0] <- NA_real_

  time_minus <- pmax(time_vec - 1e-08, 0)
  G_time_minus <- censor_eval_fn(time_minus)
  G_time_minus[!is.finite(G_time_minus) |
                 G_time_minus <= 0] <- NA_real_

  weights <- matrix(NA_real_, nrow = n, ncol = m)
  for (j in seq_len(m)) {
    t_val <- eval_times[j]
    g_t <- G_t[j]
    for (i in seq_len(n)) {
      ti <- time_vec[i]
      if (!is.finite(ti)) {
        next
      }
      if (ti <= t_val && status_vec[i] == 1) {
        denom <- G_time_minus[i]
        if (is.finite(denom) && denom > 0) {
          weights[i, j] <- 1 / denom
        }
      } else if (ti > t_val) {
        if (is.finite(g_t) && g_t > 0) {
          weights[i, j] <- 1 / g_t
        }
      }
    }
  }

  indicator <- matrix(0, nrow = n, ncol = m)
  for (j in seq_len(m)) {
    indicator[, j] <- as.numeric(time_vec > eval_times[j])
  }

  residual <- indicator - surv_mat
  residual[!is.finite(residual)] <- NA_real_
  weights[!is.finite(weights) | weights <= 0] <- NA_real_
  weighted <- weights * residual^2
  weighted[!is.finite(weighted)] <- NA_real_
  denom_counts <- colSums(!is.na(weighted))
  bs_t <- colSums(weighted, na.rm = TRUE)
  if (normalize_by == "n") {
    bs_t <- bs_t / n
  } else if (length(denom_counts) == length(bs_t)) {
    valid_cols <- denom_counts > 0
    bs_t[valid_cols] <- bs_t[valid_cols] / denom_counts[valid_cols]
    bs_t[!valid_cols] <- NA_real_
  } else {
    bs_t <- bs_t / n
  }
  bs_t[!is.finite(bs_t)] <- NA_real_

  if (!is.finite(tau) || tau <= 0) {
    tau <- suppressWarnings(max(eval_times[is.finite(eval_times)], na.rm = TRUE))
  }
  if (is.finite(tau)) {
    bs_t[eval_times > tau] <- NA_real_
  }

  valid_bs <- which(is.finite(bs_t) &
                      is.finite(eval_times) & eval_times <= tau + 1e-08)
  if (length(valid_bs) >= 2) {
    times_valid <- eval_times[valid_bs]
    bs_valid <- bs_t[valid_bs]
    ord <- order(times_valid)
    times_valid <- times_valid[ord]
    bs_valid <- bs_valid[ord]
    if (include_zero && length(times_valid) > 0 && times_valid[1] > 0) {
      times_aug <- c(0, times_valid)
      bs_aug <- c(0, bs_valid)
    } else {
      times_aug <- times_valid
      bs_aug <- bs_valid
    }
    area <- sum(diff(times_aug) * (head(bs_aug, -1) + tail(bs_aug, -1)) / 2)
    tau <- max(times_valid)
    ibs <- if (tau > 0)
      area / tau
    else
      NA_real_
  } else if (length(valid_bs) == 1) {
    ibs <- bs_t[valid_bs]
  } else {
    ibs <- NA_real_
  }

  list(ibs = ibs, curve = bs_t)
}

fastml_normalize_class_name <- function(x) {
  tolower(trimws(as.character(x)))
}

fastml_truth_levels <- function(truth_vec) {
  if (is.factor(truth_vec)) {
    levels(truth_vec)
  } else {
    unique(as.character(truth_vec))
  }
}

fastml_prob_cols_from_dots <- function(data, dots) {
  if (length(dots) == 0) {
    return(grep("^\\.pred", names(data), value = TRUE))
  }
  cols <- vapply(dots, rlang::as_name, character(1))
  cols <- cols[cols %in% names(data)]
  if (length(cols) == 0) {
    cols <- grep("^\\.pred", names(data), value = TRUE)
  }
  cols
}

fastml_prepare_prob_matrix <- function(data, prob_cols) {
  prob_cols <- intersect(as.character(prob_cols), names(data))
  if (length(prob_cols) == 0) {
    return(NULL)
  }
  num_cols <- prob_cols[vapply(data[prob_cols], is.numeric, logical(1))]
  if (length(num_cols) == 0) {
    return(NULL)
  }
  list(mat = as.matrix(data[num_cols]), cols = num_cols)
}

fastml_align_prob_matrix <- function(prob_mat, prob_cols, truth_levels) {
  if (is.null(prob_mat) || length(prob_cols) == 0 || length(truth_levels) == 0) {
    return(NULL)
  }
  prob_labels <- sub("^\\.pred_p", "", sub("^\\.pred_", "", prob_cols))
  prob_norm <- fastml_normalize_class_name(make.names(prob_labels))
  truth_norm <- fastml_normalize_class_name(make.names(truth_levels))
  match_idx <- match(truth_norm, prob_norm)
  if (any(is.na(match_idx))) {
    return(NULL)
  }
  list(mat = prob_mat[, match_idx, drop = FALSE], cols = prob_cols[match_idx])
}

fastml_binary_logloss <- function(truth_vec, prob_vec, positive_class) {
  if (is.null(positive_class) || length(prob_vec) == 0) {
    return(NA_real_)
  }
  y <- as.character(truth_vec) == as.character(positive_class)
  valid <- is.finite(prob_vec) & !is.na(y)
  if (!any(valid)) {
    return(NA_real_)
  }
  p <- prob_vec[valid]
  y <- as.numeric(y[valid])
  eps <- 1e-15
  p <- pmin(pmax(p, eps), 1 - eps)
  -mean(y * log(p) + (1 - y) * log(1 - p))
}

fastml_binary_brier <- function(truth_vec, prob_vec, positive_class) {
  if (is.null(positive_class) || length(prob_vec) == 0) {
    return(NA_real_)
  }
  y <- as.character(truth_vec) == as.character(positive_class)
  valid <- is.finite(prob_vec) & !is.na(y)
  if (!any(valid)) {
    return(NA_real_)
  }
  p <- prob_vec[valid]
  y <- as.numeric(y[valid])
  mean((p - y)^2)
}

fastml_binary_ece <- function(truth_vec, prob_vec, positive_class, n_bins = 10) {
  if (is.null(positive_class) || length(prob_vec) == 0) {
    return(NA_real_)
  }
  n_bins <- as.integer(n_bins)
  if (!is.finite(n_bins) || n_bins < 1) {
    n_bins <- 10L
  }
  y <- as.character(truth_vec) == as.character(positive_class)
  valid <- is.finite(prob_vec) & !is.na(y)
  if (!any(valid)) {
    return(NA_real_)
  }
  p <- pmin(pmax(prob_vec[valid], 0), 1)
  y <- as.numeric(y[valid])
  bins <- cut(p,
              breaks = seq(0, 1, length.out = n_bins + 1L),
              include.lowest = TRUE,
              labels = FALSE)
  n <- length(p)
  ece <- 0
  for (b in seq_len(n_bins)) {
    idx <- which(bins == b)
    if (length(idx) == 0) {
      next
    }
    conf <- mean(p[idx])
    acc <- mean(y[idx])
    ece <- ece + abs(acc - conf) * (length(idx) / n)
  }
  ece
}

fastml_multiclass_logloss <- function(truth_vec, prob_mat, truth_levels) {
  n <- length(truth_vec)
  if (is.null(prob_mat) || n == 0 || nrow(prob_mat) != n) {
    return(NA_real_)
  }
  idx <- match(as.character(truth_vec), truth_levels)
  p_true <- prob_mat[cbind(seq_len(n), idx)]
  valid <- !is.na(idx) & is.finite(p_true)
  if (!any(valid)) {
    return(NA_real_)
  }
  eps <- 1e-15
  p_true <- pmin(pmax(p_true, eps), 1 - eps)
  -mean(log(p_true[valid]))
}

fastml_multiclass_brier <- function(truth_vec, prob_mat, truth_levels) {
  n <- length(truth_vec)
  if (is.null(prob_mat) || n == 0 || nrow(prob_mat) != n) {
    return(NA_real_)
  }
  idx <- match(as.character(truth_vec), truth_levels)
  sum_sq <- rowSums(prob_mat^2)
  p_true <- prob_mat[cbind(seq_len(n), idx)]
  valid <- !is.na(idx) & is.finite(p_true) & is.finite(sum_sq)
  if (!any(valid)) {
    return(NA_real_)
  }
  mean(sum_sq[valid] - 2 * p_true[valid] + 1)
}

fastml_multiclass_ece <- function(truth_vec, prob_mat, truth_levels, n_bins = 10) {
  n <- length(truth_vec)
  if (is.null(prob_mat) || n == 0 || nrow(prob_mat) != n) {
    return(NA_real_)
  }
  n_bins <- as.integer(n_bins)
  if (!is.finite(n_bins) || n_bins < 1) {
    n_bins <- 10L
  }
  prob_mat <- pmin(pmax(prob_mat, 0), 1)
  row_has_finite <- apply(prob_mat, 1, function(x) any(is.finite(x)))
  safe_mat <- prob_mat
  safe_mat[!is.finite(safe_mat)] <- -Inf
  max_idx <- max.col(safe_mat, ties.method = "first")
  conf <- prob_mat[cbind(seq_len(n), max_idx)]
  conf[!row_has_finite] <- NA_real_
  pred_class <- truth_levels[max_idx]
  correct <- as.character(truth_vec) == as.character(pred_class)
  valid <- is.finite(conf) & !is.na(correct)
  if (!any(valid)) {
    return(NA_real_)
  }
  conf <- conf[valid]
  correct <- as.numeric(correct[valid])
  bins <- cut(conf,
              breaks = seq(0, 1, length.out = n_bins + 1L),
              include.lowest = TRUE,
              labels = FALSE)
  n_valid <- length(conf)
  ece <- 0
  for (b in seq_len(n_bins)) {
    idx <- which(bins == b)
    if (length(idx) == 0) {
      next
    }
    conf_mean <- mean(conf[idx])
    acc_mean <- mean(correct[idx])
    ece <- ece + abs(acc_mean - conf_mean) * (length(idx) / n_valid)
  }
  ece
}

fastml_class_calibration_metrics <- function(truth_vec, prob_mat, prob_cols, event_class) {
  truth_levels <- fastml_truth_levels(truth_vec)
  n_classes <- length(truth_levels)
  if (n_classes < 2 || is.null(prob_mat) || length(prob_cols) == 0) {
    return(list(logloss = NA_real_,
                brier_score = NA_real_,
                ece = NA_real_,
                estimator = NA_character_))
  }
  if (n_classes == 2) {
    if (is.null(event_class) || !event_class %in% c("first", "second")) {
      event_class <- "first"
    }
    prob_meta <- fastml_resolve_binary_prob_column(prob_cols, truth_levels, event_class)
    if (is.null(prob_meta$prob_col)) {
      return(list(logloss = NA_real_,
                  brier_score = NA_real_,
                  ece = NA_real_,
                  estimator = NA_character_))
    }
    col_idx <- match(prob_meta$prob_col, prob_cols)
    if (is.na(col_idx)) {
      return(list(logloss = NA_real_,
                  brier_score = NA_real_,
                  ece = NA_real_,
                  estimator = NA_character_))
    }
    prob_vec <- prob_mat[, col_idx]
    logloss <- fastml_binary_logloss(truth_vec, prob_vec, prob_meta$positive_class)
    brier <- fastml_binary_brier(truth_vec, prob_vec, prob_meta$positive_class)
    ece <- fastml_binary_ece(truth_vec, prob_vec, prob_meta$positive_class)
    return(list(logloss = logloss,
                brier_score = brier,
                ece = ece,
                estimator = NA_character_))
  }
  aligned <- fastml_align_prob_matrix(prob_mat, prob_cols, truth_levels)
  if (is.null(aligned)) {
    return(list(logloss = NA_real_,
                brier_score = NA_real_,
                ece = NA_real_,
                estimator = NA_character_))
  }
  prob_mat <- pmin(pmax(aligned$mat, 0), 1)
  logloss <- fastml_multiclass_logloss(truth_vec, prob_mat, truth_levels)
  brier <- fastml_multiclass_brier(truth_vec, prob_mat, truth_levels)
  ece <- fastml_multiclass_ece(truth_vec, prob_mat, truth_levels)
  list(logloss = logloss, brier_score = brier, ece = ece, estimator = "macro")
}

fastml_wrap_event_level_metric <- function(metric, event_class) {
  if (is.null(event_class) || !event_class %in% c("first", "second")) {
    return(metric)
  }

  metric_fun <- function(data, truth, estimate, na_rm = TRUE, case_weights = NULL, ...) {
    truth_quo <- rlang::enquo(truth)
    estimate_quo <- rlang::enquo(estimate)
    case_weights_quo <- rlang::enquo(case_weights)
    if (rlang::quo_is_missing(truth_quo) || rlang::quo_is_missing(estimate_quo)) {
      metric_name <- attr(metric, "metric_name")
      if (is.null(metric_name) || !nzchar(metric_name)) {
        metric_name <- "metric"
      }
      return(tibble::tibble(
        .metric = metric_name,
        .estimator = NA_character_,
        .estimate = NA_real_
      ))
    }
    truth_name <- fastml_resolve_truth_name(data, truth_quo)
    estimate_name <- fastml_resolve_estimate_name(data, estimate_quo)
    if (is.null(truth_name) || is.null(estimate_name)) {
      metric_name <- attr(metric, "metric_name")
      if (is.null(metric_name) || !nzchar(metric_name)) {
        metric_name <- "metric"
      }
      return(tibble::tibble(
        .metric = metric_name,
        .estimator = NA_character_,
        .estimate = NA_real_
      ))
    }
    dots <- rlang::dots_list(...)
    metric_args <- names(tryCatch(formals(metric), error = function(e) list()))
    call_args <- list(
      data = quote(data),
      truth = rlang::sym(truth_name),
      estimate = rlang::sym(estimate_name)
    )
    if ("na_rm" %in% metric_args) {
      call_args$na_rm <- na_rm
    }
    if ("event_level" %in% metric_args) {
      call_args$event_level <- event_class
    }
    if ("case_weights" %in% metric_args && !rlang::quo_is_missing(case_weights_quo)) {
      call_args$case_weights <- rlang::get_expr(case_weights_quo)
    }
    if ("estimator" %in% metric_args && "estimator" %in% names(dots)) {
      call_args$estimator <- dots$estimator
    }
    call <- rlang::call2(metric, !!!call_args)
    result <- tryCatch(rlang::eval_tidy(call), error = function(e) NULL)
    if (is.null(result)) {
      metric_name <- attr(metric, "metric_name")
      if (is.null(metric_name) || !nzchar(metric_name)) {
        metric_name <- "metric"
      }
      return(tibble::tibble(
        .metric = metric_name,
        .estimator = NA_character_,
        .estimate = NA_real_
      ))
    }
    result
  }

  class(metric_fun) <- class(metric)
  attr(metric_fun, "direction") <- attr(metric, "direction")
  attr(metric_fun, "metric_name") <- attr(metric, "metric_name")
  attr(metric_fun, "range") <- attr(metric, "range")
  static <- attr(metric, "static", exact = TRUE)
  if (is.null(static)) {
    static <- "event_level"
  } else {
    static <- unique(c(static, "event_level"))
  }
  attr(metric_fun, "static") <- static
  metric_fun
}

fastml_wrap_basic_metric <- function(metric) {
  metric_fun <- function(data, truth, estimate, na_rm = TRUE, case_weights = NULL, ...) {
    truth_quo <- rlang::enquo(truth)
    estimate_quo <- rlang::enquo(estimate)
    case_weights_quo <- rlang::enquo(case_weights)
    if (rlang::quo_is_missing(truth_quo) || rlang::quo_is_missing(estimate_quo)) {
      metric_name <- attr(metric, "metric_name")
      if (is.null(metric_name) || !nzchar(metric_name)) {
        metric_name <- "metric"
      }
      return(tibble::tibble(
        .metric = metric_name,
        .estimator = NA_character_,
        .estimate = NA_real_
      ))
    }
    truth_name <- fastml_resolve_truth_name(data, truth_quo)
    estimate_name <- fastml_resolve_estimate_name(data, estimate_quo)
    if (is.null(truth_name) || is.null(estimate_name)) {
      metric_name <- attr(metric, "metric_name")
      if (is.null(metric_name) || !nzchar(metric_name)) {
        metric_name <- "metric"
      }
      return(tibble::tibble(
        .metric = metric_name,
        .estimator = NA_character_,
        .estimate = NA_real_
      ))
    }
    dots <- rlang::dots_list(...)
    metric_args <- names(tryCatch(formals(metric), error = function(e) list()))
    call_args <- list(
      data = quote(data),
      truth = rlang::sym(truth_name),
      estimate = rlang::sym(estimate_name)
    )
    if ("na_rm" %in% metric_args) {
      call_args$na_rm <- na_rm
    }
    if ("case_weights" %in% metric_args && !rlang::quo_is_missing(case_weights_quo)) {
      call_args$case_weights <- rlang::get_expr(case_weights_quo)
    }
    if ("estimator" %in% metric_args && "estimator" %in% names(dots)) {
      call_args$estimator <- dots$estimator
    }
    call <- rlang::call2(metric, !!!call_args)
    result <- tryCatch(rlang::eval_tidy(call), error = function(e) NULL)
    if (is.null(result)) {
      metric_name <- attr(metric, "metric_name")
      if (is.null(metric_name) || !nzchar(metric_name)) {
        metric_name <- "metric"
      }
      return(tibble::tibble(
        .metric = metric_name,
        .estimator = NA_character_,
        .estimate = NA_real_
      ))
    }
    result
  }

  class(metric_fun) <- class(metric)
  attr(metric_fun, "direction") <- attr(metric, "direction")
  attr(metric_fun, "metric_name") <- attr(metric, "metric_name")
  attr(metric_fun, "range") <- attr(metric, "range")
  attr(metric_fun, "static") <- attr(metric, "static", exact = TRUE)
  metric_fun
}

fastml_resolve_binary_prob_column <- function(prob_cols,
                                              truth_levels,
                                              event_class,
                                              allow_fallback = FALSE) {
  if (!event_class %in% c("first", "second")) {
    stop("Invalid event_class argument. It should be either 'first' or 'second'.")
  }

  prob_cols <- as.character(prob_cols)
  truth_levels <- as.character(truth_levels)

  result <- list(
    prob_col = NULL,
    positive_class = NULL,
    candidates = character(),
    used_fallback = FALSE
  )

  if (length(prob_cols) == 0 || length(truth_levels) < 2) {
    return(result)
  }

  pos_idx <- if (event_class == "first") 1L else 2L
  positive_class <- truth_levels[pos_idx]
  result$positive_class <- positive_class

  candidate_levels <- unique(c(positive_class, make.names(positive_class)))
  candidates <- unique(c(
    paste0(".pred_", candidate_levels),
    paste0(".pred_p", candidate_levels)
  ))
  event_idx <- if (event_class == "first") 0L else 1L
  candidates <- unique(c(candidates, paste0(".pred_p", event_idx)))
  result$candidates <- candidates

  prob_col <- candidates[candidates %in% prob_cols][1]
  if (!is.na(prob_col) && length(prob_col) > 0) {
    result$prob_col <- prob_col
    return(result)
  }

  prob_levels <- sub("^\\.pred_p", "", sub("^\\.pred_", "", prob_cols))
  prob_levels_norm <- fastml_normalize_class_name(make.names(prob_levels))
  candidate_norm <- fastml_normalize_class_name(make.names(candidate_levels))
  match_idx <- match(candidate_norm, prob_levels_norm)
  match_idx <- match_idx[!is.na(match_idx)][1]
  if (!is.na(match_idx)) {
    result$prob_col <- prob_cols[match_idx]
    return(result)
  }

  if (isTRUE(allow_fallback) && length(prob_cols) == 2) {
    result$prob_col <- prob_cols[pos_idx]
    result$used_fallback <- TRUE
  }

  result
}

fastml_normalize_multiclass_auc <- function(multiclass_auc) {
  if (is.null(multiclass_auc) || length(multiclass_auc) == 0 || is.na(multiclass_auc[[1]])) {
    return("macro")
  }

  multiclass_auc <- tolower(as.character(multiclass_auc[[1]]))
  allowed <- c("macro", "macro_weighted")
  if (!multiclass_auc %in% allowed) {
    stop(
      sprintf(
        "Invalid multiclass_auc '%s'. Choose one of: %s.",
        multiclass_auc,
        paste(allowed, collapse = ", ")
      ),
      call. = FALSE
    )
  }

  multiclass_auc
}

fastml_auc_estimator_from_truth <- function(truth, multiclass_auc) {
  multiclass_auc <- fastml_normalize_multiclass_auc(multiclass_auc)
  if (is.factor(truth)) {
    n_levels <- length(levels(truth))
  } else {
    truth_vals <- truth[!is.na(truth)]
    n_levels <- length(unique(truth_vals))
  }
  if (n_levels > 2) multiclass_auc else "binary"
}

fastml_resolve_truth_name <- function(data, truth_quo) {
  if (!is.data.frame(data)) {
    return(NULL)
  }
  truth_name <- NULL
  if (!rlang::quo_is_missing(truth_quo)) {
    truth_name <- tryCatch(rlang::as_name(truth_quo), error = function(e) NULL)
  }
  if (!is.null(truth_name) && nzchar(truth_name) && truth_name %in% names(data)) {
    return(truth_name)
  }
  fallback_names <- c("truth", ".truth", ".outcome")
  fallback_names <- fallback_names[fallback_names %in% names(data)]
  if (length(fallback_names) > 0) {
    return(fallback_names[1])
  }
  NULL
}

fastml_resolve_estimate_name <- function(data, estimate_quo) {
  if (!is.data.frame(data)) {
    return(NULL)
  }
  estimate_name <- NULL
  if (!rlang::quo_is_missing(estimate_quo)) {
    estimate_name <- tryCatch(rlang::as_name(estimate_quo), error = function(e) NULL)
  }
  if (!is.null(estimate_name) && nzchar(estimate_name) && estimate_name %in% names(data)) {
    return(estimate_name)
  }
  fallback_names <- c("estimate", ".pred_class", ".pred")
  fallback_names <- fallback_names[fallback_names %in% names(data)]
  if (length(fallback_names) > 0) {
    return(fallback_names[1])
  }
  NULL
}

fastml_configured_roc_auc <- function(multiclass_auc, event_class = NULL) {
  multiclass_auc <- fastml_normalize_multiclass_auc(multiclass_auc)

  roc_fun <- function(data, truth, ..., estimator = NULL) {
    na_result <- function(estimator_value) {
      tibble::tibble(
        .metric = "roc_auc",
        .estimator = estimator_value,
        .estimate = NA_real_
      )
    }

    tryCatch({
      # Extract the truth column name from the argument
      if (!is.data.frame(data)) {
        return(na_result(NA_character_))
      }

      truth_quo <- rlang::enquo(truth)
      truth_name <- fastml_resolve_truth_name(data, truth_quo)
      if (is.null(truth_name) || !nzchar(truth_name)) {
        return(na_result(NA_character_))
      }

      truth_col <- data[[truth_name]]
      if (is.null(truth_col)) {
        return(na_result(NA_character_))
      }

      # Determine estimator type
      if (missing(estimator) || is.null(estimator) || length(estimator) == 0 || is.na(estimator[[1]])) {
        estimator <- fastml_auc_estimator_from_truth(truth_col, multiclass_auc)
      }
      estimator <- as.character(estimator)[1]
      if (is.null(estimator) || is.na(estimator) || !nzchar(estimator)) {
        estimator <- NA_character_
      }

      truth_vals <- truth_col[!is.na(truth_col)]
      if (length(unique(truth_vals)) < 2) {
        return(na_result(estimator))
      }

      dots <- rlang::enquos(...)
      prob_cols <- fastml_prob_cols_from_dots(data, dots)
      prob_info <- fastml_prepare_prob_matrix(data, prob_cols)
      if (is.null(prob_info)) {
        return(na_result(estimator))
      }

      prob_cols <- prob_info$cols
      if (length(prob_cols) == 0) {
        return(na_result(estimator))
      }

      # Use the vector-based approach to avoid NSE issues
      if (identical(estimator, "binary")) {
        truth_levels <- fastml_truth_levels(truth_col)
        event_use <- if (!is.null(event_class) && event_class %in% c("first", "second")) {
          event_class
        } else {
          "first"
        }
        prob_meta <- fastml_resolve_binary_prob_column(
          prob_cols,
          truth_levels,
          event_use,
          allow_fallback = FALSE
        )
        prob_col <- prob_meta$prob_col
        if (is.null(prob_col) || !prob_col %in% prob_cols) {
          return(na_result(estimator))
        }

        result <- tryCatch({
          # Use yardstick::roc_auc_vec for vector-based computation (no NSE)
          truth_vec <- truth_col
          prob_vec <- data[[prob_col]]

          if (!is.factor(truth_vec)) {
            truth_vec <- factor(truth_vec)
          }

          event_level_use <- if (!is.null(event_class) && event_class %in% c("first", "second")) {
            event_class
          } else {
            "first"
          }

          auc_value <- yardstick::roc_auc_vec(
            truth = truth_vec,
            estimate = prob_vec,
            event_level = event_level_use
          )

          tibble::tibble(
            .metric = "roc_auc",
            .estimator = "binary",
            .estimate = auc_value
          )
        }, error = function(e) {
          na_result("binary")
        })

        return(result)
      } else {
        # Multiclass case
        truth_levels <- fastml_truth_levels(truth_col)
        aligned <- fastml_align_prob_matrix(prob_info$mat, prob_cols, truth_levels)
        if (is.null(aligned) || length(aligned$cols) == 0) {
          return(na_result(estimator))
        }

        result <- tryCatch({
          truth_vec <- truth_col
          if (!is.factor(truth_vec)) {
            truth_vec <- factor(truth_vec, levels = truth_levels)
          }

          # Create probability matrix for multiclass
          prob_mat <- as.matrix(data[, aligned$cols, drop = FALSE])

          auc_value <- yardstick::roc_auc_vec(
            truth = truth_vec,
            estimate = prob_mat,
            estimator = estimator
          )

          tibble::tibble(
            .metric = "roc_auc",
            .estimator = estimator,
            .estimate = auc_value
          )
        }, error = function(e) {
          na_result(estimator)
        })

        return(result)
      }
    }, error = function(e) {
      na_result(NA_character_)
    })
  }

  class(roc_fun) <- class(yardstick::roc_auc)
  attr(roc_fun, "direction") <- attr(yardstick::roc_auc, "direction")
  attr(roc_fun, "metric_name") <- attr(yardstick::roc_auc, "metric_name")
  roc_fun
}

fastml_configured_logloss <- function(event_class = NULL) {
  logloss_fun <- function(data, truth, ..., estimator = NULL) {
    tryCatch({
      if (!is.data.frame(data)) {
        return(tibble::tibble(
          .metric = "logloss",
          .estimator = NA_character_,
          .estimate = NA_real_
        ))
      }
      truth_quo <- rlang::enquo(truth)
      truth_col <- fastml_resolve_truth_name(data, truth_quo)
      if (is.null(truth_col) || !nzchar(truth_col)) {
        return(tibble::tibble(
          .metric = "logloss",
          .estimator = NA_character_,
          .estimate = NA_real_
        ))
      }
      truth_vec <- data[[truth_col]]
      dots <- rlang::enquos(...)
      prob_cols <- fastml_prob_cols_from_dots(data, dots)
      prob_info <- fastml_prepare_prob_matrix(data, prob_cols)
      if (is.null(prob_info)) {
        estimate <- NA_real_
        estimator_out <- NA_character_
      } else {
        metrics <- fastml_class_calibration_metrics(truth_vec,
                                                    prob_info$mat,
                                                    prob_info$cols,
                                                    event_class)
        estimate <- metrics$logloss
        estimator_out <- metrics$estimator
      }
      tibble::tibble(
        .metric = "logloss",
        .estimator = estimator_out,
        .estimate = as.numeric(estimate)
      )
    }, error = function(e) {
      tibble::tibble(
        .metric = "logloss",
        .estimator = NA_character_,
        .estimate = NA_real_
      )
    })
  }

  class(logloss_fun) <- class(yardstick::roc_auc)
  attr(logloss_fun, "direction") <- "minimize"
  attr(logloss_fun, "metric_name") <- "logloss"
  logloss_fun
}

fastml_configured_brier_score <- function(event_class = NULL) {
  brier_fun <- function(data, truth, ..., estimator = NULL) {
    tryCatch({
      if (!is.data.frame(data)) {
        return(tibble::tibble(
          .metric = "brier_score",
          .estimator = NA_character_,
          .estimate = NA_real_
        ))
      }
      truth_quo <- rlang::enquo(truth)
      truth_col <- fastml_resolve_truth_name(data, truth_quo)
      if (is.null(truth_col) || !nzchar(truth_col)) {
        return(tibble::tibble(
          .metric = "brier_score",
          .estimator = NA_character_,
          .estimate = NA_real_
        ))
      }
      truth_vec <- data[[truth_col]]
      dots <- rlang::enquos(...)
      prob_cols <- fastml_prob_cols_from_dots(data, dots)
      prob_info <- fastml_prepare_prob_matrix(data, prob_cols)
      if (is.null(prob_info)) {
        estimate <- NA_real_
        estimator_out <- NA_character_
      } else {
        metrics <- fastml_class_calibration_metrics(truth_vec,
                                                    prob_info$mat,
                                                    prob_info$cols,
                                                    event_class)
        estimate <- metrics$brier_score
        estimator_out <- metrics$estimator
      }
      tibble::tibble(
        .metric = "brier_score",
        .estimator = estimator_out,
        .estimate = as.numeric(estimate)
      )
    }, error = function(e) {
      tibble::tibble(
        .metric = "brier_score",
        .estimator = NA_character_,
        .estimate = NA_real_
      )
    })
  }

  class(brier_fun) <- class(yardstick::roc_auc)
  attr(brier_fun, "direction") <- "minimize"
  attr(brier_fun, "metric_name") <- "brier_score"
  brier_fun
}

fastml_configured_ece <- function(event_class = NULL) {
  ece_fun <- function(data, truth, ..., estimator = NULL) {
    tryCatch({
      if (!is.data.frame(data)) {
        return(tibble::tibble(
          .metric = "ece",
          .estimator = NA_character_,
          .estimate = NA_real_
        ))
      }
      truth_quo <- rlang::enquo(truth)
      truth_col <- fastml_resolve_truth_name(data, truth_quo)
      if (is.null(truth_col) || !nzchar(truth_col)) {
        return(tibble::tibble(
          .metric = "ece",
          .estimator = NA_character_,
          .estimate = NA_real_
        ))
      }
      truth_vec <- data[[truth_col]]
      dots <- rlang::enquos(...)
      prob_cols <- fastml_prob_cols_from_dots(data, dots)
      prob_info <- fastml_prepare_prob_matrix(data, prob_cols)
      if (is.null(prob_info)) {
        estimate <- NA_real_
        estimator_out <- NA_character_
      } else {
        metrics <- fastml_class_calibration_metrics(truth_vec,
                                                    prob_info$mat,
                                                    prob_info$cols,
                                                    event_class)
        estimate <- metrics$ece
        estimator_out <- metrics$estimator
      }
      tibble::tibble(
        .metric = "ece",
        .estimator = estimator_out,
        .estimate = as.numeric(estimate)
      )
    }, error = function(e) {
      tibble::tibble(
        .metric = "ece",
        .estimator = NA_character_,
        .estimate = NA_real_
      )
    })
  }

  class(ece_fun) <- class(yardstick::roc_auc)
  attr(ece_fun, "direction") <- "minimize"
  attr(ece_fun, "metric_name") <- "ece"
  ece_fun
}

library(testthat)

test_that("binary roc_auc uses available probability columns", {
  skip_on_cran()
  required <- c("parsnip", "workflows", "recipes", "yardstick", "rsample")
  lapply(required, skip_if_not_installed)

  data(iris)
  df <- iris[iris$Species != "setosa", c("Sepal.Length", "Sepal.Width", "Species")]
  df$Species <- factor(df$Species)
  levels(df$Species) <- c("negative class", "positive class")

  fit <- fastml(
    data = df,
    label = "Species",
    algorithms = "logistic_reg",
    task = "classification",
    resampling_method = "none",
    use_default_tuning = FALSE,
    test_size = 0.3,
    seed = 123,
    bootstrap_ci = FALSE,
    event_class = "second"
  )

  perf <- fit$performance[[1]]
  if (is.list(perf) && !is.data.frame(perf)) {
    perf <- perf[[1]]
  }

  expect_true("roc_auc" %in% perf$.metric)
  roc_val <- perf$.estimate[perf$.metric == "roc_auc"][1]
  expect_true(is.finite(roc_val))

  preds <- fit$predictions[[1]]
  if (is.list(preds) && !is.data.frame(preds)) {
    preds <- preds[[1]]
  }
  expect_true(any(grepl("^\\.pred_", names(preds))))
})

test_that("binary probability column resolver handles sanitized names and explicit fallback", {
  res <- fastml:::fastml_resolve_binary_prob_column(
    prob_cols = c(".pred_pos.class", ".pred_neg.class"),
    truth_levels = c("pos class", "neg class"),
    event_class = "first"
  )
  expect_identical(res$prob_col, ".pred_pos.class")
  expect_false(isTRUE(res$used_fallback))

  res_fallback <- fastml:::fastml_resolve_binary_prob_column(
    prob_cols = c(".prob0", ".prob1"),
    truth_levels = c("no", "yes"),
    event_class = "second"
  )
  expect_null(res_fallback$prob_col)
  expect_false(isTRUE(res_fallback$used_fallback))

  res_fallback_allowed <- fastml:::fastml_resolve_binary_prob_column(
    prob_cols = c(".prob0", ".prob1"),
    truth_levels = c("no", "yes"),
    event_class = "second",
    allow_fallback = TRUE
  )
  expect_identical(res_fallback_allowed$prob_col, ".prob1")
  expect_true(isTRUE(res_fallback_allowed$used_fallback))
})

test_that("configured roc_auc accepts explicit estimator", {
  required <- c("yardstick")
  lapply(required, skip_if_not_installed)

  data(iris)
  df <- iris[iris$Species != "setosa", ]
  df$Species <- factor(df$Species)

  set.seed(123)
  df$.pred_versicolor <- runif(nrow(df))

  roc_fun <- fastml:::fastml_configured_roc_auc("macro")

  expect_silent(
    roc_fun(df, truth = Species, .pred_versicolor, estimator = "binary")
  )
})

test_that("logloss and brier_score are computed when probabilities are available", {
  skip_on_cran()
  required <- c("parsnip", "workflows", "recipes", "yardstick", "rsample")
  lapply(required, skip_if_not_installed)

  data(iris)
  df <- iris[iris$Species != "setosa", c("Sepal.Length", "Sepal.Width", "Species")]
  df$Species <- factor(df$Species)
  levels(df$Species) <- c("negative class", "positive class")

  fit <- fastml(
    data = df,
    label = "Species",
    algorithms = "logistic_reg",
    task = "classification",
    resampling_method = "none",
    use_default_tuning = FALSE,
    test_size = 0.3,
    seed = 321,
    bootstrap_ci = FALSE,
    metric = "logloss",
    event_class = "second"
  )

  perf <- fit$performance[[1]]
  if (is.list(perf) && !is.data.frame(perf)) {
    perf <- perf[[1]]
  }

  expect_true(all(c("logloss", "brier_score", "ece") %in% perf$.metric))
  calib_vals <- perf$.estimate[perf$.metric %in% c("logloss", "brier_score", "ece")]
  expect_true(all(is.finite(calib_vals)))
  expect_true(all(calib_vals >= 0))
})

#' Summary Function for fastml (Using yardstick for ROC Curves)
#'
#' @description
#' Summarizes the results of machine learning models trained using the `fastml` package.
#' Depending on the task type (classification or regression), it provides customized output such as
#' performance metrics, best hyperparameter settings, and confusion matrices.
#' It is designed to be informative and readable, helping users quickly interpret model results.

#' @details
#' For classification tasks, the summary includes metrics such as Accuracy, F1 Score, Kappa,
#' Precision, ROC AUC, Sensitivity, and Specificity. A confusion matrix is also provided for the best model(s).
#' For regression tasks, the summary reports RMSE, R-squared, and MAE.
#'
#' Users can control the type of output with the `type` argument:
#' `metrics` displays model performance metrics.
#' `params` shows the best hyperparameter settings.
#' `conf_mat` prints confusion matrices (only for classification).
#' `all` includes all of the above.
#'
#' If multiple algorithms are trained, the summary highlights the best model based on the optimized metric.
#' For survival tasks, Harrell's C-index, Uno's C-index, the integrated Brier
#' score, and (when available) the RMST difference are shown by default. Specific
#' Brier(t) horizons can be requested through the \code{brier_times} argument.
#'
#'
#' @param object An object of class \code{fastml}.
#' @param algorithm A vector of algorithm names to display summary. Default is \code{"best"}.
#' @param type Character vector indicating which outputs to produce.
#'   Options are \code{"all"} (all available outputs),
#'   \code{"metrics"} (performance metrics),
#'   \code{"params"} (best hyperparameters), and
#'   \code{"conf_mat"} (confusion matrix).
#'   Default is \code{"all"}.
#' @param sort_metric The metric to sort by. Default uses optimized metric.
#' @param show_ci Logical indicating whether to display 95\% confidence intervals
#'   for performance metrics in survival models. Defaults to \code{FALSE}.
#' @param brier_times Optional numeric or character vector that selects which
#'   time-specific Brier scores to display for survival models. When \code{NULL}
#'   (the default), time-specific Brier scores are omitted from the summary.
#' @param ... Additional arguments.
#' @return Prints summary of fastml models.
#'
#' @method summary fastml
#'
#' @importFrom dplyr filter select mutate bind_rows group_by summarise n starts_with distinct
#' @importFrom magrittr %>%
#' @importFrom reshape2 melt dcast
#' @importFrom tune extract_fit_parsnip
#' @importFrom ggplot2 ggplot aes geom_bar geom_path facet_wrap theme_bw theme element_text labs geom_point geom_line geom_histogram geom_abline coord_equal scale_color_manual theme_minimal element_blank ylim position_dodge
#' @importFrom RColorBrewer brewer.pal
#' @importFrom yardstick conf_mat
#' @importFrom pROC roc auc multiclass.roc
#' @importFrom broom tidy glance
#' @importFrom rlang get_expr get_env sym
#' @importFrom viridisLite viridis
#' @importFrom tidyr pivot_wider
#' @importFrom stats coef quantile approx AIC BIC logLik nobs setNames
#' @importFrom utils capture.output head tail
#'
#' @export
summary.fastml <- function(object,
                                 algorithm = "best",
                                 type = c("all", "metrics", "params", "conf_mat"),
                                 sort_metric = NULL,
                                 show_ci = FALSE,
                                 brier_times = NULL,
                                 ...) {

  if (!inherits(object, "fastml")) {
    stop("The input must be a 'fastml' object.")
  }

  # Validate 'type' argument
  type <- match.arg(type, several.ok = TRUE)
  if ("all" %in% type) {
    type <- c("metrics", "params", "conf_mat")
  }

  performance <- object$performance
  predictions_list <- object$predictions
  task <- object$task
  best_model <- object$best_model
  best_model_name <- object$best_model_name
  optimized_metric <- object$metric
  model_count <- length(object$models)
  positive_class <- object$positive_class
  engine_names <- object$engine_names
  brier_time_lookup <- object$survival_brier_times
  resampling_plan <- object$resampling_plan
  resampling_desc <- fastml_describe_resampling(resampling_plan)

  # Extract CV/resampling metrics for display and sorting
  selection_info <- object$selection
  cv_metrics_lookup <- list()
  cv_fold_lookup <- list()
  cv_source <- if (!is.null(selection_info$source)) selection_info$source else "none"

  if (cv_source %in% c("resampling", "nested_cv")) {
    cv_bundle <- fastml_extract_selection_performance(
      resampling_results = object$resampling_results,
      nested_results = object$nested_cv,
      task = task
    )
    if (!is.null(cv_bundle$performance) && length(cv_bundle$performance) > 0) {
      for (alg in names(cv_bundle$performance)) {
        cv_perf <- cv_bundle$performance[[alg]]
        if (is.data.frame(cv_perf) && ".metric" %in% names(cv_perf)) {
          cv_metrics_lookup[[alg]] <- cv_perf
        }
      }
    }

    if (!is.null(object$resampling_results) && length(object$resampling_results) > 0) {
      cv_fold_lookup <- lapply(object$resampling_results, function(entry) {
        if (is.list(entry) && !inherits(entry, "data.frame")) {
          folds_tbl <- entry$folds
          if (!is.null(folds_tbl) && is.data.frame(folds_tbl) && nrow(folds_tbl) > 0) {
            return(fastml_aggregate_resample_metrics(folds_tbl, task))
          }
        } else if (is.data.frame(entry)) {
          return(fastml_aggregate_resample_metrics(entry, task))
        }
        NULL
      })
      cv_fold_lookup <- cv_fold_lookup[!vapply(cv_fold_lookup, is.null, logical(1))]
    }
    # DEBUG: Print CV metrics structure
    if (getOption("fastml.debug", FALSE)) {
      message("DEBUG cv_metrics_lookup keys: ", paste(names(cv_metrics_lookup), collapse = ", "))
      message("DEBUG cv_fold_lookup keys: ", paste(names(cv_fold_lookup), collapse = ", "))
      if (length(cv_fold_lookup) > 0) {
        message("DEBUG cv_fold_lookup first entry columns: ", paste(names(cv_fold_lookup[[1]]), collapse = ", "))
        message("DEBUG cv_fold_lookup first entry:\n")
        print(cv_fold_lookup[[1]])
      } else if (length(cv_metrics_lookup) > 0) {
        message("DEBUG cv_metrics_lookup first entry columns: ", paste(names(cv_metrics_lookup[[1]]), collapse = ", "))
        message("DEBUG cv_metrics_lookup first entry:\n")
        print(cv_metrics_lookup[[1]])
      }
    }
  }

  resolve_engine_name <- function(model_name, default_engine = NA_character_) {
    if (!is.null(engine_names) && !is.null(engine_names[[model_name]])) {
      eng_candidates <- engine_names[[model_name]]
      if (length(eng_candidates) > 0) {
        eng_candidates <- as.character(eng_candidates)
        eng_candidates <- eng_candidates[!is.na(eng_candidates) & eng_candidates != ""]
        if (length(eng_candidates) > 0) {
          return(eng_candidates[[1]])
        }
      }
    }
    if (!is.null(best_model_name) && model_name %in% names(best_model_name)) {
      eng <- best_model_name[[model_name]]
      if (!is.null(eng) && !is.na(eng) && eng != "") {
        return(as.character(eng))
      }
    }
    default_engine
  }

  # Loop over the top-level names (e.g. "rand_forest", "logistic_reg")
  metrics_list <- lapply(names(performance), function(model_name) {

    perf_entry <- performance[[model_name]]

    # Two possible structures:
    # 1) A list of engines, each containing a tibble with .metric/.estimate
    # 2) A single tibble (no engine nesting), typical when only one engine was used

    if (is.list(perf_entry) && !is.data.frame(perf_entry)) {
      # Multi-engine structure
      engine_dfs <- lapply(names(perf_entry), function(engine_name) {
        df <- as.data.frame(perf_entry[[engine_name]])
        df$Engine <- engine_name
        df
      })
      combined_engines <- do.call(rbind, engine_dfs)
    } else {
      # Single tibble structure (no engine-level nesting)
      combined_engines <- as.data.frame(perf_entry)
      combined_engines$Engine <- resolve_engine_name(model_name)
    }

    # Add a "Model" column
    combined_engines$Model <- model_name
    combined_engines
  })

  # Combine all model groups into one data.frame
  performance_df <- do.call(rbind, metrics_list)

  if ("Model" %in% names(performance_df)) {
    performance_df$Model <- as.character(performance_df$Model)
  }
  if ("Engine" %in% names(performance_df)) {
    performance_df$Engine <- as.character(performance_df$Engine)
  }

  all_metric_names <- unique(performance_df$.metric)
  if (is.null(sort_metric)) {
    if (optimized_metric %in% all_metric_names) {
      main_metric <- optimized_metric
    } else {
      main_metric <- all_metric_names[1]
      warning("Optimized metric not available; using first metric.")
    }
  } else {
    if (!(sort_metric %in% all_metric_names)) {
      stop("Invalid sort_metric. Available: ", paste(all_metric_names, collapse = ", "))
    }
    main_metric <- sort_metric
  }

  if (task == "classification") {
    desired_metrics <- c("accuracy", "f_meas", "kap", "precision", "sens", "spec", "roc_auc", "logloss", "brier_score", "ece")
  } else if (task == "regression") {
    desired_metrics <- c("rmse", "rsq", "mae")
  } else if (task == "survival") {
    current_metric_names <- unique(performance_df$.metric)
    base_surv_metrics <- c("c_index", "uno_c", "ibs")
    if ("rmst_diff" %in% current_metric_names) {
      base_surv_metrics <- c(base_surv_metrics, "rmst_diff")
    }
    tau_max <- object$survival_t_max

    compute_brier_helper <- function(perf_df, time_lookup) {
      available <- sort(grep("^brier_t", unique(perf_df$.metric), value = TRUE))
      if (length(available) == 0) {
        return(list(
          available = character(0),
          time_map = stats::setNames(numeric(0), character(0)),
          numeric_map = numeric(0)
        ))
      }
      if (is.null(time_lookup) || length(time_lookup) == 0) {
        time_map <- stats::setNames(rep(NA_real_, length(available)), available)
      } else {
        time_map <- time_lookup[intersect(names(time_lookup), available)]
        missing_names <- setdiff(available, names(time_lookup))
        if (length(missing_names) > 0) {
          time_map <- c(time_map, stats::setNames(rep(NA_real_, length(missing_names)), missing_names))
        }
        time_map <- time_map[match(available, names(time_map))]
        names(time_map) <- available
      }
      numeric_map <- suppressWarnings(as.numeric(time_map))
      names(numeric_map) <- names(time_map)
      list(available = available, time_map = time_map, numeric_map = numeric_map)
    }

    order_brier_metrics <- function(metrics, numeric_map) {
      if (length(metrics) <= 1) {
        return(metrics)
      }
      idx <- match(metrics, names(numeric_map))
      times <- numeric_map[idx]
      ord <- order(ifelse(is.finite(times), times, Inf), metrics)
      metrics[ord]
    }

    match_metric_by_time <- function(target_time, numeric_map, tau_limit) {
      if (length(target_time) != 1 || !is.finite(target_time) || target_time <= 0) {
        return(NA_character_)
      }
      if (is.finite(tau_limit)) {
        target_time <- min(target_time, tau_limit)
      }
      valid_idx <- which(is.finite(numeric_map))
      if (length(valid_idx) == 0) {
        return(NA_character_)
      }
      diffs <- abs(numeric_map[valid_idx] - target_time)
      best_idx <- which.min(diffs)
      idx <- valid_idx[best_idx]
      best_diff <- diffs[best_idx]
      if (length(idx) != 1 || !is.finite(best_diff)) {
        return(NA_character_)
      }
      names(numeric_map)[idx]
    }

    default_brier_selection <- function(available_metrics, numeric_map) {
      if (length(available_metrics) == 0) {
        return(character(0))
      }
      numeric_vals <- numeric_map[is.finite(numeric_map)]
      if (length(numeric_vals) > 0) {
        quantile_targets <- stats::quantile(
          numeric_vals,
          probs = c(0.25, 0.5, 0.75),
          names = FALSE,
          na.rm = TRUE
        )
        quantile_targets <- unique(as.numeric(quantile_targets))
        quantile_targets <- quantile_targets[is.finite(quantile_targets)]
        if (length(quantile_targets) > 0) {
          matched <- vapply(
            quantile_targets,
            function(x) match_metric_by_time(x, numeric_map, tau_max),
            character(1)
          )
          matched <- matched[!is.na(matched) & nzchar(matched)]
          matched <- matched[!duplicated(matched)]
          if (length(matched) > 0) {
            matched <- order_brier_metrics(matched, numeric_map)
            return(matched[seq_len(min(3, length(matched)))])
          }
        }
      }
      fallback <- order_brier_metrics(available_metrics, numeric_map)
      fallback[seq_len(min(3, length(fallback)))]
    }

    make_combo_key <- function(model, engine) {
      engine_val <- ifelse(is.null(engine) || is.na(engine) || engine == "", "<NA>", engine)
      paste(model, engine_val, sep = "||")
    }

    compute_brier_from_curve <- function(curve_df, time_point) {
      if (is.null(curve_df) || nrow(curve_df) == 0) {
        return(NA_real_)
      }
      if (!all(c("eval_time", "brier") %in% names(curve_df))) {
        return(NA_real_)
      }
      times <- as.numeric(curve_df$eval_time)
      values <- as.numeric(curve_df$brier)
      valid <- is.finite(times) & is.finite(values)
      times <- times[valid]
      values <- values[valid]
      if (length(times) == 0) {
        return(NA_real_)
      }
      ord <- order(times)
      times <- times[ord]
      values <- values[ord]
      tol <- max(1e-08, 1e-06 * max(1, abs(time_point)))
      min_time <- min(times)
      max_time <- max(times)
      if (time_point < min_time - tol || time_point > max_time + tol) {
        return(NA_real_)
      }
      clamped_time <- time_point
      if (clamped_time < min_time) {
        clamped_time <- min_time
      }
      if (clamped_time > max_time) {
        clamped_time <- max_time
      }
      approx_res <- stats::approx(times, values, xout = clamped_time, ties = "ordered", rule = 1)
      val <- approx_res$y
      if (!is.finite(val)) {
        idx <- which.min(abs(times - clamped_time))
        if (length(idx) == 0) {
          return(NA_real_)
        }
        val <- values[idx]
      }
      pmax(pmin(val, 1), 0)
    }

    make_metric_name <- function(time_val, existing_names) {
      time_label <- format(time_val, trim = TRUE, scientific = FALSE)
      time_label <- gsub("[^0-9]+", "_", time_label)
      time_label <- gsub("_+", "_", time_label)
      time_label <- gsub("^_+|_+$", "", time_label)
      base <- if (nzchar(time_label)) {
        paste0("brier_t_at_", time_label)
      } else {
        "brier_t_custom"
      }
      candidate <- base
      counter <- 1L
      while (candidate %in% existing_names) {
        candidate <- paste0(base, "_", counter)
        counter <- counter + 1L
      }
      candidate
    }

    augment_performance_with_custom_brier <- function(perf_df, predictions, target_times, time_lookup, tau_limit) {
      if (length(target_times) == 0) {
        return(list(performance_df = perf_df, time_lookup = time_lookup, failed_times = numeric(0)))
      }
      combos <- unique(perf_df[, c("Model", "Engine")])
      combos$Model <- as.character(combos$Model)
      combos$Engine <- as.character(combos$Engine)
      if (nrow(combos) == 0) {
        return(list(performance_df = perf_df, time_lookup = time_lookup, failed_times = target_times))
      }
      brier_curve_lookup <- list()
      if (!is.null(predictions) && length(predictions) > 0) {
        for (model_name in names(predictions)) {
          pred_entry <- predictions[[model_name]]
          if (is.list(pred_entry) && !is.data.frame(pred_entry)) {
            for (engine_name in names(pred_entry)) {
              curve <- attr(pred_entry[[engine_name]], "brier_curve")
              if (!is.null(curve)) {
                brier_curve_lookup[[make_combo_key(model_name, engine_name)]] <- curve
              }
            }
          } else if (is.data.frame(pred_entry)) {
            engine_name <- resolve_engine_name(model_name)
            curve <- attr(pred_entry, "brier_curve")
            if (!is.null(curve)) {
              brier_curve_lookup[[make_combo_key(model_name, engine_name)]] <- curve
            }
          }
        }
      }
      existing_names <- unique(perf_df$.metric)
      failed <- numeric(0)
      new_rows <- list()
      for (time_val in target_times) {
        if (!is.finite(time_val) || time_val <= 0) {
          next
        }
        tol <- max(1e-08, 1e-06 * max(1, abs(time_val)))
        if (is.finite(tau_limit) && time_val > tau_limit + tol) {
          failed <- c(failed, time_val)
          next
        }
        values <- vapply(seq_len(nrow(combos)), function(idx) {
          key <- make_combo_key(combos$Model[idx], combos$Engine[idx])
          curve <- brier_curve_lookup[[key]]
          compute_brier_from_curve(curve, time_val)
        }, numeric(1))
        if (!any(is.finite(values))) {
          failed <- c(failed, time_val)
          next
        }
        metric_name <- make_metric_name(time_val, existing_names)
        existing_names <- c(existing_names, metric_name)
        row_count <- nrow(combos)
        new_df <- data.frame(
          .metric = rep(metric_name, row_count),
          .estimate = values,
          stringsAsFactors = FALSE
        )
        if (".lower" %in% names(perf_df)) new_df$.lower <- NA_real_
        if (".upper" %in% names(perf_df)) new_df$.upper <- NA_real_
        if (".n_boot" %in% names(perf_df)) new_df$.n_boot <- 0
        other_cols <- setdiff(colnames(perf_df), names(new_df))
        for (col_nm in other_cols) {
          if (col_nm == "Engine") {
            new_df[[col_nm]] <- combos$Engine
          } else if (col_nm == "Model") {
            new_df[[col_nm]] <- combos$Model
          } else {
            new_df[[col_nm]] <- NA
          }
        }
        new_df <- new_df[, colnames(perf_df), drop = FALSE]
        new_rows[[length(new_rows) + 1]] <- new_df
        time_lookup <- c(time_lookup, stats::setNames(time_val, metric_name))
      }
      if (length(new_rows) > 0) {
        perf_df <- rbind(perf_df, do.call(rbind, new_rows))
      }
      list(performance_df = perf_df, time_lookup = time_lookup, failed_times = failed)
    }

    helper <- compute_brier_helper(performance_df, brier_time_lookup)
    available_brier_metrics <- helper$available
    available_time_map <- helper$time_map
    numeric_time_map <- helper$numeric_map
    selected_brier_metrics <- character(0)
    matched_metrics <- character(0)
    unmatched_numeric <- numeric(0)
    failure_times <- numeric(0)

    if (!is.null(brier_times) && length(brier_times) > 0) {
      input_list <- as.list(brier_times)
      for (bt in input_list) {
        if (is.character(bt) && length(bt) == 1 && bt %in% available_brier_metrics) {
          matched_metrics <- c(matched_metrics, bt)
        } else {
          bt_numeric <- suppressWarnings(as.numeric(bt))
          if (length(bt_numeric) == 1 && is.finite(bt_numeric) && bt_numeric > 0) {
            metric_name <- match_metric_by_time(bt_numeric, numeric_time_map, tau_max)
            if (!is.na(metric_name)) {
              matched_metrics <- c(matched_metrics, metric_name)
            } else {
              unmatched_numeric <- c(unmatched_numeric, bt_numeric)
            }
          }
        }
      }
      unmatched_numeric <- unique(unmatched_numeric[is.finite(unmatched_numeric) & unmatched_numeric > 0])
      matched_metrics <- matched_metrics[matched_metrics %in% available_brier_metrics]
      matched_metrics <- matched_metrics[!duplicated(matched_metrics)]

      if (length(unmatched_numeric) > 0) {
        augment_result <- augment_performance_with_custom_brier(
          performance_df,
          predictions_list,
          unmatched_numeric,
          brier_time_lookup,
          tau_max
        )
        performance_df <- augment_result$performance_df
        brier_time_lookup <- augment_result$time_lookup
        failure_times <- c(failure_times, augment_result$failed_times)

        helper <- compute_brier_helper(performance_df, brier_time_lookup)
        available_brier_metrics <- helper$available
        available_time_map <- helper$time_map
        numeric_time_map <- helper$numeric_map

        to_retry <- setdiff(unmatched_numeric, failure_times)
        if (length(to_retry) > 0) {
          for (bt_numeric in to_retry) {
            metric_name <- match_metric_by_time(bt_numeric, numeric_time_map, tau_max)
            if (!is.na(metric_name)) {
              matched_metrics <- c(matched_metrics, metric_name)
            } else {
              failure_times <- c(failure_times, bt_numeric)
            }
          }
        }
        matched_metrics <- matched_metrics[matched_metrics %in% available_brier_metrics]
        matched_metrics <- matched_metrics[!duplicated(matched_metrics)]
      }

      if (length(matched_metrics) == 0) {
        defaults <- default_brier_selection(available_brier_metrics, numeric_time_map)
        if (length(defaults) > 0) {
          message(
            "None of the requested brier_times overlapped with the available follow-up; ",
            "using default horizons based on observed quartiles."
          )
        }
        selected_brier_metrics <- defaults
      } else {
        selected_brier_metrics <- order_brier_metrics(matched_metrics, numeric_time_map)
      }

      failure_times <- sort(unique(failure_times[is.finite(failure_times)]))
      if (length(failure_times) > 0) {
        warning(
          sprintf(
            "Some requested Brier time points were not available and were omitted: %s",
            paste(format(failure_times, trim = TRUE, digits = 6), collapse = ", ")
          )
        )
      }
    } else {
      selected_brier_metrics <- default_brier_selection(available_brier_metrics, numeric_time_map)
    }

    all_metric_names <- unique(performance_df$.metric)
    desired_metrics <- unique(c(base_surv_metrics, selected_brier_metrics))
  } else {
    # Fallback for any other task types
    desired_metrics <- unique(performance_df$.metric)
  }
  desired_metrics <- intersect(desired_metrics, all_metric_names)
  if (length(desired_metrics) == 0) desired_metrics <- main_metric

  metrics_for_sub <- unique(c(desired_metrics, main_metric))
  metrics_for_sub <- intersect(metrics_for_sub, all_metric_names)

  performance_sub <- performance_df[performance_df$.metric %in% metrics_for_sub, ] %>%
    dplyr::select(-dplyr::any_of(".estimator"))

  has_ci_cols <- all(c(".lower", ".upper") %in% colnames(performance_sub))
  show_ci_available <- isTRUE(show_ci) && has_ci_cols

  if (show_ci_available) {
    performance_sub <- performance_sub %>%
      dplyr::mutate(
        metric_display = dplyr::case_when(
          is.na(.data$.estimate) ~ NA_character_,
          !is.na(.data$.lower) & !is.na(.data$.upper) ~
            sprintf("%.3f (%.3f, %.3f)", .data$.estimate, .data$.lower, .data$.upper),
          TRUE ~ sprintf("%.3f", .data$.estimate)
        )
      )
  } else {
    performance_sub <- performance_sub %>%
      dplyr::mutate(
        metric_display = dplyr::case_when(
          is.na(.data$.estimate) ~ NA_character_,
          TRUE ~ sprintf("%.3f", .data$.estimate)
        )
      )
  }
  metrics_for_numeric <- metrics_for_sub
  keep_metrics <- metrics_for_numeric
  if (length(engine_names) == 1 && "LiblineaR" %in% engine_names) {
    allowed_liblinear <- c("accuracy", "kap", "sens", "spec", "precision", "f_meas")
    keep_metrics <- intersect(keep_metrics, allowed_liblinear)
    if (!(main_metric %in% keep_metrics) && main_metric %in% metrics_for_numeric) {
      keep_metrics <- c(keep_metrics, main_metric)
    }
  }
  keep_metrics <- unique(keep_metrics)
  performance_numeric <- performance_sub %>%
    dplyr::select(dplyr::all_of(c("Model", "Engine", ".metric", ".estimate"))) %>%
    dplyr::distinct()

  performance_wide <- tidyr::pivot_wider(
    performance_numeric,
    names_from = ".metric",
    values_from = ".estimate"
  )

  performance_display <- performance_sub %>%
    dplyr::select(dplyr::all_of(c("Model", "Engine", ".metric", "metric_display"))) %>%
    dplyr::distinct() %>%
    tidyr::pivot_wider(
      names_from = ".metric",
      values_from = "metric_display"
    )
  select_cols_numeric <- c("Model", "Engine", keep_metrics)
  select_cols_numeric <- intersect(select_cols_numeric, colnames(performance_wide))
  performance_wide <- dplyr::select(performance_wide, dplyr::all_of(select_cols_numeric))
  select_cols_display <- c("Model", "Engine", desired_metrics)
  select_cols_display <- intersect(select_cols_display, colnames(performance_display))
  performance_display <- dplyr::select(performance_display, dplyr::all_of(select_cols_display))
  desired_metrics <- intersect(desired_metrics, select_cols_display[!(select_cols_display %in% c("Model", "Engine"))])



  # performance_wide$Engine <- engine_names[match(performance_wide$Model, names(engine_names))]
  # performance_wide <- performance_wide[, c("Model", "Engine", setdiff(colnames(performance_wide), c("Model", "Engine")))]

  # Add CV metrics for sorting and display when resampling was used
  cv_metric_col <- paste0("cv_", main_metric)
  cv_se_col <- paste0("cv_", main_metric, "_se")
  has_cv_metrics <- length(cv_metrics_lookup) > 0 || length(cv_fold_lookup) > 0

  if (has_cv_metrics) {
    # Helper function to find CV data with flexible key matching
    find_cv_data <- function(m, e, fold_lookup, metrics_lookup) {
      combined_key <- paste0(m, " (", e, ")")
      # Try multiple key formats in cv_fold_lookup first (preferred source)
      cv_df <- fold_lookup[[combined_key]]
      if (is.null(cv_df)) cv_df <- fold_lookup[[m]]
      # Try partial matching if exact match fails
      if (is.null(cv_df) && length(fold_lookup) > 0) {
        matching_keys <- grep(paste0("^", m), names(fold_lookup), value = TRUE)
        if (length(matching_keys) > 0) cv_df <- fold_lookup[[matching_keys[1]]]
      }
      # Fall back to cv_metrics_lookup only if fold_lookup has nothing
      if (is.null(cv_df)) cv_df <- metrics_lookup[[combined_key]]
      if (is.null(cv_df)) cv_df <- metrics_lookup[[m]]
      if (is.null(cv_df) && length(metrics_lookup) > 0) {
        matching_keys <- grep(paste0("^", m), names(metrics_lookup), value = TRUE)
        if (length(matching_keys) > 0) cv_df <- metrics_lookup[[matching_keys[1]]]
      }
      cv_df
    }

    # Extract CV metric values for each model
    performance_wide[[cv_metric_col]] <- mapply(function(m, e) {
      cv_df <- find_cv_data(m, e, cv_fold_lookup, cv_metrics_lookup)
      if (is.null(cv_df)) return(NA_real_)
      row <- cv_df[cv_df$.metric == main_metric, , drop = FALSE]
      if (nrow(row) == 0) return(NA_real_)
      as.numeric(row$.estimate[1])
    }, performance_wide$Model, performance_wide$Engine)

    performance_wide[[cv_se_col]] <- mapply(function(m, e) {
      cv_df <- find_cv_data(m, e, cv_fold_lookup, cv_metrics_lookup)
      if (is.null(cv_df)) return(NA_real_)
      row <- cv_df[cv_df$.metric == main_metric, , drop = FALSE]
      if (nrow(row) == 0) return(NA_real_)
      # Try std_dev first (SD), then std_err (SE) as fallback
      if ("std_dev" %in% names(row) && !is.na(row$std_dev[1])) {
        return(as.numeric(row$std_dev[1]))
      } else if ("std_err" %in% names(row) && !is.na(row$std_err[1])) {
        return(as.numeric(row$std_err[1]))
      }
      NA_real_
    }, performance_wide$Model, performance_wide$Engine)
    # Create display column with mean +/- SE
    performance_display[[cv_metric_col]] <- sapply(seq_len(nrow(performance_wide)), function(i) {
      val <- performance_wide[[cv_metric_col]][i]
      se <- performance_wide[[cv_se_col]][i]
      if (is.na(val)) return(NA_character_)
      if (is.na(se) || !is.finite(se)) return(sprintf("%.3f", val))
      sprintf("%.3f \u00b1 %.3f", val, se)
    })
  }

  # Sort direction: lower-better metrics vs higher-better metrics
  brier_cols <- grep("^brier_t", colnames(performance_wide), value = TRUE)
  ascending_metrics <- unique(c("rmse", "mae", "ibs", "logloss", "brier_score", "ece", "mse", brier_cols))
  lower_is_better <- main_metric %in% ascending_metrics

  # Sort by CV metric when available, otherwise by holdout metric
 sort_col <- if (has_cv_metrics && cv_metric_col %in% colnames(performance_wide)) {
    cv_metric_col
  } else {
    main_metric
  }

  if (sort_col %in% colnames(performance_wide)) {
    if (lower_is_better) {
      order_idx <- order(performance_wide[[sort_col]], na.last = TRUE)
    } else {
      order_idx <- order(-performance_wide[[sort_col]], na.last = TRUE)
    }
  } else {
    warning(sprintf("Sort metric '%s' not available; retaining original order.", sort_col))
    order_idx <- seq_len(nrow(performance_wide))
  }
  performance_wide <- performance_wide[order_idx, , drop = FALSE]
  performance_display <- performance_display[order_idx, , drop = FALSE]

  performance_lookup <- performance_wide
  if (all(algorithm != "best")) {
    performance_lookup <- performance_lookup %>% dplyr::filter(Model %in% algorithm)
  }

  display_names <- c(
    accuracy = "Accuracy",
    f_meas = "F1 Score",
    kap = "Kappa",
    precision = "Precision",
    roc_auc = "ROC AUC",
    sens = "Sensitivity",
    spec = "Specificity",
    logloss = "Logloss",
    brier_score = "Brier Score",
    ece = "ECE",
    rsq = "R-squared",
    mae = "MAE",
    rmse = "RMSE",
    c_index = "Harrell C-index",
    uno_c = "Uno's C-index",
    ibs = "Integrated Brier Score",
    rmst_diff = "RMST diff (t_max)"
  )
  t_max_val <- object$survival_t_max
  if (!is.null(t_max_val) && length(t_max_val) == 1 && is.finite(t_max_val) && t_max_val > 0) {
    display_names[["rmst_diff"]] <- sprintf(
      "RMST diff (t<=%s)",
      format(t_max_val, trim = TRUE, digits = 4)
    )
  }

  if (!is.null(brier_time_lookup) && length(brier_time_lookup) > 0) {
    for (nm in names(brier_time_lookup)) {
      if (!is.null(brier_time_lookup[[nm]]) && is.finite(brier_time_lookup[[nm]])) {
        time_label <- format(brier_time_lookup[[nm]], trim = TRUE, digits = 4)
      } else {
        time_label <- nm
      }
      display_names[[nm]] <- sprintf("Brier(t=%s)", time_label)
    }
  }
  auto_brier_cols <- setdiff(grep("^brier_t", colnames(performance_display), value = TRUE), names(display_names))
  if (length(auto_brier_cols) > 0) {
    display_names[auto_brier_cols] <- auto_brier_cols
  }



  # Filter rows where the Model is in best_model_name and its Engine equals best_model_name[Model]
  best_val_df <- performance_wide %>%
    filter(Model %in% names(best_model_name)) %>%
    filter({
      if (length(Model) == 0) {
        rep(FALSE, length(Model))
      } else {
        mapply(function(m, e) {
          bm <- best_model_name[[m]]
          if (is.null(bm) || length(bm) == 0 || is.na(bm)) {
            return(FALSE)
          }
          e == bm
        }, Model, Engine)
      }
    })



  best_lookup <- if (!is.null(best_model_name)) {
    stats::setNames(as.character(best_model_name), names(best_model_name))
  } else {
    character(0)
  }
  expected_engine <- best_lookup[performance_wide$Model]
  best_model_idx <- which(!is.na(expected_engine) & performance_wide$Engine == expected_engine)


  if(length(algorithm) == 1 && algorithm == "best"){
    selected_model_idx <- best_model_idx
    desired_models <- object$best_model
  }else{

    clean_names <- sub(" \\(.*\\)", "", names(object$models))

    if(all(algorithm %in% clean_names)){

      selected_model_idx <- which(performance_wide$Model %in% algorithm)

      clean_model_names <- sub(" \\(.*\\)", "", names(object$models))
      matching_index <- match(algorithm, clean_model_names)

      if (all(!is.na(matching_index))) {
        desired_models <- object$models[matching_index]
      } else {
        desired_models <- NULL  # Handle case where algorithm is not found
      }

    }else{

      stop("Not all specified algorithms entered correctly.")
    }
  }

  if ("metrics" %in% type) {

  cat("\n===== fastml Model Summary =====\n")
  cat("Task:", task, "\n")
  cat("Number of Models Trained:", model_count, "\n\n")

  # Compute safe column widths helper

  safe_nchar <- function(x) {
    x_chr <- as.character(x)
    x_chr[is.na(x_chr)] <- ""
    nchar(x_chr)
  }

  # Helper to print a table
  print_table <- function(header, data_df, col_keys) {
    col_widths <- sapply(seq_along(header), function(i) {
      col_name <- header[i]
      col_key <- col_keys[i]
      col_data <- data_df[[col_key]]
      name_w <- safe_nchar(col_name)
      data_w <- if (length(col_data)) max(safe_nchar(col_data), na.rm = TRUE) else 0
      max(name_w, data_w, na.rm = TRUE)
    })
    header_line <- paste(mapply(function(h, w) format(h, width = w, justify = "left"), header, col_widths), collapse = "  ")
    line_sep <- paste(rep("-", sum(col_widths) + 2 * (length(col_widths) - 1)), collapse = "")
    cat(line_sep, "\n")
    cat(header_line, "\n")
    cat(line_sep, "\n")
    for (i in seq_len(nrow(data_df))) {
      row_line <- paste(mapply(function(v, w) format(v, width = w, justify = "left"),
                               data_df[i, col_keys, drop = FALSE], col_widths), collapse = "  ")
      cat(row_line, "\n")
    }
    cat(line_sep, "\n")
  }

  metrics_to_print <- c("Model", "Engine", desired_metrics)

  if (all(algorithm != "best")) {
    performance_wide <- performance_wide %>% dplyr::filter(Model %in% algorithm)
    performance_display <- performance_display %>% dplyr::filter(Model %in% algorithm)
    performance_lookup <- performance_wide
  }

  # ============================================================
  # TABLE 1: Model Selection (CV Performance) - when resampling was used
  # ============================================================
  if (has_cv_metrics && cv_metric_col %in% colnames(performance_wide)) {
    cat("-- Table 1: Model Selection (Cross-Validation) --\n")
    cat("Note: This table determines the best model.\n\n")

    # Build CV selection table
    cv_select_df <- data.frame(
      Model = performance_wide$Model,
      Engine = performance_wide$Engine,
      stringsAsFactors = FALSE
    )

    # Get display name for the main metric
    metric_display <- if (main_metric %in% names(display_names)) display_names[[main_metric]] else main_metric

    # Add CV mean column
    cv_select_df[[paste0(metric_display, " (CV mean)")]] <- sapply(performance_wide[[cv_metric_col]], function(v) {
      if (is.na(v)) "NA" else sprintf("%.4f", v)
    })

    # Add CV SD column if available
    if (cv_se_col %in% colnames(performance_wide)) {
      cv_select_df[[paste0(metric_display, " (CV SD)")]] <- sapply(performance_wide[[cv_se_col]], function(v) {
        if (is.na(v)) "NA" else sprintf("%.4f", v)
      })
    }

    # Mark selected model with dagger
    cv_select_df$Model <- as.character(cv_select_df$Model)
    if (all(algorithm == "best")) {
      cv_select_df$Model[best_model_idx] <- paste0(cv_select_df$Model[best_model_idx], "\u2020")
    }

    cv_header <- colnames(cv_select_df)
    print_table(cv_header, cv_select_df, cv_header)

    if (all(algorithm == "best")) {
      cat("\u2020 Selected based on mean ", metric_display, " across CV folds\n\n", sep = "")
    } else {
      cat("\n")
    }
  }

  # ============================================================
  # TABLE 2: Final Evaluation (Test Set Performance)
  # ============================================================
  if (has_cv_metrics) {
    cat("-- Table 2: Final Evaluation (Test Set) --\n")
    cat("Note: For reporting only; selection was based on CV above.\n\n")
  } else {
    # No CV - this is the only table, and it determines selection
    cat("-- Performance Metrics (Holdout) --\n\n")
  }

  # Build holdout table header (no CV columns)
  header <- c("Model", "Engine", sapply(desired_metrics, function(m) {
    if (m %in% names(display_names)) display_names[[m]] else m
  }))

  data_str <- performance_display
  data_str$Model <- as.character(data_str$Model)

  # Only mark with dagger if NO CV was used (i.e., holdout determines selection)
  if (!has_cv_metrics && all(algorithm == "best")) {
    data_str$Model[best_model_idx] <- paste0(data_str$Model[best_model_idx], "\u2020")
  }

  col_keys <- c("Model", "Engine", desired_metrics)
  print_table(header, data_str, col_keys)

  if (!has_cv_metrics && all(algorithm == "best")) {
    cat("\u2020 Selected based on ", main_metric, "\n\n", sep = "")
  } else {
    cat("\n")
  }

}

  if ("params" %in% type) {

    if(length(algorithm) == 1 && all(algorithm == "best")){
      cat("Best Model hyperparameters:\n\n")

    }else{

      cat("Selected Model hyperparameters:\n\n")

    }

    format_numeric_vec <- function(x, digits = 4) {
      if (is.null(x) || length(x) == 0) {
        return(character())
      }
      x <- suppressWarnings(as.numeric(x))
      if (length(x) == 0) {
        return(character())
      }
      out <- vapply(x, function(val) {
        if (is.na(val) || !is.finite(val)) {
          "<NA>"
        } else {
          format(signif(val, digits), trim = TRUE, scientific = FALSE)
        }
      }, character(1))
      out
    }

    format_numeric_single <- function(x, digits = 4) {
      vec <- format_numeric_vec(x, digits = digits)
      if (length(vec) == 0) {
        "<NA>"
      } else {
        vec[[1]]
      }
    }

    format_param_value <- function(val) {
      if (rlang::is_quosure(val)) {
        return(rlang::quo_text(val))
      }
      if (is.numeric(val)) {
        return(paste(format_numeric_vec(val), collapse = ", "))
      }
      if (is.logical(val)) {
        return(paste(ifelse(val, "TRUE", "FALSE"), collapse = ", "))
      }
      if (is.character(val)) {
        return(paste(val, collapse = ", "))
      }
      if (is.factor(val)) {
        return(paste(as.character(val), collapse = ", "))
      }
      if (length(val) == 1) {
        return(as.character(val))
      }
      paste(as.character(val), collapse = ", ")
    }

    parse_model_label <- function(label) {
      if (is.null(label)) {
        return(list(algorithm = NA_character_, engine = NA_character_))
      }
      algo <- trimws(sub("\\s*\\(.*$", "", label))
      engine <- if (grepl("\\(", label, fixed = FALSE)) {
        sub("^.*\\(([^()]*)\\)\\s*$", "\\1", label)
      } else {
        NA_character_
      }
      list(algorithm = algo, engine = trimws(engine))
    }

    extract_survival_fit <- function(label, model_obj) {
      candidates <- list(object$models[[label]], model_obj)
      for (candidate in candidates) {
        if (is.null(candidate)) {
          next
        }
        fit_candidate <- tryCatch(candidate$fit$fit$fit, error = function(e) NULL)
        if (inherits(fit_candidate, c("survreg", "coxph", "stpm2", "pstpm2", "flexsurvreg", "fastml_xgb_survival"))) {
          return(fit_candidate)
        }
        fit_candidate <- tryCatch(candidate$fit$fit, error = function(e) NULL)
        if (inherits(fit_candidate, c("survreg", "coxph", "stpm2", "pstpm2", "flexsurvreg", "fastml_xgb_survival"))) {
          return(fit_candidate)
        }
        fit_candidate <- tryCatch(candidate$fit, error = function(e) NULL)
        if (inherits(fit_candidate, c("survreg", "coxph", "stpm2", "pstpm2", "flexsurvreg", "fastml_xgb_survival"))) {
          return(fit_candidate)
        }
      }
      NULL
    }

    fastml_survival_tidy_df <- function(fit_obj) {
      tidy_df <- tryCatch(broom::tidy(fit_obj), error = function(e) NULL)
      if (!is.null(tidy_df)) {
        return(as.data.frame(tidy_df, stringsAsFactors = FALSE))
      }
      summary_fit <- tryCatch(suppressWarnings(summary(fit_obj)), error = function(e) NULL)
      if (is.null(summary_fit)) {
        return(NULL)
      }
      candidate_names <- c("coefficients", "coef", "coef.table", "coef3")
      for (nm in candidate_names) {
        comp <- tryCatch(summary_fit[[nm]], error = function(e) NULL)
        if (is.null(comp)) {
          next
        }
        if (is.matrix(comp)) {
          df <- as.data.frame(comp, stringsAsFactors = FALSE)
        } else if (is.data.frame(comp)) {
          df <- comp
        } else {
          next
        }
        if (nrow(df) == 0) {
          next
        }
        if (!"term" %in% names(df)) {
          df <- rownames(df)
        }
        if (!"term" %in% names(df)) {
          df <- seq_len(nrow(df))
        }
        df <- df[, c("term", setdiff(names(df), "term")), drop = FALSE]
        return(df)
      }
      NULL
    }

    fastml_survival_glance_df <- function(fit_obj) {
      glance_df <- tryCatch(broom::glance(fit_obj), error = function(e) NULL)
      if (!is.null(glance_df)) {
        return(as.data.frame(glance_df, stringsAsFactors = FALSE))
      }
      safe_stat <- function(expr) {
        val <- tryCatch(expr, error = function(e) NA_real_)
        if (length(val) == 0) {
          return(NA_real_)
        }
        as.numeric(val[1])
      }
      data.frame(
        logLik = safe_stat(stats::logLik(fit_obj)),
        AIC = safe_stat(stats::AIC(fit_obj)),
        BIC = safe_stat(stats::BIC(fit_obj)),
        nobs = safe_stat(stats::nobs(fit_obj)),
        stringsAsFactors = FALSE
      )
    }

    fastml_print_survival_summary <- function(fit_obj, heading = NULL) {
      if (!is.null(heading)) {
        cat("  ", heading, "\n", sep = "")
      }
      model_formula <- tryCatch(stats::formula(fit_obj), error = function(e) NULL)
      if (!is.null(model_formula)) {
        cat("  Formula: ", paste(deparse(model_formula), collapse = " "), "\n", sep = "")
      }
      glance_df <- fastml_survival_glance_df(fit_obj)
      if (!is.null(glance_df)) {
        cat("  Fit statistics:\n")
        print(glance_df, row.names = FALSE)
      }
      tidy_df <- fastml_survival_tidy_df(fit_obj)
      if (!is.null(tidy_df)) {
        cat("  Coefficients:\n")
        print(tidy_df, row.names = FALSE)
      }
      TRUE
    }

    print_stpm2_details <- function(fit_obj, model_info = NULL) {
      if (is.function(fit_obj) && !is.null(model_info)) {
        alt_fit <- tryCatch(model_info, error = function(e) NULL)
        if (inherits(alt_fit, c("stpm2", "pstpm2"))) {
          fit_obj <- alt_fit
        }
      }
      if (!inherits(fit_obj, c("stpm2", "pstpm2"))) {
        return(FALSE)
      }
      fastml_print_survival_summary(fit_obj, heading = "Royston-Parmar (rstpm2)")
    }
    print_survreg_details <- function(fit_obj) {
      coef_vec <- tryCatch(stats::coef(fit_obj), error = function(e) NULL)
      if (!is.null(coef_vec) && length(coef_vec) > 0) {
        cat("  Coefficients (coef):\n")
        coef_fmt <- format_numeric_vec(coef_vec)
        if (!is.null(names(coef_vec))) {
          names(coef_fmt) <- names(coef_vec)
        }
        coef_mat <- matrix(coef_fmt, ncol = 1)
        rownames(coef_mat) <- if (!is.null(names(coef_fmt))) names(coef_fmt) else rownames(coef_mat)
        colnames(coef_mat) <- "coef"
        print(coef_mat, quote = FALSE)
      } else {
        cat("  Coefficients (coef): <unavailable>\n")
      }

      scale_val <- tryCatch(fit_obj$scale, error = function(e) NA_real_)
      cat("  Scale:", format_numeric_single(scale_val), "\n")

      dist_val <- tryCatch(fit_obj$dist, error = function(e) NA_character_)
      dist_str <- if (!is.null(dist_val) && length(dist_val) > 0 && nzchar(as.character(dist_val)[1])) {
        as.character(dist_val)[1]
      } else {
        "<NA>"
      }
      cat("  Distribution:", dist_str, "\n")

      loglik_val <- tryCatch(fit_obj$loglik, error = function(e) NULL)
      loglik_last <- if (!is.null(loglik_val) && length(loglik_val) > 0) {
        loglik_val[length(loglik_val)]
      } else {
        NA_real_
      }
      cat("  Log-likelihood:", format_numeric_single(loglik_last), "\n")
      TRUE
    }

    print_flexsurv_details <- function(fit_obj, model_info = NULL) {
      dist_val <- tryCatch(fit_obj$dist, error = function(e) NULL)
      dist_label <- NULL
      if (!is.null(model_info)) {
        dist_label <- tryCatch(model_info$distribution_label, error = function(e) NULL)
      }
      if (is.null(dist_label) || !nzchar(as.character(dist_label)[1])) {
        if (is.character(dist_val) && length(dist_val) > 0) {
          dist_label <- dist_val[1]
        } else if (is.list(dist_val) && !is.null(dist_val$dist)) {
          dist_label <- dist_val$dist
        }
      }
      if (is.null(dist_label) || !nzchar(as.character(dist_label)[1])) {
        dist_label <- "<unknown>"
      }
      cat("  Distribution:", dist_label, "\n")

      coef_vec <- tryCatch({
        stats::coef(fit_obj)
      }, error = function(e) {
        tryCatch(fit_obj$coefficients, error = function(e2) NULL)
      })
      if (!is.null(coef_vec) && length(coef_vec) > 0) {
        cat("  Coefficients (link scale):\n")
        coef_fmt <- format_numeric_vec(coef_vec)
        coef_mat <- matrix(coef_fmt, ncol = 1)
        rownames(coef_mat) <- if (!is.null(names(coef_vec))) names(coef_vec) else rownames(coef_mat)
        colnames(coef_mat) <- "coef"
        print(coef_mat, quote = FALSE)
      }

      res_mat <- tryCatch(as.matrix(fit_obj$res), error = function(e) NULL)
      if (is.matrix(res_mat) && nrow(res_mat) > 0) {
        cat("  Parameter estimates:\n")
        col_names <- colnames(res_mat)
        fmt_mat <- apply(res_mat, 2, format_numeric_vec)
        if (is.vector(fmt_mat)) {
          fmt_mat <- matrix(fmt_mat, ncol = 1)
          colnames(fmt_mat) <- col_names[1]
        }
        rownames(fmt_mat) <- rownames(res_mat)
        print(fmt_mat, quote = FALSE)
      }

      loglik_val <- tryCatch({
        ll <- fit_obj$loglik
        if (length(ll) > 0) ll[length(ll)] else NA_real_
      }, error = function(e) NA_real_)
      if (is.finite(loglik_val)) {
        cat("  Log-likelihood:", format_numeric_single(loglik_val), "\n")
      }

      aic_val <- tryCatch(as.numeric(fit_obj$AIC), error = function(e) NA_real_)
      if (is.finite(aic_val)) {
        cat("  AIC:", format_numeric_single(aic_val), "\n")
      }

      bic_val <- tryCatch(as.numeric(fit_obj$BIC), error = function(e) NA_real_)
      if (is.finite(bic_val)) {
        cat("  BIC:", format_numeric_single(bic_val), "\n")
      }

      train_times <- NULL
      train_status <- NULL
      train_size <- NULL
      if (!is.null(model_info)) {
        train_times <- tryCatch(model_info$train_times, error = function(e) NULL)
        train_status <- tryCatch(model_info$train_status, error = function(e) NULL)
        train_size <- tryCatch(model_info$train_size, error = function(e) NULL)
      }
      n_obs <- tryCatch(as.numeric(fit_obj$N), error = function(e) NA_real_)
      events <- NA_real_
      censored <- NA_real_
      if (is.null(train_status) && is.numeric(train_size) && length(train_size) == 1 && is.finite(train_size)) {
        n_obs <- train_size
      }
      if (!is.null(train_status)) {
        status_norm <- tryCatch({
          fastml_normalize_survival_status(train_status, reference_length = length(train_status))$status
        }, error = function(e) NULL)
        if (!is.null(status_norm)) {
          n_obs <- length(status_norm)
          events <- sum(status_norm == 1, na.rm = TRUE)
          censored <- sum(status_norm == 0, na.rm = TRUE)
        }
      } else if (!is.null(train_times)) {
        times_num <- as.numeric(train_times)
        times_num <- times_num[is.finite(times_num)]
        if (length(times_num) > 0) {
          n_obs <- length(times_num)
        }
      }
      if (is.finite(n_obs)) {
        msg <- paste0("  Sample size: ", format_numeric_single(n_obs, digits = 0))
        details <- character(0)
        if (is.finite(events)) {
          details <- c(details, paste0("events = ", format_numeric_single(events, digits = 0)))
        }
        if (is.finite(censored)) {
          details <- c(details, paste0("censored = ", format_numeric_single(censored, digits = 0)))
        }
        if (length(details) > 0) {
          msg <- paste0(msg, " (", paste(details, collapse = ", "), ")")
        }
        cat(msg, "\n")
      }

      breaks_info <- NULL
      if (!is.null(model_info)) {
        breaks_info <- tryCatch(model_info$breaks, error = function(e) NULL)
      }
      if (!is.null(breaks_info) && length(breaks_info) > 0) {
        cat("  Piecewise breaks:", paste(format_numeric_vec(breaks_info), collapse = ", "), "\n")
      }

      TRUE
    }

    print_coxph_details <- function(fit_obj, model_info = NULL, performance_row = NULL) {
      summary_fit <- tryCatch(summary(fit_obj), error = function(e) NULL)
      if (is.null(summary_fit)) {
        return(FALSE)
      }

      remove_terms <- character()
      strata_details <- list()
      if (!is.null(model_info)) {
        strata_cols <- tryCatch(model_info$strata_cols, error = function(e) NULL)
        strata_dummy <- tryCatch(model_info$strata_dummy_cols, error = function(e) NULL)
        strata_base <- tryCatch(model_info$strata_base_cols, error = function(e) NULL)
        remove_terms <- unique(c(
          remove_terms,
          if (!is.null(strata_cols)) strata_cols else character(),
          if (!is.null(strata_dummy)) strata_dummy else character(),
          if (!is.null(strata_base)) strata_base else character()
        ))
        info_list <- tryCatch(model_info$strata_info, error = function(e) NULL)
        if (is.list(info_list) && length(info_list) > 0) {
          for (nm in names(info_list)) {
            entry <- info_list[[nm]]
            if (is.null(entry)) {
              next
            }
            col_nm <- if (!is.null(entry$column)) entry$column else nm
            display_nm <- entry$display
            if (is.null(display_nm) || length(display_nm) == 0 || !nzchar(as.character(display_nm)[1])) {
              display_nm <- col_nm
            }
            level_vals <- entry$levels
            if (is.null(level_vals)) {
              level_vals <- character()
            }
            level_vals <- as.character(level_vals)
            strata_details[[length(strata_details) + 1]] <- list(
              column = col_nm,
              display = display_nm,
              levels = level_vals
            )
          }
        }
      }
      remove_terms <- remove_terms[nzchar(remove_terms)]
      remove_terms <- unique(remove_terms)

      filter_matrix <- function(mat, terms_to_remove) {
        if (!is.null(mat) && is.matrix(mat) && nrow(mat) > 0) {
          rn <- rownames(mat)
          if (!is.null(rn) && length(terms_to_remove) > 0) {
            drop_idx <- rep(FALSE, length(rn))
            for (term in terms_to_remove) {
              drop_idx <- drop_idx |
                rn == term |
                startsWith(rn, paste0(term, ":")) |
                startsWith(rn, paste0(term, "=")) |
                rn == paste0("strata(", term, ")")
            }
            mat <- mat[!drop_idx, , drop = FALSE]
          }
        }
        mat
      }

      coef_mat_raw <- summary_fit$coefficients
      coef_mat <- filter_matrix(coef_mat_raw, remove_terms)
      coef_raw_is_matrix <- is.matrix(coef_mat_raw)
      removed_all_coef <- coef_raw_is_matrix && nrow(coef_mat_raw) > 0 &&
        (!is.matrix(coef_mat) || nrow(coef_mat) == 0)
      no_coef_rows <- coef_raw_is_matrix && nrow(coef_mat_raw) == 0

      if (is.matrix(coef_mat) && nrow(coef_mat) > 0 && "coef" %in% colnames(coef_mat)) {
        cat("  Coefficients (coef):\n")
        coef_fmt <- format_numeric_vec(coef_mat[, "coef"])
        if (!is.null(rownames(coef_mat))) {
          names(coef_fmt) <- rownames(coef_mat)
        }
        coef_out <- matrix(coef_fmt, ncol = 1)
        rownames(coef_out) <- if (!is.null(names(coef_fmt))) names(coef_fmt) else rownames(coef_out)
        colnames(coef_out) <- "coef"
        print(coef_out, quote = FALSE)
      } else if (removed_all_coef || no_coef_rows) {
        cat("  Coefficients (coef): <none>\n")
      } else {
        cat("  Coefficients (coef): <unavailable>\n")
      }

      if (is.matrix(coef_mat) && nrow(coef_mat) > 0 && "exp(coef)" %in% colnames(coef_mat)) {
        cat("  exp(coef):\n")
        exp_fmt <- format_numeric_vec(coef_mat[, "exp(coef)"])
        if (!is.null(rownames(coef_mat))) {
          names(exp_fmt) <- rownames(coef_mat)
        }
        exp_out <- matrix(exp_fmt, ncol = 1)
        rownames(exp_out) <- if (!is.null(names(exp_fmt))) names(exp_fmt) else rownames(exp_out)
        colnames(exp_out) <- "exp(coef)"
        print(exp_out, quote = FALSE)
      } else if (removed_all_coef || no_coef_rows) {
        cat("  exp(coef): <none>\n")
      } else {
        cat("  exp(coef): <unavailable>\n")
      }

      conf_int_raw <- summary_fit$conf.int
      conf_int <- filter_matrix(conf_int_raw, remove_terms)
      conf_raw_is_matrix <- is.matrix(conf_int_raw)
      removed_all_conf <- conf_raw_is_matrix && nrow(conf_int_raw) > 0 &&
        (!is.matrix(conf_int) || nrow(conf_int) == 0)
      no_conf_rows <- conf_raw_is_matrix && nrow(conf_int_raw) == 0

      if (is.matrix(conf_int) && nrow(conf_int) > 0 &&
          all(c("exp(coef)", "lower .95", "upper .95") %in% colnames(conf_int))) {
        cat("  Hazard Ratios (95% CI):\n")
        hr_fmt <- format_numeric_vec(conf_int[, "exp(coef)"])
        lower_fmt <- format_numeric_vec(conf_int[, "lower .95"])
        upper_fmt <- format_numeric_vec(conf_int[, "upper .95"])
        if (!is.null(rownames(conf_int))) {
          names(hr_fmt) <- rownames(conf_int)
          names(lower_fmt) <- rownames(conf_int)
          names(upper_fmt) <- rownames(conf_int)
        }
        hr_mat <- cbind(
          "HR" = hr_fmt,
          "Lower 95%" = lower_fmt,
          "Upper 95%" = upper_fmt
        )
        rownames(hr_mat) <- if (!is.null(rownames(conf_int))) rownames(conf_int) else rownames(hr_mat)
        print(hr_mat, quote = FALSE)
      } else if (removed_all_conf || removed_all_coef || no_conf_rows) {
        cat("  Hazard Ratios (95% CI): <none>\n")
      } else {
        cat("  Hazard Ratios: <unavailable>\n")
      }

      if (length(strata_details) == 0 && !is.null(model_info)) {
        fallback_cols <- tryCatch(model_info$strata_cols, error = function(e) NULL)
        if (!is.null(fallback_cols) && length(fallback_cols) > 0) {
          for (sc in fallback_cols) {
            strata_details[[length(strata_details) + 1]] <- list(
              column = sc,
              display = sc,
              levels = character()
            )
          }
        }
      }
      if (length(strata_details) > 0) {
        xlevels <- tryCatch(fit_obj$xlevels, error = function(e) NULL)
        if (!is.null(xlevels)) {
          for (i in seq_along(strata_details)) {
            col_nm <- strata_details[[i]]$column
            if (!is.null(col_nm) && col_nm %in% names(xlevels)) {
              level_vals <- as.character(xlevels[[col_nm]])
              level_vals <- level_vals[!is.na(level_vals)]
              if (length(level_vals) > 0) {
                strata_details[[i]]$levels <- unique(level_vals)
              }
            }
          }
        }
        cat("  Stratified by:\n")
        for (entry in strata_details) {
          if (is.null(entry)) {
            next
          }
          display_nm <- entry$display
          if (is.null(display_nm) || length(display_nm) == 0 || !nzchar(as.character(display_nm)[1])) {
            display_nm <- entry$column
          }
          level_vals <- entry$levels
          if (is.null(level_vals)) {
            level_vals <- character()
          }
          level_vals <- as.character(level_vals)
          level_vals <- level_vals[!is.na(level_vals) & nzchar(level_vals)]
          level_vals <- unique(level_vals)
          n_strata <- length(level_vals)
          if (n_strata > 0) {
            cat(sprintf("    %s (%d strata: %s)", display_nm, n_strata, paste(level_vals, collapse = ", ")), "\n", sep = "")
          } else {
            cat(sprintf("    %s (0 strata)", display_nm), "\n", sep = "")
          }
        }
      }

      log_test <- summary_fit$logtest
      if (!is.null(log_test) && length(log_test) >= 3) {
        lr_text <- paste0(
          format_numeric_single(log_test[1]),
          " on ",
          format_numeric_single(log_test[2], digits = 0),
          " df (p = ",
          format_numeric_single(log_test[3]),
          ")"
        )
        cat("  Likelihood ratio test:", lr_text, "\n")
      } else {
        cat("  Likelihood ratio test: <unavailable>\n")
      }

      concord_val <- NA_real_
      if (!is.null(performance_row) && nrow(performance_row) >= 1 && "c_index" %in% colnames(performance_row)) {
        concord_val <- suppressWarnings(as.numeric(performance_row$c_index[1]))
      }
      if (!is.finite(concord_val)) {
        concord_vec <- summary_fit$concordance
        if (!is.null(concord_vec) && length(concord_vec) >= 1) {
          concord_val <- suppressWarnings(as.numeric(concord_vec[1]))
        }
      }
      if (is.finite(concord_val)) {
        cat("  Concordance (Harrell C-index):", format_numeric_single(concord_val), "\n")
      } else {
        cat("  Concordance (Harrell C-index): <unavailable>\n")
      }
      TRUE
    }

    print_xgb_survival_details <- function(fit_obj) {
      if (!inherits(fit_obj, "fastml_xgb_survival")) {
        return(FALSE)
      }
      objective <- tryCatch(fit_obj$objective, error = function(e) NULL)
      objective <- if (!is.null(objective) && length(objective) > 0 && nzchar(as.character(objective)[1])) {
        as.character(objective)[1]
      } else {
        "<unknown>"
      }
      if (identical(objective, "survival:cox")) {
        cat("  Objective: survival:cox  risk ranking only\n")
      } else if (identical(objective, "survival:aft")) {
        cat("  Objective: survival:aft  full survival modeling\n")
      } else {
        cat("  Objective:", objective, "\n")
      }
      booster <- tryCatch(fit_obj$booster, error = function(e) NULL)
      rounds <- NA_real_
      if (!is.null(booster)) {
        rounds <- tryCatch(as.numeric(booster$niter), error = function(e) NA_real_)
      }
      if (is.finite(rounds) && rounds > 0) {
        cat("  Boosting rounds:", format_numeric_single(rounds, digits = 0), "\n")
      }
      feat_count <- length(tryCatch(fit_obj$feature_names, error = function(e) NULL))
      if (feat_count > 0) {
        cat("  Predictors:", format_numeric_single(feat_count, digits = 0), "\n")
      }
      if (identical(objective, "survival:aft")) {
        dist_val <- tryCatch(fit_obj$aft_distribution, error = function(e) NULL)
        dist_str <- if (!is.null(dist_val) && length(dist_val) > 0 && nzchar(as.character(dist_val)[1])) {
          as.character(dist_val)[1]
        } else {
          "<unknown>"
        }
        cat("  Distribution:", dist_str, "\n")
        scale_val <- tryCatch(as.numeric(fit_obj$aft_scale), error = function(e) NA_real_)
        if (is.finite(scale_val)) {
          cat("  Scale:", format_numeric_single(scale_val), "\n")
        } else {
          cat("  Scale: <unavailable>\n")
        }
      }
      TRUE
    }

    print_parsnip_params <- function(model_obj) {
      parsnip_fit <- tryCatch(extract_fit_parsnip(model_obj), error = function(e) NULL)
      if (is.null(parsnip_fit)) {
        return(FALSE)
      }
      params <- tryCatch(parsnip_fit$spec$args, error = function(e) NULL)
      if (is.null(params) || length(params) == 0) {
        cat("  No hyperparameters found.\n")
        return(TRUE)
      }
      for (pname in names(params)) {
        val <- params[[pname]]
        if (inherits(val, "quosure")) {
          val <- tryCatch(
            eval(rlang::get_expr(val), envir = rlang::get_env(val)),
            error = function(e) val
          )
        }
        cat("  ", pname, ": ", format_param_value(val), "\n", sep = "")
      }
      TRUE
    }

    if (is.null(desired_models) || length(desired_models) == 0) {
      cat("Could not extract final fitted model details.\n")
    } else {
      model_labels <- names(desired_models)
      if (is.null(model_labels) || length(model_labels) == 0) {
        model_labels <- paste0("Model_", seq_along(desired_models))
      }

      any_success <- FALSE
      failure_detected <- FALSE

      for (idx in seq_along(desired_models)) {
        label <- model_labels[idx]
        model_obj <- desired_models[[idx]]
        cat("Model:", label, "\n")

        components <- parse_model_label(label)
        algo_name <- components$algorithm

        handled <- FALSE
        success <- FALSE

        if (!is.na(algo_name) && algo_name %in% c("survreg", "cox_ph", "penalized_cox", "stratified_cox", "time_varying_cox", "royston_parmar", "parametric_surv", "piecewise_exp", "xgboost")) {
          fit_obj <- extract_survival_fit(label, model_obj)
          if (inherits(fit_obj, "survreg")) {
            success <- isTRUE(print_survreg_details(fit_obj))
            handled <- TRUE
          } else if (inherits(fit_obj, "coxph")) {
            perf_row <- NULL
            if (!is.null(performance_lookup) &&
                nrow(performance_lookup) > 0 &&
                !is.na(algo_name) &&
                "Model" %in% colnames(performance_lookup)) {
              perf_idx <- which(performance_lookup$Model == algo_name)
              if (length(perf_idx) > 0 && "Engine" %in% colnames(performance_lookup)) {
                engine_val <- components$engine
                if (!is.na(engine_val)) {
                  perf_idx <- perf_idx[performance_lookup$Engine[perf_idx] == engine_val]
                }
              }
              if (length(perf_idx) == 0) {
                perf_idx <- which(performance_lookup$Model == algo_name)
              }
              if (length(perf_idx) > 0) {
                perf_row <- performance_lookup[perf_idx[1], , drop = FALSE]
              }
            }
            success <- isTRUE(print_coxph_details(
              fit_obj,
              model_info = model_obj,
              performance_row = perf_row
            ))
            handled <- TRUE
          } else if (inherits(fit_obj, c("stpm2", "pstpm2"))) {
            success <- isTRUE(print_stpm2_details(
              fit_obj,
              model_info = model_obj
            ))
            handled <- TRUE
          } else if (inherits(fit_obj, "flexsurvreg")) {
            success <- isTRUE(print_flexsurv_details(
              fit_obj,
              model_info = model_obj
            ))
            handled <- TRUE
          } else if (inherits(fit_obj, "fastml_xgb_survival")) {
            success <- isTRUE(print_xgb_survival_details(fit_obj))
            handled <- TRUE
          } else if (!is.null(fit_obj)) {
            # Unexpected fit type; treat as handled to avoid duplicate warnings.
            handled <- TRUE
          }
        }

        if (!handled) {
          success <- isTRUE(print_parsnip_params(model_obj))
        }

        if (!success) {
          cat("Could not extract final fitted model details.\n")
          failure_detected <- TRUE
        } else {
          any_success <- TRUE
        }

        cat("\n")
      }

      if (!any_success && failure_detected) {
        # All attempts failed; ensure user sees the generic warning once more.
        cat("Could not extract final fitted model details.\n")
      }
    }
  }

  if ("conf_mat" %in% type) {

    if (task == "classification") {
      df_best <- list()

      for (i in seq_along(best_model_name)) {
        model_name <- names(best_model_name)[i]
        engine     <- best_model_name[i]
        name_combined <- sprintf("%s (%s)", model_name, engine)

        pred_entry <- predictions_list[[model_name]]
        if (!is.null(pred_entry)) {
          if (is.data.frame(pred_entry)) {
            df_best[[name_combined]] <- pred_entry
          } else if (!is.null(pred_entry[[engine]])) {
            df_best[[name_combined]] <- pred_entry[[engine]]
          }
        } else if (!is.null(predictions_list[[name_combined]])) {
          df_best[[name_combined]] <- predictions_list[[name_combined]]
        } else {
          cat("No predictions found for", model_name, "with engine", engine, "\n")
        }
      }

      names_df_best <- unique(unlist(lapply(df_best, names)))

      if (!is.null(df_best) && "truth" %in% names_df_best && "estimate" %in% names_df_best) {
        cat("\n===========================\n")
        cat("Confusion Matrices by Model\n")
        cat("===========================\n\n")

        for (model_name in names(df_best)) {
          cat("Model:", model_name, "\n")
          cat("---------------------------\n")

          model_predictions <- df_best[[model_name]]
          if (!("truth" %in% colnames(model_predictions)) || !("estimate" %in% colnames(model_predictions))) {
            cat("Missing `truth` or `estimate` columns in predictions for model:", model_name, "\n\n")
            next
          }

          cm <- yardstick::conf_mat(model_predictions, truth = truth, estimate = estimate)
          print(cm)
          cat("\n")  # Extra space between models
        }
      } else {
        cat("\nNo valid predictions to compute confusion matrix.\n\n")
      }
    }

  }


  invisible(object)
}

#' @title Tuning Configuration and Complexity Presets
#' @description Functions and presets for configuring hyperparameter tuning grids
#'   with explicit speed-robustness trade-offs.
#' @name tuning_config
NULL

#' Tuning Complexity Presets
#'
#' Returns the configuration for a given tuning complexity level, including
#' grid levels, parameter ranges, and expected computational characteristics.
#'
#' @param complexity Character string specifying the tuning complexity level.
#'   One of:
#'   \describe{
#'     \item{\code{"quick"}}{Minimal tuning for fast iteration. 2-3 levels per
#'       parameter, narrow ranges. Best for: initial exploration, prototyping,
#'       small datasets, time-constrained scenarios. Typical grid size: 4-27 points.}
#'     \item{\code{"balanced"}}{Moderate tuning balancing speed and thoroughness.
#'       3-4 levels per parameter, standard ranges. Best for: most production use
#'       cases, medium datasets. Typical grid size: 27-256 points. This is the default.
#'       }
#'     \item{\code{"thorough"}}{Comprehensive tuning for maximum model quality.
#'       4-5 levels per parameter, wide ranges. Best for: final model selection,
#'       publications, competitions, when compute time is not a constraint.
#'       Typical grid size: 256-3125 points.}
#'     \item{\code{"exhaustive"}}{Maximum coverage tuning. 5-7 levels per parameter,
#'       very wide ranges. Best for: research, benchmarking, when you need to be
#'       certain you've found the best hyperparameters. Warning: Can be very slow.
#'       Typical grid size: 1000-10000+ points. Consider using Bayesian tuning instead.}
#'   }
#'
#' @return A list with components:
#'   \describe{
#'     \item{grid_levels}{Integer number of levels per parameter for grid search.}
#'     \item{bayes_iterations}{Integer number of iterations for Bayesian tuning.}
#'     \item{description}{Human-readable description of the complexity level.}
#'     \item{speed_estimate}{Relative speed estimate (1 = baseline).}
#'     \item{robustness_estimate}{Relative robustness estimate (1-5 scale).}
#'   }
#'
#' @details
#' ## Speed-Robustness Trade-offs
#'
#' Hyperparameter tuning involves a fundamental trade-off between computational
#' cost and the likelihood of finding optimal hyperparameters:
#'
#' \tabular{lllll}{
#'   \strong{Level} \tab \strong{Grid Size} \tab \strong{Time} \tab \strong{Robustness} \tab \strong{Use Case} \cr
#'   quick \tab 4-27 \tab ~1x \tab Low \tab Prototyping, debugging \cr
#'   balanced \tab 27-256 \tab ~10x \tab Medium \tab Most production use \cr
#'   thorough \tab 256-3125 \tab ~100x \tab High \tab Final models, papers \cr
#'   exhaustive \tab 1000-10000+ \tab ~1000x \tab Very High \tab Research, competitions \cr
#' }
#'
#' ### Recommendations:
#'
#' 1. **Start with "quick"** during development to iterate fast
#' 2. **Use "balanced"** for most production pipelines
#' 3. **Switch to "thorough"** for final model selection
#' 4. **Consider Bayesian tuning** (`tuning_strategy = "bayes"`) for high-dimensional
#'    parameter spaces instead of exhaustive grid search
#' 5. **Use adaptive/racing** (`adaptive = TRUE`) to early-stop poor configurations
#'
#' ### Computational Scaling:
#'
#' Grid search scales as O(L^P * F * N) where:
#' - L = number of levels per parameter
#' - P = number of parameters being tuned
#' - F = number of cross-validation folds
#' - N = dataset size
#'
#' For a model with 5 tunable parameters and 10-fold CV:
#' - quick (L=2): 2^5 * 10 = 320 model fits
#' - balanced (L=3): 3^5 * 10 = 2,430 model fits
#' - thorough (L=5): 5^5 * 10 = 31,250 model fits
#'
#' @examples
#' # Get configuration for balanced tuning
#' config <- get_tuning_complexity("balanced")
#' print(config$grid_levels)  # 3
#'
#' # See all available presets
#' print_tuning_presets()
#'
#' @export
get_tuning_complexity <- function(complexity = c("balanced", "quick", "thorough", "exhaustive")) {

  complexity <- match.arg(complexity)

  presets <- list(
    quick = list(
      grid_levels = 2L,
      bayes_iterations = 5L,
      description = "Minimal tuning for fast iteration and prototyping",
      speed_estimate = 1,
      robustness_estimate = 2,
      recommended_folds = 5L,
      param_range_scale = 0.7  # Narrower ranges
    ),
    balanced = list(
      grid_levels = 3L,
      bayes_iterations = 15L,
      description = "Balanced tuning for most production use cases",
      speed_estimate = 10,
      robustness_estimate = 3,
      recommended_folds = 10L,
      param_range_scale = 1.0  # Standard ranges
    ),
    thorough = list(
      grid_levels = 5L,
      bayes_iterations = 30L,
      description = "Comprehensive tuning for final model selection",
      speed_estimate = 100,
      robustness_estimate = 4,
      recommended_folds = 10L,
      param_range_scale = 1.3  # Wider ranges
    ),
    exhaustive = list(
      grid_levels = 7L,
      bayes_iterations = 50L,
      description = "Maximum coverage for research and competitions",
      speed_estimate = 1000,
      robustness_estimate = 5,
      recommended_folds = 10L,
      param_range_scale = 1.5  # Very wide ranges
    )
  )

  presets[[complexity]]
}

#' Print Tuning Presets Summary
#'
#' Prints a formatted summary of all available tuning complexity presets
#' with their characteristics and recommended use cases.
#'
#' @return Invisibly returns a data frame with preset information.
#'
#' @export
print_tuning_presets <- function() {
  presets <- c("quick", "balanced", "thorough", "exhaustive")

  cat("\n")
  cat("=============================================================================\n")
  cat("                    fastml Tuning Complexity Presets\n")
  cat("=============================================================================\n\n")

  cat("SPEED vs ROBUSTNESS TRADE-OFF:\n")
  cat("-----------------------------------------------------------------------------\n")
  cat(sprintf("%-12s %-8s %-12s %-12s %s\n",
              "Preset", "Levels", "Bayes Iter", "Rel. Time", "Robustness"))
  cat("-----------------------------------------------------------------------------\n")

  results <- lapply(presets, function(p) {
    config <- get_tuning_complexity(p)
    cat(sprintf("%-12s %-8d %-12d %-12s %s\n",
                p,
                config$grid_levels,
                config$bayes_iterations,
                paste0("~", config$speed_estimate, "x"),
                paste(rep("*", config$robustness_estimate), collapse = "")))
    data.frame(
      preset = p,
      grid_levels = config$grid_levels,
      bayes_iterations = config$bayes_iterations,
      speed_estimate = config$speed_estimate,
      robustness_estimate = config$robustness_estimate,
      stringsAsFactors = FALSE
    )
  })

  cat("-----------------------------------------------------------------------------\n\n")

  cat("ESTIMATED GRID SIZES (for 5 tunable parameters):\n")
  cat("  quick:      2^5 =      32 combinations\n")
  cat("  balanced:   3^5 =     243 combinations\n")
  cat("  thorough:   5^5 =   3,125 combinations\n")
  cat("  exhaustive: 7^5 =  16,807 combinations\n\n")

  cat("RECOMMENDATIONS:\n")
  cat("  - Start with 'quick' during development for fast iteration\
")
  cat("  - Use 'balanced' (default) for most production pipelines\n")
  cat("  - Switch to 'thorough' for final model selection or publications\n")
  cat("  - Consider tuning_strategy='bayes' instead of 'exhaustive' grid search\n")
  cat("  - Enable adaptive=TRUE for early stopping of poor configurations\n\n")

  cat("USAGE:\n")
  cat("  fastml(..., tuning_complexity = 'balanced')     # Use preset\n")
  cat("  fastml(..., grid_levels = 4)                    # Custom grid levels\n")
  cat("  fastml(..., tuning_strategy = 'bayes',          # Bayesian tuning\n")
  cat("              tuning_iterations = 20)\n\n")

  invisible(do.call(rbind, results))
}

#' Get Tuning Parameters for Complexity Level
#'
#' Returns algorithm-specific tuning parameter ranges adjusted for the
#' specified complexity level.
#'
#' @param algo Character string specifying the algorithm name.
#' @param train_data Data frame containing the training data.
#' @param label Character string specifying the outcome variable name.
#' @param engine Character string specifying the engine.
#' @param complexity Character string specifying tuning complexity level.
#'
#' @return A list of tuning parameter ranges.
#'
#' @details
#' Parameter ranges are scaled based on the complexity level:
#' - \code{quick}: Narrower ranges (70% of standard), fewer discrete values
#' - \code{balanced}: Standard ranges (100%)
#' - \code{thorough}: Wider ranges (130% of standard)
#' - \code{exhaustive}: Very wide ranges (150% of standard)
#'
#' @export
get_tuning_params_for_complexity <- function(algo, train_data, label, engine,
                                             complexity = "balanced") {
  # Get base parameters
  base_params <- get_default_tune_params(algo, train_data, label, engine)

  if (is.null(base_params)) {
    return(NULL)
  }

  # Get complexity configuration
  config <- get_tuning_complexity(complexity)
  scale <- config$param_range_scale

  # Scale continuous parameter ranges
  scaled_params <- lapply(names(base_params), function(param_name) {
    param_range <- base_params[[param_name]]

    if (is.null(param_range) || length(param_range) < 2) {
      return(param_range)
    }

    # Check if this is a log-scale parameter (typically negative values or very small)
    is_log_scale <- any(param_range < 0) ||
      (all(param_range > 0) && max(param_range) / min(param_range) > 100)

    if (is_log_scale) {
      # For log-scale parameters, scale the range width
      mid <- mean(param_range)
      half_width <- (param_range[2] - param_range[1]) / 2
      new_half_width <- half_width * scale
      c(mid - new_half_width, mid + new_half_width)
    } else {
      # For linear-scale parameters
      mid <- mean(param_range)
      half_width <- (param_range[2] - param_range[1]) / 2
      new_half_width <- half_width * scale
      new_range <- c(mid - new_half_width, mid + new_half_width)

      # Ensure non-negative for parameters that should be positive
      if (param_name %in% c("trees", "min_n", "neighbors", "hidden_units", "epochs", "num_comp")) {
        new_range <- pmax(new_range, 1)
      }
      if (param_name %in% c("sample_size", "mixture")) {
        new_range <- pmin(pmax(new_range, 0), 1)
      }

      new_range
    }
  })
  names(scaled_params) <- names(base_params)

  scaled_params
}

#' Expanded Default Tuning Parameters
#'
#' Returns expanded tuning parameter ranges that provide better coverage
#' than the minimal defaults. These are used when \code{tuning_complexity}
#' is set to "thorough" or "exhaustive".
#'
#' @param algo Algorithm name.
#' @param train_data Training data frame.
#' @param label Outcome variable name.
#' @param engine Engine name.
#'
#' @return A list of expanded tuning parameter ranges.
#'
#' @keywords internal
#' @export
get_expanded_tune_params <- function(algo, train_data, label, engine) {
  num_predictors <- ncol(train_data) - 1

  switch(algo,
         # Random Forest - expanded ranges
         "rand_forest" = list(
           mtry = c(1, min(num_predictors, max(2, floor(num_predictors * 0.8)))),
           trees = c(100, 1000),
           min_n = c(1, 20)
         ),

         # XGBoost - comprehensive tuning
         "xgboost" = list(
           trees = c(50, 500),
           tree_depth = c(1, 10),
           learn_rate = c(-3, -0.5),  # 0.001 to ~0.3
           loss_reduction = c(0, 10),
           min_n = c(1, 15),
           sample_size = c(0.5, 1.0),
           mtry = c(0.5, 1.0)  # As proportion for xgboost
         ),

         # LightGBM - comprehensive tuning
         "lightgbm" = list(
           trees = c(50, 500),
           tree_depth = c(2, 12),
           learn_rate = c(-3, -0.5),
           loss_reduction = c(0, 10),
           min_n = c(5, 50),
           sample_size = c(0.5, 1.0),
           mtry = c(0.5, 1.0)
         ),

         # Logistic Regression (regularized)
         "logistic_reg" = {
           if (engine %in% c("glmnet", "brulee", "LiblineaR", "spark", "h2o")) {
             list(penalty = c(-6, 1), mixture = c(0, 1))
           } else {
             list(penalty = NULL, mixture = NULL)
           }
         },

         # Multinomial Regression
         "multinom_reg" = {
           if (engine %in% c("glmnet", "brulee")) {
             list(penalty = c(-6, 1), mixture = c(0, 1))
           } else if (engine == "nnet") {
             list(penalty = c(0, 1))
           } else {
             list(penalty = c(-6, 1), mixture = c(0, 1))
           }
         },

         # Decision Tree
         "decision_tree" = list(
           cost_complexity = c(-6, -1),
           tree_depth = c(1, 15),
           min_n = c(2, 20)
         ),

         # SVM Linear
         "svm_linear" = list(
           cost = c(-4, 4),
           margin = c(0.01, 0.2)
         ),

         # SVM RBF
         "svm_rbf" = list(
           cost = c(-4, 4),
           rbf_sigma = c(-10, 0)
         ),

         # K-Nearest Neighbors
         "nearest_neighbor" = list(
           neighbors = c(1, min(50, floor(sqrt(nrow(train_data))))),
           weight_func = c("rectangular", "triangular", "epanechnikov", "gaussian"),
           dist_power = c(1, 2)
         ),

         # Naive Bayes
         "naive_Bayes" = list(
           smoothness = c(0, 2),
           Laplace = c(0, 2)
         ),

         # Neural Network (MLP)
         "mlp" = list(
           hidden_units = c(2, 20),
           penalty = c(-6, 0),
           epochs = c(50, 500)
         ),

         # Elastic Net
         "elastic_net" = list(
           penalty = c(-6, 1),
           mixture = c(0, 1)
         ),

         # Ridge Regression
         "ridge_reg" = list(
           penalty = c(-6, 2)
         ),

         # Lasso Regression
         "lasso_reg" = list(
           penalty = c(-6, 2)
         ),

         # PLS
         "pls" = list(
           num_comp = c(1, min(20, num_predictors))
         ),

         # C5 Rules
         "C5_rules" = list(
           trees = c(1, 100),
           min_n = c(2, 20)
         ),

         # Bag Tree
         "bag_tree" = list(
           cost_complexity = c(-6, -1),
           tree_depth = c(1, 15),
           min_n = c(2, 20)
         ),

         # Default - return NULL
         NULL
  )
}

#' Estimate Tuning Time
#'
#' Provides a rough estimate of tuning time based on the configuration.
#'
#' @param n_params Number of parameters being tuned.
#' @param n_folds Number of cross-validation folds.
#' @param n_rows Number of rows in training data.
#' @param complexity Tuning complexity level.
#' @param tuning_strategy Tuning strategy ("grid" or "bayes").
#' @param base_fit_time Estimated time for a single model fit in seconds.
#'
#' @return A list with estimated total time and breakdown.
#'
#' @export
estimate_tuning_time <- function(n_params, n_folds = 10, n_rows = 1000,
                                 complexity = "balanced",
                                 tuning_strategy = "grid",
                                 base_fit_time = 1) {
  config <- get_tuning_complexity(complexity)

  if (tuning_strategy == "grid") {
    n_combinations <- config$grid_levels ^ n_params
  } else if (tuning_strategy == "bayes") {
    n_combinations <- config$bayes_iterations
  } else {
    n_combinations <- 1
  }

  total_fits <- n_combinations * n_folds
  estimated_seconds <- total_fits * base_fit_time * (n_rows / 1000)

  list(
    grid_combinations = n_combinations,
    total_model_fits = total_fits,
    estimated_seconds = estimated_seconds,
    estimated_minutes = estimated_seconds / 60,
    estimated_hours = estimated_seconds / 3600,
    breakdown = sprintf(
      "%d combinations x %d folds = %d fits",
      n_combinations, n_folds, total_fits
    )
  )
}

#' Recommend Tuning Configuration
#'
#' Provides recommendations for tuning configuration based on dataset
#' characteristics and time constraints.
#'
#' @param n_rows Number of rows in training data.
#' @param n_predictors Number of predictor variables.
#' @param n_algorithms Number of algorithms to tune.
#' @param max_time_minutes Maximum acceptable tuning time in minutes.
#' @param tuning_strategy Preferred tuning strategy.
#'
#' @return A list with recommended configuration.
#'
#' @export
recommend_tuning_config <- function(n_rows, n_predictors, n_algorithms = 1,
                                    max_time_minutes = 30,
                                    tuning_strategy = "grid") {
  # Estimate base fit time (rough heuristic)
  base_fit_time <- 0.01 * sqrt(n_rows) * sqrt(n_predictors)

  # Average number of tunable parameters
  avg_params <- 4

  recommendations <- list()

  for (complexity in c("quick", "balanced", "thorough", "exhaustive")) {
    est <- estimate_tuning_time(
      n_params = avg_params,
      n_folds = 10,
      n_rows = n_rows,
      complexity = complexity,
      tuning_strategy = tuning_strategy,
      base_fit_time = base_fit_time
    )

    total_time <- est$estimated_minutes * n_algorithms

    recommendations[[complexity]] <- list(
      estimated_minutes = total_time,
      within_budget = total_time <= max_time_minutes
    )
  }

  # Find best complexity within budget
  feasible <- names(recommendations)[sapply(recommendations, function(x) x$within_budget)]
  recommended <- if (length(feasible) > 0) tail(feasible, 1) else "quick"

  list(
    recommended_complexity = recommended,
    recommendations = recommendations,
    suggestion = sprintf(
      "Recommended: '%s' complexity (~%.1f minutes for %d algorithm%s)",
      recommended,
      recommendations[[recommended]]$estimated_minutes,
      n_algorithms,
      if (n_algorithms > 1) "s" else ""
    ),
    alternative = if (recommended != "thorough" && "thorough" %in% names(recommendations)) {
      sprintf(
        "For better results, consider '%s' (~%.1f minutes) if time permits",
        "thorough",
        recommendations[["thorough"]]$estimated_minutes
      )
    } else {
      NULL
    }
  )
}

#' Internal helper to prepare explainer inputs from a fastml object
#'
#' @param object A fastml object.
#' @param data Character string specifying which data to use: "train" (default) or "test".
#' @keywords internal
fastml_prepare_explainer_inputs <- function(object, data = c("train", "test")) {
  if (!inherits(object, "fastml")) {
    stop("The input must be a 'fastml' object.")
  }

  data <- match.arg(data)

  if (is.null(object$processed_train_data) || is.null(object$label) ||
      !(object$label %in% names(object$processed_train_data))) {
    stop("Processed training data with the target column is required to build an explainer.")
  }

  # Select data source based on parameter

  if (data == "test") {
    if (is.null(object$raw_test_data)) {
      stop("Test data is not available in this fastml object. Use data = 'train' instead.")
    }
    raw_data <- as.data.frame(object$raw_test_data)
    processed_data <- if (!is.null(object$processed_test_data)) {
      as.data.frame(object$processed_test_data)
    } else {
      # Bake test data if processed version not stored
      if (!is.null(object$preprocessor)) {
        as.data.frame(recipes::bake(object$preprocessor, new_data = object$raw_test_data))
      } else {
        raw_data
      }
    }
  } else {
    raw_data <- if (!is.null(object$raw_train_data)) {
      as.data.frame(object$raw_train_data)
    } else {
      as.data.frame(object$processed_train_data)
    }
    processed_data <- as.data.frame(object$processed_train_data)
  }

  # Keep references to training data for preprocessing alignment
  raw_train_data <- if (!is.null(object$raw_train_data)) {
    as.data.frame(object$raw_train_data)
  } else {
    as.data.frame(object$processed_train_data)
  }
  processed_train_data <- as.data.frame(object$processed_train_data)

  # Use selected data source (train or test) for explanations
  selected_data <- raw_data
  rownames(selected_data) <- NULL

  label <- object$label
  x_raw <- selected_data[, setdiff(names(selected_data), label), drop = FALSE]
  x_raw <- as.data.frame(x_raw)
  rownames(x_raw) <- NULL
  y_raw <- selected_data[[label]]

  # Use processed predictors for actually scoring models
  x_processed <- as.data.frame(processed_data[, setdiff(names(processed_data), label), drop = FALSE])
  rownames(x_processed) <- NULL
  y_processed <- processed_data[[label]]

  best_model <- object$best_model
  fits <- list()
  model_names <- character()
  # Use training data labels for factor levels to ensure consistency
  train_y <- raw_train_data[[label]]
  label_levels <- if (is.factor(train_y)) levels(train_y) else if (is.factor(y_raw)) levels(y_raw) else NULL

  add_fit <- function(name, mod) {
    if (is.null(mod)) {
      return()
    }
    pf <- tryCatch(tune::extract_fit_parsnip(mod), error = function(e) NULL)
    if (is.null(pf) && inherits(mod, c("model_fit", "workflow"))) {
      pf <- mod
    }
    if (!is.null(pf)) {
      fits[[length(fits) + 1]] <<- pf
      model_names[[length(model_names) + 1]] <<- name
    }
  }

  combine_names <- function(parent, child) {
    if (is.null(parent) || !nzchar(parent)) return(child)
    if (is.null(child) || !nzchar(child)) return(parent)
    if (grepl(paste0("\\(", child, "\\)$"), parent)) return(parent)
    paste0(parent, " (", child, ")")
  }

  collect_fits <- function(models, parent = NULL) {
    if (is.null(models)) return()
    if (inherits(models, c("workflow", "model_fit"))) {
      add_fit(if (!is.null(parent) && nzchar(parent)) parent else "best_model", models)
      return()
    }
    if (is.list(models)) {
      model_names_local <- names(models)
      for (i in seq_along(models)) {
        nm <- if (!is.null(model_names_local) && nzchar(model_names_local[i])) {
          model_names_local[i]
        } else {
          paste0("model_", i)
        }
        combined <- combine_names(parent, nm)
        collect_fits(models[[i]], combined)
      }
    }
  }

  collect_fits(best_model)

  if (length(fits) == 0 && !is.null(object$best_model_name) && !is.null(object$models)) {
    target_names <- paste0(names(object$best_model_name), " (", object$best_model_name, ")")
    collect_fits(object$models[target_names])
  }

  if (length(fits) == 0 && !is.null(object$models)) {
    collect_fits(object$models)
  }

  if (length(fits) == 0) {
    stop("Unable to extract fitted model(s) for explanation.")
  }

  list(
    data_source = data,
    train_data = selected_data,
    raw_train_data = raw_train_data,
    processed_train_data = processed_train_data,
    x = x_processed,
    x_raw = x_raw,
    x_processed = x_processed,
    y = y_raw,
    y_raw = y_raw,
    y_processed = y_processed,
    label = label,
    task = object$task,
    positive_class = object$positive_class,
    event_class = if (!is.null(object$event_class)) object$event_class else NULL,
    label_levels = label_levels,
    preprocessor = object$preprocessor,
    fits = fits,
    model_names = model_names
  )
}

#' Resolve the positive class for binary classification
#'
#' Determines the positive class respecting event_class settings from fastml().
#' This ensures consistency across all explainer functions.
#'
#' @param prep A list from fastml_prepare_explainer_inputs containing:
#'   positive_class, event_class, label_levels
#' @param y_levels Character vector of actual levels from the target variable
#' @return Character string of the resolved positive class name
#' @keywords internal
resolve_positive_class <- function(prep, y_levels) {
  # 1. Use explicit positive_class if set and valid
  if (!is.null(prep$positive_class) && prep$positive_class %in% y_levels) {
    return(prep$positive_class)
  }

  # 2. Use label_levels with event_class to determine positive class
  if (!is.null(prep$label_levels) && length(prep$label_levels) >= 2) {
    if (!is.null(prep$event_class) && prep$event_class %in% c("first", "second")) {
      # Respect explicit event_class setting
      idx <- if (prep$event_class == "second") {
        min(2L, length(prep$label_levels))
      } else {
        1L
      }
      candidate <- prep$label_levels[idx]
      if (candidate %in% y_levels) {
        return(candidate)
      }
    } else {
      # Default: second level is typically the positive class
      # (e.g., "Yes" > "No", "True" > "False" alphabetically)
      candidate <- prep$label_levels[min(2L, length(prep$label_levels))]
      if (candidate %in% y_levels) {
        return(candidate)
      }
    }
  }

  # 3. Fallback: use event_class directly on y_levels
  if (!is.null(prep$event_class) && prep$event_class %in% c("first", "second") &&
      length(y_levels) >= 2) {
    idx <- if (prep$event_class == "second") min(2L, length(y_levels)) else 1L
    return(y_levels[idx])
  }

  # 4. Last resort: second level if available, else first
  if (length(y_levels) >= 2) {
    return(y_levels[2L])
  }
  y_levels[1L]
}

#' Clean Column Names or Character Vectors by Removing Special Characters
#'
#' This function can operate on either a data frame or a character vector:
#' \itemize{
#'   \item \strong{Data frame}: Detects columns whose names contain any character
#'         that is not a letter, number, or underscore,
#'         removes colons, replaces slashes with underscores, and spaces
#'         with underscores.
#'   \item \strong{Character vector}: Applies the same cleaning rules to every
#'         element of the vector.
#' }
#'
#' @param x A data frame or character vector to be cleaned.
#'
#' @return
#' \itemize{
#'   \item If \code{x} is a data frame: returns a data frame with cleaned column names.
#'   \item If \code{x} is a character vector: returns a character vector with cleaned elements.
#' }
#'
#' @importFrom dplyr rename_with all_of
#' @importFrom janitor make_clean_names
#' @importFrom stringr str_detect
#' @export
sanitize <- function(x) {

  # Helper function to detect special characters in a name
  has_special_chars <- function(name) {
    str_detect(name, "[^a-zA-Z0-9_]")
  }

  # Helper function that cleans a vector of names
  clean_names_vector <- function(name_vec) {
    # Handle empty vector case - sapply returns list() for empty input
    if (length(name_vec) == 0) {
      return(character(0))
    }
    # Use make_clean_names() with a custom 'replace' to handle mu, colons, slashes, and spaces
    sapply(name_vec, function(n) {
      make_clean_names(
        n,
        replace = c(
          ":"     = "",    # Remove colons
          "/"     = "_",   # Replace slashes with underscores
          " "     = "_"    # Replace spaces with underscores
        )
      )
    }, USE.NAMES = FALSE)
  }

  # If x is a data frame, rename its columns that contain special characters
  if (is.data.frame(x)) {
    # Identify columns needing cleaning
    columns_with_special_chars <- names(x)[has_special_chars(names(x))]

    if(length(columns_with_special_chars) > 0){
      x <- rename_with(
        x,
        ~ clean_names_vector(.),
        .cols = all_of(columns_with_special_chars)
      )
    }
    return(x)

    # If x is a character vector, simply clean each element
  } else if (is.character(x)) {
    return(clean_names_vector(x))

    # Otherwise, throw an error
  } else {
    stop("Input must be a data frame or a character vector.")
  }
}

library(testthat)
library(dplyr)

test_that("sanitize errors with invalid input type", {
  expect_error(
    sanitize(123),
    "Input must be a data frame or a character vector"
  )

  expect_error(
    sanitize(list(a = 1, b = 2)),
    "Input must be a data frame or a character vector"
  )
})

test_that("sanitize cleans data frame column names with colons", {
  df <- data.frame(a = 1, `b:c` = 2, check.names = FALSE)

  result <- sanitize(df)

  expect_false("b:c" %in% names(result))
  expect_true("bc" %in% names(result) || "b_c" %in% names(result))
})

test_that("sanitize cleans data frame column names with slashes", {
  df <- data.frame(a = 1, `b/c` = 2, check.names = FALSE)

  result <- sanitize(df)

  expect_false("b/c" %in% names(result))
  expect_true(any(grepl("b_c", names(result))))
})

test_that("sanitize cleans data frame column names with spaces", {
  df <- data.frame(a = 1, `b c` = 2, check.names = FALSE)

  result <- sanitize(df)

  expect_false("b c" %in% names(result))
  expect_true(any(grepl("b_c", names(result))))
})

test_that("sanitize leaves clean column names unchanged", {
  df <- data.frame(a = 1, b_c = 2, normal_name = 3)

  result <- sanitize(df)

  expect_equal(names(result), names(df))
})

test_that("sanitize handles mixed column names", {
  df <- data.frame(
    clean_name = 1,
    `dirty:name` = 2,
    `another/dirty` = 3,
    `with spaces` = 4,
    check.names = FALSE
  )

  result <- sanitize(df)

  expect_equal(names(result)[1], "clean_name")
  expect_false(any(grepl("[: /]", names(result))))
})

test_that("sanitize works with character vectors", {
  vec <- c("clean_name", "dirty:name", "with/slash", "with space")

  result <- sanitize(vec)

  expect_true(is.character(result))
  expect_equal(length(result), 4)
  expect_equal(result[1], "clean_name")
  expect_false(any(grepl("[: /]", result)))
})

test_that("sanitize returns data frame for data frame input", {
  df <- data.frame(a = 1, `b:c` = 2, check.names = FALSE)

  result <- sanitize(df)

  expect_true(is.data.frame(result))
})

test_that("sanitize returns character vector for character input", {
  vec <- c("a", "b:c")

  result <- sanitize(vec)

  expect_true(is.character(result))
  expect_false(is.data.frame(result))
})

test_that("sanitize preserves data in data frame", {
  df <- data.frame(`value:col` = c(1, 2, 3), check.names = FALSE)

  result <- sanitize(df)

  expect_equal(nrow(result), 3)
  expect_equal(result[[1]], c(1, 2, 3))
})

test_that("sanitize handles empty data frame", {
  df <- data.frame()

  result <- sanitize(df)

  expect_true(is.data.frame(result))
  expect_equal(ncol(result), 0)
})

test_that("sanitize handles empty character vector", {
  vec <- character(0)

  result <- sanitize(vec)

  expect_true(is.character(result))
  expect_equal(length(result), 0)
})

test_that("sanitize handles data frame with no special characters", {
  df <- data.frame(a = 1, b = 2, c_d = 3)

  result <- sanitize(df)

  expect_equal(names(result), names(df))
})

test_that("sanitize handles complex special character combinations", {
  vec <- c("a:b/c d", "x::y", "m/n/o")

  result <- sanitize(vec)

  expect_false(any(grepl("[:/]", result)))
  expect_false(any(grepl(" ", result)))
})

#' Define LightGBM Model Specification
#'
#' @inheritParams define_rand_forest_spec
#' @param early_stopping Logical. If TRUE, early stopping parameters are passed
#'   to the engine.
#' @return List containing the model specification (`model_spec`).
#' @importFrom parsnip boost_tree set_mode set_engine
#' @importFrom dplyr select all_of
#' @importFrom tune tune
#' @noRd
define_lightgbm_spec <- function(task, train_data, label, tuning = FALSE, engine = "lightgbm", early_stopping = FALSE) {
  # Ensure the lightgbm package is installed.
  if (!requireNamespace("lightgbm", quietly = TRUE)) {
    stop("The 'lightgbm' package is required but is not installed.")
  }

  # Determine the number of predictors (all columns except the label).
  num_predictors <- ncol(train_data %>% dplyr::select(-dplyr::all_of(label)))

  # Retrieve default parameters for lightgbm.
  # The helper is expected to return a list with elements:
  # trees, tree_depth, learn_rate, mtry, min_n, loss_reduction, sample_size.
  defaults <- get_default_params("lightgbm", task, num_predictors, engine)

  # Define the model specification using boost_tree().
  if (tuning) {
    model_spec <- boost_tree(
      trees           = tune(),
      tree_depth      = tune(),
      learn_rate      = tune(),
      mtry            = tune(),
      min_n           = tune(),
      loss_reduction  = tune(),
      sample_size     = tune()
    )
  } else {
    model_spec <- boost_tree(
      trees           = defaults$trees,
      tree_depth      = defaults$tree_depth,
      learn_rate      = defaults$learn_rate,
      mtry            = defaults$mtry,
      min_n           = defaults$min_n,
      loss_reduction  = defaults$loss_reduction,
      sample_size     = defaults$sample_size
    )
  }

  # Set the mode and specify engine-specific parameters.
  # - counts = TRUE means that mtry is interpreted as a count (and later converted to a proportion)
  # - bagging_freq = 1 will enable bagging at every iteration when sample_size (bagging fraction) is less than 1.
  # - verbose = -1 quiets lightgbm's output.
  # - num_threads = 0 instructs lightgbm to use all available cores.
  # - seed and deterministic aid in reproducibility.
  model_spec <- model_spec %>%
    set_mode(task)

  engine_args <- list(
    counts        = TRUE,
    bagging_freq  = 1,
    verbose       = -1,
    num_threads   = 0,
    seed          = 123,
    deterministic = TRUE
  )

  if (early_stopping) {
    engine_args$early_stop <- 10
    engine_args$validation <- 0.1
  }

  model_spec <- do.call(set_engine, c(list(object = model_spec, engine = engine), engine_args))

  list(model_spec = model_spec)
}

#' Define XGBoost Model Specification
#'
#' @inheritParams define_rand_forest_spec
#' @param early_stopping Logical. If TRUE, early stopping parameters are passed
#'   to the engine.
#' @return List containing the model specification (`model_spec`).
#' @importFrom parsnip boost_tree set_mode set_engine
#' @importFrom dplyr select all_of
#' @importFrom tune tune
#' @noRd
define_xgboost_spec <- function(task, train_data, label, tuning = FALSE, engine = "xgboost", early_stopping = FALSE) {
  num_predictors <- ncol(train_data %>% dplyr::select(-dplyr::all_of(label)))
  defaults <- get_default_params("xgboost", num_predictors)

  if (tuning) {
    model_spec <- boost_tree(
      trees = tune(),
      tree_depth = tune(),
      learn_rate = tune(),
      mtry = tune(),
      min_n = tune(),
      loss_reduction = tune(),
      sample_size = tune()
    )
  } else {
    model_spec <- boost_tree(
      trees = defaults$trees,
      tree_depth = defaults$tree_depth,
      learn_rate = defaults$learn_rate,
      mtry = defaults$mtry,
      min_n = defaults$min_n,
      loss_reduction = defaults$loss_reduction,
      sample_size = defaults$sample_size
    )
  }

  model_spec <- model_spec %>%
    set_mode(task)

  engine_args <- list()
  if (early_stopping) {
    engine_args$early_stop <- 10
    engine_args$validation <- 0.1
  }

  model_spec <- do.call(set_engine, c(list(object = model_spec, engine = engine), engine_args))
  list(model_spec = model_spec)
}

#' Define C5_rules Model Specification
#'
#' @inheritParams define_rand_forest_spec
#' @return List containing the model specification (`model_spec`).
#' @importFrom parsnip boost_tree set_mode set_engine
#' @importFrom tune tune
#' @noRd
define_C5_rules_spec <- function(task, tuning = FALSE, engine = "C5.0") {
  if (task != "classification") {
    stop("C5.0 is only applicable for classification tasks.")
  }
  defaults <- get_default_params("C5_rules")

  if (tuning) {
    model_spec <- boost_tree(
      trees = tune(),
      min_n = tune(),
      sample_size = tune()
    ) %>%
      set_mode("classification") %>%
      set_engine(engine)
  } else {
    model_spec <- boost_tree(
      trees = defaults$trees,
      min_n = defaults$min_n,
      sample_size = defaults$sample_size
    ) %>%
      set_mode("classification") %>%
      set_engine(engine)
  }
  list(model_spec = model_spec)
}

#' Plot Methods for \code{fastml} Objects
#'
#' \code{plot.fastml} produces visual diagnostics for a trained \code{fastml} object.
#'
#' @param x A \code{fastml} object (output of \code{\link{fastml}()}).
#' @param algorithm Character vector specifying which algorithm(s) to include when
#'   generating certain plots (e.g., ROC curves). Defaults to \code{"best"}.
#' @param type Character vector indicating which plot(s) to produce. Options are:
#'   \describe{
#'     \item{\code{"bar"}}{Bar plot of performance metrics across all models/engines.}
#'     \item{\code{"roc"}}{ROC curve(s) for binary classification models.}
#'     \item{\code{"calibration"}}{Calibration plot for the best model(s).}
#'     \item{\code{"residual"}}{Residual diagnostics for the best model.}
#'     \item{\code{"learning_curve"}}{Learning-curve plot if recorded during training.}
#'     \item{\code{"all"}}{Produce all available plots.}
#'   }
#' @param ... Additional arguments (currently unused).
#'
#' @examples
#' \donttest{
#'   ## Create a binary classification dataset from iris
#'   data(iris)
#'   iris <- iris[iris$Species != "setosa",]
#'   iris$Species <- factor(iris$Species)
#'
#'   ## Fit fastml model on binary classification task
#'   model <- fastml(data = iris, label = "Species", algorithms = c("rand_forest", "svm_rbf"))
#'
#'   ## 1. Plot all available diagnostics
#'   plot(model, type = "all")
#'
#'   ## 2. Bar plot of performance metrics
#'   plot(model, type = "bar")
#'
#'   ## 3. ROC curves (only for classification models)
#'   plot(model, type = "roc")
#'
#'   ## 4. Calibration plot (requires 'probably' package)
#'   plot(model, type = "calibration")
#'
#'   ## 5. ROC curves for specific algorithm(s) only
#'   plot(model, type = "roc", algorithm = "rand_forest")
#'
#'   ## 6. Residual diagnostics (only available for regression tasks)
#'   model <- fastml(data = mtcars, label = "mpg", algorithms = c("linear_reg", "xgboost"))
#'   plot(model, type = "residual")
#'}
#'
#' @details
#' When \code{type = "all"}, \code{plot.fastml} will produce a bar plot of metrics,
#' ROC curves (classification), calibration plot, and residual diagnostics (regression).
#' If you specify a subset of types, only those will be drawn.
#'
#' @method plot fastml
#' @importFrom graphics plot
#' @importFrom pROC multiclass.roc
#' @export
plot.fastml <- function(x,
                        algorithm = "best",
                        type = c("all", "bar", "roc", "calibration", "residual", "learning_curve"),
                        ...) {

  if (!inherits(x, "fastml")) {
    stop("The input must be a 'fastml' object.")
  }

  # Validate 'type' argument
  type <- match.arg(type, several.ok = TRUE)
  if ("all" %in% type) {
    type <- c("bar", "roc", "calibration", "residual", "learning_curve")
  }

  performance      <- x$performance
  predictions_list <- x$predictions
  task             <- x$task
  best_model_name  <- x$best_model_name
  optimized_metric <- x$metric
  positive_class   <- x$positive_class
  engine_names     <- x$engine_names
  resampling_plan  <- x$resampling_plan
  resampling_desc  <- fastml_describe_resampling(resampling_plan)

  # Rebuild performance_wide (robust to single- or multi-engine structures)
  metrics_list <- lapply(names(performance), function(model_name) {
    perf_entry <- performance[[model_name]]

    resolve_engine <- function(default_engine = NA_character_) {
      if (!is.null(engine_names) && !is.null(engine_names[[model_name]]) && !is.na(engine_names[[model_name]])) {
        return(engine_names[[model_name]])
      }
      if (!is.null(best_model_name) && model_name %in% names(best_model_name)) {
        return(as.character(best_model_name[[model_name]]))
      }
      default_engine
    }

    if (is.list(perf_entry) && !is.data.frame(perf_entry)) {
      engine_dfs <- lapply(names(perf_entry), function(engine_name) {
        df <- as.data.frame(perf_entry[[engine_name]])
        df$Engine <- engine_name
        df
      })
      combined_engines <- do.call(rbind, engine_dfs)
    } else {
      combined_engines <- as.data.frame(perf_entry)
      combined_engines$Engine <- resolve_engine()
    }

    combined_engines$Model <- model_name
    combined_engines
  })
  performance_df <- do.call(rbind, metrics_list)

  all_metric_names <- unique(performance_df$.metric)
  if (optimized_metric %in% all_metric_names) {
    main_metric <- optimized_metric
  } else {
    main_metric <- all_metric_names[1]
    warning("Optimized metric not available; using first metric.")
  }

  if (task == "classification") {
    desired_metrics <- c("accuracy", "f_meas", "kap", "precision", "sens", "spec", "roc_auc", "logloss", "brier_score")
  } else if (task == "regression") {
    desired_metrics <- c("rmse", "rsq", "mae")
  } else if (task == "survival") {
    brier_metrics <- sort(grep("^brier_t", all_metric_names, value = TRUE))
    desired_metrics <- c("c_index", "uno_c", "ibs", "rmst_diff", brier_metrics)
  } else {
    desired_metrics <- unique(performance_df$.metric)
  }
  desired_metrics <- intersect(desired_metrics, all_metric_names)
  if (length(desired_metrics) == 0) {
    desired_metrics <- main_metric
  }

  performance_sub <- performance_df[
    performance_df$.metric %in% desired_metrics,
  ] %>%
    dplyr::select(-dplyr::any_of(".estimator"))

  if (length(engine_names) == 1 && "LiblineaR" %in% engine_names) {
    performance_wide <- tidyr::pivot_wider(
      performance_sub,
      names_from  = ".metric",
      values_from = ".estimate"
    ) %>%
      dplyr::select(Model, Engine, accuracy, kap, sens, spec, precision, f_meas)
  } else {
    keep_metrics <- desired_metrics
    if (length(engine_names) == 1 && "LiblineaR" %in% engine_names) {
      keep_metrics <- intersect(keep_metrics, c("accuracy", "kap", "sens", "spec", "precision", "f_meas"))
    }
    performance_wide <- tidyr::pivot_wider(
      performance_sub,
      names_from  = ".metric",
      values_from = ".estimate"
    )
    select_cols <- c("Model", "Engine", keep_metrics)
    select_cols <- intersect(select_cols, colnames(performance_wide))
    performance_wide <- dplyr::select(performance_wide, dplyr::all_of(select_cols))
  }

  brier_cols <- grep("^brier_t", colnames(performance_wide), value = TRUE)
  ascending_metrics <- unique(c("rmse", "mae", "ibs", "logloss", "brier_score", "mse", brier_cols))
  if (main_metric %in% ascending_metrics) {
    performance_wide <- performance_wide[order(performance_wide[[main_metric]], na.last = TRUE), ]
  } else {
    performance_wide <- performance_wide[order(-performance_wide[[main_metric]], na.last = TRUE), ]
  }

  display_names <- c(
    accuracy  = "Accuracy",
    f_meas    = "F1 Score",
    kap       = "Kappa",
    precision = "Precision",
    roc_auc   = "ROC AUC",
    sens      = "Sensitivity",
    spec      = "Specificity",
    rsq       = "R-squared",
    mae       = "MAE",
    rmse      = "RMSE",
    c_index   = "Harrell C-index",
    uno_c     = "Uno's C-index",
    ibs       = "Integrated Brier Score",
    rmst_diff = "RMST diff (t_max)"
  )
  t_max_val <- x$survival_t_max
  if (!is.null(t_max_val) && length(t_max_val) == 1 && is.finite(t_max_val) && t_max_val > 0) {
    display_names[["rmst_diff"]] <- sprintf(
      "RMST diff (t<=%s)",
      format(t_max_val, trim = TRUE, digits = 4)
    )
  }
  if (!is.null(x$survival_brier_times)) {
    for (nm in names(x$survival_brier_times)) {
      if (!is.null(x$survival_brier_times[[nm]]) && is.finite(x$survival_brier_times[[nm]])) {
        time_label <- format(x$survival_brier_times[[nm]], trim = TRUE, digits = 4)
      } else {
        time_label <- nm
      }
      display_names[[nm]] <- sprintf("Brier(t=%s)", time_label)
    }
  }
  auto_brier_cols <- setdiff(grep("^brier_t", colnames(performance_wide), value = TRUE), names(display_names))
  if (length(auto_brier_cols) > 0) {
    display_names[auto_brier_cols] <- auto_brier_cols
  }

  # ============================
  # 1. Bar plot of metrics
  # ============================
  if ("bar" %in% type) {
    performance_melt <- performance_wide %>%
      tidyr::pivot_longer(
        cols      = -c(Model, Engine),
        names_to  = "Metric",
        values_to = "Value"
      )
    performance_melt$Value <- as.numeric(performance_melt$Value)
    performance_melt$Metric <- ifelse(
      performance_melt$Metric %in% names(display_names),
      display_names[performance_melt$Metric],
      performance_melt$Metric
    )

    if (task == "classification") {
      class_order <- c("Accuracy", "F1 Score", "Kappa", "Precision", "Sensitivity", "Specificity", "ROC AUC")
      present_class_metrics <- intersect(class_order, unique(performance_melt$Metric))
      if (length(present_class_metrics) > 0) {
        performance_melt$Metric <- factor(performance_melt$Metric, levels = present_class_metrics)
      }
    } else if (task == "regression") {
      reg_order <- c("RMSE", "R-squared", "MAE")
      present_reg_metrics <- intersect(reg_order, unique(performance_melt$Metric))
      if (length(present_reg_metrics) > 0) {
        performance_melt$Metric <- factor(performance_melt$Metric, levels = present_reg_metrics)
      }
    } else if (task == "survival") {
      rmst_label <- display_names[["rmst_diff"]]
      if (is.null(rmst_label)) {
        rmst_label <- "RMST diff (t_max)"
      }
      surv_order <- c("Harrell C-index", "Uno's C-index", "Integrated Brier Score", rmst_label)
      brier_labels <- display_names[grep("^brier_t", names(display_names))]
      surv_order <- c(surv_order, brier_labels)
      present_surv_metrics <- intersect(surv_order, unique(performance_melt$Metric))
      if (length(present_surv_metrics) > 0) {
        performance_melt$Metric <- factor(performance_melt$Metric, levels = present_surv_metrics)
      }
    }

    p_bar <- ggplot2::ggplot(
      performance_melt,
      ggplot2::aes(x = .data$Model, y = .data$Value, fill = .data$Engine)
    ) +
      ggplot2::geom_bar(stat = "identity", position = ggplot2::position_dodge()) +
      ggplot2::facet_wrap(~ Metric, scales = "free_y") +
      ggplot2::theme_bw() +
      ggplot2::theme(axis.text.x = ggplot2::element_text(angle = 45, hjust = 1)) +
      ggplot2::labs(
        title = "Model Performance Comparison",
        x     = "Model",
        y     = "Metric Value"
      )

    if (task == "classification") {
      p_bar <- p_bar + ggplot2::ylim(0, 1)
    }

    print(p_bar)
  }

  # ============================
  # 2. ROC curves (classification)
  # ============================
  if ("roc" %in% type) {
    if (task == "classification" && !is.null(predictions_list) && length(predictions_list) > 0) {

      # Create a list to collect data frames from all models/engines
      dfs <- list()

      # Loop over each algorithm (e.g., "rand_forest", "logistic_reg")

      if(all(algorithm == "best")){
        pred_list = predictions_list
      }else{

        pred_list = predictions_list[algorithm]

      }

      for (algo in names(pred_list)) {
        # Check if predictions are nested by engine or flat (single engine)
        algo_preds <- predictions_list[[algo]]

        if (is.data.frame(algo_preds)) {
          # Flat structure: single engine, get engine name from best_model_name or engine_names
          eng <- if (!is.null(best_model_name) && algo %in% names(best_model_name)) {
            as.character(best_model_name[[algo]])
          } else if (!is.null(engine_names) && algo %in% names(engine_names)) {
            as.character(engine_names[[algo]])
          } else {
            "default"
          }
          engines_to_process <- stats::setNames(list(algo_preds), eng)
        } else {
          # Nested structure: multiple engines
          engines_to_process <- algo_preds
        }

        # Loop over each engine in the current algorithm group
        for (eng in names(engines_to_process)) {
          if(eng == "LiblineaR"){
            warning("Engine 'LiblineaR' does not provide probability predictions; no ROC curve (roc_auc) will be computed.")

          }else{
            df <- engines_to_process[[eng]]
            if (!is.null(df) && "truth" %in% names(df)) {
              # Get probability columns matching ".pred_"

              if(eng == "h2o"){
                prob_cols <- grep("^\\.pred_p", names(df), value = TRUE)
              } else {
                prob_cols <- grep("^\\.pred_", names(df), value = TRUE)
              }

              if (length(prob_cols) >= 2) {  # binary classification should have two probability columns
                # Determine the predictor column for the positive class
                if(eng == "h2o"){
                  pred_col <- paste0(".pred_p", positive_class)

                }else{
                  pred_col <- paste0(".pred_", positive_class)
                }

                if (pred_col %in% prob_cols) {
                  # Add columns for algorithm and engine
                  df$Model  <- algo
                  df$Engine <- eng
                  # Use a compound key to store in the list
                  key <- paste(algo, eng, sep = "_")
                  dfs[[key]] <- df
                }
              }
            }
          }
        }
      }


      if (length(dfs) > 0) {
        # Combine all prediction data frames

        dfs_standardized <- lapply(dfs, function(df) {
          # Check if any column starts with ".pred_p"
          if (all(grepl("^\\.pred_p", names(df)))) {
            df <- df %>%
              rename_with(~ sub("^\\.pred_p", ".pred_", .x), starts_with(".pred_p"))
          }
          df
        })

        dfs_roc <-  bind_rows(dfs_standardized)

        # Ensure 'truth', 'Model', and 'Engine' are factors
        dfs_roc$truth  <- factor(dfs_roc$truth)
        dfs_roc$Model  <- as.factor(dfs_roc$Model)
        dfs_roc$Engine <- as.factor(dfs_roc$Engine)

        # We'll compute ROC curves for each unique combination of Model and Engine.
        pred_col <- paste0(".pred_", positive_class)

        if (length(levels(dfs_roc$truth)) != 2) {
          cat("\nROC curves are only generated for binary classification tasks.\n\n")
        } else if (!(pred_col %in% colnames(dfs_roc))) {
          cat("\nProbability column for the positive class not found; cannot compute ROC curves.\n\n")
        } else {
          roc_list <- list()

          # Get the unique combinations by splitting the compound key back into Model and Engine
          groups <- unique(dfs_roc[, c("Model", "Engine")])
          for (i in seq_len(nrow(groups))) {
            mod <- as.character(groups$Model[i])
            eng <- as.character(groups$Engine[i])
            data_model <- subset(dfs_roc, Model == mod & Engine == eng)
            data_model <- data_model[!is.na(data_model[[pred_col]]), ]

            roc_obj <- pROC::roc(response = data_model$truth,
                                predictor = data_model[[pred_col]],
                                direction = "auto",
                                quiet = TRUE)

            key <- paste(mod, eng, sep = " - ")
            roc_list[[key]] <- roc_obj
          }

          # Compute AUC values for each model/engine combination
          auc_values <- sapply(roc_list, function(x) pROC::auc(x))

          # Sort keys by AUC in alphabetical order of names
          sorted_keys <- auc_values[order(names(auc_values))]

          # Create a data frame for ROC curves by combining the ROC objects
          roc_data <- data.frame()
          for (key in names(sorted_keys)) {
            roc_obj <- roc_list[[key]]
            roc_df <- data.frame(
              FalsePositiveRate = rev(roc_obj$specificities),
              TruePositiveRate  = rev(roc_obj$sensitivities),
              ModelEngine       = key
            )
            roc_data <- rbind(roc_data, roc_df)
          }


          # Create the ROC curve plot, using the compound ModelEngine label for color
          roc_curve_plot <- ggplot(
            data = roc_data,
            aes(
              x = 1 - .data$FalsePositiveRate,
              y = .data$TruePositiveRate,
              color = .data$ModelEngine
            )
          ) +
            geom_line(linewidth = 1) +
            geom_abline(intercept = 0, slope = 1, linetype = "dashed", color = "grey") +
            theme_minimal() +
            labs(title = "ROC Curves for Models",
                 x = "1 - Specificity",
                 y = "Sensitivity") +
            theme(plot.title = element_text(hjust = 0.5)) +
            # Use sorted keys to label the legend with AUC values
            scale_color_manual(values = 1:length(sorted_keys),
                               labels = paste0(names(sorted_keys), " (AUC = ",
                                               sprintf("%.3f", auc_values[names(sorted_keys)]), ")")) +
            theme(legend.title = element_blank())

          print(roc_curve_plot)
        }

      } else {
        cat("\nNo suitable probability predictions for ROC curves.\n\n")
      }

    } else {
      if (is.null(predictions_list) || length(predictions_list) == 0) {
        cat("\nNo predictions available to generate ROC curves.\n\n")
      }
    }

  }

  # ============================
  # 4. Calibration Plot (classification)
  # ============================
  if ("calibration" %in% type) {
    if (task == "classification") {
      if (requireNamespace("probably", quietly = TRUE)) {
        df_best <- list()
        for (model in names(best_model_name)) {
          engine <- best_model_name[model]
          name_combined <- paste(model, "(", engine, ")", sep = "")

          model_preds <- predictions_list[[model]]
          if (!is.null(model_preds)) {
            if (is.data.frame(model_preds)) {
              # Flat structure: predictions stored directly as data.frame
              df_best[[name_combined]] <- model_preds
            } else if (!is.null(model_preds[[engine]])) {
              # Nested structure: predictions stored by engine
              df_best[[name_combined]] <- model_preds[[engine]]
            } else {
              cat("\nNo predictions found for", model, "with engine", engine, "\n")
            }
          } else {
            cat("\nNo predictions found for", model, "with engine", engine, "\n")
          }
        }

        names_df_best <- unique(unlist(lapply(df_best, names)))

        prob_cols <- grep("^\\.pred", names_df_best, value = TRUE)
        if (length(prob_cols) > 1) {
          for (model_name in names(df_best)) {
            model_predictions <- df_best[[model_name]]
            if (grepl("h2o", model_name)) {
              names(model_predictions) <- sub("^\\.pred_p", ".pred_", names(model_predictions))
            }
            pred_col <- paste0(".pred_", positive_class)
            if (!pred_col %in% colnames(model_predictions)) {
              cat("Error: Column", pred_col, "not found in model predictions for", model_name, "\n")
              next
            }
            prob_cols_metric <- setdiff(names(model_predictions), c("truth", "estimate"))
            prob_cols_metric <- prob_cols_metric[
              vapply(model_predictions[prob_cols_metric], is.numeric, logical(1))
            ]
            prob_info <- fastml_prepare_prob_matrix(model_predictions, prob_cols_metric)
            calib_label <- NULL
            if (!is.null(prob_info)) {
              calib <- fastml_class_calibration_metrics(
                truth_vec = model_predictions$truth,
                prob_mat = prob_info$mat,
                prob_cols = prob_info$cols,
                event_class = x$event_class
              )
              calib_vals <- c(calib$logloss, calib$brier_score, calib$ece)
              if (any(is.finite(calib_vals))) {
                calib_label <- paste0(
                  "Logloss: ",
                  ifelse(is.finite(calib$logloss), sprintf("%.3f", calib$logloss), "NA"),
                  ", Brier: ",
                  ifelse(is.finite(calib$brier_score), sprintf("%.3f", calib$brier_score), "NA"),
                  ", ECE: ",
                  ifelse(is.finite(calib$ece), sprintf("%.3f", calib$ece), "NA")
                )
              }
            }
            p_cal <- probably::cal_plot_breaks(
              model_predictions,
              truth    = truth,
              estimate = !!rlang::sym(pred_col),
              event_level = x$event_class
            ) +
              ggplot2::labs(
                title = paste("Calibration Plot for", model_name),
                subtitle = calib_label
              )
            print(p_cal)
          }
        } else {
          cat("\nNot enough probability columns for calibration plots.\n\n")
        }
      } else {
        cat("\nInstall the 'probably' package to generate calibration plots.\n")
      }
    } else {
      cat("\nCalibration plots are only available for classification tasks.\n\n")
    }
  }

  # ============================
  # 5. Residual Diagnostics (regression)
  # ============================
  if ("residual" %in% type) {
    if (task == "regression") {
      df_best <- list()
      # Determine the single best model engine
      for (model in names(best_model_name)) {
        engine <- best_model_name[model]
        name_combined <- paste(model, "(", engine, ")", sep = "")

        model_preds <- predictions_list[[model]]
        if (!is.null(model_preds)) {
          if (is.data.frame(model_preds)) {
            # Flat structure: predictions stored directly as data.frame
            df_best[[name_combined]] <- model_preds
          } else if (!is.null(model_preds[[engine]])) {
            # Nested structure: predictions stored by engine
            df_best[[name_combined]] <- model_preds[[engine]]
          } else {
            cat("\nNo predictions found for", model, "with engine", engine, "\n")
          }
        } else {
          cat("\nNo predictions found for", model, "with engine", engine, "\n")
        }
      }


      names_df_best <- unique(unlist(lapply(df_best, names)))
      if (!is.null(df_best) && length(df_best) > 0 && "truth" %in% names_df_best && "estimate" %in% names_df_best) {
        df_best_all <- dplyr::bind_rows(df_best, .id = "ModelEngine") %>%
          dplyr::mutate(residual = .data$truth - .data$estimate)

        cat("\nResidual Diagnostics for Best Model:\n")

        p_truth_pred <- ggplot2::ggplot(
          df_best_all,
          ggplot2::aes(x = .data$estimate, y = .data$truth)
        ) +
          ggplot2::geom_point(alpha = 0.6) +
          ggplot2::geom_abline(linetype = "dashed", color = "red") +
          ggplot2::labs(
            title = "Truth vs Predicted",
            x = "Predicted",
            y = "Truth"
          ) +
          ggplot2::theme_bw()
        print(p_truth_pred)

        p_resid_hist <- ggplot2::ggplot(
          df_best_all,
          ggplot2::aes(x = .data$residual)
        ) +
          ggplot2::geom_histogram(bins = 30, fill = "steelblue", color = "white", alpha = 0.7) +
          ggplot2::labs(
            title = "Residual Distribution",
            x = "Residual",
            y = "Count"
          ) +
          ggplot2::theme_bw()
        print(p_resid_hist)
      } else {
        cat("\nNo valid predictions to compute residual diagnostics.\n\n")
      }
    } else {
      cat("\nResidual diagnostics are only available for regression tasks.\n\n")
    }
  }

  # 5. Learning curve (optional)
  if ("learning_curve" %in% type) {
    lc <- x$learning_curve
    if (!is.null(lc) && !is.null(lc$plot)) {
      print(lc$plot)
    } else if (!is.null(lc) && !is.null(lc$data)) {
      lc_plot <- ggplot2::ggplot(
        lc$data,
        ggplot2::aes(x = .data$Fraction, y = .data$Performance)
      ) +
        ggplot2::geom_line(color = "blue") +
        ggplot2::geom_point(color = "blue") +
        ggplot2::labs(
          title = "Learning Curve",
          x = "Training Set Size (fraction)",
          y = paste("Mean", optimized_metric, "across models")
        ) +
        ggplot2::theme_minimal()
      print(lc_plot)
    } else {
      cat("\nNo learning curve data available. Set `learning_curve = TRUE` when fitting to record it.\n\n")
    }
  }

  invisible(x)
}

define_lasso_reg_spec <- function(task, tuning = FALSE, engine = "glmnet") {
  if (task != "regression") {
    stop("Lasso regression is only applicable for regression tasks.")
  }
  defaults <- get_default_params("lasso_reg")

  if (tuning) {
    model_spec <- linear_reg(
      penalty = tune(),
      mixture = defaults$mixture  # Fixed mixture for Lasso
    ) %>%
      set_mode("regression") %>%
      set_engine(engine)
  } else {
    model_spec <- linear_reg(
      penalty = defaults$penalty,
      mixture = defaults$mixture
    ) %>%
      set_mode("regression") %>%
      set_engine(engine)
  }
  list(model_spec = model_spec)
}

#' Define Ridge Regression Model Specification
#'
#' @inheritParams define_elastic_net_spec
#' @return List containing the model specification (`model_spec`).
#' @importFrom parsnip linear_reg set_mode set_engine
#' @importFrom tune tune
#' @noRd
define_ridge_reg_spec <- function(task, tuning = FALSE, engine = "glmnet") {
  if (task != "regression") {
    stop("Ridge regression is only applicable for regression tasks.")
  }
  defaults <- get_default_params("ridge_reg")

  if (tuning) {
    model_spec <- linear_reg(
      penalty = tune(),
      mixture = defaults$mixture  # Fixed mixture for Ridge
    ) %>%
      set_mode("regression") %>%
      set_engine(engine)
  } else {
    model_spec <- linear_reg(
      penalty = defaults$penalty,
      mixture = defaults$mixture
    ) %>%
      set_mode("regression") %>%
      set_engine(engine)
  }
  list(model_spec = model_spec)
}

#' Define Linear Regression Model Specification
#'
#' @inheritParams define_elastic_net_spec
#' @return List containing the model specification (`model_spec`).
#' @importFrom parsnip linear_reg set_mode set_engine
#' @noRd
define_linear_reg_spec <- function(task, engine = "lm") {
  if (task != "regression") {
    stop("Linear regression is only applicable for regression tasks.")
  }
  model_spec <- linear_reg() %>%
    set_mode("regression") %>%
    set_engine(engine)
  list(model_spec = model_spec)
}

#' Define Bayesian GLM Model Specification
#'
#' @inheritParams define_elastic_net_spec
#' @return List containing the model specification (`model_spec`).
#' @importFrom parsnip linear_reg set_mode set_engine
#' @noRd
define_bayes_glm_spec <- function(task, engine = "stan") {
  if (!requireNamespace("rstanarm", quietly = TRUE)) {
    stop("The 'rstanarm' package is required but is not installed.")
  }
  if (task != "regression") {
    stop("Bayesian GLM is only applicable for regression tasks.")
  }
  model_spec <- linear_reg() %>%
    set_mode("regression") %>%
    set_engine(engine)
  list(model_spec = model_spec)
}

#' Define Elastic Net Model Specification
#'
#' @param task Character string specifying the task type ("regression").
#' @inheritParams define_rand_forest_spec
#' @return List containing the model specification (`model_spec`).
#' @importFrom parsnip linear_reg set_mode set_engine
#' @importFrom tune tune
#' @noRd
define_elastic_net_spec <- function(task, tuning = FALSE, engine = "glmnet") {
  if (task != "regression") {
    stop("Elastic Net is only applicable for regression tasks.")
  }
  defaults <- get_default_params("elastic_net")
  if (tuning) {
    model_spec <- linear_reg(
      penalty = tune(),
      mixture = tune()
    ) %>%
      set_mode("regression") %>%
      set_engine(engine)
  } else {
    model_spec <- linear_reg(
      penalty = defaults$penalty,
      mixture = defaults$mixture
    ) %>%
      set_mode("regression") %>%
      set_engine(engine)
  }
  list(model_spec = model_spec)
}

#' Define bag_tree Model Specification
#'
#' @inheritParams define_decision_tree_spec
#' @return List containing the model specification (`model_spec`).
#' @importFrom parsnip bag_tree set_mode set_engine logistic_reg
#' @importFrom tune tune
#' @noRd
define_logistic_reg_spec <- function(task, tuning = FALSE, engine = "glm") {
  if (task != "classification") {
    stop("Logistic regression is only applicable for classification tasks.")
  }

  defaults <- get_default_params("logistic_reg", task, engine = engine)

  if (tuning) {
    if (engine %in% c("glm", "gee", "glmer", "stan", "stan_glmer")) {
      model_spec <- logistic_reg(
        penalty = NULL,
        mixture = NULL
      )
      if (engine == "glm") {
        model_spec <- model_spec %>%
          set_engine("glm", family = stats::binomial())
      } else if (engine == "gee") {
        model_spec <- model_spec %>%
          set_engine("gee", corstr = "exchangeable")
      } else if (engine == "glmer") {
        model_spec <- model_spec %>%
          set_engine("glmer")
      } else if (engine %in% c("stan", "stan_glmer")) {
        model_spec <- model_spec %>%
          set_engine(engine, chains = 4L, iter = 2000L, seed = 123, cores = 1L,
                     prior = NULL, prior_intercept = NULL)
      }
      model_spec <- model_spec %>% set_mode("classification")

    } else if (engine %in% c("brulee", "glmnet", "h2o", "LiblineaR", "spark")) {
      model_spec <- logistic_reg(
        penalty = tune(),
        mixture = tune()
      )
      if (engine == "brulee") {
        model_spec <- model_spec %>%
          set_engine("brulee",
                     optimizer = "SGD",
                     epochs = 50L,
                     learn_rate = 1,    # raw 0.01 -> log10(0.01) = -2
                     momentum = 0.9,
                     batch_size = 32,
                     stop_iter = 5L,
                     class_weights = NULL)
      } else if (engine == "glmnet") {
        model_spec <- model_spec %>%
          set_engine("glmnet", standardize = TRUE)
      } else if (engine == "h2o") {
        model_spec <- model_spec %>%
          set_engine("h2o", compute_p_values = FALSE, lambda_search = TRUE)
      } else if (engine == "LiblineaR") {
        model_spec <- model_spec %>%
          set_engine("LiblineaR", cost = Inf, verbose = FALSE)
      } else if (engine == "spark") {
        model_spec <- model_spec %>%
          set_engine("spark", standardization = TRUE, reg_param = 0.0, elastic_net_param = 0.0)
      }
      model_spec <- model_spec %>% set_mode("classification")

    } else if (engine == "keras") {
      model_spec <- logistic_reg(
        penalty = tune(),
        mixture = NULL
      ) %>%
        set_engine("keras", hidden_units = 1, act = "linear") %>%
        set_mode("classification")
    } else {
      stop("Unsupported engine specified for logistic regression.")
    }

  } else {
    model_spec <- logistic_reg(
      penalty = defaults$penalty,
      mixture = defaults$mixture
    )
    if (engine == "glm") {
      model_spec <- model_spec %>%
        set_engine("glm", family = stats::binomial())
    } else if (engine == "gee") {
      model_spec <- model_spec %>%
        set_engine("gee", corstr = "exchangeable")
    } else if (engine == "glmer") {
      model_spec <- model_spec %>%
        set_engine("glmer")
    } else if (engine %in% c("stan", "stan_glmer")) {
      model_spec <- model_spec %>%
        set_engine(engine, chains = 4L, iter = 2000L, seed = 123, cores = 1L,
                   prior = NULL, prior_intercept = NULL)
    } else if (engine == "brulee") {
      model_spec <- model_spec %>%
        set_engine("brulee",
                   optimizer = "SGD",
                   epochs = 50L,
                   learn_rate = 1,
                   momentum = 0.9,
                   batch_size = 32,
                   stop_iter = 5L,
                   class_weights = NULL)
    } else if (engine == "glmnet") {
      model_spec <- model_spec %>%
        set_engine("glmnet", standardize = TRUE)
    } else if (engine == "h2o") {
      model_spec <- model_spec %>%
        set_engine("h2o", compute_p_values = FALSE, lambda_search = TRUE)
    } else if (engine == "keras") {
      model_spec <- model_spec %>%
        set_engine("keras", hidden_units = 1, act = "linear")
    } else if (engine == "LiblineaR") {
      model_spec <- model_spec %>%
        set_engine("LiblineaR", cost = Inf, verbose = FALSE)
    } else if (engine == "spark") {
      model_spec <- model_spec %>%
        set_engine("spark", standardization = TRUE, reg_param = 0.0, elastic_net_param = 0.0)
    } else {
      stop("Unsupported engine specified for logistic regression.")
    }
    model_spec <- model_spec %>% set_mode("classification")
  }

  list(model_spec = model_spec)
}



#' Define Multinomial Logistic Regression Specification
#'
#' This function defines a multinomial logistic regression model specification.
#' It supports both a basic model and one with hyperparameter tuning for the penalty parameter.
#'
#' @param task Character. The type of task. Must be "classification".
#' @param tune Logical. If `TRUE`, includes hyperparameter tuning for the penalty parameter. Default is `FALSE`.
#' @return A list containing the model specification (`model_spec`).
#' @importFrom parsnip multinom_reg set_mode set_engine
#' @importFrom tune tune
#' @noRd
define_multinomial_reg_spec <- function(task, tuning = FALSE, engine) {
  if (task != "classification") {
    stop("Multinomial logistic regression is only applicable for classification tasks.")
  }

  # Use a default engine if none is specified.
  if (missing(engine)) {
    engine <- "nnet"
  }

  # Build the base model specification based on whether tuning is requested.
  if (tuning) {
    if (engine %in% c("nnet", "keras")) {
      # For nnet and keras, only the penalty tuning parameter is supported.
      model_spec <- multinom_reg(
        penalty = tune()
      )
    } else if (engine %in% c("brulee", "glmnet", "h2o", "spark")) {
      # For these engines, both penalty and mixture can be tuned.
      model_spec <- multinom_reg(
        penalty = tune(),
        mixture = tune()
      )
    } else {
      stop("Unsupported engine specified for multinom_reg")
    }
  } else {
    # When not tuning, use the default settings provided by multinom_reg()
    model_spec <- multinom_reg()
  }

  # Set the model mode.
  model_spec <- model_spec %>% set_mode("classification")

  # Engine-specific settings.
  if (engine == "nnet") {
    # nnet::multinom() uses a 'decay' parameter corresponding to penalty.
    model_spec <- model_spec %>%
      set_engine("nnet", trace = FALSE)
  } else if (engine == "brulee") {
    # brulee supports extra arguments (optimizer, epochs, learn_rate, etc.).
    model_spec <- model_spec %>%
      set_engine("brulee",
                 optimizer    = "SGD",
                 epochs       = 50L,
                 learn_rate   = 0.01,
                 momentum     = 0.9,
                 batch_size   = 32,
                 stop_iter    = 5L,
                 class_weights = NULL)
  } else if (engine == "glmnet") {
    # glmnet requires specifying standardization and family = "multinomial"
    model_spec <- model_spec %>%
      set_engine("glmnet", standardize = TRUE, family = "multinomial")
  } else if (engine == "h2o") {
    # h2o uses agua::h2o_train_glm() behind the scenes.
    # It expects penalty and mixture translated to lambda and alpha.
    model_spec <- model_spec %>%
      set_engine("h2o", lambda_search = TRUE, solver = "AUTO")
  } else if (engine == "keras") {
    # keras_mlp() fits a linear network with one hidden unit.
    model_spec <- model_spec %>%
      set_engine("keras", hidden_units = 1, act = "linear")
  } else if (engine == "spark") {
    # For spark, the underlying function (sparklyr::ml_logistic_reg) expects
    # reg_param and elastic_net_param in place of penalty and mixture.
    model_spec <- model_spec %>%
      set_engine("spark", standardization = TRUE, reg_param = 0.0, elastic_net_param = 0.0)
  } else {
    stop("Unsupported engine specified for multinom_reg")
  }

  list(model_spec = model_spec)
}


#' Define Partial Least Squares Model Specification
#'
#' @inheritParams define_elastic_net_spec
#' @return List containing the model specification (`model_spec`).
#' @importFrom parsnip pls set_mode set_engine
#' @importFrom tune tune
#' @noRd
define_pls_spec <- function(task, tuning = FALSE, engine = "mixOmics") {
  if (task != "regression") {
    stop("PLS is only applicable for regression tasks.")
  }
  if (!requireNamespace("plsmod", quietly = TRUE)) {
    stop("Package 'plsmod' is required for PLS models. Install it with: install.packages('plsmod')")
  }
  defaults <- get_default_params("pls")

  if (tuning) {
    model_spec <- pls(
      num_comp = tune()
    ) %>%
      set_mode("regression") %>%
      set_engine(engine)
  } else {
    model_spec <- pls(
      num_comp = defaults$num_comp
    ) %>%
      set_mode("regression") %>%
      set_engine(engine)
  }
  list(model_spec = model_spec)
}

#!/bin/bash 


Dir="/groups/pupko/cohenofi/HGT/runCOG.iTOL.38.3915/Batch.V9.93.PostExpSim.FewMPcostRatios.test"
c="4"
s=38
g=3915


for i in 1 2 3 4 5 6 7 8 9 10
do


for d in  GL$c  GL$c.MP1params GL$c.MP2params GL$c.MP3params GL$c.FixedParams GL$c.FixedParams
do

for r in 0.5 1 1.5 2 2.5 3 3.5 4 4.5 5
do

echo "Perl readPostExpSimulation $Dir/AcountForMissingData.$d/RESULTS/SimulatedPostExp$i  MP with cost $r"
perl /groups/pupko/cohenofi/pupkoSVN/trunk/scripts/gainLoss/readPostExpSimulation.pl  $Dir/AcountForMissingData.$d/RESULTS/SimulatedPostExp$i/ T $s $g $r
cp  $Dir/AcountForMissingData.$d/RESULTS/SimulatedPostExp$i/outRes.txt  $Dir/AcountForMissingData.$d/RESULTS/
mv $Dir/AcountForMissingData.$d/RESULTS/outRes.txt  $Dir/AcountForMissingData.$d/RESULTS/outResMP$i.$r.txt

done


echo "Perl readPostExpSimulation $Dir/AcountForMissingData.$d/RESULTS/SimulatedPostExp$i"
perl /groups/pupko/cohenofi/pupkoSVN/trunk/scripts/gainLoss/readPostExpSimulation.pl  $Dir/AcountForMissingData.$d/RESULTS/SimulatedPostExp$i/
cp  $Dir/AcountForMissingData.$d/RESULTS/SimulatedPostExp$i/outRes.txt  $Dir/AcountForMissingData.$d/RESULTS/
mv $Dir/AcountForMissingData.$d/RESULTS/outRes.txt  $Dir/AcountForMissingData.$d/RESULTS/outRes$i.txt

echo "Perl readPostExpSimulationPerPos $Dir/AcountForMissingData.$d/RESULTS/SimulatedPostExp$i"
perl /groups/pupko/cohenofi/pupkoSVN/trunk/scripts/gainLoss/readPostExpSimulationPerPos.pl $Dir/AcountForMissingData.$d/RESULTS/SimulatedPostExp$i/
cp  $Dir/AcountForMissingData.$d/RESULTS/SimulatedPostExp$i/outResPerPos.txt  $Dir/AcountForMissingData.$d/RESULTS/
mv $Dir/AcountForMissingData.$d/RESULTS/outResPerPos.txt  $Dir/AcountForMissingData.$d/RESULTS/outResPerPos$i.txt

echo "tar $Dir/AcountForMissingData.$d/RESULTS/SimulatedPostExp$i"
tar -cf $Dir/AcountForMissingData.$d/RESULTS/SimulatedPostExp$i.tar $Dir/AcountForMissingData.$d/RESULTS/SimulatedPostExp$i
echo "rm $Dir/AcountForMissingData.$d/RESULTS/SimulatedPostExp$i"
rm -R $Dir/AcountForMissingData.$d/RESULTS/SimulatedPostExp$i


done

done

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/params_helpers.R
\name{get_default_params}
\alias{get_default_params}
\title{Get Default Parameters for an Algorithm}
\usage{
get_default_params(algo, task, num_predictors = NULL, engine = NULL)
}
\arguments{
\item{algo}{A character string specifying the algorithm name. Supported values include:
\code{"rand_forest"}, \code{"C5_rules"}, \code{"xgboost"}, \code{"lightgbm"},
\code{"logistic_reg"}, \code{"multinom_reg"}, \code{"decision_tree"}, \code{"svm_linear"},
\code{"svm_rbf"}, \code{"nearest_neighbor"}, \code{"naive_Bayes"}, \code{"mlp"},
\code{"deep_learning"}, \code{"discrim_linear"}, \code{"discrim_quad"}, \code{"bag_tree"},
\code{"elastic_net"}, \code{"bayes_glm"}, \code{"pls"}, \code{"linear_reg"},
\code{"ridge_reg"}, \code{"lasso_reg"}, and \code{"penalized_cox"}.}

\item{task}{A character string specifying the task type, typically \code{"classification"} or \code{"regression"}.}

\item{num_predictors}{An optional numeric value indicating the number of predictors. This value is used to compute default values for parameters such as \code{mtry}. Defaults to \code{NULL}.}

\item{engine}{An optional character string specifying the engine to use. If not provided, a default engine is chosen where applicable.}
}
\value{
A list of default parameter settings for the specified algorithm. If the algorithm is not recognized, the function returns \code{NULL}.
}
\description{
Returns a list of default tuning parameters for the specified algorithm based on the task type, number of predictors, and engine.
}
\details{
The function employs a \code{switch} statement to select and return a list of default parameters tailored for the given algorithm, task, and engine. The defaults vary by algorithm and, in some cases, by engine. For example:
  \itemize{
    \item For \code{"rand_forest"}, if \code{engine} is not provided, it defaults to \code{"ranger"}. The parameters such as \code{mtry}, \code{trees}, and \code{min_n} are computed based on the task and the number of predictors.
    \item For \code{"C5_rules"}, the defaults include \code{trees}, \code{min_n}, and \code{sample_size}.
    \item For \code{"xgboost"} and \code{"lightgbm"}, default values are provided for parameters like tree depth, learning rate, and sample size.
    \item For \code{"logistic_reg"} and \code{"multinom_reg"}, the function returns defaults for regularization parameters (\code{penalty} and \code{mixture}) that vary with the specified engine.
    \item For \code{"decision_tree"}, the parameters (such as \code{tree_depth}, \code{min_n}, and \code{cost_complexity}) are set based on the engine (e.g., \code{"rpart"}, \code{"C5.0"}, \code{"partykit"}, \code{"spark"}).
    \item Other algorithms, including \code{"svm_linear"}, \code{"svm_rbf"}, \code{"nearest_neighbor"}, \code{"naive_Bayes"}, \code{"mlp"}, \code{"deep_learning"}, \code{"elastic_net"}, \code{"bayes_glm"}, \code{"pls"}, \code{"linear_reg"}, \code{"ridge_reg"}, and \code{"lasso_reg"}, have their respective default parameter lists.
  }
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/process_model_helpers.R
\name{convert_survival_predictions}
\alias{convert_survival_predictions}
\title{Convert Various Prediction Formats to Survival Matrix}
\usage{
convert_survival_predictions(pred_obj, eval_times, n_obs)
}
\arguments{
\item{pred_obj}{The prediction object.}

\item{eval_times}{Numeric vector of evaluation times.}

\item{n_obs}{Expected number of observations (rows).}
}
\value{
A standardized matrix of survival probabilities, or \code{NULL}
  on failure.
}
\description{
Attempts to convert various survival prediction formats (e.g., list of
data frames from \code{predict.model_fit} with type "survival", matrices)
into a standardized \code{[n_obs, n_eval_times]} matrix.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/explain_stability.R
\name{plot.fastml_stability}
\alias{plot.fastml_stability}
\title{Plot method for fastml_stability objects}
\usage{
\method{plot}{fastml_stability}(x, top_n = 15, ...)
}
\arguments{
\item{x}{A fastml_stability object from explain_stability().}

\item{top_n}{Integer. Number of top features to display. Default is 15.}

\item{...}{Additional arguments (ignored).}
}
\value{
A ggplot object.
}
\description{
Plot method for fastml_stability objects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/counterfactual_explain.R
\name{counterfactual_explain}
\alias{counterfactual_explain}
\title{Generate counterfactual explanations for a fastml model}
\usage{
counterfactual_explain(
  object,
  observation,
  variables = NULL,
  data = c("train", "test"),
  positive_class = NULL,
  event_class = NULL,
  label_levels = NULL,
  ...
)
}
\arguments{
\item{object}{A `fastml` object.}

\item{observation}{A single observation (data frame with one row) to compute
counterfactuals for.}

\item{variables}{Optional character vector of candidate variables to vary.
Only numeric variables are used for counterfactual profiling.}

\item{data}{Character string specifying which data to use for the explainer background:
\code{"train"} (default) or \code{"test"}.}

\item{positive_class}{Optional string used to filter lines/points in the
resulting profiles for classification tasks.}

\item{event_class}{Optional event class indicator propagated from
`fastml_prepare_explainer_inputs()` (kept for compatibility).}

\item{label_levels}{Optional vector of label levels propagated from
`fastml_prepare_explainer_inputs()` (kept for compatibility).}

\item{...}{Additional arguments passed to `DALEX::predict_profile`.}
}
\value{
A list (returned invisibly) containing the DALEX profile, filtered
  lines/points when `positive_class` is supplied, and the plotted object if
  rendering succeeds.
}
\description{
Uses DALEX ceteris-paribus profiles (`predict_profile`) to compute
counterfactual-style what-if explanations for a given observation.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sanitize.R
\name{sanitize}
\alias{sanitize}
\title{Clean Column Names or Character Vectors by Removing Special Characters}
\usage{
sanitize(x)
}
\arguments{
\item{x}{A data frame or character vector to be cleaned.}
}
\value{
\itemize{
  \item If \code{x} is a data frame: returns a data frame with cleaned column names.
  \item If \code{x} is a character vector: returns a character vector with cleaned elements.
}
}
\description{
This function can operate on either a data frame or a character vector:
\itemize{
  \item \strong{Data frame}: Detects columns whose names contain any character
        that is not a letter, number, or underscore,
        removes colons, replaces slashes with underscores, and spaces
        with underscores.
  \item \strong{Character vector}: Applies the same cleaning rules to every
        element of the vector.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/explainer_helpers.R
\name{fastml_prepare_explainer_inputs}
\alias{fastml_prepare_explainer_inputs}
\title{Internal helper to prepare explainer inputs from a fastml object}
\usage{
fastml_prepare_explainer_inputs(object, data = c("train", "test"))
}
\arguments{
\item{object}{A fastml object.}

\item{data}{Character string specifying which data to use: "train" (default) or "test".}
}
\description{
Internal helper to prepare explainer inputs from a fastml object
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/explain_dalex.R
\name{explain_dalex}
\alias{explain_dalex}
\title{Generate DALEX explanations for a fastml model}
\usage{
explain_dalex(
  object,
  data = c("train", "test"),
  features = NULL,
  grid_size = 20,
  shap_sample = 5,
  vi_iterations = 10,
  seed = 123,
  loss_function = NULL
)
}
\arguments{
\item{object}{A \code{fastml} object.}

\item{data}{Character string specifying which data to use for explanations:
\code{"train"} (default) uses training data, \code{"test"} uses held-out test data.
Using test data provides explanations that better reflect model generalization,
while training data explanations may be influenced by overfitting.}

\item{features}{Character vector of feature names for partial dependence (model profiles). Default NULL.}

\item{grid_size}{Number of grid points for partial dependence. Default 20.}

\item{shap_sample}{Integer number of observations from the selected data source to compute SHAP values for. Default 5.}

\item{vi_iterations}{Integer. Number of permutations for variable importance (B). Default 10.}

\item{seed}{Integer. A value specifying the random seed.}

\item{loss_function}{Function. The loss function for \code{model_parts}.
\itemize{
  \item If \code{NULL} and task = 'classification', defaults to \code{DALEX::loss_cross_entropy}.
  \item If \code{NULL} and task = 'regression', defaults to \code{DALEX::loss_root_mean_square}.
}}
}
\value{
Invisibly returns a list with variable importance, optional model profiles and SHAP values.
}
\description{
Creates a DALEX explainer and computes permutation based variable importance,
partial dependence (model profiles) and Shapley values.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tuning_config.R
\name{get_tuning_params_for_complexity}
\alias{get_tuning_params_for_complexity}
\title{Get Tuning Parameters for Complexity Level}
\usage{
get_tuning_params_for_complexity(
  algo,
  train_data,
  label,
  engine,
  complexity = "balanced"
)
}
\arguments{
\item{algo}{Character string specifying the algorithm name.}

\item{train_data}{Data frame containing the training data.}

\item{label}{Character string specifying the outcome variable name.}

\item{engine}{Character string specifying the engine.}

\item{complexity}{Character string specifying tuning complexity level.}
}
\value{
A list of tuning parameter ranges.
}
\description{
Returns algorithm-specific tuning parameter ranges adjusted for the
specified complexity level.
}
\details{
Parameter ranges are scaled based on the complexity level:
- \code{quick}: Narrower ranges (70% of standard), fewer discrete values
- \code{balanced}: Standard ranges (100%)
- \code{thorough}: Wider ranges (130% of standard)
- \code{exhaustive}: Very wide ranges (150% of standard)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/interaction_strength.R
\name{interaction_strength}
\alias{interaction_strength}
\title{Compute feature interaction strengths for a fastml model}
\usage{
interaction_strength(object, data = c("train", "test"), ...)
}
\arguments{
\item{object}{A `fastml` object.}

\item{data}{Character string specifying which data to use: \code{"train"} (default) or \code{"test"}.}

\item{...}{Additional arguments passed to `iml::Interaction`.}
}
\value{
An `iml::Interaction` object.
}
\description{
Uses the `iml` package to quantify the strength of feature interactions.
}
\examples{
\dontrun{
data(iris)
iris <- iris[iris$Species != "setosa", ]
iris$Species <- factor(iris$Species)
model <- fastml(data = iris, label = "Species")
interaction_strength(model)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/save.fastml.R
\name{save_fastml}
\alias{save_fastml}
\title{Save Model Function}
\usage{
save_fastml(model, filepath)
}
\arguments{
\item{model}{An object of class \code{fastml}.}

\item{filepath}{A string specifying the file path to save the model.}
}
\value{
No return value, called for its side effect of saving the model object to a file.
}
\description{
Saves the trained model object to a file.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tuning_config.R
\name{get_tuning_complexity}
\alias{get_tuning_complexity}
\title{Tuning Complexity Presets}
\usage{
get_tuning_complexity(
  complexity = c("balanced", "quick", "thorough", "exhaustive")
)
}
\arguments{
\item{complexity}{Character string specifying the tuning complexity level.
One of:
\describe{
  \item{\code{"quick"}}{Minimal tuning for fast iteration. 2-3 levels per
    parameter, narrow ranges. Best for: initial exploration, prototyping,
    small datasets, time-constrained scenarios. Typical grid size: 4-27 points.}
  \item{\code{"balanced"}}{Moderate tuning balancing speed and thoroughness.
    3-4 levels per parameter, standard ranges. Best for: most production use
    cases, medium datasets. Typical grid size: 27-256 points. This is the default.
    }
  \item{\code{"thorough"}}{Comprehensive tuning for maximum model quality.
    4-5 levels per parameter, wide ranges. Best for: final model selection,
    publications, competitions, when compute time is not a constraint.
    Typical grid size: 256-3125 points.}
  \item{\code{"exhaustive"}}{Maximum coverage tuning. 5-7 levels per parameter,
    very wide ranges. Best for: research, benchmarking, when you need to be
    certain you've found the best hyperparameters. Warning: Can be very slow.
    Typical grid size: 1000-10000+ points. Consider using Bayesian tuning instead.}
}}
}
\value{
A list with components:
  \describe{
    \item{grid_levels}{Integer number of levels per parameter for grid search.}
    \item{bayes_iterations}{Integer number of iterations for Bayesian tuning.}
    \item{description}{Human-readable description of the complexity level.}
    \item{speed_estimate}{Relative speed estimate (1 = baseline).}
    \item{robustness_estimate}{Relative robustness estimate (1-5 scale).}
  }
}
\description{
Returns the configuration for a given tuning complexity level, including
grid levels, parameter ranges, and expected computational characteristics.
}
\details{
## Speed-Robustness Trade-offs

Hyperparameter tuning involves a fundamental trade-off between computational
cost and the likelihood of finding optimal hyperparameters:

\tabular{lllll}{
  \strong{Level} \tab \strong{Grid Size} \tab \strong{Time} \tab \strong{Robustness} \tab \strong{Use Case} \cr
  quick \tab 4-27 \tab ~1x \tab Low \tab Prototyping, debugging \cr
  balanced \tab 27-256 \tab ~10x \tab Medium \tab Most production use \cr
  thorough \tab 256-3125 \tab ~100x \tab High \tab Final models, papers \cr
  exhaustive \tab 1000-10000+ \tab ~1000x \tab Very High \tab Research, competitions \cr
}

### Recommendations:

1. **Start with "quick"** during development to iterate fast
2. **Use "balanced"** for most production pipelines
3. **Switch to "thorough"** for final model selection
4. **Consider Bayesian tuning** (`tuning_strategy = "bayes"`) for high-dimensional
   parameter spaces instead of exhaustive grid search
5. **Use adaptive/racing** (`adaptive = TRUE`) to early-stop poor configurations

### Computational Scaling:

Grid search scales as O(L^P * F * N) where:
- L = number of levels per parameter
- P = number of parameters being tuned
- F = number of cross-validation folds
- N = dataset size

For a model with 5 tunable parameters and 10-fold CV:
- quick (L=2): 2^5 * 10 = 320 model fits
- balanced (L=3): 3^5 * 10 = 2,430 model fits
- thorough (L=5): 5^5 * 10 = 31,250 model fits
}
\examples{
# Get configuration for balanced tuning
config <- get_tuning_complexity("balanced")
print(config$grid_levels)  # 3

# See all available presets
print_tuning_presets()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/defaults_registry.R
\name{warn_default_override}
\alias{warn_default_override}
\title{Warn About Default Overrides}
\usage{
warn_default_override(
  algo,
  task,
  fastml_engine,
  fastml_params = NULL,
  verbose = FALSE,
  warn_once = TRUE
)
}
\arguments{
\item{algo}{Algorithm name.}

\item{task}{Task type.}

\item{fastml_engine}{fastml's default engine for this algorithm.}

\item{fastml_params}{fastml's default parameters (optional).}

\item{verbose}{If TRUE, always show the message (as a message, not warning).}

\item{warn_once}{If TRUE (default), only warn once per algorithm per session.}
}
\value{
Invisibly returns the comparison result.
}
\description{
Issues a warning if fastml defaults differ from parsnip defaults and
the warning hasn't been shown yet in this session.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/engine_helpers.R
\name{get_model_engine_names}
\alias{get_model_engine_names}
\title{Get Model Engine Names}
\usage{
get_model_engine_names(models)
}
\arguments{
\item{models}{A nested list of model workflows. Each inner list should contain model objects from which a fitted model can be extracted using \code{tune::extract_fit_parsnip}.}
}
\value{
A named character vector where the names correspond to algorithm names (e.g., \code{"rand_forest"}, \code{"logistic_reg"}) and the values correspond to the associated engine names (e.g., \code{"ranger"}, \code{"glm"}).
}
\description{
Extracts and returns a named vector mapping algorithm names to engine names from a nested list of model workflows.
}
\details{
The function iterates over a nested list of model workflows and, for each workflow, attempts to extract the fitted model object using \code{tune::extract_fit_parsnip}. If successful, it retrieves the algorithm name from the first element of the class attribute of the model specification and the engine name from the specification. The results are combined into a named vector.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/process_model_helpers.R
\name{determine_round_digits}
\alias{determine_round_digits}
\title{Determine rounding digits for time horizons}
\usage{
determine_round_digits(times)
}
\arguments{
\item{times}{Numeric vector of times.}
}
\value{
Integer number of digits between 0 and 6.
}
\description{
Computes a sensible number of decimal digits to round time horizons based on the minimal
positive separation between unique finite times.
}
\details{
Uses the smallest strictly positive difference among sorted unique finite times,
then returns \code{ceiling(-log10(min_diff))} truncated to \eqn{[0, 6]}.
}
\examples{
# Not run: determine_round_digits(c(0.1, 0.12, 0.125))
NULL

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/process_model_helpers.R
\name{extract_survreg_components}
\alias{extract_survreg_components}
\title{Extract survreg Linear Predictor and Scale}
\usage{
extract_survreg_components(fit_obj, new_data)
}
\arguments{
\item{fit_obj}{A fitted \code{survreg} object.}

\item{new_data}{A data frame with predictor variables.}
}
\value{
A list with elements \code{lp} (numeric vector) and \code{scale}
  (numeric vector), or \code{NULL} on failure.
}
\description{
Computes the linear predictor (lp) and scale parameter(s) for new data
from a fitted \code{survreg} model.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fastexplore.R
\name{fastexplore}
\alias{fastexplore}
\title{Lightweight exploratory helper}
\usage{
fastexplore(
  data,
  label = NULL,
  visualize = c("histogram", "boxplot", "barplot", "heatmap", "scatterplot"),
  save_results = FALSE,
  render_report = FALSE,
  output_dir = NULL,
  sample_size = NULL,
  interactive = FALSE,
  corr_threshold = 0.9,
  auto_convert_numeric = TRUE,
  visualize_missing = TRUE,
  imputation_suggestions = FALSE,
  report_duplicate_details = TRUE,
  detect_near_duplicates = FALSE,
  auto_convert_dates = FALSE,
  feature_engineering = FALSE,
  outlier_method = c("iqr", "zscore", "dbscan", "lof"),
  run_distribution_checks = TRUE,
  normality_tests = c("shapiro"),
  pairwise_matrix = TRUE,
  max_scatter_cols = 5,
  grouped_plots = TRUE,
  use_upset_missing = TRUE
)
}
\arguments{
\item{data}{A `data.frame` to explore.}

\item{label}{Optional column name of the target/label. If supplied and
categorical, grouped plots and class balance summaries are produced.}

\item{visualize}{Character vector indicating which plot families to build.
Defaults to `c("histogram", "boxplot", "barplot", "heatmap", "scatterplot")`.}

\item{save_results}{Logical; if `TRUE`, plots/results are saved under
`output_dir` (defaults to the working directory). Default is `FALSE`.}

\item{render_report}{Logical; if `TRUE`, a short HTML report is rendered via
`rmarkdown` (if available). Default is `FALSE`.}

\item{output_dir}{Directory to save results/report when `save_results` or
`render_report` is `TRUE`.}

\item{sample_size}{Optional integer; if supplied, visualizations are produced
on a random sample of this size.}

\item{interactive}{Logical; if `TRUE` and `plotly` is available, an
interactive correlation heatmap is produced. Falls back to static ggplot
output otherwise.}

\item{corr_threshold}{Absolute correlation threshold for flagging high
correlations.}

\item{auto_convert_numeric}{Logical; convert factor/character columns that
look numeric into numeric.}

\item{visualize_missing}{Logical; if `TRUE`, include simple missingness
visualizations.}

\item{imputation_suggestions}{Logical; if `TRUE`, prints lightweight
suggestions based on missingness patterns.}

\item{report_duplicate_details}{Logical; if `TRUE`, returns a small sample of
duplicated rows when present.}

\item{detect_near_duplicates}{Placeholder for future fuzzy duplicate checks.}

\item{auto_convert_dates}{Logical; convert YYYY-MM-DD strings to `Date`.}

\item{feature_engineering}{Logical; if `TRUE`, derive day/month/year from date
columns to aid inspection of temporal structure.}

\item{outlier_method}{One of `"iqr"`, `"zscore"`, `"dbscan"`, `"lof"`.}

\item{run_distribution_checks}{Logical; if `TRUE`, run normality tests on
numeric columns.}

\item{normality_tests}{Character vector of normality tests to run; currently
supports `"shapiro"` and `"ks"`.}

\item{pairwise_matrix}{Logical; if `TRUE` and `GGally` is available, returns a
ggpairs scatterplot matrix for a subset of numeric columns.}

\item{max_scatter_cols}{Maximum number of numeric columns to include in the
pairwise matrix.}

\item{grouped_plots}{Logical; if `TRUE` and `label` is a factor, group
histograms/boxplots/density plots by label.}

\item{use_upset_missing}{Logical; retained for compatibility. When `TRUE` and
`UpSetR` is installed, an UpSet plot of missingness is returned; otherwise a
simpler missingness heatmap is used.}
}
\value{
A list of summaries (tables/tibbles) and plot objects (ggplot/plotly),
  plus any saved file paths when `save_results`/`render_report` are enabled.
}
\description{
`fastexplore()` is an optional, lightweight exploratory data analysis (EDA)
helper. It returns summary tables and plot objects; it only writes to disk or
renders a report when you explicitly request it via `save_results` or
`render_report`.
}
\details{
This helper is intentionally decoupled from the core modeling workflow. Most
of its heavy dependencies are treated as optional and loaded via
`requireNamespace()` when requested features are used.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/engine_helpers.R
\name{get_default_engine}
\alias{get_default_engine}
\title{Get Default Engine}
\usage{
get_default_engine(algo, task = NULL)
}
\arguments{
\item{algo}{A character string specifying the name of the algorithm. The value should match one of the supported algorithm names.}

\item{task}{Optional task type (e.g., \code{"classification"}, \code{"regression"}, or \code{"survival"}). Used to determine defaults that depend on the task.}
}
\value{
A character string containing the default engine name associated with the provided algorithm.
}
\description{
Returns the default engine corresponding to the specified algorithm.
}
\details{
The function uses a \code{switch} statement to select the default engine based on the given algorithm. For survival random forests, the function defaults to \code{"aorsf"}. If the provided algorithm does not have a defined default engine, the function terminates with an error.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/process_model_helpers.R
\name{align_survival_curve}
\alias{align_survival_curve}
\title{Align Survival Curve to Evaluation Times}
\usage{
align_survival_curve(curve_times, curve_surv, eval_times)
}
\arguments{
\item{curve_times}{Numeric vector of time points from the survival curve.}

\item{curve_surv}{Numeric vector of survival probabilities corresponding to
\code{curve_times}.}

\item{eval_times}{Numeric vector of new time points to evaluate at.}
}
\value{
A numeric vector of survival probabilities at \code{eval_times}.
}
\description{
Aligns a survival curve (defined by time points and survival probabilities)
to a new set of evaluation times using constant interpolation (last value
carried forward). Ensures \eqn{S(0) = 1} and monotonicity.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/explain_stability.R
\name{explain_stability}
\alias{explain_stability}
\title{Analyze Feature Importance Stability Across Cross-Validation Folds}
\usage{
explain_stability(
  object,
  model_name = NULL,
  vi_iterations = 10,
  seed = 123,
  plot = FALSE,
  conf_level = 0.95
)
}
\arguments{
\item{object}{A \code{fastml} object trained with \code{store_fold_models = TRUE}.}

\item{model_name}{Character string specifying which model to analyze. If NULL,
uses the best model. Should match the format "algorithm (engine)", e.g.,
"rand_forest (ranger)".}

\item{vi_iterations}{Integer. Number of permutations for variable importance
per fold. Default is 10 for faster computation across many folds.}

\item{seed}{Integer. Random seed for reproducibility.}

\item{plot}{Logical. If TRUE, displays a stability plot showing
mean importance with confidence intervals. Default is FALSE.}

\item{conf_level}{Numeric. Confidence level for intervals. Default is 0.95.}
}
\value{
A list with class \code{"fastml_stability"} containing:
  \describe{
    \item{importance_summary}{Data frame with aggregated feature importance
      (mean, sd, se, lower/upper CI) across folds.}
    \item{fold_importance}{List of per-fold variable importance results.}
    \item{rank_stability}{Data frame showing how feature ranks vary across folds.}
    \item{n_folds}{Number of folds analyzed.}
    \item{model_name}{Name of the model analyzed.}
  }
}
\description{
Computes feature importance for each fold model and aggregates results
to assess the stability of feature importance rankings across resamples.
This helps identify features that are consistently important vs those
whose importance varies across different data subsets.
}
\details{
This function requires that the fastml model was trained with
\code{store_fold_models = TRUE}, which stores the models fitted on each
cross-validation fold. Without stored fold models, only the final best
model is available, and cross-fold stability analysis is not possible.

The stability analysis computes permutation-based variable importance
for each fold's model using DALEX, then aggregates across folds to show:
\itemize{
  \item Mean importance and standard deviation
  \item Confidence intervals for importance
  \item Rank stability (how consistently features rank across folds)
}

Features with high mean importance but also high variance may be
important for some data subsets but not others, suggesting potential
instability in the model's reliance on those features.
}
\examples{
\donttest{
# Train model with fold models stored
model <- fastml(
  data = iris,
  label = "Species",
  algorithms = "rand_forest",
  store_fold_models = TRUE
)

# Analyze stability
stability <- explain_stability(model)
print(stability)
plot(stability)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/process_model_helpers.R
\name{get_surv_info}
\alias{get_surv_info}
\title{Extract Time and Status from Survival Matrix}
\usage{
get_surv_info(surv_matrix_vals, default_time, default_status)
}
\arguments{
\item{surv_matrix_vals}{A matrix, typically from \code{Surv(time, status)}.}

\item{default_time}{Default time vector if not found.}

\item{default_status}{Default status vector if not found.}
}
\value{
A list with elements \code{time} and \code{status}.
}
\description{
Helper function to extract "time" and "status" columns from a matrix
(like one returned by \code{survival::Surv()}), falling back to defaults.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/explain_lime.R
\name{explain_lime}
\alias{explain_lime}
\title{Generate LIME explanations for a fastml model}
\usage{
explain_lime(
  object,
  new_observation,
  data = c("train", "test"),
  n_features = 5,
  n_labels = 1,
  ...
)
}
\arguments{
\item{object}{A `fastml` object.}

\item{new_observation}{A data frame containing the new observation(s) to explain.
Must contain the same columns as the original training data (before preprocessing).
The function will apply the stored preprocessor to transform it.}

\item{data}{Character string specifying which data to use for the LIME explainer background:
\code{"train"} (default) or \code{"test"}.}

\item{n_features}{Number of features to show in the explanation. Default 5.}

\item{n_labels}{Number of labels to explain (classification only). Default 1.}

\item{...}{Additional arguments passed to `lime::explain`.}
}
\value{
An object produced by `lime::explain`.
}
\description{
Creates a `lime` explainer using processed (encoded, scaled) data and returns
feature explanations for new observations. The new observation is automatically
preprocessed using the same recipe to ensure alignment with the explainer background.
}
\examples{
\dontrun{
data(iris)
iris <- iris[iris$Species != "setosa", ]
iris$Species <- factor(iris$Species)
model <- fastml(data = iris, label = "Species")
explain_lime(model, new_observation = iris[1, ])
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_selection_helpers.R
\name{flatten_and_rename_models}
\alias{flatten_and_rename_models}
\title{Flatten and Rename Models}
\usage{
flatten_and_rename_models(models)
}
\arguments{
\item{models}{A named list of model objects or nested named lists of model
objects. Nested entries are flattened and renamed using the pattern
`algorithm (engine)`.}
}
\value{
A flat named list of model objects.
}
\description{
Flatten and Rename Models
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fastml.R
\name{fastml}
\alias{fastml}
\title{Fast Machine Learning Function}
\usage{
fastml(
  data = NULL,
  train_data = NULL,
  test_data = NULL,
  label,
  algorithms = "all",
  task = "auto",
  test_size = 0.2,
  resampling_method = if (identical(task, "survival")) "none" else "cv",
  folds = ifelse(grepl("cv", resampling_method), 10, 25),
  repeats = NULL,
  group_cols = NULL,
  block_col = NULL,
  block_size = NULL,
  initial_window = NULL,
  assess_window = NULL,
  skip = 0,
  outer_folds = NULL,
  event_class = "first",
  exclude = NULL,
  recipe = NULL,
  tune_params = NULL,
  engine_params = list(),
  metric = NULL,
  class_threshold = 0.5,
  algorithm_engines = NULL,
  use_parsnip_defaults = FALSE,
  warn_engine_defaults = TRUE,
  n_cores = 1,
  engine_threads = 1,
  stratify = TRUE,
  impute_method = "error",
  encode_categoricals = TRUE,
  scaling_methods = c("center", "scale"),
  balance_method = "none",
  resamples = NULL,
  summaryFunction = NULL,
  use_default_tuning = FALSE,
  tuning_strategy = "grid",
  tuning_iterations = 10,
  tuning_complexity = "balanced",
  grid_levels = NULL,
  early_stopping = FALSE,
  adaptive = FALSE,
  learning_curve = FALSE,
  seed = 123,
  verbose = FALSE,
  eval_times = NULL,
  survival_metric_convention = "fastml",
  bootstrap_ci = TRUE,
  bootstrap_samples = 500,
  bootstrap_seed = NULL,
  at_risk_threshold = 0.1,
  audit_mode = FALSE,
  multiclass_auc = "macro",
  store_fold_models = FALSE
)
}
\arguments{
\item{data}{A data frame containing the complete dataset. If both `train_data` and `test_data` are `NULL`, `fastml()` will split this into training and testing sets according to `test_size` and `stratify`. When `group_cols` is supplied, the holdout keeps groups intact; when `block_col` is supplied, the holdout uses the last rows in time order. Defaults to `NULL`.}

\item{train_data}{A data frame pre-split for model training. If provided, `test_data` must also be supplied, and no internal splitting will occur. Defaults to `NULL`.}

\item{test_data}{A data frame pre-split for model evaluation. If provided, `train_data` must also be supplied, and no internal splitting will occur. Defaults to `NULL`.}

\item{label}{A string specifying the name of the target variable. For
survival analysis, supply a character vector with the names of the time and
status columns.}

\item{algorithms}{A vector of algorithm names to use. Default is \code{"all"} to run all supported algorithms.}

\item{task}{Character string specifying model type selection. Use "auto" to
let the function detect whether the target is for classification, regression,
or survival based on the data. Survival is detected when `label` is a
character vector of length 2 that matches time and status columns in the data.
You may also explicitly set to "classification", "regression", or "survival".}

\item{test_size}{A numeric value between 0 and 1 indicating the proportion of the data to use for testing. For grouped holdout, this is applied to groups; for time-ordered holdout, it selects the final proportion of rows. Default is \code{0.2}.}

\item{resampling_method}{A string specifying the resampling method for model evaluation. Default is \code{"cv"}
(cross-validation) for classification/regression. Other options include \code{"none"}, \code{"boot"},
\code{"repeatedcv"}, \code{"grouped_cv"}, \code{"blocked_cv"}, \code{"rolling_origin"}, and \code{"nested_cv"}.
For survival tasks, resampling is supported for parsnip-compatible engines (e.g., censored/ranger, glmnet).
Native survival engines (flexsurv/rstpm2/custom xgboost) ignore resampling and will error if custom resamples
are supplied. When the task auto-detects survival and \code{resampling_method} is omitted, it defaults to
\code{"none"} so native engines continue to run; set it explicitly to enable resampling for parsnip survival fits.}

\item{folds}{An integer specifying the number of folds for cross-validation
(default \code{10} for methods containing "cv", \code{25} otherwise).
When \code{resampling_method = "boot"}, this controls the number of
bootstrap resamples.  When \code{resampling_method = "validation_split"},
the proportion held out for validation is derived as \code{1 - 1/folds}.}

\item{repeats}{Number of times to repeat cross-validation (only applicable for methods like "repeatedcv").}

\item{group_cols}{Character vector naming one or more grouping columns used when
\code{resampling_method = "grouped_cv"} or when grouped nested cross-validation is desired.
All rows that share the same combination of values remain together in every fold. Columns must exist
in the training data and cannot contain missing values.}

\item{block_col}{Single column name that defines the ordering variable for
\code{resampling_method = "blocked_cv"} or \code{"rolling_origin"}. Data must already be sorted in
ascending order by this column to avoid leakage from future observations.}

\item{block_size}{Positive integer specifying the block size for \code{"blocked_cv"}.}

\item{initial_window}{Positive integer giving the number of observations in the initial training
window for \code{"rolling_origin"} resampling.}

\item{assess_window}{Positive integer giving the number of observations in each assessment window for
\code{"rolling_origin"} resampling.}

\item{skip}{Non-negative integer specifying how many potential rolling windows to skip between
successive resamples when \code{resampling_method = "rolling_origin"}.}

\item{outer_folds}{Positive integer giving the number of outer folds to use when
\code{resampling_method = "nested_cv"} and no custom \code{resamples} object is supplied.}

\item{event_class}{A single string. Either "first" or "second" to specify which
level of the binary outcome factor to treat as the positive class (the "event").
For binary classification, "first" treats the first factor level as the positive
class, "second" treats the second level as positive. Use
\code{levels(your_data$outcome)} to check level order before training. Default is "first".}

\item{exclude}{A character vector specifying the names of the columns to be excluded from the training process.}

\item{recipe}{A user-defined \code{recipe} object for custom preprocessing. If provided, internal recipe steps (imputation, encoding, scaling) are skipped.}

\item{tune_params}{A named list of tuning ranges for each algorithm and engine
pair. Example: \code{list(rand_forest = list(ranger = list(mtry = c(1, 3))))}
will override the defaults for the ranger engine. Default is \code{NULL}.}

\item{engine_params}{A named list of engine-level arguments to pass directly
to the underlying model fitting functions. Use this for fixed settings that
should apply whenever an engine is fitted (for example,
\code{list(royston_parmar = list(rstpm2 = list(link = "PO")))},
\code{list(cox_ph = list(survival = list(ties = "breslow")))}, or
\code{list(rand_forest = list(ranger = list(importance = "impurity")))}).
These arguments are distinct from \code{tune_params}, which define ranges of
hyperparameters to explore during tuning. Default is an empty list.}

\item{metric}{The performance metric to optimize during training. For
classification, options include \code{"accuracy"}, \code{"roc_auc"},
\code{"logloss"}, \code{"brier_score"}, and \code{"ece"} (plus other class metrics).}

\item{class_threshold}{For binary classification, controls how class probabilities
are converted into hard class predictions during holdout evaluation. The default
is \code{0.5} (standard threshold). Numeric values in (0, 1) set a fixed threshold.
Use \code{"auto"} to tune a threshold on training data to maximize F1, or specify
a metric name (e.g., \code{"sens"}, \code{"spec"}, \code{"youden"}) to optimize
for that metric. Use \code{"model"} to keep the model's default predictions.}

\item{algorithm_engines}{A named list specifying the engine to use for each algorithm.}

\item{use_parsnip_defaults}{Logical. If \code{TRUE}, fastml uses parsnip's default
engines instead of fastml's optimized defaults. This provides compatibility with
standard tidymodels behavior. If \code{FALSE} (default), fastml uses its own
curated engine defaults which may differ from parsnip. When engines differ,
a warning is issued unless \code{algorithm_engines} explicitly specifies the engine.
Use \code{print_default_differences()} to see all differences.}

\item{warn_engine_defaults}{Logical. If \code{TRUE} (default), warns when fastml's
default engine differs from parsnip's default. Set to \code{FALSE} to suppress
these warnings. Warnings are only shown once per algorithm per session.}

\item{n_cores}{An integer specifying the number of parallel worker processes used to
evaluate resamples and tuning candidates. Default is \code{1} (sequential). This sizes
the worker pool only; it does not change how many threads an individual engine uses,
which is controlled by \code{engine_threads}.}

\item{engine_threads}{An integer specifying the number of threads passed to engines that
accept a thread count (for example \code{num.threads} for \pkg{ranger},
\code{num_threads} for \pkg{lightgbm}, \code{nthread} for \pkg{xgboost}). Default is
\code{1}. Total CPU demand is approximately \code{n_cores * engine_threads}, so raising
both above the number of available cores will oversubscribe the machine and can be
slower than either alone. Values greater than \code{1} may also make some engines
nondeterministic; a warning is issued where that applies.}

\item{stratify}{Logical indicating whether to use stratified sampling when splitting the data. Only applied to random holdout splitting. Default is \code{TRUE} for classification and \code{FALSE} for regression.}

\item{impute_method}{Method for handling missing values. Options include:
\describe{
  \item{\code{"medianImpute"}}{Impute missing values using median imputation (recipe-based).}
  \item{\code{"knnImpute"}}{Impute missing values using k-nearest neighbors (recipe-based).}
  \item{\code{"bagImpute"}}{Impute missing values using bagging (recipe-based).}
  \item{\code{"remove"}}{Remove rows with missing values from the data (recipe-based).}
  \item{\code{"error"}}{Do not perform imputation; if missing values are detected, stop execution with an error.}
  \item{\code{NULL}}{Equivalent to \code{"error"}. No imputation is performed, and the function will stop if missing values are present.}
}
All imputation occurs inside the recipe so the same trained preprocessing
can be applied at prediction time. Default is \code{"error"}.}

\item{encode_categoricals}{Logical indicating whether to encode categorical variables. Default is \code{TRUE}.}

\item{scaling_methods}{Vector of scaling methods to apply. Default is \code{c("center", "scale")}.}

\item{balance_method}{Method to handle class imbalance. One of \code{"none"},
\code{"upsample"}, or \code{"downsample"}. Applied inside the preprocessing
recipe so each resampling split is balanced independently (requires the
\code{themis} package when enabled). Default is \code{"none"}.}

\item{resamples}{Optional rsample object providing custom resampling splits.
If supplied, \code{resampling_method}, \code{folds}, and \code{repeats} are
ignored.}

\item{summaryFunction}{A custom summary function for model evaluation. Default is \code{NULL}.}

\item{use_default_tuning}{Logical. Tuning only runs when resamples are supplied and
\code{tuning_strategy} is not \code{"none"}. If \code{TRUE} and
\code{tune_params} is \code{NULL}, default grids are used; if
\code{tune_params} is provided, those values override/extend defaults. When
\code{FALSE} and no custom parameters are given, models are fitted once with
default settings. If no resamples are available or \code{tuning_strategy =
"none"}, tuning requests are ignored with a warning. Default is \code{FALSE}.}

\item{tuning_strategy}{A string specifying the tuning strategy. Must be one of
\code{"grid"}, \code{"bayes"}, or \code{"none"}. Default is \code{"grid"}.
If custom \code{tune_params} are provided while \code{tuning_strategy = "none"},
they will be ignored with a warning.}

\item{tuning_iterations}{Number of iterations for Bayesian tuning. Ignored when
\code{tuning_strategy} is not \code{"bayes"}. Validation of this argument only
occurs for the Bayesian strategy. Default is \code{10}.}

\item{tuning_complexity}{Character string specifying a tuning complexity preset
that controls grid density and parameter range width. One of:
\describe{
  \item{\code{"quick"}}{Minimal tuning (2 levels/param, ~32 combinations for 5 params).
    Best for: prototyping, debugging, time-constrained scenarios.}
  \item{\code{"balanced"}}{Standard tuning (3 levels/param, ~243 combinations).
    Best for: most production use cases. This is the default.}
  \item{\code{"thorough"}}{Comprehensive tuning (5 levels/param, ~3,125 combinations).
    Best for: final model selection, publications.}
  \item{\code{"exhaustive"}}{Maximum coverage (7 levels/param, ~16,807 combinations).
    Best for: research, competitions. Consider Bayesian tuning instead.}
}
See \code{\link{print_tuning_presets}} for detailed comparison.
Ignored if \code{grid_levels} is explicitly set.}

\item{grid_levels}{Integer specifying the number of levels per parameter for
grid search. Higher values create denser grids but increase computation time
exponentially (grid size = levels^n_params). Typical values:
\itemize{
  \item 2: Very fast, minimal coverage
  \item 3: Balanced (default via \code{tuning_complexity = "balanced"})
  \item 5: Thorough coverage
  \item 7+: Exhaustive (consider Bayesian tuning instead)
}
If \code{NULL} (default), determined by \code{tuning_complexity}.}

\item{early_stopping}{Logical indicating whether to use early stopping in Bayesian tuning methods (if supported). Default is \code{FALSE}.}

\item{adaptive}{Logical indicating whether to use adaptive/racing methods for tuning. Default is \code{FALSE}.}

\item{learning_curve}{Logical. If TRUE, generate learning curves (performance vs. training size).}

\item{seed}{An integer value specifying the random seed for reproducibility.
fastml also configures parallel backends for deterministic RNG streams when
possible; some external engines (e.g., h2o, spark, keras) may still be
nondeterministic and will emit a warning.}

\item{verbose}{Logical; if TRUE, prints progress messages during the training
and evaluation process.}

\item{eval_times}{Optional numeric vector of evaluation horizons for survival
models. When \code{NULL}, defaults to the median and 75th percentile of the
observed follow-up times (rounded to the dataset's time unit).}

\item{survival_metric_convention}{Character string specifying which survival
metric conventions to follow. `"fastml"` (default) uses fastml's internal
defaults for evaluation horizons and t_max. `"tidymodels"` uses
`eval_times` as the explicit evaluation grid and applies yardstick-style
Brier/IBS normalization; when `eval_times` is `NULL`, time-dependent Brier
metrics are omitted.}

\item{bootstrap_ci}{Logical indicating whether bootstrap confidence intervals
should be computed for performance metrics. Applies to all task types.}

\item{bootstrap_samples}{Integer giving the number of bootstrap resamples to
use when \code{bootstrap_ci = TRUE}. Defaults to 500.}

\item{bootstrap_seed}{Optional seed passed to the bootstrap procedure used to
estimate confidence intervals. When omitted, defaults to `seed` for
reproducible intervals; set to `NULL` to allow random bootstrap draws.}

\item{at_risk_threshold}{Numeric value between 0 and 1 used for survival
metrics to determine the last follow-up time (\eqn{t_{max}}). The maximum
time is set to the largest observed time where at least this proportion of
subjects remain at risk.}

\item{audit_mode}{Logical; if \code{TRUE}, enables runtime auditing of custom
preprocessing hooks and records potentially unsafe behaviour (such as global
environment access or file I/O) while flagging the run as potentially
unsafe.}

\item{multiclass_auc}{For multiclass ROC AUC, the averaging method to use:
`"macro"` (default, tidymodels) or `"macro_weighted"`. Macro weights each
class equally, while macro_weighted weights by class prevalence and can
change model rankings on imbalanced data.}

\item{store_fold_models}{Logical. If \code{TRUE}, stores the models trained
on each cross-validation fold (memory intensive). This enables
\code{\link{explain_stability}} to compute feature importance across folds
and assess explanation stability. Default is \code{FALSE}.}
}
\value{
An object of class \code{fastml} containing the best model, performance metrics, and other information.
}
\description{
Trains and evaluates multiple classification or regression models automatically detecting the task based on the target variable type.
}
\details{
Fast Machine Learning Function

Trains and evaluates multiple classification or regression models. The function automatically
detects the task based on the target variable type and can perform advanced hyperparameter tuning
using various tuning strategies.


Model selection is based exclusively on resampling metrics (cross-validation
or nested CV). The holdout split is reserved for final performance
estimation and is never used to choose the best model, mirroring
\code{tidymodels::last_fit()} semantics.

For multiclass ROC AUC, fastml defaults to macro averaging (tidymodels).
Macro treats each class equally, while macro_weighted weights by class
prevalence and can change model rankings on imbalanced data. Keep the same
setting when comparing runs.

## Tuning: Speed vs Robustness Trade-offs

Hyperparameter tuning involves a fundamental trade-off between computational
cost and the likelihood of finding optimal hyperparameters. fastml provides
presets via \code{tuning_complexity} to make this trade-off explicit:

\tabular{lllll}{
  \strong{Level} \tab \strong{Grid Size*} \tab \strong{Time} \tab \strong{Quality} \tab \strong{Use Case} \cr
  quick \tab ~32 \tab ~1x \tab Low \tab Prototyping, debugging \cr
  balanced \tab ~243 \tab ~10x \tab Medium \tab Most production use \cr
  thorough \tab ~3,125 \tab ~100x \tab High \tab Final models, papers \cr
  exhaustive \tab ~16,807 \tab ~1000x \tab Very High \tab Research, competitions \cr
}
*Grid size shown for 5 tunable parameters (levels^5)

**Recommendations:**
\itemize{
  \item Start with \code{tuning_complexity = "quick"} during development
  \item Use \code{"balanced"} (default) for most production pipelines
  \item Switch to \code{"thorough"} for final model selection
  \item Consider \code{tuning_strategy = "bayes"} instead of exhaustive grid search
  \item Enable \code{adaptive = TRUE} for early stopping of poor configurations
}

Use \code{\link{print_tuning_presets}} to see all presets and
\code{\link{estimate_tuning_time}} to estimate runtime before starting.
}
\section{Factor Level Warning}{

For binary classification, the interpretation of metrics like sensitivity, specificity,
and ROC AUC depends on which factor level is treated as the "positive" class (the event
of interest). The \code{event_class} parameter controls this:
\itemize{
  \item \code{"first"} (default): The first factor level is treated as positive
  \item \code{"second"}: The second factor level is treated as positive
}

\strong{Important:} Recipe preprocessing steps like \code{step_other()} or
\code{step_unknown()} can modify factor levels, potentially changing which level
is "first" or "second". Always verify factor levels after preprocessing.

To ensure consistent behavior, explicitly set factor levels before calling fastml:
\preformatted{
# Ensure "positive" is the second level (event_class = "second")
data$outcome <- factor(data$outcome, levels = c("negative", "positive"))

# Or ensure "positive" is the first level (event_class = "first")
data$outcome <- factor(data$outcome, levels = c("positive", "negative"))
}
}

\examples{
\donttest{
# Example 1: Using the iris dataset for binary classification (excluding 'setosa')
data(iris)
iris <- iris[iris$Species != "setosa", ]  # Binary classification
iris$Species <- factor(iris$Species)

# Define a custom tuning grid for the ranger engine
tune <- list(
  rand_forest = list(
    ranger = list(mtry = c(1, 3))
  )
)

# Train models with custom tuning
model <- fastml(
  data = iris,
  label = "Species",
  algorithms = "rand_forest",
  tune_params = tune,
  use_default_tuning = TRUE
)

# View model summary
summary(model)


  }

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xgboost_survival.R
\name{predict_survival}
\alias{predict_survival}
\alias{predict_survival.fastml_native_survival}
\alias{predict_survival.workflow}
\alias{predict_survival.default}
\title{Predict survival probabilities from a survival model}
\usage{
predict_survival(fit, newdata, times, ...)

\method{predict_survival}{fastml_native_survival}(fit, newdata, times, ...)

\method{predict_survival}{workflow}(fit, newdata, times, ...)

\method{predict_survival}{default}(fit, newdata, times, ...)
}
\arguments{
\item{fit}{A fitted survival model.}

\item{newdata}{A data frame of predictors for which to compute survival curves.}

\item{times}{Numeric vector of evaluation times.}

\item{...}{Additional arguments passed to methods.}
}
\value{
A numeric matrix with one row per observation and one column per time.
}
\description{
Predict survival probabilities from a survival model
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/defaults_registry.R
\docType{data}
\name{.fastml_warned_defaults}
\alias{.fastml_warned_defaults}
\title{Environment for Tracking Warned Defaults}
\format{
An object of class \code{environment} of length 0.
}
\usage{
.fastml_warned_defaults
}
\description{
Internal environment to track which default warnings have been shown
in the current session to avoid duplicate warnings.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/defaults_registry.R
\name{print_default_differences}
\alias{print_default_differences}
\title{Print Default Differences Table}
\usage{
print_default_differences(task = "classification", algorithms = NULL)
}
\arguments{
\item{task}{Task type ("classification", "regression", or "survival").}

\item{algorithms}{Optional character vector of algorithms to check.
If NULL, checks all available algorithms for the task.}
}
\value{
Invisibly returns the differences data frame.
}
\description{
Prints a formatted table showing differences between fastml and parsnip
defaults for the specified task type.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/explain_stability.R
\name{print.fastml_stability}
\alias{print.fastml_stability}
\title{Print method for fastml_stability objects}
\usage{
\method{print}{fastml_stability}(x, top_n = 10, ...)
}
\arguments{
\item{x}{A fastml_stability object from explain_stability().}

\item{top_n}{Integer. Number of top features to display. Default is 10.}

\item{...}{Additional arguments (ignored).}
}
\value{
The input object invisibly.
}
\description{
Print method for fastml_stability objects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.fastml.R
\name{plot.fastml}
\alias{plot.fastml}
\title{Plot Methods for \code{fastml} Objects}
\usage{
\method{plot}{fastml}(
  x,
  algorithm = "best",
  type = c("all", "bar", "roc", "calibration", "residual", "learning_curve"),
  ...
)
}
\arguments{
\item{x}{A \code{fastml} object (output of \code{\link{fastml}()}).}

\item{algorithm}{Character vector specifying which algorithm(s) to include when
generating certain plots (e.g., ROC curves). Defaults to \code{"best"}.}

\item{type}{Character vector indicating which plot(s) to produce. Options are:
\describe{
  \item{\code{"bar"}}{Bar plot of performance metrics across all models/engines.}
  \item{\code{"roc"}}{ROC curve(s) for binary classification models.}
  \item{\code{"calibration"}}{Calibration plot for the best model(s).}
  \item{\code{"residual"}}{Residual diagnostics for the best model.}
  \item{\code{"learning_curve"}}{Learning-curve plot if recorded during training.}
  \item{\code{"all"}}{Produce all available plots.}
}}

\item{...}{Additional arguments (currently unused).}
}
\description{
\code{plot.fastml} produces visual diagnostics for a trained \code{fastml} object.
}
\details{
When \code{type = "all"}, \code{plot.fastml} will produce a bar plot of metrics,
ROC curves (classification), calibration plot, and residual diagnostics (regression).
If you specify a subset of types, only those will be drawn.
}
\examples{
\donttest{
  ## Create a binary classification dataset from iris
  data(iris)
  iris <- iris[iris$Species != "setosa",]
  iris$Species <- factor(iris$Species)

  ## Fit fastml model on binary classification task
  model <- fastml(data = iris, label = "Species", algorithms = c("rand_forest", "svm_rbf"))

  ## 1. Plot all available diagnostics
  plot(model, type = "all")

  ## 2. Bar plot of performance metrics
  plot(model, type = "bar")

  ## 3. ROC curves (only for classification models)
  plot(model, type = "roc")

  ## 4. Calibration plot (requires 'probably' package)
  plot(model, type = "calibration")

  ## 5. ROC curves for specific algorithm(s) only
  plot(model, type = "roc", algorithm = "rand_forest")

  ## 6. Residual diagnostics (only available for regression tasks)
  model <- fastml(data = mtcars, label = "mpg", algorithms = c("linear_reg", "xgboost"))
  plot(model, type = "residual")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/explainer_helpers.R
\name{resolve_positive_class}
\alias{resolve_positive_class}
\title{Resolve the positive class for binary classification}
\usage{
resolve_positive_class(prep, y_levels)
}
\arguments{
\item{prep}{A list from fastml_prepare_explainer_inputs containing:
positive_class, event_class, label_levels}

\item{y_levels}{Character vector of actual levels from the target variable}
}
\value{
Character string of the resolved positive class name
}
\description{
Determines the positive class respecting event_class settings from fastml().
This ensures consistency across all explainer functions.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/explain_ale.R
\name{explain_ale}
\alias{explain_ale}
\title{Compute Accumulated Local Effects (ALE) for a fastml model}
\usage{
explain_ale(object, feature, data = c("train", "test"), ...)
}
\arguments{
\item{object}{A `fastml` object.}

\item{feature}{Character string specifying the feature name.}

\item{data}{Character string specifying which data to use: \code{"train"} (default) or \code{"test"}.}

\item{...}{Additional arguments passed to `iml::FeatureEffect`.}
}
\value{
An `iml` object containing ALE results.
}
\description{
Uses the `iml` package to calculate ALE for the specified feature.
}
\examples{
\dontrun{
data(iris)
iris <- iris[iris$Species != "setosa", ]
iris$Species <- factor(iris$Species)
model <- fastml(data = iris, label = "Species")
explain_ale(model, feature = "Sepal.Length")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/engine_helpers.R
\name{get_engine_names}
\alias{get_engine_names}
\title{Get Engine Names from Model Workflows}
\usage{
get_engine_names(models)
}
\arguments{
\item{models}{A list where each element is a list of model workflows. Each workflow is expected to contain a fitted model that can be processed with \code{tune::extract_fit_parsnip}.}
}
\value{
A list of character vectors. Each vector contains the unique engine names extracted from the corresponding element of \code{models}.
}
\description{
Extracts and returns a list of unique engine names from a list of model workflows.
}
\details{
The function applies \code{tune::extract_fit_parsnip} to each model workflow to extract the fitted model object. It then retrieves the engine name from the model specification (\code{spec$engine}). If the extraction fails, \code{NA_character_} is returned for that workflow. Finally, the function removes any duplicate engine names using \code{unique}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/process_model.R
\name{process_model}
\alias{process_model}
\title{Process and Evaluate a Model Workflow}
\usage{
process_model(
  model_obj,
  model_id,
  task,
  test_data,
  label,
  event_class,
  class_threshold = 0.5,
  start_col = NULL,
  time_col = NULL,
  status_col = NULL,
  engine,
  train_data,
  metric,
  eval_times_user = NULL,
  bootstrap_ci = TRUE,
  bootstrap_samples = 500,
  bootstrap_seed = 1234,
  at_risk_threshold = 0.1,
  survival_metric_convention = "fastml",
  metrics = NULL,
  summaryFunction = NULL,
  precomputed_predictions = NULL,
  multiclass_auc = "macro"
)
}
\arguments{
\item{model_obj}{A fitted model or a tuning result (`tune_results` object).}

\item{model_id}{A character identifier for the model (used in warnings).}

\item{task}{Type of task, either `"classification"`, `"regression"`, or
`"survival"`.}

\item{test_data}{A data frame containing the test data.}

\item{label}{The name of the outcome variable (as a character string).}

\item{event_class}{For binary classification, specifies which class is considered the positive class:
`"first"` or `"second"`.}

\item{class_threshold}{For binary classification, controls how class probabilities
are converted into hard class predictions. The default is `0.5` (standard threshold).
Numeric values in (0, 1) set a fixed threshold. Use `"auto"` to tune a threshold
on training data to maximize F1; use `"model"` to keep the model's default predictions.}

\item{start_col}{Optional string. The name of the column specifying the
start time in counting process (e.g., `(start, stop, event)`) survival
data. Only used when \code{task = "survival"}.}

\item{time_col}{String. The name of the column specifying the event or
censoring time (the "stop" time in counting process data). Only used
when \code{task = "survival"}.}

\item{status_col}{String. The name of the column specifying the event
status (e.g., 0 for censored, 1 for event). Only used when
\code{task = "survival"}.}

\item{engine}{A character string indicating the model engine (e.g., `"xgboost"`, `"randomForest"`). Used
to determine if class probabilities are supported. If `NULL`, probabilities are skipped.}

\item{train_data}{A data frame containing the training data, required to refit finalized workflows.}

\item{metric}{The name of the metric (e.g., `"roc_auc"`, `"accuracy"`, `"rmse"`) used for selecting the best tuning result.}

\item{eval_times_user}{Optional numeric vector of time horizons at which to
evaluate survival Brier scores. When `NULL`, sensible defaults based on the
observed follow-up distribution are used.}

\item{bootstrap_ci}{Logical; if `TRUE`, bootstrap confidence intervals are
estimated for performance metrics.}

\item{bootstrap_samples}{Integer giving the number of bootstrap resamples
used when computing confidence intervals.}

\item{bootstrap_seed}{Optional integer seed applied before bootstrap
resampling to make interval estimates reproducible.}

\item{at_risk_threshold}{Numeric value between 0 and 1 defining the minimum
proportion of subjects required to remain at risk when determining the
maximum follow-up time used in survival metrics.}

\item{survival_metric_convention}{Character string specifying which survival
metric conventions to follow. `"fastml"` (default) uses fastml's internal
defaults for evaluation horizons and t_max. `"tidymodels"` uses
`eval_times_user` as the explicit evaluation grid and applies
yardstick-style Brier/IBS normalization; when `eval_times_user` is `NULL`,
time-dependent Brier metrics are omitted.}

\item{metrics}{Optional yardstick metric set (e.g., `yardstick::metric_set(yardstick::rmse)`) used for computing regression performance.}

\item{summaryFunction}{Optional custom classification metric function passed to
`yardstick::new_class_metric()` and included in holdout evaluation.}

\item{precomputed_predictions}{Optional data frame or nested list of
previously generated predictions (per algorithm/engine) to reuse instead
of re-predicting; primarily used when combining results across engines.}

\item{multiclass_auc}{For multiclass ROC AUC, the averaging method to use:
`"macro"` (default, tidymodels) or `"macro_weighted"`. Macro weights each
class equally, while macro_weighted weights by class prevalence and can
change model rankings on imbalanced data.}
}
\value{
A list with two elements:
\describe{
  \item{performance}{A tibble with computed performance metrics.}
  \item{predictions}{A tibble with predicted values and corresponding truth values, and probabilities (if applicable).}
}
}
\description{
This function processes a fitted model or a tuning result, finalizes the model if tuning was used,
makes predictions on the test set, and computes performance metrics depending on the task type
(classification or regression). It supports binary and multiclass classification, and handles
probabilistic outputs when supported by the modeling engine.
}
\details{
- If the input `model_obj` is a `tune_results` object, the function finalizes the model using the
  best hyperparameters according to the specified `metric`, and refits the model on the full training data.

- For classification tasks, performance metrics include accuracy, kappa, sensitivity, specificity, precision,
  F1-score, and ROC AUC (if probabilities are available).

- For multiclass ROC AUC, the estimator is controlled by `multiclass_auc`.

- For regression tasks, RMSE, R-squared, and MAE are returned.

- For models with missing prediction lengths, a helpful imputation error is thrown to guide data preprocessing.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/process_model_helpers.R
\name{build_survfit_matrix}
\alias{build_survfit_matrix}
\title{Build Survival Matrix from survfit Object}
\usage{
build_survfit_matrix(fit_obj, eval_times, n_obs)
}
\arguments{
\item{fit_obj}{A \code{survfit} object.}

\item{eval_times}{Numeric vector of evaluation times.}

\item{n_obs}{Expected number of observations (rows).}
}
\value{
A matrix (rows=subjects, cols=eval_times) of survival
  probabilities, or \code{NULL} on failure.
}
\description{
Extracts survival probabilities from a \code{survfit} object and aligns
them to a common set of evaluation times, creating a matrix.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/process_model_helpers.R
\name{compute_tau_limit}
\alias{compute_tau_limit}
\title{Compute Tau Limit (t_max)}
\usage{
compute_tau_limit(times, threshold)
}
\arguments{
\item{times}{Numeric vector of survival times.}

\item{threshold}{Minimum proportion of subjects that must remain at risk.}
}
\value{
The computed \eqn{t_{max}} value, or \code{NA_real_} if no valid
  times are provided.
}
\description{
Finds the latest time point \eqn{t_{max}} such that at least a certain proportion
of subjects remain at risk.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/defaults_registry.R
\name{compare_defaults}
\alias{compare_defaults}
\title{Compare fastml and parsnip defaults}
\usage{
compare_defaults(algo, task, fastml_engine, fastml_params = NULL)
}
\arguments{
\item{algo}{Character string specifying the algorithm name.}

\item{task}{Character string specifying the task type.}

\item{fastml_engine}{Character string of the fastml default engine.}

\item{fastml_params}{List of fastml default parameters.}
}
\value{
A list with components:
  \describe{
    \item{engine_differs}{Logical indicating if engines differ.}
    \item{fastml_engine}{The fastml default engine.}
    \item{parsnip_engine}{The parsnip default engine.}
    \item{param_differences}{Named list of parameters that differ.}
  }
}
\description{
Compares the default engine and parameter choices between fastml and parsnip
for a given algorithm.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/explain_lime.R
\name{predict_model.model_fit}
\alias{predict_model.model_fit}
\title{Internal predict_model method for parsnip fits}
\usage{
predict_model.model_fit(x, newdata, type, ...)
}
\description{
Shim for parsnip model objects so that lime's predict_model generic ignores
unused arguments passed via `...`.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/process_model_helpers.R
\name{map_brier_values}
\alias{map_brier_values}
\title{Map Brier Curve Values to Specific Horizons}
\usage{
map_brier_values(curve, eval_times, horizons)
}
\arguments{
\item{curve}{Numeric vector of Brier scores from \code{compute_ibrier}.}

\item{eval_times}{Numeric vector of times corresponding to \code{curve}.}

\item{horizons}{Numeric vector of target time horizons to extract.}
}
\value{
A numeric vector of Brier scores corresponding to \code{horizons}.
}
\description{
Extracts Brier score values from a pre-computed curve at specific
time horizons by finding the closest matching evaluation time.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary.fastml.R
\name{summary.fastml}
\alias{summary.fastml}
\title{Summary Function for fastml (Using yardstick for ROC Curves)}
\usage{
\method{summary}{fastml}(
  object,
  algorithm = "best",
  type = c("all", "metrics", "params", "conf_mat"),
  sort_metric = NULL,
  show_ci = FALSE,
  brier_times = NULL,
  ...
)
}
\arguments{
\item{object}{An object of class \code{fastml}.}

\item{algorithm}{A vector of algorithm names to display summary. Default is \code{"best"}.}

\item{type}{Character vector indicating which outputs to produce.
Options are \code{"all"} (all available outputs),
\code{"metrics"} (performance metrics),
\code{"params"} (best hyperparameters), and
\code{"conf_mat"} (confusion matrix).
Default is \code{"all"}.}

\item{sort_metric}{The metric to sort by. Default uses optimized metric.}

\item{show_ci}{Logical indicating whether to display 95\% confidence intervals
for performance metrics in survival models. Defaults to \code{FALSE}.}

\item{brier_times}{Optional numeric or character vector that selects which
time-specific Brier scores to display for survival models. When \code{NULL}
(the default), time-specific Brier scores are omitted from the summary.}

\item{...}{Additional arguments.}
}
\value{
Prints summary of fastml models.
}
\description{
Summarizes the results of machine learning models trained using the `fastml` package.
Depending on the task type (classification or regression), it provides customized output such as
performance metrics, best hyperparameter settings, and confusion matrices.
It is designed to be informative and readable, helping users quickly interpret model results.
}
\details{
For classification tasks, the summary includes metrics such as Accuracy, F1 Score, Kappa,
Precision, ROC AUC, Sensitivity, and Specificity. A confusion matrix is also provided for the best model(s).
For regression tasks, the summary reports RMSE, R-squared, and MAE.

Users can control the type of output with the `type` argument:
`metrics` displays model performance metrics.
`params` shows the best hyperparameter settings.
`conf_mat` prints confusion matrices (only for classification).
`all` includes all of the above.

If multiple algorithms are trained, the summary highlights the best model based on the optimized metric.
For survival tasks, Harrell's C-index, Uno's C-index, the integrated Brier
score, and (when available) the RMST difference are shown by default. Specific
Brier(t) horizons can be requested through the \code{brier_times} argument.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xgboost_survival.R
\name{predict_risk}
\alias{predict_risk}
\alias{predict_risk.fastml_native_survival}
\alias{predict_risk.workflow}
\alias{predict_risk.default}
\title{Predict Risk Scores from a Survival Model}
\usage{
predict_risk(fit, newdata, ...)

\method{predict_risk}{fastml_native_survival}(fit, newdata, ...)

\method{predict_risk}{workflow}(fit, newdata, ...)

\method{predict_risk}{default}(fit, newdata, ...)
}
\arguments{
\item{fit}{A fitted survival model object.}

\item{newdata}{A data frame containing predictor variables for which to
compute risk scores.}

\item{...}{Additional arguments passed to specific methods.}
}
\value{
A numeric vector of risk scores, where higher values indicate
greater predicted risk.
}
\description{
Provides a consistent interface for computing linear predictors (risk scores)
across various survival modeling engines, including native \pkg{fastml} models
(e.g., Cox proportional hazards, XGBoost Cox) and \pkg{parsnip}/workflow objects.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/process_model_helpers.R
\name{compute_ibrier}
\alias{compute_ibrier}
\title{Compute Integrated Brier Score and Curve}
\usage{
compute_ibrier(
  eval_times,
  surv_mat,
  time_vec,
  status_vec,
  tau,
  censor_eval_fn,
  normalize_by = c("non_missing", "n"),
  include_zero = TRUE
)
}
\arguments{
\item{eval_times}{Numeric vector of evaluation time points.}

\item{surv_mat}{Matrix of predicted survival probabilities
(rows=subjects, cols=eval_times).}

\item{time_vec}{Numeric vector of test times.}

\item{status_vec}{Numeric vector of test statuses.}

\item{tau}{The time horizon \eqn{\tau} for integration.}

\item{censor_eval_fn}{A function (from \code{create_censor_eval}) that
evaluates the censoring survival function \eqn{G(t)}.}

\item{normalize_by}{Character string specifying how to normalize Brier
scores. Use \code{"non_missing"} (default) to divide by the number of
non-missing contributions or \code{"n"} to divide by the total sample size.}

\item{include_zero}{Logical; if \code{TRUE}, integrates from time 0 with a
Brier score of 0 when the evaluation grid starts after 0.}
}
\value{
A list with \code{ibs} (the scalar IBS value) and \code{curve} (a
  numeric vector of Brier scores at \code{eval_times}).
}
\description{
Calculates the Brier score at specified evaluation times and the
Integrated Brier Score (IBS) up to \eqn{\tau}, using IPCW to handle
censoring.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/evaluate_models.R
\name{fastml_compute_holdout_results}
\alias{fastml_compute_holdout_results}
\title{Evaluate Models Function}
\usage{
fastml_compute_holdout_results(
  models,
  train_data,
  test_data,
  label,
  start_col = NULL,
  time_col = NULL,
  status_col = NULL,
  task,
  metric = NULL,
  event_class,
  class_threshold = 0.5,
  eval_times = NULL,
  bootstrap_ci = TRUE,
  bootstrap_samples = 500,
  bootstrap_seed = 1234,
  at_risk_threshold = 0.1,
  survival_metric_convention = "fastml",
  precomputed_predictions = NULL,
  summaryFunction = NULL,
  multiclass_auc = "macro"
)
}
\arguments{
\item{models}{A list of trained model objects.}

\item{train_data}{Preprocessed training data frame.}

\item{test_data}{Preprocessed test data frame.}

\item{label}{Name of the target variable. For survival analysis this should
be a character vector of length two giving the names of the time and status
columns.}

\item{start_col}{Optional string. The name of the column specifying the
start time in counting process (e.g., `(start, stop, event)`) survival
data. Only used when \code{task = "survival"}.}

\item{time_col}{String. The name of the column specifying the event or
censoring time (the "stop" time in counting process data). Only used
when \code{task = "survival"}.}

\item{status_col}{String. The name of the column specifying the event
status (e.g., 0 for censored, 1 for event). Only used when
\code{task = "survival"}.}

\item{task}{Type of task: "classification", "regression", or "survival".}

\item{metric}{The performance metric to optimize (e.g., "accuracy", "rmse").}

\item{event_class}{A single string. Either "first" or "second" to specify which level of truth to consider as the "event".}

\item{class_threshold}{For binary classification, controls how class probabilities
are converted into hard class predictions. The default is `0.5` (standard threshold).
Numeric values in (0, 1) set a fixed threshold. Use `"auto"` to tune a threshold
on training data to maximize F1; use `"model"` to keep the model's default predictions.}

\item{eval_times}{Optional numeric vector of evaluation horizons for survival
metrics. Passed through to \code{process_model}.}

\item{bootstrap_ci}{Logical indicating whether bootstrap confidence intervals
should be computed for the evaluation metrics.}

\item{bootstrap_samples}{Number of bootstrap resamples used when
\code{bootstrap_ci = TRUE}.}

\item{bootstrap_seed}{Optional integer seed for the bootstrap procedure used
in metric estimation.}

\item{at_risk_threshold}{Minimum proportion of subjects that must remain at
risk to define \eqn{t_{max}} when computing survival metrics such as the
integrated Brier score.}

\item{survival_metric_convention}{Character string specifying which survival
metric conventions to follow. `"fastml"` (default) uses fastml's internal
defaults for evaluation horizons and t_max. `"tidymodels"` uses
`eval_times` as the explicit evaluation grid and applies yardstick-style
Brier/IBS normalization; when `eval_times` is `NULL`, time-dependent Brier
metrics are omitted.}

\item{precomputed_predictions}{Optional data frame or nested list of
previously generated predictions (per algorithm/engine) to reuse instead
of recomputing. This is mainly used when combining results across engines.}

\item{summaryFunction}{Optional custom classification metric function passed
through to \code{process_model} for holdout evaluation.}

\item{multiclass_auc}{For multiclass ROC AUC, the averaging method to use:
`"macro"` (default, tidymodels) or `"macro_weighted"`. Macro weights each
class equally, while macro_weighted weights by class prevalence and can
change model rankings on imbalanced data.}
}
\value{
A list with two elements:
  \describe{
    \item{performance}{A named list of performance metric tibbles for each model.}
    \item{predictions}{A named list of data frames with columns including truth, predictions, and probabilities per model.}
  }
}
\description{
Evaluates the trained models on the test data and computes performance metrics.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_selection_helpers.R
\name{get_best_workflows}
\alias{get_best_workflows}
\title{Get Best Workflows}
\usage{
get_best_workflows(models, best_model_name)
}
\arguments{
\item{models}{A nested list of model workflows.}

\item{best_model_name}{A named character vector of chosen best engines.}
}
\description{
Extracts the best workflows from a nested list of model workflows based on the provided best model names.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/defaults_registry.R
\name{reset_default_warnings}
\alias{reset_default_warnings}
\title{Reset Default Override Warnings}
\usage{
reset_default_warnings()
}
\description{
Resets the tracking of which default override warnings have been shown.
Useful for testing or when starting a new analysis session.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_selection_helpers.R
\name{get_best_model_names}
\alias{get_best_model_names}
\title{Get Best Model Names}
\usage{
get_best_model_names(models)
}
\arguments{
\item{models}{A named list where each element corresponds to an algorithm and contains a list of model workflows.
Each workflow should be compatible with \code{tune::extract_fit_parsnip}.}
}
\value{
A named character vector. The names of the vector correspond to the algorithm names, and the values represent the chosen best engine name for that algorithm.
}
\description{
Extracts and returns the best engine names from a named list of model workflows.
}
\details{
For each algorithm, the function extracts the engine names from the model workflows using \code{tune::extract_fit_parsnip}.
  It then chooses \code{"randomForest"} if it is available; otherwise, it selects the first non-\code{NA} engine.
  If no engine names can be extracted for an algorithm, \code{NA_character_} is returned.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/process_model_helpers.R
\name{compute_rmst_difference}
\alias{compute_rmst_difference}
\title{Compute Difference in Restricted Mean Survival Time (RMST)}
\usage{
compute_rmst_difference(
  time_vec,
  status_vec,
  risk_vec,
  tau,
  surv_mat = NULL,
  eval_times_full = NULL,
  model_type = "other"
)
}
\arguments{
\item{time_vec}{Numeric vector of test times.}

\item{status_vec}{Numeric vector of test statuses.}

\item{risk_vec}{Numeric vector of predicted risk scores for test data.}

\item{tau}{The time horizon \eqn{\tau} for integration.}

\item{surv_mat}{Optional. A matrix of individual survival predictions
(rows=subjects, cols=times) used for model-based RMST calculation.}

\item{eval_times_full}{Optional. A numeric vector of time points
corresponding to the columns of \code{surv_mat}.}

\item{model_type}{Optional string (e.log., "rstpm2", "flexsurv") indicating
if a model-based RMST calculation should be attempted.}
}
\value{
The RMST difference (RMST_low - RMST_high), or \code{NA_real_}.
}
\description{
Calculates the difference in RMST between "low" and "high" risk groups up to
a time horizon \eqn{\tau}. Groups are defined by median-splitting the
\code{risk_vec}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/params_helpers.R
\name{get_default_tune_params}
\alias{get_default_tune_params}
\title{Get Default Tuning Parameters}
\usage{
get_default_tune_params(algo, train_data, label, engine)
}
\arguments{
\item{algo}{A character string specifying the algorithm name. Supported values include: \code{"rand_forest"}, \code{"C5_rules"}, \code{"xgboost"}, \code{"lightgbm"}, \code{"logistic_reg"}, \code{"multinom_reg"}, \code{"decision_tree"}, \code{"svm_linear"}, \code{"svm_rbf"}, \code{"nearest_neighbor"}, \code{"naive_Bayes"}, \code{"mlp"}, \code{"deep_learning"}, \code{"discrim_linear"}, \code{"discrim_quad"}, \code{"bag_tree"}, \code{"elastic_net"}, \code{"bayes_glm"}, \code{"pls"}, \code{"linear_reg"}, \code{"ridge_reg"}, and \code{"lasso_reg"}.}

\item{train_data}{A data frame containing the training data.}

\item{label}{A character string specifying the name of the outcome variable in \code{train_data}. This column is excluded when calculating the number of predictors.}

\item{engine}{A character string specifying the engine to be used for the algorithm. Different engines may have different tuning parameter ranges.}
}
\value{
A list of tuning parameter ranges for the specified algorithm. If no tuning parameters are defined for the given algorithm, the function returns \code{NULL}.
}
\description{
Returns a list of default tuning parameter ranges for a specified algorithm based on the provided training data, outcome label, and engine.
}
\details{
The function first determines the number of predictors by removing the outcome variable (specified by \code{label}) from \code{train_data}. It then uses a \code{switch} statement to select a list of default tuning parameter ranges tailored for the specified algorithm and engine. The tuning ranges have been adjusted for efficiency and may include parameters such as \code{mtry}, \code{trees}, \code{min_n}, and others depending on the algorithm.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/defaults_registry.R
\name{validate_defaults_registry}
\alias{validate_defaults_registry}
\title{Validate Defaults Registry Against Parsnip}
\usage{
validate_defaults_registry()
}
\value{
A list of mismatches. Each element is a list with components:
  \describe{
    \item{algorithm}{The algorithm name.}
    \item{fastml_default}{The default engine recorded in fastml's registry.}
    \item{parsnip_default}{The actual default engine from parsnip.}
  }
  Returns an empty list if no mismatches are found.
}
\description{
Compares the hardcoded parsnip default engines in fastml's registry against
the actual defaults reported by \code{parsnip::show_engines()}. Returns a
list of any mismatches found, which may indicate that parsnip has updated
its defaults since fastml's registry was last updated.
}
\details{
This function queries parsnip for model specifications and compares
  against the hardcoded \code{parsnip_defaults} list in
  \code{get_parsnip_default_engine()}. Mismatches may occur when:
  \itemize{
    \item Parsnip updates its default engine for a model type
    \item New engines are added to parsnip that become the new default
    \item fastml's registry has not been updated after a parsnip release
  }

  This validation is intended for package maintenance and testing purposes.
}
\examples{
\dontrun{
mismatches <- validate_defaults_registry()
if (length(mismatches) > 0) {
  message("Found ", length(mismatches), " mismatch(es) with parsnip defaults")
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/process_model_helpers.R
\name{compute_uno_c_index}
\alias{compute_uno_c_index}
\title{Compute Uno's C-index (Time-Dependent AUC)}
\usage{
compute_uno_c_index(
  train_time,
  train_status,
  test_time,
  test_status,
  risk_vec,
  tau,
  censor_eval_fn
)
}
\arguments{
\item{train_time}{Numeric vector of training times (used for censor model).}

\item{train_status}{Numeric vector of training statuses (used for censor model).}

\item{test_time}{Numeric vector of test times.}

\item{test_status}{Numeric vector of test statuses.}

\item{risk_vec}{Numeric vector of predicted risk scores for test data.}

\item{tau}{The time horizon \eqn{\tau} for evaluation. If \code{NA} or
\code{<= 0}, the maximum finite test time is used.}

\item{censor_eval_fn}{A function (from \code{create_censor_eval}) that
evaluates the censoring survival function \eqn{G(t)}.}
}
\value{
The computed Uno's C-index, or \code{NA_real_} on failure.
}
\description{
Calculates Uno's C-index (a time-dependent AUC measure) for survival data,
weighted by the inverse probability of censoring (IPCW).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_selection_helpers.R
\name{get_best_model_idx}
\alias{get_best_model_idx}
\title{Get Best Model Indices by Metric and Group}
\usage{
get_best_model_idx(df, metric, group_cols = c("Model", "Engine"))
}
\arguments{
\item{df}{A data frame containing model performance summaries.}

\item{metric}{A single character string naming the column in `df` used to
identify the best model.}

\item{group_cols}{Character vector of column names used to define model
groups before selecting the best metric value. Defaults to
`c("Model", "Engine")`.}
}
\value{
Integer vector of row indices in `df` corresponding to the best
  model group or groups.
}
\description{
Get Best Model Indices by Metric and Group
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/process_model_helpers.R
\name{assign_risk_group}
\alias{assign_risk_group}
\title{Assign Risk Groups}
\usage{
assign_risk_group(risk_vec)
}
\arguments{
\item{risk_vec}{Numeric vector of predicted risk scores.}
}
\value{
A character vector of "low", "high", or \code{NA}.
}
\description{
Dichotomizes a continuous risk vector into "low" and "high" risk groups
based on the median.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predict.fastml.R
\name{predict.fastml}
\alias{predict.fastml}
\title{Predict method for fastml objects}
\usage{
\method{predict}{fastml}(
  object,
  newdata,
  type = "auto",
  model_name = NULL,
  verbose = FALSE,
  postprocess_fn = NULL,
  eval_time = NULL,
  ...
)
}
\arguments{
\item{object}{A fitted `fastml` object created by the `fastml()` function.}

\item{newdata}{A data frame or tibble containing new predictor data for which to generate predictions.}

\item{type}{Type of prediction to return. One of `"auto"` (default), `"class"`, `"prob"`, `"numeric"`,
`"survival"`, or `"risk"`.
- `"auto"`: chooses `"class"` for classification, `"numeric"` for regression, and `"survival"` for survival.
- `"prob"`: returns class probabilities (only for classification).
- `"class"`: returns predicted class labels.
- `"numeric"`: returns predicted numeric values (for regression).
- `"survival"`: returns survival probabilities at the supplied `eval_time` horizons (for survival tasks).
- `"risk"`: returns risk scores on the linear predictor scale (for survival tasks).}

\item{model_name}{(Optional) Name of a specific model to use when `object$best_model` contains multiple models.}

\item{verbose}{Logical; if `TRUE`, prints progress messages showing which models are used during prediction.}

\item{postprocess_fn}{(Optional) A function to apply to the final predictions (e.g., inverse transforms, thresholding).}

\item{eval_time}{Optional numeric vector of time points (on the original time
scale) at which to return survival probabilities when `type = "survival"`.
Required for survival tasks when requesting survival curves.}

\item{...}{Additional arguments (currently unused).}
}
\value{
A vector of predictions, or a named list of predictions (if multiple models are used).
        If `postprocess_fn` is supplied, its output will be returned instead.
}
\description{
Generates predictions from a trained `fastml` object on new data.
Supports both single-model and multi-model workflows, and handles classification
and regression tasks with optional post-processing and verbosity.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/save.fastml.R
\name{load_model}
\alias{load_model}
\title{Load Model Function}
\usage{
load_model(filepath)
}
\arguments{
\item{filepath}{A string specifying the file path to load the model from.}
}
\value{
An object of class \code{fastml}.
}
\description{
Loads a trained model object from a file.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/defaults_registry.R
\name{defaults_registry}
\alias{defaults_registry}
\title{Defaults Registry for Engine and Parameter Transparency}
\description{
Functions to track, compare, and warn about differences between
  fastml defaults and parsnip defaults, providing users with full transparency
  and control over model configuration.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/survival_utils.R
\name{fastml_normalize_survival_status}
\alias{fastml_normalize_survival_status}
\title{Internal helpers for survival-specific preprocessing}
\usage{
fastml_normalize_survival_status(status_vec, reference_length = NULL)
}
\arguments{
\item{status_vec}{A vector containing survival status information. May be
numeric, logical, factor, or character.}

\item{reference_length}{Optional integer specifying the desired length of the
returned vector. When `status_vec` is `NULL`, this value controls the
length of the output (defaulting to 0 when not supplied).}
}
\value{
A list with two elements: `status`, the recoded numeric vector, and
  `recoded`, a logical flag indicating whether a non-standard encoding was
  detected.
}
\description{
These utilities standardize survival status indicators so that downstream
metrics always receive the conventional coding (0 = censored, 1 = event).
The functions are intentionally unexported and are used across multiple
internal modules.
Normalize survival status coding to 0/1 representation
}
\details{
This helper attempts to coerce a status vector into a numeric format where
0 represents censoring and 1 represents the event indicator. It accepts a
variety of common encodings such as 1/2, logical values, factors, or
character labels. When the supplied values deviate from the canonical
coding, the function records that a recode was performed so callers can
communicate this to the user (once).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/process_model_helpers.R
\name{clamp01}
\alias{clamp01}
\title{Clamp Values to [0, 1]}
\usage{
clamp01(x)
}
\arguments{
\item{x}{A numeric vector.}
}
\value{
The clamped numeric vector.
}
\description{
Truncates a numeric vector so all values lie within the [0, 1] interval.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/train_models.R
\name{train_models}
\alias{train_models}
\title{Train Specified Machine Learning Algorithms on the Training Data}
\usage{
train_models(
  train_data,
  label,
  task,
  algorithms,
  resampling_method,
  folds,
  repeats,
  group_cols = NULL,
  block_col = NULL,
  block_size = NULL,
  initial_window = NULL,
  assess_window = NULL,
  skip = 0,
  outer_folds = NULL,
  resamples = NULL,
  tune_params,
  engine_params = list(),
  metric,
  summaryFunction = NULL,
  seed = 123,
  recipe,
  use_default_tuning = FALSE,
  tuning_strategy = "grid",
  tuning_iterations = 10,
  tuning_complexity = "balanced",
  grid_levels = 3L,
  early_stopping = FALSE,
  adaptive = FALSE,
  algorithm_engines = NULL,
  use_parsnip_defaults = FALSE,
  warn_engine_defaults = TRUE,
  n_cores = 1,
  engine_threads = 1,
  verbose = FALSE,
  event_class = "first",
  class_threshold = 0.5,
  start_col = NULL,
  time_col = NULL,
  status_col = NULL,
  eval_times = NULL,
  at_risk_threshold = 0.1,
  survival_metric_convention = "fastml",
  audit_env = NULL,
  multiclass_auc = "macro",
  store_fold_models = FALSE
)
}
\arguments{
\item{train_data}{Preprocessed training data frame.}

\item{label}{Name of the target variable.}

\item{task}{Type of task: "classification", "regression", or "survival".}

\item{algorithms}{Vector of algorithm names to train.}

\item{resampling_method}{Resampling method for cross-validation. Supported
options include standard \code{"cv"}, \code{"repeatedcv"}, and
\code{"boot"}, as well as grouped resampling via \code{"grouped_cv"},
blocked/rolling schemes via \code{"blocked_cv"} or \code{"rolling_origin"},
nested resampling via \code{"nested_cv"}, and the passthrough
\code{"none"} option.}

\item{folds}{Number of folds for cross-validation.}

\item{repeats}{Number of times to repeat cross-validation (only applicable for methods like "repeatedcv").}

\item{group_cols}{Optional character vector of grouping columns used with
`resampling_method = "grouped_cv"`. For classification problems the outcome
column is used to request grouped stratification where supported; if class
imbalance prevents stratification, grouped folds are still created and a
warning is emitted to document the limitation.}

\item{block_col}{Optional name of the ordering column used with blocked or
rolling resampling.}

\item{block_size}{Optional integer specifying the block size for
`resampling_method = "blocked_cv"`.}

\item{initial_window}{Optional integer specifying the initial window size for
rolling resampling.}

\item{assess_window}{Optional integer specifying the assessment window size
for rolling resampling.}

\item{skip}{Optional integer number of resamples to skip between rolling
resamples.}

\item{outer_folds}{Optional integer specifying the number of outer folds for
`resampling_method = "nested_cv"`.}

\item{resamples}{Optional rsample object. If provided, custom resampling splits
will be used instead of those created internally.}

\item{tune_params}{A named list of tuning ranges. For each algorithm, supply a
list of engine-specific parameter values, e.g.
\code{list(rand_forest = list(ranger = list(mtry = c(1, 3)))).}}

\item{engine_params}{A named list of fixed engine-level arguments passed
directly to the model fitting call for each algorithm/engine combination.
Use this to control options like \code{ties = "breslow"} for Cox models or
\code{importance = "impurity"} for ranger. Unlike \code{tune_params}, these
values are not tuned over a grid.}

\item{metric}{The performance metric to optimize. For classification, options
include \code{"accuracy"}, \code{"roc_auc"}, \code{"logloss"},
\code{"brier_score"}, and \code{"ece"} (plus other class metrics).}

\item{summaryFunction}{A custom summary function for model evaluation. Default is \code{NULL}.}

\item{seed}{An integer value specifying the random seed for reproducibility.}

\item{recipe}{A recipe object for preprocessing.}

\item{use_default_tuning}{Logical; if \code{TRUE} and \code{tune_params} is \code{NULL}, tuning is performed using default grids. Tuning also occurs when custom \code{tune_params} are supplied. When \code{FALSE} and no custom parameters are given, the model is fitted once with default settings.}

\item{tuning_strategy}{A string specifying the tuning strategy. Must be one of
\code{"grid"}, \code{"bayes"}, or \code{"none"}. Adaptive methods may be
used with \code{"grid"}. If \code{"none"} is selected, the workflow is fitted
directly without tuning.
If custom \code{tune_params} are supplied with \code{tuning_strategy = "none"},
they will be ignored with a warning.}

\item{tuning_iterations}{Number of iterations for Bayesian tuning. Ignored
when \code{tuning_strategy} is not \code{"bayes"}; validation occurs only
for the Bayesian strategy.}

\item{tuning_complexity}{Character string specifying tuning complexity preset.
One of "quick", "balanced", "thorough", or "exhaustive". Controls both
grid density and parameter range width.}

\item{grid_levels}{Integer specifying number of levels per parameter for
grid search. Higher values create denser grids but increase computation
exponentially (grid size = levels^n_params).}

\item{early_stopping}{Logical for early stopping in Bayesian tuning.}

\item{adaptive}{Logical indicating whether to use adaptive/racing methods.}

\item{algorithm_engines}{A named list specifying the engine to use for each algorithm.}

\item{use_parsnip_defaults}{Logical. If \code{TRUE}, use parsnip's default engines
instead of fastml's optimized defaults. Default is \code{FALSE}.}

\item{warn_engine_defaults}{Logical. If \code{TRUE} (default), warn when fastml's
default engine differs from parsnip's default.}

\item{n_cores}{Integer number of parallel worker processes used to evaluate
resamples and tuning candidates. Sizes the worker pool only, and decides
whether tuning/resampling runs in parallel.}

\item{engine_threads}{Integer number of threads passed to engines that accept
a thread count (\code{num.threads}, \code{num_threads}, \code{nthread}).
Total CPU demand is approximately \code{n_cores * engine_threads}.}

\item{verbose}{Logical. If \code{TRUE}, print informational messages about
engine selection and parameter overrides.}

\item{event_class}{Character string identifying the positive class when computing
classification metrics ("first" or "second").}

\item{class_threshold}{For binary classification, controls how class probabilities
are converted into hard class predictions during evaluation. The default is `0.5`
(standard threshold). Numeric values in (0, 1) set a fixed threshold. Use `"auto"`
to tune a threshold on training data to maximize F1; use `"model"` to keep the
model's default predictions.}

\item{start_col}{Optional name of the survival start time column passed through
to downstream evaluation helpers.}

\item{time_col}{Optional name of the survival stop time column.}

\item{status_col}{Optional name of the survival status/event column.}

\item{eval_times}{Optional numeric vector of time horizons for survival metrics.}

\item{at_risk_threshold}{Numeric cutoff used to determine the evaluation window
for survival metrics within guarded resampling.}

\item{survival_metric_convention}{Character string specifying which survival
metric conventions to follow. `"fastml"` (default) uses fastml's internal
defaults for evaluation horizons and t_max. `"tidymodels"` uses
`eval_times` as the explicit evaluation grid and applies yardstick-style
Brier/IBS normalization; when `eval_times` is `NULL`, time-dependent Brier
metrics are omitted.}

\item{audit_env}{Internal environment that tracks security audit findings when
custom preprocessing hooks are executed. Typically supplied by
\code{fastml()} and should be left as \code{NULL} when calling
\code{train_models()} directly.}

\item{multiclass_auc}{For multiclass ROC AUC, the averaging method to use:
`"macro"` (default, tidymodels) or `"macro_weighted"`. Macro weights each
class equally, while macro_weighted weights by class prevalence and can
change model rankings on imbalanced data.}

\item{store_fold_models}{Logical. If \code{TRUE}, store the fitted fold models
during resampling for later inspection or stability analysis.}
}
\value{
A list of trained model objects.
}
\description{
Trains specified machine learning algorithms on the preprocessed training data.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/params_helpers.R
\name{get_default_params_with_warnings}
\alias{get_default_params_with_warnings}
\title{Get Default Parameters with Transparency Warnings}
\usage{
get_default_params_with_warnings(
  algo,
  task,
  num_predictors = NULL,
  engine = NULL,
  warn_param_defaults = TRUE,
  verbose = FALSE
)
}
\arguments{
\item{algo}{Algorithm name.}

\item{task}{Task type.}

\item{num_predictors}{Number of predictors (optional).}

\item{engine}{Engine name (optional).}

\item{warn_param_defaults}{Logical; if TRUE, warn about parameter differences.}

\item{verbose}{Logical; if TRUE, print parameter selections.}
}
\value{
A list of default parameters.
}
\description{
Wrapper around \code{get_default_params} that optionally warns when fastml's
default parameters differ from parsnip/engine defaults.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/defaults_registry.R
\name{get_default_differences}
\alias{get_default_differences}
\title{Get All Default Differences Summary}
\usage{
get_default_differences(algorithms, task = "classification")
}
\arguments{
\item{algorithms}{Character vector of algorithm names.}

\item{task}{Task type ("classification", "regression", or "survival").}
}
\value{
A data frame summarizing the differences.
}
\description{
Returns a summary of all differences between fastml and parsnip defaults
for the specified algorithms.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tuning_config.R
\name{print_tuning_presets}
\alias{print_tuning_presets}
\title{Print Tuning Presets Summary}
\usage{
print_tuning_presets()
}
\value{
Invisibly returns a data frame with preset information.
}
\description{
Prints a formatted summary of all available tuning complexity presets
with their characteristics and recommended use cases.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/engine_helpers.R
\name{availableMethods}
\alias{availableMethods}
\title{Get Available Methods}
\usage{
availableMethods(type = c("classification", "regression", "survival"), ...)
}
\arguments{
\item{type}{A character string specifying the type of task. Must be one of
\code{"classification"}, \code{"regression"}, or \code{"survival"}. Defaults to
\code{c("classification", "regression", "survival")} and uses
\code{\link[base]{match.arg}} to select one.}

\item{...}{Additional arguments (currently not used).}
}
\value{
A character vector containing the names of the available algorithms for the specified task type.
}
\description{
Returns a character vector of algorithm names available for classification,
regression or survival tasks.
}
\details{
Depending on the specified \code{type}, the function returns a different set of algorithm names:
\itemize{
  \item For \code{"classification"}, it returns algorithms such as \code{"logistic_reg"}, \code{"multinom_reg"}, \code{"decision_tree"}, \code{"C5_rules"}, \code{"rand_forest"}, \code{"xgboost"}, \code{"lightgbm"}, \code{"svm_linear"}, \code{"svm_rbf"}, \code{"nearest_neighbor"}, \code{"naive_Bayes"}, \code{"mlp"}, \code{"discrim_linear"}, \code{"discrim_quad"}, and \code{"bag_tree"}.
  \item For \code{"regression"}, it returns algorithms such as \code{"linear_reg"}, \code{"ridge_reg"}, \code{"lasso_reg"}, \code{"elastic_net"}, \code{"decision_tree"}, \code{"rand_forest"}, \code{"xgboost"}, \code{"lightgbm"}, \code{"svm_linear"}, \code{"svm_rbf"}, \code{"nearest_neighbor"}, \code{"mlp"}, \code{"pls"}, and \code{"bayes_glm"}.
  \item For \code{"survival"}, it returns algorithms such as \code{"rand_forest"}, \code{"cox_ph"}, \code{"penalized_cox"}, \code{"stratified_cox"}, \code{"time_varying_cox"}, \code{"survreg"}, \code{"royston_parmar"}, \code{"parametric_surv"}, \code{"piecewise_exp"}, and \code{"xgboost"}.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tuning_config.R
\name{recommend_tuning_config}
\alias{recommend_tuning_config}
\title{Recommend Tuning Configuration}
\usage{
recommend_tuning_config(
  n_rows,
  n_predictors,
  n_algorithms = 1,
  max_time_minutes = 30,
  tuning_strategy = "grid"
)
}
\arguments{
\item{n_rows}{Number of rows in training data.}

\item{n_predictors}{Number of predictor variables.}

\item{n_algorithms}{Number of algorithms to tune.}

\item{max_time_minutes}{Maximum acceptable tuning time in minutes.}

\item{tuning_strategy}{Preferred tuning strategy.}
}
\value{
A list with recommended configuration.
}
\description{
Provides recommendations for tuning configuration based on dataset
characteristics and time constraints.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_ice.R
\name{plot_ice}
\alias{plot_ice}
\title{Plot ICE curves for a fastml model}
\usage{
plot_ice(object, features, data = c("train", "test"), target_class = NULL, ...)
}
\arguments{
\item{object}{A `fastml` object.}

\item{features}{Character vector of feature names to plot.}

\item{data}{Character string specifying which data to use: \code{"train"} (default) or \code{"test"}.}

\item{target_class}{For classification, which class probability to plot. If NULL (default),
uses the positive class determined by the model settings. For multiclass problems,
this shows the probability of the specified class vs all others.}

\item{...}{Additional arguments passed to `pdp::partial`.}
}
\value{
A list with two elements: `data` (the ICE data frame) and `plot` (the ggplot object).
}
\description{
Generates Individual Conditional Expectation (ICE) plots for selected features
using the `pdp` package (ggplot2 engine), and returns both the underlying data
and the plot object.
}
\examples{
\dontrun{
data(iris)
iris <- iris[iris$Species != "setosa", ]
iris$Species <- factor(iris$Species)
model <- fastml(data = iris, label = "Species")
plot_ice(model, features = "Sepal.Length")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/process_model_helpers.R
\name{create_censor_eval}
\alias{create_censor_eval}
\title{Create Censoring Distribution Evaluator}
\usage{
create_censor_eval(time_vec, status_vec)
}
\arguments{
\item{time_vec}{Numeric vector of survival/censoring times.}

\item{status_vec}{Numeric vector of event statuses (1=event, 0=censored).}
}
\value{
A function that takes a numeric vector of times and returns the
  estimated censoring survival probabilities \eqn{G(t)} at those times.
}
\description{
Creates a function to evaluate the survival function of the censoring
distribution, \eqn{G(t) = P(C > t)}, using a Kaplan-Meier estimator.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/defaults_registry.R
\name{format_default_override_warning}
\alias{format_default_override_warning}
\title{Format Default Override Warning Message}
\usage{
format_default_override_warning(algo, comparison, show_params = TRUE)
}
\arguments{
\item{algo}{Algorithm name.}

\item{comparison}{Result from compare_defaults().}

\item{show_params}{Logical; whether to include parameter differences.}
}
\value{
Character string with the warning message.
}
\description{
Creates a human-readable warning message about default overrides.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/defaults_registry.R
\name{get_parsnip_default_params}
\alias{get_parsnip_default_params}
\title{Get Parsnip Default Parameters for an Algorithm}
\usage{
get_parsnip_default_params(algo, engine = NULL, task = NULL)
}
\arguments{
\item{algo}{Character string specifying the algorithm name.}

\item{engine}{Character string specifying the engine.}

\item{task}{Character string specifying the task type.}
}
\value{
A named list of default parameter values, or NULL if unknown.
}
\description{
Returns the default parameter values that parsnip would use for a given
algorithm and engine combination.
}
\details{
These defaults are based on parsnip's internal defaults and the
  underlying engine defaults. They may differ from fastml's optimized
  defaults which are tuned for better out-of-box performance.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tuning_config.R
\name{tuning_config}
\alias{tuning_config}
\title{Tuning Configuration and Complexity Presets}
\description{
Functions and presets for configuring hyperparameter tuning grids
  with explicit speed-robustness trade-offs.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/defaults_registry.R
\name{get_parsnip_default_engine}
\alias{get_parsnip_default_engine}
\title{Get Parsnip Default Engine for an Algorithm}
\usage{
get_parsnip_default_engine(algo, task = NULL)
}
\arguments{
\item{algo}{Character string specifying the algorithm name.}

\item{task}{Character string specifying the task type ("classification",
"regression", or "survival").}
}
\value{
Character string of the parsnip default engine, or NULL if unknown.
}
\description{
Returns the default engine that parsnip would use for a given algorithm,
allowing comparison with fastml's optimized defaults.
}
\details{
This function documents parsnip's default engine choices as of
  tidymodels 1.x. These may change with future parsnip versions.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/process_model_helpers.R
\name{compute_survreg_matrix}
\alias{compute_survreg_matrix}
\title{Compute Survival Matrix from survreg Model}
\usage{
compute_survreg_matrix(fit_obj, new_data, eval_times)
}
\arguments{
\item{fit_obj}{A fitted \code{survreg} object.}

\item{new_data}{A data frame with predictor variables.}

\item{eval_times}{Numeric vector of evaluation times.}
}
\value{
A matrix of survival probabilities, or \code{NULL} on failure.
}
\description{
Generates a matrix of survival probabilities (rows=subjects, cols=times)
from a fitted \code{survreg} model for new data.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tuning_config.R
\name{get_expanded_tune_params}
\alias{get_expanded_tune_params}
\title{Expanded Default Tuning Parameters}
\usage{
get_expanded_tune_params(algo, train_data, label, engine)
}
\arguments{
\item{algo}{Algorithm name.}

\item{train_data}{Training data frame.}

\item{label}{Outcome variable name.}

\item{engine}{Engine name.}
}
\value{
A list of expanded tuning parameter ranges.
}
\description{
Returns expanded tuning parameter ranges that provide better coverage
than the minimal defaults. These are used when \code{tuning_complexity}
is set to "thorough" or "exhaustive".
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/save.fastml.R
\name{save.fastml}
\alias{save.fastml}
\title{Save Model Function (Deprecated)}
\usage{
save.fastml(model, filepath)
}
\arguments{
\item{model}{An object of class \code{fastml}.}

\item{filepath}{A string specifying the file path to save the model.}
}
\value{
No return value, called for its side effect of saving the model object to a file.
}
\description{
\code{save.fastml} is deprecated in favour of \code{\link{save_fastml}}.
The old name resembled an S3 method for \code{base::save()}, which is not
a generic, leading to dispatch confusion.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/resampling_guard.R
\name{fastml_guard_validate_indices}
\alias{fastml_guard_validate_indices}
\title{Guarded Resampling Utilities}
\usage{
fastml_guard_validate_indices(indices, label)
}
\arguments{
\item{indices}{Numeric vector of row indices for a resample split.}

\item{label}{Character string used to identify the index source in errors.}
}
\description{
Internal helpers that enforce the Guarded Resampling Principle by
fitting preprocessing pipelines independently within each resampling
split. These functions are not exported.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/surrogate_tree.R
\name{surrogate_tree}
\alias{surrogate_tree}
\title{Fit a surrogate decision tree for a fastml model}
\usage{
surrogate_tree(object, maxdepth = 3, data = c("train", "test"), ...)
}
\arguments{
\item{object}{A `fastml` object.}

\item{maxdepth}{Maximum depth of the surrogate tree. Default 3.}

\item{data}{Character string specifying which data to use: \code{"train"} (default) or \code{"test"}.}

\item{...}{Additional arguments passed to `iml::TreeSurrogate`.}
}
\value{
An `iml::TreeSurrogate` object.
}
\description{
Builds an interpretable tree approximating the behaviour of the underlying
model using the `iml` package.
}
\examples{
\dontrun{
data(iris)
iris <- iris[iris$Species != "setosa", ]
iris$Species <- factor(iris$Species)
model <- fastml(data = iris, label = "Species")
surrogate_tree(model)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tuning_config.R
\name{estimate_tuning_time}
\alias{estimate_tuning_time}
\title{Estimate Tuning Time}
\usage{
estimate_tuning_time(
  n_params,
  n_folds = 10,
  n_rows = 1000,
  complexity = "balanced",
  tuning_strategy = "grid",
  base_fit_time = 1
)
}
\arguments{
\item{n_params}{Number of parameters being tuned.}

\item{n_folds}{Number of cross-validation folds.}

\item{n_rows}{Number of rows in training data.}

\item{complexity}{Tuning complexity level.}

\item{tuning_strategy}{Tuning strategy ("grid" or "bayes").}

\item{base_fit_time}{Estimated time for a single model fit in seconds.}
}
\value{
A list with estimated total time and breakdown.
}
\description{
Provides a rough estimate of tuning time based on the configuration.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fastexplain.R
\name{fastexplain}
\alias{fastexplain}
\title{Explain a fastml model using various techniques}
\usage{
fastexplain(
  object,
  method = "dalex",
  data = c("train", "test"),
  features = NULL,
  n_features = 5,
  variables = NULL,
  observation = NULL,
  grid_size = 20,
  shap_sample = 5,
  vi_iterations = 10,
  seed = 123,
  loss_function = NULL,
  protected = NULL,
  ...
)
}
\arguments{
\item{object}{A \code{fastml} object.}

\item{method}{Character string specifying the explanation method.
Supported values are \code{"dalex"}, \code{"lime"}, \code{"ice"},
\code{"ale"}, \code{"surrogate"}, \code{"interaction"}, \code{"studio"},
\code{"fairness"}, \code{"breakdown"}, and \code{"counterfactual"}.
Defaults to \code{"dalex"}.}

\item{data}{Character string specifying which data to use for explanations:
\code{"train"} (default) uses training data, \code{"test"} uses held-out test data.
Using test data provides explanations that better reflect model generalization,
while training data explanations may be influenced by overfitting.}

\item{features}{Character vector of feature names for partial dependence (model profiles). Default NULL.}

\item{n_features}{Number of features to show in the explanation (used for lime). Default 5.}

\item{variables}{Character vector. Variable names to compute explanations for (used for counterfactuals).}

\item{observation}{A single observation for methods that need a new data point
(\code{method = "lime"}, \code{method = "counterfactual"}, or \code{method = "breakdown"}). Default NULL.}

\item{grid_size}{Number of grid points for partial dependence. Default 20.}

\item{shap_sample}{Integer number of observations from the selected data source to compute SHAP values for. Default 5.}

\item{vi_iterations}{Integer. Number of permutations for variable importance (B). Default 10.}

\item{seed}{Integer. A value specifying the random seed.}

\item{loss_function}{Function. The loss function for \code{model_parts}.
\itemize{
  \item If \code{NULL} and task = 'classification', defaults to \code{DALEX::loss_cross_entropy}.
  \item If \code{NULL} and task = 'regression', defaults to \code{DALEX::loss_root_mean_square}.
}}

\item{protected}{Character or factor vector of protected attribute(s) required for
\code{method = "fairness"}. Default NULL.}

\item{...}{Additional arguments passed to the underlying helper functions
for the chosen \code{method}.}
}
\value{
For DALEX-based methods, prints variable importance, model profiles, and SHAP summaries.
  Other methods return their respective explainer objects (e.g., LIME explanations, ALE plot,
  surrogate tree, interaction strengths, modelStudio dashboard, fairmodels object, breakdown
  object, or counterfactual results), usually invisibly after plotting or printing.
}
\description{
Provides model explainability across several backends. With \code{method = "dalex"} it:
\itemize{
  \item Creates a DALEX explainer from the trained model.
  \item Computes permutation-based variable importance with \code{vi_iterations} permutations and displays the table and plot.
  \item Computes partial dependence-like model profiles when \code{features} are supplied.
  \item Computes Shapley values (SHAP) for \code{shap_sample} training rows, displays the SHAP table,
  and plots a canonical SHAP summary (beeswarm) plot colored by raw feature values and ordered by
  \eqn{\text{mean}(\vert \text{SHAP value} \vert)} per feature. For classification, separate panels per class are shown.
}
}
\details{
\itemize{
   \item \bold{Data source selection:} By default, explanations are computed on training data
   (\code{data = "train"}), which reflects in-sample model behavior and may be influenced by
   overfitting. Set \code{data = "test"} to compute explanations on held-out test data for
   a more realistic assessment of how the model uses features on unseen data.
   \item \bold{Method dispatch:} \code{method} can route to LIME, ICE, ALE, surrogate tree, interaction strengths,
   DALEX/modelStudio dashboards, fairness diagnostics, iBreakDown contributions, or counterfactual search.
   \item \bold{Variable importance controls:} Use \code{vi_iterations} to tune permutation stability and \code{loss_function}
   to override the default DALEX loss (cross-entropy for classification, RMSE for regression).
   \item \bold{Fairness and breakdown support:} Provide \code{protected} for \code{method = "fairness"} and an \code{observation}
   for \code{method = "breakdown"} or \code{method = "counterfactual"}. Observations are aligned to the explainer data before scoring.
 }
}
\note{
By default, explanations use training data. For unbiased feature importance estimates
that better reflect model generalization, use \code{data = "test"} to compute explanations
on held-out test data.
}
